/*---------------------------------------------------------------------------*
  Project:  TWL - append for RTFS
  File:     rtfs_twl_append.c

  Copyright 2008 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Date:: #$
  $Rev$
  $Author$
 *---------------------------------------------------------------------------*/

#include <rtfs.h>

#if (VFAT)

#define FIRST_NAMESEG 0x40
#define NAMESEG_ORDER 0x3F


BOOLEAN pc_findin_ex( DROBJ *pobj, byte *filename, byte *fileext, int action, BOOL get_lfname);
extern byte *pc_seglist2text(DDRIVE * pdrive, SEGDESC *s, byte *lfn);
extern void pc_zeroseglist(SEGDESC *s);
extern void pc_addtoseglist(SEGDESC *s, BLOCKT my_block, int my_index);
extern BOOLEAN pc_patcmp_vfat_8_3(byte *pat, byte *name, BOOLEAN dowildcard);

/* LFNINODE - This is an image of lfn extended names in a subdirectory */
/* Note: lfn file names are null terminated unicode 00, the lfninode is
   padded mod 13 with ffff */
typedef struct lfninode {
                              /* The first LNF has 0x40 + N left */
        byte    lfnorder;     /* 0x45, 0x04, 0x03, 0x02, 0x01 they are stored in
                               reverse order */
        byte    lfname1[10];
        byte    lfnattribute; /* always 0x0F */
        byte    lfnres;       /* reserved */
        byte    lfncksum;   /* All lfninode in one dirent have the same chksum */
        byte    lfname2[12];
        word    lfncluster; /* always 0x0000 */
        byte    lfname3[4];
        } LFNINODE;


/***************************************************************************
    PC_FINDIN -  Find a filename in the same directory as the argument.

 Description
    Look for the next match of filename or pattern filename:ext in the
    subdirectory containing pobj. If found update pobj to contain the
    new information  (essentially getnext.) Called by pc_get_inode().

    Note: Filename and ext must be right filled with spaces to 8 and 3 bytes
            respectively. Null termination does not matter.

    Note the use of the action variable. This is used so we do not have to
    use match patterns in the core code like *.* and ..
    GET_INODE_MATCH   Must match the pattern exactly
    GET_INODE_WILD    Pattern may contain wild cards
    GET_INODE_STAR    Like he passed *.* (pattern will be null)
    GET_INODE_DOTDOT  Like he past .. (pattern will be null

 Returns
    Returns TRUE if found or FALSE.

****************************************************************************/
/* Find filename in the directory containing pobj. If found, load the inode
section of pobj. If the inode is already in the inode buffers we free the current inode
and stitch the existing one in, bumping its open count */
BOOLEAN pc_findin_ex( DROBJ *pobj, byte *filename, byte *fileext, int action, BOOL get_lfname)          /*__fn__*/
{
    BLKBUFF *rbuf;
    DIRBLK *pd;
    DOSINODE *pi;
    FINODE *pfi;
    BOOLEAN matchfound;
    BOOLEAN dowildcard;
    BLKBUFF *scratch;
    byte *lfn;
    byte sfn[26]; /* Leave room in case unicode */
    LFNINODE *lfn_node;
    SEGDESC s;
    byte lastsegorder;
    byte *p;

    RTFS_ARGSUSED_PVOID((void *) fileext);
    rtfs_set_errno(0);  /* Clear it here just in case */
    scratch = pc_scratch_blk();
    if (!scratch)
        return(FALSE);
    lfn = (byte *)scratch->data;

    pc_zeroseglist(&s);

    if (action == GET_INODE_WILD)
        dowildcard = TRUE;
    else
        dowildcard = FALSE;

    /* For convenience. We want to get at block info here   */
    pd = &pobj->blkinfo;

    /* Read the data   */
    pobj->pblkbuff = rbuf = pc_read_blk(pobj->pdrive, pobj->blkinfo.my_block);

    lastsegorder = 0;
    while (rbuf)
    {
        pi = (DOSINODE *) &rbuf->data[0];

        /* Look at the current inode   */
        pi += pd->my_index;

        /* And look for a match   */
        while ( pd->my_index < INOPBLOCK )
        {
            matchfound = FALSE;
            /* End of dir if name is 0   */
            if (!pi->fname[0])
            {
                rtfs_set_errno(PENOENT); /* pc_findin: file or dir not found */
                pc_free_scratch_blk(scratch);
                pc_release_buf(rbuf);
                return(FALSE);
            }

            if (pi->fname[0] == PCDELETE)
            {
                pc_zeroseglist(&s);
            }
            else
            {
                /* Support long file names   */
                if (pi->fattribute == CHICAGO_EXT)
                {
                    lfn_node = (LFNINODE *) pi;
                    if (lfn_node->lfnorder & FIRST_NAMESEG)
                    {
                        pc_addtoseglist(&s, pd->my_block, pd->my_index);
                        /*  .. we could build up the name here too    */
                        lastsegorder = lfn_node->lfnorder;
                        s.ncksum = lfn_node->lfncksum;
                    }
                    else
                    {
                        if (s.nsegs)/* if a chain already exists */
                        {
                            if ( ((lfn_node->lfnorder & NAMESEG_ORDER) == (lastsegorder & NAMESEG_ORDER) - 1) &&
                                  (lfn_node->lfncksum == s.ncksum) )
                            {
                                /* Add new segment to lfn chain   */
                                lastsegorder = lfn_node->lfnorder;
                                pc_addtoseglist(&s, pd->my_block, pd->my_index);
                            }
                            else
                            {
                               /* disconnect chain... segments do not match   */
                                lastsegorder = 0;
                                pc_zeroseglist(&s);
                            }
                        }
                    }
                }
                else
                {
               /* Note: Patcmp wo not match on deleted   */
                    if (s.nsegs)
                    {
                        if (action == GET_INODE_STAR)
                            matchfound = TRUE;
                        else if (action == GET_INODE_DOTDOT)
                            matchfound = FALSE; /* 8.3, not lfn */
                        else
                        {
                            /*---------- twl modified ----------*/
                            if( get_lfname) {
                                p = (byte*)pc_seglist2text(pobj->pdrive, &s,lfn);
                            }else{
                                p = FALSE;
                            }/*---------------------------------*/
                            if (!p)
                                matchfound = FALSE; /* 8.3, not lfn */
                            else
                                matchfound = pc_patcmp_vfat(filename, p ,dowildcard);
                        }
                    }
                    else
                        matchfound = FALSE; /* 8.3, not lfn */

                    if (matchfound)
                        matchfound = (BOOLEAN)(pc_cksum( (byte*) pi ) == s.ncksum);
                    else
                    {
                        if (action == GET_INODE_STAR)
                            matchfound = TRUE;
                        else if (action == GET_INODE_DOTDOT)
                        {
                            if (pi->fname[0] == CS_OP_ASCII('.') && pi->fname[1] == CS_OP_ASCII('.'))
                                matchfound = TRUE; /* 8.3, not lfn */
                        }
                        else
                        {
                            pc_cs_mfile(sfn,pi->fname,pi->fext);
                            matchfound = pc_patcmp_vfat_8_3(filename, sfn, dowildcard);
                        }
                    }
                    if (matchfound && pi->fattribute & AVOLUME &&
                        action != GET_INODE_STAR && action != GET_INODE_WILD)
                    {
                        /* Don't match volume labels if we are finding a specific match. */
                        matchfound = FALSE;
                    }
                    if (matchfound)
                    {
                        /* We found it   */
                        /* See if it already exists in the inode list.
                        If so.. we use the copy from the inode list */
                        pfi = pc_scani(pobj->pdrive, rbuf->blockno, pd->my_index);

                        if (pfi)
                        {
                            pc_freei(pobj->finode);
                            pobj->finode = pfi;
                        }
                        else    /* No inode in the inode list. Copy the data over
                                and mark where it came from */
                        {
                            pfi = pc_alloci();
                            if (pfi)
                            {
                                pc_freei(pobj->finode); /* Release the current */
                                pobj->finode = pfi;
                                pc_dos2inode(pobj->finode , pi );
                                pc_marki(pobj->finode , pobj->pdrive , pd->my_block,
                                        pd->my_index );
                                if (s.nsegs)
                                {
                                    if (pc_cksum( (byte*) pi ) != s.ncksum)
                                    {
                                        pc_zeroseglist(&s);
                                        lastsegorder = 0;
                                    }
                                    pobj->finode->s = s;
                                }
                                else
                                    pc_zeroseglist(&pobj->finode->s);
                            }
                            else
                            {
                                pc_free_scratch_blk(scratch);
                                pc_release_buf(rbuf);
                                return (FALSE);
                            }
                        }
                        /* Free, no error   */
                        pc_free_scratch_blk(scratch);
                        pc_release_buf(rbuf);
                        return (TRUE);
                    }                   /* if (match) */
                    else /* disconnect chain... segments do not match */
                    {
                        pc_zeroseglist(&s);
                    }
                }               /* else (CHICAGO_EXT) */
            }                   /* if (!PCDELETE) */
            pd->my_index++;
            pi++;
        }
        /* Not in that block. Try again   */
        pc_release_buf(rbuf);
        /* Update the objects block pointer   */
        if (!pc_next_block(pobj))
            break;
        pd->my_index = 0;
        pobj->pblkbuff = rbuf = pc_read_blk(pobj->pdrive, pobj->blkinfo.my_block);
    }

    if (!get_errno())
        rtfs_set_errno(PENOENT); /* pc_findin: file or dir not found */
    pc_free_scratch_blk(scratch);
    return (FALSE);
}

#endif /*(VFAT)*/
