/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfs_Error_h

  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfs_Errors.h,v $
  Revision 1.8  2008/07/11 19:14:01  paul
  Updated to match latest wfsError.h

  Revision 1.7  2007/12/04 05:59:11  paul
  bug fix

  Revision 1.6  2007/11/21 22:18:13  paul
  Added WFS_INTERNAL_RESULT_ALREADY_MOUNTED

  Revision 1.5  2007/11/06 22:44:53  paul
  Added macro to do validation checks for NULL pointers passed in

  Revision 1.4  2007/11/02 00:24:24  paul
  started logging changes in the file
 *---------------------------------------------------------------------------*/

#ifndef wfs_Error_h
#define wfs_Error_h

#define WFSReturnOnError(x) { WFSResult nResult = x; if (nResult < WFS_RESULT_OK) return nResult; }
#define WFSReturnOnInternalError(x) { WFSInternalResult nResult = x; if (nResult < WFS_INTERNAL_RESULT_OK) return nResult; }
#define WFSReturnOnNullParameter(p) { if (p == NULL) return WFS_RESULT_INVALID; }

typedef enum {
    WFS_INTERNAL_RESULT_OK                  =    0,  // Success

    // Errors that are usually detected in the client, and would result in the server not being called:
    WFS_INTERNAL_RESULT_BUSY                =   -1,  // The request queue is full.
    WFS_INTERNAL_RESULT_OUT_OF_MEMORY       =   -2,  // The client does not have enough memory to complete the requested operation. (May be due to too many outstanding aysnc calls).
    WFS_INTERNAL_RESULT_INVALID             =   -3,  // The function parameters were invalid, e.g. File name is too long, Path is too long or too deep.
    WFS_INTERNAL_RESULT_ACCESS              =   -4,  // Trying to write to a file which was opened for read-only access, or vice versa.
    WFS_INTERNAL_RESULT_LIB_NOT_INITIALIZED =   -5,  // The WFS client has not been initialized. Please call WFSInit(..) (This error message is omitted from the list of return values).
    WFS_INTERNAL_RESULT_LIB_INITIALIZED     =   -6,  // The WFS client has already been initialized.
    WFS_INTERNAL_RESULT_FILE_TOO_BIG        =   -7,  // The current write/append request would push the file over the size limit.
    WFS_INTERNAL_RESULT_NO_CHANGE_SIZE      =   -8,  // The requested operation would cause the file size to change, but this file has WFS_PERM_NO_CHANGE_SIZE flag set.

    // Errors that cannot usually be detected early, and would be passed via nResult to the Async callbacks:
    WFS_INTERNAL_RESULT_MEDIA_ERROR         =   -9,  // The device or media is not attached, or was removed before the requested operation could be completed.
    WFS_INTERNAL_RESULT_DEV_UNUSABLE        =  -10,  // Attached device is not accessible, e.g. because it is an unsupported device, or has an unknown format.
    WFS_INTERNAL_RESULT_DEV_NOT_INITIALIZED =  -11,  // Device not formatted, or does not yet contain a Wii volume
    WFS_INTERNAL_RESULT_DEV_IN_USE          =  -12,  // Trying to initialize a device which is in use (mounted) by one or more clients.
    WFS_INTERNAL_RESULT_VOL_ID_ERROR        =  -13,  // Attempted to mount a device with the same unique ID as an already mounted device (This could happen if the User backs up an entire Wii partition using a PC)
    WFS_INTERNAL_RESULT_WRITE_PROTECTED     =  -14,  // Device is physically write protected (e.g. USB Flash memory, SD Card).
    WFS_INTERNAL_RESULT_ALREADY_MOUNTED     =  -15,  // The device is already mounted.

    WFS_INTERNAL_RESULT_PERMISSION          =  -16,  // Permission flags do not permit the requested access from this application.
    WFS_INTERNAL_RESULT_PERMISSION_CL       =  -17,  // Permission error caused by control level problems.
    WFS_INTERNAL_RESULT_ACL_FULL            =  -18,  // The access list is full.
    WFS_INTERNAL_RESULT_ACL_ENTRY_NOT_FOUND =  -19,  // The specified entry does not exist in the access list.
    WFS_INTERNAL_RESULT_AUTHENTICATION      =  -20,  // Trying to access content on a different Wii, which does not have the required access rights.
    WFS_INTERNAL_RESULT_CORRUPTION          =  -21,  // A corrupted block was encountered which prevented the operation from being completed.

    WFS_INTERNAL_RESULT_DIRECTORY_QUOTA     =  -22,  // One of the ancestor directories would exceed its quota.

    WFS_INTERNAL_RESULT_MAX_HANDLES         =  -23,  // The maximum number of concurrent file handles / search handles are already in use.
    WFS_INTERNAL_RESULT_ALREADY_EXISTS      =  -24,  // The file or directory being created already exists.
    WFS_INTERNAL_RESULT_NOT_FOUND           =  -25,  // The specified file or directory or device does not exist.
    WFS_INTERNAL_RESULT_NOT_EMPTY           =  -26,  // The specified directory cannot be deleted because it is not empty.
    WFS_INTERNAL_RESULT_NOT_FILE            =  -27,  // The specified path is directory instead of a file.
    WFS_INTERNAL_RESULT_NOT_DIRECTORY       =  -28,  // The specified path is a file instead of a directory
    WFS_INTERNAL_RESULT_FILE_OPEN           =  -39,  // The file is open. Cannot delete, move, rename, change permissions of an open file. Cannot open a file more than once unless all opens are WFS_ACCESS_READ.
    WFS_INTERNAL_RESULT_LOCKED              =  -30,  // Cannot perform the requested operation since the file or directory is locked by an existing operation.
    WFS_INTERNAL_RESULT_RESOURCE_LIMIT_EXCEEDED = -31,  // A resource limit prevents this function call. Try smaller values.

    WFS_INTERNAL_RESULT_DIR_ENTRY_FOUND              =  -40,  // The search string was matched and pointed to an entry
    WFS_INTERNAL_RESULT_DIR_NODE_STRING_PREFIX       =  -41,  // The search string ended early during comparison with a node sub string. (So this name is a prefix of existing names)
    WFS_INTERNAL_RESULT_DIR_CHOICE_PREFIX            =  -42,  // The search string ended at the end of a node sub string, but the node did not have a termination choice. (This name is a prefix of existing names)
    WFS_INTERNAL_RESULT_DIR_NODE_STRING_MISMATCH     =  -43,  // A mismatch was encountered while comparing the search string to a sub node prefix string.
    WFS_INTERNAL_RESULT_DIR_NODE_CHOICE_NOT_FOUND    =  -44,  // A mismatch was encountered while comparing a character of the search string with a set of choices for that character.
    WFS_INTERNAL_RESULT_DIR_BLK_FULL                 =  -45,  // Unable to insert the specified string to the radix tree, due to lack of space in the block.

    // These result codes are not errors, but normal return values for the path cache functions
    WFS_INTERNAL_RESULT_SRV_END_OF_PATH              =  -60,  // Last part of path name parsed
    WFS_INTERNAL_RESULT_SRV_PATH_DEPTH_1             =  -61,  // Path was of depth 1, such as "/something"
    WFS_INTERNAL_RESULT_SRV_PATH_DEV                 =  -62,  // Path was "/dev"
    WFS_INTERNAL_RESULT_SRV_PATH_VOL                 =  -63,  // Path was "/vol"
    WFS_INTERNAL_RESULT_SRV_PATH_VOL_ROOT            =  -64,  // Path was a volume root path, such as "/vol/xxxxxxx/"

    WFS_INTERNAL_RESULT_PTREE_ENTRY_FOUND            =  -70,
    WFS_INTERNAL_RESULT_PTREE_ENTRY_NOT_FOUND        =  -71,
    WFS_INTERNAL_RESULT_PTREE_FULL                   =  -72,

    WFS_INTERNAL_RESULT_DEVICE_ERROR                 = -200,
    WFS_INTERNAL_RESULT_DEVICE_INVALID_PARAMETER     = -201,
    WFS_INTERNAL_RESULT_DEVICE_NOT_FOUND      = -202,

    WFS_INTERNAL_RESULT_BCACHE_ERROR                 = -300,  // unspecified error
    WFS_INTERNAL_RESULT_BCACHE_RESOURCE_LIMIT        = -301,  // exhausted resources
    WFS_INTERNAL_RESULT_BCACHE_INVALID_PARAMETER     = -302,  // at least one argument is invalid
    WFS_INTERNAL_RESULT_BCACHE_NO_MEMORY             = -303,  // could not malloc
    WFS_INTERNAL_RESULT_BCACHE_NOT_FOUND             = -304,  // sector is not cached
    WFS_INTERNAL_RESULT_BCACHE_MAX_DEVICES           = -305,  // max devices alreay allocated
    WFS_INTERNAL_RESULT_BCACHE_INVALID_DEVICE        = -306,  // device not found/open
    WFS_INTERNAL_RESULT_BCACHE_INVALID_HANDLE        = -307,  // invalid device handle
    WFS_INTERNAL_RESULT_BCACHE_INVALID_VOLUME        = -308,  // invalid volume ID (not mapped)
    WFS_INTERNAL_RESULT_BCACHE_ALREADY_MAPPED        = -309,  // vol or devH already mapped
    WFS_INTERNAL_RESULT_BCACHE_ALLOC                 = -310,  // cannot alloc (no mem, or pmem error)
    WFS_INTERNAL_RESULT_BCACHE_PMEM                  = -311,  // cannot read or write to pmem device

    WFS_INTERNAL_RESULT_VOLUME_ERROR                 = -400,  // unspecified volume error
    WFS_INTERNAL_RESULT_VOLUME_INVALID_PARAMETER     = -401,  // at least one argument is invalid
    WFS_INTERNAL_RESULT_VOLUME_BCACHE_ALLOC          = -402,  // cannot allocate a page/block in memory
    WFS_INTERNAL_RESULT_VOLUME_BCACHE_CONFIG         = -403,  // block cache not configured correctly
    WFS_INTERNAL_RESULT_VOLUME_ID                    = -404,  // bad volume ID or not initialized
    WFS_INTERNAL_RESULT_VOLUME_CORRUPT_MR            = -405,  // the MR (or its free lists) are corrupted
    WFS_INTERNAL_RESULT_VOLUME_AREA_ALLOC            = -406,  // area allocation cannot be satisfied
    WFS_INTERNAL_RESULT_VOLUME_DEVICE_PARAMETER      = -407,  // parameters incompatible with device
    WFS_INTERNAL_RESULT_VOLUME_NOT_MAPPED            = -408,  // volume is not mapped; invalid volume ID

    WFS_INTERNAL_RESULT_TRANSACTION_ERROR            = -500,  // unspecified transaction error
    WFS_INTERNAL_RESULT_TRANSACTION_INVALID_PARAMETER= -501,

    WFS_INTERNAL_RESULT_ACL_ERROR                    = -600,  // unspecified error
    WFS_INTERNAL_RESULT_ACL_INVALID_PARAMETER        = -601,  // at least one argument is invalid
    WFS_INTERNAL_RESULT_ACL_MAX_ENTRIES              = -602,  // max entries already in the access list.
    WFS_INTERNAL_RESULT_ACL_BUFFER_TOO_SMALL         = -603,  // buffer is too small to store an access list.
    WFS_INTERNAL_RESULT_ACL_CACHE                    = -610,  // cannot allocate a cache.
    WFS_INTERNAL_RESULT_ACL_FILE                     = -620,  // cannot create, modify or delete the access list file.
    WFS_INTERNAL_RESULT_ACL_ACLDIR_INCONSISTENT      = -621,  // access list files and indexlist are inconsistent.
    WFS_INTERNAL_RESULT_ACL_FILENAME                 = -630,  // cannot convert acl filename <=> acl content.
    WFS_INTERNAL_RESULT_ACL_NAMEDIR_INCONSISTENT     = -631,  // access name files and indexlist are inconsistent.
    WFS_INTERNAL_RESULT_ACL_HANDLE                   = -640,  // cannot allocate an access list handle.

    WFS_INTERNAL_RESULT_AREA_ERROR                   = -700,  // unspecified area error
    WFS_INTERNAL_RESULT_AREA_INVALID_PARAMETER       = -701,
    WFS_INTERNAL_RESULT_AREA_BCACHE_ALLOC            = -702,  // cannot allocate a page/block in memory
    WFS_INTERNAL_RESULT_AREA_CORRUPT_HDR             = -703,  // the area header is corrupted
    WFS_INTERNAL_RESULT_AREA_VOLUME_NOT_MAPPED       = -704,  // volume is not mapped; invalid volume ID
    WFS_INTERNAL_RESULT_AREA_ALLOC                   = -705,  // area allocation cannot be satisfied

    WFS_INTERNAL_RESULT_NOT_IMPLEMENTED              = -1026,  // The requested function has not yet been implemented  //ToDo: This error should be removed from the final API
    WFS_INTERNAL_RESULT_UNKNOWN                      = -1027,  // Unexpected error - could be caused by a bug in the file system.
    WFS_INTERNAL_RESULT_FATAL_ERROR                  = -1028   // May be caused by critical metadata corruption.
} WFSInternalResult;

#define WFSSetLastError(nErr) (wkg.nLastErr = (nErr))

#endif
