/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfs_Mutex.h - mutual exclusion for multi-thread case

  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfs_Mutex.h,v $
  Revision 1.11  2008/04/19 05:52:30  kondo_masahiro
  Fixed code to accept RM compiler

  Revision 1.10  2007/12/27 01:43:38  ueno
  Added WFSIsThreadTerminated() and WFSJoinThread().

  Revision 1.9  2007/12/25 13:59:47  ueno
  Added WFSSuspendThread().

  Revision 1.8  2007/11/21 04:17:28  wayne.wong
  Many changes that included single request/result RPC messages, bug fixes, code factoring, and English C comment adjustments.
    1. Combined RPC messages into a single request message and single result
        message. Except for file reads and writes.
    2. Fixed+factored some Endianess code.
    3. Enabled larger TCP/IP data transfer size. Previous setting was artifically small
        to test the code to perform large transfers.
    4. Changed the volume id parameters utf8*->WFSVolumeId for SrvGetVolumeId
        and SrvMountDevice. More consistent with the way we send this type at the
        RPC layer.
    5. Changed a couple Japanese comments to be compatible with English
        comments. In particular, the "//" comment with a backslash  at the end of
        the line causes the next line of code to be ignored.
    6. Removed Windows-specific debugging output defines for RVL build.

  Revision 1.7  2007/11/02 00:23:33  paul
  started logging changes in the file
 *---------------------------------------------------------------------------*/

#ifndef wfs_Mutex_h
#define wfs_Mutex_h

#include "wfs_Defs.h"

#ifdef _DEBUG
    #define _DEBUG_MUTEX 0  // change to 1 if debug info required
#else
    #define _DEBUG_MUTEX 0
#endif

#if _WIN32

typedef struct {
    HANDLE hHandle;
} WFSMutex;

typedef struct {
    HANDLE hAutoResetEvent;
    //HANDLE hManualResetEvent;
} WFSCond;

typedef struct {
    HANDLE hThread;
} WFSThread;

void WFSInitMutex(WFSMutex *mutex);
void WFSLockMutex(WFSMutex *mutex);
bool WFSTryLockMutex(WFSMutex *mutex);
void WFSUnlockMutex(WFSMutex *mutex);

void WFSInitCond(WFSCond *pCond);
void WFSSignalCond(WFSCond *pCond);
void WFSWaitCond(WFSCond *pCond, WFSMutex* pMx);
void WFSResetCond(WFSCond *pCond);

typedef LPTHREAD_START_ROUTINE WFSThreadFunc;

bool WFSCreateThread(WFSThread *thread,
                     WFSThreadFunc threadFunc,
                     void*      param,
                     void*      stackBase,
                     u32        stackSize,
                     OSPriority priority);

s32 WFSResumeThread(WFSThread *pThread);
void WFSCancelThread(WFSThread *pThread);
void WFSExitThread(void *pVal);
s32 WFSSuspendThread(WFSThread *pThread);
bool WFSIsThreadTerminated(WFSThread *thread);
bool WFSJoinThread(WFSThread *thread, void** val);

#elif _TWL

#define WFSMutex OSMutex
#define WFSCond OSEvent
#define WFSThread OSThread

#define WFSInitMutex osInitMutex
#define WFSLockMutex osLockMutex
#define WFSTryLockMutex osTryLockMutex
#define WFSUnlockMutex osUnlockMutex

//#define WFSInitCond osInitCond
//#define WFSSignalCond osSignalCond
//#define WFSWaitCond osWaitCond
//#define WFSResetCond(x) // This function does nothing on RVL, only needed for Windows

typedef void *(*WFSThreadFunc)(void *);

bool WFSCreateThread(WFSThread *thread,
                     WFSThreadFunc threadFunc,
                     void*      param,
                     void*      stackBase,
                     u32        stackSize,
                     OSPriority priority);

s32 WFSResumeThread(WFSThread *pThread);
void WFSCancelThread(WFSThread *pThread);
void WFSExitThread(void *pVal);
s32 WFSSuspendThread(WFSThread *pThread);
bool WFSIsThreadTerminated(WFSThread *thread);
bool WFSJoinThread(WFSThread *thread, void** val);

void WFSInitCond(WFSCond *pCond);
void WFSSignalCond(WFSCond *pCond);
void WFSWaitCond(WFSCond *pCond, WFSMutex* pMx);
void WFSResetCond(WFSCond *pCond);
/*
#define WFSResumeThread OSResumeThread
#define WFSCancelThread OSCancelThread
#define WFSExitThread OSExitThread
#define WFSSuspendThread OSSuspendThread
#define WFSIsThreadTerminated OSIsThreadTerminated
#define WFSJoinThread OSJoinThread
*/
#elif _RVL

#define WFSMutex OSMutex
#define WFSCond OSCond
#define WFSThread OSThread

#define WFSInitMutex OSInitMutex
#define WFSLockMutex OSLockMutex
#define WFSTryLockMutex OSTryLockMutex
#define WFSUnlockMutex OSUnlockMutex

#define WFSInitCond OSInitCond
#define WFSSignalCond OSSignalCond
#define WFSWaitCond OSWaitCond
#define WFSResetCond(x) // This function does nothing on RVL, only needed for Windows

typedef void *(*WFSThreadFunc)(void *);

bool WFSCreateThread(WFSThread *thread,
                     WFSThreadFunc threadFunc,
                     void*      param,
                     void*      stackBase,
                     u32        stackSize,
                     OSPriority priority);

#define WFSResumeThread OSResumeThread
#define WFSCancelThread OSCancelThread
#define WFSExitThread OSExitThread
#define WFSSuspendThread OSSuspendThread
#define WFSIsThreadTerminated OSIsThreadTerminated
#define WFSJoinThread OSJoinThread

#elif _IOP

// dummy
typedef void *WFSMutex;
typedef void *WFSThread;
typedef void *WFSCond;

#endif

#endif //define wfs_Mutex_h
