/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfs_Handles.h - Module for file handles and search directory handles


  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfscli_Handles.h,v $
  Revision 1.10  2008/10/07 06:42:05  ueno
  Modified WFSCliFileInfoAlloc to set WFSCLI_HANDLE_ALLOCATED in order to indicate the handle is allocated.

  Revision 1.9  2008/07/09 02:51:29  paul
  Fixed comments

  Revision 1.8  2008/05/12 20:30:41  paul
  Updated handle defs

  Revision 1.7  2007/11/02 00:19:58  paul
  started logging changes in the file

*---------------------------------------------------------------------------*/

// File created by Paul Donnelly on 2007.10.06


#ifndef wfs_Handles_h
#define wfs_Handles_h

#ifdef _DEBUG
    #define _DEBUG_WFS_HANDLES 1
#else
    #define _DEBUG_WFS_HANDLES 0
#endif

#include "wfs_Defs.h"
#include "wfs_Mutex.h"

#define WFSCLI_MAX_FILE_HANDLES               (1<<WFSCLI_FILE_HANDLE_IDX_BITS)
#define WFSCLI_MAX_SEARCH_DIR_HANDLES         (1<<WFSCLI_SEARCH_DIR_HANDLE_IDX_BITS)

/* Adding WFS_MAX_OUTSTANDING_ASYNC_CALLS so that we guarantee a handle is always available on the client side
#define WFSCLI_MAX_FILE_HANDLES               (16 + WFS_MAX_OUTSTANDING_ASYNC_CALLS)
#define WFSCLI_MAX_SEARCH_DIR_HANDLES         (4  + WFS_MAX_OUTSTANDING_ASYNC_CALLS)

#define WFSCLI_FILE_HANDLE_IDX_BITS           WFS_NUM_BITS_REQUIRED(WFSCLI_MAX_FILE_HANDLES)
#define WFSCLI_SEARCH_DIR_HANDLE_IDX_BITS     WFS_NUM_BITS_REQUIRED(WFSCLI_MAX_SEARCH_DIR_HANDLES)*/

#define WFSCLI_FILE_HANDLE_INC                (1<<(16+WFSCLI_FILE_HANDLE_IDX_BITS))
#define WFSCLI_SEARCH_DIR_HANDLE_INC          (1<<(16+WFSCLI_SEARCH_DIR_HANDLE_IDX_BITS))

#define WFSCLI_HANDLE_MASK                    0xFFFF0000
#define WFSCLI_HANDLE_COUNTER_MASK            0x1FFF0000
#define WFSCLI_FILE_HANDLE_TT_FIELD           0x00000000
#define WFSCLI_SEARCH_DIR_HANDLE_TT_FIELD     0x20000000
#define WFSCLI_HANDLE_ALLOCATED               0x0000FFFF // set WFSCliFileInfo.nNext to represent this handle is in use.


// Internal binary format of WFSFileHandle (u32)
//
// 0TTcccccccCCCCCCssssssssssSSSSSS
//
//  0 = always 0                          (Number of bits = 1                           )
//  T = type of handle: 00 for FileHandle (Number of bits = 2                           )
//  c = incrementally assigned by client  (Number of bits = 13 - WFSCLI_FILE_HANDLE_IDX_BITS)
//  C = idx of aCliFileInfo array         (Number of bits =      WFSCLI_FILE_HANDLE_IDX_BITS)
//  s = incrementally assigned by server  (Number of bits = 16 - WFSCLI_FILE_HANDLE_IDX_BITS)
//  S = idx of aSrvFileInfo array         (Number of bits =      WFSCLI_FILE_HANDLE_IDX_BITS)

typedef struct {
    u32 nNext;
    WFSMutex mxAccessCliFileInfo;
    WFSFileHandle nHandle;
    WFSFileAttributes fa;
    WFSFileSize nPosition;
    WFSFileSize nPreAllocateSize;
    u8  nAccess;
} WFSCliFileInfo;


// Internal binary format of WFSSearcDirHandle (u32)
//
// 0TTcccccccCCCCCCssssssssssSSSSSS
//
//  0 = always 0                               (Number of bits = 1                                 )
//  T = type of handle: 01 for SearchDirHandle (Number of bits = 2                                 )
//  c = incrementally assigned by client       (Number of bits = 13 - WFSCLI_SEARCH_DIR_HANDLE_IDX_BITS)
//  C = idx of aCliSearchDirInfo array         (Number of bits =      WFSCLI_SEARCH_DIR_HANDLE_IDX_BITS)
//  s = incrementally assigned by server       (Number of bits = 16 - WFSCLI_SEARCH_DIR_HANDLE_IDX_BITS)
//  S = idx of aSrvSearchDirInfo array         (Number of bits =      WFSCLI_SEARCH_DIR_HANDLE_IDX_BITS)

typedef struct {
    u32 nNext;
    WFSSearchDirectoryHandle nHandle;
} WFSCliSearchDirInfo;


typedef struct {
    WFSFileHandle nFileHandleCounter;  // Counter for cccc field of file handle.
    WFSSearchDirectoryHandle nSearchDirHandleCounter;  // Counter for cccc field of search directory handle.
    u32 nFirstFreeFileHandle;  // Index of the first free element of aClientFileInfo
    u32 nFirstFreeSearchDirHandle;  // Index of the first free element of aClientSearchDirInfo
    u32 nNumUsedFileHandles;
    u32 nNumUsedSearchDirHandles;
    //u32 nFirstUsedFileHandle;  // Index of the first used element of aClientFileInfo
    //u32 nFirstUsedSearchDirHandle;  // Index of the first used element of aClientSearchDirInfo
    WFSMutex mxAccessTheFileInfo;
    WFSMutex mxAccessTheSearchDirInfo;
    WFSCliFileInfo aCliFileInfo[WFSCLI_MAX_FILE_HANDLES];
    WFSCliSearchDirInfo aCliSearchDirInfo[WFSCLI_MAX_SEARCH_DIR_HANDLES];
} WFSCliHandles;


void      WFSCliInitHandles();

WFSResult WFSGetCliFileInfo(WFSCliFileInfo **ppFi, WFSFileHandle fh);
WFSResult WFSGetCliSearchDirInfo(WFSCliSearchDirInfo **ppSdi, WFSSearchDirectoryHandle sdh);

WFSResult WFSCliFileInfoAlloc(WFSCliFileInfo **ppFi);
WFSResult WFSCliSearchDirInfoAlloc(WFSCliSearchDirInfo **ppSdi);

void      WFSCliFileInfoFree(WFSCliFileInfo *pFi);
void      WFSCliSearchDirInfoFree(WFSCliSearchDirInfo *pSdi);

#endif //define wfs_Handles_h
