/*---------------------------------------------------------------------------*
  Project:  Revolution WFS library
  File:     wfstypes.h

  Copyright 2008 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfstypes.h,v $
  Revision 1.13  2008/12/10 06:55:11  kondo_masahiro
  Added the new device types WFS_DEVTYPE_USB_MSC_??.

  Revision 1.12  2008/12/03 00:13:29  kondo_masahiro
  Added WFS_MAX_WORK_SPACE_SIZE and fixed WFS_MIN_WORK_SPACE_SIZE

  Revision 1.11  2008/10/23 11:16:20  nakanose_jin
  Open and Search should give a invalid file handle if he fails to execute.

  Revision 1.10  2008/09/26 09:32:36  kondo_masahiro
  Marged with revision 1.8

  Revision 1.9  2008/09/26 09:29:54  kondo_masahiro
  Changed WFS_MIN_WORK_SPACE_SIZE.

  Revision 1.8  2008/07/30 03:57:46  paul
  Fixed WFS_FLAG_CONTIGUOUS_AREA comment

  Revision 1.7  2008/07/25 09:46:56  ooizumi
  Fixed WFSAccess definition.

  Revision 1.6  2008/07/11 09:46:03  ueno
  Added async functions.

  Revision 1.5  2008/06/20 18:35:39  paul
  Added WFS_MIN_DIRECTORY_QUOTA, WFS_FLAG_CONTIGUOUS_AREA

  Revision 1.4  2008/04/29 01:14:52  ueno
  Removed WFS_DEBUG_SYSTEM_PROCESS_ID.
  Added WFS_ROOT_ID and WFS_EVERYONE_ID.

  Revision 1.3  2008/04/08 18:24:30  paul
  Moved callbacks and other type definitions from wfs.h into this header file. Moved permissions flags to wfsPerm.h

  Revision 1.2  2008/02/25 05:35:41  paul
  Reduced WFS_MAX_FILE_NAME_SIZE and WFS_MAX_PATH_NAME_SIZE to make structs multiples of 32 bytes, and also allow space for a NULL terminator so we dont have to change existing code.

  Revision 1.1  2008/01/17 07:51:43  ooizumi
  Moved several definitions from wfs.h.


*---------------------------------------------------------------------------*/

#ifndef __WFSTYPES_H__
#define __WFSTYPES_H__


// ---------------------------------------------------------------------
// wfstypes.h - Wii File System (WFS) API
// ---------------------------------------------------------------------

#include <revolution.h>

#ifdef __cplusplus
extern "C" {
#endif


// ---------------------------------------------------------------------
// WFS attributes
// ---------------------------------------------------------------------
#define WFS_MAX_FILE_SIZE         4294967295  // ((u32)-1)
#define WFS_MAX_PATH_DEPTH        31   // (ToDo: Directories must keep track of their deepest sub-path, so we can validate directory move requests)

// The following represent the maximum size of string fields in Bytes, not including the terminating NULL character.
#define WFS_MAX_FILE_NAME_SIZE    254  // Does not include the path, just the file name.
#define WFS_MAX_PATH_NAME_SIZE    509  // (ToDo: Directories must keep track of their longest sub-path, so we can validate directory move/rename requests)
#define WFS_MAX_DEVICE_NAME_SIZE  15   // maximum size of a device name
#define WFS_MAX_VOLUME_ID_SIZE    7    // maximum size of WFS Volume Id string (This is a randomly assigned string of characters, e.g., 1a2b3c4)

#define WFS_MAX_WORK_SPACE_SIZE   ((512+40+40) * 1024)  // Used for user data cache, shim layer queueing parameters and async thread stack / queueing async parameters.
#define WFS_MIN_WORK_SPACE_SIZE   ((144+40+40) * 1024)
#define WFS_MIN_DIRECTORY_QUOTA   (1024 * 1024)
#define WFS_INVALID_HANDLE_VALUE  ((u32)-1)


typedef char utf8;           // ToDo: Perhaps utf8 should be changed to WFSChar?
typedef u32  WFSFileTime;    // Time in seconds since 00:00, 1st January, 2000
typedef u32  WFSFileSize;    // Size of file
typedef u32  WFSContentIdx;  // Used to identify files which belong to the same unit of purchased content, and control access to them
typedef u8   WFSBool;


// ---------------------------------------------------------------------
// WFS devices
// ---------------------------------------------------------------------

// Possible WFS device states:
//  1. Valid - valid WFS volume exists
//  2. Unusable - bad device
//  3. Unknown format (e.g., FAT12, NTFS, HFS)
//  4. Uninitialized - recognized format (FAT32), but not enough free space for a WFS volume

// Whenever a compatible device is attached, we check to see if a WFS volume already exists. If not,
// we try to create it automatically. If it is not created, then the device must not have enough space.

#define WFS_DEVFLAG_UNUSABLE        1       // Device is unusable (e.g. unknown format, unsupported device)
#define WFS_DEVFLAG_WRITE_PROTECTED 2       // Device is physically write protected
#define WFS_DEVFLAG_NOT_INITIALIZED 4       // Device not formatted or does not yet contain a WFS volume
#define WFS_DEVFLAG_NOT_MOUNTED     8       // Device has not yet been mounted

// ToDo: Can we force this enum to be u8?
typedef enum {
    WFS_DEVTYPE_NONE       = 0x00,
    WFS_DEVTYPE_SD         = 0x01,
    WFS_DEVTYPE_USB_MSC_10 = 0x02, // MSC device with USB1.0
    WFS_DEVTYPE_USB_MSC_11 = 0x03, // MSC device with USB1.1
    WFS_DEVTYPE_USB_MSC_20 = 0x04, // MSC device with USB2.0
    WFS_DEVTYPE_EXT_SD     = 0x05,
    WFS_DEVTYPE_NAND       = 0x06
} WFSDevType;

typedef struct {
    u64        nDevTotalCapacity;   // Total capacity of underlying device in bytes
    u8         nDevType;            // One of WFS_DEVTYPE_*
    u8         nDevFlags;           // Combination of WFS_DEVFLAG_*
} WFSDeviceInfo;

// ToDo: Should consider the case of users making an exact copy of a WFS volume causing two devices to
// have the same "unique" device Id. This may be an unavoidable problem.
// We can return WFS_RESULT_VOL_ID_ERROR if we attempt to mount a device with the same Id as an existing device.


// ---------------------------------------------------------------------
// WFSFileHandle, WFSSearchDirectoryHandle
// ---------------------------------------------------------------------

typedef u32 WFSFileHandle;
typedef u32 WFSSearchDirectoryHandle;


// ---------------------------------------------------------------------
// WFSFileInfo
// ---------------------------------------------------------------------

typedef struct {
    union {
        WFSFileSize nFileSize;    // Assuming this is a file
        u32         nNumEntries;  // Assuming this is a directory
    };
    u32           nFlags;         // Includes the permissions for the requesting user: WFS_PERM* and directory status flag: WFS_FLAG_IS_A_DIRECTORY
    WFSContentIdx nCidx;
    WFSFileTime   timeCreated;
    WFSFileTime   timeUpdated;
} WFSFileAttributes;

typedef struct {
    WFSFileAttributes attr;
    utf8              sFileName[WFS_MAX_FILE_NAME_SIZE+1];
} WFSFileInfo;


// ---------------------------------------------------------------------
// WFS File Attributes
// ---------------------------------------------------------------------

// The following flags are not stored per-title id, but they are included along with the permissions for the current title id in the nFlags field of WFSFileAttributes
#define WFS_FLAG_IS_A_DIRECTORY    0x80000000 // This flag can not be changed after creation.
#define WFS_FLAG_CONTIGUOUS_AREA   0x40000000 // Specifies that a directory has a contiguous set of blocks allocated to it, and is a self-contained area


// ---------------------------------------------------------------------
// WFSAccess
// ---------------------------------------------------------------------
typedef enum {
    WFS_ACCESS_READ  = 0x01,
    WFS_ACCESS_WRITE = 0x02,
    WFS_ACCESS_RW    = 0x03 // WFS_ACCESS_READ | WFS_ACCESS_WRITE
} WFSAccess;


// ---------------------------------------------------------------------
// WFS callback types
// ---------------------------------------------------------------------
typedef void (*WFSResultCallback          )(WFSResult nResult, void *pUserData);
typedef void (*WFSDeviceCallback          )(const utf8 *sDeviceName, void *pUserData); // For attach/detach callbacks
typedef void (*WFSDeviceInfoCallback      )(WFSResult nResult, const utf8 *sDeviceName, WFSDeviceInfo *pDi, void *pUserData);
typedef void (*WFSDeviceVolumeCallback    )(WFSResult nResult, const utf8 *sDeviceName, utf8 *sVolumeId, void *pUserData);
typedef void (*WFSUnmountVolumeCallback   )(WFSResult nResult, const utf8 *sVolumeId, void *pUserData);
typedef void (*WFSInitializeDeviceCallback)(WFSResult nResult, const utf8 *sDeviceName, void *pUserData);
typedef void (*WFSFileOpCallback          )(WFSResult nResult, WFSFileHandle fh, void *pUserData);
typedef void (*WFSGetFileSizeCallback     )(WFSResult nResult, u32 nFileSize, WFSFileHandle fh, void *pUserData);
typedef void (*WFSReadFileCallback        )(s32 nResultOrNumBytesRead, WFSFileHandle fh, void *pUserData);
typedef void (*WFSGetFreeSpaceCallback    )(WFSResult nResult, u64 nSize, void *pUserData);
typedef void (*WFSGetAttributesCallback   )(WFSResult nResult, const WFSFileAttributes *pFa, void *pUserData);
typedef void (*WFSSearchDirectoryCallback )(WFSResult nResult, WFSSearchDirectoryHandle sdh, const WFSFileInfo *pFi, void *pUserData);
typedef void (*WFSSearchDirectoryCloseCallback)(WFSResult nResult, WFSSearchDirectoryHandle sdh, void *pUserData);
typedef void (*WFSGetPermissionsForUserCallback)(WFSResult nResult, u32 nFlags, void *pUserData);
typedef void (*WFSGetAccessListCallback   )(s32 nResultOrNumAclEntries, void *pUserData);


#ifdef __cplusplus
} /* extern "C" */
#endif

#endif //define __WFSTYPES_H__
