/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfskrn_Area.h - Module for managing contiguous disk areas

  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_Area.h,v $
  Revision 1.24  2008/12/18 09:03:37  kondo_masahiro
  Fixed function name and process from BCacheSetBlkAdr() to BCacheRemap()

  Revision 1.23  2008/12/17 01:37:42  kondo_masahiro
  Added function BCacheSetBlkAdr()

  Revision 1.22  2008/11/26 01:46:19  kondo_masahiro
  Fixed arguments of several functions in order to improve access speed.

  Revision 1.21  2008/10/17 08:51:27  kondo_masahiro
  Added the new MEDIUM size category (40KB-320KB)

  Revision 1.20  2008/10/14 10:27:44  kondo_masahiro
  Added codes to access very large size file data

  Revision 1.19  2008/10/10 02:11:49  kondo_masahiro
  Fixed a bug to access large files

  Revision 1.18  2008/10/09 04:15:46  kondo_masahiro
  Fixed member of AreaHdr and AreaInfo

  Revision 1.17  2008/10/08 23:20:09  kondo_masahiro
  Added codes to access large size file data

  Revision 1.16  2008/09/28 23:30:08  kondo_masahiro
  Fixed codes to use medium size user block.

  Revision 1.15  2008/09/24 07:32:07  ueno
  Added AclHdr{} to AreaHdr{} to save configuration of accesslists in a disk.

  Revision 1.14  2008/08/27 09:53:34  paul
  Changed information in area header slightly. Added nStride to alloc, free functions

  Revision 1.13  2008/08/14 08:04:16  kondo_masahiro
  Removed struct AreaHashHdr.

  Revision 1.12  2008/08/07 10:42:39  ueno
  Modified for BTREE_BASED_ALLOCATOR.

  Revision 1.11  2008/08/05 04:17:15  kondo_masahiro
  Added error handling for hash error and detach device.

  Revision 1.10  2008/07/23 04:21:43  ueno
  Modified AreaInfo{} for B-tree based allocator.

  Revision 1.9  2008/07/17 04:54:33  kondo_masahiro
  Fixed codes for porting IOP.

  Revision 1.8  2008/07/09 01:12:34  paul
  Many changes to area structs - to allow directory module to write via block cache, support volume initialization and mounting exisiting volumes.

  Revision 1.14  2008/04/28 20:20:57  wayne.wong
  Changed the interface to block alloc slightly to return WFSKrnResult.

  Revision 1.13  2008/04/26 05:51:56  wayne.wong
  Fixed prototype name and arg type.

  Revision 1.7  2008/05/10 04:00:48  kondo_masahiro
  Merged and fixed code to accept RM compiler

  Revision 1.6  2008/04/25 19:15:19  kondo_masahiro
  Fixed code to accept RM compiler

  Revision 1.4  2008/04/25 06:22:55  ueno
  Cleanup.

  Revision 1.3  2008/04/25 04:40:47  ueno
  Added support for WFSDEV.

  Revision 1.12  2008/04/26 01:23:55  wayne.wong
  Moved area functionality from Volume to Area module. Changed implementation
  to support small area sizes for a large mapped number of sectors. Added
  block wrapper calls for Area and BCache calls.

  Revision 1.11  2008/04/23 00:38:36  paul
  Removed old defs

  Revision 1.10  2008/04/19 05:50:05  kondo_masahiro
  Fixed code to accept RM compiler

  Revision 1.9  2008/04/18 09:04:48  nakanose_jin
  merging acl

  Revision 1.8  2008/04/18 01:58:18  wayne.wong
  Created separate Area types for persistent (hdr) and non-persistent (info) types.
  Changed function names to be consistent with kernel naming convention.
  Defines are included to support the previous naming, but will be deleted in
  the future.

  Revision 1.7  2008/03/11 03:18:39  wayne.wong
  Added sector information to the area header.

  Revision 1.6  2008/02/28 13:11:58  ueno
  Revised to keep accesslist cache and indexlist handle in WFSAreaHdr{}.

  Revision 1.5  2008/02/28 06:37:24  kondo_masahiro
  Fixed code to accept PPC compiler

  Revision 1.4  2008/02/27 11:17:16  ueno
  Added fields for accesslist development to WFSAreaHdr{}.

  Revision 1.3  2008/02/20 07:10:44  paul
  Added WfsAreaFreeBlks function, WFSTransBlkAdr struct

  Revision 1.2  2007/12/27 11:39:22  paul
  preliminary API to help me proceed with the Dir module

  Revision 1.1  2007/11/16 05:44:48  paul
  first checkin


*---------------------------------------------------------------------------*/

#ifndef wfs_Area_h
#define wfs_Area_h

#ifdef _DEBUG
    #define _CHECK_BLK_USAGE 0  // change to check for invalid area block and block cache usage
    #define _CHECK_BLK_ALLOCATION 0
#else
    #define _CHECK_BLK_USAGE 0
    #define _CHECK_BLK_ALLOCATION 0
#endif

#include "wfskrn_Defs.h"
#include "wfskrn_BCache.h"
#include "wfskrn_Permission_AccessList_Type.h"

///// The flag enables block cache. /////
#define AREA_ENABLE_BLK_CACHE  1

#if AREA_ENABLE_BLK_CACHE
    #define BTREE_BASED_ALLOCATOR 1
    #define LIST_BASED_ALLOCATOR  0
    #define SIMPLE_ALLOCATOR      0
#else
    #define BTREE_BASED_ALLOCATOR 0
    #define LIST_BASED_ALLOCATOR  0
    #define SIMPLE_ALLOCATOR      1
#endif


#if BTREE_BASED_ALLOCATOR
#include "wfskrn_FreeBlkAlloc_Type.h"
#endif // BTREE_BASED_ALLOCATOR

//#define AREA_SECTOR_SIZE         512
//#define AREA_SECTOR_SIZE_BITS    9

#define AREA_PAGE_SIZE           (1<<WFS_LOG2_SMALL_BLK_SIZE)

#define AREA_NULL           (-1)         // end marker for sxx types

// wfsVolumeAreaEntry attribute bits
#define AREA_EMPTY        0x0000         // empty, not initialized
#define AREA_ALLOCATED    0x0001         // set if allocated
#define AREA_START        0x0002         // set if start of an area
// #define AREA_FIRST        0x0004         // first area has overhead
#define AREA_OVERHEAD     0x0004         // area overhead; header(s) + entry list

#define AREA_FLAG_META_DATA   0
#define AREA_FLAG_USER_DATA   BCACHE_FLAG_FILE_BLK

#ifndef _struct_WFSAccessListControlBlock_
#define _struct_WFSAccessListControlBlock_
typedef struct _WFSAccessListControlBlock WFSAccessListControlBlock;
#endif


#if LIST_BASED_ALLOCATOR
    typedef WFSBlkAdr AreaEntryIdx;

    // Area table entry.  Describe the on a min-area sized granularity.
    typedef struct _AreaEntry {
        s16          nSizeIdx;        // size enumeration
        u16          nAttributes;     // free
        AreaEntryIdx nNextIdx;        // abs area index/number of next in this list
        AreaEntryIdx nPrevIdx;        // abs area index/number of prev in this list
    } AreaEntry;

    typedef struct _AreaList {
        u32          nAreaSize;       // in sectors
        AreaEntryIdx nHeadIdx;        // index to heads of freelist
        u32          nLen;            // length; number of same-sized areas in list
    } AreaList;
#endif //LIST_BASED_ALLOCATOR

#if !_IOP
namespace wfsdev_api{
  struct tagSAreaInfo;
}
#endif

typedef struct AreaHdr_ {
    // This is the part of the AreaInfo struct which is stored on disk
    // ** Note: All block addresses and values must be independent of the absolute location of the area **
    // ToDo: Move the on-disk generic meta-data header out of the headers
    WFSMetaDataFlags    nFlags;               // 4 Bytes
    WFSHashCode         hash;                 // 20 bytes
    WFSBlkAdr           nRootBlkAdr;          // Root directory block of the area relative to the start of the area
    WFSBlkAdr           nNumBlks;             // Total size of the area in blocks (small blocks or meta data blocks)
#if LIST_BASED_ALLOCATOR
    WFSBlkAdr           nDataStartBlkAdr;     // Block address of end of free list management area  (relative to start of area)
    u16                 nMaxNumAreas;         // maximum number of minimum-sized areas on device
#endif
#if BTREE_BASED_ALLOCATOR
    FBATreeHdr          freeBlkAllocHdr;
#endif // BTREE_BASED_ALLOCATOR

    AclHdr              aclHdr;               // Block addresses of the root directories of accesslists and the namefiles.

    u8                  nLog2BlkSize;         // Log to the base 2 of the block size (small blocks or meta data blocks)
    u8                  nLog2MediumBlkSize;   // Log to the base 2 of the medium block size (for Wii, this value is 16)
    u8                  nLog2LargeBlkSize;    // Log to the base 2 of the multiple between the large block size. (large blocks are used for large files only)
} AreaHdr;


typedef struct {
#if LIST_BASED_ALLOCATOR
    AreaList            aFreeList[WFSKRN_MAX_NUM_AREA_SIZES];   // freelists
#elif SIMPLE_ALLOCATOR
    WFSBlkAdr           nFirstFreeBlkAdr;           // Block address of the first free block (relative to start of area)
    WFSBlkAdr           nNumFreeBlks;
    u8                 *pPtr;                       // pointer to a title area in memory
#endif
} AreaFreeListHdr;


typedef struct AreaInfo_ {
    struct AreaInfo_   *pParent;                        // Points to the parent area info. NULL if this is the root area.
    struct VolumeInfo_ *pVolInfo;                       // Points to the volume info
    AreaHdr             ah;                             // Area header information. This is a copy of the header stored on disk
    AreaFreeListHdr    *pAflh;                          // Pointer to the free list header
    WFSBlkAdr           nAbsStartBlkAdr;                // Absolute block address of the first block of the area
    WFSBlkAdr           nRelStartBlkAdr;                // block address of the first block of the area relative to the start of the parent area
    u32                 nBlkSize;                       // Block size in bytes (for meta data, and small file blocks). On Wii, block size is 8KB
    u32                 nMediumBlkSize;                 // Medium block size in bytes (for large file blocks). On Wii, medium block size is 64KB
    u32                 nLargeBlkSize;                  // Large block size in bytes (for large file blocks). On Wii, medium block size is 512KB
    u32                 nLog2BlksPerMediumBlk;     // nLog2 number of small blocks in medium block
    u32                 nLog2BlksPerLargeBlk;           // nLog2 number of small blocks in large block
    u32                 nLog2MediumBlkPerLargeBlk; // nLog2 number of medium blocks in large block
    u32                 nLargeBlkPtrSize;               // Used in indirect block layout calculations
    //u32                 nMaxLargeBlkPtrsPerAttrSubBlk;  // Determines when to switch file layout from large size category to very large size category
    u32                 nNumLargeBlkPtrsPerBlk;         // Used to calculate number of indirect blocks needed to index into large files
    u32                 nNumMediumBlkPtrsPerBlk;   // Used to calculate number of indirect blocks needed to index into medium files

#if LIST_BASED_ALLOCATOR
    u32                 nNumEntriesInFirstBlk;          // Number of free-list entries in first block
    u32                 nNumEntriesPerBlk;              // Number of free-list entries per block
#endif

#if _CHECK_BLK_ALLOCATION
    u32                *aBlkAllocState;
#endif
#if !_IOP
    struct wfsdev_api::tagSAreaInfo *aclWrapper;
#endif
    WFSAccessListControlBlock* accessListCb;
} AreaInfo;


//  *** Old Area module API functions (only recording memory) ****


WFSKrnResult AreaAllocBlks(AreaInfo *pAreaInfo, WFSBlkAdr nRefBlkAdr, u32 nLog2BlkSize, u32 nNumBlks, WFSBlkAdr *aBlkAdr, s32 nStride);
WFSKrnResult AreaFreeBlks(AreaInfo *pAreaInfo, u32 nLog2BlkSize, u32 nNumBlks, WFSBlkAdr *pBlkAdr, s32 nStride);
WFSKrnResult AreaInvalidateBlks(AreaInfo *pAreaInfo, u32 nNumBlks, WFSBlkAdr *pBlkAdr, s32 nStride);
WFSBool      AreaGetRegionBlkSize(AreaInfo *pAreaInfo, WFSBlkAdr nBlkAdr);
void         AreaSetRegionBlkSize(AreaInfo *pAreaInfo, WFSBlkAdr nBlkAdr, WFSBool bLargeBlk);
void         AreaFindSubArea(utf8 *pPathName, AreaInfo *pAreaInfo);
void *       AreaGetPhysicalBlk(AreaInfo *pAreaInfo, WFSBlkAdr nBlkAdr);


WFSKrnResult AreaInit(AreaInfo *pAreaInfo);         // Initializes the area, writes it's header and initial metadata to disk
WFSKrnResult AreaOpen(AreaInfo *pAreaInfo);         // Opens and reads an existing area from disk
void         AreaClose(AreaInfo *pAreaInfo);        // Closes an area


// Display the free list information for an area. (for debug)
WFSKrnResult AreaFreelistPrint(AreaInfo *pAreaInfo);

// Allocate a sub-area.
WFSKrnResult AreaAllocSubArea(AreaInfo *pAreaInfo,     // area to allocate from
                              WFSBlkAdr nNumBlks,      // size of sub area expressed as the number of parent area small blocks
                              AreaInfo *pSubAreaInfo,  // returned sub-area
                              u32       nFlags);        //


// Free a sub-area
WFSKrnResult AreaFreeSubArea(AreaInfo  *pAreaInfo,     // area to free to
                             AreaInfo  *pSubAreaInfo); // sub-area to free


// Block interface to block cache (BCache module)
WFSKrnResult AreaBlkAlloc (AreaInfo *pAreaInfo, WFSBlkAdr nRefBlkAdr, u32 nNumBlks, WFSBlkAdr aBlkAdr[]);
WFSKrnResult AreaBlkStore (AreaInfo *pAreaInfo, WFSBlkAdr blkAdr);
WFSKrnResult AreaBlkFlush (AreaInfo *pAreaInfo, WFSBlkAdr blkAdr);
WFSKrnResult AreaBlkInvalidate (AreaInfo *pAreaInfo, WFSBlkAdr blkAdr);
WFSKrnResult AreaBlkDirty (AreaInfo *pAreaInfo, WFSBlkAdr blkAdr);
WFSKrnResult AreaBlkPin   (AreaInfo *pAreaInfo, WFSBlkAdr blkAdr);
WFSKrnResult AreaBlkUnpin (AreaInfo *pAreaInfo, WFSBlkAdr blkAdr);
u32          AreaBlkGetPinCount(AreaInfo *pAreaInfo, WFSBlkAdr blkAdr);
WFSKrnResult AreaBlkGetPtr(AreaInfo *pAreaInfo, WFSBlkAdr blkAdr, void** ppPtr);
WFSKrnResult AreaBlkSetValidSize(AreaInfo *pAreaInfo, WFSBlkAdr blkAdr, u32 nValidSize);
WFSKrnResult AreaBlkRemapCache(AreaInfo *pAreaInfo, WFSBlkAdr nSrcBlkAdr, WFSBlkAdr nDstBlkAdr, WFSBlkAdr nHashBlkAdr, u32 nHashOfs);

//#define _DEBUG_WFSKRN_AREA_

#ifdef _DEBUG_WFSKRN_AREA_
#define dbga(_s)                 (_s)
#else
#define dbga(_s)
#endif // _DEBUG_WFSKRN_BCACHE_

#endif //define wfs_Area_h
