/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfskrn_BCache.h - disk block buffer cache library

  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_BCache.h,v $
  Revision 1.16  2008/12/18 09:03:37  kondo_masahiro
  Fixed function name and process from BCacheSetBlkAdr() to BCacheRemap()

  Revision 1.15  2008/12/17 01:37:42  kondo_masahiro
  Added function BCacheSetBlkAdr()

  Revision 1.14  2008/12/10 08:20:01  ueno
  Added BCacheDumpPinCount() for debug.

  Revision 1.13  2008/12/03 00:04:53  kondo_masahiro
  Added WFSBCacheStoreVolume and removed WFSBCacheFlushAll.

  Revision 1.12  2008/11/26 01:46:19  kondo_masahiro
  Fixed arguments of several functions in order to improve access speed.

  Revision 1.11  2008/11/05 05:56:30  kondo_masahiro
  Added BCacheSet/GetChainId()

  Revision 1.10  2008/10/03 08:37:18  kondo_masahiro
  Fixed codes for porting IOP

  Revision 1.9  2008/09/29 10:19:25  ueno
  Revised to support multi-B-tree-based allocator.

  Revision 1.8  2008/09/28 23:30:13  kondo_masahiro
  Fixed codes to use medium size user block.

  Revision 1.7  2008/08/27 23:17:41  paul
  Added BCACHE_FLAG_FILE_BLK

  Revision 1.6  2008/08/08 13:33:29  ueno
  Added BCacheCheck2() to check freeblock allocator.

  Revision 1.5  2008/08/05 04:17:20  kondo_masahiro
  Added error handling for hash error and detach device.

  Revision 1.4  2008/07/17 04:54:48  kondo_masahiro
  Fixed codes for porting IOP.

  Revision 1.3  2008/07/09 01:15:33  paul
  Changed block cache to use block address instead of sector address. Combined block cache globals into a single struct. Added a function to re-tag a cached block.

  Revision 1.2  2008/05/12 19:19:20  paul
  Merged in Wayne's previous changes

  Revision 1.1  2008/04/22 04:39:07  kondo_masahiro
  Fixed code to transfer the wfskrn project to RM_SDK tree

  Revision 1.9  2008/04/26 01:07:36  wayne.wong
  Removed dynamic memory allocation. User is now required to pass memory to
  use for descriptor table. A function is added to calculate the size for the user.
  Changed type name to be uniform with other modules.

  Revision 1.8  2008/03/19 07:36:32  wayne.wong
  Fixed a bug with sector tag matching at page granularity.

  Revision 1.7  2008/03/11 03:12:27  wayne.wong
  Tweaked interface a little. Most noteable is one can now specify whether a page
  will be dirty upon allocation. This eliminates a possible second BCache call
  to mark as dirty later.

  Revision 1.6  2008/02/22 00:40:20  wayne.wong
  Change BCache module's prefix for functions and types. Added some new functions.

  Revision 1.5  2008/02/20 07:44:37  paul
  Changed names to start with PMem

  Revision 1.4  2008/02/05 03:13:56  wayne.wong
  Added WfsBCacheGetConfig() call.

  Revision 1.3  2008/01/30 03:52:33  wayne.wong
  Added WfsBCacheValidVolume() call.

  Revision 1.2  2008/01/18 06:12:45  wayne.wong
  Refined api. Block(page) cache now matches on the tuple (volumeId, sector).

  Revision 1.1  2007/12/31 21:19:57  wayne.wong
  Initial version.

*---------------------------------------------------------------------------*/

#ifndef _wfskrn_BCache_h_
#define _wfskrn_BCache_h_

#include "wfskrn_Volume.h"
#include "wfskrn_Device.h"
#include "wfskrn_Area.h"

#define BCACHE_FLAG_INVALID          0x00
#define BCACHE_FLAG_READ             0x01
#define BCACHE_FLAG_DIRTY            0x02
#define BCACHE_FLAG_PINNED           0x04
#define BCACHE_FLAG_VALID            0x08
#define BCACHE_FLAG_FILE_BLK         0x10

#define BCACHE_GROUP_ID_METABLK     0x00
#define BCACHE_GROUP_ID_USERBLK_8K  0x01
#define BCACHE_GROUP_ID_USERBLK_64K 0x02

#define BCACHE_NUM_OF_GROUP          3

// The block cache (BCache) caches metadata disk blocks in memory. The metadata disk
// block size blkSize (e.g., 8KB).  The buffer cache is fully set-associative.  Let us
// call a disk block in memory a page. If no invalid pages are available, pages are
// replaced with a global LRU policy. Pages can be allocated, loaded, stored, flushed,
// invalidated, and pinned.  Pinned pages are not candidates for replacement.  If all
// pages are pinned and a new page is needed, the wfsBCache call will fail until a page
// is unpinned or invalidated (thus, generating a possible victim page for replacement).
//
// Initially, other than the alloc/read call, the interface was with a u8*.  Instead,
// let's try with the tuple (volumeId, a physical sector/LBA address).  The sector
// will be aligned to a page granularity.  Upon allocation, a pointer to the
// start of the page is returned, even if the sector address specified is in the
// middle of the page.
//
// Wayne - Why not use device handle instead of volumeId?  What about the multi-client case
// where each client may have the a different device handle to the same volume?
// Or will this never happen?
//

typedef struct BCacheListNode_ {
    void                   *pData;
    struct BCacheListNode_ *pNext;
    struct BCacheListNode_ *pPrev;
} BCacheListNode;


typedef struct BCacheList_ {
    BCacheListNode *pHead;
    BCacheListNode *pTail;
    u16             nLen;
} BCacheList;

typedef struct BCacheGroup_ BCacheGroup;

typedef struct {
    // ToDo: Later, it may be better to make the block cache use a pointer to the AreaInfo struct (which would
    //       have to be statically allocated and only reused after all of its blocks have been flushed from cache),
    //       plus the area-relative block address as the tag.
    WFSBlkAdr     nAbsBlkAdr;     // tag1 - HDD physical absolute block address that is being cached
    struct VolumeInfo_ *pVolInfo; // tag2 - pointer to volume struct2
    DeviceInfo   *pDevInfo;
    u32           nState;         // stored flags including BCACHE_FLAG_FILE_BLK and BCACHE_FLAG_SIZE_64K
    u32           nPinCount;
    u8           *pBlkPtr;
    WFSBlkAdr     nHashBlkAdr;
    WFSHashCode  *pHash;
    BCacheGroup  *pGroup;
    u32           nValidSize;
    u32           nChainId;
#if _CHECK_BLK_USAGE
    u32           nCheckSum;
#endif
#if _DEBUG
    u32           nDbgCommandCount;
#endif
} BCacheEntry;

struct BCacheGroup_{
    u8           *pBCache;       // start of actual page/block cache
    BCacheEntry  *pBCacheTable;  // cache info entries

    u8           *pMemCache;     // starting address of cached pages
    u32           nMemCacheSize; // size of cache, in bytes
    u32           nLog2BlkSize;  // base 2 logarithm of the block size for group
    u32           nBlkSize;      // block size for group (in bytes)
    u32           nNumBlks;      // size of cache in blocks

    u32           nNumMapped;    // number of mapped pages
    u32           nNumInvalid;   // number of invalid pages
    u32           nNumPinned;    // number of pinned pages

    BCacheList    bclInvalid;    // head of invalid list
    BCacheList    bclLru;        // LRU/MRU list; head=LRU; contains valid pages
    BCacheList    bclPinned;     // pinned pages; contains valid pages
    BCacheList    bclFree;
    BCacheListNode *pBclNodes;
};


typedef struct BCacheState_ {
    // previously global variables:
    u8           *pHeap;
    u32           nHeapSize;

    u32           nLog2UnitBlkSize;  // base 2 logarithm of the block size for unit block
    u32           nUnitBlkSize;      // block size for unit block (in bytes)

    u32           nNumGroups;
    BCacheGroup   aGroup[BCACHE_NUM_OF_GROUP];

#if _DEBUG
    bool          bSkipCheckPin;
#endif

    BCacheList    aHashTbl[WFSKRN_BCACHE_HASH_TABLE_SIZE]; // chained hash table
} BCacheGlobals;

extern BCacheGlobals bcg;


u32          BCacheMetadataMemSize(u32 memCacheSize, u32 nBlkSize);  // calc pMem arg to BCacheInitLib
//WFSKrnResult BCacheInitLib      (u8 *pMemCache, u32 memCacheSize, u32 nLog2BlkSize, u8 *pMem, u32 memSize);
WFSKrnResult BCacheInit           (u32 nLog2UnitBlkSize, u8 *pMem, u32 nMemSize);
WFSKrnResult BCacheGroupInit      (u32 nBCacheGroupId, u8 *pMemCache, u32 memCacheSize, u32 nLog2BlkSize);
WFSKrnResult BCacheGetState       (BCacheGroup *pGroup);
WFSKrnResult BCacheMapVolume      (struct VolumeInfo_ *pVolInfo, DeviceInfo *pDevInfo);
bool         BCacheValidVolume    (struct VolumeInfo_ *pVolInfo);
WFSKrnResult BCacheAlloc          (struct VolumeInfo_ *pVolInfo, WFSBlkAdr nAbsBlkAdr, u32 nFlags, u32 nReadSize, u32 nBCacheGroupId, 
                                   WFSBlkAdr nHashBlkAdr, u32 nHashOfs, BCacheEntry** ppBce);
WFSKrnResult BCacheSetHash        (struct VolumeInfo_ *pVolInfo, WFSBlkAdr nAbsBlkAdr, WFSBlkAdr nHashBlkAdr, u32 nHashOfs);
WFSKrnResult BCacheSetValidSize   (struct VolumeInfo_ *pVolInfo, WFSBlkAdr nAbsBlkAdr, u32 nValidSize);
WFSKrnResult BCacheSetChainId     (struct VolumeInfo_ *pVolInfo, WFSBlkAdr nAbsBlkAdr, u32 nChainId);
WFSKrnResult BCacheGetChainId     (struct VolumeInfo_ *pVolInfo, WFSBlkAdr nAbsBlkAdr, u32 *pChainId);
WFSKrnResult BCacheRemap          (struct VolumeInfo_ *pVolInfo, WFSBlkAdr nSrcAbsBlkAdr, WFSBlkAdr nDstAbsBlkAdr, WFSBlkAdr nHashAbsBlkAdr, u32 nHashOfs);
void         BCacheReTag          (struct VolumeInfo_ *pOldVolInfo, WFSBlkAdr nOldAbsBlkAdr,
                                   struct VolumeInfo_ *pNewVolInfo, WFSBlkAdr nNewAbsBlkAdr);
void*        BCacheFind           (struct VolumeInfo_ *pVolInfo, WFSBlkAdr nAbsBlkAdr);  // find a page
WFSKrnResult BCacheStore          (struct VolumeInfo_ *pVolInfo, WFSBlkAdr nAbsBlkAdr);  // store page & mark clean
WFSKrnResult BCacheFlush          (struct VolumeInfo_ *pVolInfo, WFSBlkAdr nAbsBlkAdr);  // store & invalidate page
//WFSKrnResult BCacheFlushAll       ();                                                  // flush all pages
WFSKrnResult BCacheStoreVolume    (struct VolumeInfo_ *pVolInfo);                        // store a volume's pages
WFSKrnResult BCacheFlushVolume    (struct VolumeInfo_ *pVolInfo);                        // flush a volume's pages
WFSKrnResult BCacheInvalidate     (struct VolumeInfo_ *pVolInfo, WFSBlkAdr nAbsBlkAdr);  // invalidate a page
WFSKrnResult BCacheInvalidateBlks (struct VolumeInfo_ *pVolInfo, WFSBlkAdr nAbsBlkAdr, u32 nNumBlks);  // invalidate a page
WFSKrnResult BCacheTouch          (struct VolumeInfo_ *pVolInfo, WFSBlkAdr nAbsBlkAdr);  // make a page MRU
WFSKrnResult BCacheDirty          (struct VolumeInfo_ *pVolInfo, WFSBlkAdr nAbsBlkAdr);  // mark a page as modified
WFSKrnResult BCachePin            (struct VolumeInfo_ *pVolInfo, WFSBlkAdr nAbsBlkAdr);  // page cannot be evicted
WFSKrnResult BCacheUnpin          (struct VolumeInfo_ *pVolInfo, WFSBlkAdr nAbsBlkAdr);  // page now can be evicted
u32          BCacheGetPinCount    (struct VolumeInfo_ *pVolInfo, WFSBlkAdr nAbsBlkAdr);
void*        BCachePointer        (struct VolumeInfo_ *pVolInfo, WFSBlkAdr nAbsBlkAdr);
WFSKrnResult BCacheCheck          ();
WFSKrnResult BCacheCheck2         ();

static inline
WFSKrnResult BCacheAllocMetaBlk(struct VolumeInfo_ *pVolInfo, WFSBlkAdr nAbsBlkAdr, u32 nFlags, BCacheEntry** ppBce)
{
    return BCacheAlloc(pVolInfo, nAbsBlkAdr, nFlags, 1<<bcg.aGroup[BCACHE_GROUP_ID_METABLK].nLog2BlkSize, BCACHE_GROUP_ID_METABLK, nAbsBlkAdr, sizeof(WFSMetaDataFlags), ppBce);
}

u32          BCacheDumpPinCount    ( void );
u32          BCacheSkipCheckPin    ();
WFSKrnResult BCacheStopSkipCheckPin( u32 initialPinCount );

#if _DEBUG_WFSKRN_BCACHE_
#define dbgb(_s)                 (_s)
#else
#define dbgb(_s)
#endif // _DEBUG_WFSKRN_BCACHE_

#endif // _wfskrn_BCache_h_
