/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfskrn_BitField.h - functions for manipulating bitfields within bit arrays

  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_BitField.h,v $
  Revision 1.1  2008/04/22 04:39:07  kondo_masahiro
  Fixed code to transfer the wfskrn project to RM_SDK tree

  Revision 1.2  2008/04/05 03:09:45  paul
  Added WfsSetBitField, WfsTestAndSetBitField

  Revision 1.1  2007/12/27 11:39:44  paul
  *** empty log message ***

*---------------------------------------------------------------------------*/

#ifndef wfskrn_BitField_h
#define wfskrn_BitField_h

#ifndef max
#define max(a,b)            (((a) > (b)) ? (a) : (b))
#endif

#ifndef min
#define min(a,b)            (((a) < (b)) ? (a) : (b))
#endif

#define BITFIELD_USE_TEMPLATES 0

#include "wfskrn_Defs.h"

//-----------------------------------------------------------------------------------------------------------------

u32 WfsCalculateFieldSizeInBytes(u32 nVal);
u8 WfsCalculateFieldSizeInBits(u32 nVal);

#if BITFIELD_USE_TEMPLATES
    template<int nNumBits> unsigned WfsReadBitField (unsigned *aBitArray, int nStartBit);
#endif
unsigned WfsReadBitField(unsigned *aBitArray, int nStartBit, int nNumBits);
// Arguments:
// ----------
// aBitArray is an array of machine sized unsigned int which are interpreted as a bit array.
// (The first bit of the array is the most significant bit of aBitArray[0])
// nStartBit is the bit offset in the bit array to start reading from.
// Pre Conditions: 1 <= nNumBits <= BITS_PER_INT
//
// Return Values:
// --------------
// Returns a machine sized unsigned int whose least significant nNumBits are the extracted bit field.
//
// Description:
// ------------
// This function reads a bitfield of width nNumBits from a bit array.
// NOTE: It is up to the caller to mask off the least significant nNumBits bits of the return value
//       if necessary, using a mask such as ((1<<nNumBits)-1).


//-----------------------------------------------------------------------------------------------------------------

#if BITFIELD_USE_TEMPLATES
    template<unsigned nNumBits> void WfsOverwriteBitField(unsigned *aBitArray, int nStartBit, unsigned nValue);
#endif
void WfsOverwriteBitField(unsigned *aBitArray, int nStartBit, unsigned nValue, int nNumBits, unsigned nMask);
// Arguments:
// ----------
// aBitArray is an array of machine sized words which are interpreted as a bit array.
// (The first bit of the array is the most significant bit of aBitArray[0])
// nValue is the value to write.
// nNumBits is the bit width of nValue.
// nStartBit is the bit offset in the bit array to start writing to.
// Pre Conditions: 1 <= nNumBits <= BITS_PER_INT; 0 < nValue < (1<<nNumBits)
//
// Description:
// ------------
// This function overwrites a bitfield of width nNumBits in a bit array.


//-----------------------------------------------------------------------------------------------------------------

void WfsSetBitField(unsigned *aBitArray, int nStartBit, int nNumBits);
// Arguments:
// ----------
// aBitArray is an array of machine sized words which are interpreted as a bit array.
// nStartBit is the first bit offset within the bit array to set.
// nNumBits  is the number of bits to set.
//
// Description:
// ------------
// This function sets (makes equal to 1) a atring of bits from nStartBit to (nStartBit + nNumBits - 1) inclusive.


//-----------------------------------------------------------------------------------------------------------------

unsigned WfsTestAndSetBitField(unsigned *aBitArray, int nStartBit, int nNumBits);
// Arguments:
// ----------
// aBitArray is an array of machine sized words which are interpreted as a bit array.
// nStartBit is the first bit offset within the bit array to set.
// nNumBits  is the number of bits to set.
//
// Description:
// ------------
// This function sets (makes equal to 1) a atring of bits from nStartBit to (nStartBit + nNumBits - 1) inclusive.
// It returns a non-zero value if any of the bits covered by the string were already set, and 0 otherwise.


//-----------------------------------------------------------------------------------------------------------------

void WfsBitArrayInsertGap(unsigned *aBitArray, int nBitArrayLength, int nGapStartBit, int nNumGapBits);
// Arguments:
// ----------
// aBitArray is an array of machine sized words which are interpreted as a bit array.
// nBitArrayLength is the length of the array in bits.
// nGapStartBit is the bit offset within the bit array to insert the gap.
// nNumGapBits is the number of gap bits to insert.
//
// Description:
// ------------
// This function inserts a gap into a bit array.
// The string of bits from nGapStartBit to nEndBit are shifted nNumGapBits bits to the right,
// leaving a gap which is filled with 0s.
// NOTE: nNumGapBits can be larger than BITS_PER_INT.


//-----------------------------------------------------------------------------------------------------------------

void WfsBitArrayDeleteBitField(unsigned *aBitArray, int nBitArrayLength, int nBitPos, int nNumBits);
// Arguments:
// ----------
// aBitArray is an array of machine sized words which are interpreted as a bit array.
// (The first bit of the array is the most significant bit of aBitArray[0])
// nBitArrayLength is the length of the array in bits.
// nBitPos  is the bit offset within the bit array to start deleting from.
// nValue is the value to insert.
// nNumBits is the number of bits to delete.
//
// Description:
// ------------
// This function deletes a bitfield of width nNumBits from a bit array.
// The string of bits from nBitPos to nBitArrayLength are shifted nNumBits bits to the left,


//-----------------------------------------------------------------------------------------------------------------

void WfsBitArrayInsertBitField(unsigned *aBitArray, int nBitArrayLength, int nInsertBit, int nNumBits, unsigned nValue);
// Arguments:
// ----------
// aBitArray is an array of machine sized words which are interpreted as a bit array.
// (The first bit of the array is the most significant bit of aBitArray[0])
// nBitArrayLength is the length of the array in bits.
// nInsertBit is the bit offset within the bit array to insert to.
// nValue is the value to insert.
// nNumBits is the bit width of nValue.
// Pre Conditions: 1 <= nNumBits <= BITS_PER_INT; 0 < nValue < (1<<nNumBits)
//
// Description:
// ------------
// This function inserts a bitfield of width nNumBits into a bit array.
// The string of bits from nInsertBit to nBitArrayLength are shifted nNumBits bits to the right,
// and nValue is inserted in the resulting gap.


//-----------------------------------------------------------------------------------------------------------------

void WfsBitFieldCopy(unsigned *aDstBitArray, int nDstOffset, unsigned *aSrcBitArray, int nSrcOffset, int nCopyLength);
// Arguments:
// ----------
// aDstBitArray is the destination bit array.
// nDstOffset is the bit offset within the destination bit array
// aSrcBitArray is the source bit array.
// nSrcOffset is the bit offset within the source bit array
// nCopyLength is the number of bits to copy
//
// Description:
// ------------
// This function copies nCopyLength bits from a specified offset within a source bit array to a specified offset
// within a destination bit array.


//-----------------------------------------------------------------------------------------------------------------

void WfsTestBitFieldModule();

#endif //wfskrn_BitField_h
