/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfskrn_Device.h - Storage device abstraction.

  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_Device.h,v $
  Revision 1.14  2008/12/10 06:54:02  kondo_masahiro
  Added a argument WFSDevType nDevType in DeviceAttachDetachCallback.

  Revision 1.13  2008/11/26 02:43:58  kondo_masahiro
  Added comments of DeviceRead/Write.

  Revision 1.12  2008/11/26 01:46:01  kondo_masahiro
  Fixed arguments of several functions in order to improve access speed.
  Fixed source code for Win32 version to enable easier debug.

  Revision 1.11  2008/10/09 01:40:14  kondo_masahiro
  Fixed name of the function DeviceRead/WriteSingleBlk

  Revision 1.10  2008/10/08 23:20:09  kondo_masahiro
  Added codes to access large size file data

  Revision 1.9  2008/10/03 08:37:18  kondo_masahiro
  Fixed codes for porting IOP

  Revision 1.8  2008/09/30 02:02:27  ueno
  Removed an obsolete define, WFSKRN_SIMULATED_DISK_SIZE.

  Revision 1.7  2008/09/28 23:30:24  kondo_masahiro
  Fixed codes to use medium size user block.

  Revision 1.6  2008/08/29 00:12:36  kondo_masahiro
  Added DeviceGetVolumeId.

  Revision 1.5  2008/08/14 08:20:14  kondo_masahiro
  Changed WFSKRN_SIMULATED_DISK_PATH

  Revision 1.4  2008/08/14 08:07:27  kondo_masahiro
  Changed several APIs in order to access the large size file.

  Revision 1.3  2008/08/05 04:17:29  kondo_masahiro
  Added error handling for hash error and detach device.

  Revision 1.2  2008/07/17 05:07:43  kondo_masahiro
  Added codes for IOP

  Revision 1.1  2008/07/09 01:39:12  paul
  Merged device definitions from wfskrn_PMem.h and wfskrn_Api.h.

*---------------------------------------------------------------------------*/

#ifndef _WFSKRN_DEVICE_H_
#define _WFSKRN_DEVICE_H_

#if _WIN32
#include <stdlib.h>
#include <stdio.h>
#endif

#include "revolution/wfs.h"
#include "wfs_Names.h"
#include "wfskrn_Defs.h"
#include "wfskrn_Config.h"

#if _WIN32
#include "msc_error.h"
#elif _TWL
#include "msc_error.h"
#elif _IOP
#include "msc_includes.h"
#endif

#define DEVICE_EMBEDDED_HASH_DATA        1  // If it is on, the hash data is embedded to the written data.
#define DEVICE_VERIFY_HASH_DATA          1  // If it is on, DeviceRead can outputs error when two hash data discord.
#define DEVICE_DEBUG_PSEUDO_DEVICE_ERROR 1

#define DEVICE_IOP_MAX_ASYNC_NUM         2

#if _WIN32
    #define IOSError                      s32
    #define WFSKRN_SIMULATED_DISK_PATH    "Y:\\TEMP\\shirait\\WFS_VOL_S\\"
    typedef HANDLE                        DeviceHandle;        // Windows handle
#elif _TWL
    #define IOSError                      s32
    typedef u32 DeviceHandle;
#elif _RVL
    typedef u32       DeviceHandle;
#elif _IOP
    #define DEVICE_IOP_MSC_CHECK(mscfunc) if(mscfunc < IOS_ERROR_OK){ return WFSKRN_RESULT_DEVICE_ERROR; } 
    typedef HardDisc* DeviceHandle;
#else
#endif //_WIN32


//#define _DEBUG_WFSKRN_DEVICE_

#ifdef _DEBUG_WFSKRN_DEVICE_
#define dbgd(_s)                 (_s)
#else
#define dbgd(_s)
#endif // _DEBUG_WFSKRN_DEVICE_


typedef struct DeviceHdr_ {
    u32            nLog2SectorSize;    // 1st field: Base 2 logarithm of the sector size (default = 9 corresponding to 512 byte sectors)
    u32            nNumSectors;        // 2nd field; capacity of device
} DeviceHdr;

typedef struct DeviceInfo_ {
    DeviceHdr      dh;
#if _WIN32
    HANDLE         hSimulatedDeviceFile;
#endif
    //
    bool           bInUse;
    u32            nSectorSize;
    u64            nTotalCapacity;
    DeviceHandle   nDeviceDevHandle;
    struct VolumeInfo_  *pVolInfo;
    WFSDeviceName  devName;
    WFSDevType     nDevType;
    u8             nFlags;
} DeviceInfo;


// Initialize Device library
void DeviceInitLib();

#if _IOP
// Regist callback function for attach/detach device
typedef void (*DeviceAttachDetachCallback)(u32 bAttach, u32 nDevIdx, u32 nLog2SectorSize, u32 nTotalSize, WFSDevType nDevType, void *pUserData);
void DeviceAttachDetachRegistCallback(DeviceAttachDetachCallback cbAttach, void *pUserData);
#endif

WFSKrnResult DeviceRead(DeviceInfo *pDevInfo, u32 nSectorAdr, u8 *pBuffer, u32 nSectors, WFSHashCode *pHash, u32 nLog2HashSectors, u32 nOffsetHashBlks, s32 nHashStride, s32 nHashStridePerMaxLBlks);
// Arguments:
// ----------
// pDevInfo                [IN]  .. A pointer of DeviceInfo.
// nSecrorAdr              [IN]  .. A head sector address of HDD to read.
// pBuffer                 [OUT] .. A pointer of read buffer.
// nSectors                [IN]  .. A read sector size of HDD.
// pHash                   [IN]  .. A pointer to compare the hash value.
// nLog2HashSectors        [IN]  .. A sector size unit for hash value. The hash value is calculated for each the unit.
// nOffsetHashBlks         [IN]  .. A offset block size for nHashStridePerMaxLBlks. 
// nHashStride             [IN]  .. A stride of hash address for each hash size unit.
// nHashStridePerMaxLBlks  [IN]  .. A stride of hash address for each 512KB unit.
//
// Description:
// ------------
// Read nSector from device pDevInfo starting at LBA nSectorAdr into buffer pBuffer.
// nLog2HashSector is the hash size unit. The argument nLog2HashSector must be indicate up to 64KB (~16).
// The hash value is calculated and compared with pHash for each buffer size nLog2HashSector.
// The hash address pHash is added by nHashStride for each buffer size nLog2HashSector.
// After total hash calculation size is reached to the multiples of 512KB,
// The hash address pHash is added by nHashStride and nHashStridePerMaxLBlks for "SIZE_CATEGORY_LARGE".
// For example, When nLog2HashSectors is 16(64KB) and nOffsetHashBlks is 3,
// The hash address up to 5th block are added by nHashStride. (512 / 64 - 3 = 5)
// The hash address of 6th block is added by nHashStride and nHashStridePerMaxLBlks.

static inline WFSKrnResult DeviceReadSingleBlk(DeviceInfo *pDevInfo, u32 nSectorAdr, u8 *pBuffer, u32 nSectors, WFSHashCode *pHash, u32 nLog2HashSectors){
    dbgd(printf("DeviceRead()\n"));
    ASSERT(nSectors <= 1<<(WFS_LOG2_MAX_HASHABLE_BLK_SIZE-pDevInfo->dh.nLog2SectorSize) ); // ~64KB
    return DeviceRead(pDevInfo, nSectorAdr, pBuffer, nSectors, pHash, nLog2HashSectors, 0, 0, 0);
}

WFSKrnResult DeviceWrite(DeviceInfo *pDevInfo, u32 nSectorAdr, u8 *pBuffer, u32 nSectors, WFSHashCode *pHash, u32 nLog2HashSectors, u32 nOffsetHashBlks, s32 nHashStride, s32 nHashStridePerMaxLBlks, bool bDecryption);
// Arguments:
// ----------
// pDevInfo                [IN]  .. A pointer of DeviceInfo.
// nSecrorAdr              [IN]  .. A head sector address of HDD to read.
// pBuffer                 [IN]  .. A pointer of read buffer.
// nSectors                [IN]  .. A read sector size of HDD.
// pHash                   [OUT] .. A pointer to store the hash value.
// nLog2HashSectors        [IN]  .. A sector size unit for hash value. The hash value is calculated for each the unit.
// nOffsetHashBlks         [IN]  .. A offset block size for nHashStridePerMaxLBlks. 
// nHashStride             [IN]  .. A stride of hash address for each hash size unit.
// nHashStridePerMaxLBlks  [IN]  .. A stride of hash address for each 512KB unit.
// bDecryption             [IN]  .. A flag to indicate whether buffer is decrypted after write access or not. 
//
// Description:
// ------------
// Write nSector from device pDevInfo starting at LBA nSectorAdr into buffer pBuffer.
// nLog2HashSector is the hash size unit. The argument nLog2HashSector must be indicate up to 64KB (~16).
// The hash value is calculated and stored to pHash for each buffer size nLog2HashSector.
// The hash address pHash is added by nHashStride for each buffer size nLog2HashSector.
// After total hash calculation size is reached to the multiples of 512KB,
// The hash address pHash is added by nHashStride and nHashStridePerMaxLBlks for "SIZE_CATEGORY_LARGE".
// For example, When nLog2HashSectors is 16(64KB) and nOffsetHashBlks is 3,
// The hash address up to 5th block are added by nHashStride. (512 / 64 - 3 = 5)
// The hash address of 6th block is added by nHashStride and nHashStridePerMaxLBlks.

static inline WFSKrnResult DeviceWriteSingleBlk(DeviceInfo *pDevInfo, u32 nSectorAdr, u8 *pBuffer, u32 nSectors, WFSHashCode *pHash, u32 nLog2HashSectors, bool bDecryption){
    dbgd(printf("DeviceWrite()\n"));
    ASSERT(nSectors <= 1<<(WFS_LOG2_MAX_HASHABLE_BLK_SIZE-pDevInfo->dh.nLog2SectorSize) ); // ~64KB
    return DeviceWrite(pDevInfo, nSectorAdr, pBuffer, nSectors, pHash, nLog2HashSectors, 0, 0, 0, bDecryption);
}

// Get the device info from the device name (from the path cache)
DeviceInfo *DeviceGetDeviceInfo(const WFSDeviceName *pDevName);

// Get a volume id from the device name even if the device is not mounted.
WFSKrnResult DeviceGetVolumeId(const WFSDeviceName *pDevName, WFSVolumeId *pVolId);

// Mount a volume
WFSKrnResult DeviceMountVolume(DeviceInfo *pDevInfo);

// Attach a device - called when user physically attaches a drive
WFSKrnResult DeviceAttach(DeviceInfo *pDevInfo, WFSDevType nDevType, u32 nDevIdx, u32 Log2SectorSize, u32 nNumSectors);

// Detach a device - called when user physically detaches a drive
WFSKrnResult DeviceDetach(DeviceInfo *pDevInfo);

//// Debug functions
void DeviceDebugSetPseudoDetachDevice(s32 cnt);
void DeviceDebugSetPseudoHashInconsistent(s32 cnt);
s32  DeviceDebugGetPseudoDetachDevice();
s32  DeviceDebugGetPseudoHashInconsistent();

#if _WIN32 // * Windows version *
#define      DeviceOverhead()                        (2*sizeof(u32))     // overhead for simulating a device
#else
#define      DeviceOverhead()                        0
#endif     // _WIN32

#endif // _WFSKRN_DEVICE_H_
