/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfskrn_DirRxTree.h - Find functions for the Directory Module

  Copyright 2007-2008 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_DirRxTree.h,v $
  Revision 1.4  2008/07/30 21:21:08  paul
  Changed the default value of _CHECK_RXT to 0 to improve the speed of the stress test

  Revision 1.3  2008/07/21 22:49:55  paul
  Added callback to RxtCheck()

  Revision 1.2  2008/06/09 18:06:12  paul
  Rewrote DirRxTree to support PathCache. New version includes parent pointers in each node.

  Revision 1.1  2008/05/10 04:00:23  kondo_masahiro
  Initial check-in

*---------------------------------------------------------------------------*/

#ifndef wfskrn_DirRxTree_h
#define wfskrn_DirRxTree_h

#include "wfskrn_Bitfield.h"
#include "wfskrn_Dir.h"


#define RXT_LOG2_RESERVE_SIZE    9
#define RXT_MIN_LOG2_NODE_SIZE   4
#define RXT_MAX_LOG2_SIZE       16

#if _DEBUG
    #define _CHECK_RXT           0      // Set to 1 to enable radix tree and path cache check. However, it makes the test program run a lot slower
#else
    #define _CHECK_RXT           0
#endif

#pragma pack(1)
typedef struct {
    u8              nLog2Size;      // The value n such that 2^n is the size of the node
    u8              nEntryIdx;      // The index of the entry in the parent block which points to this node
    u16             nParentOfs;     // The offset to the parent of this node
    u16             nStrLen;        // The length of the prefix string  of this node
    u8              nNumEntries;    // The number of entries in this node
    utf8            aFirstChar[1];  // The first character of the variable-size part of the node: format is prefix string followed by choice characters
    // The table of u16 offsets to child nodes are stored in reverse order starting from the end of the node
} RxtNodeHdr;
//} _ATTRIBUTE_PACKED RxtNodeHdr;
#pragma pack()


typedef struct {
    u16            *pOfs;         // A pointer to the offset pointer to the node. This would need to be updated if the sub block is relocated.
    RxtNodeHdr     *pNodeHdr;     // A pointer to the sub block containing the node.
    utf8           *pStrPtr;      // A string pointer to the current cursor position within the search name. After RxtFind(), it points to the first mismatch.
    utf8           *pNodeStrPtr;  // A string pointer to the current cursor position within the node string. After RxtFind(), it points to the first mismatch.
    s32             nEntryIdx;    // The entry index of the choice within the current node which is closest but <= search name after RxtFind().
    u32             nDepth;       // Current depth of iterator
} RxtItr;


typedef struct {
    WFSSubBlkAllocHdr sbah;
    u16               nRootOfs;
    u16               nReserveOfs;  // Offset to a sub-block which can be used to guarantee Delete will always succeed
    u16               nNumRecs;   // Number of records in the block, not the whole directory
} RxtHdr;


typedef struct {
    u8              nLog2Size;
    u8              nEntryIdx;
    u16             nParentOfs;
} RxtAttrHdr;

WFSBool DirBinarySplitSearchChoices(utf8 *aSortedArray, u32 nNumEntries, utf8 cSearchChar, u32 *pEntryIdx);

void         RxtInit    (RxtHdr *pRxtHdr, u32 nLog2Size);
WFSKrnResult RxtFind    (RxtHdr *pRxtHdr, RxtItr *pRxi, u32 nStrLen);
WFSKrnResult RxtFindNext(RxtHdr *pRxtHdr, RxtItr *pRxi, utf8 cSeparator);
WFSKrnResult RxtInsert  (RxtHdr *pRxtHdr, RxtItr *pRxi, u32 nSuffixLen, RxtAttrHdr *pAttrHdr);
WFSKrnResult RxtDeleteNode(RxtHdr *pRxtHdr, RxtItr *pRxi, u32 nEntryIdx, u32 nNodeOfs, RxtNodeHdr *pNodeHdr);
WFSKrnResult RxtDelete  (RxtHdr *pRxtHdr, RxtItr *pRxi);
WFSKrnResult RxtNext    (RxtHdr *pRxtHdr, RxtItr *pRxi);
WFSKrnResult RxtNextUntilSeparator(RxtHdr *pRxtHdr, RxtItr *pRxi, utf8 cSeparator);

void         RxtGetString(RxtHdr *pRxtHdr, RxtNodeHdr *pNodeHdr, utf8 *sStr, u32 nStrLen);


#if _CHECK_RXT
    typedef void (*RxtCheckEntry)(u32 nRecIdx, u32 nDepth, utf8 *sPath, void *pPtr);
    void RxtCheck(RxtHdr *pRxtHdr, u32 nTotalSize, RxtCheckEntry cbCheckEntry);
#endif

void         RxtTest();

// Inline Function Definitions
static inline u32 RxtGetNodeLog2Size_2(u32 nStrLen, u32 nNumEntries) {
    //u32 nNumBytesRequired = (u32)&((RxtNodeHdr*)0)->aFirstChar + nStrLen + (nNumEntries*3);
    u32 nNumBytesRequired = sizeof(RxtNodeHdr)-1 + nStrLen + (nNumEntries*3);
    u32 nLog2Size = WfsCalculateFieldSizeInBits(nNumBytesRequired-1);
    if (nLog2Size < SBA_MIN_LOG2_SUB_BLK_SIZE) {
        nLog2Size = SBA_MIN_LOG2_SUB_BLK_SIZE;
    }
    return nLog2Size;
}

static inline u32 RxtGetNodeLog2Size(RxtNodeHdr *pNodeHdr) {
    return RxtGetNodeLog2Size_2(pNodeHdr->nStrLen, pNodeHdr->nNumEntries);
}

static inline u16 *RxtGetChildOfsPtr_ote(s32 nIdx, u16 *pOfsTblEnd) {
    return &pOfsTblEnd[-nIdx];
}

static inline u16* RxtGetChildOfsPtr(s32 nIdx, RxtNodeHdr *pNodeHdr) {
    u16 *pOfsTblEnd = (u16*)((u8*)pNodeHdr + (1<<pNodeHdr->nLog2Size));
    return RxtGetChildOfsPtr_ote(nIdx, pOfsTblEnd);
}

static inline u16 RxtGetChildOfs_ote(s32 nIdx, u16 *pOfsTblEnd) {
    return *RxtGetChildOfsPtr_ote(nIdx, pOfsTblEnd);
}

static inline u16 RxtGetChildOfs(s32 nIdx, RxtNodeHdr *pNodeHdr) {
    return *RxtGetChildOfsPtr(nIdx, pNodeHdr);
}

static inline void RxtSetChildOfs_ote(s32 nIdx, u16 nOfs, u16 *pOfsTblEnd) {
    *RxtGetChildOfsPtr_ote(nIdx, pOfsTblEnd) = nOfs;
}

static inline void RxtSetChildOfs(s32 nIdx, u16 nOfs, RxtNodeHdr *pNodeHdr) {
    *RxtGetChildOfsPtr(nIdx, pNodeHdr) = nOfs;
}


#endif //define wfskrn_DirRxTree_h
