/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfskrn_FTree.h


  Copyright 2007-2008 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_FTree.h,v $
  Revision 1.15  2008/12/18 05:09:43  ueno
  Added "extern" for VC++.

  Revision 1.14  2008/12/18 00:31:45  ueno
  Moved FTreeBlkLog2Ofs{} for IOP.

  Revision 1.13  2008/12/17 05:08:17  ueno
  Cleanup.

  Revision 1.12  2008/12/02 00:32:23  ueno
  Started implementing EPTree.

  Revision 1.11  2008/10/30 04:33:12  ueno
  Modified for release build.

  Revision 1.10  2008/10/17 08:51:27  kondo_masahiro
  Added the new MEDIUM size category (40KB-320KB)

  Revision 1.9  2008/10/14 11:25:13  ueno
  Disabled runtime check.

  Revision 1.8  2008/09/29 10:19:25  ueno
  Revised to support multi-B-tree-based allocator.

  Revision 1.7  2008/09/02 07:01:05  kondo_masahiro
  Fixed VerifyMemoryBlock -> VerifyMemoryBlockFTree

  Revision 1.6  2008/08/09 03:34:29  ueno
  Disabled runtime check.

  Revision 1.5  2008/08/09 02:02:02  ueno
  Cleanup.

  Revision 1.4  2008/08/08 13:27:22  ueno
  Fixed BLOCK_SIZE.

  Revision 1.3  2008/08/07 10:46:50  ueno
  Modified for BTREE_BASED_ALLOCATOR (now debugging).

  Revision 1.2  2008/08/02 11:52:10  ueno
  Revised the positions of FTreeHdr{} and FBAPTreeHdr{} in 8KB-blocks to protect hash code.

  Revision 1.1  2008/08/01 06:57:39  ueno
  Initial Check-in.

*---------------------------------------------------------------------------*/

#ifndef __WFSKRN_FTREE_H__
#define __WFSKRN_FTREE_H__

#include "wfskrn_Area.h"
#if BTREE_BASED_ALLOCATOR

#define MAX_SUB_TREE  7

#include "wfskrn_Defs.h"
#include "wfskrn_Api.h"
#include "wfskrn_Area.h"
#include "wfskrn_Errors.h"
#include "wfskrn_PTree.h"

#define FTREE_LEAF_FAN_OUT  7

#define FTREE_LEAF_APPROPRIATE_NUM  5  // 70% of FTREE_LEAF_FAN_OUT
#define FTREE_NODE_APPROPRIATE_NUM  4  // 70% of PTREE_NODE_FAN_OUT

enum // treeType
{
    FTREE_TYPE_A = 0,
    FTREE_TYPE_B,
    FTREE_TYPE_C,
    FTREE_TYPE_D,
    FTREE_TYPE_E,
    FTREE_TYPE_F,
    FTREE_TYPE_G,
    FTREE_TYPE_END // this is not used.
};

#define GET_SUBTREE(header, treeType) (&(header)->tree[(treeType)])
#define GET_SUBTREE_TYPE(header, fti) ((fti)->subTree - &(header)->tree[0])

typedef PTreeNode FTreeNode;


/*
 [check] should save the number of data.
    4 bits * 7 (remainder 4 bits)
 <----------------------------------->
 +---+---+---+---+---+---+---+---+---+
 | x | 7 | 6 | 5 | 4 | 3 | 2 | 1 | 0 |
 +---+---+---+---+---+---+---+---+---+
*/
typedef struct
{
    u32         aKey[FTREE_LEAF_FAN_OUT];
    u32         data; // 4 bits * 7 (remainder 4 bits)
} FTreeLeaf;

#define LEAF_GET_LEN(leaf, index)  ((((leaf)->data) >> (4 * (index))) & 0x0f)
#define FTREE_SBA_OFFSET            24

/*
       an 8kb-block.
       +-------------------------+ A
       |                         | |
       |   hash code, areaInfo,  | | offset
       |   volumeInfo, etc...    | |
       |                         | |
       +-------------------------+ V
       |                         | A
       |                         | |
       |                         | |
       |                         | |
       |                         | |
       |                         | |
       |                         | |
       |       Sub-blocks        | |
       |                         | | size
       |                         | |
       |                         | |
       |                         | |
       |                         | |
       |                         | |
       |                         | V
       +-------------------------+ A
       |                         | |
       +-------------------------+ | FTreeHdr (footer?)
       | SubBlockAllocatorHeader | |
       +-------------------------+ V
*/
typedef struct _SubBlockAllocatorHeader
{
    u16 freeBlock; // start address of free blocks.
    u16 numBlock;  // the number of blocks in this 8kB block.

    u16 offset;
    u16 size;
} SubBlockAllocatorHeader;

typedef struct
{
    PTreeHdr tree[MAX_SUB_TREE];              // 8 bytes * 7
    SubBlockAllocatorHeader sbaHdr;           // 8 bytes
} FTreeHdr; // sizeof(FTreeHdr) == sizeof(FTreeBlock) * NUM_HEADER_BLOCK.

typedef struct
{
    u32 prev;
    u32 next;
    u16 num;
} FTreeFreeBlock;

typedef struct
{
    u32 prev;
    u32 next;
    u16 num;
    u8 padding[22]; // sizeof(FTreeBlock) must be 32.
} FTreeBlock;

typedef struct
{
    PTreeItr    ptree; // inherit PTreeItr.
    FTreeLeaf*  pLeaf;
    PTreeHdr*   subTree;
    WFSBlkAdr   splitKey; // [ToDo] should separete FTreeItr.
    WFSBlkAdr   newChild; // [ToDo] should separete FTreeItr.
    WFSBlkAdr   newChildSubTreeType; // [ToDo] should separete FTreeItr.
} FTreeItr;

#define FTREE_GET_KEY(pFti)    ((pFti)->ptree.nKey)
#define FTREE_GET_DATA(pFti)   ((pFti)->ptree.nData)

#define PTREE_GET_KEY(pPti)    ((pPti)->nKey)
#define PTREE_GET_DATA(pPti)   ((pPti)->nData)

#define FTREE_MASK_LENGTH 0x07FF
#define FTREE_FLAG_BTREE  0x8000
#define FTREE_BLOCK_SIZE        (32)

#define BLOCK_SIZE              (1<<WFS_LOG2_SMALL_BLK_SIZE) // [check] should use area->nBlkSize as a block size.

#define FTREE_LOG2_OFS_A              0
#define FTREE_LOG2_OFS_B              3
#define FTREE_LOG2_OFS_C              6
#define FTREE_LOG2_OFS_D              10
#define FTREE_LOG2_OFS_E              14
#define FTREE_LOG2_OFS_F              18
#define FTREE_LOG2_OFS_G              22
#define FTREE_LOG2_OFS_END            26

extern const u32 FTreeBlkLog2Ofs[MAX_SUB_TREE+1];

#define SBAHDR_OFFSET                  (BLOCK_SIZE - sizeof(SubBlockAllocatorHeader))
#define GET_SBAHDR(pBlock)             ((SubBlockAllocatorHeader*) ((u8*) pBlock + SBAHDR_OFFSET))
#define SET_FREEBLOCK(sbaHdr, pBlock)  (sbaHdr->freeBlock = ((FTreeBlock*) (pBlock) - CACHE(sbaHdr)))
#define FREEBLOCK_NUMBER(sbaHdr)       (sbaHdr->freeBlock)
#define CACHE(sbaHdr)                  ((FTreeBlock*) ((u8*) (sbaHdr) - (SBAHDR_OFFSET - (sbaHdr)->offset)))
#define END_OF_CACHE(sbaHdr)           ((FTreeBlock*) ((u8*) CACHE(sbaHdr) + (sbaHdr)->size))
#define GET_FREEBLOCK(sbaHdr)          (CACHE(sbaHdr) + (sbaHdr)->freeBlock)
#define MAX_FREEBLOCK(sbaHdr)          ((sbaHdr)->size / sizeof(FTreeBlock))
#define GET_BLOCK_HEAD(treeHdr /* EPTreeHdr or FTreeHdr */)        ((FTreeBlock*) ((u8*) &((treeHdr)->sbaHdr) - SBAHDR_OFFSET))

#define FTREEHDR_OFFSET                (BLOCK_SIZE - sizeof(FTreeHdr))
#define GET_FTREEHDR(pBlock)           ((FTreeHdr*) ((u8*) (pBlock) + FTREEHDR_OFFSET))
#define GET_SUBBLOCK(ftreeHdr, offset) ((FTreeBlock*) ((u8*) GET_BLOCK_HEAD(ftreeHdr) + (offset)))

#ifdef _DEBUG
// #define FTREE_RUNTIME_CHECK // for debug.
#endif // _DEBUG


#ifdef FTREE_RUNTIME_CHECK
#define WFSKRN_FTREE_RUNTIME_CHECK(x)   {                               \
                                            if (FBADebugVerbose)        \
                                            {                           \
                                                WFSKRN_FTREE_ASSERT(x); \
                                            }                           \
                                        }
#else
#define WFSKRN_FTREE_RUNTIME_CHECK(x)
#endif

#ifdef _DEBUG
#ifdef WIN32
#define WFSKRN_FTREE_ASSERT(x)  assert(x)
#else  // Win32
#define WFSKRN_FTREE_ASSERT(x)  ASSERT(x)
#endif // WIN32
#else
#define WFSKRN_FTREE_ASSERT(x)
#endif

typedef struct _FTreeExtent
{
    u32 start;
    u32 len; // the unit is 8kB.
} FTreeExtent;

FTreeHdr* FTreeBlockInit(AreaInfo* area, void* block /* 8kB block */);

WFSKrnResult FTreeFind(FTreeHdr* header, FTreeItr* iter, u32 nKey, WFSBool strictSearch);

WFSKrnResult FTreeInsert(AreaInfo* area, FTreeHdr* header, FTreeItr* iter, u32 nKey, u32 nData);
WFSKrnResult FTreeDelete(FTreeHdr* header, FTreeItr* iter, u32 nKey);
WFSKrnResult FTreeModify(FTreeHdr* header, FTreeItr* iter, u32 originalKey, u32 modKey, u32 modData);

WFSKrnResult FTreeFirst(FTreeHdr* header, FTreeItr* iter);
WFSKrnResult FTreeNext(FTreeHdr* header, FTreeItr* iter);
WFSKrnResult FTreeLast(FTreeHdr* header, FTreeItr* iter);
WFSKrnResult FTreePrev(FTreeHdr* header, FTreeItr* iter);
WFSKrnResult PTreeLast(PTreeHdr *pPTreeHdr, PTreeItr *pPti, PTreeAllocator *pAllocator);
WFSKrnResult PTreePrev(PTreeHdr *pPTreeHdr, PTreeItr *pPti, PTreeAllocator *pAllocator);
PTreeHdr*    FTreeSelectSubTree(FTreeHdr* header, WFSBlkAdr unitLen);

u16 FTreeAllocEntry(void *pBase);
WFSKrnResult FTreeAllocEntries(void *pBase, u16 *pNodeOfs, u32 nNumNodes);
void FTreeFreeEntry(void *pBase, u16 nOfs);

FTreeBlock* MemoryBlockAlloc(SubBlockAllocatorHeader* header, int numBlocks);
int MemoryBlockFree(SubBlockAllocatorHeader* header, FTreeBlock* block, u32 numBlocks);

void FTreeSbaInit(AreaInfo* area, void* block, u32 footerSize, u32 offset);

WFSBool VerifyMemoryBlockFTree(SubBlockAllocatorHeader* header); // [check] debug API.
WFSBool FTreeVerifyTree(FTreeHdr* header);

#endif // BTREE_BASED_ALLOCATOR
#endif // __WFSKRN_FTREE_H__


