/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfskrn_Handles.h - Module for file handles and search directory handles


  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_Handles.h,v $
  Revision 1.6  2008/10/07 06:01:16  ueno
  Modified WFSKrnFileInfoAlloc() to set WFSKRN_HANDLE_ALLOCATED in order to indicate the handle is allocated.

  Revision 1.5  2008/07/17 22:27:36  paul
  Removed nRootBlkAdr from WFSKrnSearchDirInfo. (Now in PathCacheItr)

  Revision 1.4  2008/06/09 18:12:18  paul
  Made changes to support PathCache. Handles now contain a path cache iterator.

  Revision 1.3  2008/05/16 18:53:34  paul
  Added nRootBlkAdr to WFSKrnSearchDirInfo

  Revision 1.2  2008/05/10 04:00:48  kondo_masahiro
  Merged and fixed code to accept RM compiler

  Revision 1.1  2008/04/24 23:21:13  kondo_masahiro
  Initial check-in

  Revision 1.1  2008/04/23 00:27:45  paul
  Copied and modified from wfscli_Handles.h

*---------------------------------------------------------------------------*/

#ifndef wfskrn_Handles_h
#define wfskrn_Handles_h

#ifdef _DEBUG
    #define _DEBUG_WFSKRN_HANDLES 1
#else
    #define _DEBUG_WFSKRN_HANDLES 0
#endif

#include "wfskrn_Defs.h"
#include "wfskrn_Mutex.h"
#include "wfskrn_PathCache.h"


#define WFSKRN_MAX_FILE_HANDLES               (1<<WFSKRN_FILE_HANDLE_IDX_BITS)
#define WFSKRN_MAX_SEARCH_DIR_HANDLES         (1<<WFSKRN_SEARCH_DIR_HANDLE_IDX_BITS)
#define WFSKRN_LOG2_MAX_NUM_HANDLES           WFS_NUM_BITS_REQUIRED(WFSKRN_MAX_FILE_HANDLES + WFSKRN_MAX_SEARCH_DIR_HANDLES)

#define WFSKRN_FILE_HANDLE_INC                (1<<WFSKRN_FILE_HANDLE_IDX_BITS)
#define WFSKRN_SEARCH_DIR_HANDLE_INC          (1<<WFSKRN_SEARCH_DIR_HANDLE_IDX_BITS)


#define WFSKRN_HANDLE_MASK                    0xFFFF
#define WFSKRN_HANDLE_COUNTER_MASK            0xFFFF
#define WFSKRN_HANDLE_ALLOCATED               0xFFFF

// Internal binary format of WFSFileHandle (u32)
//
// 0TTcccccccccCCCCssssssssssSSSSSS
//
//  0 = always 0                          (Number of bits = 1                               )
//  T = type of handle: 00 for FileHandle (Number of bits = 2                               ) 
//  c = incrementally assigned by client  (Number of bits = 13 - WFSCLI_FILE_HANDLE_IDX_BITS)
//  C = idx of aCliFileInfo array         (Number of bits =      WFSCLI_FILE_HANDLE_IDX_BITS)
//  s = incrementally assigned by server  (Number of bits = 16 - WFSKRN_FILE_HANDLE_IDX_BITS)
//  S = idx of aSrvFileInfo array         (Number of bits =      WFSKRN_FILE_HANDLE_IDX_BITS)

typedef struct {
    u32                 nNext;
    WFSKrnMutex         mxAccessCliFileInfo;
    WFSSrvFileHandle    nHandle;
    DirEntryAttrHdr     attr;
    u8                  nAccess;
    PathCacheItr        pci;
    DirFilePos          fpos;
#if _DEBUG
    WFSPathName         pathDebug;
#endif
} WFSKrnFileInfo;


// Internal binary format of WFSSearcDirHandle (u32)
//
// 0TTcccccccccccCCssssssssssssSSSS
//
//  0 = always 0                               (Number of bits = 1                                     )
//  T = type of handle: 01 for SearchDirHandle (Number of bits = 2                                     ) 
//  c = incrementally assigned by client       (Number of bits = 13 - WFSCLI_SEARCH_DIR_HANDLE_IDX_BITS)
//  C = idx of aCliSearchDirInfo array         (Number of bits =      WFSCLI_SEARCH_DIR_HANDLE_IDX_BITS)
//  s = incrementally assigned by server       (Number of bits = 16 - WFSKRN_SEARCH_DIR_HANDLE_IDX_BITS)
//  S = idx of aSrvSearchDirInfo array         (Number of bits =      WFSKRN_SEARCH_DIR_HANDLE_IDX_BITS)

typedef struct {
    u32                         nNext;
    WFSSrvSearchDirectoryHandle nHandle;
    WFSFileName                 pattern;
    PathCacheItr                pci;
#if _DEBUG
    WFSPathName                 pathDebug;
#endif
} WFSKrnSearchDirInfo;


typedef struct {
    WFSSrvFileHandle            nFileHandleCounter;         // Counter for ssss field of file handle.
    WFSSrvSearchDirectoryHandle nSearchDirHandleCounter;    // Counter for ssss field of search directory handle.
    u32                         nFirstFreeFileHandle;       // Index of the first free element of aClientFileInfo
    u32                         nFirstFreeSearchDirHandle;  // Index of the first free element of aClientSearchDirInfo
    u32                         nNumUsedFileHandles;
    u32                         nNumUsedSearchDirHandles;
    WFSKrnMutex                 mxAccessTheFileInfo;
    WFSKrnMutex                 mxAccessTheSearchDirInfo;
    WFSKrnFileInfo              aFileInfo[WFSKRN_MAX_FILE_HANDLES];
    WFSKrnSearchDirInfo         aSearchDirInfo[WFSKRN_MAX_SEARCH_DIR_HANDLES];
} WFSKrnHandles;


void         WFSKrnInitHandles();

WFSKrnResult WFSKrnGetFileInfo(WFSKrnFileInfo **ppFi, WFSSrvFileHandle fh);
WFSKrnResult WFSKrnGetSearchDirInfo(WFSKrnSearchDirInfo **ppSdi, WFSSrvSearchDirectoryHandle ssdh);

WFSKrnResult WFSKrnFileInfoAlloc(WFSKrnFileInfo **ppFi);
WFSKrnResult WFSKrnSearchDirInfoAlloc(WFSKrnSearchDirInfo **ppSdi);

void         WFSKrnFileInfoFree(WFSKrnFileInfo *pFi);
void         WFSKrnSearchDirInfoFree(WFSKrnSearchDirInfo *pSdi);

#endif //define wfskrn_Handles_h
