/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfskrn_PathCache.h - Caches recently used path names plus root path names "/dev", "/vol"

  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_PathCache.h,v $
  Revision 1.13  2008/12/04 00:42:37  ooizumi
  Changed definition name to enable permission.

  Revision 1.12  2008/10/31 09:51:42  ueno
  Modified WFSSrvAPIs to return WFSKRN_RESULT_RESOURCE_LIMIT_EXCEEDED when path cache is short of memory.

  Revision 1.11  2008/10/07 10:30:01  ooizumi
  Defined PathCachePerm.
  Added nObjCreator to PathCacheData.

  Revision 1.10  2008/07/30 21:25:01  paul
  Removed some unused definitions

  Revision 1.9  2008/07/30 06:04:34  ooizumi
  Added definitions for permission.

  Revision 1.8  2008/07/21 22:48:35  paul
  Added PathCacheCheck()

  Revision 1.7  2008/07/17 22:31:12  paul
  Added nRootBlkAdr to PathCacheItr.

  Revision 1.6  2008/07/09 01:46:06  paul
  Minor change.

  Revision 1.5  2008/06/23 13:44:04  nakanose_jin
  add cast for avoid warning

  Revision 1.4  2008/06/09 18:06:42  paul
  Rewrote PathCache to use a radix tree. Now actually provides PathCache functionality.
  Also stores a list of open file / directory search paths so that invalid WFSSrv calls can be detected.

  Revision 1.3  2008/05/14 02:14:54  paul
  Added some comments

  Revision 1.2  2008/05/12 19:22:12  paul
  Made naming more consistent

  Revision 1.1  2008/05/10 04:00:29  kondo_masahiro
  Initial check-in

*---------------------------------------------------------------------------*/

#ifndef wfskrn_PathCache_h
#define wfskrn_PathCache_h

#include "wfskrn_Config.h"
#include "wfskrn_DirRxTree.h"

#ifdef _DEBUG
    #define _DEBUG_PATH_CACHE 0  // change to 1 if debug info required
#else
    #define _DEBUG_PATH_CACHE 0
#endif

#if _CHECK_RXT
    #define _CHECK_PATH_CACHE 1
#else
    #define _CHECK_PATH_CACHE 0
#endif

#define WFS_DEVICE_DIRECTORY "/dev"
#define WFS_VOLUME_DIRECTORY "/vol"


#define PATH_CACHE_MAX_ENTRIES       ((1<<WFSKRN_FILE_HANDLE_IDX_BITS) + (1<<WFSKRN_SEARCH_DIR_HANDLE_IDX_BITS) + 20 + WFSKRN_MAX_VOLUMES)

#define PATH_CACHE_ENABLED            1

#define PATH_CACHE_FLAG_OPEN_FOR_WRITE  0x8000


typedef struct {
    struct PathCacheEntry_ *pNext;
    struct PathCacheEntry_ *pPrev;
} PathCacheLink;

#ifdef PERMISSION_ENABLED
#define PATH_CACHE_PERM_MAX 2
typedef struct {
    WFSTitleId nTitleId;
    WFSTitleId nEntryCreatorId;
    u32        nFlags;
    u32        nEntryCreatorCl;
} PathCachePerm;
#endif

typedef struct {
    union {
        void        *pData;
        AreaInfo    *pAreaInfo;
    };
    u32              nBlkAdr;
    u16              nOpenHandleCount;  // This is calculated as the sum of nPinCount for all sub-paths currently in the cache
    u16              nPinCount;
    u16              nFlags;
    u16              nStrLen;
#ifdef PERMISSION_ENABLED
    u32              nAccessListIdx;
    WFSTitleId       nObjCreator;
//    PathCachePerm    aPerm[PATH_CACHE_PERM_MAX];
#endif
} PathCacheData;

typedef struct PathCacheEntry_ {
    RxtAttrHdr       ah;
    PathCacheLink    ml;
    PathCacheData    pcd;
} PathCacheEntry;

#define PATH_CACHE_LOG2_NODE_SIZE  WFS_NUM_BITS_REQUIRED(sizeof(PathCacheEntry))


typedef struct {
    u8               aRxtBuf[1<<WFS_LOG2_PATH_CACHE_SIZE];       // WFS_LOG2_PATH_CACHE_SIZE <= RXT_MAX_LOG2_SIZE
    RxtHdr          *pRxtHdr;
    PathCacheLink    mruAnchor;
    PathCacheEntry  *pMruAnchor;    // List from Most Recently Used to Least Recently Used
    PathCacheLink    pinAnchor;
    PathCacheEntry  *pPinAnchor;    // List of directories that are pinned - do not get automatically deleted to make room for other
#if _DEBUG_PATH_CACHE
    u32              nNumPinned;
#endif
} PathCache;


typedef struct {
    RxtItr       rxi;
    DirItr       di;
    WFSBlkAdr    nRootBlkAdr;
    const utf8  *pNameEnd;
    const utf8  *pRawNameStart;
    PathCacheEntry *pEntry;
} PathCacheItr;


#if _DEBUG
    void PathCacheDebugOutput();
#endif

PathCacheEntry *PathCacheInsertAfterCursorEntry3(PathCacheEntry *pCursorEntry, utf8 *sStr, u32 nStrLen);

static inline PathCacheEntry *PathCacheInsertAfterCursorEntry2(PathCacheEntry *pCursorEntry, utf8 *sStr) {
    u32 nStrLen = (u32)STRNLEN(sStr, WFS_MAX_PATH_NAME_SIZE);
    return PathCacheInsertAfterCursorEntry3(pCursorEntry, sStr, nStrLen);
}

PathCacheEntry *PathCacheFind3(RxtItr *pRxi, u32 nStrLen, void **ppData);

static inline PathCacheEntry *PathCacheFind2(RxtItr *pRxi, void **ppData) {
    u32 nStrLen = (u32)STRNLEN(pRxi->pStrPtr, WFS_MAX_PATH_NAME_SIZE);
    return PathCacheFind3(pRxi, nStrLen, ppData);
}

static inline PathCacheEntry *PathCacheFind1(RxtItr *pRxi) {
    void *pData;
    return PathCacheFind2(pRxi, &pData);
}

PathCacheEntry *PathCacheIncrementalFindEntry(RxtItr *pRxi, u32 nStrLen, void **ppData);

void PathCacheDeleteEntry(PathCacheEntry *pEntry);
WFSKrnResult PathCacheDeleteSubTree(PathCacheEntry *pEntry);
WFSKrnResult PathCacheDeleteLru();

PathCacheEntry *PathCacheIncrementalInsertAfterCursorEntry(PathCacheEntry *pCursorEntry, RxtItr *pRxi, u32 nPrefixLen, u32 nSuffixLen);
WFSKrnResult PathCacheIncrementalInsertPinnedEntry(PathCacheItr *pPci, u32 nStrLen);
#if _DEBUG_PATH_CACHE
    WFSKrnResult PathCacheIncreaseEntryOpenHandleCount(PathCacheItr *pPci, WFSAccess nDesiredAccess, utf8 *sStr);
    WFSKrnResult PathCacheDecreaseEntryOpenHandleCount(PathCacheItr *pPci, utf8 *sStr);
    void PathCachePinEntry(PathCacheEntry *pEntry, utf8 *sStr);
#else
    WFSKrnResult PathCacheIncreaseEntryOpenHandleCount(PathCacheItr *pPci, WFSAccess nDesiredAccess);
    WFSKrnResult PathCacheDecreaseEntryOpenHandleCount(PathCacheItr *pPci);
    void PathCachePinEntry(PathCacheEntry *pEntry);
#endif

#if _CHECK_PATH_CACHE
    void PathCacheCheck();
#endif


void PathCacheInit();
void PathCacheTest();

WFSKrnResult PathCacheFindParentDir(const WFSPathName *pAbsPathName, utf8 *sLowerCaseStr, PathCacheItr *pPci);
WFSKrnResult PathCacheFindPath     (const WFSPathName *pAbsPathName, utf8 *sLowerCaseStr, PathCacheItr *pPci);

#endif //define wfskrn_PathCache_h
