/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfskrn_Types.h - global types used by the WFS kernel

  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_Types.h,v $
  Revision 1.12  2008/10/24 09:03:09  kondo_masahiro
  Added decrease block process of DirResize.

  Revision 1.11  2008/10/17 08:51:27  kondo_masahiro
  Added the new MEDIUM size category (40KB-320KB)

  Revision 1.10  2008/10/10 02:11:31  kondo_masahiro
  Fixed a bug to access large files

  Revision 1.9  2008/10/08 23:20:09  kondo_masahiro
  Added codes to access large size file data

  Revision 1.8  2008/09/28 23:30:56  kondo_masahiro
  Fixed codes to use medium size user block.

  Revision 1.7  2008/08/06 02:24:17  paul
  Added WFSFileBlkPtr and DIR_ENTRY_ATTR* defs

  Revision 1.6  2008/07/28 22:30:56  paul
  Moved defs related to update counter here

  Revision 1.5  2008/07/25 08:47:53  ooizumi
  Added WFSKrnAccess.

  Revision 1.4  2008/07/17 05:15:08  kondo_masahiro
  Fixed codes for porting IOP.

  Revision 1.3  2008/07/09 01:52:00  paul
  Updated definitions for meta-data flags.

  Revision 1.2  2008/05/10 04:00:48  kondo_masahiro
  Merged and fixed code to accept RM compiler

  Revision 1.1  2008/04/22 04:39:08  kondo_masahiro
  Fixed code to transfer the wfskrn project to RM_SDK tree

  Revision 1.5  2008/04/26 01:25:56  wayne.wong
  Added magic number values.

  Revision 1.4  2008/04/05 22:43:54  paul
  Added WFSMetaDataFlags to make wfskrn_Volume.h compile

  Revision 1.3  2008/02/21 00:12:44  paul
  Added WFSMetaDataHdr

  Revision 1.2  2007/12/27 11:33:28  paul
  added various types

  Revision 1.1  2007/11/05 19:03:13  paul
  first checkin

*---------------------------------------------------------------------------*/

#ifndef wfskrn_Types_h
#define wfskrn_Types_h

#include "wfskrn_Defs.h"

// Magic numbers for WFS block types.  Should be the first u32(4B) of a meta-data block (post decryption)
typedef enum {  // Flags to identify persistent block types
    WFS_MDF_NULL             = 0x00000000,

    WFS_MDF_DIR_BLK          = 0x80000000, // Directory Block
    WFS_MDF_DIR_ROOT_BLK     = 0x40000000, // Always combined with WFS_MDF_DIR_BLK or WFS_MDF_FREE_LIST
    WFS_MDF_DIR_LEAF_BLK     = 0x20000000, // Always combined with WFS_MDF_DIR_BLK or WFS_MDF_FREE_LIST

    WFS_MDF_INDIRECT_BLK     = 0x40000000, // Indirect block. (Block of pointers to user data blocks).
    WFS_MDF_FREE_LIST_BLK    = 0x20000000, // Free List Management (B-tree) block
    WFS_MDF_VOLUME_HDR       = 0x10000000, // Master Record. Not combined with anything else
    WFS_MDF_AREA_HDR         = 0x08000000, // Area Header. Not combined with anything else

    WFS_MDF_MAX              = 0xffffffff
} WFSMetaDataFlags;


#define WFSKRN_MAX_FILE_HANDLES               (1<<WFSKRN_FILE_HANDLE_IDX_BITS)
#define WFSKRN_MAX_SEARCH_DIR_HANDLES         (1<<WFSKRN_SEARCH_DIR_HANDLE_IDX_BITS)
#define WFSKRN_LOG2_MAX_NUM_HANDLES           WFS_NUM_BITS_REQUIRED(WFSKRN_MAX_FILE_HANDLES + WFSKRN_MAX_SEARCH_DIR_HANDLES)

#define WFSKRN_FILE_HANDLE_INC                (1<<WFSKRN_FILE_HANDLE_IDX_BITS)
#define WFSKRN_SEARCH_DIR_HANDLE_INC          (1<<WFSKRN_SEARCH_DIR_HANDLE_IDX_BITS)

#define DIR_BLK_REF_COUNT_INC                 (1<<(32-WFSKRN_LOG2_MAX_NUM_HANDLES))
#define WFS_MDF_DIR_COUNTER_MASK              (0x1FFFFFFF & (DIR_BLK_REF_COUNT_INC-1)) // counter for DIR blocks - to detect updates

typedef u32 WFSTransIdx;    // Transaction Index
typedef u32 WFSVolIdx;      // Volume Index

typedef u32 WFSBlkAdr;      // Block Address
typedef u32 WFSLglBlkAdr;   // Logical Block Address
typedef u32 WFSPhyBlkAdr;   // Physical Block Address
typedef u16 WFSIntraBlkOfs; // Intra Block Offset (for offsets within meta-data blocks)
typedef WFSTitleId WFSUserId;

typedef u32 WFSHashCode[5]; // SHA-1 hash code per block

#pragma pack(1)
typedef struct {
    // This is the format for the start of all metadata blocks
    WFSMetaDataFlags nFlags;    // Contains a few bits for type identification, and the rest update-counter bits
    WFSHashCode      hash;
} WFSMetaDataHdr;
#pragma pack()

#pragma pack(1)
typedef struct {
    // This is the format for pointers from meta-data to user-data blocks. (Note: not needed for pointers to meta-data blocks)
    WFSBlkAdr    nBlkAdr;
    WFSHashCode  hash;
} WFSFileBlkPtr;
#pragma pack()

#pragma pack(1)
typedef struct {
    // This is the format for pointers from meta-data to user-data blocks. (Note: not needed for pointers to meta-data blocks)
    WFSBlkAdr    nBlkAdr;
    WFSHashCode  aHash[1<<(WFS_LOG2_LARGE_BLK_SIZE-WFS_LOG2_MAX_HASHABLE_BLK_SIZE)];
} WFSFileLargeBlkPtr;
#pragma pack()

#pragma pack(1)
typedef struct {
    WFSBlkAdr    nBlkAdr;
    WFSHashCode  *pHash;
} WFSFileBlkInfo;
#pragma pack()

#define DIR_ENTRY_ATTR_MAX_SMALL_SIZE                    0xA000   // ( 40KB) : Max size = 0x50000  (320KB)
#define DIR_ENTRY_ATTR_MAX_MEDIUM_SIZE                   0x50000  // (320KB) : Max size = 0x280000 (2.5MB)
#define DIR_ENTRY_ATTR_MAX_LARGE_SIZE                    0x600000 // (  6MB) : Max size = 0x600000 (  6MB)

#define DIR_ENTRY_ATTR_MAX_SMALL_SIZE_DECREASE           0x2000   // (  8KB)
#define DIR_ENTRY_ATTR_MAX_MEDIUM_SIZE_DECREASE          0x10000  // ( 64KB)
#define DIR_ENTRY_ATTR_MAX_LARGE_SIZE_DECREASE           0x80000  // (512KB)

#define DIR_ENTRY_ATTR_MAX_HDR_SIZE                      ((size_t)(&((DirEntryAttrHdr*)0)->aCaseBitArray) + (WFS_MAX_FILE_NAME_SIZE>>3)) // 28+31 = 59byte
#define DIR_ENTRY_ATTR_MAX_PTR_ARY_SIZE                  ((1<<SBA_MAX_LOG2_SUB_BLK_SIZE) - DIR_ENTRY_ATTR_MAX_HDR_SIZE)

#define DIR_ENTRY_ATTR_MAX_SMALL_BLK_PTRS                (DIR_ENTRY_ATTR_MAX_SMALL_SIZE  >> WFS_LOG2_SMALL_BLK_SIZE)
#define DIR_ENTRY_ATTR_MAX_MEDIUM_BLK_PTRS               (DIR_ENTRY_ATTR_MAX_MEDIUM_SIZE >> WFS_LOG2_MEDIUM_BLK_SIZE)
#define DIR_ENTRY_ATTR_MAX_LARGE_BLK_PTRS                (DIR_ENTRY_ATTR_MAX_LARGE_SIZE  >> WFS_LOG2_LARGE_BLK_SIZE)

#define DIR_ENTRY_ATTR_MAX_SMALL_BLK_PTRS_DECREASE       (DIR_ENTRY_ATTR_MAX_SMALL_SIZE_DECREASE  >> WFS_LOG2_SMALL_BLK_SIZE)
#define DIR_ENTRY_ATTR_MAX_MEDIUM_BLK_PTRS_DECREASE      (DIR_ENTRY_ATTR_MAX_MEDIUM_SIZE_DECREASE >> WFS_LOG2_MEDIUM_BLK_SIZE)
#define DIR_ENTRY_ATTR_MAX_LARGE_BLK_PTRS_DECREASE       (DIR_ENTRY_ATTR_MAX_LARGE_SIZE_DECREASE  >> WFS_LOG2_LARGE_BLK_SIZE)

// Internal WFSAccess value
typedef enum {
    WFS_ACCESS_CHANGE_SIZE = 0x04
} WFSKrnAccess;

#endif //define wfskrn_Types_h
