/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfs_AsyncUtils.cpp


  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfs_AsyncUtils.cpp,v $
  Revision 1.42  2008/12/03 00:17:05  kondo_masahiro
  Fixed case WFSSRV_FLUSH.

  Revision 1.41  2008/10/31 05:37:09  ooizumi
  Fixed a bug in WFSExecFunction().

  Revision 1.40  2008/10/23 11:15:28  nakanose_jin
  Open and Search should give a invalid file handle if he fails to execute.

  Revision 1.39  2008/10/21 10:45:20  kondo_masahiro
  Added new function WFSWriteFileAndDiscard

  Revision 1.38  2008/07/18 02:59:46  ueno
  Renamed WFSDeletePermissions() -> WFSDeleteAccessListEntry().

  Revision 1.37  2008/07/14 07:15:24  ueno
  Implemented WFSGetAccessListAsync(), WFSSetAccessListEntryAsync() and WFSDeletePermissionsAsync().

  Revision 1.36  2008/04/24 04:26:29  nakanose_jin
  byebye nPermissions param

  Revision 1.35  2008/04/18 07:45:03  nakanose_jin
  merge acl

  Revision 1.34  2008/04/10 13:35:38  ueno
  Added WFSSetPermissionsForUser[Async] and WFSGetPermissionsForUser[Async].

  Revision 1.33  2008/02/25 08:33:20  paul
  BOOL -> WFSBool

  Revision 1.32  2008/02/07 13:02:21  nakanose_jin
  WFSUnmountVolume : Add force-close option (but only in detach condition)

  Revision 1.31  2008/01/23 00:44:30  ueno
  Fixed to copy volume ID only when the result is WFS_RESULT_OK.

  Revision 1.30  2008/01/23 00:07:02  paul
  Fixed WFSGetVolumeIdAsync to only overwrite buffer when result==WFS_RESULT_OK

  Revision 1.29  2008/01/10 09:23:26  paul
  bugfix for WFSSRV_SET_CURRENT_DIRECTORY, WFSSRV_SET_HOME_DIRECTORY

  Revision 1.28  2008/01/10 05:33:23  nakanose_jin
  validatePathName => validateDirectory

  Revision 1.27  2008/01/08 09:49:46  ueno
  Modified WFSExecFunction() to support WFSSRV_MOUNT_DEVICE.

  Revision 1.26  2007/12/27 06:02:37  paul
  bugfix for WFSSRV_SET_HOME_DIRECTORY, WFSSRV_SET_CURRENT_DIRECTORY

  Revision 1.25  2007/12/27 01:43:19  ueno
  Added WFSIsThreadTerminated() and WFSJoinThread().

  Revision 1.24  2007/12/26 05:43:45  ueno
  Fixed WFSAsyncUtilsInit() to create a thread.
  Fixed WFSAsyncUtilsClose() to send signal to wake up async thread.

  Revision 1.23  2007/12/25 14:03:50  ueno
  Modified to initialize mutex, thread and heap only once.

  Revision 1.22  2007/12/12 19:57:07  paul
  Changed WFSSearchDirectoryFirstAsync to not create a search directory handle, and pass sdh=0 to the callback if WFSSrvSearchDirectoryFirst returns an error code.

  Revision 1.21  2007/12/12 07:52:10  paul
  Added WFSSRV_GET_FILE_SIZE

  Revision 1.20  2007/12/11 00:10:06  paul
  Fixed a bug in WFSWriteFileAsync

  Revision 1.19  2007/12/10 02:48:49  paul
  Fixed WFSSetCurrentDirectoryAsync, WFSSetHomeDirectoryAsync

  Revision 1.18  2007/11/21 04:17:45  wayne.wong
  Many changes that included single request/result RPC messages, bug fixes, code factoring, and English C comment adjustments.
    1. Combined RPC messages into a single request message and single result
        message. Except for file reads and writes.
    2. Fixed+factored some Endianess code.
    3. Enabled larger TCP/IP data transfer size. Previous setting was artifically small
        to test the code to perform large transfers.
    4. Changed the volume id parameters utf8*->WFSVolumeId for SrvGetVolumeId
        and SrvMountDevice. More consistent with the way we send this type at the
        RPC layer.
    5. Changed a couple Japanese comments to be compatible with English
        comments. In particular, the "//" comment with a backslash  at the end of
        the line causes the next line of code to be ignored.
    6. Removed Windows-specific debugging output defines for RVL build.

  Revision 1.17  2007/11/16 21:55:08  paul
  Changed WFSSrvCloseFile to pass nNewSize instead of bTruncateFlag

  Revision 1.16  2007/11/16 02:06:40  paul
  Added WFSFlush()

  Revision 1.15  2007/11/09 01:17:29  paul
  Added sub-directory validation check for WFSMoveLocal

  Revision 1.14  2007/11/02 01:57:40  paul
  bugfix

  Revision 1.13  2007/11/02 00:52:21  paul
  now logging changes
  added nStartPosition to WFSSrvReadFile, WFSSrvWriteFile

---------------------------------------------------------------------------*/

#include "wfs_Client.h"

#undef dbg
#if _DEBUG_ASYNC_UTILS
#define dbg(s) s
#else
#define dbg(s)
#endif

static BOOL bMutexInitialized;

void WFSExecFunction(WFSAsyncParamHdr *pParamHdr);

void WFSExecFunction(WFSAsyncParamHdr *pParamHdr) {
    dbg(MyOSReport("Processing Function: %d\n", pParamHdr->nFunction));
    void *pPtr = (void*)((size_t)pParamHdr + sizeof(WFSAsyncParamHdr));
    WFSResult nResult = WFS_RESULT_OK;
    switch(pParamHdr->nFunction) {
        case WFSSRV_EXIT: {
            //WFSAsyncUtilsClose();
            ((WFSResultCallback)pParamHdr->cb)(nResult, pParamHdr->pUserData);
            WFSExitThread(NULL);
            break;
        }
        case WFSSRV_GET_DEVICE_INFO: {
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM_THEN_INC_PTR(pPtr, utf8 *, sDeviceName);
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM_THEN_INC_PTR(pPtr, WFSDeviceInfo *, pDi);
            WFSDeviceName *pDevName = (WFSDeviceName *)pPtr;
            nResult = WFSSrvGetDeviceInfo(pDevName, pDi);
            ((WFSDeviceInfoCallback)pParamHdr->cb)(nResult, sDeviceName, pDi, pParamHdr->pUserData);
            break;
        }
        case WFSSRV_MOUNT_DEVICE: {
            WFSVolumeId volId;
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM_THEN_INC_PTR(pPtr, const utf8 *, sDeviceName);
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM_THEN_INC_PTR(pPtr, utf8 *, sVolumeIdBuffer);
            WFSDeviceName *pDevName = (WFSDeviceName *)pPtr;
            nResult = WFSSrvMountDevice(pDevName, &volId);
            if (nResult == WFS_RESULT_OK)
            {
                strncpy(sVolumeIdBuffer, volId.sStr, WFS_MAX_VOLUME_ID_SIZE+1);
            }
            ((WFSDeviceVolumeCallback)pParamHdr->cb)(nResult, sDeviceName, sVolumeIdBuffer, pParamHdr->pUserData);
            break;
        }
        case WFSSRV_UNMOUNT_VOLUME: {
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM_THEN_INC_PTR(pPtr, const utf8 *, sVolumeId);
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM_THEN_INC_PTR(pPtr, WFSBool, bForce);
            WFSVolumeId *pVolumeId = (WFSVolumeId *)pPtr;
            nResult = WFSSrvUnmountVolume(pVolumeId, bForce);
            ((WFSUnmountVolumeCallback)pParamHdr->cb)(nResult, sVolumeId, pParamHdr->pUserData);
            break;
        }
        case WFSSRV_INITIALIZE_DEVICE: {
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM_THEN_INC_PTR(pPtr, const utf8 *, sDeviceName);
            WFSDeviceName *pDevName = (WFSDeviceName *)pPtr;
            nResult = WFSSrvInitializeDevice(pDevName);
            ((WFSInitializeDeviceCallback)pParamHdr->cb)(nResult, sDeviceName, pParamHdr->pUserData);
            break;
        }
        case WFSSRV_GET_VOLUME_ID: {
            WFSVolumeId volId;
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM_THEN_INC_PTR(pPtr, const utf8 *, sDeviceName);
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM_THEN_INC_PTR(pPtr, utf8 *, sVolumeIdBuffer);
            WFSDeviceName *pDevName = (WFSDeviceName *)pPtr;
            nResult = WFSSrvGetVolumeId(pDevName, &volId);
            if (nResult == WFS_RESULT_OK) {
                strncpy(sVolumeIdBuffer, volId.sStr, WFS_MAX_VOLUME_ID_SIZE+1);
            }
            ((WFSDeviceVolumeCallback)pParamHdr->cb)(nResult, sDeviceName, sVolumeIdBuffer, pParamHdr->pUserData);
            break;
        }
        case WFSSRV_GET_FREE_SPACE_SIZE: {
            WFSPathName *pAbsPathName = (WFSPathName *)pPtr;
            u64 nSize;
            nResult = WFSSrvGetFreeSpaceSize(pAbsPathName, &nSize);
            ((WFSGetFreeSpaceCallback)pParamHdr->cb)(nResult, nSize, pParamHdr->pUserData);
            break;
        }
        case WFSSRV_SET_HOME_DIRECTORY: {
            WFSPathName *pAbsPathName = (WFSPathName *)pPtr;
            nResult = WFSSrvValidateDirectory(pAbsPathName);
            if (nResult == WFS_RESULT_OK) {
                WFSLockMutex(&wcg.pg.mxAccessPathGlobals);
                if (pAbsPathName->sStr[pAbsPathName->nLen-1]!='/') {
                    pAbsPathName->sStr[pAbsPathName->nLen++] ='/';
                    pAbsPathName->sStr[pAbsPathName->nLen]   =0;
                }
                WFSPathResolve(&wcg.pg.pathHome.pn, &wcg.pg.pathHome.ppoa, pAbsPathName->sStr, PATH_TYPE_ABS_DIR_ONLY);
                WFSUnlockMutex(&wcg.pg.mxAccessPathGlobals);
            }
            ((WFSResultCallback)pParamHdr->cb)(nResult, pParamHdr->pUserData);
            break;
        }
        case WFSSRV_SET_CURRENT_DIRECTORY: {
            WFSPathName *pAbsPathName = (WFSPathName *)pPtr;
            nResult = WFSSrvValidateDirectory(pAbsPathName);
            if (nResult == WFS_RESULT_OK) {
                WFSLockMutex(&wcg.pg.mxAccessPathGlobals);
                if (pAbsPathName->sStr[pAbsPathName->nLen-1]!='/') {
                    pAbsPathName->sStr[pAbsPathName->nLen++] ='/';
                    pAbsPathName->sStr[pAbsPathName->nLen]   =0;
                }
                WFSPathResolve(&wcg.pg.pathCurrent.pn, &wcg.pg.pathCurrent.ppoa, pAbsPathName->sStr, PATH_TYPE_ABS_DIR_ONLY);
                WFSUnlockMutex(&wcg.pg.mxAccessPathGlobals);
            }
            ((WFSResultCallback)pParamHdr->cb)(nResult, pParamHdr->pUserData);
            break;
        }
        case WFSSRV_VALIDATE_PATH_NAME: {
            WFSPathName *pAbsPathName = (WFSPathName *)pPtr;
            nResult = WFSSrvValidateDirectory(pAbsPathName);
            ((WFSResultCallback)pParamHdr->cb)(nResult, pParamHdr->pUserData);
            break;
        }
        case WFSSRV_CREATE_DIRECTORY: {
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM_THEN_INC_PTR(pPtr, u32, nFlags);
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM_THEN_INC_PTR(pPtr, u64, nQuota);
            WFSPathName *pAbsPathName = (WFSPathName *)pPtr;
            nResult = WFSSrvCreateDirectory(pAbsPathName, nFlags, nQuota);
            ((WFSResultCallback)pParamHdr->cb)(nResult, pParamHdr->pUserData);
            break;
        }
        case WFSSRV_GET_ATTRIBUTES: {
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM_THEN_INC_PTR(pPtr, WFSFileAttributes *, pFa);
            WFSPathName *pAbsPathName = (WFSPathName *)pPtr;
            nResult = WFSSrvGetAttributes(pAbsPathName, pFa);
            ((WFSGetAttributesCallback)pParamHdr->cb)(nResult, pFa, pParamHdr->pUserData);
            break;
        }
        case WFSSRV_CHANGE_PERMISSIONS: {
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM_THEN_INC_PTR(pPtr, u32, nFlags);
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM_THEN_INC_PTR(pPtr, u32, nMask);
            WFSPathName *pAbsPathName = (WFSPathName *)pPtr;
            nResult = WFSSrvChangePermissions(pAbsPathName, nFlags, nMask);
            ((WFSResultCallback)pParamHdr->cb)(nResult, pParamHdr->pUserData);
            break;
        }
        case WFSSRV_DELETE: {
            WFSPathName *pAbsPathName = (WFSPathName *)pPtr;
            nResult = WFSSrvDelete(pAbsPathName);
            ((WFSResultCallback)pParamHdr->cb)(nResult, pParamHdr->pUserData);
            break;
        }
        case WFSSRV_DELETE_ACL_ENTRY: {
            WFSPathName *pAbsPathName;

            // restore arguments.
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM_THEN_INC_PTR(pPtr, WFSTitleId, titleId);
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM_THEN_INC_PTR(pPtr, WFSTitleId, entryCreatorId);
            pAbsPathName = (WFSPathName *)pPtr;

            nResult = WFSSrvDeleteAccessListEntry(pAbsPathName, titleId, entryCreatorId);
            ((WFSResultCallback)pParamHdr->cb)(nResult, pParamHdr->pUserData);
            break;
        }
        case WFSSRV_SET_ACL_ENTRY: {
            WFSPathName *pAbsPathName;

            // restore arguments.
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM_THEN_INC_PTR(pPtr, u32, nFlags);
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM_THEN_INC_PTR(pPtr, u32, nMask);
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM_THEN_INC_PTR(pPtr, WFSTitleId, titleId);
            pAbsPathName = (WFSPathName *)pPtr;
            nResult = WFSSrvSetAccessListEntry(pAbsPathName, nFlags, nMask, titleId);

            ((WFSResultCallback)pParamHdr->cb)(nResult, pParamHdr->pUserData);
            break;
        }
        case WFSSRV_GET_ACL: {
            WFSPathName *pAbsPathName;
            s32 nResultOrNumAclEntries;

            // restore arguments.
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM_THEN_INC_PTR(pPtr, WFSAccessListEntry*, pAccessList);
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM_THEN_INC_PTR(pPtr, u32, nMaxElements);
            pAbsPathName = (WFSPathName *)pPtr;
            nResultOrNumAclEntries = WFSSrvGetAccessList(pAbsPathName, pAccessList, nMaxElements);

            ((WFSGetAccessListCallback)pParamHdr->cb)(nResultOrNumAclEntries, pParamHdr->pUserData);
            break;
        }
        case WFSSRV_MOVE_LOCAL: {
            WFSPathName *pSrcAbsPathName = (WFSPathName *)pPtr;
            WFSPathName *pDstAbsPathName = (WFSPathName *)((size_t)pPtr + pSrcAbsPathName->nLen + sizeof(WFSPathName) - WFS_MAX_PATH_NAME_SIZE - 1);
            nResult = WFSSrvMoveLocal(pSrcAbsPathName, pDstAbsPathName);
            ((WFSResultCallback)pParamHdr->cb)(nResult, pParamHdr->pUserData);
            break;
        }
        case WFSSRV_CREATE_AND_OPEN_FILE: {
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM_THEN_INC_PTR(pPtr, u32, nFlags);
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM_THEN_INC_PTR(pPtr, WFSContentIdx, nCidx);
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM_THEN_INC_PTR(pPtr, u32, nPreAllocateSize);
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM_THEN_INC_PTR(pPtr, WFSCliFileInfo *, pFi);
            WFSSrvFileHandle sfh;
            WFSPathName *pAbsPathName = (WFSPathName *)pPtr;
            nResult = WFSSrvCreateAndOpenFile(pAbsPathName, nFlags, nCidx, nPreAllocateSize, &sfh);
            WFSFileHandle fh;
            if (nResult == WFS_RESULT_OK) {
                fh = pFi->nHandle |= sfh;
                pFi->fa.nFileSize = pFi->nPosition = 0;
                pFi->fa.nFlags       = nFlags;
                pFi->fa.nCidx = nCidx;
                pFi->nAccess = WFS_ACCESS_RW;
            } else {
                WFSCliFileInfoFree(pFi);
                fh = NULL;
            }
            ((WFSFileOpCallback)pParamHdr->cb)(nResult, fh, pParamHdr->pUserData);
            break;
        }
        case WFSSRV_OPEN_FILE: {
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM_THEN_INC_PTR(pPtr, WFSAccess, nDesiredAccess);
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM_THEN_INC_PTR(pPtr, WFSCliFileInfo *, pFi);
            WFSSrvFileHandle sfh;
            WFSPathName *pAbsPathName = (WFSPathName *)pPtr;
            nResult = WFSSrvOpenFile(pAbsPathName, nDesiredAccess, &sfh, &pFi->fa);
            WFSFileHandle fh;
            if (nResult == WFS_RESULT_OK) {
                fh = pFi->nHandle |= sfh;
                pFi->nPosition = 0;
                pFi->nAccess = nDesiredAccess;
            } else {
                WFSCliFileInfoFree(pFi);                // fh = NULL;
                fh = WFS_INVALID_HANDLE_VALUE;
            }
            ((WFSFileOpCallback)pParamHdr->cb)(nResult, fh, pParamHdr->pUserData);
            break;
        }
        case WFSSRV_GET_FILE_SIZE: {
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM(pPtr, WFSFileHandle, fh);
            WFSSrvFileHandle sfh = (WFSSrvFileHandle)fh;
            WFSFileSize nFileSize;
            nResult = WFSSrvGetFileSize(sfh, &nFileSize);
            ((WFSGetFileSizeCallback)pParamHdr->cb)(nResult, nFileSize, fh, pParamHdr->pUserData);
            break;
        }
        case WFSSRV_READ_FILE: {
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM_THEN_INC_PTR(pPtr, WFSFileHandle, fh);
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM_THEN_INC_PTR(pPtr, void *, pFileDataBuffer);
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM_THEN_INC_PTR(pPtr, WFSFileSize, nStartPosition);
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM(pPtr, s32, nSize);
            WFSSrvFileHandle sfh = (WFSSrvFileHandle)fh;
            s32 nResultOrNumBytesRead = WFSSrvReadFile(sfh, pFileDataBuffer, nStartPosition, nSize);
#if 0       // This code has been disabled, because we are no longer updating file position after async read/write
            if (nResultOrNumBytesRead >= 0) {
                WFSCliFileInfo *pFi;
                WFSGetCliFileInfo(&pFi, fh);
                pFi->nPosition += nResultOrNumBytesRead;
                if (pFi->fa.nFileSize < pFi->nPosition) {
                    pFi->fa.nFileSize = pFi->nPosition;
                }
            }
#endif
            ((WFSReadFileCallback)pParamHdr->cb)(nResultOrNumBytesRead, fh, pParamHdr->pUserData);
            break;
        }
        case WFSSRV_WRITE_FILE: {
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM_THEN_INC_PTR(pPtr, WFSFileHandle, fh);
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM_THEN_INC_PTR(pPtr, const void *, pFileData);
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM_THEN_INC_PTR(pPtr, WFSFileSize, nStartPosition);
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM(pPtr, WFSFileSize, nSize);
            WFSSrvFileHandle sfh = (WFSSrvFileHandle)fh;
            nResult = WFSSrvWriteFile(sfh, pFileData, nStartPosition, nSize, FALSE);
            if (nResult == WFS_RESULT_OK) {
                WFSCliFileInfo *pFi;
                WFSGetCliFileInfo(&pFi, fh);
                u32 nEndPosition = nStartPosition + nSize;
                if (pFi->fa.nFileSize < nEndPosition) {
                    pFi->fa.nFileSize = nEndPosition;
                }
            }
            ((WFSFileOpCallback)pParamHdr->cb)(nResult, fh, pParamHdr->pUserData);
            break;
        }
        case WFSSRV_WRITE_FILE_AND_DISCARD: {
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM_THEN_INC_PTR(pPtr, WFSFileHandle, fh);
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM_THEN_INC_PTR(pPtr, const void *, pFileData);
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM_THEN_INC_PTR(pPtr, WFSFileSize, nStartPosition);
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM(pPtr, WFSFileSize, nSize);
            WFSSrvFileHandle sfh = (WFSSrvFileHandle)fh;
            nResult = WFSSrvWriteFile(sfh, pFileData, nStartPosition, nSize, TRUE);
            if (nResult == WFS_RESULT_OK) {
                WFSCliFileInfo *pFi;
                WFSGetCliFileInfo(&pFi, fh);
                u32 nEndPosition = nStartPosition + nSize;
                if (pFi->fa.nFileSize < nEndPosition) {
                    pFi->fa.nFileSize = nEndPosition;
                }
            }
            ((WFSFileOpCallback)pParamHdr->cb)(nResult, fh, pParamHdr->pUserData);
            break;
        }
        case WFSSRV_CLOSE_FILE: {
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM_THEN_INC_PTR(pPtr, WFSFileHandle, fh);
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM(pPtr, WFSFileSize, nNewSize);
            WFSSrvFileHandle sfh = (WFSSrvFileHandle)fh;
            nResult = WFSSrvCloseFile(sfh, nNewSize);
            if (nResult == WFS_RESULT_OK) {
                WFSCliFileInfo *pFi;
                WFSGetCliFileInfo(&pFi, fh);
                WFSCliFileInfoFree(pFi);
            }
            ((WFSFileOpCallback)pParamHdr->cb)(nResult, fh, pParamHdr->pUserData);
            break;
        }
        case WFSSRV_FLUSH: {
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM_THEN_INC_PTR(pPtr, const utf8 *, sVolumeId);
            WFSVolumeId *pVolumeId = (WFSVolumeId *)pPtr;
            nResult = WFSSrvFlush(pVolumeId);
            ((WFSResultCallback)pParamHdr->cb)(nResult, pParamHdr->pUserData);
            break;
        }
        case WFSSRV_SEARCH_DIRECTORY_FIRST: {
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM_THEN_INC_PTR(pPtr, WFSCliSearchDirInfo *, pSdi);
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM_THEN_INC_PTR(pPtr, WFSFileInfo *, pFi);
            WFSPathName *pAbsPattern = (WFSPathName *)pPtr;
            WFSSrvSearchDirectoryHandle ssdh;
            nResult = WFSSrvSearchDirectoryFirst(pAbsPattern, &ssdh, pFi);
            pSdi->nHandle |= ssdh;
            WFSFileHandle fh = pSdi->nHandle;
            if (nResult != WFS_RESULT_OK) {
                WFSCliSearchDirInfoFree(pSdi);
                fh = WFS_INVALID_HANDLE_VALUE;                // fh = 0; 
            }
            ((WFSSearchDirectoryCallback)pParamHdr->cb)(nResult, fh, pFi, pParamHdr->pUserData);
            break;
        }
        case WFSSRV_SEARCH_DIRECTORY_NEXT: {
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM_THEN_INC_PTR(pPtr, WFSSearchDirectoryHandle, sdh);
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM(pPtr, WFSFileInfo *, pFi);
            WFSSrvSearchDirectoryHandle ssdh = (WFSSrvSearchDirectoryHandle)sdh;
            nResult = WFSSrvSearchDirectoryNext(ssdh, pFi);
            ((WFSSearchDirectoryCallback)pParamHdr->cb)(nResult, sdh, pFi, pParamHdr->pUserData);
            break;
        }
        case WFSSRV_SEARCH_DIRECTORY_CLOSE: {
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM_THEN_INC_PTR(pPtr, WFSSearchDirectoryHandle, sdh);
            WFSSrvSearchDirectoryHandle ssdh = (WFSSrvSearchDirectoryHandle)sdh;
            nResult = WFSSrvSearchDirectoryClose(ssdh);
            if (nResult == WFS_RESULT_OK) {
                WFSCliSearchDirInfo *pSdi;
                WFSGetCliSearchDirInfo(&pSdi, sdh);
                WFSCliSearchDirInfoFree(pSdi);
            }
            ((WFSSearchDirectoryCloseCallback)pParamHdr->cb)(nResult, sdh, pParamHdr->pUserData);
            break;
        }
        case WFSSRV_GET_PERMISSIONS: {
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM_THEN_INC_PTR(pPtr, WFSTitleId, titleId);
            WFSPathName *pAbsPathName = (WFSPathName *)pPtr;
            u32 nFlags;
            nResult = WFSSrvGetPermissionForUser(pAbsPathName, &nFlags, titleId);
            ((WFSGetPermissionsForUserCallback)pParamHdr->cb)(nResult, nFlags, pParamHdr->pUserData);
            break;
        }
        case WFSSRV_SET_PERMISSIONS: {
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM_THEN_INC_PTR(pPtr, u32, nFlags);
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM_THEN_INC_PTR(pPtr, u32, nMask);
            WFSCLI_DECLARE_AND_READ_ASYNC_PARAM_THEN_INC_PTR(pPtr, WFSTitleId, titleId);
            WFSPathName *pAbsPathName = (WFSPathName *)pPtr;
            nResult = WFSSrvSetPermissionForUser(pAbsPathName, nFlags, nMask, titleId);
            ((WFSResultCallback)pParamHdr->cb)(nResult, pParamHdr->pUserData);
            break;
        }
    }
}

#if _WIN32
DWORD WINAPI WFSAsyncExecThread(LPVOID pParam);
#elif _RVL
void *WFSAsyncExecThread(void *pParam);
#elif _TWL
void *WFSAsyncExecThread(void *pParam);
#endif

#if _WIN32
DWORD WINAPI WFSAsyncExecThread(LPVOID pParam)
#elif _RVL
void *WFSAsyncExecThread(void *pParam)
#elif _TWL
void *WFSAsyncExecThread(void *pParam)
#endif
{
#pragma unused(pParam)

    (void)osEnableInterrupts();
    dbg(MyOSReport("Starting WFSAsyncExecThread()\n"));
    while(1) {
        WFSLockMutex(&wcg.ag.mxAccessTheParamQueue);
        while (wcg.ag.anchor.pHead == (WFSAsyncParamHdr*)&wcg.ag.anchor) {
            if (wcg.bLibInitialized==false) {
                WFSSignalCond(&wcg.ag.condParamQueueEmpty);
                return NULL;
            }
            WFSWaitCond(&wcg.ag.condParamQueueNotEmpty, &wcg.ag.mxAccessTheParamQueue);
        }
        WFSResetCond(&wcg.ag.condParamQueueNotEmpty);
        WFSAsyncParamHdr *pParamHdr = wcg.ag.anchor.pHead;
        wcg.ag.anchor.pHead = pParamHdr->l.pNext;
        pParamHdr->l.pNext->l.pPrev = (WFSAsyncParamHdr*)&wcg.ag.anchor;
        WFSUnlockMutex(&wcg.ag.mxAccessTheParamQueue);
        WFSExecFunction(pParamHdr);
        WFSHeapFree(&wcg.heap, pParamHdr);
    }
}

void WFSAsyncUtilsInit() {
    wcg.ag.anchor.pHead = wcg.ag.anchor.pTail = (WFSAsyncParamHdr*)&wcg.ag.anchor;
    if (!bMutexInitialized)
    {
        bMutexInitialized = true;
        WFSInitMutex(&wcg.ag.mxAccessTheParamQueue);
        WFSInitCond(&wcg.ag.condParamQueueNotEmpty);
    }
    wcg.ag.pAsyncExecStack = WFSHeapAlloc(&wcg.heap, WFS_ASYNC_EXEC_STACK_SIZE);
    WFSCreateThread(&wcg.ag.asyncExecThread, WFSAsyncExecThread, 0, wcg.ag.pAsyncExecStack, WFS_ASYNC_EXEC_STACK_SIZE, 15);
    WFSResumeThread(&wcg.ag.asyncExecThread);
}

void WFSAsyncUtilsClose() {
    WFSLockMutex(&wcg.ag.mxAccessTheParamQueue);
    WFSSignalCond(&wcg.ag.condParamQueueNotEmpty); // [check]
    if (wcg.ag.anchor.pHead != wcg.ag.anchor.pTail) {
        WFSWaitCond(&wcg.ag.condParamQueueEmpty, &wcg.ag.mxAccessTheParamQueue);
    }
    WFSUnlockMutex(&wcg.ag.mxAccessTheParamQueue);
    WFSJoinThread(&wcg.ag.asyncExecThread, NULL);
    // Assumes all workspace memory will be deleted, so does not bother to free space
    // ToDo: blocking wait for last async to finish
}

static void WfsPrepareAsyncCallCommon(void **ppPtr, u8 nFunction, void *pUserData, void *cb, size_t nOtherParamSize, u32 nSize, WFSAsyncParamHdr *pParamHdr) {
    pParamHdr->nSize = nSize;
    pParamHdr->nFunction = nFunction;
    pParamHdr->cb = cb;
    pParamHdr->pUserData = pUserData;
    WFSLockMutex(&wcg.ag.mxAccessTheParamQueue);
    if (wcg.ag.anchor.pHead == (WFSAsyncParamHdr*)&wcg.ag.anchor) {
        // ParamQueue was previously empty, but will now have an entry added.
        // It is OK to signal in advance of actually adding the entry, becuase
        // the async execution thread won't be able to proceed until we unlock the mutex.
        WFSSignalCond(&wcg.ag.condParamQueueNotEmpty);
    }
    pParamHdr->l.pNext = (WFSAsyncParamHdr*)&wcg.ag.anchor;
    pParamHdr->l.pPrev = wcg.ag.anchor.pTail;
    wcg.ag.anchor.pTail->l.pNext = pParamHdr;
    wcg.ag.anchor.pTail = pParamHdr;
    if (nOtherParamSize==0) {
        WFSUnlockMutex(&wcg.ag.mxAccessTheParamQueue);
    }
    if (ppPtr) {
        *ppPtr = (void*)((size_t)pParamHdr + sizeof(WFSAsyncParamHdr));
    }
}

WFSResult WFSPrepareAsyncCall(void **ppPtr, u8 nFunction, void *pUserData, void *cb, size_t nOtherParamSize) {
    // This function allocates memory for storing away parameters of asynchronous function calls which do not include path names.
    // If any errors occur, it releases the allocated memory and returns with an error code.
    // If no errors occur, the memory remains allocated, and a parameter header entry is added to the async execution ParamQueue.
    // If nOtherParamSize is 0, the ParamQueue item is complete, and the access mutex is unlocked.
    //    otherwise, it is assumed that the calling function will fill out the remaining elements and will unlock the mutex.
    size_t nTotalAllocSize = sizeof(WFSAsyncParamHdr) + nOtherParamSize;
    WFSAsyncParamHdr *pParamHdr = (WFSAsyncParamHdr *)WFSHeapAlloc(&wcg.heap, nTotalAllocSize);
    if (pParamHdr == 0) {
        return WFS_RESULT_OUT_OF_MEMORY;
    }
    WfsPrepareAsyncCallCommon(ppPtr, nFunction, pUserData, cb, nOtherParamSize, nTotalAllocSize, pParamHdr);
    return WFS_RESULT_OK;
}

WFSResult WFSPrepareAsyncCallWithPathName(void **ppPtr, const utf8 *sInPathName,
    WFSPathNameType nPathType, u8 nFunction, void *pUserData, void *cb, size_t nOtherParamSize)
{
    // This function allocates memory for storing away parameters of asynchronous function calls, including a pathname parameter.
    // It resolves, stores, and validates the specified pathname.
    // If any errors occur, it releases the allocated memory and returns with an error code.
    // If no errors occur, the memory remains allocated, and a parameter header entry is added to the async execution ParamQueue.
    // If nOtherParamSize is 0, the ParamQueue item is complete, and the access mutex is unlocked.
    //    otherwise, it is assumed that the calling function will fill out the remaining elements and will unlock the mutex.
    size_t nOffset = sizeof(WFSAsyncParamHdr) + nOtherParamSize;
    size_t nTotalAllocSize = nOffset + sizeof(WFSPathName);
    WFSAsyncParamHdr *pParamHdr = (WFSAsyncParamHdr *)WFSHeapAlloc(&wcg.heap, nTotalAllocSize);
    if (pParamHdr == 0) {
        return WFS_RESULT_OUT_OF_MEMORY;
    }
    WFSPathPartOffsetArray ppoa;
    WFSPathName *pPnAbs = (WFSPathName *)((u8*)pParamHdr + nOffset);
    WFSResult nResult = WFSPathResolve(pPnAbs, &ppoa, sInPathName, nPathType);
    if (nResult != WFS_RESULT_OK) {
        WFSHeapFree(&wcg.heap, pParamHdr);
        return nResult;
    }
    WfsPrepareAsyncCallCommon(ppPtr, nFunction, pUserData, cb, nOtherParamSize, nTotalAllocSize, pParamHdr);
    return WFS_RESULT_OK;
}

WFSResult WFSPrepareMoveLocalAsyncCall(void **ppPtr, const utf8 *sSrcPathName, const utf8 *sDstPathName,
    WFSPathNameType nPathType, u8 nFunction, void *pUserData, void *cb, size_t nOtherParamSize)
{
    // This function allocates memory for storing away parameters of asynchronous function calls, including a pathname parameter.
    // It resolves, stores, and validates the specified pathname.
    // If any errors occur, it releases the allocated memory and returns with an error code.
    // If no errors occur, the memory remains allocated, and a parameter header entry is added to the async execution ParamQueue.
    // If nOtherParamSize is 0, the ParamQueue item is complete, and the access mutex is unlocked.
    //    otherwise, it is assumed that the calling function will fill out the remaining elements and will unlock the mutex.
    size_t nOffset = sizeof(WFSAsyncParamHdr) + nOtherParamSize;
    size_t nTotalAllocSize = nOffset + (2 * sizeof(WFSPathName));
    WFSAsyncParamHdr *pParamHdr = (WFSAsyncParamHdr *)WFSHeapAlloc(&wcg.heap, nTotalAllocSize);
    if (pParamHdr == 0) {
        return WFS_RESULT_OUT_OF_MEMORY;
    }
    WFSPathPartOffsetArray ppoa;
    WFSPathName *pPnSrcAbs = (WFSPathName *)((u8*)pParamHdr + nOffset);
    WFSResult nResult = WFSPathResolve(pPnSrcAbs, &ppoa, sSrcPathName, nPathType);
    if (nResult != WFS_RESULT_OK) {
        goto Error;
    }
    nOffset += pPnSrcAbs->nLen + sizeof(WFSPathName) - WFS_MAX_PATH_NAME_SIZE - 1;
    WFSPathName *pPnDstAbs = (WFSPathName *)((u8*)pParamHdr + nOffset);
    nResult = WFSPathResolve(pPnDstAbs, &ppoa, sDstPathName, nPathType);
    if (nResult != WFS_RESULT_OK) {
        goto Error;
    }
    if (strncmp(pPnSrcAbs->sStr, pPnDstAbs->sStr, pPnSrcAbs->nLen)==0) {
        if (pPnDstAbs->sStr[pPnSrcAbs->nLen]=='/') {
            // trying to move a directory to it's own sub-directory
            nResult = WFS_RESULT_INVALID;
            goto Error;
        }
    }
    nOffset += pPnDstAbs->nLen + sizeof(WFSPathName) - WFS_MAX_PATH_NAME_SIZE - 1;
    WfsPrepareAsyncCallCommon(ppPtr, nFunction, pUserData, cb, nOtherParamSize, nOffset, pParamHdr);
    return WFS_RESULT_OK;
Error:
    WFSHeapFree(&wcg.heap, pParamHdr);
    return nResult;
}

WFSResult WFSPrepareAsyncCallWithDeviceName(void **ppPtr, const utf8 *sDeviceName,
    u8 nFunction, void *pUserData, void *cb, size_t nOtherParamSize)
{
    size_t nDeviceNameOffset = sizeof(WFSAsyncParamHdr) + sizeof(const utf8 *) + nOtherParamSize;
    size_t nTotalAllocSize = nDeviceNameOffset + sizeof(WFSDeviceName);
    WFSAsyncParamHdr *pParamHdr = (WFSAsyncParamHdr *)WFSHeapAlloc(&wcg.heap, nTotalAllocSize);
    if (pParamHdr == 0) {
        return WFS_RESULT_OUT_OF_MEMORY;
    }
    void *pPtr = (void*)((size_t)pParamHdr + sizeof(WFSAsyncParamHdr));
    WFSCLI_WRITE_ASYNC_PARAM(pPtr, const utf8 *, sDeviceName);
    WFSDeviceName *pDeviceName = (WFSDeviceName *)((u8*)pParamHdr + nDeviceNameOffset);
    WFSResult nResult = WFSSetDeviceName(pDeviceName, sDeviceName);
    if (nResult != WFS_RESULT_OK) {
        WFSHeapFree(&wcg.heap, pParamHdr);
        return nResult;
    }
    WfsPrepareAsyncCallCommon(ppPtr, nFunction, pUserData, cb, nOtherParamSize, nTotalAllocSize, pParamHdr);
    if (ppPtr) {
        *ppPtr = (void*)((size_t)*ppPtr + sizeof(const utf8 *));
    }
    return WFS_RESULT_OK;
}

WFSResult WFSPrepareAsyncCallWithVolumeId(void **ppPtr, const utf8 *sVolumeId,
    u8 nFunction, void *pUserData, void *cb, size_t nOtherParamSize)
{
    size_t nVolumeIdOffset = sizeof(WFSAsyncParamHdr) + sizeof(const utf8 *) + nOtherParamSize;
    size_t nTotalAllocSize = nVolumeIdOffset + sizeof(WFSVolumeId);
    WFSAsyncParamHdr *pParamHdr = (WFSAsyncParamHdr *)WFSHeapAlloc(&wcg.heap, nTotalAllocSize);
    if (pParamHdr == 0) {
        return WFS_RESULT_OUT_OF_MEMORY;
    }
    void *pPtr = (void*)((size_t)pParamHdr + sizeof(WFSAsyncParamHdr));
    WFSCLI_WRITE_ASYNC_PARAM(pPtr, const utf8 *, sVolumeId);
    WFSVolumeId *pVolumeId = (WFSVolumeId *)((u8*)pParamHdr + nVolumeIdOffset);
    WFSResult nResult = WFSSetVolumeId(pVolumeId, sVolumeId);
    if (nResult != WFS_RESULT_OK) {
        WFSHeapFree(&wcg.heap, pParamHdr);
        return nResult;
    }
    WfsPrepareAsyncCallCommon(ppPtr, nFunction, pUserData, cb, nOtherParamSize, nTotalAllocSize, pParamHdr);
    if (ppPtr) {
        *ppPtr = (void*)((size_t)*ppPtr + sizeof(const utf8 *));
    }
    return WFS_RESULT_OK;
}


