/*---------------------------------------------------------------------------*
  Project:  Revolution WFS library
  File:     wfs_Debug.cpp - implementation of debugging routines for WFS 
            implementation

  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfs_Debug.cpp,v $
  Revision 1.14  2008/12/10 06:55:50  kondo_masahiro
  Added the new device types WFS_DEVTYPE_USB_MSC_??.

  Revision 1.13  2008/10/23 11:15:28  nakanose_jin
  Open and Search should give a invalid file handle if he fails to execute.

  Revision 1.12  2008/07/15 08:43:57  nakanose_jin
  change arg order

  Revision 1.11  2008/06/25 01:00:08  nakanose_jin
  add control level error

  Revision 1.10  2008/04/21 04:39:42  nakanose_jin
  merging acl

  Revision 1.9  2008/04/18 07:45:03  nakanose_jin
  merge acl

  Revision 1.8  2008/03/26 01:18:02  nakanose_jin
  move wfs_Debug.h from local to private folder.

  Revision 1.7  2008/02/08 09:52:05  paul
  Added WFS_RESULT_DEV_IN_USE

  Revision 1.6  2008/02/07 13:02:21  nakanose_jin
  WFSUnmountVolume : Add force-close option (but only in detach condition)

  Revision 1.5  2007/12/26 10:25:51  paul
  Changed to use WFS_RANGE_CHK, WFS_INDEX_COUNT from wfs_defs.h, instead of linking jut_utility.cpp

  Revision 1.4  2007/12/17 13:29:17  nakanose_jin
  add debug utility function

  Revision 1.3  2007/11/02 00:55:00  paul
  now logging changes

*---------------------------------------------------------------------------*/

#include "wfs_Defs.h"
#include "private/wfs_Debug.h"

// Print to stdout for now via printf's

void WFSDevTypePrint(WFSDevType devType) {
    switch(devType) {
    case WFS_DEVTYPE_NONE:
        printf("WFS_DEVTYPE_NONE");
        break;
    case WFS_DEVTYPE_SD:
        printf("WFS_DEVTYPE_SD");
        break;
    case WFS_DEVTYPE_USB_MSC_10:
        printf("WFS_DEVTYPE_USB_MSC_10");
        break;
    case WFS_DEVTYPE_USB_MSC_11:
        printf("WFS_DEVTYPE_USB_MSC_11");
        break;
    case WFS_DEVTYPE_USB_MSC_20:
        printf("WFS_DEVTYPE_USB_MSC_20");
        break;
    default:
        printf("unknown");
    }
}

void WFSDevInfoPrint(WFSDeviceInfo *pDevInfo) {
    printf("DevTotalCapacity (MB) = %I64u\n", pDevInfo->nDevTotalCapacity/(1024*1024));
    printf("DevType = "); WFSDevTypePrint((WFSDevType)pDevInfo->nDevType); printf("\n");
    printf("DevFlags = 0x%1x\n", pDevInfo->nDevFlags);
}

void WFSFileAttrPrint(WFSFileAttributes *pFileAttr) {
    if (pFileAttr->nFlags & WFS_FLAG_IS_A_DIRECTORY)
        printf("NumEntries = %u\n", pFileAttr->nNumEntries);
    else
        printf("FileSize (MB) = %I64u\n", pFileAttr->nFileSize/(1024*1024));

    printf("Permissions = 0x%8x\n", pFileAttr->nFlags);
    // printf("ContentIdx = %u\n", pFileAttr->nNumEntries);
    printf("timeCreated = %u\n", pFileAttr->timeCreated);
    printf("timeUpdated = %u\n", pFileAttr->timeUpdated);
}

void WFSFileInfoPrint(WFSFileInfo *pFileInfo) {
    WFSFileAttrPrint(&pFileInfo->attr);
    printf("sFilename = %s\n", pFileInfo->sFileName);
}

static char *_result_strings[] = {
  "OK             ", 
  "BUSY           ", 
  "OUT_OF_MEMORY  ", 
  "INVALID        ", 
  "ACCESS         ", 
  "LIB_NOT_INITIAL", 
  "LIB_ALREADY_INI", 
  "FILE_TOO_BIG   ", 
  "NO_CHANGE_SIZE ", 
  "MEDIA_ERROR    ", 
  "DEV_UNUSABLE   ",
  "DEV_NOT_INITIAL", 
  "DEV_IN_USE     ", 
  "VOL_ID_ERROR   ", 
  "WRITE_PROTECTED", 
  "ALREADY_MOUNTED", 
  "PERMISSION     ",
  "PERMISSION_CL  ", 
  "ACL_FULL       ", 
  "ACL_NOT_FOUND  ",
  "AUTHENTICATION ", 
  "CORRUPTION     ", 
  "DIRECTORY_QUOTA", 
  "MAX_HANDLES    ", 
  "ALREADY_EXISTS ", 
  "NOT_FOUND      ", 
  "NOT_EMPTY      ", 
  "NOT_FILE       ", 
  "NOT_DIRECTORY  ", 
  "FILE_OPEN      ", 
  "LOCKED         ", 
  "RESOURCE_LIMIT_", 
  ""
  }; 

static char *_result_strings2[] = {
    "NOT_IMPLEMENTED",
    "UNKNOWN        ",       
    "FATAL_ERROR    "
    };

/*--------------------------------------------------------------------------*/
//!  \fn       iiFSDev_PrintResult(WFSResult result)            
//!  \par :                      
//!            
//!  \param    result             
//!  \return                         
/*--------------------------------------------------------------------------*/

const char * 
WFSDebug_GetResultString(WFSResult result)
{
  s32 idx = -1 * (s32)result;
  if (WFS_RANGE_CHK(idx , 0 , WFS_INDEX_COUNT(_result_strings)))
    {
      return _result_strings[idx];
    }
  else if (WFS_RANGE_CHK(idx , -WFS_RESULT_NOT_IMPLEMENTED , -WFS_RESULT_NOT_IMPLEMENTED+WFS_INDEX_COUNT(_result_strings2)) ) 
    {
      return _result_strings2[idx+WFS_RESULT_NOT_IMPLEMENTED];
    }
  else
    {
      return "???";
    }
}

