/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfs_Mutex.cpp


  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfs_Mutex.cpp,v $
  Revision 1.9  2008/04/19 05:53:50  kondo_masahiro
  Fixed code to accept RM compiler

  Revision 1.8  2007/12/27 01:43:19  ueno
  Added WFSIsThreadTerminated() and WFSJoinThread().

  Revision 1.7  2007/12/25 14:01:49  ueno
  Added WFSSuspendThread().

  Revision 1.6  2007/11/02 01:00:36  paul
  now logging changes

*---------------------------------------------------------------------------*/

#include "wfs_Mutex.h"

#undef dbg
#if _DEBUG_MUTEX
#define dbg(s) s
#else
#define dbg(s)
#endif

#if _WIN32

dbg(u32 aSignalCounter[0x2000]);

void WFSInitMutex(WFSMutex *pMx) {
    pMx->hHandle = CreateMutex(NULL, false, NULL);
    ReleaseMutex(pMx->hHandle);
    dbg(aSignalCounter[(u32)pMx->hHandle&0x1fff]=1;
        MyOSReport("InitMutex(%x:%d)\n", pMx->hHandle, aSignalCounter[(u32)pMx->hHandle&0x1fff]));
}

void WFSLockMutex(WFSMutex *pMx) {
    dbg(MyOSReport("Wait(%x:%d)\n", pMx->hHandle, aSignalCounter[(u32)pMx->hHandle&0x1fff]));
    WaitForSingleObject(pMx->hHandle, INFINITE);
    dbg(aSignalCounter[(u32)pMx->hHandle&0x1fff]--);
}

bool WFSTryLockMutex(WFSMutex *pMx) {
    return (WaitForSingleObject(pMx->hHandle, 0) != WAIT_TIMEOUT);
}

void WFSUnlockMutex(WFSMutex *pMx) {
    dbg(aSignalCounter[(u32)pMx->hHandle&0x1fff]++;
        MyOSReport("Release(%x:%d)\n", pMx->hHandle, aSignalCounter[(u32)pMx->hHandle&0x1fff]));
    ReleaseMutex(pMx->hHandle);
}

void WFSInitCond(WFSCond *pCond) {
    pCond->hAutoResetEvent = CreateEvent(NULL, false, false, NULL);
    dbg(aSignalCounter[(u32)pCond->hAutoResetEvent&0x1fff]=0;
        MyOSReport("InitCond(%x:%d)\n", pCond->hAutoResetEvent, aSignalCounter[(u32)pCond->hAutoResetEvent&0x1fff]));
    //ResetEvent(pCond->hAutoResetEvent);
    //pCond->hManualResetEvent = CreateEvent(NULL, true, false, NULL);
}

void WFSSignalCond(WFSCond *pCond) {
    //PulseEvent(pCond->hAutoResetEvent);
    dbg(aSignalCounter[(u32)pCond->hAutoResetEvent&0x1fff]++;
        MyOSReport("Signal(%x:%d)\n", pCond->hAutoResetEvent, aSignalCounter[(u32)pCond->hAutoResetEvent&0x1fff]));
    SetEvent(pCond->hAutoResetEvent);
}

void WFSWaitCond(WFSCond *pCond, WFSMutex* pMx) {
#if 1
    dbg(aSignalCounter[(u32)pMx->hHandle&0x1fff]++;
        MyOSReport("Signal(%x:%d) And Wait(%x:%d)\n", pMx->hHandle, aSignalCounter[(u32)pMx->hHandle&0x1fff], pCond->hAutoResetEvent, aSignalCounter[(u32)pCond->hAutoResetEvent&0x1fff]));
    DWORD nRet = SignalObjectAndWait(pMx->hHandle, pCond->hAutoResetEvent, INFINITE, false);
    dbg(aSignalCounter[(u32)pCond->hAutoResetEvent&0x1fff]--);
#else
    ReleaseMutex(pMx->hHandle);
    WaitForSingleObject(pCond->hAutoResetEvent, INFINITE);
#endif
    dbg(MyOSReport("Wait(%x:%d)\n", pMx->hHandle, aSignalCounter[(u32)pMx->hHandle&0x1fff]));
    WaitForSingleObject(pMx->hHandle, INFINITE);
    dbg(aSignalCounter[(u32)pMx->hHandle&0x1fff]--);
}

void WFSResetCond(WFSCond *pCond) {
    ResetEvent(pCond->hAutoResetEvent);
}

bool WFSCreateThread(WFSThread *pThread,
                     WFSThreadFunc threadFunc,
                     void*      param,
                     void*      stackBase,
                     u32        stackSize,
                     OSPriority priority)
{
    pThread->hThread = CreateThread(NULL, stackSize, threadFunc, param, CREATE_SUSPENDED, 0);
    SetThreadPriority(pThread->hThread, 16-priority);
    return true;
}

s32 WFSResumeThread(WFSThread *pThread) {
    return ResumeThread(pThread->hThread);
}

void WFSCancelThread(WFSThread *pThread) {
    // Note: On windows, this does not do any cleanup - should avoid using
    TerminateThread(pThread->hThread, 0);
}

void WFSExitThread(void *pVal) {
    ExitThread((DWORD)pVal);
}

s32 WFSSuspendThread(WFSThread *pThread)
{
    return SuspendThread(pThread->hThread);
}

bool WFSIsThreadTerminated(WFSThread *pThread)
{
    switch (WaitForSingleObject(pThread->hThread, 0))
    {
      case WAIT_TIMEOUT:
        return false;
      case WAIT_OBJECT_0:  // the thread has exited.
      case WAIT_ABANDONED: // [check]
      case WAIT_FAILED:    // the thread does not exist. [check] should check error code?
      default:
        break;
    }

    return true;
}

bool WFSJoinThread(WFSThread *pThread, void** val)
{
#pragma unused (val) // exit value is not supported.
    switch (WaitForSingleObject(pThread->hThread, INFINITE))
    {
      case WAIT_OBJECT_0:
        return true;
      case WAIT_TIMEOUT:
      case WAIT_ABANDONED:
      case WAIT_FAILED:
      default:
        break;
    }
    return false;
}


#elif _RVL


bool WFSCreateThread(WFSThread *pThread,
                     WFSThreadFunc threadFunc,
                     void*      param,
                     void*      stackBase,
                     u32        stackSize,
                     OSPriority priority)
{
    return OSCreateThread(pThread, threadFunc, param, (void*)((u32)stackBase+stackSize), stackSize, priority, 0); // joinable
}


//-- twl modified
#elif _TWL
typedef void (*MyThreadFunc)(void*);
bool WFSCreateThread(WFSThread *pThread,
                     WFSThreadFunc threadFunc,
                     void*      param,
                     void*      stackBase,
                     u32        stackSize,
                     OSPriority priority)
{
    osCreateThread( pThread, (MyThreadFunc)threadFunc, param, (void*)((u32)stackBase+stackSize), stackSize, priority);
    return true;
}

s32 WFSResumeThread(WFSThread *pThread) {
    osWakeupThreadDirect( pThread);
//    osTPrintf( "Need TWL modifies.\n");
    return 0;
}

void WFSCancelThread(WFSThread *pThread) {
    osDestroyThread( pThread);
}

void WFSExitThread(void *pVal) {
    osExitThread();
}

s32 WFSSuspendThread(WFSThread *pThread)
{
    osTPrintf( "Need TWL modifies.\n");
    return 0;
}

bool WFSIsThreadTerminated(WFSThread *pThread)
{
    return( osIsThreadTerminated( pThread));
}

bool WFSJoinThread(WFSThread *pThread, void** val)
{
#pragma unused (val) // exit value is not supported.
    osJoinThread( pThread);
    return true;
}

void WFSInitCond(WFSCond *pCond) {
    osInitEvent( pCond);
}

void WFSSignalCond(WFSCond *pCond) {
    osSignalEvent( pCond, 1);
}

void WFSWaitCond(WFSCond *pCond, WFSMutex* pMx) {
    osUnlockMutex( pMx);
    osWaitEventEx( pCond, 1, OS_EVENT_MODE_OR, 1); //Cxg҂ɃNA𔺂
}

void WFSResetCond(WFSCond *pCond) {
    osClearAllEvent( pCond);
    return;
}

//-- twl modified

#endif
