/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfs_PathNames.cpp - Functions for concatenating relative and absolute path names


  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfs_PathNames.cpp,v $
  Revision 1.23  2008/08/19 14:12:41  nakanose_jin
  bug : relative dir from long path current dir causes buffer-overflow

  Revision 1.22  2008/07/15 02:32:17  paul
  Fixed so that single "/" is still accepted

  Revision 1.21  2008/07/09 02:50:34  paul
  Changed to ensure absolute paths do not end with a '/'

  Revision 1.20  2008/04/28 18:37:29  paul
  Fixed a bug with parsing relative patterns which start with a wildcard character

  Revision 1.19  2008/04/19 05:53:44  kondo_masahiro
  Fixed code to accept RM compiler

  Revision 1.18  2008/02/29 07:21:55  paul
  The previous fix was not adequate. Should be fixed now.

  Revision 1.17  2008/02/29 02:07:16  paul
  Fixed a bug where pathnames which ended with a filename that was 1 byte too long were incorrectly accepted.

  Revision 1.16  2008/02/25 05:37:06  paul
  Fixed bug in pathname length check

  Revision 1.15  2008/02/13 05:47:08  paul
  Changed spec to make empty pathnames "" invalid. Previously it was treated as current directory.

  Revision 1.14  2008/02/07 06:32:06  paul
  fixed a problem with filenames such as ~File

  Revision 1.13  2008/02/04 05:44:00  paul
  Fixed a bug with checking the length of a directory entry

  Revision 1.12  2008/01/25 04:33:39  paul
  Fixed a bug with WFSPathResolve for directories relative to the home directory "~/dir"

  Revision 1.11  2008/01/17 10:18:11  paul
  Fixed a bug where characters >=0x80 were being treated as -ve indexes in the table lookup

  Revision 1.10  2007/12/25 14:03:49  ueno
  Modified to initialize mutex, thread and heap only once.

  Revision 1.9  2007/12/10 02:49:38  paul
  Added debug output

  Revision 1.8  2007/11/09 01:19:32  paul
  Added a new start state to allow validation of path name depth for paths ending in directories

  Revision 1.7  2007/11/06 22:43:12  paul
  Added validation checks for NULL pointers passed in

  Revision 1.6  2007/11/02 01:06:49  paul
  now logging changes

*---------------------------------------------------------------------------*/

// File created by Paul Donnelly on 2007.10.05

#include "wfs_Client.h"

#undef dbg
#if _DEBUG_PATHNAMES
#define dbg(s) s
#else
#define dbg(s)
#endif

// A Lookup table is used to categorize characters for the purposes of resolving path names.
// It maps a single utf-8 byte to one of the following categories:
#define CHAR_CATEGORY_NORMAL 0
#define CHAR_CATEGORY_DOT    1
#define CHAR_CATEGORY_SLASH  2
#define CHAR_CATEGORY_TILDA  3
#define CHAR_CATEGORY_NULL   4
#define CHAR_CATEGORY_BAD    5
#define NUM_CHAR_CATEGORIES  6

// Invalid Utf-8 bytes, control characters, and wildcards are not allowed in path names
utf8 aCharCategory[256] = {
    CHAR_CATEGORY_NULL, CHAR_CATEGORY_BAD, CHAR_CATEGORY_BAD, CHAR_CATEGORY_BAD,
    CHAR_CATEGORY_BAD,  CHAR_CATEGORY_BAD, CHAR_CATEGORY_BAD, CHAR_CATEGORY_BAD,
    CHAR_CATEGORY_BAD,  CHAR_CATEGORY_BAD, CHAR_CATEGORY_BAD, CHAR_CATEGORY_BAD,
    CHAR_CATEGORY_BAD,  CHAR_CATEGORY_BAD, CHAR_CATEGORY_BAD, CHAR_CATEGORY_BAD,

    CHAR_CATEGORY_BAD,  CHAR_CATEGORY_BAD, CHAR_CATEGORY_BAD, CHAR_CATEGORY_BAD,
    CHAR_CATEGORY_BAD,  CHAR_CATEGORY_BAD, CHAR_CATEGORY_BAD, CHAR_CATEGORY_BAD,
    CHAR_CATEGORY_BAD,  CHAR_CATEGORY_BAD, CHAR_CATEGORY_BAD, CHAR_CATEGORY_BAD,
    CHAR_CATEGORY_BAD,  CHAR_CATEGORY_BAD, CHAR_CATEGORY_BAD, CHAR_CATEGORY_BAD,

//     !  "  #  $  %  &  '    (  )         *           +  ,  -         .                   /
    0, 0, 0, 0, 0, 0, 0, 0,   0, 0, CHAR_CATEGORY_BAD, 0, 0, 0, CHAR_CATEGORY_DOT, CHAR_CATEGORY_SLASH,

//  0  1  2  3  4  5  6  7    8  9  :  ;  <  =  >         ?
    0, 0, 0, 0, 0, 0, 0, 0,   0, 0, 0, 0, 0, 0, 0, CHAR_CATEGORY_BAD,

//  @  A  B  C  D  E  F  G    H  I  J  K  L  M  N  O
    0, 0, 0, 0, 0, 0, 0, 0,   0, 0, 0, 0, 0, 0, 0, 0,

//  P  Q  R  S  T  U  V  W    X  Y  Z  [  \  ]  ^  _
    0, 0, 0, 0, 0, 0, 0, 0,   0, 0, 0, 0, 0, 0, 0, 0,

//  `  a  b  c  d  e  f  g    h  i  j  k  l  m  n  o
    0, 0, 0, 0, 0, 0, 0, 0,   0, 0, 0, 0, 0, 0, 0, 0,

//  p  q  r  s  t  u  v  w    x  y  z  {  |  }         ~                   DEL
    0, 0, 0, 0, 0, 0, 0, 0,   0, 0, 0, 0, 0, 0, CHAR_CATEGORY_TILDA, CHAR_CATEGORY_BAD,

// 0x80..0x8F  (128..143)
    0, 0, 0, 0, 0, 0, 0, 0,   0, 0, 0, 0, 0, 0, 0, 0,

// 0x90..0x9F (144..159)
    0, 0, 0, 0, 0, 0, 0, 0,   0, 0, 0, 0, 0, 0, 0, 0,

// 0xA0..0xAF (160..175)
    0, 0, 0, 0, 0, 0, 0, 0,   0, 0, 0, 0, 0, 0, 0, 0,

// 0xB0..0xBF (176..191)
    0, 0, 0, 0, 0, 0, 0, 0,   0, 0, 0, 0, 0, 0, 0, 0,

// 0xC0..0xCF (192..207)
    CHAR_CATEGORY_BAD, CHAR_CATEGORY_BAD, 0, 0, 0, 0, 0, 0,   0, 0, 0, 0, 0, 0, 0, 0,

// 0xD0..0xDF (208..223)
    0, 0, 0, 0, 0, 0, 0, 0,   0, 0, 0, 0, 0, 0, 0, 0,

// 0xE0..0xEF (224..239)
    0, 0, 0, 0, 0, 0, 0, 0,   0, 0, 0, 0, 0, 0, 0, 0,

// 0xF0..0xFF (240..255)
    0, 0, 0, 0, 0, CHAR_CATEGORY_BAD, CHAR_CATEGORY_BAD, CHAR_CATEGORY_BAD,
    CHAR_CATEGORY_BAD, CHAR_CATEGORY_BAD, CHAR_CATEGORY_BAD, CHAR_CATEGORY_BAD,
    CHAR_CATEGORY_BAD, CHAR_CATEGORY_BAD, CHAR_CATEGORY_BAD, CHAR_CATEGORY_BAD,
};

// Finite State Machine for resolving and normalizing path names
typedef enum {
    // Initial states - see wfs_PathNames.h
    STATE_START_REL      = PATH_TYPE_REL_OR_ABS,
    STATE_START_REL_DIR  = PATH_TYPE_REL_OR_ABS_DIR,
    STATE_START_ABS_DIR  = PATH_TYPE_ABS_DIR_ONLY,
    STATE_START_PTN      = PATH_TYPE_PATTERN,

    // Mid-parsing states
    STATE_START_DOT,
    STATE_START_SLASH,
    STATE_START_TILDA,
    STATE_START_NAME,
    STATE_START_WILD,
    STATE_TWO_DOTS,
    STATE_IGNORE,
    STATE_MID,
    STATE_DOT,
    STATE_TILDA,
    STATE_HOME,
    STATE_SLASH,
    STATE_PARENT,

    // for parsing search patterns including wild-cards
    STATE_CHK_WILD,
    STATE_WILD,

    // End states (Do not require an entry in the state table, since no further transition is possible)
    STATE_END,
    STATE_HOME_END,
    STATE_NAME_END,
    STATE_PARENT_END,
    STATE_ERROR
} WFSPathFsmState;


u8 aStateMachine[][NUM_CHAR_CATEGORIES] = {
    //    A-Z                  .                  /                    ~                  NULL                 * ?
    { STATE_START_NAME,   STATE_START_DOT,   STATE_START_SLASH,   STATE_START_TILDA,   STATE_ERROR,        STATE_ERROR    }, // STATE_START_REL
    { STATE_START_NAME,   STATE_START_DOT,   STATE_START_SLASH,   STATE_START_TILDA,   STATE_ERROR,        STATE_ERROR    }, // STATE_START_REL_DIR
    { STATE_ERROR,        STATE_ERROR,       STATE_START_SLASH,   STATE_ERROR,         STATE_ERROR,        STATE_ERROR    }, // STATE_START_ABS_DIR
    { STATE_START_NAME,   STATE_START_DOT,   STATE_START_SLASH,   STATE_START_TILDA,   STATE_ERROR,        STATE_START_WILD}, // STATE_START_PTN

    { STATE_MID,          STATE_TWO_DOTS,    STATE_IGNORE,        STATE_MID,           STATE_END,          STATE_CHK_WILD }, // STATE_START_DOT
    { STATE_MID,          STATE_DOT,         STATE_IGNORE,        STATE_TILDA,         STATE_END,          STATE_CHK_WILD }, // STATE_START_SLASH
    { STATE_START_NAME,   STATE_START_NAME,  STATE_HOME,          STATE_START_NAME,    STATE_HOME_END,     STATE_CHK_WILD }, // STATE_START_TILDA
    { STATE_MID,          STATE_MID,         STATE_SLASH,         STATE_MID,           STATE_NAME_END,     STATE_CHK_WILD }, // STATE_START_NAME
    { STATE_WILD,         STATE_WILD,        STATE_ERROR,         STATE_WILD,          STATE_NAME_END,     STATE_WILD     }, // STATE_START_WILD
    { STATE_MID,          STATE_MID,         STATE_PARENT,        STATE_MID,           STATE_PARENT_END,   STATE_CHK_WILD }, // STATE_TWO_DOTS
    { STATE_MID,          STATE_DOT,         STATE_IGNORE,        STATE_TILDA,         STATE_END,          STATE_CHK_WILD }, // STATE_IGNORE
    { STATE_MID,          STATE_MID,         STATE_SLASH,         STATE_MID,           STATE_NAME_END,     STATE_CHK_WILD }, // STATE_MID
    { STATE_MID,          STATE_TWO_DOTS,    STATE_IGNORE,        STATE_MID,           STATE_END,          STATE_CHK_WILD }, // STATE_DOT
    { STATE_MID,          STATE_MID,         STATE_ERROR,         STATE_MID,           STATE_ERROR,        STATE_CHK_WILD }, // STATE_TILDA
    { STATE_MID,          STATE_DOT,         STATE_IGNORE,        STATE_TILDA,         STATE_END,          STATE_CHK_WILD }, // STATE_HOME
    { STATE_MID,          STATE_DOT,         STATE_IGNORE,        STATE_TILDA,         STATE_END,          STATE_CHK_WILD }, // STATE_SLASH
    { STATE_MID,          STATE_DOT,         STATE_IGNORE,        STATE_TILDA,         STATE_END,          STATE_CHK_WILD }, // STATE_PARENT
    { STATE_WILD,         STATE_WILD,        STATE_ERROR,         STATE_WILD,          STATE_NAME_END,     STATE_WILD     }, // STATE_CHK_WILD
    { STATE_WILD,         STATE_WILD,        STATE_ERROR,         STATE_WILD,          STATE_NAME_END,     STATE_WILD     }, // STATE_WILD
};

static bool bMutexInitialized;

static void WfsPathClear(WFSPathName *pPn, WFSPathPartOffsetArray *pPpoa) {
    pPn->nLen = 1;
    pPpoa->nNumParts = 0;
    pPpoa->aPart[pPpoa->nNumParts] = pPn->nLen;
    pPn->sStr[0] = '/';
    pPn->sStr[1] = 0;
}

void WFSPathInit() {
    if (!bMutexInitialized)
    {
        bMutexInitialized = true;
        WFSInitMutex(&wcg.pg.mxAccessPathGlobals);
    }
    WfsPathClear(&wcg.pg.pathCurrent.pn, &wcg.pg.pathCurrent.ppoa);
    WfsPathClear(&wcg.pg.pathHome.pn, &wcg.pg.pathHome.ppoa);
}

static void WfsPathParent(WFSPathName *pPn, WFSPathPartOffsetArray *pPpoa) {
    if (pPpoa->nNumParts == 0) {
        return;
    }
    pPpoa->nNumParts--;
    pPn->nLen = pPpoa->aPart[pPpoa->nNumParts];
    pPn->sStr[pPn->nLen] = 0;
}

void WFSPathCopy(WFSPathName *pPnDst, WFSPathPartOffsetArray *pPpoaDst, WFSPathName *pPnSrc, WFSPathPartOffsetArray *pPpoaSrc, bool bAppendSeparator, bool bNullTermination) {
    pPnDst->nLen = pPnSrc->nLen;
    pPpoaDst->nNumParts = pPpoaSrc->nNumParts;
    memcpy(pPpoaDst->aPart, pPpoaSrc->aPart, sizeof(pPpoaSrc->aPart[0]) * (pPpoaSrc->nNumParts+1));
    memcpy(pPnDst->sStr, pPnSrc->sStr, sizeof(pPnSrc->sStr[0]) * pPnSrc->nLen);
    if ((bAppendSeparator) && (pPnDst->sStr[pPnDst->nLen-1]!='/')) {
        pPnDst->sStr[pPnDst->nLen++] = '/';
        pPpoaDst->aPart[pPpoaDst->nNumParts]++;
    }

    if (bNullTermination)
    {
       pPnDst->sStr[pPnDst->nLen] = 0;
    }
}

WFSResult WFSPathResolve(WFSPathName *pPnOutAbs, WFSPathPartOffsetArray *pPpoaOutAbs, const utf8 *sInPathName, WFSPathNameType nPathType) {
    // Arguments:
    // ----------
    // pOutAbsPath  .. The output normalized absolute path name
    // sInPathName  .. The input path name
    // nPathType    .. Specifies whether to allow relative or absolute, absolute-only, or patterns
    //
    // Description:
    // ------------
    // Uses a state machine to process input path names or search patterns.
    // Relative paths are concatenated to the current directory or home directory.
    // The output is a normalized absolute path name or pattern.
    // normalized means "../" == "go back one directory" are processed and removed.
    // Invalid path names are detected, and cause a WFS_RESULT_INVALID error.
    WFSPathPartOffsetArray ppoa;
    WFSReturnOnNullParameter(sInPathName);
    if (pPpoaOutAbs == NULL) {
        pPpoaOutAbs = &ppoa;
    }
    const utf8 *pParse = sInPathName;
    const utf8 *pName = pParse;
    s32 nState = (WFSPathFsmState)nPathType;
    s32 nDepthLimit = WFS_MAX_PATH_DEPTH;
    s32 nLenAdjust = 0;
    if ((nPathType==STATE_START_REL_DIR)||(nPathType==STATE_START_ABS_DIR)) {
        nDepthLimit--;
    }
    WFSLockMutex(&wcg.pg.mxAccessPathGlobals);
    WfsPathClear(pPnOutAbs, pPpoaOutAbs);
    while(nState < STATE_END) {
        s32 nCharCategory = aCharCategory[*((u8*)pParse++)];
        nState = aStateMachine[nState][nCharCategory];
        switch(nState) {
            case STATE_START_DOT:   // ".  "
            case STATE_START_NAME:   // "A  "
            case STATE_START_WILD:   // "*  "
                WFSPathCopy(pPnOutAbs, pPpoaOutAbs, &wcg.pg.pathCurrent.pn, &wcg.pg.pathCurrent.ppoa, true, false);
                break;
            case STATE_START_SLASH: // "/  "
                WfsPathClear(pPnOutAbs, pPpoaOutAbs);
                pName++;
                break;
            case STATE_HOME:
            case STATE_HOME_END:
                WFSPathCopy(pPnOutAbs, pPpoaOutAbs, &wcg.pg.pathHome.pn, &wcg.pg.pathHome.ppoa, true, false);
                // Falls through
            case STATE_IGNORE:
                pName = pParse;
                break;
            case STATE_NAME_END:
                nLenAdjust = -1;
                // Falls through
            case STATE_SLASH: {
                s32 nPartLen  = pParse - pName;
                if (nPartLen>(WFS_MAX_FILE_NAME_SIZE+1)) {
                    goto Error;
                }
                s32 nNewLen   = pPnOutAbs->nLen + nPartLen + nLenAdjust;
                s32 nNewDepth = pPpoaOutAbs->nNumParts+1;
                if ((nNewLen>WFS_MAX_PATH_NAME_SIZE)||(nNewDepth>nDepthLimit)) {
                    goto Error;
                }
                memcpy(&pPnOutAbs->sStr[pPnOutAbs->nLen], pName, (u32)nPartLen);
                pPnOutAbs->nLen = (WFSPathNameLengthType)nNewLen;
                pPpoaOutAbs->nNumParts++;
                pPpoaOutAbs->aPart[pPpoaOutAbs->nNumParts] = (WFSPathNameLengthType)pPnOutAbs->nLen;
                pName = pParse;
                break;
            }
            case STATE_PARENT:
            case STATE_PARENT_END:
                WfsPathParent(pPnOutAbs, pPpoaOutAbs);
                pName = pParse;
                break;
            case STATE_CHK_WILD:
                if (nPathType != PATH_TYPE_PATTERN) {
                    goto Error;
                }
                break;
            case STATE_ERROR:
                    goto Error;
            case STATE_START_TILDA: case STATE_TILDA:    case STATE_DOT:
            case STATE_START_REL:   case STATE_TWO_DOTS: case STATE_MID:   case STATE_END:
            case STATE_START_REL_DIR:  case STATE_START_ABS_DIR:
                break;
        }
    }
    if ((pPnOutAbs->sStr[pPnOutAbs->nLen-1] == '/') && (pPnOutAbs->nLen>1)) {
        pPnOutAbs->sStr[--pPnOutAbs->nLen] = 0;
        --pPpoaOutAbs->aPart[pPpoaOutAbs->nNumParts];
    }
    pPnOutAbs->sStr[pPnOutAbs->nLen] = 0;
    WFSUnlockMutex(&wcg.pg.mxAccessPathGlobals);
    dbg(MyOSReport("path: %s\n", pPnOutAbs->sStr));
    return WFS_RESULT_OK;
Error:
    WFSUnlockMutex(&wcg.pg.mxAccessPathGlobals);
    dbg(MyOSReport("path: WFS_RESULT_INVALID\n"));
    return WFS_RESULT_INVALID;
}
