/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfscli_Handles.cpp - Module for file handles and search directory handles


  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfscli_Handles.cpp,v $
  Revision 1.11  2008/10/07 06:41:51  ueno
  Modified WFSCliFileInfoAlloc to set WFSCLI_HANDLE_ALLOCATED in order to indicate the handle is allocated.

  Revision 1.10  2008/04/19 05:53:38  kondo_masahiro
  Fixed code to accept RM compiler

  Revision 1.9  2007/12/25 14:03:49  ueno
  Modified to initialize mutex, thread and heap only once.

  Revision 1.8  2007/11/16 02:15:59  wayne.wong
  Added enumeration and stubs for WFSFlush call.

  Revision 1.7  2007/11/02 01:07:38  paul
  now logging changes

*---------------------------------------------------------------------------*/

// File created by Paul Donnelly on 2007.10.06

#include "wfs_Client.h"

static bool bMutexInitialized; // [check]

void WFSCliInitHandles() {
    if (!bMutexInitialized)
    {
        bMutexInitialized = true;
        WFSInitMutex(&wcg.hnd.mxAccessTheFileInfo);
        WFSInitMutex(&wcg.hnd.mxAccessTheSearchDirInfo);
    }
    wcg.hnd.nFileHandleCounter        = 0;
    wcg.hnd.nSearchDirHandleCounter   = 0;
    wcg.hnd.nFirstFreeFileHandle      = 0;
    wcg.hnd.nFirstFreeSearchDirHandle = 0;
    wcg.hnd.nNumUsedFileHandles       = 0;
    wcg.hnd.nNumUsedSearchDirHandles  = 0;
    //wcg.hnd.nFirstUsedFileHandle      = WFSCLI_MAX_FILE_HANDLES;  // A value of WFSCLI_MAX_FILE_HANDLES is the list terminator
    //wcg.hnd.nFirstUsedSearchDirHandle = WFSCLI_MAX_SEARCH_DIR_HANDLES;  // A value of WFSCLI_MAX_SEARCH_DIR_HANDLES is the list terminator
    for(u32 nI=0; nI<WFSCLI_MAX_FILE_HANDLES; nI++) {
        wcg.hnd.aCliFileInfo[nI].nNext = nI+1;
    }
    for(u32 nI=0; nI<WFSCLI_MAX_SEARCH_DIR_HANDLES; nI++) {
        wcg.hnd.aCliSearchDirInfo[nI].nNext = nI+1;
    }
}

WFSResult WFSGetCliFileInfo(WFSCliFileInfo **ppFi, WFSFileHandle fh) {
    *ppFi = &wcg.hnd.aCliFileInfo[(fh>>16)&(WFSCLI_MAX_FILE_HANDLES-1)];
    return ((*ppFi)->nHandle == fh)?WFS_RESULT_OK:WFS_RESULT_INVALID;
}

WFSResult WFSGetCliSearchDirInfo(WFSCliSearchDirInfo **ppSdi, WFSSearchDirectoryHandle sdh) {
    *ppSdi = &wcg.hnd.aCliSearchDirInfo[(sdh>>16)&(WFSCLI_MAX_SEARCH_DIR_HANDLES-1)];
    return ((*ppSdi)->nHandle == sdh)?WFS_RESULT_OK:WFS_RESULT_INVALID;
}

WFSResult WFSCliFileInfoAlloc(WFSCliFileInfo **ppFi) {
    WFSLockMutex(&wcg.hnd.mxAccessTheFileInfo);
    if (wcg.hnd.nFirstFreeFileHandle == WFSCLI_MAX_FILE_HANDLES) {
        WFSUnlockMutex(&wcg.hnd.mxAccessTheFileInfo);
        return WFS_RESULT_MAX_HANDLES;
    }
    wcg.hnd.nNumUsedFileHandles++;
    u32 nIdx = wcg.hnd.nFirstFreeFileHandle;
    wcg.hnd.nFirstFreeFileHandle = wcg.hnd.aCliFileInfo[nIdx].nNext;
    wcg.hnd.aCliFileInfo[nIdx].nNext = WFSCLI_HANDLE_ALLOCATED; // this means the handle is open.
    *ppFi = &wcg.hnd.aCliFileInfo[nIdx];
    wcg.hnd.aCliFileInfo[nIdx].nHandle = WFSCLI_FILE_HANDLE_TT_FIELD | wcg.hnd.nFileHandleCounter | (nIdx<<16);
    wcg.hnd.nFileHandleCounter += WFSCLI_FILE_HANDLE_INC;
    wcg.hnd.nFileHandleCounter &= WFSCLI_HANDLE_COUNTER_MASK;
    WFSUnlockMutex(&wcg.hnd.mxAccessTheFileInfo);
    return WFS_RESULT_OK;
}

WFSResult WFSCliSearchDirInfoAlloc(WFSCliSearchDirInfo **ppSdi) {
    WFSLockMutex(&wcg.hnd.mxAccessTheSearchDirInfo);
    if (wcg.hnd.nFirstFreeSearchDirHandle == WFSCLI_MAX_SEARCH_DIR_HANDLES) {
        WFSUnlockMutex(&wcg.hnd.mxAccessTheSearchDirInfo);
        return WFS_RESULT_MAX_HANDLES;
    }
    wcg.hnd.nNumUsedSearchDirHandles++;
    u32 nIdx = wcg.hnd.nFirstFreeSearchDirHandle;
    wcg.hnd.nFirstFreeSearchDirHandle = wcg.hnd.aCliSearchDirInfo[nIdx].nNext;
    wcg.hnd.aCliSearchDirInfo[nIdx].nNext = WFSCLI_HANDLE_ALLOCATED; // this means the handle is open.
    *ppSdi = &wcg.hnd.aCliSearchDirInfo[nIdx];
    wcg.hnd.aCliSearchDirInfo[nIdx].nHandle = WFSCLI_SEARCH_DIR_HANDLE_TT_FIELD | wcg.hnd.nSearchDirHandleCounter | (nIdx<<16);
    wcg.hnd.nSearchDirHandleCounter += WFSCLI_SEARCH_DIR_HANDLE_INC;
    wcg.hnd.nSearchDirHandleCounter &= WFSCLI_HANDLE_COUNTER_MASK;
    WFSUnlockMutex(&wcg.hnd.mxAccessTheSearchDirInfo);
    return WFS_RESULT_OK;
}

void WFSCliFileInfoFree(WFSCliFileInfo *pFi) {
    WFSLockMutex(&wcg.hnd.mxAccessTheFileInfo);
    u32 nIdx = (u32)(pFi - wcg.hnd.aCliFileInfo);
    pFi->nHandle ^= 0x55555555; // ensures the deleted handle is invalid
    pFi->nNext = wcg.hnd.nFirstFreeFileHandle;
    wcg.hnd.nNumUsedFileHandles--;
    wcg.hnd.nFirstFreeFileHandle = nIdx;
    WFSUnlockMutex(&wcg.hnd.mxAccessTheFileInfo);
}

void WFSCliSearchDirInfoFree(WFSCliSearchDirInfo *pSdi) {
    WFSLockMutex(&wcg.hnd.mxAccessTheSearchDirInfo);
    u32 nIdx = (u32)(pSdi - wcg.hnd.aCliSearchDirInfo);
    pSdi->nHandle ^= 0x55555555; // ensures the deleted handle is invalid
    pSdi->nNext = wcg.hnd.nFirstFreeSearchDirHandle;
    wcg.hnd.nNumUsedSearchDirHandles--;
    wcg.hnd.nFirstFreeSearchDirHandle = nIdx;
    WFSUnlockMutex(&wcg.hnd.mxAccessTheSearchDirInfo);
}
