/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfskrn_Device.cpp - storage device abstraction.

  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

*---------------------------------------------------------------------------*/

#include "wfskrn_Device.h"
#include "wfskrn_Api.h"
#include <brom/nand/nand.h>
#include <drnand.h>


typedef struct{
    u32 cmd;
    s32 status;
    WFSHashCode *pHash;
    WFSHashCode readHash;
} DeviceMscResourceRequest;

static DeviceMscResourceRequest gPmrr[DEVICE_IOP_MAX_ASYNC_NUM];

#if _WIN32
char sDevicePathWin32[sizeof(WFSKRN_SIMULATED_DISK_PATH)+WFS_MAX_DEVICE_NAME_SIZE+1] = WFSKRN_SIMULATED_DISK_PATH;

char *DeviceGetWin32PathName(const utf8 *sDevName) {
    memcpy(&sDevicePathWin32[sizeof(WFSKRN_SIMULATED_DISK_PATH)-1], sDevName, WFS_MAX_DEVICE_NAME_SIZE+1);
    return sDevicePathWin32;
}

WFSHashCode aHashDummyData = {0xbeefbeef,0xbeefbeef,0xbeefbeef,0xbeefbeef,0xbeefbeef};

#elif _IOP

static IOSMessageQueueId      gMqidPmrr;
static IOSMessage             gMqPmrr[DEVICE_IOP_MAX_ASYNC_NUM];

static IOSMessageQueueId      gMqidIrrFb;
static IOSMessage             gMqIrrFb[1];
static IOSResourceRequest     gIrrFuncBlock;

static DeviceAttachDetachCallback gCbAttach = 0;

#endif

static s32 cntDebugPseudoDetachDrive = -1;
static s32 cntDebugPseudoHashInconsistent = -1;

#define WFS_MAX_DEVICE_PATH (sizeof(WFS_DEVICE_DIRECTORY) + WFS_MAX_DEVICE_NAME_SIZE + 2)


// Initialize Device library
// x86 - initialize parameter
void DeviceInitLib() {
    dbgd(osTPrintf("DeviceInitLib()\n"));
    wkg.nNumDevices = 0;
    memset(&wkg.aDevInfo, 0, sizeof(wkg.aDevInfo));
    memset(&gPmrr, 0, sizeof(DeviceMscResourceRequest)*DEVICE_IOP_MAX_ASYNC_NUM);
}


void DeviceDebugSetPseudoDetachDevice(s32 cnt)
{
    cntDebugPseudoDetachDrive = cnt;
}

void DeviceDebugSetPseudoHashInconsistent(s32 cnt)
{
    cntDebugPseudoHashInconsistent = cnt;
}

s32 DeviceDebugGetPseudoDetachDevice()
{
    return cntDebugPseudoDetachDrive;
}

s32 DeviceDebugGetPseudoHashInconsistent()
{
    return cntDebugPseudoHashInconsistent;
}

#if _WIN32

static IOSError _WIN32DeviceRead(DeviceInfo *pDevInfo, u32 nSectorAdr, u8 *pData, u16 nSectors, u8 *pHash, u16 nHashSectors, s32 nHashStride)
{
    u64  filePos = (u64)nSectorAdr * pDevInfo->nSectorSize;
    filePos += 2 * sizeof(u32);    // initial bytes have sector size + dev capacity
    LONG posLow = (LONG) filePos & 0xffffffff;
    LONG posHigh = filePos >> 32;
    SetFilePointer(pDevInfo->hSimulatedDeviceFile, posLow, &posHigh, FILE_BEGIN);

    DWORD numBytes = pDevInfo->nSectorSize * nSectors;
    DWORD bytesRead;
    if (!ReadFile(pDevInfo->hSimulatedDeviceFile, pData, numBytes, &bytesRead, NULL)) {
        dbgd(osTPrintf("DeviceRead: read failed\n"));
        return IOS_ERROR_MSC_COMMAND_FAILED;
    }
    if(pData[0] == 0xaa){
        int a; a=0;
    }
    if (numBytes != bytesRead) {
        dbgd(osTPrintf("DeviceRead: bytes read does not match\n"));
        return IOS_ERROR_MSC_COMMAND_FAILED;
    }

	s32 nSize = numBytes;
    int nNumBlks = 0;
    while(nSize>0){
        u8 *pHashPtr = (u8*)pHash+(nNumBlks*nHashStride);
        u32 *pHashDummyData32 = (u32*)aHashDummyData;
        *pHashDummyData32++ = nSectorAdr+(nNumBlks*nHashSectors);
        *pHashDummyData32++ = nSize < (nHashSectors<<pDevInfo->dh.nLog2SectorSize) ? nSize : (nHashSectors<<pDevInfo->dh.nLog2SectorSize);
        if(memcmp(pHashPtr, (u8*)aHashDummyData, sizeof(WFSHashCode))){
            return IOS_ERROR_MSC_HASH_INCONSISTENT;
        }
        nSize -= (nHashSectors<<pDevInfo->dh.nLog2SectorSize);
		++nNumBlks;
    }
    return IOS_ERROR_MSC_OK;
}

static IOSError _WIN32DeviceWrite(DeviceInfo *pDevInfo, u32 nSectorAdr, u8 *pData, u16 nSectors, u8 *pHash, u16 nHashSectors, s32 nHashStride, bool bDecryption)
{
    u64  filePos = (u64)nSectorAdr * pDevInfo->nSectorSize;
    filePos += sizeof(pDevInfo->dh);    // initial bytes have sector size + dev capacity
    LONG posLow = (LONG) filePos & 0xffffffff;
    LONG posHigh = filePos >> 32;
    SetFilePointer(pDevInfo->hSimulatedDeviceFile, posLow, &posHigh, FILE_BEGIN);

    DWORD numBytes = pDevInfo->nSectorSize * nSectors;
    DWORD bytesWritten;
    s32 nSize = numBytes;
    int nNumBlks = 0;
	while(nSize>0){
        u8 *pHashPtr = (u8*)pHash+(nNumBlks*nHashStride);
        u32 *pHashDummyData32 = (u32*)aHashDummyData;
        *pHashDummyData32++ = nSectorAdr+(nNumBlks*nHashSectors);
		if(nSectorAdr+(nNumBlks*nHashSectors) == 0x180){
		    int a; a=0;
		}
        *pHashDummyData32++ = nSize < (nHashSectors<<pDevInfo->dh.nLog2SectorSize) ? nSize : (nHashSectors<<pDevInfo->dh.nLog2SectorSize);
        memcpy(pHashPtr, (u8*)aHashDummyData, sizeof(WFSHashCode));
        nSize -= (nHashSectors<<pDevInfo->dh.nLog2SectorSize);
		++nNumBlks;
    }
    if (!WriteFile(pDevInfo->hSimulatedDeviceFile, pData, numBytes, &bytesWritten, NULL)) {
        dbgd(osTPrintf("DeviceWrite: write failed\n"));
        return IOS_ERROR_MSC_COMMAND_FAILED;
    }
    if(pData[0] == 0xaa){
        int a; a=0;
    }
    if (numBytes != bytesWritten) {
        dbgd(osTPrintf("DeviceWrite: bytes written does not match\n"));
        return IOS_ERROR_MSC_COMMAND_FAILED;
    }
    if(!bDecryption){
        memset(pData, 0xaa, numBytes);
    }
    return IOS_ERROR_MSC_OK;
}

#endif

// Read nNumSectors sectors from device devHandle starting at LBA nSectorAdr into buffer
// pBuffer. nSectorAdr == LBA.
WFSKrnResult DeviceRead(DeviceInfo *pDevInfo, u32 nSectorAdr, u8 *pBuffer, u32 nSectors, WFSHashCode *pHash, u32 nLog2HashSectors, u32 nOffsetHashBlks, s32 nHashStride, s32 nHashStridePerMaxLBlks)
{
    dbgd(osTPrintf("DeviceRead()\n"));
    ASSERT(nLog2HashSectors <= (WFS_LOG2_MAX_HASHABLE_BLK_SIZE-pDevInfo->dh.nLog2SectorSize)); // ~64KB
#if _DEBUG
    //osTPrintf("Read: nSectorAdr = 0x%x, pDevInfo->dh.nNumSectors = 0x%x\n", nSectorAdr, pDevInfo->dh.nNumSectors);
    if (nSectorAdr >= pDevInfo->dh.nNumSectors) {
        WFSKrnOutputErrorStr("Attempted access beyond end of device");
    }
#endif
    WFSKrnResult nResult = WFSKRN_RESULT_OK;
    DeviceMscResourceRequest *pCurrentReply;
    u32 nMaxSectors  = 1<<(WFS_LOG2_LARGE_BLK_SIZE-pDevInfo->dh.nLog2SectorSize);
    u32 nHashSectors = (1<<nLog2HashSectors);
    u8  *pData       = pBuffer;
    u32 nNumBlks     = nSectors ? ((nSectors-1)>>nLog2HashSectors)+1 : 0;
    u32 nNumLoop     = (((nHashSectors<<pDevInfo->dh.nLog2SectorSize)*(nNumBlks+nOffsetHashBlks)-1)>>WFS_LOG2_LARGE_BLK_SIZE)+1;
    u32 nRcvLoop     = nNumLoop < DEVICE_IOP_MAX_ASYNC_NUM ? nNumLoop : DEVICE_IOP_MAX_ASYNC_NUM;
    int i;
    for(i=0;i<nNumLoop+nRcvLoop;i++){
#if _TWL
        pCurrentReply = &gPmrr[0];
        if(i>=nRcvLoop){
#endif
            switch(pCurrentReply->status){
            case IOS_ERROR_MSC_DEVICE_DETACH:
            case IOS_ERROR_MSC_COMMAND_FAILED:
            case IOS_ERROR_MSC_PHASE_ERROR:
            case IOS_ERROR_MSC_INVALID_CBW_SIGNATURE:
            case IOS_ERROR_MSC_INVALID_CBW_TAG:
                nResult = WFSKRN_RESULT_DEVICE_NOT_FOUND;
                break;
#if DEVICE_VERIFY_HASH_DATA
            case IOS_ERROR_MSC_HASH_INCONSISTENT:
                pDevInfo->nFlags |= WFS_DEVFLAG_NOT_INITIALIZED;
                nResult = WFSKRN_RESULT_DEVICE_HASH_INCONSISTENT;
                break;
#endif
            }
#if DEVICE_DEBUG_PSEUDO_DEVICE_ERROR
            if(cntDebugPseudoDetachDrive == 0){
                cntDebugPseudoDetachDrive = -1;
                nResult = WFSKRN_RESULT_DEVICE_NOT_FOUND;
            }
            else if(cntDebugPseudoDetachDrive){
                --cntDebugPseudoDetachDrive;
            }
#if DEVICE_VERIFY_HASH_DATA
            if(cntDebugPseudoHashInconsistent == 0){
                pDevInfo->nFlags |= WFS_DEVFLAG_NOT_INITIALIZED;
                nResult = WFSKRN_RESULT_DEVICE_HASH_INCONSISTENT;
                cntDebugPseudoHashInconsistent = -1;
            }
            else if(cntDebugPseudoHashInconsistent){
                --cntDebugPseudoHashInconsistent;
            }
#endif
#endif
        }
        if(i<nNumLoop){
            u32 nSec = (nSectors+(nOffsetHashBlks<<nLog2HashSectors)) < nMaxSectors ? nSectors : (nMaxSectors-(nOffsetHashBlks<<nLog2HashSectors));
            pCurrentReply->pHash = pHash;
            IOSError ie;
#if _WIN32
			ie = _WIN32DeviceRead(pDevInfo, nSectorAdr, pData, (u16)nSec, (u8*)pCurrentReply->pHash, (u16)(1<<nLog2HashSectors), nHashStride);
#elif _TWL
        if( nandNfsIo( 0, nSectorAdr, pData, nSec, TRUE) == FALSE) {
            ie = IOS_ERROR_MSC_COMMAND_FAILED;
            nResult = WFSKRN_RESULT_MEDIA_ERROR;
        }else{
            ie = IOS_ERROR_MSC_OK;
        }
#endif
            if (ie < IOS_ERROR_MSC_OK){
                pCurrentReply->status = ie;
            }
            nSectors    -= nSec;
            pData       += nSec<<pDevInfo->dh.nLog2SectorSize;
            nSectorAdr  += nSec;
            u32 nBlks    = ((nSec-1)>>nLog2HashSectors)+1;
            pHash        = (WFSHashCode*)((u8*)pHash+nBlks*nHashStride+nHashStridePerMaxLBlks);
            nOffsetHashBlks  = 0;
        }
#if _TWL
    }
#endif
    return nResult;
}


// Write nNumSectors sectors to device devHandle starting at LBA nSectorAdr from buffer
// pBuffer. nSectorAdr == LBA.
WFSKrnResult DeviceWrite(DeviceInfo *pDevInfo, u32 nSectorAdr, u8 *pBuffer, u32 nSectors, WFSHashCode *pHash, u32 nLog2HashSectors, u32 nOffsetHashBlks, s32 nHashStride, s32 nHashStridePerMaxLBlks, bool bDecryption)
{
#if _DEBUG_BREAK_POINT
    if (nSectorAdr==0x28800) {
        int a;a=0;
    }
#endif
    dbgd(osTPrintf("DeviceWrite()\n"));
    ASSERT(nLog2HashSectors <= (WFS_LOG2_MAX_HASHABLE_BLK_SIZE-pDevInfo->dh.nLog2SectorSize)); // ~64KB
#if _DEBUG
    //osTPrintf("Write: nSectorAdr = 0x%x, pDevInfo->dh.nNumSectors = 0x%x\n", nSectorAdr, pDevInfo->dh.nNumSectors);
    if (nSectorAdr >= pDevInfo->dh.nNumSectors) {
        WFSKrnOutputErrorStr("Attempted access beyond end of device");
    }
#endif
    WFSKrnResult nResult = WFSKRN_RESULT_OK;
    DeviceMscResourceRequest *pCurrentReply;
    u32 nMaxSectors  = 1<<(WFS_LOG2_LARGE_BLK_SIZE-pDevInfo->dh.nLog2SectorSize);
    u32 nHashSectors = (1<<nLog2HashSectors);
    u8  *pData       = pBuffer;
    u32 nNumBlks     = nSectors ? ((nSectors-1)>>nLog2HashSectors)+1 : 0;
    u32 nNumLoop    = (((nHashSectors<<pDevInfo->dh.nLog2SectorSize)*(nNumBlks+nOffsetHashBlks)-1)>>WFS_LOG2_LARGE_BLK_SIZE)+1;
	u32 nRcvLoop    = nNumLoop < DEVICE_IOP_MAX_ASYNC_NUM ? nNumLoop : DEVICE_IOP_MAX_ASYNC_NUM;
    int i;
    for(i=0;i<nNumLoop+nRcvLoop;i++){
#if _TWL
        pCurrentReply = &gPmrr[0];
        if( i>=1 && i<=nNumLoop ){
#endif
            switch(pCurrentReply->status){
            case IOS_ERROR_MSC_DEVICE_DETACH:
            case IOS_ERROR_MSC_COMMAND_FAILED:
            case IOS_ERROR_MSC_PHASE_ERROR:
            case IOS_ERROR_MSC_INVALID_CBW_SIGNATURE:
            case IOS_ERROR_MSC_INVALID_CBW_TAG:
                nResult = WFSKRN_RESULT_DEVICE_NOT_FOUND;
            }
#if DEVICE_DEBUG_PSEUDO_DEVICE_ERROR
            if(cntDebugPseudoDetachDrive == 0){
                nResult = WFSKRN_RESULT_DEVICE_NOT_FOUND;
                cntDebugPseudoDetachDrive = -1;
            }
            else if(cntDebugPseudoDetachDrive){
                --cntDebugPseudoDetachDrive;
            }
#endif
        }
        if(i<nNumLoop){
            u32 nSec = (nSectors+(nOffsetHashBlks<<nLog2HashSectors)) < nMaxSectors ? nSectors : (nMaxSectors-(nOffsetHashBlks<<nLog2HashSectors));
            pCurrentReply->pHash = pHash;
            IOSError ie;
#if _WIN32
			ie = _WIN32DeviceWrite(pDevInfo, nSectorAdr, pData, (u16)nSec, (u8*)pCurrentReply->pHash, (u16)(1<<nLog2HashSectors), nHashStride, bDecryption);
#elif _TWL
        if( nandNfsIo( 0, nSectorAdr, pData, nSec, FALSE) == FALSE) {
            ie = IOS_ERROR_MSC_COMMAND_FAILED;
            nResult = WFSKRN_RESULT_MEDIA_ERROR;
        }else{
            ie = IOS_ERROR_MSC_OK;
        }
#endif
            if (ie < IOS_ERROR_MSC_OK){
                pCurrentReply->status = ie;
            }
            nSectors    -= nSec;
            pData       += nSec<<pDevInfo->dh.nLog2SectorSize;
            nSectorAdr  += nSec;
            u32 nBlks    = ((nSec-1)>>nLog2HashSectors)+1;
            pHash        = (WFSHashCode*)((u8*)pHash+nBlks*nHashStride+nHashStridePerMaxLBlks);
            nOffsetHashBlks  = 0;
        }
#if _TWL
	}
#endif
    return WFSKRN_RESULT_OK;
}

DeviceInfo *DeviceGetDeviceInfo(const WFSDeviceName *pDevName) {
    dbgd(osTPrintf("DeviceGetDeviceInfo()\n"));
    utf8 sDevPath[WFS_MAX_DEVICE_NAME_SIZE + 6];
    void *pDevInfo;
    RxtItr rxi;
    rxi.pStrPtr = sDevPath;
    osSNPrintf(sDevPath, WFS_MAX_DEVICE_NAME_SIZE + 6,"/dev/%s", pDevName->sStr);
    if ((!PathCacheFind3(&rxi, pDevName->nLen+5, &pDevInfo)) ||
        ((DeviceInfo*)pDevInfo < wkg.aDevInfo) ||
        ((DeviceInfo*)pDevInfo >= &wkg.aDevInfo[WFSKRN_MAX_DEVICES]))
    {
        return 0;
    }
    return (DeviceInfo*)pDevInfo;
}


WFSKrnResult DeviceGetVolumeId(const WFSDeviceName *pDevName, WFSVolumeId *pVolId) {
    dbgd(osTPrintf("DeviceGetVolumeInfo()\n"));
    DeviceInfo *pDevInfo = DeviceGetDeviceInfo(pDevName);
    if(!pDevInfo || !pDevInfo->bInUse) {
        return WFSKRN_RESULT_NOT_FOUND;
    }
    else if (pDevInfo->nFlags & WFS_DEVFLAG_UNUSABLE) {
        return WFSKRN_RESULT_DEV_UNUSABLE;
    }
    else if (pDevInfo->nFlags & WFS_DEVFLAG_NOT_INITIALIZED) {
        return WFSKRN_RESULT_DEV_NOT_INITIALIZED;
    }
    else if(pDevInfo->pVolInfo){
        VolumeConvertU32ToVolId(pVolId, pDevInfo->pVolInfo->vh.nVolId);
        return WFSKRN_RESULT_OK;
    }
    VolumeInfo tempVolInfo;
    tempVolInfo.pDevInfo = pDevInfo;
    pDevInfo->pVolInfo = &tempVolInfo;
    BCacheEntry* pBce;
    WFSKrnResult nResult = BCacheAllocMetaBlk(&tempVolInfo, VOLUME_HDR_BLK_ADR, BCACHE_FLAG_PINNED | BCACHE_FLAG_READ, &pBce);
    if(nResult < WFSKRN_RESULT_OK){
        return nResult;
    }
    VolumeHdr *pVh = (VolumeHdr *)pBce->pBlkPtr;
    VolumeConvertU32ToVolId(pVolId, pVh->nVolId);
    nResult = BCacheUnpin(&tempVolInfo, VOLUME_HDR_BLK_ADR);
    pDevInfo->pVolInfo = 0;
    return nResult;
}


WFSKrnResult DeviceMountVolume(DeviceInfo *pDevInfo) {
    dbgd(osTPrintf("DeviceMountVolume()\n"));
    WFSKrnResult nResult = WFSKRN_RESULT_OK;
    VolumeInfo tempVolInfo;
    tempVolInfo.pDevInfo = pDevInfo;
    pDevInfo->pVolInfo = &tempVolInfo;
    BCacheEntry* pBce;
    nResult = BCacheAllocMetaBlk(&tempVolInfo, VOLUME_HDR_BLK_ADR, BCACHE_FLAG_PINNED | BCACHE_FLAG_READ, &pBce);
    if(nResult < WFSKRN_RESULT_OK){
        return nResult;
    }
    VolumeHdr *pVh = (VolumeHdr *)pBce->pBlkPtr;
    tempVolInfo.vh = *(VolumeHdr *)pVh;
    BCacheUnpin(&tempVolInfo, VOLUME_HDR_BLK_ADR);
    VolumeInfo *pVolInfo = wkg.pMountedVolListAnchor->vl.pNext;
    while(pVolInfo != wkg.pMountedVolListAnchor) {
        if (pVh->nVolId == pVolInfo->vh.nVolId) {
            // The volume Id matches a currently mounted volume
            if (pVolInfo->bDetached) {
                // ToDo: Check Session ID
                pVolInfo->bDetached = false;
                pDevInfo->pVolInfo = pVolInfo;
                //pDevInfo->nFlags &= ~WFS_DEVFLAG_NOT_MOUNTED;
            } else {
                nResult = WFSKRN_RESULT_VOL_ID_ERROR;
            }
            goto Exit;
        }
    }
    BCacheReTag(&tempVolInfo, VOLUME_HDR_BLK_ADR, pVolInfo, VOLUME_HDR_BLK_ADR);
    // The volume Id did not match any currently mounted volumes
    pVolInfo = pDevInfo->pVolInfo = VolumeAlloc();
    pVolInfo->pDevInfo = pDevInfo;
    pVolInfo->vh = tempVolInfo.vh;
    VolumeAddToMountedList(pVolInfo);
    pVolInfo->rootAreaInfo.pVolInfo        = pVolInfo;
    pVolInfo->rootAreaInfo.pParent         = 0;
    pVolInfo->rootAreaInfo.nRelStartBlkAdr = 0;
    //pVolInfo->rootAreaInfo.ah.nLog2BlkSize = nLog2BlkSize;
    //pVolInfo->rootAreaInfo.ah.nNumBlks     = nNumBlks;
    nResult = AreaOpen(&pVolInfo->rootAreaInfo); // for debug wfskrn
    BCacheUnpin(pVolInfo, VOLUME_HDR_BLK_ADR);   // kondo_masahiro 080729
Exit:
    //BCacheUnpin(pVolInfo, VOLUME_HDR_BLK_ADR);
    return nResult;
}


WFSKrnResult DeviceAttach(DeviceInfo *pDevInfo, WFSDevType nDevType, u32 nDevIdx, u32 nLog2SectorSize, u32 nNumSectors) {
    dbgd(osTPrintf("DeviceAttach()\n"));
    // Called when user physically attaches a drive
    if(pDevInfo->bInUse){
//        return WFSKRN_RESULT_OK;
        DeviceDetach(pDevInfo);
    }

    nandNfsCtrl( 0, NFS_DEVCTL_WARMSTART, NULL);
  
    WFSKrnResult nResult = WFSKRN_RESULT_OK;
    ++wkg.nNumDevices;
    pDevInfo->bInUse = true;
    pDevInfo->nDevType = nDevType;
    pDevInfo->dh.nLog2SectorSize = nLog2SectorSize;
    pDevInfo->nSectorSize = 1<<nLog2SectorSize;
    pDevInfo->dh.nNumSectors = nNumSectors;
    pDevInfo->nTotalCapacity = (u64)nNumSectors * (u64)(pDevInfo->nSectorSize);
    WFSPathName path;
    switch(nDevType) {
        case WFS_DEVTYPE_USB_MSC_10:
            osTPrintf("Attach device with USB1.0\n");
            break;
        case WFS_DEVTYPE_USB_MSC_11:
            osTPrintf("Attach device with USB1.1\n");
            break;
        case WFS_DEVTYPE_USB_MSC_20:
            osTPrintf("Attach device with USB2.0\n");
            break;
        default:
            return WFSKRN_RESULT_DEVICE_ERROR;
    }
    pDevInfo->devName.nLen = 5;
    osSNPrintf(pDevInfo->devName.sStr, pDevInfo->devName.nLen+1, "msc%02d", nDevIdx+1);
    path.nLen = sizeof(WFS_MAX_DEVICE_PATH)+1+ pDevInfo->devName.nLen;
    osSNPrintf(path.sStr, WFS_MAX_DEVICE_PATH, WFS_DEVICE_DIRECTORY"/%s", pDevInfo->devName.sStr);
    PathCacheEntry *pEntry = PathCacheInsertAfterCursorEntry2(wkg.pathCache.pPinAnchor, path.sStr);
    pEntry->pcd.pData = pDevInfo;
#if _IOP
    if(__p_hd[nDevIdx]->state < HARDDISC_STATUS_IDLE){
        return WFSKRN_RESULT_DEVICE_ERROR;
    }
    pDevInfo->nDeviceDevHandle = __p_hd[nDevIdx];
#endif
    if (wkg.bMountedDeviceWasDetached) {
        DeviceMountVolume(pDevInfo);
    } else {
        pDevInfo->pVolInfo = 0;
        //pDevInfo->nFlags = WFS_DEVFLAG_NOT_MOUNTED;
    }
    return nResult;
}


WFSKrnResult DeviceDetach(DeviceInfo *pDevInfo) {
    dbgd(osTPrintf("DeviceDetach()\n"));
    if(!pDevInfo->bInUse){
        return WFSKRN_RESULT_OK;
    }
    // Called when user physically detaches a drive
    utf8 sDevicePathWfs[WFS_MAX_DEVICE_PATH];
    osSNPrintf(sDevicePathWfs, WFS_MAX_DEVICE_PATH, WFS_DEVICE_DIRECTORY"/%s", pDevInfo->devName.sStr);
    WFSPathName path;
    path.nLen = sizeof(WFS_MAX_DEVICE_PATH)+1+ pDevInfo->devName.nLen;
    PathCacheItr pci;
    WFSKrnResult nResult = WFSKRN_RESULT_OK;
    WFSKrnExitOnError(WFSKrnValidateCopyStrAndConvertToLowerCase(path.sStr, sDevicePathWfs, path.nLen, WFS_MAX_FILE_NAME_SIZE));
    WFSKrnExitOnError(PathCacheFindPath(&path, path.sStr, &pci));
    PathCacheDeleteEntry(pci.pEntry);
#if _DEBUG
    WFSDeviceName devName;
    //WFSSetDeviceName(&devName, pDevInfo->devName.sStr);
    devName.nLen = STRNLEN(pDevInfo->devName.sStr, WFS_MAX_DEVICE_NAME_SIZE);
    strncpy(devName.sStr, pDevInfo->devName.sStr, WFS_MAX_DEVICE_NAME_SIZE);
    DeviceInfo *pFoundDevInfo = DeviceGetDeviceInfo(&devName);
    if (pFoundDevInfo != 0) {
        WFSKrnOutputErrorStr("Delete failed");
    }
#endif
    pDevInfo->bInUse = false;
    VolumeInfo *pVolInfo = pDevInfo->pVolInfo;
    if (pVolInfo){
        if (pVolInfo->nClientsMountedBy) {
            wkg.bMountedDeviceWasDetached = true;
            pVolInfo->bDetached = true;
        } else {
            VolumeRemoveFromMountedList(pDevInfo->pVolInfo);
            VolumeFree(pDevInfo->pVolInfo);
            pDevInfo->pVolInfo = 0;
        }
    }
    --wkg.nNumDevices;
Exit:
    return nResult;
}

