/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfskrn_Area.cpp - Module for managing contiguous disk areas


  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_Area.cpp,v $
  Revision 1.27  2008/12/18 09:03:10  kondo_masahiro
  Fixed function name and process from BCacheSetBlkAdr() to BCacheRemap()

  Revision 1.26  2008/12/17 01:37:27  kondo_masahiro
  Added function BCacheSetBlkAdr()

  Revision 1.25  2008/11/26 01:43:38  kondo_masahiro
  Fixed arguments of several functions in order to improve access speed.

  Revision 1.24  2008/10/31 09:51:11  ueno
  Modified WFSSrvAPIs to return WFSKRN_RESULT_RESOURCE_LIMIT_EXCEEDED when path cache is short of memory.

  Revision 1.23  2008/10/17 08:50:48  kondo_masahiro
  Added the new MEDIUM size category (40KB-320KB)

  Revision 1.22  2008/10/14 10:27:14  kondo_masahiro
  Added #define _DEBUG_BREAK_POINT

  Revision 1.21  2008/10/10 02:11:19  kondo_masahiro
  Fixed a bug to access large files

  Revision 1.20  2008/10/09 04:16:53  kondo_masahiro
  Fixed member of AreaHdr and AreaInfo

  Revision 1.19  2008/10/08 23:19:54  kondo_masahiro
  Added codes to access large size file data

  Revision 1.18  2008/10/03 08:37:56  kondo_masahiro
  Fixed codes for porting IOP

  Revision 1.17  2008/09/29 01:58:57  ueno
  Modified to call BCacheInvalidateBlks() in FreeBlkFree() to invalidate caches of user data.

  Revision 1.16  2008/09/28 23:31:16  kondo_masahiro
  Fixed codes to use medium size user block.

  Revision 1.15  2008/09/24 07:33:21  ueno
  Added AclHdr{} to AreaHdr{} to save configuration of accesslists in a disk.

  Revision 1.14  2008/08/27 23:23:28  ooizumi
  Fixed newline to CR+LF(Windows format).

  Revision 1.13  2008/08/27 09:41:24  paul
  Added stride to alloc, free functions

  Revision 1.12  2008/08/16 04:35:31  ueno
  Fixed hardcoding.

  Revision 1.11  2008/08/15 03:36:20  ueno
  Fixed stride.

  Revision 1.10  2008/08/14 08:11:40  kondo_masahiro
  Removed struct AreaHashHdr.

  Revision 1.9  2008/08/09 03:17:51  ueno
  Modified AreaFreeBlks() to call FreeBlkFree() if BTREE_BASED_ALLOCATOR == 1.

  Revision 1.8  2008/08/08 13:31:34  ueno
  Added comment.

  Revision 1.7  2008/08/07 10:46:33  ueno
  Modified for BTREE_BASED_ALLOCATOR (now debugging).

  Revision 1.6  2008/08/05 04:07:51  kondo_masahiro
  Added error handling for hash error and detach device.

  Revision 1.5  2008/07/23 04:14:52  ueno
  Modified to initialize accesslist in AreaOpen().

  Revision 1.4  2008/07/17 05:31:06  kondo_masahiro
  Added codes for debug.

  Revision 1.3  2008/07/08 23:48:25  paul
  Many Changes - made the block cache and associated allocator work on blocks instead of sectors. Removed handle-to-pointer mapping code - now using pointers to static structs instead of device or volume handles. Fixed some bugs in the list-based allocator.

  Revision 1.2  2008/05/12 19:05:44  paul
  Merged in Wayne's previous changes

  Revision 1.1  2008/04/22 04:39:08  kondo_masahiro
  Fixed code to transfer the wfskrn project to RM_SDK tree

  Revision 1.12  2008/04/28 20:20:26  wayne.wong
  Changed the interface to block alloc slightly to return WFSKrnResult.

  Revision 1.11  2008/04/26 05:53:49  wayne.wong
  Small fix. Function name.

  Revision 1.10  2008/04/26 01:24:41  wayne.wong
  Moved area functionality from Volume to Area module. Changed implementation
  to support small area sizes for a large mapped number of sectors. Added
  block wrapper calls for Area and BCache calls.

  Revision 1.9  2008/04/18 01:58:44  wayne.wong
  Created separate Area types for persistent (hdr) and non-persistent (info) types.
  Changed function names to be consistent with kernel naming convention.
  Defines are included to support the previous naming, but will be deleted in
  the future.

  Revision 1.7  2008/04/03 23:39:00  kondo_masahiro
  change spec of debug proc

  Revision 1.6  2008/02/28 07:01:41  kondo_masahiro
  Fixed code to accept PPC compiler

  Revision 1.4  2008/02/27 00:27:17  paul
  increased block cache size to run bigger tests

  Revision 1.3  2008/02/20 08:09:25  paul
  Added some more temporary code for block alloc/free in order to proceed with wfskrn_Dir.

  Revision 1.2  2007/12/27 11:00:05  paul
  Added various functions - still only a preliminary API to allow me to progress with
  directory implementation

  Revision 1.1  2007/11/16 05:45:15  paul
  first checkin
 *---------------------------------------------------------------------------*/

#include "wfskrn_Api.h"
#include "wfskrn_Area.h"
#include "wfskrn_Bcache.h"
#include "wfskrn_Volume.h"
#include "wfskrn_Permission_AccessList.h"

#if BTREE_BASED_ALLOCATOR
#include "wfskrn_FreeBlkAlloc.h"
#endif // BTREE_BASED_ALLOCATOR

#if SIMPLE_ALLOCATOR
void *AreaGetPhysicalBlk(AreaInfo *pAreaInfo, WFSBlkAdr nBlkAdr) {
    return &pAreaInfo->ah.pPtr[(nBlkAdr-1) * pAreaInfo->nBlkSize];
}
#endif


WFSKrnResult AreaPinFreeListHdr(AreaInfo *pAreaInfo, u32 nFlags, AreaFreeListHdr **ppAflh) {
    WFSBlkAdr nAbsStartBlkAdr = pAreaInfo->nAbsStartBlkAdr;
    BCacheEntry *pBce;
    WFSKrnResult nResult = BCacheAllocMetaBlk(pAreaInfo->pVolInfo, nAbsStartBlkAdr, BCACHE_FLAG_PINNED | nFlags, &pBce);
    if (nResult == WFSKRN_RESULT_OK) {
        u8* pBlkPtr = pBce->pBlkPtr;
        if (nAbsStartBlkAdr == 0) {
            pBlkPtr += sizeof(VolumeHdr);
        }
        *ppAflh = (AreaFreeListHdr *)(pBlkPtr + sizeof(AreaHdr));
    }
    else{
        *ppAflh = NULL;
    }
    return nResult;
}

void AreaUnpinFreeListHdr(AreaInfo *pAreaInfo) {
    BCacheUnpin(pAreaInfo->pVolInfo, pAreaInfo->nAbsStartBlkAdr);
}


WFSKrnResult AreaAllocBlks(AreaInfo *pAreaInfo, WFSBlkAdr nRefBlkAdr, u32 nLog2BlkSize, u32 nNumBlks, WFSBlkAdr *pBlkAdr, s32 nStride) {
    // This function allocates the requested number of physical blocks or fails.
    // nRefBlkAdr    (IN)  A reference block which we wish to allocate close to.
    // nNumBlks      (IN)  Number of blocks to allocate.
    // aBlkAdr       (OUT) Array to pass back the index of allocated blocks.

    // In the final version, this function should try to allocate sequentially, close to
    // the specified reference block address. (These two goals may be conflicting, so a
    // heuristic approach is needed).
    // The list of allocated block addresses should be written to the specified array.

    // The current version
    // Re-mapping is not implemented in this function, it should be a separate function.
    // ToDo: For now, assumes small blocks. May need to add a flag to specify large/small blocks.
    //WFSBlkAdr nrba = nRefBlkAdr; // for warning

#if _DEBUG_BREAK_POINT
    if(nRefBlkAdr == 463){
        int a; a=0;
    }
#endif

    dbga(OS_TPrintf("AreaAllocBlks()\n"));
    WFSKrnResult nResult = AreaPinFreeListHdr(pAreaInfo, BCACHE_FLAG_READ | BCACHE_FLAG_DIRTY, &pAreaInfo->pAflh);
    if(nResult < WFSKRN_RESULT_OK){
        goto Done;
    }
/*
    if(pAreaInfo->pVolInfo->pDevInfo->nFlags & WFS_DEVFLAG_UNUSABLE){ // device detached
        nResult = WFSKRN_RESULT_MEDIA_ERROR;
        goto Done;
    }
    else if(pAreaInfo->pVolInfo->pDevInfo->nFlags & WFS_DEVFLAG_NOT_INITIALIZED){ // hash error
        nResult = WFSKRN_RESULT_CORRUPTION;
        goto Done;
    }
    */
#if BTREE_BASED_ALLOCATOR

    nResult = FreeBlkAlloc(pAreaInfo, nRefBlkAdr, nNumBlks, pBlkAdr, nStride, nLog2BlkSize); // [check] hard-coding.


#elif LIST_BASED_ALLOCATOR
    u32 nI;
    for(nI=0; nI<nNumBlks; nI++) {
        AreaInfo subAreaInfo;
        nResult = AreaAllocSubArea(pAreaInfo, 1, &subAreaInfo, BCACHE_FLAG_READ | BCACHE_FLAG_DIRTY);
        if (nResult != WFSKRN_RESULT_OK) {
            if (nResult == WFSKRN_RESULT_DIRECTORY_QUOTA) {
                pBlkAdr = (WFSBlkAdr *)((u8*)pBlkAdr - nStride);
                AreaFreeBlks(pAreaInfo, nI, pBlkAdr, -nStride);
            }
            goto Done;
        }
        *pBlkAdr = subAreaInfo.nRelStartBlkAdr;
    #if _CHECK_BLK_ALLOCATION
        u32 nAllocState = (pAreaInfo->aBlkAllocState[*pBlkAdr>>5]>>(*pBlkAdr&31))&1;
        if (nAllocState == 1) {
            WFSKrnOutputErrorStr("Block already allocated!");
        }
        pAreaInfo->aBlkAllocState[*pBlkAdr>>5] |= (1<<(*pBlkAdr&31));
    #endif //_CHECK_BLK_ALLOCATION
        pBlkAdr = (WFSBlkAdr *)((u8*)pBlkAdr + nStride);
    }
#else //SIMPLE_ALLOCATOR
    if (pAreaInfo->ah.nNumFreeBlks < nNumBlks) {
        return WFSKRN_RESULT_DIRECTORY_QUOTA;
    }
    pAreaInfo->ah.nNumFreeBlks -= nNumBlks;
    u32 nI;
    for(nI=0; nI<nNumBlks; nI++) {
        *pBlkAdr = pAreaInfo->ah.nFirstFreeBlkAdr;
        WFSBlkAdr *pNextFreeBlock = (WFSBlkAdr *)AreaGetPhysicalBlk(pAreaInfo, *pBlkAdr);
        pAreaInfo->ah.nFirstFreeBlkAdr = *pNextFreeBlock;
        pBlkAdr = (WFSBlkAdr *)((u8*)pBlkAdr + nStride);
    }
#endif //BTREE_BASED_ALLOCATOR/LIST_BASED_ALLOCATOR/SIMPLE_ALLOCATOR
Done:
    AreaUnpinFreeListHdr(pAreaInfo);
    return nResult;
}

WFSKrnResult AreaFreeBlks(AreaInfo *pAreaInfo, u32 nLog2BlkSize, u32 nNumBlks, WFSBlkAdr *pBlkAdr, s32 nStride) {
    // This function frees the physical blocks passed in.
    dbga(OS_TPrintf("AreaFreeBlks()\n"));
    WFSKrnResult nResult = AreaPinFreeListHdr(pAreaInfo, BCACHE_FLAG_READ | BCACHE_FLAG_DIRTY, &pAreaInfo->pAflh);
    if(nResult < WFSKRN_RESULT_OK){
        goto Done;
    }
    /*
    if(pAreaInfo->pVolInfo->pDevInfo->nFlags & WFS_DEVFLAG_UNUSABLE){ // device detached
        nResult = WFSKRN_RESULT_MEDIA_ERROR;
        goto Done;
    }
    else if(pAreaInfo->pVolInfo->pDevInfo->nFlags & WFS_DEVFLAG_NOT_INITIALIZED){ // hash error
        nResult = WFSKRN_RESULT_CORRUPTION;
        goto Done;
    }
    */
#if BTREE_BASED_ALLOCATOR
    nResult = FreeBlkFree(pAreaInfo, nNumBlks, pBlkAdr, nStride, nLog2BlkSize);
#elif LIST_BASED_ALLOCATOR
    AreaInfo subAreaInfo;
    subAreaInfo.pVolInfo = pAreaInfo->pVolInfo;
    u32 nI;
    for (nI=0; nI<nNumBlks; nI++) {
    #if _CHECK_BLK_ALLOCATION
        u32 nAllocState = (pAreaInfo->aBlkAllocState[*pBlkAdr>>5]>>(*pBlkAdr&31))&1;
        if (nAllocState == 0) {
            WFSKrnOutputErrorStr("Block already free!");
        }
        pAreaInfo->aBlkAllocState[*pBlkAdr>>5] &= ~(1<<(*pBlkAdr&31));
    #endif //_CHECK_BLK_ALLOCATION
        subAreaInfo.nRelStartBlkAdr = *pBlkAdr; // * pAreaInfo->pAflh->aFreeList[0].nAreaSize;
        nResult = AreaFreeSubArea(pAreaInfo, &subAreaInfo);
    #if _DEBUG
        if (nResult != WFSKRN_RESULT_OK) {
            return nResult;
            //WFSKrnOutputErrorCode(nResult);
        }
    #endif
        pBlkAdr = (WFSBlkAdr *)((u8*)pBlkAdr + nStride);
    }
#else //SIMPLE_ALLOCATOR
    pAreaInfo->ah.nNumFreeBlks += nNumBlks;
    u32 nI;
    for(nI=0; nI<nNumBlks; nI++) {
        WFSBlkAdr *pNextFreeBlock = (WFSBlkAdr *)AreaGetPhysicalBlk(pAreaInfo, *pBlkAdr);
        *pNextFreeBlock = pAreaInfo->ah.nFirstFreeBlkAdr;
        pAreaInfo->ah.nFirstFreeBlkAdr = *pBlkAdr;
        pBlkAdr = (WFSBlkAdr *)((u8*)pBlkAdr + nStride);
    }
#endif //BTREE_BASED_ALLOCATOR/LIST_BASED_ALLOCATOR/SIMPLE_ALLOCATOR
Done:;
    AreaUnpinFreeListHdr(pAreaInfo);
    return nResult;
}


WFSKrnResult AreaInvalidateBlks(AreaInfo *pAreaInfo, u32 nNumBlks, WFSBlkAdr *pBlkAdr, s32 nStride)
{
    WFSKrnResult nResult = WFSKRN_RESULT_UNKNOWN;
    int nI;
    for(nI=0;nI<nNumBlks;nI++){
        nResult = AreaBlkInvalidate(pAreaInfo, *pBlkAdr);
        if(nResult < WFSKRN_RESULT_OK){
            WFSKrnOutputErrorCode(nResult);
            break;
        }
        pBlkAdr = (WFSBlkAdr *)((u8*)pBlkAdr + nStride);
    }
    return nResult;
}


// Initialize an area.  Current only support one level of areas.
WFSKrnResult AreaHdrInit(AreaInfo *pAreaInfo,        // area to initialize
                         u32 nNumSectors,            // number of sectors to map on device
                         u32 numAreaSizes,           // number of different sub-area sizes
                         u32 areaSizes[]);           // array of sub-area sizes (in sectors)

// Get the area's header.
WFSKrnResult AreaGetHdr(AreaInfo *pAreaInfo,            // area to get info about
                        AreaHdr  *pAreaHdr);            // returned area header


#if BTREE_BASED_ALLOCATOR

#elif LIST_BASED_ALLOCATOR

    u32 gaAllocationSizes[] = {   // Multiples of the small block size
        1, (1<<2), (1<<4), (1<<6), (1<<8), (1<<10), (1<<12), (1<<14), (1<<16), (1<<18)
    };
    u32 gnNumAreaSizes = sizeof(gaAllocationSizes)/sizeof(u32);

    static
    WFSBlkAdr AreaEntryBlkAdr(AreaInfo *pAreaInfo, AreaEntryIdx nAreaEntryIdx, u32 *pOfs) {
        dbga(OS_TPrintf("AreaEntryBlkAdr()\n"));
        u32 nBlkAdr;
        if (nAreaEntryIdx < pAreaInfo->nNumEntriesInFirstBlk) {
            nBlkAdr = 0;
            *pOfs = sizeof(AreaHdr) + sizeof(AreaFreeListHdr) + (nAreaEntryIdx * sizeof(AreaEntry));
            if (pAreaInfo->nAbsStartBlkAdr == VOLUME_HDR_BLK_ADR) {
                *pOfs += sizeof(VolumeHdr);
            }
        } else {
            nAreaEntryIdx -= pAreaInfo->nNumEntriesInFirstBlk;
            nBlkAdr =  (nAreaEntryIdx / pAreaInfo->nNumEntriesPerBlk) + 1;
            *pOfs   = ((nAreaEntryIdx % pAreaInfo->nNumEntriesPerBlk) * sizeof(AreaEntry)) + sizeof(WFSMetaDataHdr);
            //nBlkAdr = (nAreaEntryIdx / pAreaInfo->nNumEntriesPerBlk) + 1;
            //*pOfs   = (nAreaEntryIdx % pAreaInfo->nNumEntriesPerBlk) * sizeof(AreaEntry);
        }
        return nBlkAdr;
    }


    // Loads the specified entry into memory, and returns a pointer to it.
    // Otherise NULL is returned to indicate an error. Assumes the device+volume is
    // already mapped. No checks here if nAreaEntryIdx is within bounds.
    // Also assumes that we don't need to pin the page (i.e., no one else is going
    // to access the BCache before the caller is done.
    static
    WFSKrnResult _AreaEntryMap(AreaInfo *pAreaInfo, AreaEntryIdx nAreaEntryIdx, u32 nFlags, AreaEntry** ppAe) {
        dbga(OS_TPrintf("_AreaEntryMap()\n"));
        static int c=0; ++c;
#if _DEBUG_BREAK_POINT
        if(c>=43) {
            int a;a=0;
        }
#endif
        u32 nOfs;
        u32 nAbsBlkAdr = pAreaInfo->nAbsStartBlkAdr + AreaEntryBlkAdr(pAreaInfo, nAreaEntryIdx, &nOfs);
        //u32 nFlags = (bModify?BCACHE_FLAG_DIRTY:0) | BCACHE_FLAG_READ; // no pin; assumes no interruptions to caller
        if (!nAbsBlkAdr){
            nFlags &= ~BCACHE_FLAG_PINNED;
        }
        BCacheEntry *pBce;
        //u8 *pBlkPtr;
        WFSKrnResult nResult = BCacheAllocMetaData(pAreaInfo->pVolInfo, nAbsBlkAdr, nFlags, &pBce);

    #if _DEBUG
        if (!pBce->pBlkPtr || nResult < WFSKRN_RESULT_OK) {
            *ppAe = NULL;
            return nResult;
        }
        if (nOfs > (pAreaInfo->nBlkSize - sizeof(AreaEntry))) {
            *ppAe = NULL;
            return WFSKRN_RESULT_AREA_BCACHE_ALLOC;
        }
    #endif
        //pBlkPtr += nOfs;
        //return (AreaEntry*)pBlkPtr;
        *ppAe = (AreaEntry*)(pBce->pBlkPtr + nOfs);
        return WFSKRN_RESULT_OK;
    }


    static
    WFSKrnResult _AreaEntryUnpin(AreaInfo *pAreaInfo, AreaEntryIdx nAreaEntryIdx){
        dbga(OS_TPrintf("_AreaEntryUnmap()\n"));
        u32 nOfs;
        u32 nAbsBlkAdr = pAreaInfo->nAbsStartBlkAdr + AreaEntryBlkAdr(pAreaInfo, nAreaEntryIdx, &nOfs);
        if (!nAbsBlkAdr){
            return WFSKRN_RESULT_OK;
        }
        return BCacheUnpin(pAreaInfo->pVolInfo, nAbsBlkAdr);
    }


    static
    u32 AreaEntryGetPinCount(AreaInfo *pAreaInfo, AreaEntryIdx nAreaEntryIdx){
        dbga(OS_TPrintf("_AreaEntryGetPinCount()\n"));
        u32 nOfs;
        u32 nAbsBlkAdr = pAreaInfo->nAbsStartBlkAdr + AreaEntryBlkAdr(pAreaInfo, nAreaEntryIdx, &nOfs);
        return BCacheGetPinCount(pAreaInfo->pVolInfo, nAbsBlkAdr);
    }


    static
    u32 _AreaGetAbsBlkAdr(AreaInfo *pAreaInfo, AreaEntryIdx nAreaEntryIdx) {
        dbga(OS_TPrintf("_AreaGetAbsBlkAdr()\n"));
        u32 nOfs;
        return pAreaInfo->nAbsStartBlkAdr + AreaEntryBlkAdr(pAreaInfo, nAreaEntryIdx, &nOfs);
    }


    // Add to the head of the free list.  Assumes area header should be already be in-core memory.
    // Its page should be marked dirty if a successful add. No checks on if the entry is already
    // in another (or this) list.
    static
    WFSKrnResult _AreaListAdd(AreaInfo *pAreaInfo, u32 nFreeListIdx, AreaEntryIdx nAreaEntryIdx) {
        dbga(OS_TPrintf("_AreaListAdd()\n"));
        AreaEntry *pEntry;
        WFSKrnResult nResult = _AreaEntryMap(pAreaInfo, nAreaEntryIdx, BCACHE_FLAG_READ | BCACHE_FLAG_DIRTY | BCACHE_FLAG_PINNED, &pEntry);
        if(nResult < WFSKRN_RESULT_OK){
            return nResult;
        }
        AreaList *pList = &(pAreaInfo->pAflh->aFreeList[nFreeListIdx]);
        if (pList->nLen > 0) {                        // link to rest of list?
            AreaEntry *pNextEntry;
            nResult = _AreaEntryMap(pAreaInfo, pList->nHeadIdx, BCACHE_FLAG_READ | BCACHE_FLAG_DIRTY, &pNextEntry);
            if(nResult < WFSKRN_RESULT_OK){
                return nResult;
            }
            pNextEntry->nPrevIdx = nAreaEntryIdx;
            pEntry->nNextIdx = pList->nHeadIdx;
        } else {                                    // else, just attach as head
            pEntry->nNextIdx = AREA_NULL;
        }
        pEntry->nPrevIdx = AREA_NULL;
        pList->nHeadIdx = nAreaEntryIdx;
        ++pList->nLen;
        nResult = _AreaEntryUnpin(pAreaInfo, nAreaEntryIdx);
        return WFSKRN_RESULT_OK;
    }


    // Delete entry from the free list.  Assumes volume's MR should be already be in-core memory.
    // Its page should be marked dirty if a successful delete.
    static
    WFSKrnResult _AreaListDelete(AreaInfo *pAreaInfo, u32 nFreeListIdx, AreaEntryIdx nAreaEntryIdx) {
        dbga(OS_TPrintf("_AreaListDelete()\n"));
#if _DEBUG_BREAK_POINT
        static int c=0; ++c; if (c>=3) {
            int a;a=0;
        }
#endif
        AreaEntry *pEntry;
        WFSKrnResult nResult = _AreaEntryMap(pAreaInfo, nAreaEntryIdx, BCACHE_FLAG_READ | BCACHE_FLAG_DIRTY | BCACHE_FLAG_PINNED, &pEntry);
        if(nResult < WFSKRN_RESULT_OK){
            return nResult;
        }
        if (pEntry->nSizeIdx != nFreeListIdx) {
            return WFSKRN_RESULT_AREA_CORRUPT_HDR;
        }

        AreaList *pList = &(pAreaInfo->pAflh->aFreeList[nFreeListIdx]);
        if (pList->nHeadIdx == nAreaEntryIdx) {
            pList->nHeadIdx = pEntry->nNextIdx;
        }
        if (pEntry->nNextIdx != AREA_NULL) {
            AreaEntry *pNextEntry;
            nResult = _AreaEntryMap(pAreaInfo, pEntry->nNextIdx, BCACHE_FLAG_READ | BCACHE_FLAG_DIRTY, &pNextEntry);
            if(nResult < WFSKRN_RESULT_OK){
                return nResult;
            }
            pNextEntry->nPrevIdx = pEntry->nPrevIdx;
        }
        if (pEntry->nPrevIdx != AREA_NULL) {
            AreaEntry *pPrevEntry;
            nResult = _AreaEntryMap(pAreaInfo, pEntry->nPrevIdx, BCACHE_FLAG_READ | BCACHE_FLAG_DIRTY, &pPrevEntry);
            if(nResult < WFSKRN_RESULT_OK){
                return nResult;
            }
            pPrevEntry->nNextIdx = pEntry->nNextIdx;
        }
        pEntry->nPrevIdx = AREA_NULL;
        pEntry->nNextIdx = AREA_NULL;
        --pList->nLen;
        nResult = _AreaEntryUnpin(pAreaInfo, nAreaEntryIdx);
        return nResult;
    }


    static
    void _AreaListPrint(AreaInfo *pAreaInfo, AreaEntryIdx start) {
        dbga(OS_TPrintf("_AreaListPrint()\n"));
        AreaEntryIdx index = start;
        while (AREA_NULL != index) {
            AreaEntry *pEntry;
            WFSKrnResult nResult = _AreaEntryMap(pAreaInfo, index, BCACHE_FLAG_READ, &pEntry);
            if (pEntry && nResult >= WFSKRN_RESULT_OK) {
                OS_TPrintf("\t%d: attr=0x%02x, next=%d, prev=%d\n", index, pEntry->nAttributes, pEntry->nNextIdx, pEntry->nPrevIdx);
                index = pEntry->nNextIdx;
            } else {
                OS_TPrintf("ERROR: cannot allocate page (%d)\n", nResult);
                break;
            }
        }
    }


    // Check that freelists are consistent. Print out the free list.
    static
    WFSKrnResult _AreaFreelistCheck(AreaInfo *pAreaInfo) {
        dbga(OS_TPrintf("_AreaFreelistCheck()\n"));
#if _DEBUG_BREAK_POINT
        static int c=0; ++c; if (c>=3) {
            int a;a=0;
        }
#endif
        //u8 *pMem = (u8*)BCacheAlloc(pAreaInfo->pVolInfo, pAreaInfo->nAbsStartBlkAdr, BCACHE_FLAG_READ); //true, false, false);
        // OS_TPrintf("\nvolume 0x%8x\n", pVolInfo);
        u32 nSizeIdx;
        for (nSizeIdx=0; nSizeIdx<gnNumAreaSizes; nSizeIdx++) {
            AreaList *pList = &pAreaInfo->pAflh->aFreeList[nSizeIdx];
            // OS_TPrintf("list %2d, length=%d, head=%d\n", nSizeIdx, pList->nLen, pList->nHeadIdx);
            if (pList->nLen > 0) {
                u32 nNumEntries = 0;
                AreaEntryIdx nEntryIdx = pList->nHeadIdx;
                while (AREA_NULL != nEntryIdx) {
                    nNumEntries++;
                    AreaEntry *pEntry;
                    WFSKrnReturnOnError(_AreaEntryMap(pAreaInfo, nEntryIdx, BCACHE_FLAG_READ, &pEntry));
                    if (pEntry) {
                        nEntryIdx = pEntry->nNextIdx;
                        if (pEntry->nAttributes & AREA_ALLOCATED) {
                            WFSKrnOutputErrorStr("non-free block on free list");
                            return WFSKRN_RESULT_AREA_CORRUPT_HDR;
                        }
                    } else {
                        break;
                    }
                }
                if (nNumEntries != pList->nLen) {
                    OS_TPrintf("ERROR: incorrect length. %d != %d\n", nNumEntries, pList->nLen);
                    _AreaListPrint(pAreaInfo, pList->nHeadIdx);
                    WFSKrnOutputErrorCode(WFSKRN_RESULT_AREA_CORRUPT_HDR);
                    return WFSKRN_RESULT_AREA_CORRUPT_HDR;
                }
            } else if (pList->nHeadIdx != AREA_NULL) {
                OS_TPrintf("ERROR: incorrect head. head=%d,  len=%d\n", pList->nHeadIdx, pList->nLen);
                _AreaListPrint(pAreaInfo, pList->nHeadIdx);
                WFSKrnOutputErrorCode(WFSKRN_RESULT_AREA_CORRUPT_HDR);
                return WFSKRN_RESULT_AREA_CORRUPT_HDR;
            }
        }
        return WFSKRN_RESULT_OK;
    }


    WFSKrnResult AreaFreelistPrint(AreaInfo *pAreaInfo) {
        return _AreaFreelistCheck(pAreaInfo);
    }


    // Allocate an sub-area. Assumes nSizeIdx < gnNumAreaSizes. Check above here?
    static
    WFSKrnResult _AreaAllocate(AreaInfo *pAreaInfo, u32 nSizeIdx, AreaEntryIdx *pEntryIndex, u32 nFlags) {
        dbga(OS_TPrintf("_AreaAllocate()\n"));
#if _DEBUG_BREAK_POINT
        static int c=0;++c;if(c>=6){
            int a;a=0;
        }
#endif
        AreaEntryIdx freeIndex;                   // entry of free area to return
        WFSKrnResult nResult;
        if (pAreaInfo->pAflh->aFreeList[nSizeIdx].nLen > 0) {       // area available?
            freeIndex = pAreaInfo->pAflh->aFreeList[nSizeIdx].nHeadIdx;
            nResult = _AreaListDelete(pAreaInfo, nSizeIdx, freeIndex);
            if (WFSKRN_RESULT_OK != nResult) {
                return nResult;
            }
        } else {  // else, try to split a large free sub-area
            if (nSizeIdx < (gnNumAreaSizes-1)) {// try to allocate from larger free list?
                nResult = _AreaAllocate(pAreaInfo, nSizeIdx+1, &freeIndex, nFlags);
                if (WFSKRN_RESULT_OK == nResult) {    // split larger area into n smaller ones
                    *pEntryIndex = freeIndex;        // first child is returned
                    AreaList *pList = &pAreaInfo->pAflh->aFreeList[nSizeIdx];
                    u32 nNumSiblings = pAreaInfo->pAflh->aFreeList[nSizeIdx+1].nAreaSize/pList->nAreaSize;
                    u32 nStride = pList->nAreaSize/pAreaInfo->pAflh->aFreeList[0].nAreaSize;
                    AreaEntryIdx nSiblingIdx;
                    u32 nS;
                    for (nS=1, nSiblingIdx=*pEntryIndex+nStride; nS<nNumSiblings; nS++, nSiblingIdx+=nStride) {
                        AreaEntry *pAreaEntry;
                        WFSKrnResult nResult = _AreaEntryMap(pAreaInfo, nSiblingIdx, nFlags, &pAreaEntry);
                        if(nResult < WFSKRN_RESULT_OK){
                            return nResult;
                        }
                        pAreaEntry->nSizeIdx = nSizeIdx;
                        pAreaEntry->nAttributes = AREA_START;
                        _AreaListAdd(pAreaInfo, nSizeIdx, nSiblingIdx);
                    }
                } else {
                    return nResult;
                }
            } else {
                return WFSKRN_RESULT_DIRECTORY_QUOTA;
                //return WFSKRN_RESULT_AREA_ALLOC;
            }
        }

        // Wayne - calling routine will initialize the area's header?
        *pEntryIndex = freeIndex;
        AreaEntry *pAreaEntry;
        nResult = _AreaEntryMap(pAreaInfo, freeIndex, BCACHE_FLAG_READ | BCACHE_FLAG_DIRTY, &pAreaEntry);
        if(nResult < WFSKRN_RESULT_OK){
            return nResult;
        }

        pAreaEntry->nSizeIdx  = nSizeIdx;
        pAreaEntry->nAttributes = (0==freeIndex) ? AREA_OVERHEAD : AREA_EMPTY;
        pAreaEntry->nAttributes |= AREA_ALLOCATED | AREA_START;
        pAreaEntry->nNextIdx       = AREA_NULL;
        pAreaEntry->nPrevIdx       = AREA_NULL;
        return WFSKRN_RESULT_OK;
    }


    // Recursive core routine to free a volume area.  Coalesce into a larger area if
    // possible. The larger area Assumes that *pAreaEntry is not currently on any free list.
    //
    // Assumes this call is not interrupted by another user of the BCache (i.e., no need to
    // pin pages).
    static
    WFSKrnResult _AreaFree(AreaInfo *pAreaInfo, s16 nSizeIdx, AreaEntryIdx nEntryIdx) {
        dbga(OS_TPrintf("_AreaFree()\n"));
#if _DEBUG_BREAK_POINT
        static int c=0;++c;if (c>=6){
            int a;a=0;
        }
#endif
        // try to coalesce aggressively
        AreaList *pList = &pAreaInfo->pAflh->aFreeList[nSizeIdx];

        // Wayne - can we move pDevInfo and pAreaEntry above to eliminate multiple calls?
        AreaEntry *pAreaEntry;
        WFSKrnResult nResult = _AreaEntryMap(pAreaInfo, nEntryIdx, BCACHE_FLAG_READ | BCACHE_FLAG_DIRTY, &pAreaEntry);
        if(nResult < WFSKRN_RESULT_OK){
            return nResult;
        }
        // Wayne - need to pin pAreaEntry?

        if ((nSizeIdx < (gnNumAreaSizes-1)) &&             // larger area freelist exists?
            (pList->nAreaSize < (pAreaInfo->ah.nNumBlks>>1))) {  // larger area possible?
            u32 nLargerSize = pAreaInfo->pAflh->aFreeList[nSizeIdx+1].nAreaSize;
            u32 nStride = pList->nAreaSize / pAreaInfo->pAflh->aFreeList[0].nAreaSize;
            u32 nNumSiblings = nLargerSize / pList->nAreaSize;
            AreaEntryIdx nFirstSiblingIdx = nEntryIdx - (nEntryIdx % (nLargerSize/pAreaInfo->pAflh->aFreeList[0].nAreaSize));
            AreaEntryIdx nSiblingIdx;
            u32 nS;
            for (nS=0, nSiblingIdx=nFirstSiblingIdx; nS<nNumSiblings; ++nS) {  // for each siblings
                if (nSiblingIdx >= pAreaInfo->ah.nMaxNumAreas) break;
                if (nSiblingIdx != nEntryIdx) {
                    AreaEntry *pSiblingEntry;
                    nResult = _AreaEntryMap(pAreaInfo, nSiblingIdx, BCACHE_FLAG_READ, &pSiblingEntry);
                    if(nResult < WFSKRN_RESULT_OK){
                        return nResult;
                    }
                    if ((pSiblingEntry->nAttributes & AREA_START) &&   // start of an area
                        (pSiblingEntry->nSizeIdx == nSizeIdx)) {           // and same size
                        if (pSiblingEntry->nAttributes & AREA_ALLOCATED) break;
                    } else {
                        break;
                    }
                }
                nSiblingIdx += nStride;
            }

            if (nS==nNumSiblings) {   // all siblings are free?
                for (nS=0, nSiblingIdx=nFirstSiblingIdx; nS<nNumSiblings; nS++, nSiblingIdx+=nStride) {
                    AreaEntry *pSiblingEntry;
                    WFSKrnResult nResult = _AreaEntryMap(pAreaInfo, nSiblingIdx, BCACHE_FLAG_READ | BCACHE_FLAG_DIRTY, &pSiblingEntry);
                    if(nResult < WFSKRN_RESULT_OK){
                        return nResult;
                    }
                    if (nSiblingIdx != nEntryIdx) {   // remove siblings from list; nEntryIdx is not in the list.
                        nResult = _AreaListDelete(pAreaInfo, nSizeIdx, nSiblingIdx);
                        if (WFSKRN_RESULT_OK != nResult) return nResult;
                    }
                    if (nSiblingIdx == nFirstSiblingIdx) {
                        pSiblingEntry->nAttributes &= ~AREA_ALLOCATED;
                    } else {
                        pSiblingEntry->nAttributes = AREA_EMPTY;
                    }
                    // Wayne - anything with AREA_FIRST?  or only on allocations?
                }
                return _AreaFree(pAreaInfo, nSizeIdx+1, nFirstSiblingIdx);
                //nResult = _AreaFree(pAreaInfo, nSizeIdx+1, nFirstSiblingIdx);
                //if (WFSKRN_RESULT_OK != nResult) return nResult;
            }
        }

        // not able to coalesce
        nResult = _AreaListAdd(pAreaInfo, nSizeIdx, nEntryIdx);
        pAreaEntry->nAttributes &= ~AREA_ALLOCATED;
        pAreaEntry->nSizeIdx = nSizeIdx;
        return nResult;
    }


    // Allocate an area large enough to fit the desired request.  Actual number of free
    // sectors in the area may be less due to overhead (e.g., first area on disk has
    // the volume header and entry table. The area's header should tell the actual
    // usable starting nSector and the ending nSector.
    //
    // Actually, current hack is to allocate an initial area covering the overhead.
    // Thus, all subsequent request should be no overhead.
    WFSKrnResult AreaAllocSubArea(AreaInfo *pAreaInfo, WFSBlkAdr nNumBlks, AreaInfo *pSubAreaInfo, u32 nFlags) {
        dbga(OS_TPrintf("AreaAllocSubArea()\n"));
#if _DEBUG_BREAK_POINT
        static int c=0;++c; if(c>=3){
            int a;a=0;
        }
#endif
        VolumeInfo *pVolInfo = pAreaInfo->pVolInfo;
        //DeviceInfo *pDevInfo = pVolInfo->pDevInfo;
        //u8 *pMem = (u8*)BCacheAlloc(pVolInfo, pAreaInfo->nAbsStartBlkAdr, BCACHE_FLAG_READ | BCACHE_FLAG_PINNED); //true, true, false);
        //AreaInfo *pAreaInfo = (pAreaInfo->nAbsStartBlkAdr==VOLUME_HDR_BLK_ADR) ? (AreaHdr*)(pMem + sizeof(VolumeHdr)) : (AreaHdr*)pMem;
        WFSKrnResult nResult = WFSKRN_RESULT_OK;
        u32 nSizeIdx = 0;
        while ((nSizeIdx < gnNumAreaSizes) &&        // find matching area size
            (nNumBlks > pAreaInfo->pAflh->aFreeList[nSizeIdx].nAreaSize)) {
            ++nSizeIdx;
        }
        if (nSizeIdx < gnNumAreaSizes) {
            AreaEntryIdx nEntryIdx;
            nResult = _AreaAllocate(pAreaInfo, nSizeIdx, &nEntryIdx, nFlags);
            u32 nNumBlksAllocated = pAreaInfo->pAflh->aFreeList[nSizeIdx].nAreaSize;
            if (WFSKRN_RESULT_OK == nResult) {
                pSubAreaInfo->pParent = pAreaInfo;
                pSubAreaInfo->pVolInfo = pVolInfo;
                pSubAreaInfo->nRelStartBlkAdr = (WFSBlkAdr)nEntryIdx;// * pAreaInfo->pAflh->aFreeList[0].nAreaSize;
                pSubAreaInfo->ah.nNumBlks = nNumBlks;
                //pSubAreaInfo->nNumSectorsPerSmallBlk = pAreaInfo->pAflh->aFreeList[0].nAreaSize;
            }
            // Free up any unused portion of the area just allocated
            while(nSizeIdx > 0) {
                --nSizeIdx;
                while((nNumBlksAllocated - nNumBlks) >= pAreaInfo->pAflh->aFreeList[nSizeIdx].nAreaSize) {
                    nNumBlksAllocated -= pAreaInfo->pAflh->aFreeList[nSizeIdx].nAreaSize;
                    _AreaFree(pAreaInfo, nSizeIdx, nEntryIdx + nNumBlksAllocated);
                }
            }
        } else {
            nResult = WFSKRN_RESULT_AREA_INVALID_PARAMETER;  // no area large enough
        }
        if (WFSKRN_RESULT_OK == nResult) {
            BCacheDirty(pVolInfo, pAreaInfo->nAbsStartBlkAdr);
            //nResult = BCacheUnpin(pVolInfo, pAreaInfo->nAbsStartBlkAdr);
        } else {
            //BCacheUnpin(pVolInfo, pAreaInfo->nAbsStartBlkAdr);
        }
        WFSKrnReturnOnError(_AreaFreelistCheck(pAreaInfo));
        return nResult;
    }


    WFSKrnResult AreaFreeSubArea(AreaInfo *pAreaInfo, AreaInfo *pSubAreaInfo) {
        dbga(OS_TPrintf("AreaFreeSubArea()\n"));
#if _DEBUG_BREAK_POINT
        static int c=0; ++c; if (c>=831) {
            int a;a=0;
        }
#endif
        WFSKrnReturnOnError(_AreaFreelistCheck(pAreaInfo));
        //VolumeInfo *pVolInfo = pAreaInfo->pVolInfo;
        //DeviceInfo *pDevInfo = pVolInfo->pDevInfo;
        //u8 *pMem = (u8*)BCacheAlloc(pVolInfo, pAreaInfo->nAbsStartBlkAdr, BCACHE_FLAG_READ | BCACHE_FLAG_PINNED | BCACHE_FLAG_DIRTY); //true, true, true);
        //if (!pMem) return WFSKRN_RESULT_AREA_BCACHE_ALLOC;
        //AreaInfo *pAreaInfo = (pAreaInfo->pParent==NULL) ? (AreaHdr*)(pMem + sizeof(VolumeHdr)) : (AreaHdr*)pMem;
        AreaEntryIdx nEntryIdx = pSubAreaInfo->nRelStartBlkAdr / pAreaInfo->pAflh->aFreeList[0].nAreaSize;
        AreaEntry *pAreaEntry;
        WFSKrnResult nResult = _AreaEntryMap(pAreaInfo, nEntryIdx, BCACHE_FLAG_READ | BCACHE_FLAG_DIRTY, &pAreaEntry);
        if(nResult < WFSKRN_RESULT_OK){
            return nResult;
        }
        //WFSKrnReturnOnError(_AreaFreelistCheck(pAreaInfo));
        nResult = _AreaFree(pAreaInfo, pAreaEntry->nSizeIdx, nEntryIdx);
        WFSKrnReturnOnError(_AreaFreelistCheck(pAreaInfo));
        //nResult = _AreaEntryUnpin(pAreaInfo, nEntryIdx);
    //Done:
        //BCacheUnpin(pVolInfo, pAreaInfo->nAbsStartBlkAdr);
        return nResult;
    }


#else //SIMPLE_ALLOCATOR

#endif //LIST_BASED_ALLOCATOR


WFSKrnResult AreaBlkStore(AreaInfo *pAreaInfo, WFSBlkAdr nBlkAdr) {
    dbga(OS_TPrintf("AreaBlkStore()\n"));
    return BCacheStore(pAreaInfo->pVolInfo, pAreaInfo->nAbsStartBlkAdr + nBlkAdr);
}


WFSKrnResult AreaBlkFlush(AreaInfo *pAreaInfo, WFSBlkAdr nBlkAdr) {
    dbga(OS_TPrintf("AreaBlkFlush()\n"));
#if _CHECK_BLK_USAGE
    u8 *pPtr;
    AreaBlkGetPtr(pAreaInfo, nBlkAdr, (void **)&pPtr);
#endif
    WFSKrnResult nResult = BCacheFlush(pAreaInfo->pVolInfo, pAreaInfo->nAbsStartBlkAdr + nBlkAdr);
#if _CHECK_BLK_USAGE
    memset(pPtr, 0xff, AREA_PAGE_SIZE);
#endif
    return nResult;
}


WFSKrnResult AreaBlkInvalidate(AreaInfo *pAreaInfo, WFSBlkAdr nBlkAdr){
    dbga(OS_TPrintf("AreaBlkFlush()\n"));
#if _CHECK_BLK_USAGE
    u8 *pPtr;
    AreaBlkGetPtr(pAreaInfo, nBlkAdr, (void **)&pPtr);
#endif
    dbga(OS_TPrintf("AreaBlkFlush()\n"));
    WFSKrnResult nResult = BCacheInvalidate(pAreaInfo->pVolInfo, pAreaInfo->nAbsStartBlkAdr + nBlkAdr);
#if _CHECK_BLK_USAGE
    memset(pPtr, 0xff, AREA_PAGE_SIZE);
#endif
    return nResult;
}


WFSKrnResult AreaBlkDirty(AreaInfo *pAreaInfo, WFSBlkAdr nBlkAdr) {
    dbga(OS_TPrintf("AreaBlkDirty()\n"));
    return BCacheDirty(pAreaInfo->pVolInfo, pAreaInfo->nAbsStartBlkAdr + nBlkAdr);
}


WFSKrnResult AreaBlkPin(AreaInfo *pAreaInfo, WFSBlkAdr nBlkAdr) {
    dbga(OS_TPrintf("AreaBlkPin()\n"));
    return BCachePin(pAreaInfo->pVolInfo, pAreaInfo->nAbsStartBlkAdr + nBlkAdr);
}


WFSKrnResult AreaBlkUnpin(AreaInfo *pAreaInfo, WFSBlkAdr nBlkAdr) {
    dbga(OS_TPrintf("AreaBlkUnpin()\n"));
    WFSKrnResult nResult = BCacheUnpin(pAreaInfo->pVolInfo, pAreaInfo->nAbsStartBlkAdr + nBlkAdr);
#if _CHECK_BLK_USAGE
// if nPinCount become 0, the cache data is crashed. (for debug)
    if(!BCacheGetPinCount(pAreaInfo->pVolInfo, pAreaInfo->nAbsStartBlkAdr + nBlkAdr)){
        AreaBlkFlush(pAreaInfo, nBlkAdr);
    }
#endif
    return nResult;
}


u32 AreaBlkGetPinCount(AreaInfo *pAreaInfo, WFSBlkAdr nBlkAdr){
    dbga(OS_TPrintf("AreaBlkUnpin()\n"));
    return BCacheGetPinCount(pAreaInfo->pVolInfo, pAreaInfo->nAbsStartBlkAdr + nBlkAdr);
}


WFSKrnResult AreaBlkGetPtr(AreaInfo *pAreaInfo, WFSBlkAdr nBlkAdr, void** ppPtr){
    dbga(OS_TPrintf("AreaBlkGetPtr()\n"));
    u8 *pMem = (u8*)BCachePointer(pAreaInfo->pVolInfo, pAreaInfo->nAbsStartBlkAdr + nBlkAdr );
    if((s32)pMem < WFSKRN_RESULT_OK){
        return WFSKRN_RESULT_AREA_INVALID_PARAMETER;
    }
    *ppPtr = pMem;
    return WFSKRN_RESULT_OK;
}


WFSKrnResult AreaBlkSetValidSize(AreaInfo *pAreaInfo, WFSBlkAdr nBlkAdr, u32 nValidSize){
    dbga(OS_TPrintf("AreaBlkSetValidSize()\n"));
    return BCacheSetValidSize(pAreaInfo->pVolInfo, pAreaInfo->nAbsStartBlkAdr + nBlkAdr, nValidSize);
}


WFSKrnResult AreaBlkRemapCache(AreaInfo *pAreaInfo, WFSBlkAdr nSrcBlkAdr, WFSBlkAdr nDstBlkAdr, WFSBlkAdr nHashBlkAdr, u32 nHashOfs){
    dbga(OS_TPrintf("AreaBlkSetBlkAdr()\n"));
    return BCacheRemap(pAreaInfo->pVolInfo, pAreaInfo->nAbsStartBlkAdr + nSrcBlkAdr, pAreaInfo->nAbsStartBlkAdr + nDstBlkAdr, 
                       pAreaInfo->nAbsStartBlkAdr + nHashBlkAdr, nHashOfs);
}


WFSKrnResult AreaInitAllocator(AreaInfo *pAreaInfo, u32 nSizeRemaining) {
    dbga(OS_TPrintf("AreaInitAllocator()\n"));
    // This function initializes the block and sub-area allocation data structures within a new area
    // Arguments:
    //      pAreaInfo   [IN/OUT] .. The area whose free block allocator we wish to initilize
    //      pRootBlkPtr     [IN] .. A pointer to the root block of the area in memory (block cache)
    WFSKrnResult nResult = WFSKRN_RESULT_OK;
#if BTREE_BASED_ALLOCATOR

    nResult = FreeBlkInit(pAreaInfo, pAreaInfo->nBlkSize - nSizeRemaining);

#elif LIST_BASED_ALLOCATOR
    u32 *aAreaSizes = gaAllocationSizes;
    #if _DEBUG
        if ((gnNumAreaSizes==0) || (gnNumAreaSizes>WFSKRN_MAX_NUM_AREA_SIZES)) {
            return WFSKRN_RESULT_AREA_INVALID_PARAMETER;
        }
        // Check that larger areas are an integral multiples of the smaller
        s32 nSizeIdx;
        for (nSizeIdx=1; nSizeIdx<gnNumAreaSizes; nSizeIdx++) {
            if (0 != (aAreaSizes[nSizeIdx] % aAreaSizes[nSizeIdx-1])) {
                nResult = WFSKRN_RESULT_AREA_INVALID_PARAMETER;
                goto Done;
            }
        }
    #endif //_DEBUG

    nSizeRemaining -= sizeof(AreaFreeListHdr);
    pAreaInfo->nNumEntriesInFirstBlk = nSizeRemaining / sizeof(AreaEntry);
    pAreaInfo->nNumEntriesPerBlk     = (pAreaInfo->nBlkSize-sizeof(WFSMetaDataHdr)) / sizeof(AreaEntry);
    //pAreaInfo->nNumEntriesPerBlk     = (pAreaInfo->nBlkSize) / sizeof(AreaEntry);
    pAreaInfo->ah.nMaxNumAreas       = pAreaInfo->ah.nNumBlks / aAreaSizes[0];
    pAreaInfo->ah.nDataStartBlkAdr   = pAreaInfo->nRelStartBlkAdr;
    pAreaInfo->ah.nDataStartBlkAdr  += ((pAreaInfo->ah.nMaxNumAreas * sizeof(AreaEntry)) + (pAreaInfo->nBlkSize<<1) -nSizeRemaining-1) >> pAreaInfo->ah.nLog2BlkSize;

    u32 nNumUnclaimedBlks = pAreaInfo->ah.nNumBlks;
    AreaEntryIdx nEntryIdx = 0;
    //u32 sectorNum = pAreaInfo->nRelStartBlkAdr;
    for (nSizeIdx=gnNumAreaSizes-1; nSizeIdx>=0; nSizeIdx--) {
        pAreaInfo->pAflh->aFreeList[nSizeIdx].nAreaSize = aAreaSizes[nSizeIdx];
        pAreaInfo->pAflh->aFreeList[nSizeIdx].nHeadIdx = AREA_NULL;
        u32 nEntryStride = aAreaSizes[nSizeIdx]/aAreaSizes[0]; // # entries between areas
        u32 nNumAreas = 0;                                     // # areas of this size created
        while (aAreaSizes[nSizeIdx] <= nNumUnclaimedBlks) {
            nNumUnclaimedBlks -= aAreaSizes[nSizeIdx];
            AreaEntry *pEntry;
            WFSKrnResult nResult = _AreaEntryMap(pAreaInfo, nEntryIdx, BCACHE_FLAG_DIRTY, &pEntry);
            if(nResult < WFSKRN_RESULT_OK){
                return nResult;
            }
            pEntry->nSizeIdx = nSizeIdx;
            pEntry->nAttributes = AREA_START;
            ++nNumAreas;
            if (nNumAreas == 1) {
                pAreaInfo->pAflh->aFreeList[nSizeIdx].nHeadIdx = nEntryIdx;    // 1st is start of free list
                pEntry->nPrevIdx = AREA_NULL;
            } else {
                pEntry->nPrevIdx = nEntryIdx - nEntryStride;
            }
            pEntry->nNextIdx = (aAreaSizes[nSizeIdx] <= nNumUnclaimedBlks) ? nEntryIdx+nEntryStride : AREA_NULL;
            nEntryIdx += nEntryStride;
        }
        pAreaInfo->pAflh->aFreeList[nSizeIdx].nLen = nNumAreas;
    }
    //// The blocks are mapped to the bcache once because the blocks have to be written the hash data.
    //// The hash data is written when BCacheFlushVolume() will be called.
    u32 nAbsBlkAdr = _AreaGetAbsBlkAdr(pAreaInfo, nEntryIdx-1)+1;
    u32 nBlkAdr;
    for(nBlkAdr=0;nBlkAdr<nAbsBlkAdr;nBlkAdr++){
        nResult = BCacheAllocMetaData(pAreaInfo->pVolInfo, nBlkAdr, BCACHE_FLAG_DIRTY, 0);
        if (nResult < WFSKRN_RESULT_OK) {
            return nResult;
        }
    }
    //BCacheUnpin(pAreaInfo->pVolInfo, pAreaInfo->nAbsStartBlkAdr);
    AreaInfo subAreaInfo; // sub area allocated for pAreaHdr header
    nResult = AreaAllocSubArea(pAreaInfo, pAreaInfo->ah.nDataStartBlkAdr, &subAreaInfo, BCACHE_FLAG_DIRTY);

#else //SIMPLE_ALLOCATOR

    pAreaInfo->ah.pPtr = gpPtr;

    pAreaInfo->ah.nFirstFreeBlkAdr = 1;
    pAreaInfo->ah.nNumFreeBlks = pAreaInfo->ah.nNumBlks-1;
    u32 nI;
    for(nI=1; nI<pAreaInfo->ah.nNumFreeBlks; nI++) {
        WFSBlkAdr *pNextFreeBlock = (WFSBlkAdr *)AreaGetPhysicalBlk(pAreaInfo, nI);
        *pNextFreeBlock = nI+1;
    }
    WFSBlkAdr *pNextFreeBlock = (WFSBlkAdr *)AreaGetPhysicalBlk(pAreaInfo, nI);
    *pNextFreeBlock = 0;

#endif
#if LIST_BASED_ALLOCATOR
Done:
#endif
    return nResult;
}


void AreaInfoInit(AreaInfo *pAreaInfo) {
    dbga(OS_TPrintf("AreaInfoInit()\n"));
    // This function initializes the AreaInfo from the header on disk
    pAreaInfo->nBlkSize = (u32)(1<<pAreaInfo->ah.nLog2BlkSize);
    pAreaInfo->nMediumBlkSize = (u32)(1<<pAreaInfo->ah.nLog2MediumBlkSize);
    pAreaInfo->nLargeBlkSize = (u32)(1<<pAreaInfo->ah.nLog2LargeBlkSize);
    pAreaInfo->nLog2BlksPerMediumBlk = (u32)(pAreaInfo->ah.nLog2MediumBlkSize-pAreaInfo->ah.nLog2BlkSize);
    pAreaInfo->nLog2BlksPerLargeBlk = (u32)(pAreaInfo->ah.nLog2LargeBlkSize-pAreaInfo->ah.nLog2BlkSize);
    pAreaInfo->nLog2MediumBlkPerLargeBlk = (u32)(pAreaInfo->ah.nLog2LargeBlkSize-pAreaInfo->ah.nLog2MediumBlkSize);
    pAreaInfo->nLargeBlkPtrSize = sizeof(WFSBlkAdr) + (sizeof(WFSHashCode) << pAreaInfo->nLog2MediumBlkPerLargeBlk);
    //pAreaInfo->nMaxMediumBlkPtrsPerAttrSubBlk = DIR_ENTRY_ATTR_MAX_PTR_ARY_SIZE / pAreaInfo->nLargeBlkPtrSize;
    //pAreaInfo->nMaxLargeBlkPtrsPerAttrSubBlk  = DIR_ENTRY_ATTR_MAX_PTR_ARY_SIZE / pAreaInfo->nLargeBlkPtrSize;
    pAreaInfo->nNumLargeBlkPtrsPerBlk = (pAreaInfo->nBlkSize - sizeof(WFSMetaDataHdr)) / pAreaInfo->nLargeBlkPtrSize;
    pAreaInfo->nNumMediumBlkPtrsPerBlk = pAreaInfo->nNumLargeBlkPtrsPerBlk<<pAreaInfo->nLog2MediumBlkPerLargeBlk;
}


static WFSKrnResult FreeAclCache(AreaInfo *pAreaInfo)
{
    WFSKrnHeapFree(&wkg.heap, pAreaInfo->accessListCb); // release acl control block and acl cache.
    pAreaInfo->accessListCb = 0;
    return WFSKRN_RESULT_OK;
}


static WFSKrnResult AllocAclCache(AreaInfo *pAreaInfo)
{
    WFSKrnResult result = WFSKRN_RESULT_OUT_OF_MEMORY;
    WFSAccessListControlBlock* cb;
    u32 cbSize = ROUNDUP32BYTE(sizeof(WFSAccessListControlBlock));
    u32 aclCacheSize = ACL_DEFAULT_CACHESIZE;

    // Accesslist cache must be able to contain 3 accesslists, at least.
    u32 minBlocks = 3 * (WFS_ACL_MAX_ENTRIES + sizeof(WFSKrnAccessList) / sizeof(WFSKrnAccessListBlock));

    cb = (WFSAccessListControlBlock*) WFSKrnHeapAlloc(&wkg.heap, cbSize + aclCacheSize);
    if (cb)
    {
        cb->memoryBlock = (WFSKrnAccessListBlock*) ((u8*) cb + cbSize);
        cb->numBlock = aclCacheSize / sizeof(WFSKrnAccessListBlock);
        cb->initialized = (WFSBool) FALSE;
        pAreaInfo->accessListCb = cb;

        if (minBlocks <= cb->numBlock)
        {
            result = WFSKRN_RESULT_OK;
        }
        else
        {
            WFSKrnHeapFree(&wkg.heap, pAreaInfo->accessListCb); // release acl control block and acl cache.
        }
    }

    return result;
}


static WFSKrnResult AclInit(AreaInfo *pAreaInfo)
{
    WFSKrnResult result = AllocAclCache(pAreaInfo);
    if (result == WFSKRN_RESULT_OK)
    {
        result = WFSKrnAccessListInit(pAreaInfo);
    }
    return result;
}


WFSKrnResult AreaInit(AreaInfo *pAreaInfo) {
    dbga(OS_TPrintf("AreaInit()\n"));
    // Arguments:
    //     pAreaInfo              [IN/OUT] .. Pointer to area info struct for the new area to create.
    //   Input parameters used (remainder is output):
    //     pAreaInfo->pVolInfo        [IN] .. Points to the volume struct of the area.
    //     pAreaInfo->pParent         [IN] .. Points to the parent area. If this is NULL, then there is no parent (pAreaInfo is the root area)
    //     pAreaInfo->nRelStartBlkAdr [IN] .. Specifies the location of the new area within the parent area
    //     pAreaInfo->ah.nLog2BlkSize [IN] .. Log to the base 2 of the block size for this area (small block or meta data blocks)
    //     pAreaInfo->ah.nNumBlks     [IN] .. Number of blocks in the area.  Should be a power of 2.
    // Synopsis:
    //    This function can be called to initialize a new area
    WFSKrnResult nResult = WFSKRN_RESULT_OK;
    VolumeInfo *pVolInfo = pAreaInfo->pVolInfo;
    AreaInfo *pParent = pAreaInfo->pParent;
    pAreaInfo->nAbsStartBlkAdr = pAreaInfo->nRelStartBlkAdr;
    if (pParent) {
        pAreaInfo->nAbsStartBlkAdr += pParent->nAbsStartBlkAdr;
    }
    WFSBlkAdr nAbsStartBlkAdr = pAreaInfo->nAbsStartBlkAdr;
#if _DEBUG
    // check size area lies within the bounds of its parent
    if (pParent) {
        if ((nAbsStartBlkAdr + pAreaInfo->ah.nNumBlks) > (pParent->nAbsStartBlkAdr + pParent->ah.nNumBlks)) {
            return WFSKRN_RESULT_AREA_INVALID_PARAMETER;
        }
    } else {
        DeviceInfo *pDevInfo = pVolInfo->pDevInfo;
        if (((nAbsStartBlkAdr + pAreaInfo->ah.nNumBlks)<<(pDevInfo->dh.nLog2SectorSize - pAreaInfo->ah.nLog2BlkSize)) > pDevInfo->dh.nNumSectors) {
            return WFSKRN_RESULT_AREA_INVALID_PARAMETER;
        }
    }
#endif
    pAreaInfo->ah.nFlags = WFS_MDF_AREA_HDR;
    pAreaInfo->ah.nLog2MediumBlkSize = WFS_LOG2_MEDIUM_BLK_SIZE;
    pAreaInfo->ah.nLog2LargeBlkSize = WFS_LOG2_LARGE_BLK_SIZE;

    AreaInfoInit(pAreaInfo);
    BCacheEntry *pBce;
    nResult = BCacheAllocMetaBlk(pVolInfo, nAbsStartBlkAdr, BCACHE_FLAG_PINNED | BCACHE_FLAG_DIRTY, &pBce);
    if (nResult < WFSKRN_RESULT_OK) {
        return nResult;
    }
    u8 *pBlkPtr = pBce->pBlkPtr;
    u32 nSizeRemaining = pAreaInfo->nBlkSize - sizeof(AreaHdr);
    if (pAreaInfo->nAbsStartBlkAdr == VOLUME_HDR_BLK_ADR) {
        pBlkPtr += sizeof(VolumeHdr);
        nSizeRemaining -= sizeof(VolumeHdr);
    }
    pAreaInfo->pAflh = (AreaFreeListHdr *)(pBlkPtr + sizeof(AreaHdr));
#if _CHECK_BLK_ALLOCATION
    pAreaInfo->aBlkAllocState = (u32*)WFSKrnHeapAlloc(&wkg.heap, ((pAreaInfo->ah.nNumBlks+31)&-32)>>3);
    memset(pAreaInfo->aBlkAllocState, 0, pAreaInfo->ah.nNumBlks>>3);
#endif
    /*
        root disk block
     +------------------+ <-- pVolInfo - - - - - - - - -A
     |                  |                               |
     +------------------+ <-- pAreaHdr (pAreaInfo->ah)  |
     |                  |                               |
     +------------------+ <-- pAreaInfo->pAflh          |
     |                  | A                             |
     |                  | |                             |
     |                  | |                             | pAreaInfo->nBlkSize
     |                  | |                             |
     |                  | | nSizeRemaining              |
     |                  | |                             |
     |                  | |                             |
     |                  | |                             |
     +------------------+ V - - - - - - - - - - - - - - V
    */
    nResult = AreaInitAllocator(pAreaInfo, nSizeRemaining);
    if (nResult != WFSKRN_RESULT_OK) {
        goto Exit;
    }
#if _DEBUG
    bcg.bSkipCheckPin = true;
#endif
    nResult = AreaAllocBlks(pAreaInfo, 0, pAreaInfo->ah.nLog2BlkSize, 1, &pAreaInfo->ah.nRootBlkAdr, sizeof(WFSBlkAdr));
#if _DEBUG
    bcg.bSkipCheckPin = false;
#endif
    if (nResult != WFSKRN_RESULT_OK) {
        goto Exit;
    }

    nResult = AclInit(pAreaInfo);
    if (nResult != WFSKRN_RESULT_OK) {
        goto Exit;
    }
    *((AreaHdr *)pBlkPtr) = pAreaInfo->ah;

Exit:
    BCacheUnpin(pVolInfo, nAbsStartBlkAdr);
    if (nResult == WFSKRN_RESULT_OK)
    {
        BCacheFlushVolume(pAreaInfo->pVolInfo); // write out all.  Needed? or close flushes?
    }
    return nResult;
}


WFSKrnResult AreaOpen(AreaInfo *pAreaInfo) {
    dbga(OS_TPrintf("AreaOpen()\n"));
    WFSKrnResult nResult = WFSKRN_RESULT_OK;
    VolumeInfo *pVolInfo = pAreaInfo->pVolInfo;
    AreaInfo *pParent = pAreaInfo->pParent;
    pAreaInfo->nAbsStartBlkAdr = pAreaInfo->nRelStartBlkAdr;
    if (pParent) {
        pAreaInfo->nAbsStartBlkAdr += pParent->nAbsStartBlkAdr;
    }
    WFSBlkAdr nAbsStartBlkAdr = pAreaInfo->nAbsStartBlkAdr;
#if _DEBUG
    // check size area lies within the bounds of its parent
    if (pParent) {
        if ((nAbsStartBlkAdr + pAreaInfo->ah.nNumBlks) > (pParent->nAbsStartBlkAdr + pParent->ah.nNumBlks)) {
            return WFSKRN_RESULT_AREA_INVALID_PARAMETER;
        }
    } else {
        DeviceInfo *pDevInfo = pVolInfo->pDevInfo;
        if (((nAbsStartBlkAdr + pAreaInfo->ah.nNumBlks)<<(pDevInfo->dh.nLog2SectorSize - pAreaInfo->ah.nLog2BlkSize)) > pDevInfo->dh.nNumSectors) {
            return WFSKRN_RESULT_AREA_INVALID_PARAMETER;
        }
    }
#endif
    BCacheEntry *pBce;
    nResult = BCacheAllocMetaBlk(pVolInfo, nAbsStartBlkAdr, BCACHE_FLAG_PINNED | BCACHE_FLAG_READ, &pBce);
    if(nResult < WFSKRN_RESULT_OK){
        return nResult;
    }
    u8 *pBlkPtr = pBce->pBlkPtr;

#if BTREE_BASED_ALLOCATOR
    if (pAreaInfo->nAbsStartBlkAdr == VOLUME_HDR_BLK_ADR) {
        pBlkPtr += sizeof(VolumeHdr);
    }

    AreaHdr *pSrcAreaHdr = (AreaHdr*)pBlkPtr;
    pAreaInfo->ah = *pSrcAreaHdr;
    pAreaInfo->pAflh = (AreaFreeListHdr *)(pBlkPtr + sizeof(AreaHdr));
    AreaInfoInit(pAreaInfo); // [check] init again?

#elif LIST_BASED_ALLOCATOR

    s32 nSizeRemaining = -(signed)(sizeof(AreaHdr) + sizeof(AreaFreeListHdr));
    if (pAreaInfo->nAbsStartBlkAdr == VOLUME_HDR_BLK_ADR) {
        pBlkPtr += sizeof(VolumeHdr);
        nSizeRemaining -= sizeof(VolumeHdr);
    }
    AreaHdr *pSrcAreaHdr = (AreaHdr*)pBlkPtr;
    pAreaInfo->ah = *pSrcAreaHdr;
    pAreaInfo->pAflh = (AreaFreeListHdr *)(pBlkPtr + sizeof(AreaHdr));
    AreaInfoInit(pAreaInfo);
    nSizeRemaining += pAreaInfo->nBlkSize;
    pAreaInfo->nNumEntriesInFirstBlk = nSizeRemaining / sizeof(AreaEntry);
    pAreaInfo->nNumEntriesPerBlk     = (pAreaInfo->nBlkSize-sizeof(WFSMetaDataHdr)) / sizeof(AreaEntry);
    //pAreaInfo->nNumEntriesPerBlk     = (pAreaInfo->nBlkSize) / sizeof(AreaEntry);

#else //SIMPLE_ALLOCATOR

#endif

#if _CHECK_BLK_ALLOCATION
    pAreaInfo->aBlkAllocState = (u32*)WFSKrnHeapAlloc(&wkg.heap, ((pAreaInfo->ah.nNumBlks+31)&-32)>>3);
    memset(pAreaInfo->aBlkAllocState, 0, pAreaInfo->ah.nNumBlks>>3);
#endif
    //nResult = AreaInitAllocator(pAreaInfo, pBlkPtr);
    if (nResult != WFSKRN_RESULT_OK) {
        goto Exit;
    }
    BCacheUnpin(pVolInfo, nAbsStartBlkAdr);

    // load accesslist config.
    nResult = AllocAclCache(pAreaInfo);
    if (nResult == WFSKRN_RESULT_OK)
    {
        nResult = WFSKrnAccessListOpen(pAreaInfo);
    }
    if (nResult != WFSKRN_RESULT_OK) {
        goto Exit;
    }

    BCacheFlushVolume(pAreaInfo->pVolInfo); // write out all.  Needed? or close flushes?
Exit:
    return nResult;
}


void AreaClose(AreaInfo *pAreaInfo) {
    dbga(OS_TPrintf("AreaClose()\n"));
    FreeAclCache(pAreaInfo);
#if _CHECK_BLK_ALLOCATION
    WFSKrnHeapFree(&wkg.heap, pAreaInfo->aBlkAllocState);
#endif
}
