/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfskrn_BCache.cpp - disk block cache library.
               Intended for metadata block caching only.

  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_BCache.cpp,v $
  Revision 1.27  2008/12/22 08:56:12  ueno
  Avoided null reference.

  Revision 1.26  2008/12/18 09:03:10  kondo_masahiro
  Fixed function name and process from BCacheSetBlkAdr() to BCacheRemap()

  Revision 1.25  2008/12/17 01:37:27  kondo_masahiro
  Added function BCacheSetBlkAdr()

  Revision 1.24  2008/12/10 08:20:08  ueno
  Added BCacheDumpPinCount() for debug.

  Revision 1.23  2008/12/03 00:08:59  kondo_masahiro
  Added WFSBCacheStoreVolume and removed WFSBCacheFlushAll.

  Revision 1.22  2008/11/26 01:43:24  kondo_masahiro
  Fixed arguments of several functions in order to improve access speed.

  Revision 1.21  2008/11/05 16:48:37  ueno
  Fixed.

  Revision 1.20  2008/11/05 05:53:40  kondo_masahiro
  Added ChainId

  Revision 1.19  2008/10/16 09:30:39  ooizumi
  Fixed definitions for debug build.

  Revision 1.18  2008/10/14 10:27:14  kondo_masahiro
  Added #define _DEBUG_BREAK_POINT

  Revision 1.17  2008/10/09 01:41:13  kondo_masahiro
  Fixed name of the function DeviceRead/WriteSingleBlk

  Revision 1.16  2008/10/08 23:59:27  ueno
  Modified BCacheAlloc() to return WFSKRN_RESULT_MEDIA_ERROR when the volume is unmounted.

  Revision 1.15  2008/10/08 23:19:54  kondo_masahiro
  Added codes to access large size file data

  Revision 1.14  2008/10/07 05:56:41  ueno
  Modified BCacheAlloc() to return WFSKRN_RESULT_MEDIA_ERROR when the volume is unmounted.

  Revision 1.13  2008/10/03 08:37:56  kondo_masahiro
  Fixed codes for porting IOP

  Revision 1.12  2008/09/29 10:19:36  ueno
  Revised to support multi-B-tree-based allocator.

  Revision 1.11  2008/09/28 23:31:21  kondo_masahiro
  Fixed codes to use medium size user block.

  Revision 1.10  2008/09/04 03:31:07  kondo_masahiro
  Fixed error handling

  Revision 1.9  2008/08/14 08:10:57  kondo_masahiro
  Fixed functions related on DeviceRead / Write.

  Revision 1.8  2008/08/08 13:32:28  ueno
  Added BCacheCheck2() to check freeblock allocator.

  Revision 1.7  2008/08/05 04:07:46  kondo_masahiro
  Added error handling for hash error and detach device.

  Revision 1.6  2008/07/28 21:47:22  paul
  Removed "Increment update counter" section. This is now specific to Dir blocks.

  Revision 1.5  2008/07/17 22:45:10  paul
  Change for updateMap

  Revision 1.4  2008/07/17 05:31:15  kondo_masahiro
  Added codes for debug.

  Revision 1.3  2008/07/09 00:18:45  paul
  Changed to use pointer to static volume struct as tag instead of handle. Removed handle mapping code. Moved globals to a single struct. Removed some of the less useful parameter checking. Added more stringent usage checks in debug builds.

  Revision 1.2  2008/05/12 19:06:39  paul
  Merged in Wayne's previous changes

  Revision 1.1  2008/04/22 04:39:08  kondo_masahiro
  Fixed code to transfer the wfskrn project to RM_SDK tree

  Revision 1.8  2008/04/26 01:28:08  wayne.wong
  Moved BCache paramers to wfskrn_config.h. Changed type prefix.

  Revision 1.7  2008/03/19 07:42:26  wayne.wong
  Fixed a bug with nAbsBlkAdr tag matching at page granularity.

  Revision 1.6  2008/03/11 03:15:53  wayne.wong
  Added a call to map a device without knowing the volume's ID. Here, the
  device is assumed to have been initialized with a volume and the volume's
  header is read.

  Revision 1.5  2008/02/22 00:40:51  wayne.wong
  Change BCache module's prefix for functions and types. Added some new functions.

  Revision 1.4  2008/02/20 08:01:39  paul
  Changed wfsKrnResult to WFSKrnResult, wfsDevice to Device.

  Revision 1.3  2008/02/05 03:10:52  wayne.wong
  Small fixes.

  Revision 1.2  2008/01/30 03:55:49  wayne.wong
  Multiple fixes. Checkpoint.

  Revision 1.1  2008/01/18 06:13:50  wayne.wong
  Initial version.

*---------------------------------------------------------------------------*/


// #define _DEBUG_WFSKRN_PMEM_    1
// #define _DEBUG_WFSKRN_BCACHE  1

#include "wfskrn_Api.h"
#include "wfskrn_BCache.h"
#include "wfskrn_Config.h"
#include "wfskrn_Device.h"
#include "wfskrn_Volume.h"
#include "wfskrn_Area.h"
#include "wfskrn_Utils.h"

// moved to wfskrn_Config.h

BCacheGlobals bcg ATTRIBUTE_ALIGN(512);

#define PAGE_NUMBER_CE(_ce_addr)             (_ce_addr - pGroup->pBCacheTable)

#define BCACHE_CACHE_ENTRY_INVALIDATE(_pCE) {   \
        _pCE->nAbsBlkAdr = 0;                   \
        _pCE->pVolInfo   = 0;                   \
        _pCE->pDevInfo   = 0;                   \
        _pCE->nState     = BCACHE_FLAG_INVALID; \
    }


#if !DEVICE_EMBEDDED_HASH_DATA
static WFSHashCode hashData;
#endif


// internal BCacheList functions
static BCacheListNode *BCacheListNodeAlloc(BCacheGroup *pGroup) {
    if (pGroup->bclFree.nLen > 0) {
        BCacheListNode *pNode = pGroup->bclFree.pHead;
        pGroup->bclFree.pHead = pNode->pNext;
        --pGroup->bclFree.nLen;
        return pNode;
    }
    return NULL;
}


static void BCacheListNodeFree(BCacheGroup *pGroup, BCacheListNode *pNode) {
    pNode->pNext = pGroup->bclFree.pHead;
    pGroup->bclFree.pHead = pNode;
    ++pGroup->bclFree.nLen;
}


static WFSKrnResult BCacheListNodeInit(BCacheGroup *pGroup, u32 numPages) {
    dbgb(osTPrintf("BCacheListNodeInit()\n"));
    // bcg.pBclNodes = (BCacheListNode*)malloc(2*numPages * sizeof(BCacheListNode));
    pGroup->pBclNodes = (BCacheListNode*)bcg.pHeap;
    bcg.pHeap     += 2*numPages * sizeof(BCacheListNode);
    bcg.nHeapSize -= 2*numPages * sizeof(BCacheListNode);
    if (!pGroup->pBclNodes) return WFSKRN_RESULT_BCACHE_NO_MEMORY;
    pGroup->bclFree.pHead = NULL;
    pGroup->bclFree.pTail = NULL;
    u32 nI;
    for (nI=0; nI<(2*numPages); nI++) {  // need to add to list and hash table
        BCacheListNodeFree(pGroup, &pGroup->pBclNodes[nI]);
    }
    return WFSKRN_RESULT_OK;
}


static void BCacheListInit(BCacheList *pList) {
    dbgb(osTPrintf("BCacheListInit()\n"));
    pList->pHead = NULL;
    pList->pTail = NULL;
    pList->nLen  = 0;
}


// A match is found if the nVolumeId and nAbsBlkAdr are the same.
static BCacheEntry* BCacheListFind(BCacheList *pList, VolumeInfo *pVolInfo, WFSBlkAdr nAbsBlkAdr) {
    dbgb(osTPrintf("BCacheListFind()\n"));
    BCacheListNode *c = pList->pHead;
    while (NULL != c) {
        BCacheEntry *pBce = (BCacheEntry*)c->pData;
        if ((pVolInfo == pBce->pVolInfo) && (pBce->nAbsBlkAdr == nAbsBlkAdr)) {
            return pBce;
        }
        c = c->pNext;
    }
    return NULL;
}


// Add a BCacheEntry to the head of the list. Be careful to not add an entry to the same list
// more than once.
static BCacheEntry* BCacheListAddToHead(BCacheList *pList, BCacheEntry *pBce) {
    dbgb(osTPrintf("BCacheListAddToHead()\n"));
    BCacheListNode *n = BCacheListNodeAlloc(pBce->pGroup);
    if (!n) return NULL;
    n->pData = (void*)pBce;

    BCacheListNode *c = pList->pHead;
    if (c != NULL) {
        c->pPrev = n;
        n->pNext = c;
    } else {
        n->pNext  = NULL;
        pList->pTail = n;
    }
    n->pPrev = NULL;
    pList->pHead = n;
    pList->nLen++;
    return pBce;
}


// Add a BCacheEntry to the tail of the list. Be careful to not add an empty to the same list
// more than once!
static BCacheEntry* BCacheListAddToTail(BCacheList *pList, BCacheEntry *pBce) {
    dbgb(osTPrintf("BCacheListAddToTail()\n"));
    BCacheListNode *n = BCacheListNodeAlloc(pBce->pGroup);
    if (!n) return NULL;
    n->pData = (void*)pBce;

    BCacheListNode *c = pList->pTail;
    if (c != NULL) {
        c->pNext = n;
        n->pPrev = c;
    } else {
        n->pPrev  = NULL;
        pList->pHead = n;
    }
    n->pNext  = NULL;
    pList->pTail = n;
    pList->nLen++;
    return pBce;
}


// Returns a pointer to the entry if the delete was successful.  Otherwise, the
// entry was not found.
static BCacheEntry* BCacheListDelete(BCacheList *pList, BCacheEntry *pBce) {
    dbgb(osTPrintf("BCacheListDelete()\n"));
    BCacheListNode *c = pList->pHead;
    while (NULL != c) {
        BCacheEntry *currCE = (BCacheEntry*)c->pData;
        if (pBce == currCE) {
            if (pList->pHead == c) {
                pList->pHead = c->pNext;
            } else {
                (c->pPrev)->pNext = c->pNext;
            }
            if (pList->pTail == c) {
                pList->pTail = c->pPrev;
            } else {
                (c->pNext)->pPrev = c->pPrev;
            }
            pList->nLen--;
            BCacheListNodeFree(pBce->pGroup, c);
            return pBce;
        }
        c = c->pNext;
    }
    return NULL;
}


static BCacheEntry* BCacheNodeDelete(BCacheList *pList, BCacheListNode *pNode) {
    dbgb(osTPrintf("BCacheNodeDelete()\n"));
    if (NULL != pNode) {
        BCacheEntry *pBce = (BCacheEntry*)pNode->pData;
        if (pList->pHead == pNode) {
            pList->pHead = pNode->pNext;
        } else {
            (pNode->pPrev)->pNext = pNode->pNext;
        }
        if (pList->pTail == pNode) {
            pList->pTail = pNode->pPrev;
        } else {
            (pNode->pNext)->pPrev = pNode->pPrev;
        }
        pList->nLen--;
        BCacheListNodeFree(pBce->pGroup, pNode);
        return pBce;
    }
    return NULL;
}


// Returns a pointer to the entry if the move was successful
static BCacheEntry* BCacheListMoveToHead(BCacheList *pList, BCacheEntry *pBce) {
    dbgb(osTPrintf("BCacheListMoveToHead()\n"));
    if ((BCacheEntry*)pList->pHead->pData == pBce) return pBce;
    if (!BCacheListDelete(pList, pBce)) return NULL;
    return BCacheListAddToHead(pList, pBce);
}


// Returns a pointer to the entry if the move was successful
static BCacheEntry* BCacheListMoveToTail(BCacheList *pList, BCacheEntry *pBce) {
    dbgb(osTPrintf("BCacheListMoveToTail()\n"));
    if ((BCacheEntry*)pList->pTail->pData == pBce) return pBce;
    if (!BCacheListDelete(pList, pBce)) return NULL;
    return BCacheListAddToTail(pList, pBce);
}


static void BCacheListPrint(BCacheList *pList) {
    dbgb(osTPrintf("BCacheListPrint()\n"));
    u32 cnt = 0;
    BCacheListNode *c = pList->pHead;
    //osTPrintf("\nlist 0x%08x\n", pList);
    while (NULL != c) {
        BCacheEntry *pBce = (BCacheEntry*)c->pData;
        //osTPrintf("%d: 0x%08x, nAbsBlkAdr=%d\n", cnt, pBce, pBce->nAbsBlkAdr);
        cnt++;
        c = c->pNext;
    }
}


// Exposed BCache functions

// Calculate the number of bytes need to support the cache geometry.
u32 BCacheMetadataMemSize(u32 nMemCacheSize, u32 nBlkSize) {
    dbgb(osTPrintf("BCacheMetadataMemSize()\n"));
    u32 frames = nMemCacheSize/nBlkSize;
    u32 result =  frames*sizeof(BCacheEntry);   // page table
    result += 2*frames * sizeof(BCacheListNode);
    return result;
}


WFSKrnResult BCacheInit(u32 nLog2UnitBlkSize, u8 *pMem, u32 nMemSize)
{
    dbgb(osTPrintf("BCacheInitMetadata()\n"));
    bcg.pHeap            = pMem;
    bcg.nHeapSize        = nMemSize;
    bcg.nNumGroups       = 0;
    bcg.nLog2UnitBlkSize = nLog2UnitBlkSize;
    bcg.nUnitBlkSize     = 1<<nLog2UnitBlkSize;

    // pBCache is used as empty flag.
    u32 nI;
    for (nI=0; nI<BCACHE_NUM_OF_GROUP; nI++) {
        BCacheGroup *pGroup = &bcg.aGroup[nI];
        pGroup->pBCache = 0;
    }
    // init hash table
    for (nI=0; nI<WFSKRN_BCACHE_HASH_TABLE_SIZE; nI++) {  // initialize hash table
        BCacheListInit(&bcg.aHashTbl[nI]);
    }

    return WFSKRN_RESULT_OK;
}


WFSKrnResult BCacheGroupInit(u32 nBCacheGroupId, u8 *pMemCache, u32 nMemCacheSize, u32 nLog2BlkSize)
{
    // Wayne - check the alignment(512B) of pMemCache for IOP port
    dbgb(osTPrintf("BCacheInitGroup()\n"));
    BCacheGroup *pGroup  = &bcg.aGroup[nBCacheGroupId];
    if ( (pGroup->pBCache) ||
         (bcg.pHeap == NULL) || (pMemCache == NULL) || (nMemCacheSize==0) || (nLog2BlkSize==0) ||
         ((u32)bcg.pHeap & 0x3) || ((u32)pMemCache & 0x1ff) ||
         (bcg.nHeapSize < BCacheMetadataMemSize(nMemCacheSize, 1<<nLog2BlkSize)) ) {
        return WFSKRN_RESULT_BCACHE_INVALID_PARAMETER;
    }
    pGroup->pBCache      = pMemCache;
    pGroup->nLog2BlkSize = nLog2BlkSize;
    pGroup->nBlkSize     = 1<<nLog2BlkSize;
    pGroup->nNumBlks     = nMemCacheSize>>nLog2BlkSize;

    BCacheListNodeInit(pGroup, pGroup->nNumBlks);

    BCacheListInit(&pGroup->bclInvalid);
    BCacheListInit(&pGroup->bclLru);
    BCacheListInit(&pGroup->bclPinned);

    // init cache table
    pGroup->pBCacheTable = (BCacheEntry*) bcg.pHeap;
    bcg.pHeap     += pGroup->nNumBlks*sizeof(BCacheEntry);
    bcg.nHeapSize -= pGroup->nNumBlks*sizeof(BCacheEntry);

    if (!pGroup->pBCacheTable) return WFSKRN_RESULT_BCACHE_ERROR;
    dbgb(osTPrintf("bcg.pBCacheTable @ 0x%08x\n", bcg.pBCacheTable));
    int nI;
    for (nI=0; nI<pGroup->nNumBlks; nI++) {             // initialize pages and link together
        BCacheEntry *pBce = &pGroup->pBCacheTable[nI];
        BCACHE_CACHE_ENTRY_INVALIDATE(pBce);
        pBce->pGroup = pGroup;
        pBce->pHash = 0;
        BCacheListAddToTail(&pGroup->bclInvalid, pBce);
    }

    ++bcg.nNumGroups;

    return WFSKRN_RESULT_OK;
}


WFSKrnResult BCacheGetState(BCacheGroup *pGroup) {
    dbgb(osTPrintf("BCacheGetState()\n"));
    pGroup->pMemCache     = pGroup->pBCache;
    pGroup->nMemCacheSize = pGroup->nNumBlks * pGroup->nBlkSize;
    pGroup->nBlkSize      = pGroup->nBlkSize;
    pGroup->nNumMapped    = pGroup->bclLru.nLen + pGroup->bclPinned.nLen;
    pGroup->nNumInvalid   = pGroup->bclInvalid.nLen;
    pGroup->nNumPinned    = pGroup->bclPinned.nLen;
    return WFSKRN_RESULT_OK;
}


// Allocate a page in the cache.
static
WFSKrnResult BCacheAllocBlk(BCacheGroup *pGroup, u8 **ppBlkPtr, BCacheEntry** ppBce) {
    dbgb(osTPrintf("BCacheAllocBlk()\n"));
    // find a victim list
    *ppBce = NULL;
    BCacheList *pList;
    if (pGroup->bclInvalid.pHead) {                // if invalid blocks available
        pList = &pGroup->bclInvalid;
    } else if (pGroup->bclLru.pHead) {             // else, try a non-pinned LRU page
        pList = &pGroup->bclLru;
    } else {                                 // otherwise, no free page available
        return WFSKRN_RESULT_BCACHE_RESOURCE_LIMIT;
    }

    BCacheEntry *pBce = (BCacheEntry*) pList->pHead->pData;  // victim is head of invalid or LRU list
    u8 *pBlkPtr = pGroup->pBCache + (pGroup->nBlkSize * PAGE_NUMBER_CE(pBce));
#if _DEBUG_BREAK_POINT
    if(pBce->nAbsBlkAdr == 0x1c5d){
        int a; a=0;
    }
#endif
    if (pBce->nState & BCACHE_FLAG_DIRTY) {
        DeviceInfo *pDevInfo = pBce->pDevInfo;
        WFSKrnResult nResult = DeviceWriteSingleBlk(pDevInfo, pBce->nAbsBlkAdr<<(bcg.nLog2UnitBlkSize-pDevInfo->dh.nLog2SectorSize),
                                                    pBlkPtr, pBce->nValidSize>>pDevInfo->dh.nLog2SectorSize,
                                                    pBce->pHash, pGroup->nLog2BlkSize-pDevInfo->dh.nLog2SectorSize, false);
        if(nResult < WFSKRN_RESULT_OK) {
            return WFSKRN_RESULT_MEDIA_ERROR;
        }
        if(pBce->nAbsBlkAdr != pBce->nHashBlkAdr){
            nResult = BCacheDirty(pDevInfo->pVolInfo, pBce->nHashBlkAdr);
            nResult = BCacheUnpin(pDevInfo->pVolInfo, pBce->nHashBlkAdr);
            if(nResult < WFSKRN_RESULT_OK){
                return nResult;
            }
        }
        pBce->nState &= ~BCACHE_FLAG_DIRTY;    // make clean; just in case we don't really replace

        if(pBce->nChainId && pList == &pGroup->bclLru){
            BCacheEntry *pBceWrite = pBce;
            while(1){
                WFSBlkAdr nBlkAdrTmp = pBceWrite->nAbsBlkAdr + (1<<(pGroup->nLog2BlkSize-bcg.nLog2UnitBlkSize));
                pBceWrite = BCacheListFind(&pGroup->bclLru, pBceWrite->pDevInfo->pVolInfo, nBlkAdrTmp);
                if(pBceWrite == NULL || pBceWrite->nChainId != pBce->nChainId){
                    break;
                }
                pBceWrite->nChainId = 0;
                nResult = BCacheStore(pBceWrite->pDevInfo->pVolInfo, nBlkAdrTmp);
                if(nResult < WFSKRN_RESULT_OK) {
                    return WFSKRN_RESULT_MEDIA_ERROR;
                }
            }
        }
    }
    if (pList == &pGroup->bclLru) { // if mapped, remove mapping
        BCacheListDelete(&bcg.aHashTbl[WFSKRN_BCACHE_HASH_FN(pBce->nAbsBlkAdr)], pBce);
    }
    if (!BCacheListDelete(pList, pBce)){ // remove victim entry from its list; if fail, return NULL
        return WFSKRN_RESULT_BCACHE_ERROR;
    }

    pBce->nState = BCACHE_FLAG_INVALID;        // make clean; just in case we don't really replace
    pBce->pHash = 0;
    pBce->nChainId = 0;

    if (ppBlkPtr) *ppBlkPtr = pBlkPtr;
    *ppBce = pBce;
    return WFSKRN_RESULT_OK;
}


#if _CHECK_BLK_USAGE
u32 BCacheCalcCheckSum(u8 *pBuffer, u32 nSize) {
    dbgb(osTPrintf("BCacheCalcCheckSum()\n"));
    // pBuffer should be 16 byte aligned address, and nSize should be a multiple of 16 bytes
    u32 *pPtr = (u32*)pBuffer;
    u32 *pPtrEnd = (u32*)((u8*)pBuffer + nSize);
    u32 nCheckSum = 0x12345678;
    while(pPtr<pPtrEnd) {
        nCheckSum += *pPtr++;
        nCheckSum ^= *pPtr++ + (nCheckSum>>3);
        nCheckSum += *pPtr++;
        nCheckSum ^= *pPtr++ + (nCheckSum*0x8020);
    }
    return nCheckSum;
}
#endif


void BCacheReTag(VolumeInfo *pOldVolInfo, WFSBlkAdr nOldAbsBlkAdr, VolumeInfo *pNewVolInfo, WFSBlkAdr nNewAbsBlkAdr) {
    dbgb(osTPrintf("BCacheReTag()\n"));
    BCacheEntry *pBce = BCacheListFind(&bcg.aHashTbl[WFSKRN_BCACHE_HASH_FN(nOldAbsBlkAdr)], pOldVolInfo, nOldAbsBlkAdr);
    if (!pBce) {
        // Doesn't exist in the cache, so nothing to re-tag
        return;
    }
    if (WFSKRN_BCACHE_HASH_FN(nOldAbsBlkAdr) == WFSKRN_BCACHE_HASH_FN(nOldAbsBlkAdr)) {
        pBce->nAbsBlkAdr = nNewAbsBlkAdr;
        pBce->pVolInfo = pNewVolInfo;
    } else {
        // Cache entry needs to be moved to a new hash table address
        // ToDo: This branch has not yet been tested
        BCacheListDelete(&bcg.aHashTbl[WFSKRN_BCACHE_HASH_FN(nOldAbsBlkAdr)], pBce);
        BCacheListAddToHead(&bcg.aHashTbl[WFSKRN_BCACHE_HASH_FN(nNewAbsBlkAdr)], pBce);
    }
}


static
WFSKrnResult BCacheSetHashPtrToEntry(struct VolumeInfo_ *pVolInfo, BCacheEntry *pBce, WFSBlkAdr nHashBlkAdr, u32 nHashOfs)
{
#if DEVICE_EMBEDDED_HASH_DATA
    BCacheEntry *pBceHash;
    if (nHashBlkAdr == pBce->nAbsBlkAdr){
        pBceHash = pBce;
    } else {
        pBceHash = BCacheListFind(&bcg.aHashTbl[WFSKRN_BCACHE_HASH_FN(nHashBlkAdr)], pVolInfo, nHashBlkAdr);
        if (!pBceHash){ // nHashBlkAdr have to exist on the block cache.
            return WFSKRN_RESULT_BCACHE_HASH_BLK_NOT_MAPPED;
        }
        pBceHash->nState |= BCACHE_FLAG_DIRTY;
    }
    if(pBce->nHashBlkAdr != nHashBlkAdr && pBce->nState & BCACHE_FLAG_DIRTY){
        BCacheUnpin(pVolInfo, pBce->nHashBlkAdr);
        BCachePin(pVolInfo, nHashBlkAdr);
    }
    pBce->pHash = (WFSHashCode*)(pBceHash->pBlkPtr+nHashOfs);
    pBce->nHashBlkAdr = nHashBlkAdr;
#elif
    pBce->pHash = &hashData;
#endif
    return WFSKRN_RESULT_OK;
}


// nAbsBlkAdr and nHashBlkAdr have to be mapped to use this function
WFSKrnResult BCacheSetHash(struct VolumeInfo_ *pVolInfo, WFSBlkAdr nAbsBlkAdr, WFSBlkAdr nHashBlkAdr, u32 nHashOfs)
{
    BCacheEntry *pBce = BCacheListFind(&bcg.aHashTbl[WFSKRN_BCACHE_HASH_FN(nAbsBlkAdr)], pVolInfo, nAbsBlkAdr);
    if (!pBce) {
        return WFSKRN_RESULT_BCACHE_NOT_FOUND;
    }
    return BCacheSetHashPtrToEntry(pVolInfo, pBce, nHashBlkAdr, nHashOfs);
}


WFSKrnResult BCacheSetValidSize(struct VolumeInfo_ *pVolInfo, WFSBlkAdr nAbsBlkAdr, u32 nValidSize)
{
    BCacheEntry *pBce = BCacheListFind(&bcg.aHashTbl[WFSKRN_BCACHE_HASH_FN(nAbsBlkAdr)], pVolInfo, nAbsBlkAdr);
    if (!pBce) {
        return WFSKRN_RESULT_BCACHE_NOT_FOUND;
    }
    if (nValidSize > pBce->pGroup->nBlkSize){
        return WFSKRN_RESULT_INVALID;
    }
    pBce->nValidSize = WFS_ROUND_UP(nValidSize,pVolInfo->pDevInfo->nSectorSize);
    return WFSKRN_RESULT_OK;
}


WFSKrnResult BCacheSetChainId(struct VolumeInfo_ *pVolInfo, WFSBlkAdr nAbsBlkAdr, u32 nChainId)
{
    BCacheEntry *pBce = BCacheListFind(&bcg.aHashTbl[WFSKRN_BCACHE_HASH_FN(nAbsBlkAdr)], pVolInfo, nAbsBlkAdr);
    if (!pBce) {
        return WFSKRN_RESULT_BCACHE_NOT_FOUND;
    }
    pBce->nChainId = nChainId;
    return WFSKRN_RESULT_OK;
}


WFSKrnResult BCacheGetChainId(struct VolumeInfo_ *pVolInfo, WFSBlkAdr nAbsBlkAdr, u32 *pChainId)
{
    BCacheEntry *pBce = BCacheListFind(&bcg.aHashTbl[WFSKRN_BCACHE_HASH_FN(nAbsBlkAdr)], pVolInfo, nAbsBlkAdr);
    if (!pBce) {
        return WFSKRN_RESULT_BCACHE_NOT_FOUND;
    }
    *pChainId = pBce->nChainId;
    return WFSKRN_RESULT_OK;
}


WFSKrnResult BCacheRemap(struct VolumeInfo_ *pVolInfo, WFSBlkAdr nSrcAbsBlkAdr, WFSBlkAdr nDstAbsBlkAdr, WFSBlkAdr nHashAbsBlkAdr, u32 nHashOfs)
{
    BCacheList *pHMapList = &bcg.aHashTbl[WFSKRN_BCACHE_HASH_FN(nSrcAbsBlkAdr)];
    BCacheEntry *pBce = BCacheListFind(pHMapList, pVolInfo, nSrcAbsBlkAdr);
    if (!pBce) {
        return WFSKRN_RESULT_BCACHE_NOT_FOUND;
    }
    if (!BCacheListDelete(pHMapList, pBce))
    {
        return WFSKRN_RESULT_BCACHE_ERROR;
    }
    pBce->nAbsBlkAdr = nDstAbsBlkAdr;
    if ( (!BCacheListAddToHead(&bcg.aHashTbl[WFSKRN_BCACHE_HASH_FN(nDstAbsBlkAdr)], pBce)) ||
         (BCacheDirty(pVolInfo, nDstAbsBlkAdr) != WFSKRN_RESULT_OK)                        ||
         (BCacheSetHash(pVolInfo, nDstAbsBlkAdr, nHashAbsBlkAdr, nHashOfs) != WFSKRN_RESULT_OK) )
    {
        return WFSKRN_RESULT_BCACHE_ERROR;
    }
    return WFSKRN_RESULT_OK;
}


// Alloc a page.  A page is mapped memory <-> persistent memory.
// (Wayne - make sure that we do not double-map, map a page already mapped?
// or is that checked above here?) For now, we check if it's already
// mapped/cached here. If so, return that page.
// nHashBlkAdr have to be mapped previously or nHashBlkAdr have to be indicated the same value as nAbsBlkAdr.
WFSKrnResult BCacheAlloc(struct VolumeInfo_ *pVolInfo, WFSBlkAdr nAbsBlkAdr, u32 nFlags, u32 nValidSize, u32 nBCacheGroupId,
                         WFSBlkAdr nHashBlkAdr, u32 nHashOfs, BCacheEntry** ppBce)
{
    if (!pVolInfo){
        return WFSKRN_RESULT_MEDIA_ERROR;
    }
    DeviceInfo *pDevInfo = pVolInfo->pDevInfo;
    if (!pDevInfo){
        return WFSKRN_RESULT_MEDIA_ERROR;
    }
    if (nValidSize > (1<<bcg.aGroup[nBCacheGroupId].nLog2BlkSize)){
        return WFSKRN_RESULT_INVALID;
    }
#if _DEBUG_BREAK_POINT
    if(nAbsBlkAdr > 0x10000000){
        int a; a=0;
    }
#endif
    WFSKrnResult nResult = WFSKRN_RESULT_UNKNOWN;
    dbgb(osTPrintf("BCacheAlloc()\n"));
    BCacheEntry *pBce = BCacheListFind(&bcg.aHashTbl[WFSKRN_BCACHE_HASH_FN(nAbsBlkAdr)], pVolInfo, nAbsBlkAdr);
    BCacheGroup *pGroup;
    if (pBce) { // if already cached
        if(&bcg.aGroup[nBCacheGroupId] != pBce->pGroup){
            WFSKrnOutputErrorStr("BCacheAllocMetaData: GroupIds do not correspond.\n");
        }
        pGroup = pBce->pGroup;
        if (nFlags & BCACHE_FLAG_PINNED) {
            if(!pBce->nPinCount) { // need to move to pinned list?
                BCacheListDelete(&pGroup->bclLru, pBce);
                BCacheListAddToHead(&pGroup->bclPinned, pBce);
            }
#if _DEBUG
#if !BTREE_BASED_ALLOCATOR
            else{
                if (bcg.bSkipCheckPin == false) {
                    WFSKrnOutputErrorStr("BCacheAllocMetaData: PinCount should be 0 or 1.\n");
                }
            }
#endif
#endif
            ++pBce->nPinCount;
#if _CHECK_BLK_USAGE
            pBce->nDbgCommandCount = nDbgCommandCount+1;
#endif
        }
        pBce->pBlkPtr = pGroup->pBCache + (pGroup->nBlkSize * PAGE_NUMBER_CE(pBce));
#if _CHECK_BLK_USAGE
        if ((pBce->nState & BCACHE_FLAG_DIRTY)==0) {
            u32 nCheckSum = BCacheCalcCheckSum((u8*)pBlkPtr, pGroup->nBlkSize);
            if (nCheckSum != pBce->nCheckSum) {
                // This is not necessarily a bug because some of Wayne's allocator code marks dirty just before calling Unpin.
                // The directory code should always mark the block dirty in advance though.
                MyOSReport("Warning: Change detected to block that is not marked dirty. nAbsBlkAdr=%d\n", nAbsBlkAdr);
                //WFSKrnOutputErrorStr("Checksum mismatch");
            }
        }
#endif
        nResult = BCacheSetHashPtrToEntry(pVolInfo, pBce, nHashBlkAdr, nHashOfs);
        pBce->nValidSize  = WFS_ROUND_UP(nValidSize, pDevInfo->dh.nLog2SectorSize); // round up secter byte
        if(nResult < WFSKRN_RESULT_OK){
            return nResult;
        }
        goto Done;
    }

    u8 *pBlkPtr;
    pGroup = &bcg.aGroup[nBCacheGroupId];
    nResult = BCacheAllocBlk(pGroup, (u8**)&pBlkPtr, &pBce);
    if(nResult < WFSKRN_RESULT_OK){
        return nResult;
    }
    pBce->pBlkPtr     = pBlkPtr;
    pBce->nAbsBlkAdr  = nAbsBlkAdr;
    pBce->pVolInfo    = pVolInfo;
    pBce->pDevInfo    = pDevInfo;
    pBce->nState      = BCACHE_FLAG_VALID;
    pBce->nValidSize  = WFS_ROUND_UP(nValidSize, 1<<pDevInfo->dh.nLog2SectorSize); // round up secter byte
    pBce->nChainId    = 0;

    nResult = BCacheSetHashPtrToEntry(pVolInfo, pBce, nHashBlkAdr, nHashOfs);
    if(nResult < WFSKRN_RESULT_OK){
        BCacheListAddToHead(&pGroup->bclInvalid, pBce);
        return nResult;
    }

    if (nFlags & BCACHE_FLAG_READ) { // intend to read existing value of page
        // Check hash data have been recorded.
        u32 nHashOfs = 0;
        if(pBce->nAbsBlkAdr != pBce->nHashBlkAdr){
            for(; nHashOfs<sizeof(WFSHashCode); nHashOfs+=sizeof(u32)){
                if(*(u32*)((size_t)pBce->pHash+nHashOfs) != 0 ){
                    break;
                }
            }
        }
        if(nHashOfs != sizeof(WFSHashCode) && pBce->nValidSize){
            nResult = DeviceReadSingleBlk(pDevInfo, nAbsBlkAdr<<(bcg.nLog2UnitBlkSize-pDevInfo->dh.nLog2SectorSize), (u8*)pBce->pBlkPtr,
                                          pBce->nValidSize>>pDevInfo->dh.nLog2SectorSize, pBce->pHash, pGroup->nLog2BlkSize-pDevInfo->dh.nLog2SectorSize);
        }
        /*
        if(nAbsBlkAdr == 6402 && pBce->nValidSize == 8192){
            int a; a=0;
        }
        static int c=0; c++;
        if(c > 10950){
            osTPrintf("c=%d ,nAbsBlkAdr = %d, nValidSize = %d\n", c, nAbsBlkAdr, pBce->nValidSize);
        }
        if(c == 11001){
            int a; a=0;
        }
        */
        if(nResult < WFSKRN_RESULT_OK){
            pBce->nState = BCACHE_FLAG_INVALID;
            pBce->nValidSize = nValidSize;
            BCacheListAddToHead(&pGroup->bclInvalid, pBce);
            switch(nResult){
            case WFSKRN_RESULT_DEVICE_HASH_INCONSISTENT:
                osTPrintf("Hash Error: nAbsBlkAdr = %d, nValidSize = %d\n", nAbsBlkAdr, pBce->nValidSize);
                return WFSKRN_RESULT_CORRUPTION;
            case WFSKRN_RESULT_DEVICE_NOT_FOUND:
            default:
                return WFSKRN_RESULT_MEDIA_ERROR;
                //return WFSKRN_RESULT_BCACHE_ERROR;
            }
        }
#if _CHECK_BLK_USAGE
        pBce->nCheckSum = BCacheCalcCheckSum((u8*)pBlkPtr, pGroup->nBlkSize);
    } else {
        pBce->nCheckSum = 0;
#endif
    }

    if (nFlags & BCACHE_FLAG_PINNED) {
#if _CHECK_BLK_USAGE
        pBce->nDbgCommandCount = nDbgCommandCount+1;
#endif
        pBce->nPinCount = 1;
        BCacheListAddToHead(&pGroup->bclPinned, pBce);
    } else {
        pBce->nPinCount = 0;
        BCacheListAddToTail(&pGroup->bclLru, pBce);     // add as MRU
    }
    BCacheListAddToHead(&bcg.aHashTbl[WFSKRN_BCACHE_HASH_FN(nAbsBlkAdr)], pBce);
    dbgb(osTPrintf("BCacheAlloc(0x%08x, %C%C%C) -> 0x%08x\n",
        pVolInfo, nAbsBlkAdr, (nFlags&BCACHE_FLAG_READ)?'R':'-', (nFlags&BCACHE_FLAG_PINNED)?'P':'-', (nFlags&BCACHE_FLAG_DIRTY)?'D':'-'));
Done:;
    if(nAbsBlkAdr != nHashBlkAdr && !(pBce->nState & BCACHE_FLAG_DIRTY) && (nFlags & BCACHE_FLAG_DIRTY)){
        nResult = BCachePin(pVolInfo, nHashBlkAdr);
        if(nResult < WFSKRN_RESULT_OK){
            return nResult;
        }
    }
    pBce->nState |= nFlags & BCACHE_FLAG_DIRTY;
    if(ppBce){
        *ppBce = pBce;
    }
#if _DEBUG_BREAK_POINT
    if(nAbsBlkAdr == 5693){
        int a; a=0;
    }
#endif
    return nResult;
}


void* BCacheFind(VolumeInfo *pVolInfo, WFSBlkAdr nAbsBlkAdr) {
    BCacheEntry *pBce = BCacheListFind(&bcg.aHashTbl[WFSKRN_BCACHE_HASH_FN(nAbsBlkAdr)], pVolInfo, nAbsBlkAdr);
    if (!pBce) return NULL;
    BCacheGroup *pGroup = pBce->pGroup;
    u8 *pBlkPtr = pGroup->pBCache + (pGroup->nBlkSize * PAGE_NUMBER_CE(pBce));
    return pBlkPtr;
}


// Write page to disk, mark as clean, and if non-pinned, move to head of LRU list since
// it is now clean.
WFSKrnResult BCacheStore(VolumeInfo *pVolInfo, WFSBlkAdr nAbsBlkAdr) {
    dbgb(osTPrintf("BCacheStore()\n"));
    DeviceInfo *pDevInfo = pVolInfo->pDevInfo;
    BCacheEntry *pBce = BCacheListFind(&bcg.aHashTbl[WFSKRN_BCACHE_HASH_FN(nAbsBlkAdr)], pVolInfo, nAbsBlkAdr);
    if (!pBce) return WFSKRN_RESULT_BCACHE_NOT_FOUND;
    if (!(pBce->nState & BCACHE_FLAG_VALID)) return WFSKRN_RESULT_BCACHE_ERROR;
    BCacheGroup *pGroup = pBce->pGroup;
    if (pBce->nState & BCACHE_FLAG_DIRTY) {
        u8 *pBlkPtr = pGroup->pBCache + (pGroup->nBlkSize * PAGE_NUMBER_CE(pBce));
        WFSKrnResult nResult = DeviceWriteSingleBlk(pDevInfo, pBce->nAbsBlkAdr<<(bcg.nLog2UnitBlkSize-pDevInfo->dh.nLog2SectorSize), pBlkPtr,
                                                    pBce->nValidSize>>pDevInfo->dh.nLog2SectorSize, pBce->pHash, pGroup->nLog2BlkSize-pDevInfo->dh.nLog2SectorSize, true);
        if(nResult < WFSKRN_RESULT_OK) {
            return WFSKRN_RESULT_MEDIA_ERROR;
        }
        if(pBce->nAbsBlkAdr != pBce->nHashBlkAdr){
            nResult = BCacheDirty(pDevInfo->pVolInfo, pBce->nHashBlkAdr);
            nResult = BCacheUnpin(pDevInfo->pVolInfo, pBce->nHashBlkAdr);
            if(nResult < WFSKRN_RESULT_OK){
                return nResult;
            }
        }
        pBce->nState &= ~BCACHE_FLAG_DIRTY;    // make clean; just in case we don't really replace
        if (!pBce->nPinCount){
            BCacheListMoveToHead(&pGroup->bclLru, pBce);
        }
        dbgb(osTPrintf("BCacheStore(0x%08x, %d)\n", pVolInfo, nAbsBlkAdr));
    }
#if _CHECK_BLK_USAGE
    else {
        u8 *pBlkPtr = pGroup->pBCache + (pGroup->nBlkSize * PAGE_NUMBER_CE(pBce));
        if (pBce->nCheckSum != BCacheCalcCheckSum(pBlkPtr, pGroup->nBlkSize)) {
            WFSKrnOutputErrorStr("Block contents changed, but was not marked dirty\n");
        }
    }
#endif
    return WFSKRN_RESULT_OK;
}


// Write page to disk, invalidate, and move to head of invalid list.  Applies to both
// non-pinned and pinned pages.  Illegal for invalid pages.
WFSKrnResult BCacheFlush(VolumeInfo *pVolInfo, WFSBlkAdr nAbsBlkAdr) {
    dbgb(osTPrintf("BCacheFlush()\n"));
    WFSKrnResult nResult = WFSKRN_RESULT_OK;
    DeviceInfo *pDevInfo = pVolInfo->pDevInfo;
    BCacheList *pHMapList = &bcg.aHashTbl[WFSKRN_BCACHE_HASH_FN(nAbsBlkAdr)];
    BCacheEntry *pBce = BCacheListFind(pHMapList, pVolInfo, nAbsBlkAdr);
    if (!pBce) return WFSKRN_RESULT_BCACHE_NOT_FOUND;
    if (!(pBce->nState & BCACHE_FLAG_VALID)) return WFSKRN_RESULT_BCACHE_ERROR;
    if (pBce->nPinCount){
        osTPrintf("BCacheFlush: The pinned block can not be flushed.");
    }
    BCacheGroup *pGroup = pBce->pGroup;
    if (pBce->nState & BCACHE_FLAG_DIRTY) {  // if writeback needed
        u8 *pBlkPtr = pGroup->pBCache + (pGroup->nBlkSize * PAGE_NUMBER_CE(pBce));
        nResult = DeviceWriteSingleBlk(pDevInfo, pBce->nAbsBlkAdr<<(bcg.nLog2UnitBlkSize-pDevInfo->dh.nLog2SectorSize), pBlkPtr,
                                       pBce->nValidSize>>pDevInfo->dh.nLog2SectorSize, pBce->pHash, pGroup->nLog2BlkSize-pDevInfo->dh.nLog2SectorSize, false);
        if(nResult < WFSKRN_RESULT_OK) {
            return WFSKRN_RESULT_MEDIA_ERROR;
        }
        if(pBce->nAbsBlkAdr != pBce->nHashBlkAdr){
            nResult = BCacheDirty(pDevInfo->pVolInfo, pBce->nHashBlkAdr);
            nResult = BCacheUnpin(pDevInfo->pVolInfo, pBce->nHashBlkAdr);
            if(nResult < WFSKRN_RESULT_OK){
                return nResult;
            }
        }
        pBce->nState &= ~BCACHE_FLAG_DIRTY;
        if (!pBce->nPinCount) BCacheListMoveToHead(&pGroup->bclLru, pBce);
    }
    BCacheList *pList = pBce->nPinCount ? &pGroup->bclPinned : &pGroup->bclLru;
    if (!BCacheListDelete(pList, pBce)) return WFSKRN_RESULT_BCACHE_ERROR;
    if (!BCacheListDelete(pHMapList, pBce)) return WFSKRN_RESULT_BCACHE_ERROR;
    BCACHE_CACHE_ENTRY_INVALIDATE(pBce);
    BCacheListAddToHead(&pGroup->bclInvalid, pBce);
    return nResult;
}



WFSKrnResult BCacheStoreVolume(VolumeInfo *pVolInfo) {
    dbgb(osTPrintf("BCacheStoreVolume()\n"));

#if WFS_BLK_CACHE_SIZE >= WFS_MAX_USER_BLK_CACHE_SIZE
    WFSBlkAdr aStoreBlkAdr[WFS_BLK_CACHE_SIZE>>WFS_LOG2_SMALL_BLK_SIZE];
#else
    WFSBlkAdr aStoreBlkAdr[WFS_MAX_USER_BLK_CACHE_SIZE>>WFS_LOG2_SMALL_BLK_SIZE];
#endif
    u32 nNumStoreBlks = 0;

    // Make list to store flushed block address for user cache.
    BCacheGroup *pGroup = 0;
    BCacheList *pList = 0;
    int nI,nJ;
    for(nI=0;nI<BCACHE_NUM_OF_GROUP*2;nI++){
        switch(nI){
            case 0:
                pGroup = &bcg.aGroup[BCACHE_GROUP_ID_USERBLK_8K];
                pList = (BCacheList*)&pGroup->bclPinned;
                break;
            case 2:
                pGroup = &bcg.aGroup[BCACHE_GROUP_ID_USERBLK_64K];
                pList = (BCacheList*)&pGroup->bclPinned;
                break;
            case 4:
                pGroup = &bcg.aGroup[BCACHE_GROUP_ID_METABLK];
                pList = (BCacheList*)&pGroup->bclPinned;
                break;
            case 1:
            case 3:
            case 5:
                pList = (BCacheList*)&pGroup->bclLru;
                break;
        }
        BCacheListNode *pBln = pList->pHead;
        for(nJ=0;nJ<pList->nLen;nJ++){
            BCacheEntry* pBce = (BCacheEntry*)pBln->pData;
            if(pVolInfo == pBce->pVolInfo){
                if(pBce->nState & BCACHE_FLAG_DIRTY){
                    aStoreBlkAdr[nNumStoreBlks++] = pBce->nAbsBlkAdr;
                }
            }
            pBln = pBln->pNext;
        }
        if (nI == 3 || nI == 5){
            if(nNumStoreBlks){
                _qsort((void*)aStoreBlkAdr, nNumStoreBlks, sizeof(WFSBlkAdr), CompareBlkAdrForward);
                for(nJ=0;nJ<nNumStoreBlks;nJ++){
                    WFSKrnReturnOnError(BCacheStore(pVolInfo, aStoreBlkAdr[nJ]));
                }
                nNumStoreBlks = 0;
            }
        }
    }
    return WFSKRN_RESULT_OK;
}


WFSKrnResult BCacheFlushVolume(VolumeInfo *pVolInfo) {
    dbgb(osTPrintf("BCacheFlushVolume()\n"));

#if WFS_BLK_CACHE_SIZE >= WFS_MAX_USER_BLK_CACHE_SIZE
    WFSBlkAdr aFlushBlkAdr[WFS_BLK_CACHE_SIZE>>WFS_LOG2_SMALL_BLK_SIZE];
    WFSBlkAdr aInvalidBlkAdr[WFS_BLK_CACHE_SIZE>>WFS_LOG2_SMALL_BLK_SIZE];
#else
    WFSBlkAdr aFlushBlkAdr[WFS_MAX_USER_BLK_CACHE_SIZE>>WFS_LOG2_SMALL_BLK_SIZE];
    WFSBlkAdr aInvalidBlkAdr[WFS_MAX_USER_BLK_CACHE_SIZE>>WFS_LOG2_SMALL_BLK_SIZE];
#endif
    u32 nNumFlushBlks = 0;
    u32 nNumInvalidBlks = 0;

    // Make list to store flushed block address for user cache.
    BCacheGroup *pGroup = 0;
    BCacheList *pList = 0;
    int nI,nJ;
    for(nI=0;nI<BCACHE_NUM_OF_GROUP*2;nI++){
        switch(nI){
            case 0:
                pGroup = &bcg.aGroup[BCACHE_GROUP_ID_USERBLK_8K];
                pList = (BCacheList*)&pGroup->bclPinned;
                break;
            case 2:
                pGroup = &bcg.aGroup[BCACHE_GROUP_ID_USERBLK_64K];
                pList = (BCacheList*)&pGroup->bclPinned;
                break;
            case 4:
                pGroup = &bcg.aGroup[BCACHE_GROUP_ID_METABLK];
                pList = (BCacheList*)&pGroup->bclPinned;
                break;
            case 1:
            case 3:
            case 5:
                pList = (BCacheList*)&pGroup->bclLru;
                break;
        }
        BCacheListNode *pBln = pList->pHead;
        for(nJ=0;nJ<pList->nLen;nJ++){
            BCacheEntry* pBce = (BCacheEntry*)pBln->pData;
            if(pVolInfo == pBce->pVolInfo){
                if(pBce->nState & BCACHE_FLAG_DIRTY){
                    aFlushBlkAdr[nNumFlushBlks++] = pBce->nAbsBlkAdr;
                } else if (!pBce->nPinCount){
                    aInvalidBlkAdr[nNumInvalidBlks++] = pBce->nAbsBlkAdr;
                }
            }
            pBln = pBln->pNext;
        }
        if (nI == 3 || nI == 5){
            if(nNumFlushBlks){
                _qsort((void*)aFlushBlkAdr, nNumFlushBlks, sizeof(WFSBlkAdr), CompareBlkAdrForward);
                for(nJ=0;nJ<nNumFlushBlks;nJ++){
                    WFSKrnReturnOnError(BCacheFlush(pVolInfo, aFlushBlkAdr[nJ]));
                }
                nNumFlushBlks = 0;
            }
            if(nNumInvalidBlks){
                _qsort((void*)aInvalidBlkAdr, nNumInvalidBlks, sizeof(WFSBlkAdr), CompareBlkAdrForward);
                for(nJ=0;nJ<nNumInvalidBlks;nJ++){
                    WFSKrnReturnOnError(BCacheInvalidate(pVolInfo, aInvalidBlkAdr[nJ]));
                }
                nNumInvalidBlks = 0;
            }
        }
    }
    return WFSKRN_RESULT_OK;
}



// Invalidate block. No write-backs.
WFSKrnResult BCacheInvalidate(VolumeInfo *pVolInfo, WFSBlkAdr nAbsBlkAdr) {
    dbgb(osTPrintf("BCacheInvalidate()\n"));
    BCacheList  *pHMapList = &bcg.aHashTbl[WFSKRN_BCACHE_HASH_FN(nAbsBlkAdr)];
    BCacheEntry *pBce = BCacheListFind(pHMapList, pVolInfo, nAbsBlkAdr);
    if (!pBce) return WFSKRN_RESULT_OK;
    if (pBce->nPinCount){
        osTPrintf("BCacheInvalidate: The pinned block can not be invalidated.\n");
    }
    //BCacheList *pList = (pBce->nState & BCACHE_FLAG_PINNED) ? &bcg.bclPinned : &bcg.bclLru;
    BCacheGroup *pGroup = pBce->pGroup;
    BCacheList *pList = pBce->nPinCount ? &pGroup->bclPinned : &pGroup->bclLru;
    if (pBce->nState & BCACHE_FLAG_DIRTY && pBce->nAbsBlkAdr != pBce->nHashBlkAdr) {
        WFSKrnReturnOnError(BCacheUnpin(pVolInfo, pBce->nHashBlkAdr));
    }
    if (!BCacheListDelete(pList, pBce)) return WFSKRN_RESULT_BCACHE_ERROR;
    if (!BCacheListDelete(pHMapList, pBce)) return WFSKRN_RESULT_BCACHE_ERROR;
    BCACHE_CACHE_ENTRY_INVALIDATE(pBce);
    BCacheListAddToHead(&pGroup->bclInvalid, pBce);
    return WFSKRN_RESULT_OK;
}


// Invalidate blocks. The function search the indicated block in the cache lists.
WFSKrnResult BCacheInvalidateBlks (struct VolumeInfo_ *pVolInfo, WFSBlkAdr nAbsBlkAdr, u32 nNumBlks){
    dbgb(osTPrintf("BCacheInvalidateBlks()\n"));
    s32 nI;
    u32 InvalidBlks = 0;
    for (nI=bcg.nNumGroups-1;nI>=0;nI--){
        BCacheGroup *pGroup = &bcg.aGroup[nI];
        BCacheList *pList = &pGroup->bclLru;
        BCacheListNode *pNode = pList->pHead;
        while (NULL != pNode)
        {
            BCacheEntry *pBce = (BCacheEntry*)pNode->pData;
            if (!pBce)
            {
                return WFSKRN_RESULT_OK;
            }

            if (pVolInfo == pBce->pVolInfo && pBce->nAbsBlkAdr >= nAbsBlkAdr && pBce->nAbsBlkAdr < nAbsBlkAdr+nNumBlks)
            {
                if (pBce->nPinCount){
                    osTPrintf("BCacheInvalidateBlks: The pinned block can not be invalidated.\n");
                }
                if (pBce->nState & BCACHE_FLAG_DIRTY && pBce->nAbsBlkAdr != pBce->nHashBlkAdr) {
                    WFSKrnReturnOnError(BCacheUnpin(pVolInfo, pBce->nHashBlkAdr));
                }
                BCacheListNode *pNodeNext = pNode->pNext;
                if (!BCacheNodeDelete(pList, pNode)) return WFSKRN_RESULT_BCACHE_ERROR;
                BCacheList *pHMapList = &bcg.aHashTbl[WFSKRN_BCACHE_HASH_FN(pBce->nAbsBlkAdr)];
                if (!BCacheListDelete(pHMapList, pBce)) return WFSKRN_RESULT_BCACHE_ERROR;
                BCACHE_CACHE_ENTRY_INVALIDATE(pBce);
                BCacheListAddToHead(&pGroup->bclInvalid, pBce);
                ++InvalidBlks;
                if (InvalidBlks == nNumBlks) return WFSKRN_RESULT_OK;
                pNode = pNodeNext;
            }
            else{
                pNode = pNode->pNext;
            }
        }
#if _DEBUG
        pList = &pGroup->bclPinned;
        pNode = pList->pHead;
        while (NULL != pNode) {
            BCacheEntry *pBce = (BCacheEntry*)pNode->pData;
            if (pVolInfo == pBce->pVolInfo && pBce->nAbsBlkAdr >= nAbsBlkAdr && pBce->nAbsBlkAdr < nAbsBlkAdr+nNumBlks) {
                osTPrintf("BCacheInvalidateBlks.\n");
            }
            pNode = pNode->pNext;
        }
#endif
    }
    return WFSKRN_RESULT_OK;
}


// If valid and non-pinned, make MRU. No-op for pinned block.
WFSKrnResult BCacheTouch(VolumeInfo *pVolInfo, WFSBlkAdr nAbsBlkAdr) {
    dbgb(osTPrintf("BCacheTouch()\n"));
    BCacheEntry *pBce = BCacheListFind(&bcg.aHashTbl[WFSKRN_BCACHE_HASH_FN(nAbsBlkAdr)], pVolInfo, nAbsBlkAdr);
    if (!pBce) return WFSKRN_RESULT_BCACHE_NOT_FOUND;
    if (!(pBce->nState & BCACHE_FLAG_VALID)) return WFSKRN_RESULT_BCACHE_ERROR;
    BCacheGroup *pGroup = pBce->pGroup;
    if (!pBce->nPinCount) {
        if (BCacheListMoveToHead(&pGroup->bclLru, pBce)) {
            return WFSKRN_RESULT_OK;
        } else {
            return WFSKRN_RESULT_BCACHE_ERROR;
        }
    }
    return WFSKRN_RESULT_OK;
}


// Make dirty.  Illegal for non-valid pages.  Make MRU or not?
WFSKrnResult BCacheDirty(VolumeInfo *pVolInfo, WFSBlkAdr nAbsBlkAdr) {
    dbgb(osTPrintf("BCacheDirty()\n"));
    BCacheEntry *pBce = BCacheListFind(&bcg.aHashTbl[WFSKRN_BCACHE_HASH_FN(nAbsBlkAdr)], pVolInfo, nAbsBlkAdr);
    if (!pBce) return WFSKRN_RESULT_BCACHE_NOT_FOUND;
    if (!(pBce->nState & BCACHE_FLAG_VALID)) return WFSKRN_RESULT_BCACHE_ERROR;
    pBce->nState |= BCACHE_FLAG_DIRTY;
    if(pBce->nAbsBlkAdr != pBce->nHashBlkAdr){
        WFSKrnResult nResult = BCacheAllocMetaBlk(pVolInfo, pBce->nHashBlkAdr, BCACHE_FLAG_READ | BCACHE_FLAG_PINNED | BCACHE_FLAG_DIRTY, 0);
        if(nResult < WFSKRN_RESULT_OK){
            return nResult;
        }
    }
    return WFSKRN_RESULT_OK;
}


// Pin a block in memory. Move it from the LRU/MRU list to the pinned list.
WFSKrnResult BCachePin(VolumeInfo *pVolInfo, WFSBlkAdr nAbsBlkAdr) {
    dbgb(osTPrintf("BCachePin()\n"));
    BCacheEntry *pBce = BCacheListFind(&bcg.aHashTbl[WFSKRN_BCACHE_HASH_FN(nAbsBlkAdr)], pVolInfo, nAbsBlkAdr);
    if (!pBce) return WFSKRN_RESULT_BCACHE_NOT_FOUND;
    if (!(pBce->nState & BCACHE_FLAG_VALID)) return WFSKRN_RESULT_BCACHE_ERROR;
    if (!pBce->nPinCount) {
        BCacheGroup *pGroup = pBce->pGroup;
        if (!BCacheListDelete(&pGroup->bclLru, pBce)) return WFSKRN_RESULT_BCACHE_ERROR;
        if (!BCacheListAddToHead(&pGroup->bclPinned, pBce))  return WFSKRN_RESULT_BCACHE_ERROR;
    }
    //else{
    //    WFSKrnOutputErrorStr("BCacheUnpin: PinCount have to be 0 or 1.\n");
    //}
    ++pBce->nPinCount;
#if _DEBUG_BREAK_POINT
    if(nAbsBlkAdr == 5693){
        int a; a=0;
    }
#endif
    return WFSKRN_RESULT_OK;
}


// Unpin a block in memory; can be replaced.  Make MRU.
WFSKrnResult BCacheUnpin(VolumeInfo *pVolInfo, WFSBlkAdr nAbsBlkAdr) {
    dbgb(osTPrintf("BCacheUnpin()\n"));
    BCacheEntry *pBce = BCacheListFind(&bcg.aHashTbl[WFSKRN_BCACHE_HASH_FN(nAbsBlkAdr)], pVolInfo, nAbsBlkAdr);
    if (!pBce) return WFSKRN_RESULT_BCACHE_NOT_FOUND;
    if (!(pBce->nState & BCACHE_FLAG_VALID)) return WFSKRN_RESULT_BCACHE_ERROR;

    if(!pBce->nPinCount){
        WFSKrnOutputErrorStr("BCacheUnpin: PinCount cannot be reduced.\n");
    } else if (pBce->nPinCount == 1) {
        BCacheGroup *pGroup = pBce->pGroup;
        if (!BCacheListDelete(&pGroup->bclPinned, pBce)) return WFSKRN_RESULT_BCACHE_ERROR;
        if (!BCacheListAddToTail(&pGroup->bclLru, pBce))  return WFSKRN_RESULT_BCACHE_ERROR;
    }
    --pBce->nPinCount;
#if _DEBUG_BREAK_POINT
    if(nAbsBlkAdr == 5693){
        int a; a=0;
    }
#endif
    return WFSKRN_RESULT_OK;
}


u32 BCacheGetPinCount(VolumeInfo *pVolInfo, WFSBlkAdr nAbsBlkAdr) {
    dbgb(osTPrintf("BCacheGetPinCount()\n"));
    BCacheEntry *pBce = BCacheListFind(&bcg.aHashTbl[WFSKRN_BCACHE_HASH_FN(nAbsBlkAdr)], pVolInfo, nAbsBlkAdr);
    if (!pBce) return (u32)WFSKRN_RESULT_BCACHE_NOT_FOUND;
    if (!(pBce->nState & BCACHE_FLAG_VALID)) return (u32)WFSKRN_RESULT_BCACHE_ERROR;
    return pBce->nPinCount;
}


// Check to see that the BCache is consistent.
//  - check the lists' lengths and pages belong
WFSKrnResult BCacheCheck(void) {
    dbgb(osTPrintf("BCacheCheck()\n"));
    // check LRU list
    u32 nI;
    for (nI=0;nI<bcg.nNumGroups;nI++){
        BCacheGroup *pGroup = &bcg.aGroup[nI];
        u32 nLen = 0;
        BCacheListNode *pListNode = pGroup->bclLru.pHead;
        while (pListNode != NULL) {
            BCacheEntry *pBce = (BCacheEntry*) pListNode->pData;
            if (!(pBce->nState & BCACHE_FLAG_VALID)) {
                dbgb(osTPrintf("BCacheCheck: invalid block (%d) in LRU list\n", PAGE_NUMBER_CE(pBce)));
                return WFSKRN_RESULT_BCACHE_ERROR;
            }
            if (pBce->nPinCount) {
                dbgb(osTPrintf("BCacheCheck: pinned block (%d) in LRU list\n", PAGE_NUMBER_CE(pBce)));
                return WFSKRN_RESULT_BCACHE_ERROR;
            }
            pListNode = pListNode->pNext;
            nLen++;
        }
        if (pGroup->bclLru.nLen != nLen) {
            dbgb(osTPrintf("BCacheCheck: LRU list length inconsistent, %d != %d\n", bcg.bclLru.nLen, nLen));
            return WFSKRN_RESULT_BCACHE_ERROR;
        }

        // check pinned list
        nLen = 0;
        pListNode = pGroup->bclPinned.pHead;
        while (pListNode != NULL) {
            BCacheEntry *pBce = (BCacheEntry*) pListNode->pData;
            if (!(pBce->nState & BCACHE_FLAG_VALID)) {
                dbgb(osTPrintf("BCacheCheck: invalid block (%d) in pinned list\n", PAGE_NUMBER_CE(pBce)));
                return WFSKRN_RESULT_BCACHE_ERROR;
            }
            if (!pBce->nPinCount) {
                dbgb(osTPrintf("BCacheCheck: non-pinned block (%d) in pinned list\n", PAGE_NUMBER_CE(pBce)));
                return WFSKRN_RESULT_BCACHE_ERROR;
            }
            pListNode = pListNode->pNext;
            nLen++;
        }
        if (pGroup->bclPinned.nLen != nLen) {
            dbgb(osTPrintf("BCacheCheck: pinned list length inconsistent, %d != %d\n", bcg.bclPinned.nLen, nLen));
            return WFSKRN_RESULT_BCACHE_ERROR;
        }

        // check invalid list
        nLen = 0;
        pListNode = pGroup->bclInvalid.pHead;
        while (pListNode != NULL) {
            BCacheEntry *pBce = (BCacheEntry*) pListNode->pData;
            if (pBce->nState & BCACHE_FLAG_VALID) {
                dbgb(osTPrintf("BCacheCheck: valid block (%d) in invalid list\n", PAGE_NUMBER_CE(pBce)));
                return WFSKRN_RESULT_BCACHE_ERROR;
            }
            pListNode = pListNode->pNext;
            nLen++;
        }
        if (pGroup->bclInvalid.nLen != nLen) {
            dbgb(osTPrintf("BCacheCheck: invalid list length inconsistent, %d != %d\n", bcg.bclInvalid.nLen, nLen));
            return WFSKRN_RESULT_BCACHE_ERROR;
        }

        if (pGroup->nNumBlks != (pGroup->bclLru.nLen + pGroup->bclPinned.nLen + pGroup->bclInvalid.nLen)) {
            dbgb(osTPrintf("BCacheCheck: missing a block, %d != %d+%d+%d\n",
                pGroup->nNumBlks, pGroup->bclLru.nLen, pGroup->bclPinned.nLen, pGroup->bclInvalid.nLen));
        }
    }
    return WFSKRN_RESULT_OK;
}


void* BCachePointer(VolumeInfo *pVolInfo, WFSBlkAdr nAbsBlkAdr) {
    dbgb(osTPrintf("BCachePointer()\n"));
    BCacheEntry *pBce = BCacheListFind(&bcg.aHashTbl[WFSKRN_BCACHE_HASH_FN(nAbsBlkAdr)], pVolInfo, nAbsBlkAdr);
    if (!pBce) {                                   // if already cached
        return (void *)WFSKRN_RESULT_BCACHE_NOT_FOUND;
    }
    BCacheGroup *pGroup = pBce->pGroup;
    return pGroup->pBCache + (pGroup->nBlkSize * PAGE_NUMBER_CE(pBce));
}

// This must be called at the end of a test program.
// In addition to BCacheCheck(),
// BCacheCheck2() checks if all caches are unpinned.
WFSKrnResult BCacheCheck2(void)
{
    dbgb(osTPrintf("BCacheCheck2()\n"));
    // check LRU list
    u32 nI;
    for (nI=0;nI<bcg.nNumGroups;nI++){
        BCacheGroup *pGroup = &bcg.aGroup[nI];
        u32 nLen = 0;
        WFSBool pinned = (WFSBool) FALSE;
        BCacheListNode *pListNode = pGroup->bclLru.pHead;
        while (pListNode != NULL) {
            BCacheEntry *pBce = (BCacheEntry*) pListNode->pData;
            if (!(pBce->nState & BCACHE_FLAG_VALID)) {
                dbgb(osTPrintf("BCacheCheck2: invalid block (%d) in LRU list\n", PAGE_NUMBER_CE(pBce)));
                return WFSKRN_RESULT_BCACHE_ERROR;
            }
            if (pBce->nPinCount) {
                dbgb(osTPrintf("BCacheCheck2: pinned block (%d) in LRU list\n", PAGE_NUMBER_CE(pBce)));
                return WFSKRN_RESULT_BCACHE_ERROR;
            }
            if (pBce->nPinCount)
            {
                osTPrintf("BCacheCheck2: blk %u in LRU list: unexpectedly pinned. (%u)\n",
                    pBce->nAbsBlkAdr, pBce->nPinCount);
                return WFSKRN_RESULT_BCACHE_ERROR;
            }
            pListNode = pListNode->pNext;
            nLen++;
        }
        if (pGroup->bclLru.nLen != nLen) {
            dbgb(osTPrintf("BCacheCheck2: LRU list length inconsistent, %d != %d\n", pGroup->bclLru.nLen, nLen));
            return WFSKRN_RESULT_BCACHE_ERROR;
        }

        // check pinned list
        nLen = 0;
        pListNode = pGroup->bclPinned.pHead;
        if (pListNode)
        {
            pinned = (WFSBool) TRUE;
        }

        while (pListNode != NULL) {
            BCacheEntry *pBce = (BCacheEntry*) pListNode->pData;
            if (!(pBce->nState & BCACHE_FLAG_VALID)) {
                dbgb(osTPrintf("BCacheCheck2: invalid block (%d) in pinned list\n", PAGE_NUMBER_CE(pBce)));
                return WFSKRN_RESULT_BCACHE_ERROR;
            }
            if (!pBce->nPinCount) {
                dbgb(osTPrintf("BCacheCheck2: non-pinned block (%d) in pinned list\n", PAGE_NUMBER_CE(pBce)));
                return WFSKRN_RESULT_BCACHE_ERROR;
            }
            if (pBce->nPinCount)
            {
                osTPrintf("BCacheCheck2: blk %u in pinned list: unexpectedly pinned. (%u)\n",
                    pBce->nAbsBlkAdr, pBce->nPinCount);
                // return WFSKRN_RESULT_BCACHE_ERROR;
            }
            pListNode = pListNode->pNext;
            nLen++;
        }
        if (pinned)
        {
            return WFSKRN_RESULT_BCACHE_ERROR;
        }

        if (pGroup->bclPinned.nLen != nLen) {
            dbgb(osTPrintf("BCacheCheck2: pinned list length inconsistent, %d != %d\n", pGroup->bclPinned.nLen, nLen));
            return WFSKRN_RESULT_BCACHE_ERROR;
        }

        // check invalid list
        nLen = 0;
        pListNode = pGroup->bclInvalid.pHead;
        while (pListNode != NULL) {
            BCacheEntry *pBce = (BCacheEntry*) pListNode->pData;
            if (pBce->nState & BCACHE_FLAG_VALID) {
                dbgb(osTPrintf("BCacheCheck2: valid block (%d) in invalid list\n", PAGE_NUMBER_CE(pBce)));
                return WFSKRN_RESULT_BCACHE_ERROR;
            }
            /* [check] Should pBce->nPinCount be zero?
            if (pBce->nPinCount)
            {
            osTPrintf("BCacheCheck2: blk %u in invalid list: unexpectedly pinned. (%u)\n",
            pBce->nAbsBlkAdr, pBce->nPinCount);
            // return WFSKRN_RESULT_BCACHE_ERROR;
            }
            */
            pListNode = pListNode->pNext;
            nLen++;
        }
        if (pGroup->bclInvalid.nLen != nLen) {
            dbgb(osTPrintf("BCacheCheck2: invalid list length inconsistent, %d != %d\n", pGroup->bclInvalid.nLen, nLen));
            return WFSKRN_RESULT_BCACHE_ERROR;
        }

        if (pGroup->nNumBlks != (pGroup->bclLru.nLen + pGroup->bclPinned.nLen + pGroup->bclInvalid.nLen)) {
            dbgb(osTPrintf("BCacheCheck2: missing a block, %d != %d+%d+%d\n",
                pGroup->nNumBlks, pGroup->bclLru.nLen, pGroup->bclPinned.nLen, pGroup->bclInvalid.nLen));
        }
    }
    return WFSKRN_RESULT_OK;
}

static u32 GetTotalPinCount()
{
    u32 totalPinCount = 0;
    BCacheGroup* pGroup;
    BCacheListNode* pListNode;
    BCacheEntry* pBce;

    int index;
    for (index = 0; index < bcg.nNumGroups; ++index)
    {
        pGroup = &bcg.aGroup[index];
        pListNode = pGroup->bclPinned.pHead;

        while (pListNode != NULL)
        {
            pBce = (BCacheEntry*) pListNode->pData;
            totalPinCount += pBce->nPinCount;
            pListNode = pListNode->pNext;
        }
    }

    return totalPinCount;
}

// stop pin count check.
u32 BCacheSkipCheckPin(void)
{
    u32 totalPinCount = GetTotalPinCount();
#if _DEBUG
    bcg.bSkipCheckPin = true;
#endif
    return totalPinCount;
}

// restart pin count check.
WFSKrnResult BCacheStopSkipCheckPin(u32 initialPinCount)
{
    u32 totalPinCount = GetTotalPinCount();
#if _DEBUG
    bcg.bSkipCheckPin = false;
#endif
    if (totalPinCount == initialPinCount)
    {
        return WFSKRN_RESULT_OK;
    }
    BCacheDumpPinCount();
    return WFSKRN_RESULT_FATAL_ERROR; // Pin counts are inconsistent.
}

u32 BCacheDumpPinCount(void)
{
    u32 totalPinCount = 0;
    BCacheGroup* pGroup;
    BCacheListNode* pListNode;
    BCacheEntry* pBce;

    int index;
    for (index = 0; index < bcg.nNumGroups; ++index)
    {
        pGroup = &bcg.aGroup[index];
        pListNode = pGroup->bclPinned.pHead;

        while (pListNode != NULL)
        {
            pBce = (BCacheEntry*) pListNode->pData;
            osTPrintf("Addr %d, PC %d\n", pBce->nAbsBlkAdr, pBce->nPinCount);
            totalPinCount += pBce->nPinCount;
            pListNode = pListNode->pNext;
        }
    }
    return totalPinCount;
}
