/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfskrn_BitField.cpp - functions for manipulating bitfields within bit arrays


  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_BitField.cpp,v $
  Revision 1.4  2008/12/22 08:57:15  ueno
  Fixed WfsSetBitField() when 32 < nNumBits.

  Revision 1.3  2008/07/17 22:45:41  paul
  fixed a performance problem

  Revision 1.2  2008/05/10 03:59:43  kondo_masahiro
  Merged and fixed code to accept RM compiler

  Revision 1.4  2008/04/19 05:51:14  kondo_masahiro
  Fixed code to accept RM compiler

  Revision 1.3  2008/04/04 20:54:35  paul
  Added WfsSetBitField, WfsTestAndSetBitField (used by dir check code to map coverage of used and free parts of block)

  Revision 1.2  2008/02/28 06:32:02  kondo_masahiro
  Fixed code to accept PPC compiler

  Revision 1.1  2007/12/27 11:14:47  paul
  Most of this is not used at the moment.. but I'll leave it here for now

 *---------------------------------------------------------------------------*/

#include "wfskrn_BitField.h"
#include "wfskrn_Platform.h"

#ifdef _DEBUG
    #define _BITFIELD_NUM_TESTS 10000
#else
    #define _BITFIELD_NUM_TESTS 10000
#endif

#define BITFIELD_MASK ((1<<nNumBits)-1)


u32 WfsCalculateFieldSizeInBytes(u32 nVal) {
    if (nVal<256) return 1; else if (nVal<256) return 2; else if (nVal<256) return 3; else return 4;
}

u8 WfsCalculateFieldSizeInBits(u32 nVal) {
    u8 nRet=0;
    static const u8 aBitTbl[16] = { 0, 1, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4 };
    for(; nVal>=16; nVal>>=4) {
        nRet+=4;
    }
    return (u8)( nRet + aBitTbl[nVal] );
}

unsigned WfsReadBitField(unsigned *aBitArray, int nStartBit, int nNumBits) {
    // Masking should be performed by the caller
    unsigned *pBitArray = &aBitArray[nStartBit>>BIT_OFFSET_TO_INT_OFFSET_SHIFT];
    int nShift=BITS_PER_INT-(nStartBit&BIT_OFFSET_MASK)-nNumBits;
    if (nShift>0) {
        return (pBitArray[0]>>nShift);
    } else {
        if (nShift==0) {
            return pBitArray[0];
        } else {
            return pBitArray[1]>>(BITS_PER_INT+nShift) | pBitArray[0]<<(-nShift);
        }
    }
}


#if BITFIELD_USE_TEMPLATES
template<unsigned nNumBits>
unsigned WfsReadBitField (unsigned *aBitArray, int nStartBit) {
    unsigned *pBitArray = &aBitArray[nStartBit>>BIT_OFFSET_TO_INT_OFFSET_SHIFT];
    int nShift=BITS_PER_INT-(nStartBit&BIT_OFFSET_MASK)-nNumBits;
    if (nShift>0) {
        return (pBitArray[0]>>nShift) & BITFIELD_MASK;
    } else {
        if (nShift==0) {
            return pBitArray[0] & BITFIELD_MASK;
        } else {
            return (pBitArray[1]>>(BITS_PER_INT+nShift) | pBitArray[0]<<(-nShift)) & BITFIELD_MASK;
        }
    }
}
#endif //BITFIELD_USE_TEMPLATES


void WfsOverwriteBitField(unsigned *aBitArray, int nStartBit, unsigned nValue, int nNumBits, unsigned nMask) {
    unsigned *pBitArray = &aBitArray[nStartBit>>BIT_OFFSET_TO_INT_OFFSET_SHIFT];
    int nShift=BITS_PER_INT-(nStartBit&BIT_OFFSET_MASK)-nNumBits;
    if (nShift>0)
        pBitArray[0] = (pBitArray[0]&~(nMask<<nShift)) | (nValue<<nShift);
    else if (nShift<0) {
        int nS1 = -nShift;
        pBitArray[0] = (pBitArray[0]&~(nMask>>nS1)) | (nValue>>nS1);
        int nS2 = BITS_PER_INT+nShift;
        pBitArray[1] = (pBitArray[1]&~(nMask<<nS2)) | (nValue<<nS2);
    } else {
        pBitArray[0] = (pBitArray[0]&~nMask) | nValue;
    }
}

#if BITFIELD_USE_TEMPLATES
template<unsigned nNumBits>
void WfsOverwriteBitField(unsigned *aBitArray, int nStartBit, int nValue) {
    unsigned *pBitArray = &aBitArray[nStartBit>>BIT_OFFSET_TO_INT_OFFSET_SHIFT];
    int nShift=BITS_PER_INT-(nStartBit&BIT_OFFSET_MASK)-nNumBits;
    if (nShift>0)
        pBitArray[0] = (pBitArray[0]&~(BITFIELD_MASK<<nShift)) | (nValue<<nShift);
    else {
        int nS1 = -nShift;
        pBitArray[0] = (pBitArray[0]&~(BITFIELD_MASK>>nS1)) | (nValue>>nS1);
        int nS2 = BITS_PER_INT+nShift;
        pBitArray[1] = (pBitArray[1]&~(BITFIELD_MASK<<nS2)) | (nValue<<nS2);
    }
}
#endif //BITFIELD_USE_TEMPLATES


void WfsBitArrayInsertBitField(unsigned *aBitArray, int nBitArrayLength, int nBitPos, int nNumBits, unsigned nValue) {
    unsigned *pCopyFrom    = &aBitArray[nBitArrayLength>>BIT_OFFSET_TO_INT_OFFSET_SHIFT];
    unsigned *pCopyTo      = &aBitArray[(nBitArrayLength+nNumBits)>>BIT_OFFSET_TO_INT_OFFSET_SHIFT];
    unsigned *pStart       = &aBitArray[nBitPos>>BIT_OFFSET_TO_INT_OFFSET_SHIFT];
    unsigned nOldStartWord = *pStart;
    unsigned nStartShift   = (u32)(nBitPos&BIT_OFFSET_MASK);
    unsigned nStartMask    = ((unsigned)-1) >> nStartShift;
    // Insert right side of value
    s32 nValueShift = (s32)(BITS_PER_INT - nStartShift);
    *pStart &= nStartMask;
    if (nStartShift) {
        *pStart |= (nValue<<nValueShift);
    }
    if (nNumBits!=BITS_PER_INT) {
        unsigned nShiftLeft    = (u32)(BITS_PER_INT - nNumBits);
        unsigned nTrailingBits = (u32)(nBitArrayLength & BIT_OFFSET_MASK);
        if (nTrailingBits >= nShiftLeft) {
            if (nTrailingBits > nShiftLeft) {
                *pCopyTo = *pCopyFrom<<nShiftLeft;
            }
            pCopyTo--;
        }
        while(pCopyFrom > pStart) {
            *pCopyTo-- = (pCopyFrom[-1]<<nShiftLeft) | (pCopyFrom[0]>>nNumBits);
            pCopyFrom--;
        }
        *pCopyTo-- = pCopyFrom[0]>>nNumBits;
    } else {
        // Need a special case for no shift right, becuase shift left by 32 does nothing
        if ((nBitArrayLength & BIT_OFFSET_MASK)==0) {
            pCopyTo--;
            pCopyFrom--;
        }
        while(pCopyFrom >= pStart) {
            *pCopyTo-- = *pCopyFrom--;
        }
        *pCopyTo = 0;
    }
    // Insert left side of value
    *pStart |= nOldStartWord & ~nStartMask;
    nValueShift -= nNumBits;
    if (nValueShift > 0) {
        *pStart |= nValue<<nValueShift;
    } else {
        *pStart |= nValue>>-nValueShift;
    }
}

void WfsBitArrayDeleteBitField(unsigned *aBitArray, int nBitArrayLength, int nBitPos, int nNumBits) {
    unsigned *pCopyStart   = &aBitArray[ nBitPos          >>BIT_OFFSET_TO_INT_OFFSET_SHIFT];
    unsigned nOldCopyStart = *pCopyStart & ~(((unsigned)-1)>>(nBitPos&BIT_OFFSET_MASK));
    unsigned nFromBit      = (u32)(nBitPos+nNumBits);
    unsigned *pCopyFrom    = &aBitArray[nFromBit>>BIT_OFFSET_TO_INT_OFFSET_SHIFT];
    unsigned *pCopyTo      = pCopyStart;
    unsigned *pCopyEnd     = &aBitArray[(nBitArrayLength-nNumBits)>>BIT_OFFSET_TO_INT_OFFSET_SHIFT];
    unsigned nShiftLeft    = (u32)(nNumBits & BIT_OFFSET_MASK);
    unsigned nShiftRight   = BITS_PER_INT - nShiftLeft;
   *pCopyFrom &= (((unsigned)-1) >> (nFromBit&BIT_OFFSET_MASK));
    if (pCopyTo == pCopyFrom) {
        while(pCopyTo <= pCopyEnd) {
            pCopyTo[0] = (pCopyTo[0]<<nShiftLeft) | (pCopyTo[1]>>nShiftRight);
            pCopyTo++;
        }
    } else {
        if (nShiftLeft) {
            if ((((pCopyFrom-pCopyTo)*BITS_PER_INT) + nShiftLeft) > nNumBits) {
                *pCopyTo++ = pCopyFrom[0]>>nShiftRight;
            }
            while(pCopyTo <= pCopyEnd) {
                pCopyTo[0] = (pCopyFrom[0]<<nShiftLeft) | (pCopyFrom[1]>>nShiftRight);
                pCopyTo++; pCopyFrom++;
            }
        } else {
            while(pCopyTo <= pCopyEnd) {
                *pCopyTo++ = *pCopyFrom++;
            }
        }
    }
    *pCopyStart |= nOldCopyStart;
}

void WfsSetBitField(unsigned *aBitArray, int nStartBit, int nNumBits) {
    unsigned *pBitArray = &aBitArray[nStartBit>>BIT_OFFSET_TO_INT_OFFSET_SHIFT];
    unsigned nEndBit    = (unsigned)(nStartBit + nNumBits - 1);
    unsigned *pEnd      = &aBitArray[nEndBit>>BIT_OFFSET_TO_INT_OFFSET_SHIFT];
    unsigned nLeftShift = nStartBit&BIT_OFFSET_MASK;
    unsigned nRightShift= (nEndBit&BIT_OFFSET_MASK)^BIT_OFFSET_MASK;
    if (pBitArray == pEnd) {
        *pBitArray |= (((unsigned)-1)<<nLeftShift) & (((unsigned)-1)>>nRightShift);
    } else {
        *pBitArray |= ((u32)-1)<<nLeftShift;
        nNumBits -= (BITS_PER_INT - nLeftShift);
        while(nNumBits >= BITS_PER_INT) {
            *(++pBitArray) = ((u32)-1);
            nNumBits -= BITS_PER_INT;
        }
        *(++pBitArray) |= ((u32)-1)>>nRightShift;
    }
}

unsigned WfsTestAndSetBitField(unsigned *aBitArray, int nStartBit, int nNumBits) {
    unsigned *pBitArray = &aBitArray[nStartBit>>BIT_OFFSET_TO_INT_OFFSET_SHIFT];
    unsigned nEndBit    = (unsigned)(nStartBit + nNumBits - 1);
    unsigned *pEnd      = &aBitArray[nEndBit>>BIT_OFFSET_TO_INT_OFFSET_SHIFT];
    unsigned nLeftShift = nStartBit&BIT_OFFSET_MASK;
    unsigned nRightShift= (nEndBit&BIT_OFFSET_MASK)^BIT_OFFSET_MASK;
    unsigned nTest;
    if (pBitArray == pEnd) {
        nTest = *pBitArray & (((unsigned)-1)<<nLeftShift) & (((unsigned)-1)>>nRightShift);
        *pBitArray |= (((unsigned)-1)<<nLeftShift) & (((unsigned)-1)>>nRightShift);
    } else {
        nTest = *pBitArray  & (((u32)-1)<<nLeftShift);
        *pBitArray |= ((u32)-1)<<nLeftShift;
        nNumBits -= nLeftShift;
        while(++pBitArray < pEnd) {
            nTest |= *pBitArray;
            *pBitArray = ((u32)-1);
        }
        nTest |= *pBitArray & (((u32)-1)>>nRightShift);
        *pBitArray |= ((u32)-1)>>nRightShift;
    }
    return nTest;
}

void WfsBitFieldCopy(unsigned *aDstBitArray, int nDstOffset, unsigned *aSrcBitArray, int nSrcOffset, int nCopyLength) {
    unsigned *pDstStart    = &aDstBitArray[ nDstOffset>>BIT_OFFSET_TO_INT_OFFSET_SHIFT];
    unsigned *pSrcStart    = &aSrcBitArray[ nSrcOffset>>BIT_OFFSET_TO_INT_OFFSET_SHIFT];
    unsigned nDstEndBit    = (u32)(nDstOffset + nCopyLength - 1);
    unsigned *pDstEnd      = &aDstBitArray[nDstEndBit>>BIT_OFFSET_TO_INT_OFFSET_SHIFT];
    unsigned nOldDstStart  = *pDstStart;
    unsigned nOldDstEnd    = *pDstEnd;
    int      nShift        = (nDstOffset&BIT_OFFSET_MASK) - (nSrcOffset&BIT_OFFSET_MASK);
    unsigned nShiftRight   = (u32)(nShift & BIT_OFFSET_MASK);
    unsigned nShiftLeft    = BITS_PER_INT - nShiftRight;
    if (nShift<0) {
        if (pDstStart <= pSrcStart) {
            unsigned *pDst = pDstStart;
            unsigned *pSrc = pSrcStart;
            while(pDst <= pDstEnd) {
                pDst[0] = (pSrc[0]<<nShiftLeft) | (pSrc[1]>>nShiftRight);
                pDst++; pSrc++;
            }
        } else {
            unsigned *pDst = pDstEnd;
            unsigned *pSrc = pSrcStart+(pDstEnd-pDstStart);
            while(pDst >= pDstStart) {
                pDst[0] = (pSrc[0]<<nShiftLeft) | (pSrc[1]>>nShiftRight);
                pDst--; pSrc--;
            }
        }
    } else if (nShift>0) {
        if (pDstStart < pSrcStart) {
            unsigned *pDst = pDstStart;
            unsigned *pSrc = pSrcStart;
            while(pDst <= pDstEnd) {
                pDst[0] = (pSrc[-1]<<nShiftLeft) | (pSrc[0]>>nShiftRight);
                pDst++; pSrc++;
            }
        } else {
            unsigned *pDst = pDstEnd;
            unsigned *pSrc = pSrcStart+(pDstEnd-pDstStart);
            while(pDst >= pDstStart) {
                pDst[0] = (pSrc[-1]<<nShiftLeft) | (pSrc[0]>>nShiftRight);
                pDst--; pSrc--;
            }
        }
    } else {
        if (pDstStart < pSrcStart) {
            unsigned *pDst = pDstStart;
            unsigned *pSrc = pSrcStart;
            while(pDst <= pDstEnd) {
                *pDst++ = *pSrc++;
            }
        } else {
            unsigned *pDst = pDstEnd;
            unsigned *pSrc = pSrcStart+(pDstEnd-pDstStart);
            while(pDst >= pDstStart) {
                *pDst-- = *pSrc--;
            }
        }
    }
    unsigned nDstStartMask =  (((unsigned)-1)    >>(nDstOffset&BIT_OFFSET_MASK));
    unsigned nDstEndMask   = ((((unsigned)-1)>>1)>>(nDstEndBit&BIT_OFFSET_MASK));
    *pDstStart = (*pDstStart &  nDstStartMask) | (nOldDstStart & ~nDstStartMask);
    *pDstEnd   = (*pDstEnd   & ~nDstEndMask  ) | (nOldDstEnd   &  nDstEndMask);
}


#define BitMask(b) (1<<(BIT_OFFSET_MASK-((b)&BIT_OFFSET_MASK)))
#include "randomlib.h"

void WfsTestBitFieldModule() {
#if _BITFIELD_NUM_TESTS
    int nDstOffset, nSrcOffset, nCopyLength;
    unsigned aOldBitArray[8];
    unsigned aNewBitArray[8];

    s32 nRandomSeed;
#ifdef WIN32
    LARGE_INTEGER sysClock;
    QueryPerformanceCounter(&sysClock);
    nRandomSeed = ((u32*)&sysClock)[0] ^ ((u32*)&sysClock)[1];
#else
    nRandomSeed = 0x1;
#endif
    RandomSetSeedU32((u32)nRandomSeed);
    osTPrintf("nRandomSeed=0x%08x\n", nRandomSeed);

    osTPrintf("WfsBitArrayInsertBitField() Test\n");
    int nTest;
    for(nTest=1; nTest<=_BITFIELD_NUM_TESTS; nTest++) {
        //for(unsigned nI=0; nI<8; nI++) { aOldBitArray[nI] = 0xFFFFFFFF; } //RandomU32(); }
        unsigned nI;
        for(nI=0; nI<8; nI++) { aOldBitArray[nI] = RandomU32(); }
        int nNumBits = RandomInt(1, 32);
        int nBitArrayLength = RandomInt(0, 256-nNumBits);
        int nInsertBit = RandomInt(0, nBitArrayLength);
        unsigned nValue = RandomU32();
        nValue >>= (32-nNumBits);
        aOldBitArray[(nBitArrayLength-1)>>BIT_OFFSET_TO_INT_OFFSET_SHIFT] |= BitMask(nBitArrayLength);
        int nBit;
        for(nBit=nBitArrayLength; nBit<256; nBit++) {
            aOldBitArray[nBit>>BIT_OFFSET_TO_INT_OFFSET_SHIFT] &= ~BitMask(nBit);
        }
        for(nI=0; nI<8; nI++) { aNewBitArray[nI] = aOldBitArray[nI]; } //RandomU32(); }
        WfsBitArrayInsertBitField(aNewBitArray, nBitArrayLength, nInsertBit, nNumBits, nValue);
        for(nBit=0; nBit<nInsertBit; nBit++) {
            if ((aOldBitArray[nBit>>BIT_OFFSET_TO_INT_OFFSET_SHIFT]&BitMask(nBit)) !=
                (aNewBitArray[nBit>>BIT_OFFSET_TO_INT_OFFSET_SHIFT]&BitMask(nBit))) {
                osTPrintf("ERROR! Test:%d Mismatch before insertion point at bit %02X\n", nTest, nBit);
                goto Error;
            }
        }
        for(; nBit<nInsertBit+nNumBits; nBit++) {
            if (((aNewBitArray[nBit>>BIT_OFFSET_TO_INT_OFFSET_SHIFT]&BitMask(nBit))==0) ^
                ((nValue&(1<<(nNumBits-1-(nBit-nInsertBit))))==0)) {
                    osTPrintf("ERROR! Test:%d Mismatch in inserted Value at bit %02X:%02x\n", nTest, nBit, nNumBits-1-(nBit-nInsertBit));
                goto Error;
            }
        }
        for(; nBit<256; nBit++) {
            if (((aOldBitArray[(nBit-nNumBits)>>BIT_OFFSET_TO_INT_OFFSET_SHIFT]&BitMask(nBit-nNumBits))==0) ^
                ((aNewBitArray[nBit>>BIT_OFFSET_TO_INT_OFFSET_SHIFT]&BitMask(nBit))==0)) {
                osTPrintf("ERROR! Test:%d Mismatch after insertion point at bit %02x:%02x\n(aOldBitArray[%d] & %08x = %08x)\n(aNewBitArray[%d] & %08x = %08x)\n",
                    nTest, nBit, nBit-nNumBits,
                    (nBit-nNumBits)>>BIT_OFFSET_TO_INT_OFFSET_SHIFT, BitMask(nBit-nNumBits), 
                    (aOldBitArray[(nBit-nNumBits)>>BIT_OFFSET_TO_INT_OFFSET_SHIFT]&BitMask(nBit-nNumBits)),
                    nBit>>BIT_OFFSET_TO_INT_OFFSET_SHIFT, BitMask(nBit),
                    (aNewBitArray[nBit>>BIT_OFFSET_TO_INT_OFFSET_SHIFT]&BitMask(nBit)));
                goto Error;
            }
        }
    }

    osTPrintf("WfsDeleteBitField() Test\n");
    for(nTest=1; nTest<=_BITFIELD_NUM_TESTS; nTest++) {
        unsigned nI;
        for(nI=0; nI<8; nI++) { aOldBitArray[nI] = RandomU32(); }
        int nBitArrayLength = RandomInt(0, 256);
        int nDeleteStartBit = RandomInt(0, nBitArrayLength-1);
        int nDeleteNumBits = RandomInt(1, 96);
        if (nDeleteNumBits >(nBitArrayLength-nDeleteStartBit)) {
            nDeleteNumBits = nBitArrayLength-nDeleteStartBit;
        }
        int nBit;
        for(nBit=nBitArrayLength; nBit<256; nBit++) {
            aOldBitArray[nBit>>BIT_OFFSET_TO_INT_OFFSET_SHIFT] &= ~BitMask(nBit);
        }
        for(nI=0; nI<8; nI++) { aNewBitArray[nI] = aOldBitArray[nI]; }
        WfsBitArrayDeleteBitField(aNewBitArray, nBitArrayLength, nDeleteStartBit, nDeleteNumBits);
        for(nBit=0; nBit<nDeleteStartBit; nBit++) {
            if ((aOldBitArray[nBit>>BIT_OFFSET_TO_INT_OFFSET_SHIFT]&BitMask(nBit)) !=
                (aNewBitArray[nBit>>BIT_OFFSET_TO_INT_OFFSET_SHIFT]&BitMask(nBit))) {
                osTPrintf("ERROR! Test:%d Mismatch at bit %02X\n", nTest, nBit);
                goto Error;
            }
        }
        nBit += nDeleteNumBits;
        for(; nBit<nBitArrayLength; nBit++) {
            if (((aNewBitArray[(nBit-nDeleteNumBits)>>BIT_OFFSET_TO_INT_OFFSET_SHIFT]&BitMask(nBit-nDeleteNumBits))==0) ^
                ((aOldBitArray[nBit>>BIT_OFFSET_TO_INT_OFFSET_SHIFT]&BitMask(nBit))==0)) {
                    osTPrintf("ERROR! Test:%d Mismatch between bits %02x:%02x\n(aOldBitArray[%d] & %08x = %08x)\n(aNewBitArray[%d] & %08x = %08x)\n",
                    nTest, nBit, nBit-nDeleteNumBits,
                    (nBit-nDeleteNumBits)>>BIT_OFFSET_TO_INT_OFFSET_SHIFT, BitMask(nBit-nDeleteNumBits), 
                    (aOldBitArray[(nBit-nDeleteNumBits)>>BIT_OFFSET_TO_INT_OFFSET_SHIFT]&BitMask(nBit-nDeleteNumBits)),
                    nBit>>BIT_OFFSET_TO_INT_OFFSET_SHIFT, BitMask(nBit),
                    (aNewBitArray[nBit>>BIT_OFFSET_TO_INT_OFFSET_SHIFT]&BitMask(nBit)));
                goto Error;
            }
        }
    }

    osTPrintf("WfsBitFieldCopy() Test\n");
    for(nTest=1; nTest<=_BITFIELD_NUM_TESTS; nTest++) {
        unsigned nI;
        for(nI=0; nI<8; nI++) { aOldBitArray[nI] = RandomU32(); }
        nDstOffset = RandomInt(0, 255);
        nSrcOffset = RandomInt(0, 255);
        int nMaxLength = 256 - max(nSrcOffset, nDstOffset);
        nCopyLength = RandomInt(1, nMaxLength);
        int nBit;
        /*for(nBit=nBitArrayLength; nBit<256; nBit++) {
            aOldBitArray[nBit>>BIT_OFFSET_TO_INT_OFFSET_SHIFT] &= ~BitMask(nBit);
        }*/
        for(nI=0; nI<8; nI++) { aNewBitArray[nI] = aOldBitArray[nI]; }
        WfsBitFieldCopy(aNewBitArray, nDstOffset, aNewBitArray, nSrcOffset, nCopyLength);
        for(nBit=0; nBit<nDstOffset; nBit++) {
            if ((aOldBitArray[nBit>>BIT_OFFSET_TO_INT_OFFSET_SHIFT]&BitMask(nBit)) !=
                (aNewBitArray[nBit>>BIT_OFFSET_TO_INT_OFFSET_SHIFT]&BitMask(nBit))) {
                osTPrintf("ERROR! Test:%d Mismatch at bit %02X\n", nTest, nBit);
                goto WfsBitFieldCopyError;
            }
        }
        for(; nBit<nDstOffset+nCopyLength; nBit++) {
            int nSrcBit = nBit-nDstOffset+nSrcOffset;
            if (((aOldBitArray[nSrcBit>>BIT_OFFSET_TO_INT_OFFSET_SHIFT]&BitMask(nSrcBit))==0) ^
                ((aNewBitArray[nBit>>BIT_OFFSET_TO_INT_OFFSET_SHIFT]&BitMask(nBit))==0)) {
                osTPrintf("ERROR! Test:%d Mismatch at bit %02X\n", nTest, nBit);
                goto WfsBitFieldCopyError;
            }
        }
        for(; nBit<256; nBit++) {
            if ((aOldBitArray[nBit>>BIT_OFFSET_TO_INT_OFFSET_SHIFT]&BitMask(nBit)) !=
                (aNewBitArray[nBit>>BIT_OFFSET_TO_INT_OFFSET_SHIFT]&BitMask(nBit))) {
                osTPrintf("ERROR! Test:%d Mismatch at bit %02X\n", nTest, nBit);
                goto WfsBitFieldCopyError;
            }
        }
    }
    return;
WfsBitFieldCopyError:;
    osTPrintf("WfsBitFieldCopy(New, nDstOffset=%02x, Src, nSrcOffset=%02x, nCopyLength=%02x)\n", nDstOffset, nSrcOffset, nCopyLength);
    goto Error;
Error:;
    unsigned nI;
    osTPrintf("Old:    "); for(nI=0; nI<8; nI++) { osTPrintf("%08x", aOldBitArray[nI]); } osTPrintf("\n");
    osTPrintf("New:    "); for(nI=0; nI<8; nI++) { osTPrintf("%08x", aNewBitArray[nI]); } osTPrintf("\n");
    osTPrintf("        0   1   2   3   4   5   6   7   8   9   a   b   c   d   e   f   \n");
    osTPrintf("        048c048c048c048c048c048c048c048c048c048c048c048c048c048c048c048c\n");
#endif
}

// nStartBit = 9
// nBitArrayLength = 45
// nNumGapBits = 18
// nBitArrayLength - nNumGapBits = 36
// pStart    = &aBitArray[0]
// pCopyTo   = &aBitArray[1]
// pCopyFrom = &aBitArray[1]
