/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfskrn_DirCheck.cpp


  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_DirCheck.cpp,v $
  Revision 1.14  2008/11/18 04:57:12  saito_tomoya
  Modified SetBit() and added SetBitArray().

  Revision 1.13  2008/10/30 04:50:45  kondo_masahiro
  Added DirCheckDisk

  Revision 1.12  2008/10/17 08:50:48  kondo_masahiro
  Added the new MEDIUM size category (40KB-320KB)

  Revision 1.11  2008/10/16 09:30:39  ooizumi
  Fixed definitions for debug build.

  Revision 1.10  2008/09/01 01:57:26  kondo_masahiro
  Minor fixed for IOP compiler.

  Revision 1.9  2008/08/27 23:02:55  ooizumi
  Fixed newline to CR+LF(Windows format).

  Revision 1.8  2008/08/27 09:47:43  paul
  Removed nTotalFree from block header. Added pAttrHdr->nAttrLog2Size

  Revision 1.7  2008/07/09 00:31:36  paul
  Minor changes

  Revision 1.6  2008/06/09 17:22:38  paul
  Changed comment

  Revision 1.5  2008/05/27 18:44:19  paul
  Made utf8 variable definitions more consistent, and use (unsigned) for comparisons

  Revision 1.4  2008/05/14 02:16:52  paul
  Fixed the node choice ordering check to prevent utf8 being upgraded to signed int

  Revision 1.3  2008/05/10 03:59:44  kondo_masahiro
  Merged and fixed code to accept RM compiler

  Revision 1.2  2008/04/25 17:29:46  kondo_masahiro
  Fixed code to transfer the wfskrn project to RM_SDK tree

  Revision 1.1  2008/04/22 04:39:08  kondo_masahiro
  Fixed code to transfer the wfskrn project to RM_SDK tree

  Revision 1.5  2008/04/22 23:26:55  paul
  Changed areaHdrRoot to areaInfoRoot

  Revision 1.4  2008/04/19 05:50:57  kondo_masahiro
  Fixed code to accept RM compiler

  Revision 1.3  2008/04/15 18:51:26  paul
  Made some changes to debug code

  Revision 1.2  2008/04/08 00:25:34  paul
  Added an extra check for radix tree nodes with 0 entries

  Revision 1.1  2008/04/05 00:03:30  paul
  Integrity checks for directory node blocks and directory leaf blocks

 *---------------------------------------------------------------------------*/

#undef  _KRN_FILE_
#define _KRN_FILE_  "DirCheck"

#include "wfskrn_Dir.h"
#include "wfskrn_SubBlkAlloc.h"
#include "wfskrn_DirNodeStack.h"
#include "wfskrn_Area.h"
#include "wfskrn_BitField.h"
#include "wfskrn_Heap.h"
#include "wfskrn_Api.h"
#include "wfskrn_Mutex.h"

#if _DEBUG_DIR
bool bCheckDirBlks = true; // extern
#endif

static
void DirSubBlkDebugPrint(DirSubBlkHdr *pSubBlkHdr) {
    utf8 *pChoiceChar = &pSubBlkHdr->aFirstChar[pSubBlkHdr->nStrLen];
    utf8 cTemp = *pChoiceChar;
    *pChoiceChar = 0;
    MyOSReport("%d,%d,\"%s\"", pSubBlkHdr->nStrLen, pSubBlkHdr->nNumEntries, pSubBlkHdr->aFirstChar);
    *pChoiceChar = cTemp;
    u32 nI;
    for(nI=0; nI<pSubBlkHdr->nNumEntries; nI++) {
        if ((*pChoiceChar<32)||(*pChoiceChar>127)) {
            MyOSReport(",%d", *pChoiceChar);
        } else {
            MyOSReport(",'%c'", *pChoiceChar);
        }
        ++pChoiceChar;
    }
    MyOSReport("\n");
}

void DirLeafCheckBlk(DirItr *pDi) {
#if _DEBUG_DIR
    // Checks that all allocated sub blocks of pDi->pBlkHdr are reachable, and all non-allocated blocks are on the free list
    if (!bCheckDirBlks) {
        return;
    }
#endif
#if _DEBUG_DIR_NODE_STACK
    bool bOldCheckDirNodeStack = bCheckDirNodeStack;
    bCheckDirNodeStack = false;
#endif
    DirItr di = *pDi;

    di.dni.nDnsDepth = 0;
    WFSKrnResult nResult = DirNodeStackGetEntry(&di.dni);
    di.ppDnsRootParent = &di.dni.pDns->pParent;

    DirBlkHdr *pBlkHdr = (DirBlkHdr *)di.pBlkHdr;
    u32 nSrcNodeOfs = pBlkHdr->nRootOfs;
    utf8 *pChoiceChar;
    u32 nConnectedSize=(1<<WFS_NUM_BITS_REQUIRED(sizeof(DirBlkHdr)));
    u32 aUtilization[1<<(WFS_LOG2_MEDIUM_BLK_SIZE - SBA_MIN_LOG2_SUB_BLK_SIZE - 5)];
    memset(aUtilization, 0, sizeof(aUtilization));
    WfsSetBitField((unsigned*)aUtilization, 0, 1<<(WFS_NUM_BITS_REQUIRED(sizeof(DirBlkHdr))-SBA_MIN_LOG2_SUB_BLK_SIZE));
    u32 nRecIdx = 0;
    u32 nSrcLeafStartDepth = di.dni.nDnsDepth;
    while(1) {
        di.dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + nSrcNodeOfs);
        WFSSubBlkFreeHdr *pSbfHdr = (WFSSubBlkFreeHdr *)di.dni.pSubBlkHdr;
        if (pSbfHdr->nMagic == SBA_FREE_SUB_BLK_MAGIC) {
            WFSKrnOutputErrorStr("Pointer to deleted sub block detected!");
        }
        di.dni.nLog2Size = DirLeafGetSubBlkLog2Size(di.dni.pSubBlkHdr);
        di.dni.pDns->nBlkAdr = di.tba.nBlkAdr;
        di.dni.pDns->nSubBlkOfs = nSrcNodeOfs;
#if _DEBUG_DIR
        if ((nDirTest>=nDirTestBreakPoint-1)&&(nDirTest<=nDirTestBreakPoint)) {
            u32 nI;
            for(nI=nSrcLeafStartDepth; nI<di.dni.nDnsDepth; nI++) {
                MyOSReport("    ");
            }
            MyOSReport("%03x: ", nSrcNodeOfs);
            DirSubBlkDebugPrint(di.dni.pSubBlkHdr);
        }
#endif
        if (WfsTestAndSetBitField((unsigned*)aUtilization, nSrcNodeOfs>>SBA_MIN_LOG2_SUB_BLK_SIZE, 1<<(di.dni.nLog2Size-SBA_MIN_LOG2_SUB_BLK_SIZE))) {
            WFSKrnOutputErrorStr("Overlap!");
        }
        nConnectedSize += (u32)(1<<di.dni.nLog2Size);
        u16 *pSrcNodeOfsTblEnd = (u16*)((u8*)di.dni.pSubBlkHdr + (1<<di.dni.nLog2Size));
        di.dni.pDns->nEntryIdx = 1;
        if (di.dni.pSubBlkHdr->nNumEntries == 0) {
            if (pBlkHdr->nNumRecs) {
                WFSKrnOutputErrorStr("No entries!");
            }
        }
        pChoiceChar = &di.dni.pSubBlkHdr->aFirstChar[di.dni.pSubBlkHdr->nStrLen];
        if (*pChoiceChar == 0) {
            // String terminator char indicates this is a record, so increment the record count
            u32 nAttrOfs = pSrcNodeOfsTblEnd[-1];
            SbaCheckOfs(nAttrOfs);
            DirEntryAttrHdr *pAttrHdr = (DirEntryAttrHdr *)((u8*)pBlkHdr + nAttrOfs);
            WFSSubBlkFreeHdr *pSbfHdr = (WFSSubBlkFreeHdr *)pAttrHdr;
            if (pSbfHdr->nMagic == SBA_FREE_SUB_BLK_MAGIC) {
                WFSKrnOutputErrorStr("Pointer to deleted attr block detected!");
            }
            u8 nSizeCategory;
            u16 nAttrLog2Size = (u16)DirCalculateAttrSubBlkLog2Size(di.tba.pAreaInfo, pAttrHdr->nAllocSize, pAttrHdr->nNameLen, &nSizeCategory);
            if (nSizeCategory != pAttrHdr->nSizeCategory) {
                nAttrLog2Size = (u16)DirCalculateAttrSubBlkLog2SizeDecrease(di.tba.pAreaInfo, pAttrHdr->nAllocSize, pAttrHdr->nNameLen, &nSizeCategory);
                if (nSizeCategory != pAttrHdr->nSizeCategory) {
                    WFSKrnOutputErrorStr("Size category incorrect");
                }
            }
            if (WfsTestAndSetBitField((unsigned*)aUtilization, nAttrOfs>>SBA_MIN_LOG2_SUB_BLK_SIZE, 1<<(nAttrLog2Size-SBA_MIN_LOG2_SUB_BLK_SIZE))) {
                WFSKrnOutputErrorStr("Overlap!");
            }
            if (nAttrLog2Size != pAttrHdr->nAttrLog2Size) {
                WFSKrnOutputErrorStr("Attr log 2 size incorrect");
            }
            if (nSizeCategory != pAttrHdr->nSizeCategory) {
                WFSKrnOutputErrorStr("Size category incorrect");
            }
            nConnectedSize += (u32)(1<<nAttrLog2Size);
            ++nRecIdx;
            ++di.dni.pDns->nEntryIdx;
        } else if (di.dni.pSubBlkHdr->nNumEntries == 1) {
            WFSKrnOutputErrorStr("Non-terminating single-choice node!");
        }
        if (di.dni.pDns->nEntryIdx > di.dni.pSubBlkHdr->nNumEntries) {
            do {
                if (di.dni.nDnsDepth == nSrcLeafStartDepth) {
                    // We reached the top of the stack, so this should be the last record
                    goto CheckFreeSpace;
                }
                DirPopNodeStackEntry(&di.dni);
                di.dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + di.dni.pDns->nSubBlkOfs);
                ++di.dni.pDns->nEntryIdx;
            } while(di.dni.pDns->nEntryIdx > di.dni.pSubBlkHdr->nNumEntries);
            pChoiceChar = &di.dni.pSubBlkHdr->aFirstChar[di.dni.pSubBlkHdr->nStrLen + di.dni.pDns->nEntryIdx-1];
            if ((unsigned)pChoiceChar[-1] >= (unsigned)*pChoiceChar) {
                WFSKrnOutputErrorStr("Choice out of order!");
            }
        }

        nSrcNodeOfs = DirLeafGetIntraBlkOfs(di.dni.pDns->nEntryIdx, di.dni.pSubBlkHdr);
        SbaCheckOfs(nSrcNodeOfs);
        nResult = DirNodeStackGetEntry(&di.dni);
        if (nResult != WFSKRN_RESULT_OK) {
            WFSKrnOutputErrorCode(nResult);
            break;
        }
    }
CheckFreeSpace:;
    // Next check the free space
    u32 nTotalFree = SbaCheckFreeSpace((WFSSubBlkAllocHdr *)pBlkHdr, aUtilization);
    if ((nTotalFree + nConnectedSize) != pDi->tba.pAreaInfo->nBlkSize) {
        WFSKrnOutputErrorStr("Memory Leak detected within block!");
    }
    if (nRecIdx != pBlkHdr->nNumRecs) {
        WFSKrnOutputErrorStr("NumRecs Error!");
    }
    DirNodeStackFree(&di.dni, di.ppDnsRootParent);
#if _DEBUG_DIR
    if ((nDirTest>=nDirTestBreakPoint-1)&&(nDirTest<=nDirTestBreakPoint)) {
        MyOSReport("\n");
    }
#endif
#if _DEBUG_DIR_NODE_STACK
    bCheckDirNodeStack = bOldCheckDirNodeStack;
#endif
}

void DirNodeCheckBlk(DirItr *pDi) {
#if _DEBUG_DIR
    // Checks that all allocated sub blocks of pDi->pBlkHdr are reachable, and all non-allocated blocks are on the free list
    if (!bCheckDirBlks) {
        return;
    }
#endif
#if _DEBUG_DIR_NODE_STACK
    bool bOldCheckDirNodeStack = bCheckDirNodeStack;
    bCheckDirNodeStack = false;
#endif

    DirItr di = *pDi;
    DirBlkHdr *pBlkHdr = (DirBlkHdr *)di.pBlkHdr;
    u32 nSrcNodeOfs = pBlkHdr->nRootOfs;
    utf8 *pChoiceChar;
    u32 nConnectedSize=(1<<WFS_NUM_BITS_REQUIRED(sizeof(DirBlkHdr)));
    u32 aUtilization[1<<(WFS_LOG2_MEDIUM_BLK_SIZE - SBA_MIN_LOG2_SUB_BLK_SIZE - 5)];
    memset(aUtilization, 0, sizeof(aUtilization));
    WfsSetBitField((unsigned*)aUtilization, 0, 1<<(WFS_NUM_BITS_REQUIRED(sizeof(DirBlkHdr))-SBA_MIN_LOG2_SUB_BLK_SIZE));
    u32 nRecIdx = 0;
    u32 nSrcNodeStartDepth = di.dni.nDnsDepth;
    WFSKrnResult nResult;
    while(1) {
        di.dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + nSrcNodeOfs);
        WFSSubBlkFreeHdr *pSbfHdr = (WFSSubBlkFreeHdr *)di.dni.pSubBlkHdr;
        if (pSbfHdr->nMagic == SBA_FREE_SUB_BLK_MAGIC) {
            WFSKrnOutputErrorStr("Pointer to deleted sub block detected!");
        }
        di.dni.nLog2Size = DirNodeGetSubBlkLog2Size(di.dni.pSubBlkHdr);
        di.dni.pDns->nBlkAdr = di.tba.nBlkAdr;
        di.dni.pDns->nSubBlkOfs = nSrcNodeOfs;
#if _DEBUG_DIR
        if ((nDirTest>=nDirTestBreakPoint-1)&&(nDirTest<=nDirTestBreakPoint)) {
            u32 nI;
            for(nI=nSrcNodeStartDepth; nI<di.dni.nDnsDepth; nI++) {
                MyOSReport("    ");
            }
            MyOSReport("%03x: ", nSrcNodeOfs);
            DirSubBlkDebugPrint(di.dni.pSubBlkHdr);
        }
#endif
        if (WfsTestAndSetBitField((unsigned*)aUtilization, nSrcNodeOfs>>SBA_MIN_LOG2_SUB_BLK_SIZE, 1<<(di.dni.nLog2Size-SBA_MIN_LOG2_SUB_BLK_SIZE))) {
            WFSKrnOutputErrorStr("Overlap!");
        }
        nConnectedSize += (u32)(1<<di.dni.nLog2Size);
        di.dni.pDns->nEntryIdx = 1;
        if (di.dni.pSubBlkHdr->nNumEntries == 0) {
            if (pBlkHdr->nNumRecs) {
                WFSKrnOutputErrorStr("No entries!");
            }
        }
        pChoiceChar = &di.dni.pSubBlkHdr->aFirstChar[di.dni.pSubBlkHdr->nStrLen];
        if (*pChoiceChar == 0) {
            // String terminator char indicates this is a record, so increment the record count
            ++nRecIdx;
            ++di.dni.pDns->nEntryIdx;
        } else if (di.dni.pSubBlkHdr->nNumEntries == 1) {
            WFSKrnOutputErrorStr("Non-terminating single-choice node!");
        }
        if (di.dni.pDns->nEntryIdx > di.dni.pSubBlkHdr->nNumEntries) {
            do {
                if (di.dni.nDnsDepth == nSrcNodeStartDepth) {
                    // We reached the top of the stack, so this should be the last record
                    goto CheckFreeSpace;
                }
                DirPopNodeStackEntry(&di.dni);
                di.dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + di.dni.pDns->nSubBlkOfs);
                ++di.dni.pDns->nEntryIdx;
            } while(di.dni.pDns->nEntryIdx > di.dni.pSubBlkHdr->nNumEntries);
            pChoiceChar = &di.dni.pSubBlkHdr->aFirstChar[di.dni.pSubBlkHdr->nStrLen + di.dni.pDns->nEntryIdx -1];
            if ((unsigned)pChoiceChar[-1] >= (unsigned)*pChoiceChar) {
                WFSKrnOutputErrorStr("Choice out of order!");
            }
        }
        nSrcNodeOfs = DirNodeGetIntraBlkOfs(di.dni.pDns->nEntryIdx, di.dni.pSubBlkHdr);
        SbaCheckOfs(nSrcNodeOfs);
        nResult = DirNodeStackGetEntry(&di.dni);
        if (nResult != WFSKRN_RESULT_OK) {
            WFSKrnOutputErrorCode(nResult);
            break;
        }
    }
CheckFreeSpace:;
    // Next check the free space
    u32 nTotalFree = SbaCheckFreeSpace((WFSSubBlkAllocHdr *)pBlkHdr, aUtilization);
    if ((nTotalFree + nConnectedSize) != pDi->tba.pAreaInfo->nBlkSize) {
        WFSKrnOutputErrorStr("Memory Leak detected within block!");
    }
    if (nRecIdx != pBlkHdr->nNumRecs) {
        WFSKrnOutputErrorStr("NumRecs Error!");
    }
#if _DEBUG_DIR
    if ((nDirTest>=nDirTestBreakPoint-1)&&(nDirTest<=nDirTestBreakPoint)) {
        MyOSReport("\n");
    }
#endif
#if _DEBUG_DIR_NODE_STACK
    bCheckDirNodeStack = bOldCheckDirNodeStack;
#endif
}

void DirCheckBlk(DirItr *pDi) {
#if _DEBUG_DIR
    // Checks that all allocated blocks are reachable, and all non-allocated blocks are on the free list
    if (!bCheckDirBlks) {
        return;
    }
#endif
    if (pDi->pBlkHdr->sbah.mdh.nFlags & WFS_MDF_DIR_LEAF_BLK) {
        DirLeafCheckBlk(pDi);
    } else {
        DirNodeCheckBlk(pDi);
    }
}


static void SetBit(WFSSrvFileHandle fh, u32 diskAddr)
{
    u32 byte = diskAddr / 8;
    u32 bit = diskAddr % 8;
    u8 flag = (1 << bit);

    u8 dst;
    
    WFSResult nResult;
    nResult = (WFSResult)WFSSrvReadFile(fh, &dst, byte, 1);
    if(nResult < WFS_RESULT_OK){
        osTPrintf("SetBit() error %d\n", nResult);
    }

    ASSERT(0 == (dst & flag));
    dst |= flag;
    nResult = WFSSrvWriteFile(fh, &dst, byte, 1, true);

}

static void SetBitArrayInternal(WFSSrvFileHandle fh, u8* pBuffer, u32 bufferSize, u32 diskAddr, u32 length){

    u32 byte = diskAddr / 8;
    u32 bit  = diskAddr % 8;
    u32 byteLength = (diskAddr + length - 1) / 8  - byte + 1;

    WFSResult nResult;

    // read
    nResult = (WFSResult)WFSSrvReadFile(fh, pBuffer, byte, byteLength);
    if(nResult < WFS_RESULT_OK){
        osTPrintf("SetBitArray() error %d\n", nResult);
    }

    // set bits
    int i;
    for(i=bit; i<bit+length; ++i){
        u8 flag = 1 << (i & 0x07);
        ASSERT( !(pBuffer[i/8] & flag) );
        pBuffer[i/8] |= flag;
    }

    // write
    nResult = WFSSrvWriteFile(fh, pBuffer, byte, byteLength, true);

}

static void SetBitArray(WFSSrvFileHandle fh, u8* pBuffer, u32 bufferSize, u32 diskAddr, u32 length){

    while(length > (bufferSize-1)*8){
        SetBitArrayInternal(fh, pBuffer, bufferSize, diskAddr, (bufferSize-1)*8);
        length   -= (bufferSize-1)*8;
        diskAddr += (bufferSize-1)*8;
    }
    SetBitArrayInternal(fh, pBuffer, bufferSize, diskAddr, length);
}


WFSKrnResult DirCheckDisk(WFSBlkAdr rootBlkAdr, DirItr *pDi, WFSSrvFileHandle fh, u8* pBuffer, u32 bufferSize, u32 nFlag) {

    DirBlkHdr *pBlkHdr;
    pDi->tba.nBlkAdr = rootBlkAdr;

    ///
    //pCheckDiskArray[rootBlkAdr/8] |= 1<<(rootBlkAdr%8);
    SetBit(fh, rootBlkAdr);

    pDi->nBlkDepth = 0;
    WFSKrnReturnOnError(DirGetBlk(pDi, TRANS_FLAG_READ_BLK | BCACHE_FLAG_PINNED, &pBlkHdr));

    pDi->dni.nDnsDepth = 0;
    WFSKrnReturnOnError(DirNodeStackGetEntry(&pDi->dni));
    pDi->ppDnsRootParent = &pDi->dni.pDns->pParent;

    pDi->dni.pDns->nBlkAdr = 0; //tba.nBlkAdr;
    pDi->dni.pDns->nSubBlkOfs = pBlkHdr->nRootOfs;
    SbaCheckOfs(pDi->dni.pDns->nSubBlkOfs);
    pDi->dni.pDns->nEntryIdx = 1;

    pDi->dni.nEntryIdx = 1;
    pDi->dni.pOfs = &pBlkHdr->nRootOfs;
    SbaCheckOfs(*pDi->dni.pOfs);
    pDi->dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + *pDi->dni.pOfs);
    if (pBlkHdr->sbah.mdh.nFlags & WFS_MDF_DIR_LEAF_BLK) {
        pDi->dni.nLog2Size = DirLeafGetSubBlkLog2Size(pDi->dni.pSubBlkHdr);
    } else {
        pDi->dni.nLog2Size = DirNodeGetSubBlkLog2Size(pDi->dni.pSubBlkHdr);
    }

    while(1){
#if _DEBUG_BREAK_POINT
        static int c=0; c++; if(c==0x2b){
            int a; a=0;
        } 
#endif
        while (pDi->dni.nEntryIdx > pDi->dni.pSubBlkHdr->nNumEntries) {
            if(pDi->dni.pDns->nBlkAdr == 0){
                TransUnpinBlk(&pDi->tba);
                DirPopNodeStackEntry(&pDi->dni);
                pDi->nBlkDepth = 0;
                return WFSKRN_RESULT_OK;
            }
            if(pDi->tba.nBlkAdr != pDi->dni.pDns->nBlkAdr){
                TransUnpinBlk(&pDi->tba);
                pDi->tba.nBlkAdr = pDi->dni.pDns->nBlkAdr;
                WFSKrnReturnOnError( DirGetBlk(pDi, TRANS_FLAG_READ_BLK | BCACHE_FLAG_PINNED, &pBlkHdr) );
            }
            pDi->dni.nEntryIdx  = ++pDi->dni.pDns->nEntryIdx;
            pDi->dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + pDi->dni.pDns->nSubBlkOfs);
            if (pDi->dni.nEntryIdx <= pDi->dni.pSubBlkHdr->nNumEntries){
                if (pBlkHdr->sbah.mdh.nFlags & WFS_MDF_DIR_LEAF_BLK) {
                    pDi->dni.nLog2Size = DirLeafGetSubBlkLog2Size(pDi->dni.pSubBlkHdr);
                    pDi->dni.pOfs = DirLeafGetIntraBlkOfsPtr_l2s((s32)pDi->dni.nEntryIdx, pDi->dni.pSubBlkHdr, pDi->dni.nLog2Size);
                    pDi->dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + *pDi->dni.pOfs);
                    pDi->dni.nLog2Size = DirLeafGetSubBlkLog2Size(pDi->dni.pSubBlkHdr);
                } else {
                    pDi->dni.nLog2Size = DirNodeGetSubBlkLog2Size(pDi->dni.pSubBlkHdr);
                    pDi->dni.pOfs = DirNodeGetIntraBlkOfsPtr_l2s((s32)pDi->dni.nEntryIdx, pDi->dni.pSubBlkHdr, pDi->dni.nLog2Size);
                    pDi->dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + *pDi->dni.pOfs);
                    pDi->dni.nLog2Size = DirNodeGetSubBlkLog2Size(pDi->dni.pSubBlkHdr);
                }
                pDi->dni.nEntryIdx = 1;
            } else {
                DirPopNodeStackEntry(&pDi->dni);
                --pDi->nBlkDepth;
            }
        }
        WFSKrnReturnOnError(DirPushNodeStackEntry(&pDi->dni, pDi->tba.nBlkAdr));
        ++pDi->nBlkDepth;
        pDi->dni.pDns->nUpdateCtr = pDi->nUpdateCtr;

        if (pBlkHdr->sbah.mdh.nFlags & WFS_MDF_DIR_LEAF_BLK) {
            if(pDi->dni.nEntryIdx == 1 && pDi->dni.pSubBlkHdr->aFirstChar[pDi->dni.pSubBlkHdr->nStrLen] == 0){
                u16 nOfs = *DirLeafGetIntraBlkOfsPtr_l2s(1, pDi->dni.pSubBlkHdr, pDi->dni.nLog2Size);
                pDi->pAttrHdr = (DirEntryAttrHdr *)((u8*)pBlkHdr + nOfs);
                if (pDi->pAttrHdr->nAccessListIdx & WFS_FLAG_IS_A_DIRECTORY) {
                    TransUnpinBlk(&pDi->tba);
                    pDi->tba.nBlkAdr = pDi->pAttrHdr->dir.nSubDirBlkAdr;

                    ///
                    //pCheckDiskArray[pDi->tba.nBlkAdr/8] |= 1<<(pDi->tba.nBlkAdr%8);
                    SetBit(fh, pDi->tba.nBlkAdr);

                    WFSKrnReturnOnError( DirGetBlk(pDi, TRANS_FLAG_READ_BLK | BCACHE_FLAG_PINNED, &pBlkHdr) );
                    pDi->dni.pOfs = &pBlkHdr->nRootOfs;
                    SbaCheckOfs(*pDi->dni.pOfs);
                    pDi->dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + *pDi->dni.pOfs);
                    if (pBlkHdr->sbah.mdh.nFlags & WFS_MDF_DIR_LEAF_BLK) {
                        pDi->dni.nLog2Size = DirLeafGetSubBlkLog2Size(pDi->dni.pSubBlkHdr);
                    } else {
                        pDi->dni.nLog2Size = DirNodeGetSubBlkLog2Size(pDi->dni.pSubBlkHdr);
                    }
                    pDi->dni.nEntryIdx = 1;
                } else {
                    if(nFlag == DIR_CHECK_DISK_FOR_DIR_TREE){
                        switch(pDi->pAttrHdr->nSizeCategory){
                        case FILE_SIZE_CATEGORY_SMALL:{
                            u32 nI, nNumSmallBlks = pDi->pAttrHdr->nAllocSize>>pDi->tba.pAreaInfo->ah.nLog2BlkSize;
                            for(nI=0;nI<nNumSmallBlks;nI++){
                                WFSFileBlkPtr *pFileBlkPtr = WFSKrnGetFileBlkPtrForCategorySmall(pDi->pAttrHdr, pDi->pAttrHdr->nAttrLog2Size, nI);
                                //if(pCheckDiskArray[pFileBlkPtr->nBlkAdr/8]&(1<<(pFileBlkPtr->nBlkAdr%8))){
                                //    int a; a=0;
                                //}

                                ///
                                //pCheckDiskArray[pFileBlkPtr->nBlkAdr/8] |= 1<<(pFileBlkPtr->nBlkAdr%8);
                                SetBit(fh, pFileBlkPtr->nBlkAdr);

                            }
                            break;
                        }
                        case FILE_SIZE_CATEGORY_MEDIUM:{
                            u32 nI, nNumMediumBlks = pDi->pAttrHdr->nAllocSize>>pDi->tba.pAreaInfo->ah.nLog2MediumBlkSize;
                            for(nI=0;nI<nNumMediumBlks;nI++){
                                WFSFileBlkPtr *pFileBlkPtr = WFSKrnGetFileBlkPtrForCategoryMedium(pDi->pAttrHdr, pDi->pAttrHdr->nAttrLog2Size, nI);
                                //u8 flag = 0xff<<(pFileBlkPtr->nBlkAdr%8);

                                ///
                                SetBitArray(fh, pBuffer, bufferSize, pFileBlkPtr->nBlkAdr, 8); // array!
                                /*
                                pCheckDiskArray[pFileBlkPtr->nBlkAdr/8] |= flag;
                                
                                if(~flag){

                                    ///
                                    pCheckDiskArray[pFileBlkPtr->nBlkAdr/8+1] |= ~flag;
                                }
                                */
                            }
                            break;
                        }
                        case FILE_SIZE_CATEGORY_LARGE:{
                            u32 nI, nJ, nNumLargeBlks = pDi->pAttrHdr->nAllocSize>>pDi->tba.pAreaInfo->ah.nLog2LargeBlkSize;
                            for(nJ=0;nJ<nNumLargeBlks;nJ++){
                                WFSFileLargeBlkPtr *pFileLargeBlkPtr = WFSKrnGetFileLargeBlkPtrForCategoryLarge(pDi->pAttrHdr, pDi->pAttrHdr->nAttrLog2Size, nJ);
                                for(nI=0;nI<(1<<pDi->tba.pAreaInfo->nLog2MediumBlkPerLargeBlk);nI++){
                                    WFSFileBlkInfo fileBlkInfo = WFSKrnGetFileBlkInfoForCategoryLarge(pDi->tba.pAreaInfo, pFileLargeBlkPtr, nI);
                                    //u8 flag = 0xff<<(fileBlkInfo.nBlkAdr%8);

                                    SetBitArray(fh, pBuffer, bufferSize, fileBlkInfo.nBlkAdr, 8); // array!
                                    /*
                                    ///
                                    pCheckDiskArray[fileBlkInfo.nBlkAdr/8] |= flag;
                                    if(~flag){

                                        ///
                                        pCheckDiskArray[fileBlkInfo.nBlkAdr/8+1] |= ~flag;
                                    }*/
                                }
                            }
                            break;
                        }
                        case FILE_SIZE_CATEGORY_VERY_LARGE: {
                            WFSBlkAdr nBlkAdr = pDi->tba.nBlkAdr;
                            u32 nI, nJ, nK;
                            u32 nNumLargeBlks = pDi->pAttrHdr->nAllocSize>>pDi->tba.pAreaInfo->ah.nLog2LargeBlkSize;
                            u32 nNumIndirectBlks = (nNumLargeBlks-1)/pDi->tba.pAreaInfo->nNumLargeBlkPtrsPerBlk+1;
                            for(nK=0;nK<nNumIndirectBlks;nK++){
                                u8* pIndirectPtr;
                                WFSBlkAdr *pBlkAdr = WFSKrnGetBlkAdrPtrFromAttrHdrAndIndirectBlkIdxForCategoryVeryLarge(pDi->tba.pAreaInfo, pDi->pAttrHdr, pDi->pAttrHdr->nAttrLog2Size, nK);

                                ///
                                //pCheckDiskArray[*pBlkAdr/8] |= 1<<(*pBlkAdr%8);
                                SetBit(fh, *pBlkAdr);
                                pDi->tba.nBlkAdr = *pBlkAdr;
                                WFSKrnReturnOnError( DirGetBlk(pDi, TRANS_FLAG_READ_BLK | BCACHE_FLAG_PINNED, (DirBlkHdr**)&pIndirectPtr) );
                                for(nJ=0;nJ<pDi->tba.pAreaInfo->nNumLargeBlkPtrsPerBlk;nJ++){
                                    WFSFileLargeBlkPtr *pFileLargeBlkPtr = WFSKrnGetFileLargeBlkPtrFromIndirectBlkForCategoryVeryLarge(pIndirectPtr, nJ);
                                    for(nI=0;nI<(1<<pDi->tba.pAreaInfo->nLog2MediumBlkPerLargeBlk);nI++){
                                        WFSFileBlkInfo fileBlkInfo = WFSKrnGetFileBlkInfoForCategoryLarge(pDi->tba.pAreaInfo, pFileLargeBlkPtr, nI);
                                        //u8 nFlag = 0xff<<(fileBlkInfo.nBlkAdr%8);

                                        SetBitArray(fh, pBuffer, bufferSize, fileBlkInfo.nBlkAdr, 8); // array!

                                        /*
                                        ///
                                        pCheckDiskArray[fileBlkInfo.nBlkAdr/8] |= nFlag;
                                        nFlag = ~nFlag;
                                        if(nFlag){

                                            ///
                                            pCheckDiskArray[fileBlkInfo.nBlkAdr/8+1] |= nFlag;
                                        }
                                        */
                                    }
                                    if(--nNumLargeBlks == 0){
                                        break;
                                    }
                                }
                                TransUnpinBlk(&pDi->tba);
                            }
                            pDi->tba.nBlkAdr = nBlkAdr;
                            break;
                        }
                        }
                    }
                    DirPopNodeStackEntry(&pDi->dni);
                    pDi->dni.nEntryIdx = 2;
                }
            }
            else{
                pDi->dni.pOfs = DirLeafGetIntraBlkOfsPtr_l2s((s32)pDi->dni.nEntryIdx, pDi->dni.pSubBlkHdr, pDi->dni.nLog2Size);
                SbaCheckOfs(*pDi->dni.pOfs);
                pDi->dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + *pDi->dni.pOfs);
                pDi->dni.nLog2Size = DirLeafGetSubBlkLog2Size(pDi->dni.pSubBlkHdr);
                pDi->dni.nEntryIdx = 1;
            }
        } else {
            if (pDi->dni.nEntryIdx == 1 && pDi->dni.pSubBlkHdr->aFirstChar[pDi->dni.pSubBlkHdr->nStrLen] == 0){
                u32 nNewBlkAdr = DirNodeGetBlkAdr_l2s(pDi->dni.pSubBlkHdr, pDi->dni.nLog2Size);

                ///
                //pCheckDiskArray[nNewBlkAdr/8] |= 1<<(nNewBlkAdr%8);
                SetBit(fh, nNewBlkAdr);

                TransUnpinBlk(&pDi->tba);
                pDi->tba.nBlkAdr = nNewBlkAdr;
                WFSKrnReturnOnError( DirGetBlk(pDi, TRANS_FLAG_READ_BLK | BCACHE_FLAG_PINNED, &pBlkHdr) );
                pDi->dni.pOfs = &pBlkHdr->nRootOfs;
                SbaCheckOfs(*pDi->dni.pOfs);
                pDi->dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + *pDi->dni.pOfs);
                if (pBlkHdr->sbah.mdh.nFlags & WFS_MDF_DIR_LEAF_BLK) {
                    pDi->dni.nLog2Size = DirLeafGetSubBlkLog2Size(pDi->dni.pSubBlkHdr);
                } else {
                    pDi->dni.nLog2Size = DirNodeGetSubBlkLog2Size(pDi->dni.pSubBlkHdr);
                }        
            } else {
                pDi->dni.pOfs = DirNodeGetIntraBlkOfsPtr_l2s((s32)pDi->dni.nEntryIdx, pDi->dni.pSubBlkHdr, pDi->dni.nLog2Size);
                SbaCheckOfs(*pDi->dni.pOfs);
                pDi->dni.pSubBlkHdr = (DirSubBlkHdr *)((u8*)pBlkHdr + *pDi->dni.pOfs);
                pDi->dni.nLog2Size = DirNodeGetSubBlkLog2Size(pDi->dni.pSubBlkHdr);
            }
            pDi->dni.nEntryIdx = 1;
        }
    }
}

