/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfskrn_DirRxTree.cpp - Functions for creating and using radix trees


  Copyright 2007-2008 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_DirRxTree.cpp,v $
  Revision 1.20  2008/12/08 00:01:42  kondo_masahiro
  Fixed RxtMergeNodeWithItsSingleChild().

  Revision 1.19  2008/12/05 05:47:26  kondo_masahiro
  Fixed error handling in RxtDeleteNode()
  ToDo: PathCache have to be fixed to the efficient method of delete node and to prevent the error.

  Revision 1.18  2008/11/27 00:03:28  ooizumi
  Fixed a bug in RxtInsert().

  Revision 1.17  2008/11/13 12:44:18  ooizumi
  Fixed a bug in RxtInsert().

  Revision 1.16  2008/10/29 02:22:16  ooizumi
  Fixed codes for IOP compiler.

  Revision 1.15  2008/10/28 08:13:34  ooizumi
  Modified RxtFindNext() to fix node offsets in currently accessing search directory infos.
  This code must be fixed in the future.

  Revision 1.14  2008/10/14 10:27:14  kondo_masahiro
  Added #define _DEBUG_BREAK_POINT

  Revision 1.13  2008/10/10 10:48:41  ooizumi
  Fixed RxtNextUntilSeparator().

  Revision 1.12  2008/09/26 07:37:22  ooizumi
  Fixed a bug RxtFindNext cannot handle separator.

  Revision 1.11  2008/09/02 00:09:22  ooizumi
  Fixed a bug with pRxi->pNodeStrPtr when inserting a new choice character and resizing node size.

  Revision 1.10  2008/08/28 05:54:38  paul
  Fixed a bug with pRxi->pOfs when inserting a prefix to an existing name

  Revision 1.9  2008/08/27 23:02:55  ooizumi
  Fixed newline to CR+LF(Windows format).

  Revision 1.8  2008/08/27 09:49:10  paul
  Added debug asserts

  Revision 1.7  2008/07/28 22:18:22  paul
  Fixed a bug in RxtDeleteNode().

  Revision 1.6  2008/07/21 22:37:04  paul
  Added a callback parameter to RxtCheck() to allow path cache entries to be checked for consistency

  Revision 1.5  2008/07/17 22:51:30  paul
  Fixed a bug

  Revision 1.4  2008/07/09 00:38:54  paul
  Minor changes

  Revision 1.3  2008/06/09 17:53:07  paul
  Rewrote DirRxTree to support PathCache. New version includes parent pointers in each node.

  Revision 1.2  2008/05/14 02:17:46  paul
  Fixed the DirBinarySplitSearchChoices() node choice ordering check to prevent utf8 being upgraded to signed int

  Revision 1.1  2008/05/10 03:57:03  kondo_masahiro
  Initial check-in

*---------------------------------------------------------------------------*/

#undef  _KRN_FILE_
#define _KRN_FILE_  "DirRxTree"

#include "wfskrn_SubBlkAlloc.h"

#include "wfskrn_Dir.h"
#include "wfskrn_DirFind.h"
#include "wfskrn_DirNodeStack.h"
#include "wfskrn_DirRxTree.h"

#include "wfskrn_Api.h"

#undef dbg
#if _DEBUG_DIR
    #define dbg(s) s
#else
    #define dbg(s)
#endif


WFSBool DirBinarySplitSearchChoices(utf8 *aSortedArray, u32 nNumEntries, utf8 cSearchChar, u32 *pEntryIdx) {
    if (nNumEntries==0) {
        return false;
    }
    u32  nLow  = 0;
    u32  nHigh = nNumEntries-1;
    utf8 cVal;
    while(1) {
        *pEntryIdx = (nLow + nHigh + 1)>>1;
        cVal = aSortedArray[*pEntryIdx];
        if (nLow >= nHigh) {
            if ((unsigned)cSearchChar >= (unsigned)aSortedArray[0]) {
                ++(*pEntryIdx);
            }
            return ((*pEntryIdx) && (cVal==cSearchChar));
        }
        if ((unsigned)cVal>(unsigned)cSearchChar) {
            nHigh = *pEntryIdx-1;
        } else {
            nLow = *pEntryIdx;
        }
    }
}


void RxtInit(RxtHdr *pRxtHdr, u32 nLog2Size) {
    SbaBlkInit(pRxtHdr, nLog2Size, WFS_NUM_BITS_REQUIRED(sizeof(RxtHdr)));
    u32 nRootNodeLog2Size = RxtGetNodeLog2Size_2(0, 0);
    pRxtHdr->nRootOfs = SbaAllocSubBlk((WFSSubBlkAllocHdr *)pRxtHdr, nRootNodeLog2Size);
    pRxtHdr->nReserveOfs = 0;
    RxtNodeHdr *pNodeHdr = (RxtNodeHdr *)((u8*)pRxtHdr + pRxtHdr->nRootOfs);
    pNodeHdr->nParentOfs = 0;
    pNodeHdr->nStrLen = 0;
    pNodeHdr->nNumEntries = 0;
    pNodeHdr->aFirstChar[0] = (utf8)0xC1;
    pNodeHdr->nLog2Size = nRootNodeLog2Size;
}


void RxtGetString(RxtHdr *pRxtHdr, RxtNodeHdr *pNodeHdr, utf8 *sStr, u32 nStrLen) {
    utf8 *pStr = &sStr[nStrLen];
    *pStr=0;
    while(1) {
        pStr -= pNodeHdr->nStrLen;
#if _DEBUG
        if (pStr < sStr) {
            WFSKrnOutputErrorStr("String too long");
        }
#endif
        memcpy(pStr, pNodeHdr->aFirstChar, pNodeHdr->nStrLen);
        if (pNodeHdr->nParentOfs == 0) {
            return;
        }
        u32 nEntryIdx = pNodeHdr->nEntryIdx;
        pNodeHdr = (RxtNodeHdr *)((u8*)pRxtHdr + pNodeHdr->nParentOfs);
#if _DEBUG
        if (pStr == sStr) {
            WFSKrnOutputErrorStr("String too long");
        }
#endif
        *(--pStr) = pNodeHdr->aFirstChar[pNodeHdr->nStrLen + nEntryIdx];
    }
}


static
void RxtNodeDebugPrint(RxtNodeHdr *pNodeHdr) {
    utf8 *pChoiceChar = &pNodeHdr->aFirstChar[pNodeHdr->nStrLen];
    utf8 cTemp = *pChoiceChar;
    *pChoiceChar = 0;
    MyOSReport("%d,%d,\"%s\"", pNodeHdr->nStrLen, pNodeHdr->nNumEntries, pNodeHdr->aFirstChar);
    *pChoiceChar = cTemp;
    u32 nI;
    for(nI=0; nI<pNodeHdr->nNumEntries; nI++) {
        if ((*pChoiceChar<32)||(*pChoiceChar>127)) {
            MyOSReport(",%d", *pChoiceChar);
        } else {
            MyOSReport(",'%c'", *pChoiceChar);
        }
        ++pChoiceChar;
    }
    //MyOSReport("\n");
}


#if _CHECK_RXT
void RxtCheck(RxtHdr *pRxtHdr, u32 nTotalSize, RxtCheckEntry cbCheckEntry) {
    // Checks the integrity of the radix tree:
    //  * That all allocated sub blocks of pRxtHdr are reachable
    //  * That all non-allocated blocks are on the free list
    //  * That no blocks overlap
    //  * That each node within the tree contains a sorted list of choices
    //  * That calculated node sizes match the stored node sizes
    //  * That parent pointers point to the correct parent
    //  * That nEntryIdx matches
    RxtItr rxi;
    rxi.nDepth = 0;
    u32 nParentOfs = 0;
    u32 nOfs = pRxtHdr->nRootOfs;
    utf8 *pChoiceChar;
    u32 nConnectedSize=1<<WFS_NUM_BITS_REQUIRED(sizeof(RxtHdr));
    if (pRxtHdr->nReserveOfs) {
        nConnectedSize += 1<<RXT_LOG2_RESERVE_SIZE;
    }
    u32 aUtilization[1<<(RXT_MAX_LOG2_SIZE - SBA_MIN_LOG2_SUB_BLK_SIZE - 5)];
    memset(aUtilization, 0, sizeof(aUtilization));
    //WfsSetBitField((unsigned*)aUtilization, 0, 1<<(WFS_NUM_BITS_REQUIRED(sizeof(RxtHdr))-SBA_MIN_LOG2_SUB_BLK_SIZE));
    u32 nRecIdx = 0;
    u32 nStrIdx = (u32)-1;
    rxi.nEntryIdx = 1;
    while(1) {
        rxi.pNodeHdr = (RxtNodeHdr *)((u8*)pRxtHdr + nOfs);
        WFSSubBlkFreeHdr *pSbfHdr = (WFSSubBlkFreeHdr *)rxi.pNodeHdr;
        if (pSbfHdr->nMagic == SBA_FREE_SUB_BLK_MAGIC) {
            WFSKrnOutputErrorStr("Pointer to deleted sub block detected!");
        }
#if _DEBUG_PATH_CACHE
        if ((nDirTest>=nDirTestBreakPoint-1)&&(nDirTest<=nDirTestBreakPoint)) {
            u32 nI;
            for(nI=0; nI<rxi.nDepth; nI++) {
                MyOSReport("    ");
            }
            MyOSReport("%03x: ", nOfs);
            RxtNodeDebugPrint(rxi.pNodeHdr);
        }
#endif
        u32 nNodeLog2Size = RxtGetNodeLog2Size(rxi.pNodeHdr);
        if (nNodeLog2Size != rxi.pNodeHdr->nLog2Size) {
            WFSKrnOutputErrorStr("pNodeHdr->nLog2Size incorrect!");
        }
        if (nParentOfs != rxi.pNodeHdr->nParentOfs) {
            WFSKrnOutputErrorStr("pNodeHdr->nParentOfs incorrect!");
        }
        if (rxi.nEntryIdx != rxi.pNodeHdr->nEntryIdx + 1) {
            WFSKrnOutputErrorStr("pNodeHdr->nEntryIdx incorrect!");
        }
        if (WfsTestAndSetBitField((unsigned*)aUtilization, nOfs>>SBA_MIN_LOG2_SUB_BLK_SIZE, 1<<(nNodeLog2Size-SBA_MIN_LOG2_SUB_BLK_SIZE))) {
            WFSKrnOutputErrorStr("DirRxTree: Overlap!");
        }
        nConnectedSize += (u32)(1<<nNodeLog2Size);
        nStrIdx += rxi.pNodeHdr->nStrLen+1;
        u16 *pSrcNodeOfsTblEnd = (u16*)((u8*)rxi.pNodeHdr + (1<<nNodeLog2Size));
        rxi.nEntryIdx = 1;
        if (rxi.pNodeHdr->nNumEntries == 0) {
            if (pRxtHdr->nNumRecs) {
                WFSKrnOutputErrorStr("No entries!");
            }
        }
        pChoiceChar = &rxi.pNodeHdr->aFirstChar[rxi.pNodeHdr->nStrLen];
        if (*pChoiceChar == 0) {
            // String terminator char indicates this is a record, so increment the record count
            u32 nAttrOfs = pSrcNodeOfsTblEnd[-1];
            RxtAttrHdr *pAttrHdr = (RxtAttrHdr *)((u8*)pRxtHdr + nAttrOfs);
            WFSSubBlkFreeHdr *pSbfHdr = (WFSSubBlkFreeHdr *)pAttrHdr;
            if (pSbfHdr->nMagic == SBA_FREE_SUB_BLK_MAGIC) {
                WFSKrnOutputErrorStr("Pointer to deleted attr block detected!");
            }
            u16 nAttrLog2Size = pAttrHdr->nLog2Size;
            if (nOfs != pAttrHdr->nParentOfs) {
                WFSKrnOutputErrorStr("pAttrHdr->nParentOfs incorrect!");
            }
            if (WfsTestAndSetBitField((unsigned*)aUtilization, nAttrOfs>>SBA_MIN_LOG2_SUB_BLK_SIZE, 1<<(nAttrLog2Size-SBA_MIN_LOG2_SUB_BLK_SIZE))) {
                WFSKrnOutputErrorStr("DirRxTree: Overlap!");
            }
            nConnectedSize += (u32)(1<<nAttrLog2Size);
            if (cbCheckEntry) {
                WFSPathName path;
                RxtGetString(pRxtHdr, rxi.pNodeHdr, path.sStr, nStrIdx);
                cbCheckEntry(nRecIdx, rxi.nDepth, path.sStr, pAttrHdr);
            }
            ++nRecIdx;
            ++rxi.nEntryIdx;
#if _DEBUG_PATH_CACHE
            if ((nDirTest>=nDirTestBreakPoint-1)&&(nDirTest<=nDirTestBreakPoint)) {
                WFSPathName path;
                RxtGetString(pRxtHdr, rxi.pNodeHdr, path.sStr, nStrIdx);
                MyOSReport("  %s\n", path.sStr);
            }
#endif
        } else if (rxi.pNodeHdr->nNumEntries == 1) {
            WFSKrnOutputErrorStr("Non-terminating single-choice node!");
        } else {
#if _DEBUG_PATH_CACHE
            if ((nDirTest>=nDirTestBreakPoint-1)&&(nDirTest<=nDirTestBreakPoint)) {
                MyOSReport("\n");
            }
#endif
        }
        if (rxi.nEntryIdx > rxi.pNodeHdr->nNumEntries) {
            do {
                if (rxi.nDepth == 0) {
                    // We reached the top of the stack, so this should be the last record
                    goto CheckFreeSpace;
                }
                rxi.nEntryIdx = rxi.pNodeHdr->nEntryIdx + 2;
                nOfs = rxi.pNodeHdr->nParentOfs;
                nStrIdx -= rxi.pNodeHdr->nStrLen+1;
                rxi.pNodeHdr = (RxtNodeHdr *)((u8*)pRxtHdr + nOfs);
                --rxi.nDepth;
            } while(rxi.nEntryIdx > rxi.pNodeHdr->nNumEntries);
            pChoiceChar = &rxi.pNodeHdr->aFirstChar[rxi.pNodeHdr->nStrLen + rxi.nEntryIdx-1];
            if ((unsigned)pChoiceChar[-1] >= (unsigned)*pChoiceChar) {
                WFSKrnOutputErrorStr("Choice out of order!");
            }
        }
        ++rxi.nDepth;
        nParentOfs = nOfs;
        nOfs = RxtGetChildOfs(rxi.nEntryIdx, rxi.pNodeHdr);
    }
CheckFreeSpace:
    // Next check the free space
    SbaCheckFreeSpace((WFSSubBlkAllocHdr *)pRxtHdr, aUtilization);
    if (nTotalSize) {
        if ((pRxtHdr->sbah.nTotalFree + nConnectedSize) != nTotalSize) {
            WFSKrnOutputErrorStr("DirRxTree: Memory Leak detected within block!");
        }
    }
    if (nRecIdx != pRxtHdr->nNumRecs) {
        WFSKrnOutputErrorStr("NumRecs Error!");
    }
}
#endif //_CHECK_RXT


WFSKrnResult RxtFind(RxtHdr *pRxtHdr, RxtItr *pRxi, u32 nStrLen) {
    pRxi->nDepth = 0;
    pRxi->pOfs = &pRxtHdr->nRootOfs;
    WFSKrnResult nResult;
    utf8 *pStrEnd = pRxi->pStrPtr + nStrLen;
    RxtNodeHdr *pNodeHdr;
    while(1) {
        pNodeHdr = (RxtNodeHdr *)((u8*)pRxtHdr + *pRxi->pOfs);
        u32 nNodeStrLen = pNodeHdr->nStrLen;
        pRxi->pNodeStrPtr = pNodeHdr->aFirstChar;
        u32 nSearchStrLen = pStrEnd - pRxi->pStrPtr;
        if (nSearchStrLen < nNodeStrLen) {
            while((nSearchStrLen) && (*pRxi->pStrPtr == *pRxi->pNodeStrPtr)) {
                ++pRxi->pStrPtr; ++pRxi->pNodeStrPtr; --nSearchStrLen;
            }
            nNodeStrLen -= (pRxi->pNodeStrPtr - pNodeHdr->aFirstChar);
            pRxi->nEntryIdx = -(s32)nNodeStrLen;
            nResult = WFSKRN_RESULT_DIR_NODE_STRING_MISMATCH;
            if (nSearchStrLen == 0) {
                nResult = WFSKRN_RESULT_DIR_NODE_STRING_PREFIX;
            }
            break;
        }
        while((nNodeStrLen) && (*pRxi->pStrPtr == *pRxi->pNodeStrPtr)) {
            ++pRxi->pStrPtr; ++pRxi->pNodeStrPtr; --nNodeStrLen;
        }
        if (nNodeStrLen) {
            pRxi->nEntryIdx = pRxi->pNodeStrPtr - pNodeHdr->aFirstChar;
            nResult = WFSKRN_RESULT_DIR_NODE_STRING_MISMATCH;
            break;
        }
        // matched all the characters from the prefix. pRxi->pNodeStrPtr should now be pointing to the array of choices following the prefix.
        utf8 cSearchChar = *pRxi->pStrPtr;
        if (pRxi->pStrPtr == pStrEnd){
            cSearchChar = 0;
        }
        if (!DirBinarySplitSearchChoices(pRxi->pNodeStrPtr, pNodeHdr->nNumEntries, cSearchChar, (u32*)&pRxi->nEntryIdx)) {
            nResult = WFSKRN_RESULT_DIR_NODE_CHOICE_NOT_FOUND;
            break;
        }
        // Found the search character.
        if (cSearchChar==0) {
            // We matched a 0 .. which means we have found the search string
            nResult = WFSKRN_RESULT_DIR_ENTRY_FOUND;
            break;
        }
        ++pRxi->nDepth;
        pRxi->pOfs = RxtGetChildOfsPtr((s32)pRxi->nEntryIdx, pNodeHdr);
        ++pRxi->pStrPtr;
    }
    pRxi->pNodeHdr = pNodeHdr;
    return nResult;
}


WFSKrnResult RxtNextUntilSeparator(RxtHdr *pRxtHdr, RxtItr *pRxi, utf8 cSeparator) {
    RxtNodeHdr *pNodeHdr = pRxi->pNodeHdr;
    utf8 *pDst, *pSrc, *pSrcEnd;
    if (pRxi->nEntryIdx < 0) {
        pDst = pRxi->pStrPtr;
        pSrc = pRxi->pNodeStrPtr;
        pSrcEnd = pSrc + 1-pRxi->nEntryIdx;
        if(*pDst)
            pDst++; // Skip choice char
        while(pSrc < pSrcEnd) {
            if (*pSrc == cSeparator) {
                return WFSKRN_RESULT_DIR_NODE_STRING_PREFIX;
            }
            *pDst++ = *pSrc++;
        }
        pRxi->pStrPtr = pDst-1;
        pRxi->nEntryIdx = 1;
    } else {
        while(pRxi->nEntryIdx == pNodeHdr->nNumEntries) {
            u32 nParentOfs = (u32)pNodeHdr->nParentOfs;
            if (nParentOfs == 0) {
                return WFSKRN_RESULT_NOT_FOUND;
            }
            if (pNodeHdr->aFirstChar[pNodeHdr->nStrLen + pNodeHdr->nNumEntries-1] == cSeparator) {
                return WFSKRN_RESULT_NOT_FOUND;
            }
            pSrc = pNodeHdr->aFirstChar; 
            pSrcEnd = pSrc + pNodeHdr->nStrLen;
            while(pSrc<pSrcEnd) {
                if (*pSrc++ == cSeparator) {
                    return WFSKRN_RESULT_NOT_FOUND;
                }
            }
            pRxi->pStrPtr -= pNodeHdr->nStrLen+1;
            pRxi->nEntryIdx = pNodeHdr->nEntryIdx+1;
            pNodeHdr = (RxtNodeHdr *)((u8*)pRxtHdr + nParentOfs);
        }
        *pRxi->pStrPtr = pNodeHdr->aFirstChar[pNodeHdr->nStrLen + pRxi->nEntryIdx];
        ++pRxi->nEntryIdx;
    }
    if (*pRxi->pStrPtr == 0) {
        return WFSKRN_RESULT_OK;
    }
    ++pRxi->pStrPtr;
    u32 nOfs = RxtGetChildOfs(pRxi->nEntryIdx, pNodeHdr);
    pNodeHdr = (RxtNodeHdr *)((u8*)pRxtHdr + nOfs);
    pRxi->nEntryIdx = 1;
    while(1) {
        pDst = pRxi->pStrPtr;
        pSrc = pNodeHdr->aFirstChar;
        pSrcEnd = pSrc + pNodeHdr->nStrLen + 1;
        while(pSrc < pSrcEnd) {
            if (*pSrc == cSeparator) {
                *pDst = 0;
                return WFSKRN_RESULT_OK;
                //return WFSKRN_RESULT_DIR_NODE_STRING_PREFIX;
            }
            *pDst++ = *pSrc++;
        }
        //memcpy(pRxi->pStrPtr, pNodeHdr->aFirstChar, pNodeHdr->nStrLen+1);
        pRxi->pStrPtr += pNodeHdr->nStrLen;
        if (*pRxi->pStrPtr == 0) {
            pRxi->pNodeHdr = pNodeHdr;
            return WFSKRN_RESULT_OK;
        }
        ++pRxi->pStrPtr;
#if _DEBUG_DIR
        u32 nParentOfs = nOfs;
#endif
        nOfs = RxtGetChildOfs(1, pNodeHdr);
        pNodeHdr = (RxtNodeHdr *)((u8*)pRxtHdr + nOfs);
#if _DEBUG_DIR
        if (pNodeHdr->nParentOfs != nParentOfs) {
            WFSKrnOutputErrorStr("Parent Incorrect");
        }
#endif
    }
}


WFSKrnResult RxtNext(RxtHdr *pRxtHdr, RxtItr *pRxi) {
    RxtNodeHdr *pNodeHdr = pRxi->pNodeHdr;
    if (pRxi->nEntryIdx < 0) {
        memcpy(pRxi->pStrPtr, pRxi->pNodeStrPtr, 1-pRxi->nEntryIdx);
        pRxi->pStrPtr -= pRxi->nEntryIdx;
        pRxi->nEntryIdx = 1;
    } else {
        while(pRxi->nEntryIdx == pNodeHdr->nNumEntries) {
            u32 nParentOfs = (u32)pNodeHdr->nParentOfs;
            if (nParentOfs == 0) {
                return WFSKRN_RESULT_NOT_FOUND;
            }
            pRxi->pStrPtr -= pNodeHdr->nStrLen+1;
            pRxi->nEntryIdx = pNodeHdr->nEntryIdx+1;
            pNodeHdr = (RxtNodeHdr *)((u8*)pRxtHdr + nParentOfs);
        }
        *pRxi->pStrPtr = pNodeHdr->aFirstChar[pNodeHdr->nStrLen + pRxi->nEntryIdx];
        ++pRxi->nEntryIdx;
    }
    if (*pRxi->pStrPtr == 0) {
        return WFSKRN_RESULT_OK;
    }
    ++pRxi->pStrPtr;
    u32 nOfs = RxtGetChildOfs(pRxi->nEntryIdx, pNodeHdr);
    pNodeHdr = (RxtNodeHdr *)((u8*)pRxtHdr + nOfs);
    pRxi->nEntryIdx = 1;
    while(1) {
        memcpy(pRxi->pStrPtr, pNodeHdr->aFirstChar, pNodeHdr->nStrLen+1);
        pRxi->pStrPtr += pNodeHdr->nStrLen;
        if (*pRxi->pStrPtr == 0) {
            pRxi->pNodeHdr = pNodeHdr;
            return WFSKRN_RESULT_OK;
        }
        ++pRxi->pStrPtr;
#if _DEBUG_DIR
        u32 nParentOfs = nOfs;
#endif
        nOfs = RxtGetChildOfs(1, pNodeHdr);
        pNodeHdr = (RxtNodeHdr *)((u8*)pRxtHdr + nOfs);
#if _DEBUG_DIR
        if (pNodeHdr->nParentOfs != nParentOfs) {
            WFSKrnOutputErrorStr("Parent Incorrect");
        }
#endif    
    }
}


WFSKrnResult RxtFindNext(RxtHdr *pRxtHdr, RxtItr *pRxi, utf8 cSeparator) {
    // This function requires that pRxi->pNodeHdr is pointing to a RxtAttrHdr entry within the radix tree
    pRxi->nDepth = 0;
    pRxi->pOfs = &pRxtHdr->nRootOfs;
    //WFSKrnResult nResult;
    //utf8 *pStrEnd = pRxi->pStrPtr + nStrLen;
    RxtNodeHdr *pNodeHdr = pRxi->pNodeHdr;
    pRxi->nEntryIdx = pNodeHdr->nEntryIdx + 1;
    pNodeHdr = (RxtNodeHdr *)((u8*)pRxtHdr + pNodeHdr->nParentOfs);
    // If separator is found in choice chars of parent node, search failed.
    if (pNodeHdr->aFirstChar[pNodeHdr->nStrLen + pRxi->nEntryIdx - 1] == cSeparator) {
        return WFSKRN_RESULT_NOT_FOUND;
    }
    // To backtrack radix tree until finding next entry.
    while(pRxi->nEntryIdx == pNodeHdr->nNumEntries) {
        // If no parent node, search failed.
        u32 nParentOfs = (u32)pNodeHdr->nParentOfs;
        if (nParentOfs == 0) {
            return WFSKRN_RESULT_NOT_FOUND;
        }
        // If separator is found in common chars, search failed.
        utf8 *pPrefix = pNodeHdr->aFirstChar; 
        utf8 *pPrefixEnd = pPrefix + pNodeHdr->nStrLen;
        while(pPrefix<pPrefixEnd) {
            if (*pPrefix++ == cSeparator) {
                return WFSKRN_RESULT_NOT_FOUND;
            }
        }
        pRxi->pStrPtr -= pNodeHdr->nStrLen+1;
        pRxi->nEntryIdx = pNodeHdr->nEntryIdx+1;
        pNodeHdr = (RxtNodeHdr *)((u8*)pRxtHdr + nParentOfs);
        // If separator is found in choice chars of parent node, search failed.
        if (pNodeHdr->aFirstChar[pNodeHdr->nStrLen + pRxi->nEntryIdx - 1] == cSeparator) {
            return WFSKRN_RESULT_NOT_FOUND;
        }
    }
    *pRxi->pStrPtr = pNodeHdr->aFirstChar[pNodeHdr->nStrLen + pRxi->nEntryIdx];
    ++pRxi->nEntryIdx;

    if (*pRxi->pStrPtr == 0) {
        return WFSKRN_RESULT_OK;
    }
    ++pRxi->pStrPtr;
    u32 nOfs = RxtGetChildOfs(pRxi->nEntryIdx, pNodeHdr);
    pNodeHdr = (RxtNodeHdr *)((u8*)pRxtHdr + nOfs);
    pRxi->nEntryIdx = 1;
    while(1) {
        memcpy(pRxi->pStrPtr, pNodeHdr->aFirstChar, pNodeHdr->nStrLen+1);
        pRxi->pStrPtr += pNodeHdr->nStrLen;
        if (*pRxi->pStrPtr == 0) {
            pRxi->pNodeHdr = pNodeHdr;
            return WFSKRN_RESULT_OK;
        }
        ++pRxi->pStrPtr;
#if _DEBUG_DIR
        u32 nParentOfs = nOfs;
#endif
        nOfs = RxtGetChildOfs(1, pNodeHdr);
        pNodeHdr = (RxtNodeHdr *)((u8*)pRxtHdr + nOfs);
#if _DEBUG_DIR
        if (pNodeHdr->nParentOfs != nParentOfs) {
            WFSKrnOutputErrorStr("Parent Incorrect");
        }
#endif    
    }
/*    if (pRxi->nEntryIdx > pNodeHdr->nNumEntries) {
        return WFSKRN_RESULT_NOT_FOUND;
    }
    pNodeHdr = (RxtNodeHdr *)((u8*)pRxtHdr + RxtGetChildOfs(pRxi->nEntryIdx, pNodeHdr));
    return WFSKRN_RESULT_OK;*/
}

WFSKrnResult RxtInsert(RxtHdr *pRxtHdr, RxtItr *pRxi, u32 nSuffixLen, RxtAttrHdr *pAttrHdr) {
    //MyOSReport("RxtInsert: pRxtHdr = %x, pRxi = %x, nSuffixLen = %d, pAttrHdr = %x\n", pRxtHdr, pRxi, nSuffixLen, pAttrHdr);
    //MyOSReport("RxtInsert: pRxi->pStrPtr= %s\n", pRxi->pStrPtr);
#if _DEBUG_BREAK_POINT
    static int c=0;++c;if(c>=126307){
        int a;a=0;
    }
#endif
    if (pRxtHdr->nReserveOfs == 0) {
        // Note: This technique sacrifices a block of bytes in order to make the delete implementation simpler
        // The on-disk structures do not use this technique within blocks because it would reduce the compaction of every block.
        // It is fine for a one-off in-memory structure however.
        pRxtHdr->nReserveOfs = SbaAllocSubBlk((WFSSubBlkAllocHdr *)pRxtHdr, RXT_LOG2_RESERVE_SIZE);
        if (pRxtHdr->nReserveOfs == 0) {
            return WFSKRN_RESULT_DIR_BLK_FULL;
        }
    }
    u32 nAttrOfs = (u32)((u8*)pAttrHdr - (u8*)pRxtHdr);
    //RxtAttrHdr *pAttrHdr = (RxtAttrHdr *)((u8*)pRxtHdr + nAttrOfs);
    RxtNodeHdr *pNodeHdr = pRxi->pNodeHdr;
    u16 *pExistingNodeOfsTblEnd = (u16*)((u8*)pNodeHdr + (1<<pNodeHdr->nLog2Size));
    if (pRxtHdr->nNumRecs==0) {
        // No records in this directory yet. Add the first record.
        pNodeHdr->nStrLen = nSuffixLen;
        pNodeHdr->nNumEntries = 1;
        u32 nNewLog2Size = RxtGetNodeLog2Size(pNodeHdr);
        if (pNodeHdr->nLog2Size != nNewLog2Size) {
            u32 nOfs = (u32)((u8*)pNodeHdr - (u8*)pRxtHdr);
            dbg(SbaCheckOfs(nOfs));
            SbaFreeSubBlk((WFSSubBlkAllocHdr *)pRxtHdr, nOfs, pNodeHdr->nLog2Size);
            *pRxi->pOfs = SbaAllocSubBlk((WFSSubBlkAllocHdr *)pRxtHdr, nNewLog2Size);
            pNodeHdr = pRxi->pNodeHdr = (RxtNodeHdr *)((u8*)pRxtHdr + *pRxi->pOfs);
            //MyOSReport("pRxi->pOfs was updated\n");
        }
        pNodeHdr->nLog2Size = nNewLog2Size;
        pNodeHdr->nEntryIdx = 0;
        pNodeHdr->nParentOfs = 0;
        pNodeHdr->nStrLen = nSuffixLen;
        pNodeHdr->nNumEntries = 1;
        memcpy(pNodeHdr->aFirstChar, pRxi->pStrPtr, (size_t)(pNodeHdr->nStrLen+1));
        RxtSetChildOfs(1, nAttrOfs, pNodeHdr);
        pAttrHdr->nParentOfs = pRxtHdr->nRootOfs;
        ++pRxtHdr->nNumRecs;
#if (_CHECK_RXT >= 2)
        RxtCheck(pRxtHdr, 0, 0); //WFS_LOG2_PATH_CACHE_SIZE, NULL);
#endif
        return WFSKRN_RESULT_OK;
    }

    // Insert the new name at the location found
    //
    // To insert a new file name into a directory block requires to allocate up to two new sub blocks
    // for name nodes, resize an existing name node, and add a new sub block for file attributes and
    // block pointers. It needs to be determined up-front whether this will require splitting the block.

    u32 nNumMatching = (u32)(pRxi->pNodeStrPtr - pNodeHdr->aFirstChar);
    u32 nNodeNumLeft = pNodeHdr->nStrLen - nNumMatching;
    u16 nNameNodeLog2Size = 0;
    u16 nNameNodeOfs;
    u16 nForkNodeLog2Size;
    u16 nForkNodeOfs;
    u16 *pForkNodeOfsTblEnd;
    RxtNodeHdr *pNameNodeHdr = 0;
    u16 *pAttrOfs = 0;
    u8 nEntryIdx;
    utf8 cNameChar = pRxi->pStrPtr[0];

    if (nSuffixLen) {
        // allocate a sub block for the new name entry
        u32 nNodeStrLen = nSuffixLen-1;
        nNameNodeLog2Size = (u16)RxtGetNodeLog2Size_2(nNodeStrLen, 1);
        nNameNodeOfs = SbaAllocSubBlk((WFSSubBlkAllocHdr *)pRxtHdr, nNameNodeLog2Size);
        if (nNameNodeOfs == 0) {
            return WFSKRN_RESULT_DIR_BLK_FULL;
        }
        pNameNodeHdr = (RxtNodeHdr *)((u8*)pRxtHdr + nNameNodeOfs);
        pNameNodeHdr->nLog2Size = nNameNodeLog2Size;
        pNameNodeHdr->nNumEntries = 1;
        pNameNodeHdr->nStrLen = (u8)nNodeStrLen;
        memcpy(pNameNodeHdr->aFirstChar, &pRxi->pStrPtr[1], nNodeStrLen);
        pRxi->pNodeHdr = pNameNodeHdr;
        pRxi->pNodeStrPtr = &pNameNodeHdr->aFirstChar[nNodeStrLen];
        *pRxi->pNodeStrPtr = 0;
        pAttrOfs = RxtGetChildOfsPtr(1, pNameNodeHdr);
        *pAttrOfs = nAttrOfs;
        pAttrHdr->nParentOfs = nNameNodeOfs;
    } else {
        nNameNodeOfs = nAttrOfs;
        cNameChar = 0;
    }

    if (nNodeNumLeft == 0) {
        // Add a new option to the existing node
        u16 nUpdatedNodeOfs;
        u16 nUpdatedNodeLog2Size = (u16)RxtGetNodeLog2Size_2(pNodeHdr->nStrLen, (u32)(pNodeHdr->nNumEntries+1));
        if (nUpdatedNodeLog2Size != pNodeHdr->nLog2Size) {
            // The node needs to be copied to a larger sub block
            nUpdatedNodeOfs = SbaAllocSubBlk((WFSSubBlkAllocHdr *)pRxtHdr, nUpdatedNodeLog2Size);
            if (nUpdatedNodeOfs == 0) {
                // Allocation failed, so undo any changes so far, split the block, and try again
                if (nSuffixLen) {
                    // Undo the allocation of the name sub block
                    SbaFreeSubBlk((WFSSubBlkAllocHdr *)pRxtHdr, nNameNodeOfs, nNameNodeLog2Size);
                }
                //goto NameNotAllocated;
                return WFSKRN_RESULT_DIR_BLK_FULL;
            }
            RxtNodeHdr *pUpdatedNodeHdr = (RxtNodeHdr *)((u8*)pRxtHdr + nUpdatedNodeOfs);
            pUpdatedNodeHdr->nLog2Size = nUpdatedNodeLog2Size;
            u16 *pUpdatedNodeOfsTblEnd = (u16*)((u8*)pUpdatedNodeHdr + (1<<nUpdatedNodeLog2Size));
            // Copy the old choices to the new sub block while inserting the new choice
            u8 *pCopyDst = (u8*)&pUpdatedNodeHdr->nEntryIdx;   // Note: This code may need to be changed if the RxtNodeHdr struct changes
            u8 *pCopySrc = (u8*)&pNodeHdr->nEntryIdx;
            s32 nCopySize1 = (u8*)&pNodeHdr->aFirstChar[pNodeHdr->nStrLen + pRxi->nEntryIdx] - pCopySrc;
            memcpy(pCopyDst, pCopySrc, nCopySize1);
            pCopyDst += nCopySize1;
            pCopySrc += nCopySize1;
            *pCopyDst++ = (u8)cNameChar;
            s32 nCopySize2 = (s32)(pNodeHdr->nNumEntries - pRxi->nEntryIdx);
            memcpy(pCopyDst, pCopySrc, (u32)nCopySize2);
            pUpdatedNodeHdr->nNumEntries = pNodeHdr->nNumEntries+1;
            // Copy the table of child offset pointers from the old block to the new block while inserting the new offset.
            // Fix up parent pointers of existing children to point to new location.
            u16 *pDst = &pUpdatedNodeOfsTblEnd[-(s32)pUpdatedNodeHdr->nNumEntries];
            u16 *pDstEnd = pDst + nCopySize2;
            u16 *pSrcStart = &pExistingNodeOfsTblEnd[-(s32)pNodeHdr->nNumEntries];
            memcpy(pDst, pSrcStart, (u8*)pDstEnd-(u8*)pDst);
            nEntryIdx = pNodeHdr->nNumEntries;
            while(pDst<pDstEnd) {
                RxtNodeHdr *pChildNodeHdr = (RxtNodeHdr *)((u8*)pRxtHdr + *pDst++);
                pChildNodeHdr->nParentOfs = nUpdatedNodeOfs;
                pChildNodeHdr->nEntryIdx  = nEntryIdx--;
            }
            if (pAttrOfs == 0) {
                pAttrOfs = pDst;
                pAttrHdr->nParentOfs = nUpdatedNodeOfs;
                pRxi->pNodeHdr    = pUpdatedNodeHdr;
                pRxi->pNodeStrPtr = &pUpdatedNodeHdr->aFirstChar[pUpdatedNodeHdr->nStrLen];
            }
            u16 *pOfs = pRxi->pOfs;
            pRxi->pOfs = pDst;
            *pDst++ = nNameNodeOfs;
            if (pNameNodeHdr) {
                pNameNodeHdr->nParentOfs = nUpdatedNodeOfs;
                pNameNodeHdr->nEntryIdx  = nEntryIdx;
            }
            pSrcStart = &pExistingNodeOfsTblEnd[nCopySize2-pNodeHdr->nNumEntries];
            pDstEnd = pDst + pRxi->nEntryIdx;
            memcpy(pDst, pSrcStart, pRxi->nEntryIdx * sizeof(u16));
            while(pDst<pDstEnd) {
                RxtNodeHdr *pChildNodeHdr = (RxtNodeHdr *)((u8*)pRxtHdr + *pDst++);
                pChildNodeHdr->nParentOfs = nUpdatedNodeOfs;
            }
            SbaFreeSubBlk((WFSSubBlkAllocHdr *)pRxtHdr, *pOfs, pNodeHdr->nLog2Size);
            // Fixed node offsets in currently accessing search directory infos.
            // ToDo: Fix only when issuing search directory functions.
            u32 nJ;
            for(nJ = 0; nJ < WFSKRN_MAX_SEARCH_DIR_HANDLES; nJ++)
            {
                if(wkg.hnd.aSearchDirInfo[nJ].pci.rxi.pNodeHdr == (RxtNodeHdr *)((u8*)wkg.pathCache.pRxtHdr + *pOfs))
                    wkg.hnd.aSearchDirInfo[nJ].pci.rxi.pNodeHdr = (RxtNodeHdr *)((u8*)wkg.pathCache.pRxtHdr + nUpdatedNodeOfs);
            }
            *pOfs = nUpdatedNodeOfs; // Update the parent pointer
        } else {
            // The new option can be added without changing the size of the node
            utf8 *pCharInsertPoint = pNodeHdr->aFirstChar + pNodeHdr->nStrLen + pRxi->nEntryIdx;
            s32 nCopySize = (s32)(pNodeHdr->nNumEntries - pRxi->nEntryIdx);
            memmove(pCharInsertPoint+1, pCharInsertPoint, (u32)nCopySize);
            pCharInsertPoint[0] = cNameChar;
            u16 *pSrcStart = &pExistingNodeOfsTblEnd[-(s32)pNodeHdr->nNumEntries];
            u16 *pDst = pSrcStart-1;
            u16 *pDstEnd = pDst + nCopySize;
            memmove(pDst, pSrcStart, sizeof(u16) * nCopySize);
            nEntryIdx = pNodeHdr->nNumEntries;
            while(pDst<pDstEnd) {
                RxtNodeHdr *pChildNodeHdr = (RxtNodeHdr *)((u8*)pRxtHdr + *pDst++);
                pChildNodeHdr->nEntryIdx = nEntryIdx--;
            }
            if (pAttrOfs == 0) {
                pAttrOfs = pDst;
                pAttrHdr->nParentOfs = (u16)((u8*)pNodeHdr - (u8*)pRxtHdr);
            } else {
#if _DEBUG_DIR
                if ((pNameNodeHdr==0)||(*pRxi->pOfs != (u8*)pNodeHdr - (u8*)pRxtHdr)) {
                    WFSKrnOutputErrorStr("RxtInsert: Unexpected");
                }
#endif
                //pNameNodeHdr->nParentOfs = *pRxi->pOfs; // Note: Sometimes *pRxi->pOfs points invalid offset.
                // ToDo: Comfirm when *pRxi->pOfs points unexpected offset.
                pNameNodeHdr->nParentOfs = (u8*)pNodeHdr - (u8*)pRxtHdr;
                pNameNodeHdr->nEntryIdx = nEntryIdx;
            }
            // If allocated a sub block for the new name entry, it is need to update the parent pointer.
            if(nSuffixLen)
                pRxi->pOfs = pDst;
            *pDst = nNameNodeOfs;
            ++pNodeHdr->nNumEntries;
        }
    } else {
        // Replace the start of the existing node with a new 2-choice node
        // and update the existing node to contain only the end of the string,
        // resizing the node if its size drops to a smaller power-of-2 allocation size.
        // (note: the resize is compulsory due to the algorithms used elsewhere).
        //
        // First, try to allocate a sub block for the new 2-choice node.
        nForkNodeLog2Size = (u16)RxtGetNodeLog2Size_2(nNumMatching, 2);
        nForkNodeOfs = SbaAllocSubBlk((WFSSubBlkAllocHdr *)pRxtHdr, nForkNodeLog2Size);
        if (nForkNodeOfs == 0) {
            // Allocation failed, so undo any changes so far, split the block, and try again
            if (nSuffixLen) {
                // Undo the allocation of the name sub block
                SbaFreeSubBlk((WFSSubBlkAllocHdr *)pRxtHdr, nNameNodeOfs, nNameNodeLog2Size);
            }
            return WFSKRN_RESULT_DIR_BLK_FULL;
        }
        RxtNodeHdr *pForkNodeHdr = (RxtNodeHdr *)((u8*)pRxtHdr + nForkNodeOfs);
        pForkNodeHdr->nLog2Size = nForkNodeLog2Size;
        pForkNodeOfsTblEnd = (u16*)((u8*)pForkNodeHdr + (1<<nForkNodeLog2Size));
        pForkNodeHdr->nParentOfs = pNodeHdr->nParentOfs;
        pForkNodeHdr->nEntryIdx  = pNodeHdr->nEntryIdx;
        if (pNameNodeHdr) {
            pNameNodeHdr->nParentOfs = nForkNodeOfs;
        }
        pForkNodeHdr->nNumEntries = 2;
        pForkNodeHdr->nStrLen = (u8)nNumMatching;
        memcpy(pForkNodeHdr->aFirstChar, pNodeHdr->aFirstChar, nNumMatching);
        utf8 cNodeChar = pNodeHdr->aFirstChar[nNumMatching];

        // Now, update the existing node
        u16 nUpdatedNodeOfs;
        u16 nUpdatedNodeLog2Size = (u16)RxtGetNodeLog2Size_2(nNodeNumLeft-1, pNodeHdr->nNumEntries);
        RxtNodeHdr *pUpdatedNodeHdr;
        if (nUpdatedNodeLog2Size != pNodeHdr->nLog2Size) {
            // The existing node should be reduced in size.
            // First try to allocate a new smaller sub block.
            // If this is not possible, then resize in place.
            nUpdatedNodeOfs = SbaAllocSubBlk((WFSSubBlkAllocHdr *)pRxtHdr, nUpdatedNodeLog2Size);
            u16 *pUpdatedNodeOfsTblEnd;
            u32 nOldNodeOfs = (u16)((u8*)pNodeHdr - (u8*)pRxtHdr);
            if (nUpdatedNodeOfs == 0) {
                // allocation failed, so resize in place
                // shift/copy the end of the string and the following choices backwards over the start of the string
                pUpdatedNodeHdr = pNodeHdr;
                nUpdatedNodeOfs = (u16)nOldNodeOfs;
                pUpdatedNodeOfsTblEnd = (u16*)((u8*)pUpdatedNodeHdr + (1<<nUpdatedNodeLog2Size));
                memmove(pUpdatedNodeHdr->aFirstChar, &pUpdatedNodeHdr->aFirstChar[nNumMatching+1], nNodeNumLeft + pUpdatedNodeHdr->nNumEntries);
                // shift over the table of pointers at the end of the sub block
                memmove(&pUpdatedNodeOfsTblEnd[-(s32)pUpdatedNodeHdr->nNumEntries], &pExistingNodeOfsTblEnd[-(s32)pUpdatedNodeHdr->nNumEntries], sizeof(u16) * pUpdatedNodeHdr->nNumEntries);
                SbaReduceSubBlkSize((WFSSubBlkAllocHdr *)pRxtHdr, nUpdatedNodeOfs, pNodeHdr->nLog2Size, nUpdatedNodeLog2Size);
            } else {
                // copy the end of the node to the new node, and free the old node
                pUpdatedNodeHdr = (RxtNodeHdr *)((u8*)pRxtHdr + nUpdatedNodeOfs);
                // Copy the old choices to the new sub block while inserting the new choice
                *((u64*)pUpdatedNodeHdr) = *((u64*)pNodeHdr);   // Note: This code may need to be changed if the RxtNodeHdr struct changes
                memcpy(pUpdatedNodeHdr->aFirstChar, &pNodeHdr->aFirstChar[nNumMatching+1], nNodeNumLeft + pUpdatedNodeHdr->nNumEntries - 1);
                pUpdatedNodeOfsTblEnd = (u16*)((u8*)pUpdatedNodeHdr + (1<<nUpdatedNodeLog2Size));
                u16 *pDst = &pUpdatedNodeOfsTblEnd[-(s32)pUpdatedNodeHdr->nNumEntries];
                u16 *pSrcStart = &pExistingNodeOfsTblEnd[-(s32)pUpdatedNodeHdr->nNumEntries];
                u16 *pDstEnd = pDst + pUpdatedNodeHdr->nNumEntries;
                memcpy(pDst, pSrcStart, (u8*)pDstEnd - (u8*)pDst);
                while(pDst<pDstEnd) {
                    RxtNodeHdr *pChildNodeHdr = (RxtNodeHdr *)((u8*)pRxtHdr + *pDst++);
                    pChildNodeHdr->nParentOfs = nUpdatedNodeOfs;
                }
                SbaFreeSubBlk((WFSSubBlkAllocHdr *)pRxtHdr, nOldNodeOfs, pNodeHdr->nLog2Size);
            }
        } else {
            // the sub block remains the same size
            pUpdatedNodeHdr = pNodeHdr;
            nUpdatedNodeOfs = (u16)((u8*)pUpdatedNodeHdr - (u8*)pRxtHdr);
            dbg(SbaCheckOfs(nUpdatedNodeOfs));
            // Remove characters from the start of the string
            memmove(pUpdatedNodeHdr->aFirstChar, &pUpdatedNodeHdr->aFirstChar[nNumMatching+1], nNodeNumLeft + pUpdatedNodeHdr->nNumEntries - 1);
        }
        pUpdatedNodeHdr->nLog2Size = nUpdatedNodeLog2Size;
        pUpdatedNodeHdr->nParentOfs = nForkNodeOfs;
        pUpdatedNodeHdr->nStrLen = (u8)(nNodeNumLeft-1);
        *pRxi->pOfs = nForkNodeOfs;
        //pRxi->pNodeHdr = pForkNodeHdr;
        if ((unsigned)cNodeChar < (unsigned)cNameChar) {
            pForkNodeHdr->aFirstChar[nNumMatching] = cNodeChar;
            pForkNodeHdr->aFirstChar[nNumMatching+1] = cNameChar;
            pForkNodeOfsTblEnd[-1] = nUpdatedNodeOfs;
            pForkNodeOfsTblEnd[-2] = nNameNodeOfs;
            pRxi->pOfs = &pForkNodeOfsTblEnd[-2];
            if (pAttrOfs == 0) {
                WFSKrnOutputErrorStr("RxtInsert: Unexpected Fork");
            }
            pUpdatedNodeHdr->nEntryIdx = 0;
            if (pNameNodeHdr) {
                pNameNodeHdr->nEntryIdx = 1;
            }
        } else {
            pForkNodeHdr->aFirstChar[nNumMatching] = cNameChar;
            pForkNodeHdr->aFirstChar[nNumMatching+1] = cNodeChar;
            pForkNodeOfsTblEnd[-1] = nNameNodeOfs;
            pForkNodeOfsTblEnd[-2] = nUpdatedNodeOfs;
            if (pAttrOfs == 0) {
                pAttrOfs = &pForkNodeOfsTblEnd[-1];
                pAttrHdr->nParentOfs = nForkNodeOfs;
                pRxi->pNodeHdr = pForkNodeHdr;
                pRxi->pNodeStrPtr = &pForkNodeHdr->aFirstChar[pForkNodeHdr->nStrLen];
            } else {
                pRxi->pOfs = &pForkNodeOfsTblEnd[-1];
            }
            pUpdatedNodeHdr->nEntryIdx = 1;
            if (pNameNodeHdr) {
                pNameNodeHdr->nEntryIdx = 0;
            }
        }
    }
    ++pRxtHdr->nNumRecs;
    //pRxi->pOfs = pAttrOfs;
    pRxi->pStrPtr += nSuffixLen;
#if (_CHECK_RXT >= 2)
    RxtCheck(pRxtHdr, 0, 0); //WFS_LOG2_PATH_CACHE_SIZE, NULL);
#endif
    return WFSKRN_RESULT_OK;
}


static
u16 RxtMergeNodeWithItsSingleChild(RxtHdr *pRxtHdr, u16 nOldOfs) {
    //MyOSReport("RxtMergeNodeWithItsSingleChild: pRxtHdr = %x, nOldOfs = %x\n", pRxtHdr, nOldOfs);
#if _DEBUG_DIR
    if (nDirTest == nDirTestBreakPoint) {
        int a; a=0;
    }
#endif
    // A non-leaf node of the radix tree is not allowed to end up with exactly 1 child. This function can be used to merge it with the child.
    RxtNodeHdr *pOldNodeHdr = (RxtNodeHdr *)((u8*)pRxtHdr + nOldOfs);
    u32 nOldLog2Size = pOldNodeHdr->nLog2Size; //RxtGetNodeLog2Size(pOldNodeHdr);
    u16 *pOldNodeOfsTblEnd = (u16*)((u8*)pOldNodeHdr + (1<<nOldLog2Size));
    u32 nOldChildOfs = pOldNodeOfsTblEnd[-1];
    dbg(SbaCheckOfs(nOldChildOfs));
    RxtNodeHdr *pOldChildHdr = (RxtNodeHdr *)((u8*)pRxtHdr + nOldChildOfs);
    u32 nOldChildLog2Size = pOldChildHdr->nLog2Size;
    u32 nNewStrLen = (u32)(pOldNodeHdr->nStrLen + pOldChildHdr->nStrLen + 1);
    u32 nNewNodeLog2Size = RxtGetNodeLog2Size_2(nNewStrLen, pOldChildHdr->nNumEntries);
    if (nOldChildLog2Size == nNewNodeLog2Size) {
        memmove(&pOldChildHdr->aFirstChar[pOldNodeHdr->nStrLen+1], pOldChildHdr->aFirstChar, (u32)(pOldChildHdr->nStrLen + pOldChildHdr->nNumEntries));
        memcpy(pOldChildHdr->aFirstChar, pOldNodeHdr->aFirstChar, (u32)(pOldNodeHdr->nStrLen+1));
        pOldChildHdr->nStrLen = (u8)nNewStrLen;
        pOldChildHdr->nEntryIdx  = pOldNodeHdr->nEntryIdx;
        pOldChildHdr->nParentOfs = pOldNodeHdr->nParentOfs;
        if (pOldChildHdr->nParentOfs != 0) {
            RxtNodeHdr *pParentNodeHdr = (RxtNodeHdr *)((u8*)pRxtHdr + pOldChildHdr->nParentOfs);
            RxtSetChildOfs(pOldChildHdr->nEntryIdx+1, nOldChildOfs, pParentNodeHdr);
        } else {
            pRxtHdr->nRootOfs = nOldChildOfs;
        }
        SbaFreeSubBlk((WFSSubBlkAllocHdr *)pRxtHdr, nOldOfs, nOldLog2Size);
        return (u16)nOldChildOfs;
    }
    // The merged node will require more space than the original, so it must be reallocated
    u32 nNewNodeOfs = SbaAllocSubBlk((WFSSubBlkAllocHdr *)pRxtHdr, nNewNodeLog2Size);
    if (nNewNodeOfs == 0) {
        // Could not allocate a larger block
        return 0;
    }
    RxtNodeHdr *pNewNodeHdr = (RxtNodeHdr *)((u8*)pRxtHdr + nNewNodeOfs);
    u32 nCopySize1 = sizeof(RxtNodeHdr) + pOldNodeHdr->nStrLen;
    memcpy(pNewNodeHdr, pOldNodeHdr, nCopySize1);
    u32 nCopySize2 = pOldChildHdr->nStrLen + pOldChildHdr->nNumEntries;
    memcpy((u8*)pNewNodeHdr + nCopySize1, pOldChildHdr->aFirstChar, nCopySize2);
    pNewNodeHdr->nLog2Size = nNewNodeLog2Size;
    pNewNodeHdr->nStrLen = (u16)nNewStrLen;
    pNewNodeHdr->nNumEntries = pOldChildHdr->nNumEntries;
    s32 nNumEntries = (s32)pOldChildHdr->nNumEntries;
    u16 *pDstEnd = (u16*)((u8*)pNewNodeHdr + (1<<nNewNodeLog2Size));
    u16 *pDst = RxtGetChildOfsPtr(nNumEntries, pNewNodeHdr);
    u16 *pSrcStart = RxtGetChildOfsPtr(nNumEntries, pOldChildHdr);
    memcpy(pDst, pSrcStart, (u8*)pDstEnd - (u8*)pDst);
    while(pDst<pDstEnd) {
        RxtNodeHdr *pChildNodeHdr = (RxtNodeHdr *)((u8*)pRxtHdr + *pDst++);
        pChildNodeHdr->nParentOfs = nNewNodeOfs;
    }
    if (pOldChildHdr->nParentOfs != 0) {
        RxtNodeHdr *pParentNodeHdr = (RxtNodeHdr *)((u8*)pRxtHdr + pOldNodeHdr->nParentOfs);
        RxtSetChildOfs(pOldNodeHdr->nEntryIdx+1, nNewNodeOfs, pParentNodeHdr);
    } else {
        pRxtHdr->nRootOfs = nNewNodeOfs;
    }
    SbaFreeSubBlk((WFSSubBlkAllocHdr *)pRxtHdr, nOldChildOfs, nOldChildLog2Size);
    SbaFreeSubBlk((WFSSubBlkAllocHdr *)pRxtHdr, nOldOfs, nOldLog2Size);
    return (u16)nNewNodeOfs;
}


WFSKrnResult RxtDeleteNode(RxtHdr *pRxtHdr, RxtItr *pRxi, u32 nEntryIdx, u32 nNodeOfs, RxtNodeHdr *pNodeHdr) {
#if _DEBUG_BREAK_POINT
    static int c=0; c++; if(c==0xa842){
        int a; a=0;
    }
#endif
    //MyOSReport("RxtDeleteNode: pRxtHdr = %x, pRxi = %x, nEntryIdx = %d, nNodeOfs = %x, pNodeHdr = %x\n", pRxtHdr, pRxi, nEntryIdx, nNodeOfs, pNodeHdr);
    //MyOSReport("RxtDeleteNode: pRxi->pStrPtr= %s\n", pRxi->pStrPtr);
    u16 *pSrcStart = RxtGetChildOfsPtr(pNodeHdr->nNumEntries, pNodeHdr);
    // Delete the choice character from the original block
    utf8 *pSrc = &pNodeHdr->aFirstChar[pNodeHdr->nStrLen + nEntryIdx];
    memmove(pSrc-1, pSrc, pNodeHdr->nNumEntries - nEntryIdx);
    --pNodeHdr->nNumEntries;
    u32 nUpdatedNodeLog2Size = RxtGetNodeLog2Size(pNodeHdr);
    if (nUpdatedNodeLog2Size != pNodeHdr->nLog2Size) {
        u32 nUpdatedNodeOfs = SbaAllocSubBlk((WFSSubBlkAllocHdr *)pRxtHdr, nUpdatedNodeLog2Size);
        if (nUpdatedNodeOfs == 0) {
            // Could not allocate a new block, so resize in place
            u32 nOldLog2Size = pNodeHdr->nLog2Size;
            pNodeHdr->nLog2Size = nUpdatedNodeLog2Size;
            u16 *pDst = RxtGetChildOfsPtr(pNodeHdr->nNumEntries, pNodeHdr);
            u16 *pDstEnd = RxtGetChildOfsPtr(nEntryIdx-1, pNodeHdr);
            u32 nCopySize2 = (u8*)pDstEnd - (u8*)pDst;
            memmove(pDst, pSrcStart, nCopySize2);
            nEntryIdx = pNodeHdr->nNumEntries;
            while(pDst<pDstEnd) {
                RxtNodeHdr *pChildNodeHdr = (RxtNodeHdr *)((u8*)pRxtHdr + *pDst++);
                pChildNodeHdr->nEntryIdx  = --nEntryIdx;
            }
            if (nEntryIdx) {
                pSrcStart = (u16*)((u8*)pSrcStart + nCopySize2 + 2);
                memcpy(pDst, pSrcStart, sizeof(u16) * nEntryIdx);
            }
            SbaReduceSubBlkSize((WFSSubBlkAllocHdr *)pRxtHdr, nNodeOfs, nOldLog2Size, nUpdatedNodeLog2Size);
        } else {
            // Delete offset pointer while copying to a new sub block
            RxtNodeHdr *pUpdatedNodeHdr = (RxtNodeHdr *)((u8*)pRxtHdr + nUpdatedNodeOfs);
            u32 nCopySize1 = sizeof(RxtNodeHdr)-1 + pNodeHdr->nStrLen + pNodeHdr->nNumEntries;
            memcpy(pUpdatedNodeHdr, pNodeHdr, nCopySize1);
            pUpdatedNodeHdr->nLog2Size = nUpdatedNodeLog2Size;
            u16 *pDst = RxtGetChildOfsPtr(pNodeHdr->nNumEntries, pUpdatedNodeHdr);
            u16 *pDstEnd = RxtGetChildOfsPtr(nEntryIdx-1, pUpdatedNodeHdr);
            u32 nCopySize2 = (u8*)pDstEnd - (u8*)pDst;
            memcpy(pDst, pSrcStart, nCopySize2);
            nEntryIdx = pNodeHdr->nNumEntries;
            while(pDst<pDstEnd) {
                RxtNodeHdr *pChildNodeHdr = (RxtNodeHdr *)((u8*)pRxtHdr + *pDst++);
                pChildNodeHdr->nEntryIdx  = --nEntryIdx;
                pChildNodeHdr->nParentOfs = nUpdatedNodeOfs;
            }
            pSrcStart = (u16*)((u8*)pSrcStart + nCopySize2 + 2);
            memcpy(pDst, pSrcStart, sizeof(u16) * nEntryIdx);
            pDstEnd = pDst + nEntryIdx;
            while(pDst<pDstEnd) {
                RxtNodeHdr *pChildNodeHdr = (RxtNodeHdr *)((u8*)pRxtHdr + *pDst++);
                pChildNodeHdr->nParentOfs = nUpdatedNodeOfs;
            }
            // Update parent to point to the new node. Parent may be the nRootOfs of the block
            if (pUpdatedNodeHdr->nParentOfs != 0) {
                RxtNodeHdr *pParentNodeHdr = (RxtNodeHdr *)((u8*)pRxtHdr + pUpdatedNodeHdr->nParentOfs);
                RxtSetChildOfs(pUpdatedNodeHdr->nEntryIdx+1, nUpdatedNodeOfs, pParentNodeHdr);
            } else {
                pRxtHdr->nRootOfs = nUpdatedNodeOfs;
            }
            SbaFreeSubBlk((WFSSubBlkAllocHdr *)pRxtHdr, nNodeOfs, pNodeHdr->nLog2Size);
            pNodeHdr = pUpdatedNodeHdr;
            nNodeOfs = nUpdatedNodeOfs;
            pRxi->pNodeHdr = pUpdatedNodeHdr;
        }
    } else {
        // Delete entry from existing node
        u16 *pDst = RxtGetChildOfsPtr(pNodeHdr->nNumEntries, pNodeHdr);
        u16 *pDstEnd = RxtGetChildOfsPtr(nEntryIdx-1, pNodeHdr);
        memmove(pDst, pSrcStart, (u8*)pDstEnd - (u8*)pDst);
        nEntryIdx = pNodeHdr->nNumEntries;
        while(pDst<pDstEnd) {
            RxtNodeHdr *pChildNodeHdr = (RxtNodeHdr *)((u8*)pRxtHdr + *pDst++);
            pChildNodeHdr->nEntryIdx  = --nEntryIdx;
        }
    }
    if ((pNodeHdr->nNumEntries==1) && (pNodeHdr->aFirstChar[pNodeHdr->nStrLen] != 0)) {
        // The sub block has been left with one entry after deletion.
        // If it has a child pointer, it must be concatenated with the child.
        u16 nNewOfs = RxtMergeNodeWithItsSingleChild(pRxtHdr, nNodeOfs);
        if (nNewOfs == 0) {
            if (pRxtHdr->nReserveOfs) {
                SbaFreeSubBlk((WFSSubBlkAllocHdr *)pRxtHdr, pRxtHdr->nReserveOfs, RXT_LOG2_RESERVE_SIZE);
                pRxtHdr->nReserveOfs = 0;
                nNewOfs = RxtMergeNodeWithItsSingleChild(pRxtHdr, nNodeOfs);
            }
            if (nNewOfs == 0) {
                // WFSKrnOutputErrorCode(WFSKRN_RESULT_DIR_BLK_FULL);
                // ToDo : PathCache tree achieve the error reraly.
                //        PathCache have to be fixed to the efficient method of delete node and to prevent the error.
                MyOSReport("Output error WFSKRN_RESULT_DIR_BLK_FULL in RxtDeleteNode().\n");
                return WFSKRN_RESULT_DIR_BLK_FULL;
            }
        }
        pRxi->pNodeHdr = (RxtNodeHdr *)((u8*)pRxtHdr + nNewOfs);
    }
#if (_CHECK_RXT >= 2)
    RxtCheck(pRxtHdr, 0, 0); //WFS_LOG2_PATH_CACHE_SIZE, NULL);
#endif
    return WFSKRN_RESULT_OK;
}


WFSKrnResult RxtDelete(RxtHdr *pRxtHdr, RxtItr *pRxi) {
    //MyOSReport("RxtDelete: pRxtHdr = %x, pRxi = %x\n", pRxtHdr, pRxi);
    //MyOSReport("RxtDelete: pRxi->pStrPtr = %s\n", pRxi->pStrPtr);
    --pRxtHdr->nNumRecs;
    // Backtrack within leaf block
    RxtNodeHdr *pNodeHdr = pRxi->pNodeHdr;
    u32 nNodeOfs = (u8*)pNodeHdr - (u8*)pRxtHdr;
    u32 nParentOfs, nEntryIdx = 1; //pNodeHdr->nEntryIdx+1;
    if (pNodeHdr->nNumEntries==1) {
        // This is the only entry in the sub-block. Delete it, and delete the corresponding entry from its parent.
        nParentOfs = pNodeHdr->nParentOfs;
        nEntryIdx = pNodeHdr->nEntryIdx+1;
        SbaFreeSubBlk((WFSSubBlkAllocHdr *)pRxtHdr, nNodeOfs, pNodeHdr->nLog2Size);
        if (nNodeOfs == pRxtHdr->nRootOfs) {
            // This was the last sub block of the block
            // ToDo: Add in a NULL node
            return WFSKRN_RESULT_OK;
        }
        nNodeOfs = nParentOfs;
        pNodeHdr = (RxtNodeHdr *)((u8*)pRxtHdr + nNodeOfs);
    }
    return RxtDeleteNode(pRxtHdr, pRxi, nEntryIdx, nNodeOfs, pNodeHdr);
}

#if (_WIN32 && _DEBUG)

    #include "randomlib.h"

    typedef struct {
        u8               aRxtBuf[1<<RXT_MAX_LOG2_SIZE];
        RxtHdr          *pRxtHdr;
    } RxtGlobals;

    RxtGlobals rxtg;
    WFSPathName testPath;
    extern utf8 **aTestWords;

    extern int nNumTestWords;
    extern utf8 *pTestWordsData;

    void RxtQuickCheck(bool bOutput) {
        u32 nI=0;
        RxtItr rxi;
        WFSSetPathName(&testPath, "");
        rxi.pStrPtr = testPath.sStr;
        WFSKrnResult nResult = RxtFind(rxtg.pRxtHdr, &rxi, testPath.nLen);
        nResult = RxtNext(rxtg.pRxtHdr, &rxi);
        while(nResult == WFSKRN_RESULT_OK) {
            if (bOutput) {
                MyOSReport("%3d. %s\n", nI++, testPath.sStr);
            }
            nResult = RxtNext(rxtg.pRxtHdr, &rxi);
        }
        if (bOutput) {
            MyOSReport("\n");
        }
    }

#define RXT_TEST_MAX_ENTRIES 1000

    void RxtTest() {
        rxtg.pRxtHdr = (RxtHdr *)rxtg.aRxtBuf;
        for(u32 nI=0; nI<100000000; nI++) {
            RxtItr rxi;
            u32 nWordIdx;
            if ((nI%RXT_TEST_MAX_ENTRIES) == 0) {
                MyOSReport("%d\n", nI);
                RandomSetSeedU32(0x10000+(nI/RXT_TEST_MAX_ENTRIES));
                RxtInit(rxtg.pRxtHdr, RXT_MAX_LOG2_SIZE);
            }
            nWordIdx = RandomInt(0, nNumTestWords);
            rxi.pStrPtr = aTestWords[nWordIdx];
            u32 nStrLen = strlen(rxi.pStrPtr);
#if _DEBUG_BREAK_POINT
            if (nI==18841) {
                int a; a=0;
            }
#endif
            WFSKrnResult nResult = RxtFind(rxtg.pRxtHdr, &rxi, nStrLen);
            if (nResult == WFSKRN_RESULT_DIR_ENTRY_FOUND) {
                SbaFreeSubBlk((WFSSubBlkAllocHdr *)rxtg.pRxtHdr, RxtGetChildOfs(1, rxi.pNodeHdr), SBA_MIN_LOG2_SUB_BLK_SIZE);
                nResult = RxtDelete(rxtg.pRxtHdr, &rxi);
                if (nResult != WFSKRN_RESULT_OK) {
                    WFSKrnOutputErrorCode(nResult);
                }
                continue;
            }
            u32 nSuffixLen = nStrLen - (rxi.pStrPtr - aTestWords[nWordIdx]);
            u32 nAttrOfs = SbaAllocSubBlk((WFSSubBlkAllocHdr *)rxtg.pRxtHdr, SBA_MIN_LOG2_SUB_BLK_SIZE);
            RxtAttrHdr *pAttrHdr = (RxtAttrHdr *)((u8*)rxtg.pRxtHdr + nAttrOfs);
            pAttrHdr->nEntryIdx = 0;
            nResult = RxtInsert(rxtg.pRxtHdr, &rxi, nSuffixLen, pAttrHdr);
            RxtQuickCheck(nI==18840);
        }
    }

#endif
