/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfskrn_EPTree.cpp

  Copyright 2007-2008 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_EPTree.cpp,v $
  Revision 1.5  2008/12/18 08:53:11  ueno
  Refactoring to avoid recursive call.

  Revision 1.4  2008/12/17 05:06:49  ueno
  Cleanup.

  Revision 1.3  2008/12/15 04:14:40  ueno
  Cleanup.

  Revision 1.2  2008/12/15 02:40:07  ueno
  Added support for multi p-tree.

  Revision 1.1  2008/12/02 04:55:13  ueno
  Added EPTree to support multi-PTree.


*---------------------------------------------------------------------------*/
#include "wfskrn_Area.h"
#include "wfskrn_FTree.h"
#include "wfskrn_FreeBlkAlloc.h"

#define ROUNDUP32BYTE(val)   (((size_t)(val)+31)&~31)
#define ROUNDDOWN(val, unit) ((u32) (val) & ~((unit)-1))
#define ROUNDUP(val, unit)   ((u32) ((val) + ((unit)-1)) & ~((unit)-1))

#define FBA_PTREE_MIN_HEIGHT 1

typedef struct _EPTreeChildInfo
{
    EPTreeHdr* eptreeHdr;
    u32 address;
    u32 numEntries;
    u32 subTreeType;
    u32 firstKey;
} EPTreeChildInfo;

typedef struct _EPTreeSplitInfo
{
    EPTreeChildInfo right;
    EPTreeChildInfo left;
    u32 splitKey;
    u32 newChild;
    u32 totalNumRecs;
} EPTreeSplitInfo;

EPTreeItr* EPTreeGetLastIter(MEPTreeItr* iter)
{
    ASSERT(iter->eptreeHdr[0]);
    u8 height = iter->eptreeHdr[0]->height;
    EPTreeItr* last = &iter->epti[height-1];
    return last;
}

static inline void AllocatorInit(PTreeAllocator* allocator, EPTreeHdr* eptreeHdr)
{
    allocator->pBase = GET_BLOCK_HEAD(eptreeHdr);
    allocator->fpAllocNode = FTreeAllocEntry;
    allocator->fpAllocNodes = FTreeAllocEntries;
    allocator->fpFreeNode = FTreeFreeEntry;
}

static WFSKrnResult GetAndPinEPTreeHdr(AreaInfo* area,
    MEPTreeItr* iter, u32 level, WFSBlkAdr addr, u32 flags)
{
    WFSKrnResult result = WFSKRN_RESULT_OK;
    void* block;

    if (level == 0)
    {
        // root ptree block is already pinned.
        return WFSKRN_RESULT_OK;
    }

    if (iter->eptreeHdr[level] &&
        iter->eptreeHdr[level]->addr == addr)
    {
        return WFSKRN_RESULT_OK;
    }

    if (iter->eptreeHdr[level])
    {
        ASSERT(iter->eptreeHdr[level]->addr != 0);
        FBAUnpinCache(area, iter->eptreeHdr[level]->addr); // unpin old ptree block.
    }

    result = FBAGetCache(area, addr, flags, &block);
    if (result == WFSKRN_RESULT_OK)
    {
        iter->eptreeHdr[level] = GET_EPTREEHDR(block);
        ASSERT(iter->eptreeHdr[level]->addr == addr);
    }
    else
    {
        iter->eptreeHdr[level] = 0;
    }
    return result;
}

EPTreeHdr* EPTreeBlockInit(AreaInfo* area, void* block, WFSBlkAdr addr, u32 offset, u8 height)
{
    EPTreeHdr* eptreeHdr;
    FTreeSbaInit(area, block, sizeof(EPTreeHdr), offset);
    eptreeHdr = GET_EPTREEHDR(block);
    eptreeHdr->addr = addr;
    eptreeHdr->height = height;

    PTreeInit(&eptreeHdr->ptreeHdr, NULL); // 2nd argument is not used in PTreeInit().
    return eptreeHdr; // keep pinning.
};

/*

 * initial state.

[ToDo] update the figures.

 disk address (FBA_ROOT_BLK). (PLeaf in 8kB)
    +----------+
    |  sbaHdr  |
    +----------+
    | PTreeHdr |--->PLeaf
    +----------+    +---+---+---+---+
    |  PLeaf   |    | 0 | 0 | 0 | 0 |
    +----------+    +---+---+---+---+
    :          :    | 1 | ? | ? | ? |
    :          :    +---+---+---+---+
    |          |
    +----------+

 disk address (FBA_INITIAL_FTREE_BLK)
    +----------+
    |  sbaHdr  |
    +- - - - - +
    | FTreeHdr |
    +----------+
    :  empty   :
    :          :
    |          |
    +----------+

*/

WFSKrnResult EPTreeInit(AreaInfo* pAreaInfo, u32 offset)
{
    WFSKrnResult result;
    FTreeHdr* ftreeHdr;
    EPTreeHdr* eptreeHdr;
    void* cache;
    void* eptreeBlock;

    // FBA_ROOT_BLK is already pinned.

    eptreeBlock = (u8*) pAreaInfo->pAflh - offset; // FBA_ROOT_BLK is already pinned.
    eptreeHdr = EPTreeBlockInit(pAreaInfo, eptreeBlock, FBA_ROOT_BLK, ROUNDUP32BYTE(offset), FBA_PTREE_MIN_HEIGHT);

    // create an F-tree block.
    result = FBAGetCache(pAreaInfo, FBA_INITIAL_FTREE_BLK, BCACHE_FLAG_PINNED | BCACHE_FLAG_DIRTY, &cache);

    if (result == WFSKRN_RESULT_OK)
    {
        PTreeAllocator allocator;
        EPTreeItr epti;

        ftreeHdr = FTreeBlockInit(pAreaInfo, cache);
        BCacheUnpin(pAreaInfo->pVolInfo, pAreaInfo->nAbsStartBlkAdr + FBA_INITIAL_FTREE_BLK);

        // +---+---+---+---+
        // | 0 | 0 | 0 | 0 |
        // +---+---+---+---+
        // | 1 | ? | ? | ? |
        // +---+---+---+---+
        //   A
        //   |
        // FBA_INITIAL_FTREE_BLK

        AllocatorInit(&allocator, eptreeHdr);
        result = PTreeInsert(&eptreeHdr->ptreeHdr, &epti, 0, FBA_INITIAL_FTREE_BLK, &allocator);
        WFSKRN_FTREE_ASSERT(result == WFSKRN_RESULT_OK);
    }

    // the EP-tree block is pinned.
    return result;
}

WFSKrnResult EPTreeReinit(AreaInfo* pAreaInfo, u32 offset)
{
    WFSKrnResult result;
    EPTreeHdr* eptreeHdr;
    void* eptreeBlock;

    // FBA_ROOT_BLK is already pinned.

    eptreeBlock = (u8*) pAreaInfo->pAflh - offset; // FBA_ROOT_BLK is already pinned.
    eptreeHdr = EPTreeBlockInit(pAreaInfo, eptreeBlock, FBA_ROOT_BLK, ROUNDUP32BYTE(offset), FBA_PTREE_MIN_HEIGHT);

    PTreeAllocator allocator;
    EPTreeItr epti;

    // +---+---+---+---+
    // | 0 | 0 | 0 | 0 |
    // +---+---+---+---+
    // | 1 | ? | ? | ? |
    // +---+---+---+---+
    //   A
    //   |
    // FBA_INITIAL_FTREE_BLK

    // FBA_INITIAL_FTREE_BLK is not initialized, because it may include some freeblocks.

    AllocatorInit(&allocator, eptreeHdr);
    result = PTreeInsert(&eptreeHdr->ptreeHdr, &epti, 0, FBA_INITIAL_FTREE_BLK, &allocator);
    WFSKRN_FTREE_ASSERT(result == WFSKRN_RESULT_OK);

    // the EP-tree block is pinned.
    return result;
}

// determine the threshold block address.
static u32 GetSplitAddress(PTreeHdr* ptreeHdr, EPTreeItr* iter, PTreeAllocator* allocator)
{
    WFSKrnResult result;
    u32 halfNumRecs = ptreeHdr->nNumRecs / 2;

    result = PTreeFirst(ptreeHdr, iter, allocator);
    while (halfNumRecs--)
    {
        result = PTreeNext(ptreeHdr, iter, allocator);
        ASSERT(result == WFSKRN_RESULT_OK);
    }

    return PTREE_GET_KEY(iter);
}

static WFSKrnResult GetNewControlBlock(AreaInfo* area, WFSBlkAdr nNGBlkAdr1, WFSBlkAdr nNGBlkAdr2, WFSBlkAdr nRefBlkAdr, FTreeExtent* extent, u32* subTreeIdx)
{
    WFSKrnResult result = WFSKRN_RESULT_NOT_FOUND;
    FBATreeItr iter;
    FBAItrOpen(area, &iter);

    u32 idx;
    for (idx = FTREE_TYPE_A; idx < MAX_SUB_TREE; ++idx)
    {
        result = FreeBlkFind(area, &iter, nRefBlkAdr, FTreeBlkLog2Ofs[idx]);
        while ((result == WFSKRN_RESULT_OK) &&
               (nNGBlkAdr1 == FBA_RESULT_KEY(&iter) ||
                nNGBlkAdr2 == FBA_RESULT_KEY(&iter)))
        {
            result = FreeBlkPrev(area, &iter);
        }
        if (result == WFSKRN_RESULT_OK)
        {
            extent->start = FBA_RESULT_KEY(&iter);
            extent->len = (FBA_RESULT_DATA(&iter) + 1) * (1<<FTreeBlkLog2Ofs[idx]);
            *subTreeIdx = idx;
            break;
        }
    }

    FBAItrClose(area, &iter);
    return result;
}

static EPTreeHdr* ControlBlockInit(AreaInfo* area, u32 addr, u8 height)
{
    void* pBlock;
    EPTreeHdr* eptreeHdr;
    WFSKrnResult result;

    result = FBAGetCache(area, addr, BCACHE_FLAG_DIRTY | BCACHE_FLAG_PINNED, &pBlock);
    ASSERT(result == WFSKRN_RESULT_OK);

    u32 offset = (sizeof(WFSHashCode) + sizeof(WFSMetaDataHdr));
    EPTreeBlockInit(area, pBlock, addr, offset, height);

    eptreeHdr = GET_EPTREEHDR(pBlock);
    eptreeHdr->addr = addr;
    return eptreeHdr;
}

// allocate 2 blocks. (must not remove them from FBA tree, because it may cause another p-tree splitting.)
static WFSKrnResult SplitTreeInit(AreaInfo* area, EPTreeHdr* eptreeHdr, EPTreeSplitInfo* info, PTreeAllocator* allocator)
{
    EPTreeItr iter;
    WFSKrnResult result;
    FTreeExtent newCb;
    u32 subTreeIdxNewCb;

    // choose split key.
    info->splitKey = GetSplitAddress(&eptreeHdr->ptreeHdr, &iter, allocator);

    result = GetNewControlBlock(area, 0 /* no NG address */, 0 /* no NG address */,
        info->splitKey, &newCb, &subTreeIdxNewCb);
    ASSERT(result == WFSKRN_RESULT_OK);

    info->left.address = newCb.start;
    info->left.subTreeType = subTreeIdxNewCb;

    if (newCb.len == 1)
    {
        result = GetNewControlBlock(area, info->left.address, 0, newCb.start + 1, &newCb, &subTreeIdxNewCb);
        ASSERT(result == WFSKRN_RESULT_OK);
        info->right.address = newCb.start;
    }
    else
    {
        info->right.address = info->left.address + 1;
    }
    info->right.subTreeType = subTreeIdxNewCb;

    // height field is only used by root ptree block.
    info->left.eptreeHdr = ControlBlockInit(area, info->left.address, eptreeHdr->height);
    info->right.eptreeHdr = ControlBlockInit(area, info->right.address, eptreeHdr->height);

    return WFSKRN_RESULT_OK;
}

static WFSKrnResult CreateChildBlocks(AreaInfo* area, EPTreeSplitInfo* info, PTreeHdr* ptreeHdr, PTreeAllocator* allocatorParent)
{
    WFSKrnResult result;
    PTreeAllocator allocator;
    PTreeItr ptiSearch;
    PTreeItr ptiInsert;

    // insert to left child.
    result = PTreeFirst(ptreeHdr, &ptiSearch, allocatorParent);
    ASSERT(result == WFSKRN_RESULT_OK);

    AllocatorInit(&allocator, info->left.eptreeHdr); // [ToDo] single ptree.

    info->left.firstKey = PTREE_GET_KEY(&ptiSearch);
    int count = ptreeHdr->nNumRecs;
    while ((int) (ptreeHdr->nNumRecs - ptreeHdr->nNumRecs / 2) < count)
    {
        ASSERT(PTREE_GET_KEY(&ptiSearch) < info->splitKey);
        result = PTreeInsert(&info->left.eptreeHdr->ptreeHdr, &ptiInsert,
            PTREE_GET_KEY(&ptiSearch), PTREE_GET_DATA(&ptiSearch), &allocator);
        if (result != WFSKRN_RESULT_OK)
        {
            ASSERT(result == WFSKRN_RESULT_OK);
            goto ERROR_CREATECHILDBLOCKS;
        }

        result = PTreeNext(ptreeHdr, &ptiSearch, allocatorParent);
        if (result != WFSKRN_RESULT_OK)
        {
            ASSERT(result == WFSKRN_RESULT_OK);
            goto ERROR_CREATECHILDBLOCKS;
        }
        --count;
    }

    ASSERT(result == WFSKRN_RESULT_OK);

    // insert to right child.
    info->right.firstKey = PTREE_GET_KEY(&ptiSearch);
    AllocatorInit(&allocator, info->right.eptreeHdr); // [ToDo] single ptree.
    while (0 < count)
    {
        ASSERT(info->splitKey <= PTREE_GET_KEY(&ptiSearch));
        result = PTreeInsert(&info->right.eptreeHdr->ptreeHdr, &ptiInsert,
            PTREE_GET_KEY(&ptiSearch), PTREE_GET_DATA(&ptiSearch), &allocator);
        if (result != WFSKRN_RESULT_OK)
        {
            ASSERT(result == WFSKRN_RESULT_OK);
            goto ERROR_CREATECHILDBLOCKS;
        }

        if (--count == 0)
        {
            break;
        }

        ASSERT(result == WFSKRN_RESULT_OK);
        result = PTreeNext(ptreeHdr, &ptiSearch, allocatorParent);
        if (result != WFSKRN_RESULT_OK)
        {
            ASSERT(result == WFSKRN_RESULT_OK);
            goto ERROR_CREATECHILDBLOCKS;
        }
    }

ERROR_CREATECHILDBLOCKS:
    return result;
}

static inline WFSBool IsRootBlock(AreaInfo* area, EPTreeHdr* eptreeHdr)
{
    FBATreeHdr* fbaTreeHdr = GET_FBATREE_HDR(area);
    return ((u32) (eptreeHdr->height) == fbaTreeHdr->ptreeHeight);
}

static WFSKrnResult InsertToParentBlocks(AreaInfo* area, EPTreeHdr* eptreeHdr, EPTreeSplitInfo* info)
{
    WFSKrnResult result;
    EPTreeHdr* tempEptreeHdr;
    PTreeItr pti;
    PTreeAllocator allocator;

    if (IsRootBlock(area, eptreeHdr) == (WFSBool) TRUE)
    {
        u32 offset;
        u32 subTreeIdxTempCb;
        u32 originalAddr;
        FBATreeHdr* fbaTreeHdr = GET_FBATREE_HDR(area);
        FTreeExtent tempCb;

        // create a new level.
        ++fbaTreeHdr->ptreeHeight;
        ASSERT(fbaTreeHdr->ptreeHeight <= EPTREE_MAX_HEIGHT);


        result = GetNewControlBlock(area, info->left.address, info->right.address,
            1, &tempCb, &subTreeIdxTempCb);
        ASSERT(result == WFSKRN_RESULT_OK);

        tempEptreeHdr = ControlBlockInit(area, tempCb.start, (u8) (fbaTreeHdr->ptreeHeight));

        // copy original parent ptree.
        memmove((u8*) GET_BLOCK_FROM_EPTREEHDR(tempEptreeHdr), GET_BLOCK_FROM_EPTREEHDR(eptreeHdr), area->nBlkSize);

        // clear all entries
        offset = eptreeHdr->sbaHdr.offset;
        EPTreeBlockInit(area, GET_BLOCK_FROM_EPTREEHDR(tempEptreeHdr), tempCb.start, offset, (u8) (fbaTreeHdr->ptreeHeight));

        // insert child blocks into the parent.
        AllocatorInit(&allocator, tempEptreeHdr);

        result = PTreeInsert(&tempEptreeHdr->ptreeHdr, &pti, info->left.firstKey, info->left.address, &allocator);
        ASSERT(result == WFSKRN_RESULT_OK);

        result = PTreeInsert(&tempEptreeHdr->ptreeHdr, &pti, info->right.firstKey, info->right.address, &allocator);
        ASSERT(result == WFSKRN_RESULT_OK);

        originalAddr = eptreeHdr->addr;
        memmove(GET_BLOCK_FROM_EPTREEHDR(eptreeHdr), GET_BLOCK_FROM_EPTREEHDR(tempEptreeHdr), area->nBlkSize); // remap
        eptreeHdr->addr = originalAddr;
        FBAFreeCache(area, tempCb.start); // unpin the temporary parent block.
    }
    else
    {
        // insert the split key into the parent block.
        AllocatorInit(&allocator, eptreeHdr);
        result = PTreeInsert(&eptreeHdr->ptreeHdr, &pti, info->splitKey, 1, &allocator);
    }

    // remove left and right child blocks from the FBA tree.
    return result;
}

static WFSKrnResult EPTreeSplit(AreaInfo* area, MEPTreeItr* iter, EPTreeHdr* eptreeHdr, EPTreeSplitInfo* info)
{
    WFSKrnResult result;
    PTreeAllocator allocator;

    if (EPTREE_MAX_HEIGHT == iter->eptreeHdr[0]->height)
    {
        return WFSKRN_RESULT_RESOURCE_LIMIT_EXCEEDED;
    }

    AllocatorInit(&allocator, eptreeHdr);

    result = SplitTreeInit(area, eptreeHdr, info, &allocator);
    if (result != WFSKRN_RESULT_OK)
    {
        return result;
    }

    // split the original block to two children.
    result = CreateChildBlocks(area, info, &eptreeHdr->ptreeHdr, &allocator);

    FBAUnpinCache(area, info->left.address);
    FBAUnpinCache(area, info->right.address);

    if (result != WFSKRN_RESULT_OK)
    {
        return result;
    }

    // create new parent block and remap it to the original block.
    result = InsertToParentBlocks(area, eptreeHdr, info);
    if (result == WFSKRN_RESULT_OK)
    {
        FBATreeItr iterDelete;

        FBAItrOpen(area, &iterDelete);
        iterDelete.fti.subTree = FTreeSelectSubTree(iterDelete.ftreeHdr, info->left.subTreeType);

        // remove child block address from the FBA tree.
        result = FreeBlkDelete(area, &iterDelete,
            info->left.address, 1, FTreeBlkLog2Ofs[info->left.subTreeType]);
        ASSERT(result == WFSKRN_RESULT_OK);

        iterDelete.fti.subTree = FTreeSelectSubTree(iterDelete.ftreeHdr, info->right.subTreeType);
        result = FreeBlkDelete(area, &iterDelete,
            info->right.address, 1, FTreeBlkLog2Ofs[info->right.subTreeType]);
        ASSERT(result == WFSKRN_RESULT_OK);

        FBAItrClose(area, &iterDelete);
    }

    return result;
}

WFSKrnResult EPTreeFirst(AreaInfo* area, MEPTreeItr* iter)
{
    WFSKrnResult result;
    PTreeAllocator allocator;
    int level;
    u8 ptreeHeight;

    AllocatorInit(&allocator, iter->eptreeHdr[0]);
    result = PTreeFirst(&iter->eptreeHdr[0]->ptreeHdr, (PTreeItr*) &iter->epti[0], &allocator);

    ptreeHeight = iter->eptreeHdr[0]->height;
    for (level = 1; level < ptreeHeight; ++level)
    {
        if (result != WFSKRN_RESULT_OK)
        {
            break;
        }

        result = GetAndPinEPTreeHdr(area, iter, level,
            EPTREE_GET_DATA(&iter->epti[level-1]), BCACHE_FLAG_READ | BCACHE_FLAG_PINNED);

        if (result != WFSKRN_RESULT_OK)
        {
            break;
        }

        AllocatorInit(&allocator, iter->eptreeHdr[level]);
        result = PTreeFirst(&iter->eptreeHdr[level]->ptreeHdr, (PTreeItr*) &iter->epti[level], &allocator);
    }

    return result;
}

WFSKrnResult EPTreeLast(AreaInfo* area, MEPTreeItr* iter)
{
    WFSKrnResult result;
    PTreeAllocator allocator;
    int level;
    u8 ptreeHeight;

    AllocatorInit(&allocator, iter->eptreeHdr[0]);
    result = PTreeLast(&iter->eptreeHdr[0]->ptreeHdr, (PTreeItr*) &iter->epti[0], &allocator);

    ptreeHeight = iter->eptreeHdr[0]->height;
    for (level = 1; level < ptreeHeight; ++level)
    {
        if (result != WFSKRN_RESULT_OK)
        {
            break;
        }

        result = GetAndPinEPTreeHdr(area, iter, level,
            EPTREE_GET_DATA(&iter->epti[level-1]), BCACHE_FLAG_READ | BCACHE_FLAG_PINNED);

        if (result != WFSKRN_RESULT_OK)
        {
            break;
        }

        AllocatorInit(&allocator, iter->eptreeHdr[level]);
        result = PTreeLast(&iter->eptreeHdr[level]->ptreeHdr, (PTreeItr*) &iter->epti[level], &allocator);
    }

    return result;
}

WFSKrnResult EPTreeNext(AreaInfo* area, MEPTreeItr* iter)
{
    WFSKrnResult result = WFSKRN_RESULT_PTREE_ENTRY_NOT_FOUND;
    PTreeAllocator allocator;
    EPTreeHdr* eptreeHdr;
    u8 ptreeHeight = iter->eptreeHdr[0]->height;

    eptreeHdr = iter->eptreeHdr[ptreeHeight-1];

    AllocatorInit(&allocator, eptreeHdr);
    result = PTreeNext(&eptreeHdr->ptreeHdr, (PTreeItr*) &iter->epti[ptreeHeight-1], &allocator);
    if (result == WFSKRN_RESULT_PTREE_ENTRY_FOUND)
    {
        return result;
    }

    // go back and lookup an entry in the next node.
    int level = ptreeHeight;
    while (result == WFSKRN_RESULT_PTREE_ENTRY_NOT_FOUND && --level)
    {
        eptreeHdr = iter->eptreeHdr[level-1];
        AllocatorInit(&allocator, eptreeHdr);
        result = PTreeNext(&eptreeHdr->ptreeHdr, (PTreeItr*) &iter->epti[level-1], &allocator);
    }

    if (result == WFSKRN_RESULT_OK)
    {
        for (;level < ptreeHeight; ++level)
        {
            result = GetAndPinEPTreeHdr(area, iter, level,
                EPTREE_GET_DATA(&iter->epti[level-1]), BCACHE_FLAG_READ | BCACHE_FLAG_PINNED);
            if (result != WFSKRN_RESULT_OK)
            {
                break;
            }
            eptreeHdr = iter->eptreeHdr[level];
            AllocatorInit(&allocator, eptreeHdr);
            result = PTreeFirst(&eptreeHdr->ptreeHdr, (PTreeItr*) &iter->epti[level], &allocator);
            if (result != WFSKRN_RESULT_OK)
            {
                break;
            }
        }
    }

    return result;
}

WFSKrnResult EPTreePrev(AreaInfo* area, MEPTreeItr* iter)
{
    WFSKrnResult result = WFSKRN_RESULT_PTREE_ENTRY_NOT_FOUND;
    PTreeAllocator allocator;
    EPTreeHdr* eptreeHdr;
    u8 ptreeHeight = iter->eptreeHdr[0]->height;

    eptreeHdr = iter->eptreeHdr[ptreeHeight-1];

    AllocatorInit(&allocator, eptreeHdr);
    result = PTreePrev(&eptreeHdr->ptreeHdr, (PTreeItr*) &iter->epti[ptreeHeight-1], &allocator);
    if (result == WFSKRN_RESULT_PTREE_ENTRY_FOUND)
    {
        return result;
    }

    // go back and lookup an entry in the next node.
    int level = ptreeHeight;
    while (result == WFSKRN_RESULT_PTREE_ENTRY_NOT_FOUND && --level)
    {
        eptreeHdr = iter->eptreeHdr[level-1];
        AllocatorInit(&allocator, eptreeHdr);
        result = PTreePrev(&eptreeHdr->ptreeHdr, (PTreeItr*) &iter->epti[level-1], &allocator);
    }

    if (result == WFSKRN_RESULT_OK)
    {
        for (;level < ptreeHeight; ++level)
        {
            result = GetAndPinEPTreeHdr(area, iter, level,
                EPTREE_GET_DATA(&iter->epti[level-1]), BCACHE_FLAG_READ | BCACHE_FLAG_PINNED);
            if (result != WFSKRN_RESULT_OK)
            {
                break;
            }
            eptreeHdr = iter->eptreeHdr[level];
            AllocatorInit(&allocator, eptreeHdr);
            result = PTreeLast(&eptreeHdr->ptreeHdr, (PTreeItr*) &iter->epti[level], &allocator);
            if (result != WFSKRN_RESULT_OK)
            {
                break;
            }
        }
    }

    return result;
}

static WFSKrnResult EPTreeInsertWithoutSplitting(AreaInfo* area, MEPTreeItr* iter, u32 key, u32 data)
{
    WFSKrnResult result;
    EPTreeHdr* eptreeHdr;
    EPTreeItr* epti;
    PTreeAllocator allocator;
    u8 height;

    result = EPTreeFind(area, iter, key);
    if (result == WFSKRN_RESULT_OK)
    {
        return WFSKRN_RESULT_ALREADY_EXISTS;
    }

    if (result != WFSKRN_RESULT_PTREE_ENTRY_NOT_FOUND)
    {
        return result;
    }

    height = iter->eptreeHdr[0]->height;
    epti = &iter->epti[height-1];
    eptreeHdr = iter->eptreeHdr[height-1];

    AllocatorInit(&allocator, eptreeHdr);
    result = PTreeInsert(&eptreeHdr->ptreeHdr, (PTreeItr*) epti, key, data, &allocator);
    return result;
}

WFSKrnResult EPTreeInsert(AreaInfo* area, MEPTreeItr* iter, u32 key, u32 data)
{
    WFSKrnResult result;
    while (WFSKRN_RESULT_PTREE_FULL == (result = EPTreeInsertWithoutSplitting(area, iter, key, data)))
    {
        EPTreeSplitInfo info;
        result = EPTreeSplit(area, (MEPTreeItr*) iter, iter->eptreeHdr[0], &info);
        if (result != WFSKRN_RESULT_OK)
        {
#ifdef _DEBUG
            OS_TPrintf("*** EPTreeInsert(): split error (%d) ***\n", result);
#endif // _DEBUG
            ASSERT(result == WFSKRN_RESULT_OK ||
                   result == WFSKRN_RESULT_DIRECTORY_QUOTA);
            break;
        }
    }
    return result;
}

static WFSKrnResult DelteEmptyBlocks(AreaInfo* area, MEPTreeItr* iter, WFSBlkAdr* freeBlks)
{
    WFSKrnResult result = WFSKRN_RESULT_OK;
    EPTreeHdr* eptreeHdr;
    EPTreeItr* epti;
    PTreeAllocator allocator;

    WFSBlkAdr addr;
    int level;
    int height = (int) iter->eptreeHdr[0]->height;
    for (level = height - 2; 0 <= level; --level)
    {
        eptreeHdr = iter->eptreeHdr[level];

        epti = &iter->epti[level];
        addr = EPTREE_GET_DATA(epti);
        AllocatorInit(&allocator, eptreeHdr);
        result = PTreeDelete(&eptreeHdr->ptreeHdr, (PTreeItr*) epti, &allocator);
        if (result != WFSKRN_RESULT_OK)
        {
            break;
        }
        iter->eptreeHdr[level+1] = 0;
        FBAFreeCache(area, addr);
        freeBlks[level] = addr; // save to insert it into FBA tree, later.
        if (0 < eptreeHdr->ptreeHdr.nNumRecs)
        {
            break;
        }
    }
    return result;
}

WFSKrnResult EPTreeShrink(AreaInfo* area, MEPTreeItr* iter, WFSBlkAdr* freeBlks)
{
    WFSKrnResult result = EPTreeLast(area, iter);
    if (result == WFSKRN_RESULT_OK)
    {
        u32 height = (u32) iter->eptreeHdr[0]->height;
        EPTreeItr* epti = EPTreeGetLastIter(iter);
        if (EPTREE_GET_DATA(epti) == FBA_INITIAL_FTREE_BLK &&
            EPTREE_GET_KEY(epti) == 0)
        {
            MEPTreeItr iterInsert;
            iterInsert.eptreeHdr[0] = GET_EPTREEHDR_FROM_AREAINFO(area);
            u32 level;
            for (level = 1; level < EPTREE_MAX_HEIGHT; ++level)
            {
                iterInsert.eptreeHdr[level] = 0;
            }

            FBATreeHdr* fbaTreeHdr = GET_FBATREE_HDR(area);
            fbaTreeHdr->ptreeHeight = 1;
            u32 offset = area->nBlkSize - GET_FBATREE_HDR(area)->ofsEPtreeHdr - sizeof(EPTreeHdr);
            EPTreeReinit(area, ROUNDUP32BYTE(offset));

            for (level = 0; level < (int) height; ++level)
            {
                u32 addr = iter->epti[level].nData;
                if (addr != FBA_INITIAL_FTREE_BLK)
                {
                    FBAFreeCache(area, addr);
                    freeBlks[level + (EPTREE_MAX_HEIGHT-1)] = addr;
                }
            }
        }
    }
    return result;
}

WFSKrnResult EPTreeDelete(AreaInfo* area, MEPTreeItr* iter, u32 key, WFSBlkAdr* freeBlks)
{
    WFSKrnResult result;
    EPTreeHdr* eptreeHdr;
    PTreeAllocator allocator;

    result = EPTreeFind(area, iter, key);
    if (result == WFSKRN_RESULT_OK ||
        result == WFSKRN_RESULT_PTREE_ENTRY_NOT_FOUND) // not found?
    {
        u8 height = iter->eptreeHdr[0]->height;
        EPTreeItr* epti = &iter->epti[height-1];
        eptreeHdr = iter->eptreeHdr[height-1];

        AllocatorInit(&allocator, eptreeHdr);
        result = PTreeDelete(&eptreeHdr->ptreeHdr, (PTreeItr*) epti, &allocator);
        if (eptreeHdr->ptreeHdr.nNumRecs <= 1 && 2 <= height)
        {
            if (eptreeHdr->ptreeHdr.nNumRecs == 0)
            {
                result = DelteEmptyBlocks(area, iter, freeBlks);
                if (result == WFSKRN_RESULT_OK)
                {
                    result = EPTreeShrink(area, iter, freeBlks);
                }
            }
            else if (epti->pLeaf->aKey[0] == 0 &&
                     epti->pLeaf->aData[0] == FBA_INITIAL_FTREE_BLK)
            {
                // eptreeHdr->ptreeHdr.nNumRecs == 1.
                result = EPTreeShrink(area, iter, freeBlks);
            }
        }
    }
    return result;
}

// search a P-tree by key.
/*
                P-tree node
             +-----------------+
             |      P-Tree     |
             |key:    100      |
             |                 |
             +-----------------+
                 |           |
    p-tree leaf  V           V  p-tree leaf
  +-----------------+     +-----------------+
  |      P-Tree     |     |      P-Tree     |
  |key: 80 96       |     |key: 110 116 119 |
  |                 |     |                 |
  +-----------------+     +-----------------+
                       A         A     A      A
                       |         |     |      |
                      (1)       (2)   (3)    (4)

     key       result                    pPti->nKey
 (1) 101   =>  WFSKRN_RESULT_NOT_FOUND    -  (no value is set).
 (2) 110   =>  WFSKRN_RESULT_OK          110
 (3) 118   =>  WFSKRN_RESULT_OK          116
 (4) 120   =>  WFSKRN_RESULT_OK          119
*/
WFSKrnResult EPTreeFind(AreaInfo* area, MEPTreeItr* iter, u32 key)
{
    WFSKrnResult result = WFSKRN_RESULT_PTREE_ENTRY_NOT_FOUND;
    PTreeAllocator allocator;
    EPTreeHdr* eptreeHdr;
    EPTreeItr* epti;
    int level;
    u8 ptreeHeight = iter->eptreeHdr[0]->height;

    for (level = 0; level < ptreeHeight; ++level)
    {
        eptreeHdr = iter->eptreeHdr[level];
        AllocatorInit(&allocator, eptreeHdr);
        result = PTreeFind(&eptreeHdr->ptreeHdr, (PTreeItr*) &iter->epti[level], key, &allocator);

        epti = &iter->epti[level];
        if (result == WFSKRN_RESULT_PTREE_ENTRY_NOT_FOUND)
        {
            if (iter->eptreeHdr[level]->ptreeHdr.nNumRecs == 0)
            {
                result = WFSKRN_RESULT_PTREE_EMPTY;
                goto RESULT_EPTREEFIND;
            }
            else if (epti->aEntryIdx[0] == 0)
            {
                // the key is smaller than the minimum key in this P-tree.
                result = EPTreePrev(area, iter);
                if (result != WFSKRN_RESULT_OK)
                {
                    result = WFSKRN_RESULT_NOT_FOUND;
                    goto RESULT_EPTREEFIND;
                }
                if (key == EPTREE_GET_KEY(epti))
                {
                    result = WFSKRN_RESULT_PTREE_ENTRY_FOUND;
                }
                else
                {
                    result = WFSKRN_RESULT_PTREE_ENTRY_NOT_FOUND;
                }
            }
            else
            {
                // In this case, PTreeFind() saved the nearest key.
                --epti->aEntryIdx[0];
            }
        }

        if (result != WFSKRN_RESULT_PTREE_ENTRY_FOUND &&
            result != WFSKRN_RESULT_PTREE_ENTRY_NOT_FOUND)
        {
            break;
        }

        epti->nKey = epti->pLeaf->aKey[epti->aEntryIdx[0]];
        epti->nData = epti->pLeaf->aData[epti->aEntryIdx[0]];

        if (level == ptreeHeight - 1)
        {
            // the bottom of multi EP-Tree.
            break;
        }

        result = GetAndPinEPTreeHdr(area, iter, level + 1, EPTREE_GET_DATA(epti), BCACHE_FLAG_READ | BCACHE_FLAG_PINNED);
        if (result != WFSKRN_RESULT_OK)
        {
            break;
        }
    }

    if (result == WFSKRN_RESULT_PTREE_ENTRY_FOUND)
    {
        result = WFSKRN_RESULT_OK;
    }

RESULT_EPTREEFIND:
    return result;
}

