/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfskrn_FTree.cpp

  Copyright 2007-2008 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_FTree.cpp,v $
  Revision 1.20  2008/12/18 08:37:04  ueno
  Cleanup.

  Revision 1.19  2008/12/18 00:31:27  ueno
  Moved FTreeBlkLog2Ofs{} for IOP.

  Revision 1.18  2008/12/17 05:06:49  ueno
  Cleanup.

  Revision 1.17  2008/12/15 00:57:39  ueno
  Modified to prevent from overwriting a new F-tree block.

  Revision 1.16  2008/10/14 10:27:14  kondo_masahiro
  Added #define _DEBUG_BREAK_POINT

  Revision 1.15  2008/10/09 12:51:03  ueno
  Fixed FTreeInsert().

  Revision 1.14  2008/10/03 08:37:56  kondo_masahiro
  Fixed codes for porting IOP

  Revision 1.12  2008/09/30 05:55:52  ueno
  Cleanup.

  Revision 1.10  2008/09/29 10:19:36  ueno
  Revised to support multi-B-tree-based allocator.

  Revision 1.9  2008/09/29 01:57:23  ueno
  Modified RemapControlBlocks() not to copy hash and flags.

  Revision 1.8  2008/09/28 23:32:04  kondo_masahiro
  Fixed codes to use medium size user block.

  Revision 1.17  2008/09/29 01:58:57  ueno
  Modified to call BCacheInvalidateBlks() in FreeBlkFree() to invalidate caches of user data.

  Revision 1.7  2008/09/02 07:00:45  kondo_masahiro
  Fixed VerifyMemoryBlock -> VerifyMemoryBlockFTree

  Revision 1.6  2008/08/09 03:32:51  ueno
  Disabled runtime check.

  Revision 1.5  2008/08/09 02:01:41  ueno
  Cleanup.

  Revision 1.4  2008/08/07 10:46:33  ueno
  Modified for BTREE_BASED_ALLOCATOR (now debugging).

  Revision 1.3  2008/08/02 11:52:40  ueno
  Revised the positions of FTreeHdr{} and FBAPTreeHdr{} in 8KB-blocks to protect hash code.

  Revision 1.2  2008/08/01 07:00:05  ueno
  Added copyrights.

  Revision 1.1  2008/08/01 06:57:39  ueno
  Initial Check-in.

*---------------------------------------------------------------------------*/
#include "wfskrn_Area.h"
#if BTREE_BASED_ALLOCATOR

#include "wfskrn_FreeBlkAlloc.h"
#include "wfskrn_FTree.h"

#ifdef FBA_DEBUG_VERBOSE
extern WFSBool FBADebugVerbose;
#endif // FBA_DEBUG_VERBOSE

#define ROUNDUP32BYTE(val) (((size_t)val+31)&~31)
#define IS_OUT_OF_RANGE(pLeaf, key)  ((key) < (u32*) (pLeaf))

typedef struct _ChildInfo
{
    FTreeHdr* ftreeHdr;
    u32 address;
    u32 numEntries[PTREE_MAX_HEIGHT+1];
    u32 totalEntries; // leaves and nodes.
    u32 firstEntry;   // byte offset of the first leaf (or node).
    u32 firstKey;     // the minimum key in the tree.
    u32 nRootHeight;
    u16 nNumRecs;     // the number of keys in the tree.
    u16 nRootOfs;
} ChildInfo;

typedef struct _SplitInfo
{
    ChildInfo right;
    ChildInfo left;
    u32 splitKey;
    u32 newChild;
    u32 newChildSubTreeType;
    u32 totalNumRecs;
} SplitInfo;

const u32 FTreeBlkLog2Ofs[MAX_SUB_TREE+1] =
{
    FTREE_LOG2_OFS_A,
    FTREE_LOG2_OFS_B,
    FTREE_LOG2_OFS_C,
    FTREE_LOG2_OFS_D,
    FTREE_LOG2_OFS_E,
    FTREE_LOG2_OFS_F,
    FTREE_LOG2_OFS_G,
    FTREE_LOG2_OFS_END
};

// p-tree
WFSKrnResult PTreeInsertIntoNode(PTreeHdr *pPTreeHdr, PTreeItr *pPti, u32 nKey, u32 *pChildOfs, PTreeAllocator *pAllocator);
u32 PTreeFindLeafOfs(PTreeHdr *pPTreeHdr, PTreeItr *pPti, u32 nKey, PTreeAllocator *pAllocator);
u32 PTreeGetOfs(void *pNode, PTreeAllocator *pAllocator);
PTreeLeaf *PTreeGetLeaf(PTreeAllocator *pAllocator, u32 nOfs);
PTreeNode *PTreeGetNode(PTreeAllocator *pAllocator, u32 nOfs);

// for BCache
static WFSBool VerifySBA(FTreeHdr* header);
static WFSBool VerifyTree(FTreeHdr* header);

PTreeHdr* FTreeSelectSubTree(FTreeHdr* header, WFSBlkAdr unitLen)
{
    PTreeHdr* subTree;

    // An extent in lower subtree will be frequently referred.
    if (unitLen == FTREE_LOG2_OFS_A)
    {
        subTree = &header->tree[0];
    }
    else if (unitLen == FTREE_LOG2_OFS_B)
    {
        subTree = &header->tree[1];
    }
    else if (unitLen == FTREE_LOG2_OFS_C)
    {
        subTree = &header->tree[2];
    }
    else if (unitLen == FTREE_LOG2_OFS_D)
    {
        subTree = &header->tree[3];
    }
    else if (unitLen == FTREE_LOG2_OFS_E)
    {
        subTree = &header->tree[4];
    }
    else if (unitLen == FTREE_LOG2_OFS_F)
    {
        subTree = &header->tree[5];
    }
    else if (unitLen == FTREE_LOG2_OFS_G)
    {
        subTree = &header->tree[6];
    }
    else
    {
        // must not reach here.
        WFSKRN_FTREE_ASSERT(0);
        return NULL;
    }
    return subTree;
}

static inline void LEAF_SET_LEN(FTreeLeaf* pLeaf, u32 index, u32 value)
{
    u32 mask = 0xf << (4 * index);
    pLeaf->data &= ~mask;
    pLeaf->data |= ((value & 0x0f) << (4 * (index)));
}

static inline u16 GetOfs(FTreeHdr* header, void* subblock)
{
    WFSKRN_FTREE_ASSERT(GET_BLOCK_HEAD(header) < (FTreeBlock*)subblock); // kondo_masahiro 08/08/15
    WFSKRN_FTREE_ASSERT((FTreeHdr*)subblock < header);                   // kondo_masahiro 08/08/15
    return (u16) ((u8*) subblock - (u8*) GET_BLOCK_HEAD(header));
}

static inline void AllocatorInit(PTreeAllocator* allocator, FTreeHdr* header)
{
    allocator->pBase = GET_BLOCK_HEAD(header);
    allocator->fpAllocNode = FTreeAllocEntry;
    allocator->fpAllocNodes = FTreeAllocEntries;
    allocator->fpFreeNode = FTreeFreeEntry;
}

static inline FTreeLeaf* FTreeGetLeaf(FTreeHdr* header, u16 ofs)
{
    return (FTreeLeaf*) GET_SUBBLOCK(header, ofs);
}

/*

 If key[n] <= nKey = key[n+1], return key[n].

 If key[last] <= nKey < key[0]_of_the_next_leaf, return key[last].

        leaf                       the next leaf
 |---------------------|       |---------------------|
 0    1     2      3   4       0        1     2      3
 |--->|---->|----->|-->|------>|------->|---->|----->|- - ->


SearchKey():
search the leaf for the specified key.
If the key is found, return the ponter to the key. (c)
If not found, return the nearest key which is smaller then the specified key. (b, d)
Note that the key returned may not be in the leaf. (a)

          leaf
   |---------------------|
   0    1     2      3   4
   |--->|---->|----->|-->|
 A         A         A     A
 |         |         |     |
(a)       (b)       (c)   (d) <--- the specified key.

    return value (k_result)
    (a)  k_result < k0
    (b)  k_result = k1
    (c)  k_result = k3
    (d)  k_result = k4

*/
static inline u32* SearchKey(FTreeLeaf *pLeaf, u32 nKey)
{
    u32* key;
    key = &pLeaf->aKey[3];
    if (nKey < *key || *key == 0)
    {
        key -= 2; // &pLeaf->aKey[1];
    }
    else
    {
        key += 2; // &pLeaf->aKey[5];
    }

    // [1] or [5]
    if (nKey < *key || *key == 0)
    {
        --key; // [0] or [4].
        if (nKey < *key || *key == 0)
        {
            if (pLeaf->aKey[0] == 0 && key == &pLeaf->aKey[0])
            {
                // [0] is a valid key.
            }
            else
            {
                --key; // error or [3].
            }
        }
    }
    else
    {
        ++key; // [2] or [6]
        if (nKey < *key || *key == 0)
        {
            --key; // [1] or [5].
        }
    }
    return key;
}

//
// split F-tree block.
//

// calculate the number of nodes and leaves in the new F-tree.
static WFSKrnResult EstimateNewTree(ChildInfo* childInfo)
{
    u32 count;
    u32 level = 0;

    WFSKRN_FTREE_RUNTIME_CHECK(VerifySBA(childInfo->ftreeHdr));

    childInfo->numEntries[level] = (childInfo->nNumRecs + FTREE_LEAF_APPROPRIATE_NUM-1) / (FTREE_LEAF_APPROPRIATE_NUM);
    count = childInfo->numEntries[0]; // the number of leaves.
    while (1 < childInfo->numEntries[level])
    {
        childInfo->numEntries[level+1] = (childInfo->numEntries[level] + FTREE_NODE_APPROPRIATE_NUM) / (FTREE_NODE_APPROPRIATE_NUM + 1);
        count += childInfo->numEntries[level+1]; // add the number of nodes in (level + 1).
        ++level;
        if (PTREE_MAX_HEIGHT - 1 < level)
        {
            return WFSKRN_RESULT_PTREE_FULL; // cannot split. It must not happen...
        }
        WFSKRN_FTREE_RUNTIME_CHECK(VerifySBA(childInfo->ftreeHdr));
    }
    childInfo->nRootHeight = level;
    childInfo->numEntries[++level] = 0; // this level does not exist.

    childInfo->totalEntries = count; // leaves and nodes.

    WFSKRN_FTREE_RUNTIME_CHECK(VerifySBA(childInfo->ftreeHdr));
    return WFSKRN_RESULT_OK;
}

// determine the threshold block address.
static u32 GetSplitAddress(FTreeHdr* header, FTreeItr* iter)
{
    WFSKrnResult result;
    u32 halfNumRecs;
    int index;
    int indexMajority = -1;
    int nNumRecsMax = 0;
    PTreeHdr* subTree;

    // search majority.
    for (index = 0; index < MAX_SUB_TREE; ++index)
    {
        subTree = &header->tree[index];

        if (nNumRecsMax < subTree->nNumRecs)
        {
            nNumRecsMax = subTree->nNumRecs;
            indexMajority = index;
        }
    }

    WFSKRN_FTREE_ASSERT(0 <= indexMajority && nNumRecsMax != 0);
    iter->subTree = &header->tree[indexMajority];

    halfNumRecs = iter->subTree->nNumRecs / 2;
    result = FTreeFirst(header, iter); // [check] error handling.
    WFSKRN_FTREE_ASSERT(result == WFSKRN_RESULT_OK);
    while (halfNumRecs--)
    {
        result = FTreeNext(header, iter); // [check] error handling.
        WFSKRN_FTREE_ASSERT(result == WFSKRN_RESULT_OK);
    }

    return FTREE_GET_KEY(iter);
}

static WFSKrnResult GetTemporaryControlBlock(FTreeHdr* header, FTreeItr* iter, u32 key)
{
    WFSKrnResult result = WFSKRN_RESULT_NOT_FOUND;
    result = FTreeFind(header, iter, key, (WFSBool) FALSE);
    WFSKRN_FTREE_ASSERT(result == WFSKRN_RESULT_OK); // key must be found, because the subtree has the most entry.
    return result;
}

static FTreeHdr* ControlBlockInit(AreaInfo* area, u32 addr)
{
    void* pBlock;
    FTreeHdr* hdr;

    WFSKrnResult result = FBAGetCache(area, addr, BCACHE_FLAG_DIRTY | BCACHE_FLAG_PINNED, &pBlock);
    if (result != WFSKRN_RESULT_OK)
    {
        return NULL;
    }
    FTreeBlockInit(area, pBlock);

    hdr = GET_FTREEHDR(pBlock);
    return hdr;
}

// prepare split.
static WFSKrnResult SplitTreeInit(AreaInfo* area, FTreeHdr* header, SplitInfo* splitInfo)
{
    FTreeItr iter;
    WFSKrnResult result;

    splitInfo->splitKey = GetSplitAddress(header, &iter);

    result = GetTemporaryControlBlock(header, &iter, splitInfo->splitKey);
    WFSKRN_FTREE_ASSERT(result == WFSKRN_RESULT_OK);

    splitInfo->right.address = FTREE_GET_KEY(&iter);
    if (FTREE_GET_DATA(&iter) == 0 &&
        iter.subTree == &header->tree[FTREE_TYPE_A])
    {
        // need one more block for the left child.
        result = FTreeNext(header, &iter);
        WFSKRN_FTREE_ASSERT(result == WFSKRN_RESULT_OK);
        splitInfo->left.address = FTREE_GET_KEY(&iter);
    }
    else
    {
        splitInfo->left.address = splitInfo->right.address + 1;
    }

    splitInfo->right.ftreeHdr = ControlBlockInit(area, splitInfo->right.address);
    splitInfo->left.ftreeHdr = ControlBlockInit(area, splitInfo->left.address);

    return WFSKRN_RESULT_OK;
}

static void InsertToLeafAtOnce(FTreeLeaf* leaf, u32* key, u32* data, u32 num)
{
    int i;

    leaf->data = 0;
    for (i = 0; i < num; ++i)
    {
        WFSKRN_FTREE_ASSERT(data[i] < 16);
        leaf->aKey[i] = key[i];
        leaf->data |= ((data[i] & 0xf) << 4 * i);

        // the root area can be larger than 16GB.
        WFSKRN_FTREE_ASSERT(key[i] < 1024 * 1024 * 1024); // [check] 8TB.
        /*
        WFSKRN_FTREE_ASSERT(key[i] < 1024 * 1024 * 2); // the maximum area size is 16GB, for now.
                                                       // 2 * 1024 * 1024 means 16GB (= 2 * 1k * 1k * 8KB).
        */
    }

    for (;i < FTREE_LEAF_FAN_OUT; ++i)
    {
        leaf->aKey[i] = 0;
    }
}

static inline u32 GetKey(FTreeHdr* child, ChildInfo* childInfo, u32 index)
{
    u32 ofs = childInfo->firstEntry + (index / 5) * sizeof(FTreeBlock); // [check] hard-coding.
    WFSKRN_FTREE_ASSERT(ofs < BLOCK_SIZE);
    WFSKRN_FTREE_ASSERT(sizeof(WFSMetaDataFlags) + sizeof(WFSHashCode) <= ofs);
    FTreeLeaf* pLeaf = FTreeGetLeaf(child, ofs);
    u32 key = pLeaf->aKey[index % 5]; // [check] hard-coding.

    return key;
}

// connect nodes in a new F-tree block.
static WFSKrnResult LinkNodes(FTreeHdr* child, ChildInfo* childInfo)
{
    PTreeNode* node = 0; // kondo_masahiro 08/08/15
    FTreeBlock* block;
    u32 height;
    u32 index; // key index
    u32 lastIndex;
    int count;
    u32 numNodes;
    WFSKrnResult result;

    int i;
    u32 factor = 5; // [check] hard-coding.
    u32 leafOfs;
    u32 ofsLeaves = childInfo->numEntries[0] * sizeof(FTreeBlock);
    block = GET_SUBBLOCK(child, childInfo->firstEntry + ofsLeaves);

    lastIndex = childInfo->nNumRecs;

    WFSKRN_FTREE_RUNTIME_CHECK(VerifySBA(childInfo->ftreeHdr));

    numNodes = childInfo->totalEntries - childInfo->numEntries[0];
    if (numNodes == 0)
    {
        if (0 < childInfo->totalEntries)
        {
            childInfo->nRootOfs = childInfo->firstEntry;
        }
        return WFSKRN_RESULT_OK;
    }
    result = FTreeAllocEntries(GET_BLOCK_HEAD(child), NULL, numNodes);
    if (result != WFSKRN_RESULT_OK)
    {
        return result;
    }

    WFSKRN_FTREE_RUNTIME_CHECK(VerifySBA(childInfo->ftreeHdr));

    leafOfs = childInfo->firstEntry;
    for (height = 1; height <= childInfo->nRootHeight; ++height)
    {
        index = 0;
        count = childInfo->numEntries[height];
        u32 numChild = childInfo->numEntries[height-1];
        while (count--)
        {
            index += factor;
            node = (PTreeNode*) block;
            ++block;
            memset(&node->aKey[0], 0, sizeof(node->aKey));

#ifdef FBA_DEBUG_VERBOSE
        FTreeLeaf* pLeaf = 0;
        if (FBADebugVerbose)
        {
            OS_TPrintf("node ofs %d\n", (u8*) node - (u8*) GET_BLOCK_HEAD(child));
        }
#endif // FBA_DEBUG_VERBOSE

            if (numChild == 1)
            {
                // If child is only one, pass through.

                /* pass through node
                 PTreeNode
                 +-----+-----+ - -
                 |  0  |  0  |
                 +---+---+---+ - -
                 |ofs| 0 | 0
                 +---+---+---+ - -
                */
#ifdef FBA_DEBUG_VERBOSE
                if (FBADebugVerbose)
                {
                    pLeaf = (FTreeLeaf*) GET_SUBBLOCK(child, leafOfs);
                }
#endif // FBA_DEBUG_VERBOSE
                node->aOfs[0] = leafOfs;
                node->aOfs[1] = 0;
                leafOfs += sizeof(FTreeBlock);

#ifdef FBA_DEBUG_VERBOSE
                if (FBADebugVerbose)
                {
                    OS_TPrintf("key : %5d, ofs : %5d, leaf key[0] %d\n",
                        node->aKey[0], node->aOfs[0], pLeaf->aKey[0]);
                    OS_TPrintf("             ofs : %5d\n\n", node->aOfs[1]);
                }
#endif // FBA_DEBUG_VERBOSE
            }
            else
            {
                for (i = 0; i < FTREE_NODE_APPROPRIATE_NUM && 1 < numChild; ++i)
                {
                    if (lastIndex < index && factor * (i+1) <= lastIndex)
                    {
                        index = factor * (i+1);
                    }

                    node->aKey[i] = GetKey(child, childInfo, index);
                    node->aOfs[i] = leafOfs;
#ifdef FBA_DEBUG_VERBOSE
                    if (FBADebugVerbose)
                    {
                        pLeaf = (FTreeLeaf*) GET_SUBBLOCK(child, leafOfs);
                        OS_TPrintf("key : %5d, ofs : %5d, leaf key[0] %d\n",
                            node->aKey[i], node->aOfs[i], pLeaf->aKey[0]);

                        WFSKRN_FTREE_ASSERT(leafOfs < BLOCK_SIZE);
                    }
#endif // FBA_DEBUG_VERBOSE
                    leafOfs += sizeof(FTreeBlock);
                    index += factor;
                    --numChild;
                }

                --numChild;
                node->aOfs[i] = leafOfs;
#ifdef FBA_DEBUG_VERBOSE
                if (FBADebugVerbose)
                {
                    pLeaf = (FTreeLeaf*) GET_SUBBLOCK(child, leafOfs);
                    OS_TPrintf("             ofs : %5d, leaf key[0] %d\n\n",
                        node->aOfs[i], pLeaf->aKey[0]);
                }
#endif // FBA_DEBUG_VERBOSE
                leafOfs += sizeof(FTreeBlock);
            }
        }

        lastIndex = (childInfo->nNumRecs / factor) * factor;
        factor *= 5; // hard-coding.
    }
    childInfo->nRootOfs = GetOfs(child, node);
    WFSKRN_FTREE_ASSERT(childInfo->nRootOfs < BLOCK_SIZE);
    return WFSKRN_RESULT_OK;
}

static WFSKrnResult AllocMiddleBlock(FTreeHdr* right, SplitInfo* splitInfo, WFSBlkAdr ngAddr)
{
    WFSKrnResult result = WFSKRN_RESULT_UNKNOWN;
    FTreeItr iter;
    u32 halfNumRecs;
    u32 subTreeIndex = 0;

    iter.subTree = &right->tree[subTreeIndex];

    while (subTreeIndex < FTREE_TYPE_END)
    {
        iter.subTree = &right->tree[subTreeIndex];
        if (0 < iter.subTree->nNumRecs)
        {
            halfNumRecs = iter.subTree->nNumRecs / 2;
            result = FTreeFirst(right, &iter);
            while (halfNumRecs-- && result == WFSKRN_RESULT_OK)
            {
                result = FTreeNext(right, &iter);
            }
            if (ngAddr == FTREE_GET_KEY(&iter) &&
                result == WFSKRN_RESULT_OK)
            {
#ifdef _DEBUG
                OS_TPrintf("*** prevent from overwriting the left block. ***\n");
#endif // _DEBUG
                result = FTreeNext(right, &iter);
            }
            break;
        }
        ++subTreeIndex;
    }

    if (subTreeIndex == FTREE_TYPE_END)
    {
        return WFSKRN_RESULT_FTREE_EMPTY;
    }

    if (result == WFSKRN_RESULT_OK)
    {
        // It is ok even if
        // splitInfo->newChild == splitInfo.right.address or
        // splitInfo->newChild == splitInfo.left.address.
        // Because FBA remaps splitInfo.left.address to the original, then
        // remaps splitInfo.right.address to splitInfo->newChild.
        splitInfo->newChild = FTREE_GET_KEY(&iter);
        splitInfo->newChildSubTreeType = subTreeIndex;
    }

    return result;
}

/*
 header (original)     left child (temporary)
 +---------+           +---------+
 |         |   remap   |         |
 |      8kb|<----------|      8kb|
 +---------+           +---------+

 right child (new)     right child (temporary)
 +---------+           +---------+
 |         |   remap   |         |
 |      8kb|<----------|      8kb|
 +---------+           +---------+

 The new right child is in the middle of the right child area.
 */

static WFSKrnResult RemapControlBlocks(AreaInfo* area, FTreeHdr* header, SplitInfo* splitInfo)
{
    WFSKrnResult result;
    FTreeHdr* newRight;
    void* tmpBlock;
    void* newBlock;

    result = AllocMiddleBlock(splitInfo->right.ftreeHdr, splitInfo, splitInfo->left.address);

    WFSKRN_FTREE_ASSERT(result == WFSKRN_RESULT_OK);

    if (result == WFSKRN_RESULT_OK)
    {
        u32 offset = sizeof(WFSMetaDataFlags) + sizeof(WFSHashCode); // to keep hash and flag in 8KB block.

        result = FBAGetCache(area, splitInfo->newChild, BCACHE_FLAG_DIRTY | BCACHE_FLAG_PINNED, &newBlock);
        WFSKRN_FTREE_ASSERT(result == WFSKRN_RESULT_OK);

        newRight = GET_FTREEHDR(newBlock);
        WFSKRN_FTREE_RUNTIME_CHECK(VerifySBA(splitInfo->right.ftreeHdr));

        // remap left child, first.
        // splitInfo->left.address may be reused as a splitInfo->newChild.
        tmpBlock = GET_BLOCK_HEAD(splitInfo->left.ftreeHdr);
        memmove((u8*) GET_BLOCK_HEAD(header) + offset, (u8*) tmpBlock + offset, area->nBlkSize - offset); // [check] must use BCache API.
        FBAFreeCache(area, splitInfo->left.address);

        // remap right child.
        tmpBlock = GET_BLOCK_HEAD(splitInfo->right.ftreeHdr);
        memmove((u8*) newBlock + offset, (u8*) tmpBlock + offset, area->nBlkSize - offset);
        FBAFreeCache(area, splitInfo->right.address);

        WFSKRN_FTREE_RUNTIME_CHECK(VerifySBA(newRight));

        FBAUnpinCache(area, splitInfo->newChild); // unpin the new child block.
    }
    return result;
}

static inline WFSKrnResult InsertKeysToLeaf(FTreeHdr* header, SplitInfo* splitInfo, FTreeHdr* child, int count, FTreeItr* iter, u32* numKeys)
{
    u32 key[FTREE_LEAF_APPROPRIATE_NUM];
    u32 data[FTREE_LEAF_APPROPRIATE_NUM];
    u32 offset;
    int index = 0;
    FTreeLeaf* leaf;
    WFSKrnResult result = WFSKRN_RESULT_UNKNOWN;

    *numKeys = 0;
    while (count--)
    {
        key[index] = FTREE_GET_KEY(iter);
        data[index] = FTREE_GET_DATA(iter);
        if (index == FTREE_LEAF_APPROPRIATE_NUM-1)
        {
            // insert
            WFSKRN_FTREE_RUNTIME_CHECK(VerifySBA(splitInfo->right.ftreeHdr));

            offset = FTreeAllocEntry(GET_BLOCK_HEAD(child));

            WFSKRN_FTREE_RUNTIME_CHECK(VerifySBA(splitInfo->right.ftreeHdr));

            leaf = FTreeGetLeaf(child, offset);

            WFSKRN_FTREE_RUNTIME_CHECK(VerifySBA(splitInfo->right.ftreeHdr));

            InsertToLeafAtOnce(leaf, key, data, FTREE_LEAF_APPROPRIATE_NUM);
            *numKeys += FTREE_LEAF_APPROPRIATE_NUM;
            index = 0;

            WFSKRN_FTREE_RUNTIME_CHECK(VerifySBA(splitInfo->right.ftreeHdr));
        }
        else
        {
            ++index;
        }

        result = FTreeNext(header, iter);
        if (result != WFSKRN_RESULT_OK)
        {
            break;
        }
    }

    WFSKRN_FTREE_RUNTIME_CHECK(VerifySBA(splitInfo->right.ftreeHdr));

    if (0 < index)
    {
        // remainder
        offset = FTreeAllocEntry(GET_BLOCK_HEAD(child));
        leaf = FTreeGetLeaf(child, offset);
        InsertToLeafAtOnce(leaf, key, data, index);
        *numKeys += index;
    }

    splitInfo->right.nNumRecs = *numKeys;
    if (count == 0)
    {
        result = WFSKRN_RESULT_OK;
    }

    return result;
}

static WFSKrnResult CreateLeavesRight(FTreeHdr* header, SplitInfo* splitInfo, FTreeItr* iter)
{
    FTreeHdr* child;
    int count;
    u32 numKeys;
    WFSKrnResult result = WFSKRN_RESULT_OK;

    WFSKRN_FTREE_RUNTIME_CHECK(VerifySBA(splitInfo->right.ftreeHdr));

    child = splitInfo->right.ftreeHdr;
    splitInfo->right.firstEntry = splitInfo->right.ftreeHdr->sbaHdr.offset + splitInfo->right.ftreeHdr->sbaHdr.freeBlock * sizeof(FTreeBlock);

    count = splitInfo->totalNumRecs - splitInfo->left.nNumRecs;
    if (0 < count)
    {
        result = InsertKeysToLeaf(header, splitInfo, child, count, iter, &numKeys);
    }
    else
    {
        splitInfo->right.nNumRecs = 0;
    }

    WFSKRN_FTREE_RUNTIME_CHECK(VerifySBA(splitInfo->right.ftreeHdr));
    return result;
}

static WFSKrnResult CreateLeavesLeft(FTreeHdr* header, SplitInfo* splitInfo, FTreeItr* iter)
{
    u32 key[FTREE_LEAF_APPROPRIATE_NUM];
    u32 data[FTREE_LEAF_APPROPRIATE_NUM];
    WFSKrnResult result;
    FTreeHdr* child;
    FTreeLeaf* leaf;
    u32 numKeys = 0;
    int index = 0;
    u32 offset;

    if (iter->subTree->nNumRecs == 0)
    {
        // this subtree is empty. // [check] must check this routine.
        splitInfo->left.nNumRecs = 0;
        return WFSKRN_RESULT_OK;
    }

    splitInfo->left.firstEntry = splitInfo->left.ftreeHdr->sbaHdr.offset + splitInfo->left.ftreeHdr->sbaHdr.freeBlock * sizeof(FTreeBlock);
    child = splitInfo->left.ftreeHdr;

    result = FTreeFirst(header, iter);
    if (result != WFSKRN_RESULT_OK)
    {
        return result;
    }

    while (FTREE_GET_KEY(iter) < splitInfo->splitKey)
    {
        key[index] = FTREE_GET_KEY(iter);
        data[index] = FTREE_GET_DATA(iter);
        if (index == FTREE_LEAF_APPROPRIATE_NUM-1)
        {
            // insert
            offset = FTreeAllocEntry(GET_BLOCK_HEAD(child));

            leaf = FTreeGetLeaf(child, offset);
            InsertToLeafAtOnce(leaf, key, data, FTREE_LEAF_APPROPRIATE_NUM);
            numKeys += FTREE_LEAF_APPROPRIATE_NUM;
            index = 0;
        }
        else
        {
            ++index;
        }

        result = FTreeNext(header, iter);
        if (result != WFSKRN_RESULT_OK)
        {
            break;
        }
    }

    if (0 < index)
    {
        // remainder
        offset = FTreeAllocEntry(GET_BLOCK_HEAD(child));
        leaf = FTreeGetLeaf(child, offset);
        InsertToLeafAtOnce(leaf, key, data, index);
        numKeys += index;
    }

    splitInfo->left.nNumRecs = numKeys;

    if (result != WFSKRN_RESULT_OK)
    {
        if (iter->subTree->nNumRecs == numKeys)
        {
            result = WFSKRN_RESULT_OK;
        }
    }

    return result;
}

static WFSKrnResult CreateRightChild(FTreeHdr* header, SplitInfo* splitInfo, FTreeItr* iter, int treeType)
{
    WFSKrnResult result;
    PTreeHdr* subtree;

    WFSKRN_FTREE_RUNTIME_CHECK(VerifySBA(splitInfo->right.ftreeHdr));

    result = CreateLeavesRight(header, splitInfo, iter); // [check] error handling.
    if (result != WFSKRN_RESULT_OK)
    {
        goto CREATERIGHTCHILD_ERROR;
    }

    WFSKRN_FTREE_RUNTIME_CHECK(VerifySBA(splitInfo->right.ftreeHdr));

    result = EstimateNewTree(&splitInfo->right); // [check] name should be geometry??
    if (result != WFSKRN_RESULT_OK)
    {
        goto CREATERIGHTCHILD_ERROR;
    }

    WFSKRN_FTREE_RUNTIME_CHECK(VerifySBA(splitInfo->right.ftreeHdr));

    result = LinkNodes(splitInfo->right.ftreeHdr, &splitInfo->right); // [check] error handling.
    if (result != WFSKRN_RESULT_OK)
    {
        goto CREATERIGHTCHILD_ERROR;
    }

    WFSKRN_FTREE_RUNTIME_CHECK(VerifySBA(splitInfo->right.ftreeHdr));

    subtree = GET_SUBTREE(splitInfo->right.ftreeHdr, treeType);
    subtree->nRootHeight = splitInfo->right.nRootHeight;
    subtree->nRootOfs = splitInfo->right.nRootOfs;
    subtree->nNumRecs = splitInfo->right.nNumRecs;

CREATERIGHTCHILD_ERROR:
    return result;
}

static WFSKrnResult CreateLeftChild(FTreeHdr* header, SplitInfo* splitInfo, FTreeItr* iter, int treeType)
{
    WFSKrnResult result;
    PTreeHdr* subtree;

    iter->subTree = GET_SUBTREE(header, treeType);

    result = CreateLeavesLeft(header, splitInfo, iter); // [check] error handling.
    if (result != WFSKRN_RESULT_OK)
    {
        goto CREATELEFTCHILD_ERROR;
    }

    result = EstimateNewTree(&splitInfo->left); // [check] name should be geometry??
    if (result != WFSKRN_RESULT_OK)
    {
        goto CREATELEFTCHILD_ERROR;
    }

    result = LinkNodes(splitInfo->left.ftreeHdr, &splitInfo->left); // [check] error handling.
    if (result != WFSKRN_RESULT_OK)
    {
        goto CREATELEFTCHILD_ERROR;
    }

    subtree = GET_SUBTREE(splitInfo->left.ftreeHdr, treeType);
    subtree->nRootHeight = splitInfo->left.nRootHeight;
    subtree->nRootOfs = splitInfo->left.nRootOfs;
    subtree->nNumRecs = splitInfo->left.nNumRecs;

CREATELEFTCHILD_ERROR:
    return result;
}

/*
  nodes  (lv.3)
  nodes  (lv.2)
  nodes  (lv.1)
  leaves (lv.0)
*/

static WFSKrnResult SplitTree(AreaInfo* area, FTreeHdr* header, FTreeItr* iter)
{
    WFSKrnResult result;
    FTreeItr searchIter;
    SplitInfo splitInfo;
    int treeType;

#ifdef FBA_DEBUG_VERBOSE
        if (FBADebugVerbose)
        {
            OS_TPrintf("[SPLIT TREE]\n");
        }
#endif // FBA_DEBUG_VERBOSE

    // prepare two 8kB-blocks for left and right child-blocks.
    // (Both blocks are used as a work buffer.)
    result = SplitTreeInit(area, header, &splitInfo);
    if (result != WFSKRN_RESULT_OK)
    {
        return result;
    }

    for (treeType = FTREE_TYPE_A; treeType < FTREE_TYPE_G; ++treeType)
    {
        splitInfo.totalNumRecs = header->tree[treeType].nNumRecs;
        result = CreateLeftChild(header, &splitInfo, &searchIter, treeType);
        if (result != WFSKRN_RESULT_OK)
        {
            return result; // [check]
        }
        result = CreateRightChild(header, &splitInfo, &searchIter, treeType);
        if (result != WFSKRN_RESULT_OK)
        {
            return result; // [check]
        }
    }
    // remap control blocks.
    result = RemapControlBlocks(area, header, &splitInfo);

    iter->splitKey = splitInfo.splitKey;
    iter->newChild = splitInfo.newChild;
    iter->newChildSubTreeType = splitInfo.newChildSubTreeType;

#ifdef FBA_DEBUG_VERBOSE
    if (FBADebugVerbose)
    {
        OS_TPrintf("newChild block address %u\n", iter->newChild);
    }
#endif // FBA_DEBUG_VERBOSE

    return result;
}

static WFSKrnResult SplitLeaf(AreaInfo* area, FTreeHdr* header, FTreeItr* iter, FTreeLeaf** ppLeaf, u32 nKey)
{
    FTreeLeaf* pLeaf = *ppLeaf;
    FTreeLeaf* pNewLeaf;
    u32 nOfs;
    u32 key;
    WFSKrnResult result;

    // [check] should refine.
    if (pLeaf->aKey[4] < nKey)
    {
        key = pLeaf->aKey[4];
    }
    else if (pLeaf->aKey[3] < nKey)
    {
        key = nKey;
    }
    else
    {
        key = pLeaf->aKey[3];
    }

    PTreeAllocator allocator;
    AllocatorInit(&allocator, header);
    nOfs = GetOfs(header, pLeaf);
    WFSKRN_FTREE_ASSERT(nOfs < area->nBlkSize);
    result = PTreeInsertIntoNode(iter->subTree, &iter->ptree, key, &nOfs, &allocator);
    if (result == WFSKRN_RESULT_PTREE_FULL)
    {
        WFSKRN_FTREE_RUNTIME_CHECK(VerifyTree(header));
        // split 8kB block.
        result = SplitTree(area, header, iter); // WFSKRN_RESULT_FTREE_RETRY or WFSKRN_RESULT_FTREE_SPLIT
                                          // the history in iter became ivalid. retry.
        WFSKRN_FTREE_RUNTIME_CHECK(VerifyTree(header));
        if (result == WFSKRN_RESULT_OK)
        {
            result = WFSKRN_RESULT_FTREE_RETRY;
        }
        return result;
    }
    pNewLeaf = (FTreeLeaf*) PTreeGetLeaf(&allocator, nOfs);

    if (pLeaf->aKey[3] < nKey)
    {
        // left child : key[0]-[3].
        // right child: key[4]-[6] <-- insert nKey later.
        memmove(&pNewLeaf->aKey[0], &pLeaf->aKey[4], 3 * sizeof(u32));
        pNewLeaf->data = (pLeaf->data >> 16);
        memset(&pLeaf->aKey[4], 0, 3 * sizeof(u32));
        memset(&pNewLeaf->aKey[3], 0, 4 * sizeof(u32));
        pLeaf = pNewLeaf;
    }
    else
    {
        // left child : key[0]-[2] <-- insert nKey later.
        // right child: key[3]-[6].
        memmove(&pNewLeaf->aKey[0], &pLeaf->aKey[3], 4 * sizeof(u32));
        pNewLeaf->data = (pLeaf->data >> 12);
        memset(&pLeaf->aKey[3], 0, 4 * sizeof(u32));
        memset(&pNewLeaf->aKey[4], 0, 3 * sizeof(u32));
    }

    *ppLeaf = pLeaf;
    return WFSKRN_RESULT_OK;
}

//
// Insert a key into F-tree.
//
WFSKrnResult FTreeInsert(AreaInfo* area, FTreeHdr* header, FTreeItr* iter, u32 nKey, u32 nData)
{
    u32* key;
    u32 shift;
    int index;
    u32 mask;
    FTreeLeaf* pLeaf;
    u32 ofs;

#ifdef FBA_DEBUG_VERBOSE
    if (FBADebugVerbose)
    {
        OS_TPrintf("                    ");
        OS_TPrintf("FTreeInsert(): %u (blk %u) type %d\n", nKey, nData + 1, (iter->subTree - &header->tree[0]));
    }
#endif // FBA_DEBUG_VERBOSE

    WFSKRN_FTREE_RUNTIME_CHECK(VerifyTree(header));

    PTreeAllocator allocator;
    AllocatorInit(&allocator, header);

    PTreeHdr* subTree = iter->subTree;
    if (subTree->nNumRecs == 0)
    {
        subTree->nRootHeight = 0;
        subTree->nRootOfs = FTreeAllocEntry(GET_BLOCK_HEAD(header));
        if (subTree->nRootOfs == 0)
        {
            // split 8kB block.
            WFSKrnResult result = SplitTree(area, header, iter); // WFSKRN_RESULT_FTREE_RETRY or WFSKRN_RESULT_FTREE_SPLIT
                                                    // the history in iter became ivalid. retry.
            WFSKRN_FTREE_RUNTIME_CHECK(VerifyTree(header));
            if (result == WFSKRN_RESULT_OK)
            {
                result = WFSKRN_RESULT_FTREE_RETRY;
            }
            return result;
        }

        WFSKRN_FTREE_ASSERT(subTree->nRootOfs < area->nBlkSize);
        pLeaf = iter->pLeaf = FTreeGetLeaf(header, (u32) subTree->nRootOfs);
        memset(&pLeaf->aKey[1], 0, (FTREE_LEAF_FAN_OUT-1) * sizeof(u32));    // It is necessary to initialize to 0.
        pLeaf->aKey[0] = nKey;
        LEAF_SET_LEN(pLeaf, 0, nData);
        goto Done;
    }

    ofs = PTreeFindLeafOfs(iter->subTree, &iter->ptree, nKey, &allocator);
    WFSKRN_FTREE_ASSERT(ofs < area->nBlkSize);
    WFSKRN_FTREE_ASSERT(header->sbaHdr.offset <= ofs);

    pLeaf = iter->pLeaf = FTreeGetLeaf(header, ofs);
    if (pLeaf->aKey[FTREE_LEAF_FAN_OUT-1] != 0)
    {
        WFSKRN_FTREE_RUNTIME_CHECK(VerifyTree(header));
        WFSKrnResult result = SplitLeaf(area, header, iter, &pLeaf, nKey);
        WFSKRN_FTREE_RUNTIME_CHECK(VerifyTree(header));
        if (result != WFSKRN_RESULT_OK)
        {
            return result; // WFSKRN_RESULT_FTREE_RETRY or WFSKRN_RESULT_PTREE_FULL
        }
    }

    key = SearchKey(pLeaf, nKey);
    if (!IS_OUT_OF_RANGE(pLeaf, key) && *key == nKey)
    {
        // cannot overwrite the existing key.
        return WFSKRN_RESULT_ALREADY_EXISTS;
    }
    ++key;
    index = (key - &pLeaf->aKey[0]);
    WFSKRN_FTREE_ASSERT(0 <= index);

    // insert data.
    shift = FTREE_LEAF_FAN_OUT - 1 - index;
    memmove(key+1, key, shift * sizeof(u32));
    mask = (0xffffffff << (4 * index));
    pLeaf->data = ((pLeaf->data & mask) << 4) | (pLeaf->data & ~mask);

    *key = nKey;
    LEAF_SET_LEN(pLeaf, index, nData);

#ifdef _DEBUG
    int i = 0;
    for (i = 0; i < FTREE_LEAF_FAN_OUT; ++i)
    {
        WFSKRN_FTREE_ASSERT(pLeaf->aKey[i] < 1024 * 1024 * 1024 /* 8TB */);
    }
#endif // _DEBUG

Done:
    ++subTree->nNumRecs;
    WFSKRN_FTREE_RUNTIME_CHECK(VerifyTree(header));

    return WFSKRN_RESULT_OK;
}

static WFSKrnResult ModifyEntry(FTreeLeaf* pLeaf, u32 origKey, u32 modKey, u32 modData)
{
    u32* key;
    int index;
    u32 origData;

    key = SearchKey(pLeaf, origKey);
    if (IS_OUT_OF_RANGE(pLeaf, key) || *key != origKey)
    {
        return WFSKRN_RESULT_NOT_FOUND;
    }
    index = (key - &pLeaf->aKey[0]);
    WFSKRN_FTREE_ASSERT(0 <= index);

    origData = LEAF_GET_LEN(pLeaf, index);

#ifdef FBA_DEBUG_VERBOSE
        if (FBADebugVerbose)
        {
            OS_TPrintf("%u (blk %u) -> %u (blk %u)\n",
                origKey, LEAF_GET_LEN(pLeaf, index) + 1, modKey, modData + 1);
        }
#endif // FBA_DEBUG_VERBOSE

    pLeaf->aKey[index] = modKey;
    LEAF_SET_LEN(pLeaf, index, modData);

    return WFSKRN_RESULT_OK;
}

WFSKrnResult FTreeModify(FTreeHdr* header, FTreeItr* iter, u32 origKey, u32 modKey, u32 modData)
{
    WFSKrnResult result;
    FTreeLeaf* pLeaf;
    u32 ofs;
    PTreeHdr* subTree;
    PTreeAllocator allocator;

#ifdef FBA_DEBUG_VERBOSE
    if (FBADebugVerbose)
    {
        OS_TPrintf("                    ");
        OS_TPrintf("FTreeModify(): type %d: ", (iter->subTree - &header->tree[0]));
    }
#endif // FBA_DEBUG_VERBOSE

    AllocatorInit(&allocator, header);

    subTree = iter->subTree;
    if (subTree->nNumRecs == 0)
    {
        return WFSKRN_RESULT_NOT_FOUND;
    }

    ofs = PTreeFindLeafOfs(iter->subTree, &iter->ptree, origKey, &allocator);
    WFSKRN_FTREE_ASSERT(ofs < BLOCK_SIZE);
    WFSKRN_FTREE_ASSERT(header->sbaHdr.offset <= ofs);
    pLeaf = iter->pLeaf = FTreeGetLeaf(header, ofs);

    result = ModifyEntry(pLeaf, origKey, modKey, modData);
    if (result == WFSKRN_RESULT_OK)
    {
        FTREE_GET_KEY(iter) = modKey;
        FTREE_GET_DATA(iter) = modData;
    }
    WFSKRN_FTREE_RUNTIME_CHECK(VerifyTree(header));
    return result;
}

static WFSKrnResult DeleteKey(FTreeLeaf* pLeaf, u32 nKey)
{
    u32* key;
    u32 shift;
    int index;
    u32 maskLow;
    u32 maskHigh;

    key = SearchKey(pLeaf, nKey);
    if (IS_OUT_OF_RANGE(pLeaf, key) || *key != nKey)
    {
        return WFSKRN_RESULT_NOT_FOUND;
    }
    index = (key - &pLeaf->aKey[0]);
    WFSKRN_FTREE_ASSERT(0 <= index);

    // insert data.
    shift = FTREE_LEAF_FAN_OUT - 1 - index;
    memmove(key, key+1, shift * sizeof(u32));
    pLeaf->aKey[FTREE_LEAF_FAN_OUT - 1] = 0;

    maskLow  = (0xffffffff << (4 * (index)));
    maskHigh = (0xffffffff << (4 * (index+1)));
    pLeaf->data = ((pLeaf->data & maskHigh) >> 4) | (pLeaf->data & ~maskLow);

    return WFSKRN_RESULT_OK;
}

static WFSKrnResult DeleteEntryFromParentNode(PTreeHdr* pPTreeHdr, PTreeItr* pPti, PTreeAllocator *pAllocator)
{
    u32 nOfs;
    u32 *pEntryIdx = pPti->aEntryIdx;

    // Delete the entry from the parent node
    u32 nLevel = 1;
    while(nLevel <= pPTreeHdr->nRootHeight) {
        ++pEntryIdx;
        PTreeNode *pNode = pPti->apNode[nLevel-1];
        switch(*pEntryIdx) {
            case 0: if (pNode->aKey[0]==0) {
                        break;
                    }
                    pNode->aOfs[0] = pNode->aOfs[1];
            case 1: pNode->aKey[0] = pNode->aKey[1];
                    pNode->aOfs[1] = pNode->aOfs[2];
            case 2: pNode->aKey[1] = pNode->aKey[2];
                    pNode->aOfs[2] = pNode->aOfs[3];
            case 3: pNode->aKey[2] = pNode->aKey[3];
                    pNode->aOfs[3] = pNode->aOfs[4];
            case 4: pNode->aKey[3] = pNode->aKey[4];
                    pNode->aOfs[4] = pNode->aOfs[5];
            case 5: pNode->aKey[4] = 0;
                    return WFSKRN_RESULT_OK;
        }
        nOfs = PTreeGetOfs(pNode, pAllocator);
        WFSKRN_FTREE_ASSERT(nOfs < BLOCK_SIZE);
        pAllocator->fpFreeNode(pAllocator->pBase, nOfs);
        ++nLevel;
    }
    return WFSKRN_RESULT_OK;
}

//
// Delete the specified key.
//
WFSKrnResult FTreeDelete(FTreeHdr* header, FTreeItr* iter, u32 nKey)
{
    WFSKrnResult result;
    FTreeLeaf* pLeaf;
    u32 ofs;

#ifdef FBA_DEBUG_VERBOSE
    if (FBADebugVerbose)
    {
        OS_TPrintf("                    ");
        OS_TPrintf("FTreeDelete(): %u (blk %u) type %d\n", nKey, FTREE_GET_DATA(iter) + 1, (iter->subTree - &header->tree[0]));
    }
#endif // FBA_DEBUG_VERBOSE

    if (iter->subTree->nNumRecs == 0)
    {
        return WFSKRN_RESULT_NOT_FOUND;
    }

    WFSKRN_FTREE_RUNTIME_CHECK(VerifyTree(header));

    PTreeAllocator allocator;
    AllocatorInit(&allocator, header);
    ofs = PTreeFindLeafOfs(iter->subTree, &iter->ptree, nKey, &allocator);
    WFSKRN_FTREE_ASSERT(ofs < BLOCK_SIZE);
    WFSKRN_FTREE_ASSERT(header->sbaHdr.offset <= ofs);
    pLeaf = iter->pLeaf = FTreeGetLeaf(header, ofs);
    PTreeHdr* subTree = iter->subTree;

    if (subTree->nNumRecs == 0)
    {
        return WFSKRN_RESULT_NOT_FOUND;
    }

    if (pLeaf->aKey[1] == 0)
    {
        if (pLeaf->aKey[0] != nKey)
        {
            return WFSKRN_RESULT_NOT_FOUND;
        }
        FTreeFreeEntry(GET_BLOCK_HEAD(header), ofs);
        result = DeleteEntryFromParentNode(iter->subTree, &iter->ptree, &allocator);
    }
    else
    {
        result = DeleteKey(pLeaf, nKey);
    }

    if (result == WFSKRN_RESULT_OK &&
        --iter->subTree->nNumRecs == 0)
    {
        iter->subTree->nRootHeight = 0; // reset.
    }

    WFSKRN_FTREE_RUNTIME_CHECK(VerifyTree(header)); // [check] header was deleted?
    return result;
}

//
// Sub-block allocator
//
FTreeBlock* MemoryBlockAlloc(SubBlockAllocatorHeader* header, int numBlocks)
{
    FTreeBlock* free = GET_FREEBLOCK(header);
    FTreeBlock* prev = NULL;
    FTreeBlock* newFree;

    if ((END_OF_CACHE(header) <= GET_FREEBLOCK(header)) || numBlocks == 0)
    {
        return NULL; // no freeblock.
    }

    while (free->num < numBlocks)
    {
        if (free->next == MAX_FREEBLOCK(header))
        {
            WFSKRN_FTREE_RUNTIME_CHECK(VerifyMemoryBlockFTree(header));
            return NULL;
        }
        WFSKRN_FTREE_ASSERT(0 < free->next);
        prev = free;
        free = CACHE(header) + free->next;
    }

    if (free->num == numBlocks)
    {
        if (prev)
        {
            prev->next = free->next;
        }
        else
        {
            SET_FREEBLOCK(header, CACHE(header) + free->next);
        }
    }
    else
    {
        newFree = free + numBlocks;
        newFree->next = free->next;
        newFree->num = free->num - numBlocks;

        if (prev)
        {
            prev->next = newFree - CACHE(header);
        }
        else
        {
            SET_FREEBLOCK(header, newFree);
        }
    }

    free->next = 0; // allocated.
    free->num = numBlocks;

    WFSKRN_FTREE_RUNTIME_CHECK(VerifyMemoryBlockFTree(header));

    return free;
}

int MemoryBlockFree(SubBlockAllocatorHeader* header, FTreeBlock* block, u32 numBlocks)
{
    if ((block < CACHE(header)) || END_OF_CACHE(header) <= block)
    {
        return -1; // out of memory blocks.
    }
    if (numBlocks == 0)
    {
        return 0;
    }

#ifdef _DEBUG
    memset((u8*) block, 0xef, sizeof(FTreeBlock) * numBlocks); // clear
#endif // DEBUG

    if (block < GET_FREEBLOCK(header))
    {
        if (GET_FREEBLOCK(header) == block + numBlocks)
        {
            // combine
            if (GET_FREEBLOCK(header) == END_OF_CACHE(header))
            {
                block->num = numBlocks; // set flags.
                block->next = MAX_FREEBLOCK(header); // the tail of memory block.
            }
            else
            {
                block->num = GET_FREEBLOCK(header)->num + numBlocks;
                block->next = GET_FREEBLOCK(header)->next;
            }
        }
        else
        {
            block->num = numBlocks;
            block->next = GET_FREEBLOCK(header) - CACHE(header);
        }
        SET_FREEBLOCK(header, block);
        WFSKRN_FTREE_RUNTIME_CHECK(VerifyMemoryBlockFTree(header));
    }
    else
    {
        block->num = numBlocks;

        FTreeBlock* free = GET_FREEBLOCK(header);
        FTreeBlock* next = CACHE(header) + GET_FREEBLOCK(header)->next;

        while (next < block)
        {
            free = next;
            next = CACHE(header) + next->next;
        }

        if (free + free->num == block)
        {
            // combine
            numBlocks += free->num;
            free->num = numBlocks;
            block = free;
        }
        else
        {
            free->next = block - CACHE(header);
        }

        if (block + block->num == next && next < END_OF_CACHE(header))
        {
            // combine
            block->num =  next->num + numBlocks;
            block->next = next->next;
        }
        else
        {
            block->next = next - CACHE(header);
        }
        WFSKRN_FTREE_RUNTIME_CHECK(VerifyMemoryBlockFTree(header));
    }
    return 0;
}

WFSKrnResult FTreeAllocEntries(void *pBase, u16 *pNodeOfs, u32 nNumNodes)
{
    WFSKrnResult result = WFSKRN_RESULT_OUT_OF_MEMORY;
    SubBlockAllocatorHeader* header = GET_SBAHDR(pBase);
    FTreeBlock* block;

    if (nNumNodes == 0)
    {
        return WFSKRN_RESULT_OK;
    }

    block = MemoryBlockAlloc(header, nNumNodes);
    if (block)
    {
        result = WFSKRN_RESULT_OK;
        header->numBlock += nNumNodes;
        if (pNodeOfs)
        {
            int i;
            for (i = 0; i < nNumNodes; ++i)
            {
                pNodeOfs[i] = (u16) ((u8*) &block[i] - (u8*) pBase);
                WFSKRN_FTREE_ASSERT(0 < pNodeOfs[i]);
                WFSKRN_FTREE_ASSERT(pNodeOfs[i] < BLOCK_SIZE);
            }
        }
    }

    return result;
}

u16 FTreeAllocEntry(void *pBase)
{
    SubBlockAllocatorHeader* header = GET_SBAHDR(pBase);
    FTreeBlock* block = MemoryBlockAlloc(header, 1);
    u32 ret = 0;
    if (block)
    {
        WFSKRN_FTREE_ASSERT(header->numBlock + 1 <= MAX_FREEBLOCK(header));
        ++header->numBlock;
        ret = ((u8*) block - (u8*) pBase); // byte-offset from the top of the 8kB block.
        WFSKRN_FTREE_ASSERT(0 < ret);
        WFSKRN_FTREE_ASSERT(ret < BLOCK_SIZE);
    }
    return ret;
}

// nOfs is the byte offset from the top of the 8kB block.
void FTreeFreeEntry(void *pBase, u16 nOfs)
{
    SubBlockAllocatorHeader* header = GET_SBAHDR(pBase);

    FTreeBlock* block = (FTreeBlock*) ((u8*) pBase + nOfs);
    if (MemoryBlockFree(header, block, 1) == 0)
    {
        WFSKRN_FTREE_ASSERT(0 <= header->numBlock - 1);
        --header->numBlock;
    }
}

//
// Initialize an 8kB block.
//
/*
       an 8kb-block.
       +-------------------------+ A <--- block
       |                         | |
       |        hash code        | | sbaHdr->offset ... this must be the multiple of 32 bytes.
       |        etc..            | |
       +-------------------------+ V <--- freeBlock
       |                         | A
       |                         | |
       |                         | |
       |                         | |
       |                         | |
       |                         | |
       |                         | |
       |       Sub-blocks        | | sbaHdr->size
       |                         | |
       |                         | |
       |                         | |
       |                         | |
       |                         | |
       |                         | V
       +-------------------------+ <--- ftreeHdr or ptreeHdr A
       |                         |                           |
       +-------------------------+ <--- sbaHdr               | footerSize
       | SubBlockAllocatorHeader |                           |
       +-------------------------+                           V
*/

void FTreeSbaInit(AreaInfo* area, void* block, u32 footerSize, u32 offset)
{
    SubBlockAllocatorHeader* sbaHdr;
    FTreeBlock* freeBlock;
    WFSKRN_FTREE_ASSERT(sizeof(SubBlockAllocatorHeader) < footerSize);

    sbaHdr = GET_SBAHDR(block);
    sbaHdr->numBlock = 0;
    sbaHdr->freeBlock = 0;
    sbaHdr->offset = offset;
    sbaHdr->size = area->nBlkSize - (sbaHdr->offset + footerSize);

    WFSKRN_FTREE_ASSERT((sbaHdr->offset + footerSize) < area->nBlkSize);

    freeBlock = CACHE(sbaHdr);
    freeBlock->num = sbaHdr->size / sizeof(FTreeBlock);
    freeBlock->prev = 0; // first entry.
    freeBlock->next = MAX_FREEBLOCK(sbaHdr);

    // initialize ftreeHdr or ptreeHdr.
    memset((u8*) freeBlock + sbaHdr->size, 0, footerSize - sizeof(SubBlockAllocatorHeader));

}

FTreeHdr* FTreeBlockInit(AreaInfo* area, void* block /* 8kB block */)
{
    FTreeSbaInit(area, block, sizeof(FTreeHdr),
        ROUNDUP32BYTE(sizeof(WFSMetaDataFlags) + sizeof(WFSHashCode))); // offset for hash.
    return GET_FTREEHDR(block);
}

//
// Search an F-tree for the specified key.
// If strictSearch == FALSE, FTreeFind() returns the nearest value,
// which is smaller than nKey, when result == WFSKRN_RESULT_OK.
//
/*
 (ex)
          F-Tree leaf
    +--------------------+
    | 13    16   20  100 |
    +--------------------+
    | 100  200  300  400 |
    +--------------------+
  A       A      A         A
  |       |      |         |
 (1)     (2)    (3)       (4)  <--- nKey

* result

         |     strictSearch
    Case |  TRUE       |   FALSE
    -----+-------------+-------------
    (1)  |  NOT_FOUND  |   NOT_FOUND
    (2)  |  NOT_FOUND  |   OK
    (3)  |  OK         |   OK
    (4)  |  NOT_FOUND  |   OK

* FTREE_GET_KEY(iter), FTREE_GET_DATA(iter)

         |     strictSearch
    Case |  TRUE      |   FALSE
    -----+------------+-------------
    (1)  |  x,   x    |   13, 100 <--- note the key, 13 is not smaller than nKey.
    (2)  |  x,   x    |   13, 100
    (3)  |  20, 300   |   20, 300
    (4)  |  x,   x    |  100, 400

            x means no value is set.
*/

WFSKrnResult FTreeFind(FTreeHdr* header, FTreeItr* iter, u32 nKey, WFSBool strictSearch)
{
    u32* key;
    int index;
    u32 ofs;
    FTreeLeaf* pLeaf;

    PTreeAllocator allocator;
    AllocatorInit(&allocator, header);

    // sub-tree must be selected before FTreeFind().
    WFSKRN_FTREE_ASSERT(&header->tree[0] <= iter->subTree);
    WFSKRN_FTREE_ASSERT(iter->subTree <= &header->tree[MAX_SUB_TREE-1]);
    PTreeHdr* subTree = iter->subTree;
    if (subTree->nNumRecs == 0)
    {
        return WFSKRN_RESULT_FTREE_EMPTY;
    }

    ofs = PTreeFindLeafOfs(iter->subTree, &iter->ptree, nKey, &allocator);
    WFSKRN_FTREE_ASSERT(ofs < BLOCK_SIZE);
    WFSKRN_FTREE_ASSERT(header->sbaHdr.offset <= ofs);

    pLeaf = FTreeGetLeaf(header, ofs);
    iter->ptree.pLeaf = (PTreeLeaf*) pLeaf;

    key = SearchKey(pLeaf, nKey);
    if (strictSearch &&
        (IS_OUT_OF_RANGE(pLeaf, key) || *key != nKey))
    {
        return WFSKRN_RESULT_NOT_FOUND;
    }

    index = (key - &pLeaf->aKey[0]);
    if (index < 0)
    {
        // nKey < the minimum key.
        iter->ptree.aEntryIdx[0] = 0; // save the nearest key. (note: nKey < this key)

        iter->pLeaf = (FTreeLeaf*) iter->ptree.pLeaf; // [check]
        FTREE_GET_KEY(iter) = iter->pLeaf->aKey[0]; // [check]
        FTREE_GET_DATA(iter) = LEAF_GET_LEN(iter->pLeaf, 0); // [check]
        return WFSKRN_RESULT_NOT_FOUND;
    }

    iter->ptree.aEntryIdx[0] = index;

    // save data
    iter->pLeaf = (FTreeLeaf*) iter->ptree.pLeaf;
    FTREE_GET_KEY(iter) = iter->pLeaf->aKey[index];
    FTREE_GET_DATA(iter) = LEAF_GET_LEN(iter->pLeaf, index);

    return WFSKRN_RESULT_OK;
}

//
// Get the first key in this F-tree.
//
WFSKrnResult FTreeFirst(FTreeHdr* header, FTreeItr* iter)
{
    WFSKrnResult result;
    PTreeAllocator allocator;
    AllocatorInit(&allocator, header);

    result = PTreeFirst(iter->subTree, &iter->ptree, &allocator);
    if (result == WFSKRN_RESULT_OK)
    {
        iter->pLeaf = (FTreeLeaf*) iter->ptree.pLeaf;
        FTREE_GET_KEY(iter) = iter->pLeaf->aKey[0];
        FTREE_GET_DATA(iter) = LEAF_GET_LEN(iter->pLeaf, 0);
        result = WFSKRN_RESULT_OK;
    }
    return result;
}

//
// Get the next key.
// Must call FTreeFind() or FTreeFirst() before FTreeNext().
//
// [check] should be unified with PTreeNext.
//         parameterize LEAF_FAN_OUT and data format.
WFSKrnResult FTreeNext(FTreeHdr* header, FTreeItr* iter)
{
    PTreeHdr *pPTreeHdr = iter->subTree;
    PTreeItr *pPti = &iter->ptree;
    PTreeAllocator allocator;
    PTreeAllocator *pAllocator = &allocator;
    AllocatorInit(&allocator, header);

    WFSKrnResult nResult = WFSKRN_RESULT_OK;
    u32 *pEntryIdx = &pPti->aEntryIdx[0];
    ++(*pEntryIdx);
    u32 *aKey = pPti->pLeaf->aKey;
    u32 *pKey = &aKey[*pEntryIdx];
    if ((*pEntryIdx==FTREE_LEAF_FAN_OUT) || (*pKey==0)) {
        u32 *pEntryIdxEnd = &pPti->aEntryIdx[pPTreeHdr->nRootHeight];
        PTreeNode **ppNode = pPti->apNode;
        while(1) {
            ++pEntryIdx;
            if (pEntryIdx > pEntryIdxEnd) {
                return WFSKRN_RESULT_PTREE_ENTRY_NOT_FOUND;
            }
            aKey = (*ppNode)->aKey;
            ++ppNode;
            if (*pEntryIdx >= PTREE_NODE_FAN_OUT-1) {
                continue;
            }
            pKey = &aKey[*pEntryIdx];
            ++(*pEntryIdx);
            if (*pKey != 0) {
                break;
            }
        }
        --ppNode;
        u32 nOfs = (*ppNode)->aOfs[pKey - aKey + 1];
        WFSKRN_FTREE_ASSERT(nOfs < BLOCK_SIZE);
        while(1) {
            --pEntryIdx;
            *pEntryIdx = 0;
            if (ppNode == pPti->apNode) {
                break;
            }
            --ppNode;
            *ppNode = PTreeGetNode(pAllocator, nOfs);
            nOfs = (*ppNode)->aOfs[0];
            WFSKRN_FTREE_ASSERT(nOfs < BLOCK_SIZE);
        }
        pPti->pLeaf = PTreeGetLeaf(pAllocator, nOfs);
        WFSKRN_FTREE_ASSERT(pPti->pLeaf); // [check] debug code.
        aKey = pPti->pLeaf->aKey;
        pKey = &aKey[0];
    }
    pPti->nKey = *pKey;
    pPti->nData = LEAF_GET_LEN((FTreeLeaf*) pPti->pLeaf, pKey - aKey);
    return nResult;
}

// [check] should be refined.
static inline int GetMaxKey(u32 *aKey, u32 maxKeys)
{
    int count = (int) maxKeys;
    while (count)
    {
        if (aKey[count] != 0)
        {
            break;
        }
        --count;
    }

    if (count == 0 && aKey[count] == 0)
    {
        // pass through
        --count;
    }
    return count;
}

WFSKrnResult static GetPrev(PTreeHdr *pPTreeHdr, PTreeItr *pPti,
    PTreeAllocator *pAllocator, u32 leafFanout, u32** ppKey)
{
    WFSKrnResult nResult = WFSKRN_RESULT_OK;
    int count;
    u32 *pEntryIdx = pPti->aEntryIdx;

    if (pPTreeHdr->nNumRecs == 0)
    {
        return WFSKRN_RESULT_PTREE_ENTRY_NOT_FOUND;
    }

    // search the current leaf.
    WFSKRN_FTREE_ASSERT(*pEntryIdx != 0xffffffff);

    u32 *aKey = pPti->pLeaf->aKey;
    u32 *pKey;
    if (*pEntryIdx == 0)
    {
        // search up.
        u32 *pEntryIdxEnd = &pPti->aEntryIdx[pPTreeHdr->nRootHeight];
        PTreeNode **ppNode = pPti->apNode; // parent node.
        while(1) {
            ++pEntryIdx; // parent entry index.
            if (pEntryIdxEnd < pEntryIdx) {
                return WFSKRN_RESULT_PTREE_ENTRY_NOT_FOUND;
            }
            aKey = (*ppNode)->aKey; // parent "key"
            ++ppNode;
            if (*pEntryIdx == 0) // if parent index is zero, check the next parent.
            {
                continue;
            }
            --(*pEntryIdx);
            pKey = &aKey[*pEntryIdx];
            if (*pKey != 0) { // [check] necessary?
                break;
            }
        }
        // search down.
        --ppNode;
        u32 nOfs = (*ppNode)->aOfs[pKey - aKey];
        WFSKRN_FTREE_ASSERT(nOfs < BLOCK_SIZE);
        while(1) {
            --pEntryIdx;
            if (ppNode == pPti->apNode)
            {
                // select the max key in leaf.
                break;
            }
            --ppNode; // select child node.
            *ppNode = PTreeGetNode(pAllocator, nOfs);

            // select the max key in the child.
            count = GetMaxKey((*ppNode)->aKey, PTREE_NODE_FAN_OUT-2); //
            if (count < 0)
            {
                *pEntryIdx = 0; // [check] pass through must be refined.
            }
            else
            {
                *pEntryIdx = count+1; // ofs_index =  key_index + 1.
            }
            nOfs = (*ppNode)->aOfs[*pEntryIdx];
            WFSKRN_FTREE_ASSERT(nOfs < BLOCK_SIZE);
        }
        pPti->pLeaf = PTreeGetLeaf(pAllocator, nOfs);
        aKey = pPti->pLeaf->aKey;
        count = GetMaxKey(aKey, leafFanout-1);
        if (count < 0)
        {
            count = 0; // [check] pass through must be refined.
        }

        WFSKRN_FTREE_ASSERT(0 <= count && count <= leafFanout-1);
        *pEntryIdx = count;
        pKey = &aKey[count];
    }
    else
    {
        --(*pEntryIdx); // [check]
        pKey = &aKey[*pEntryIdx];
    }
    *ppKey = pKey;
    return nResult;
}

//
// Get the previous key in this P-tree.
// Must call PTreeFind() or PTreeLast() before PTreePrev().
//
WFSKrnResult PTreePrev(PTreeHdr *pPTreeHdr, PTreeItr *pPti, PTreeAllocator *pAllocator)
{
    WFSKrnResult nResult;
    u32* pKey;
    u32* aKey;

    nResult = GetPrev(pPTreeHdr, pPti, pAllocator, PTREE_LEAF_FAN_OUT, &pKey);
    if (nResult == WFSKRN_RESULT_OK)
    {
        aKey = pPti->pLeaf->aKey;
        pPti->nKey = *pKey;
        pPti->nData = pPti->pLeaf->aData[pKey - aKey];
    }
    return nResult;
}

//
// Get the previous key in this F-tree.
// Must call FTreeFind() or FTreeLast() before FTreePrev().
//
WFSKrnResult FTreePrev(FTreeHdr* header, FTreeItr* iter)
{
    WFSKrnResult nResult;
    PTreeItr *pPti = &iter->ptree;
    PTreeAllocator allocator;
    AllocatorInit(&allocator, header);

    u32* pKey;
    u32* aKey;
    nResult = GetPrev(iter->subTree, pPti, &allocator, FTREE_LEAF_FAN_OUT, &pKey);
    if (nResult == WFSKRN_RESULT_OK)
    {
        aKey = pPti->pLeaf->aKey;
        pPti->nKey = *pKey;
        pPti->nData = LEAF_GET_LEN((FTreeLeaf*) pPti->pLeaf, pKey - aKey);
    }
    return nResult;
}

//
// Get the last key in this P-tree.
//
WFSKrnResult PTreeLast(PTreeHdr *pPTreeHdr, PTreeItr *pPti, PTreeAllocator *pAllocator) {
    if (pPTreeHdr->nNumRecs == 0) {
        return WFSKRN_RESULT_PTREE_ENTRY_NOT_FOUND;
    }
    int index;
    u32 nLevel = pPTreeHdr->nRootHeight;
    u32 *pEntryIdx = &pPti->aEntryIdx[nLevel];
    u32 nOfs = (u32)pPTreeHdr->nRootOfs;
    WFSKRN_FTREE_ASSERT(nOfs < BLOCK_SIZE);

    while (nLevel) {
        --nLevel;
        PTreeNode *pNode = pPti->apNode[nLevel] = PTreeGetNode(pAllocator, nOfs);
        index = GetMaxKey(pNode->aKey, PTREE_NODE_FAN_OUT-2); // [check] should refine.
        nOfs = pNode->aOfs[index+1]; //*pEntryIdx
        WFSKRN_FTREE_ASSERT(nOfs < BLOCK_SIZE);
        *pEntryIdx = index+1;
        --pEntryIdx;
    }
    PTreeLeaf *pLeaf = pPti->pLeaf = PTreeGetLeaf(pAllocator, nOfs);

    index = GetMaxKey(pLeaf->aKey, PTREE_LEAF_FAN_OUT-1); // [check] should refine.
    if (index < 0)
    {
        // path-through
        index = 0;
    }
    *pEntryIdx = index;
    pPti->nKey = pLeaf->aKey[index];
    pPti->nData = pLeaf->aData[index];

    return WFSKRN_RESULT_OK;
}

//
// Get the last key in this F-tree.
//
// [check] must refine.
WFSKrnResult FTreeLast(FTreeHdr* header, FTreeItr* iter)
{
    return FTreeFind(header, iter, 0xffffffff, (WFSBool) FALSE); // find the max key.
}

//
// debug code
//
static WFSBool VerifySBA(FTreeHdr* header)
{
    SubBlockAllocatorHeader* sba = &header->sbaHdr;
    return VerifyMemoryBlockFTree(sba);
}

static WFSBool DoubleCheck(FTreeHdr* header, u32 key, int subTreeIndex)
{
    WFSKrnResult result;
    FTreeItr iter;
    WFSBool strictSearch;
    iter.subTree = &header->tree[subTreeIndex];

    strictSearch = (WFSBool) TRUE;
    result = FTreeFind(header, &iter, key, strictSearch); // find
    if (result == WFSKRN_RESULT_OK)
    {
        return (WFSBool) TRUE;
    }
    return (WFSBool) FALSE;
}

static WFSBool VerifyTree(FTreeHdr* header)
{
    u32 count;
    u32 key;
    u32 data;
    u32 lastKey;
    u32 lastData;
    FTreeItr iter;

    int subTreeIndex;
    for (subTreeIndex = 0; subTreeIndex < MAX_SUB_TREE; ++subTreeIndex)
    {
        iter.subTree = &header->tree[subTreeIndex];
        count = iter.subTree->nNumRecs;
        if (count == 0)
        {
            continue;
        }

        if (FTreeFirst(header, &iter) != WFSKRN_RESULT_OK)
        {
            return (WFSBool) FALSE;
        }

        key = FTREE_GET_KEY(&iter);
        if (DoubleCheck(header, key, subTreeIndex) == (WFSBool) FALSE)
        {
            return (WFSBool) FALSE;
        }

        // if (2 * 1024 * 1024 < FTREE_GET_KEY(&iter)) // the maximum area size is 16GB, for now.
        if (1024 * 1024 * 1024 < FTREE_GET_KEY(&iter)) // [check] 8TB
        {
            return (WFSBool) FALSE;
        }
        if (16 <= FTREE_GET_DATA(&iter))
        {
            return (WFSBool) FALSE;
        }
        --count;

        lastKey = FTREE_GET_KEY(&iter);
        lastData = FTREE_GET_DATA(&iter);
        while (count--)
        {
            if (FTreeNext(header, &iter) != WFSKRN_RESULT_OK)
            {
                return (WFSBool) FALSE;
            }
            if (FTREE_GET_KEY(&iter) <= key)
            {
                return (WFSBool) FALSE;
            }
            key = FTREE_GET_KEY(&iter);
            data = FTREE_GET_DATA(&iter);
            if (DoubleCheck(header, key, subTreeIndex) == (WFSBool) FALSE)
            {
                return (WFSBool) FALSE;
            }
            // if (2 * 1024 * 1024 < FTREE_GET_KEY(&iter)) // the maximum area size is 16GB, for now.
            if (1024 * 1024 * 1024 < FTREE_GET_KEY(&iter)) // [check] 8TB.
            {
                return (WFSBool) FALSE;
            }
            if (16 <= FTREE_GET_DATA(&iter))
            {
                return (WFSBool) FALSE;
            }
            if (key < lastKey + lastData + 1)
            {
                return (WFSBool) FALSE;
            }
            lastKey = key;
            lastData = data;
        }
    }
    return (WFSBool) TRUE;
}

WFSBool FTreeVerifyTree(FTreeHdr* header)
{
    return VerifyTree(header);
}

WFSBool VerifyMemoryBlockFTree(SubBlockAllocatorHeader* header)
{
    int numFreeBlock;
    FTreeBlock* free;
    WFSBool result = (WFSBool) FALSE;

    free = GET_FREEBLOCK(header);
    numFreeBlock = 0;

    while (free < END_OF_CACHE(header))
    {
        if (MAX_FREEBLOCK(header) < free->next)
        {
            goto VERIFY_ERROR;
        }
        if (free->num == 0)
        {
            goto VERIFY_ERROR;
        }
        if (MAX_FREEBLOCK(header) < free->num)
        {
            goto VERIFY_ERROR;
        }
        numFreeBlock += free->num;
        free = CACHE(header) + free->next;
    }

    result = (WFSBool) TRUE;

VERIFY_ERROR:
    return result;
}

#endif // BTREE_BASED_ALLOCATOR
