/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfskrn_FreeBlkAlloc.cpp

  Copyright 2007-2008 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_FreeBlkAlloc.cpp,v $
  Revision 1.43  2008/12/19 07:57:06  ueno
  Cleanup.

  Revision 1.42  2008/12/18 00:36:37  ueno
  Modified to always select free blocks considering fragmentation and distance from the specified reference address.

  Revision 1.41  2008/12/17 07:43:09  ueno
  Modified the priority of free extents.

  Revision 1.40  2008/12/17 06:51:24  ueno
  Fixed IsFarawayFromRefAddr() to identify a half of the area size is "far".

  Revision 1.39  2008/12/17 05:07:39  ueno
  Cleanup.
  Fixed MergeExtents() to check the length of the smaller extent.

  Revision 1.38  2008/12/15 04:14:40  ueno
  Cleanup.

  Revision 1.37  2008/12/15 02:40:07  ueno
  Added support for multi p-tree.

  Revision 1.36  2008/12/05 04:29:19  nakanose_jin
  fix error on Release Target.

  Revision 1.35  2008/12/02 04:55:13  ueno
  Added EPTree to support multi-PTree.

  Revision 1.33  2008/11/18 04:57:12  saito_tomoya
  Modified SetBit() and added SetBitArray().

  Revision 1.32  2008/11/12 13:34:17  ueno
  Fixed uninitialized value.

  Revision 1.31  2008/10/30 04:32:32  ueno
  Modified for release build.

  Revision 1.30  2008/10/27 13:17:28  ueno
  Modified to sequentially allocate blocks to improve performance.

  Revision 1.29  2008/10/27 05:44:57  ueno
  Modified FreeBlkAlloc() to get blocks near the refernce address when the number of blocks <= 5.

  Revision 1.28  2008/10/17 08:50:48  kondo_masahiro
  Added the new MEDIUM size category (40KB-320KB)

  Revision 1.27  2008/10/15 07:41:22  saito_tomoya
  fix for porting IOP

  Revision 1.26  2008/10/14 11:22:13  ueno
  Committed revision 1.22 again.
  - Modified FTreeInsert() to be able to split the block when creating a root node.
  - Modified FreeBlkAlloc() to improve performance.

  Revision 1.25  2008/10/06 23:16:17  ueno
  Fixed null pointer reference in InsertIntoThisSubTree().

  Revision 1.24  2008/10/03 10:22:46  ueno
  Updated debug function.

  Revision 1.23  2008/10/03 08:37:56  kondo_masahiro
  Fixed codes for porting IOP

  Revision 1.21  2008/09/30 05:55:52  ueno
  Cleanup.

  Revision 1.20  2008/09/30 05:21:16  ueno
  Improved performance of FreeBlkInsert().

  Revision 1.19  2008/09/30 04:48:41  ueno
  Reduced stack usage of FreeBlkInsert().
  Improved performance of FreeBlkNext() and Prev().

  Revision 1.18  2008/09/30 02:01:19  ueno
  Modified to support large disk ( < 32TB).

  Revision 1.17  2008/09/29 10:19:36  ueno
  Revised to support multi-B-tree-based allocator.

  Revision 1.16  2008/09/29 01:58:57  ueno
  Modified to call BCacheInvalidateBlks() in FreeBlkFree() to invalidate caches of user data.

  Revision 1.15  2008/09/28 23:32:04  kondo_masahiro
  Fixed codes to use medium size user block.

  Revision 1.15  2008/09/28 23:32:04  kondo_masahiro
  Fixed codes to use medium size user block.

  Revision 1.13  2008/09/02 07:02:49  kondo_masahiro
  Fixed several points for IOP compiler.

  Revision 1.12  2008/08/28 01:17:41  ueno
  Modified FreeBlkAlloc() to resurn WFSKRN_RESULT_DIRECTORY_QUOTA.

  Revision 1.11  2008/08/28 00:23:18  ueno
  Modified to skip pin count check in FreeBlkAlloc() and FreeBlkVerify().

  Revision 1.10  2008/08/13 04:13:28  ueno
  Modified FreeBlkAlloc() to allocate blocks near nRefBlkAdr.

  Revision 1.9  2008/08/13 00:12:09  paul
  Changed specification of nStride to include sizeof(nBlkAdr) and support negative strides

  Revision 1.8  2008/08/09 03:32:51  ueno
  Disabled runtime check.

  Revision 1.7  2008/08/09 02:01:41  ueno
  Cleanup.

  Revision 1.6  2008/08/09 00:40:09  ueno
  Fixed FreeBlkAlloc() to unpin and pin blocks correctly.

  Revision 1.5  2008/08/08 13:26:50  ueno
  Modified to unpin properly.

  Revision 1.4  2008/08/07 10:46:33  ueno
  Modified for BTREE_BASED_ALLOCATOR (now debugging).

  Revision 1.3  2008/08/02 11:52:40  ueno
  Revised the positions of FTreeHdr{} and FBAPTreeHdr{} in 8KB-blocks to protect hash code.

  Revision 1.2  2008/08/01 07:00:05  ueno
  Added copyrights.

  Revision 1.1  2008/08/01 06:57:39  ueno
  Initial Check-in.

*---------------------------------------------------------------------------*/
#include "wfskrn_Area.h"
#if BTREE_BASED_ALLOCATOR

#include "wfskrn_FTree.h"
#include "wfskrn_FreeBlkAlloc.h"

#define ROUNDUP32BYTE(val)   (((size_t)(val)+31)&~31)
#define ROUNDDOWN(val, unit) ((u32) (val) & ~((unit)-1))
#define ROUNDUP(val, unit)   ((u32) ((val) + ((unit)-1)) & ~((unit)-1))

#define FBA_PTREE_MIN_HEIGHT 1

#ifdef FBA_DEBUG_VERBOSE
WFSBool FBADebugVerbose = (WFSBool) FALSE; // enable debug message.
#endif //FBA_DEBUG_VERBOSE

typedef struct _CacheMap
{
    u32 diskAddr;
    void* cache;
} CacheMap;

void FBAUnpinCache(AreaInfo* area, u32 diskAddr)
{
    BCacheUnpin(area->pVolInfo, area->nAbsStartBlkAdr + diskAddr);
}

void FBAUnpinFTreeBlk(AreaInfo* area, FBATreeItr* iter)
{
    if (iter->ftreeHdr)
    {
        FBAUnpinCache(area, FBA_GET_FTREE_BLK(iter));
        iter->ftreeHdr = 0; // no f-tree is pinned.
    }
}

void FBAFreeCache(AreaInfo* area, u32 diskAddr)
{
    BCacheUnpin(area->pVolInfo, area->nAbsStartBlkAdr + diskAddr);
    BCacheInvalidate(area->pVolInfo, area->nAbsStartBlkAdr + diskAddr);
}

WFSKrnResult FBAGetCache(AreaInfo* area, u32 diskAddr, u32 flags, void** cache)
{
    WFSKrnResult result;

    BCacheEntry* pBce;
    result = BCacheAllocMetaBlk(area->pVolInfo, area->nAbsStartBlkAdr + diskAddr, flags, &pBce);
    if (result == WFSKRN_RESULT_OK)
    {
        *cache = pBce->pBlkPtr;
    }
    return result;
}

static void FBADirtyFtreeBlock(AreaInfo* area, FBATreeItr* iter)
{
#ifdef _DEBUG
    int index;
    EPTreeItr* epti = EPTreeGetLastIter((MEPTreeItr*)iter);
    index = epti->aEntryIdx[0];
    WFSKRN_FTREE_ASSERT(index < PTREE_LEAF_FAN_OUT);
    WFSKRN_FTREE_ASSERT(iter->ftreeAddr == epti->pLeaf->aData[index]);
#endif // _DEBUG
    BCacheDirty(area->pVolInfo, area->nAbsStartBlkAdr + iter->ftreeAddr);
}

static WFSKrnResult FBAGetAndPinFtreeBlock(AreaInfo* area, FBATreeItr* iter, u32 flags)
{
    int index;
    EPTreeItr* epti;
    WFSKrnResult result;
    void* cache;
    u8 ptreeHeight = iter->eptreeHdr[0]->height;

    epti = &iter->epti[ptreeHeight-1];
    index = epti->aEntryIdx[0];

    if (iter->ftreeAddr == epti->pLeaf->aData[index] &&
        iter->ftreeHdr)
    {
        // already pinned.
        return WFSKRN_RESULT_OK;
    }

    if (iter->ftreeHdr)
    {
        FBAUnpinCache(area, iter->ftreeAddr);
    }

    WFSKRN_FTREE_ASSERT(index < PTREE_LEAF_FAN_OUT);
    iter->ftreeAddr = epti->pLeaf->aData[index];

    result = FBAGetCache(area, iter->ftreeAddr, flags, &cache);
    if (result == WFSKRN_RESULT_OK)
    {
        iter->ftreeHdr = GET_FTREEHDR(cache);
    }
    return result;
}

void FBAItrOpen(AreaInfo* area, FBATreeItr* iter)
{
    iter->eptreeHdr[0] = GET_EPTREEHDR_FROM_AREAINFO(area);

    u32 i;
    for (i = 1; i < EPTREE_MAX_HEIGHT; ++i)
    {
        iter->eptreeHdr[i] = 0;
    }

    iter->ftreeHdr = 0;
    iter->isInserted = (WFSBool) FALSE;
    iter->height = iter->eptreeHdr[0]->height; // keep the height to know the bottom p-tree in this iter,
                                               // because another iter can change FBA tree height (iter->eptreeHdr[0]->height).
}

void FBAItrClose(AreaInfo* area, FBATreeItr* iter)
{
    u32 i;
    u8 ptreeHeight = iter->eptreeHdr[0]->height;
    ASSERT(ptreeHeight <= EPTREE_MAX_HEIGHT);

    iter->eptreeHdr[0] = GET_EPTREEHDR_FROM_AREAINFO(area);
    for (i = 1; i < ptreeHeight; ++i)
    {
        if (iter->eptreeHdr[i])
        {
            u32 diskAddr = EPTREE_GET_DATA(&iter->epti[i-1]);
            FBAUnpinCache(area, diskAddr);
            iter->eptreeHdr[i] = 0;
        }
    }

    FBAUnpinFTreeBlk(area, iter);
}

static WFSKrnResult FBAPTreeFind(AreaInfo* area, FBATreeItr* iter, u32 nKey)
{
    WFSKrnResult result = EPTreeFind(area, (MEPTreeItr*) iter, nKey);
    if (result == WFSKRN_RESULT_PTREE_ENTRY_NOT_FOUND)
    {
        // NOT_FOUND means the specified key does not match the keys in this P-tree,
        // but EPTreeFind() could find a candidate key.
        // The candidate is the nearest neighbor of the specified key and
        // the candidate key <= the specified key.
        result = WFSKRN_RESULT_OK;
    }
    return result;
}

static WFSKrnResult FBALookupStrict(AreaInfo* area, FBATreeItr* iter, WFSBlkAdr nBlkAdr, u32 nBlkLog2Size)
{
    WFSKrnResult result;

    // P-tree block is already pinned.
    result = FBAPTreeFind(area, iter, nBlkAdr);
    if (result == WFSKRN_RESULT_OK)
    {
        result = FBAGetAndPinFtreeBlock(area, iter, BCACHE_FLAG_READ | BCACHE_FLAG_PINNED);
        if (result == WFSKRN_RESULT_OK)
        {
            FTreeItr* fti = &iter->fti;
            fti->subTree = FTreeSelectSubTree(iter->ftreeHdr, nBlkLog2Size);
            result = FTreeFind(iter->ftreeHdr, fti, nBlkAdr, (WFSBool) TRUE);
        }
    }

    // the P-tree block is pinned.
    // an F-tree block is pinned.
    return result;
}

static WFSKrnResult FBALookup(AreaInfo* area, FBATreeItr* iter, WFSBlkAdr nBlkAdr, u32 nBlkLog2Size)
{
    WFSKrnResult result;

    // P-tree block is already pinned.
    result = FBAPTreeFind(area, iter, nBlkAdr);
    if (result != WFSKRN_RESULT_OK)
    {
        // FBAPTreeFind() case (1).
        result = EPTreePrev(area, (MEPTreeItr*) iter);
    }

    if (result == WFSKRN_RESULT_OK)
    {
        iter->isInserted = (WFSBool) FALSE; // reset
        result = FBAGetAndPinFtreeBlock(area, iter, BCACHE_FLAG_READ | BCACHE_FLAG_PINNED);
        if (result == WFSKRN_RESULT_OK)
        {
            FTreeItr* fti = &iter->fti;
            fti->subTree = FTreeSelectSubTree(iter->ftreeHdr, nBlkLog2Size);
            result = FTreeFind(iter->ftreeHdr, fti, nBlkAdr, (WFSBool) FALSE);
            if (result == WFSKRN_RESULT_NOT_FOUND ||
                result == WFSKRN_RESULT_FTREE_EMPTY)
            {
                result = FreeBlkPrev(area, iter);
            }
        }
    }

    // the P-tree block is pinned.
    // an F-tree block is pinned.
    return result;
}

// search where to insert the new key.
static WFSKrnResult FBALookupBeforeInsert(AreaInfo* area, FBATreeItr* iter, WFSBlkAdr nBlkAdr, u32 nBlkLog2Size)
{
    WFSKrnResult result;

    // the P-tree block is already pinned.
    result = FBAPTreeFind(area, iter, nBlkAdr);
    if (result != WFSKRN_RESULT_OK)
    {
        EPTreeItr* epti;
        /*
         (ex)
             +-----------------+
             |      P-Tree     |
             |key: 10 16 19... |
             |                 |
             +-----------------+
          A
          |           |
      nBlkAdr (5)     | insert nBlkAdr.
                      |
                      V
             +-------------------+
             |       P-Tree      |
             |key: 5 16 19...    |
             |     A             |
             +-----|-------------+
                   |
                 insert the head of the leaf after FBALookupBeforeInsert().
        */

        epti = EPTreeGetLastIter((MEPTreeItr*) iter);
        WFSKRN_FTREE_ASSERT(nBlkAdr < epti->pLeaf->aKey[0]);
#ifdef _DEBUG
        if (0 < epti->aEntryIdx[1])
        {
            WFSKRN_FTREE_ASSERT(epti->apNode[0]->aKey[epti->aEntryIdx[1]-1] <= nBlkAdr);
        }
#endif // _DEBUG
        epti->aEntryIdx[0] = 0;
        epti->pLeaf->aKey[0] = nBlkAdr;

        epti->nKey = epti->pLeaf->aKey[0];
        epti->nData = epti->pLeaf->aData[0];
        result = WFSKRN_RESULT_OK;
    }

    if (result == WFSKRN_RESULT_OK)
    {
        FTreeItr* fti = &iter->fti;
        result = FBAGetAndPinFtreeBlock(area, iter, BCACHE_FLAG_READ | BCACHE_FLAG_PINNED);
        if (result != WFSKRN_RESULT_OK)
        {
            return result;
        }
        fti->subTree = FTreeSelectSubTree(iter->ftreeHdr, nBlkLog2Size);
        result = FTreeFind(iter->ftreeHdr, fti, nBlkAdr, (WFSBool) FALSE);
    }

    iter->isInserted = (WFSBool) FALSE;

    // the P-tree block is pinned.
    // an F-tree block is pinned.
    return result;
}

WFSKrnResult InsertNewChild(AreaInfo* area, FBATreeItr* iter,
    FTreeExtent* extent, u32 nBlkLog2Size)
{
    WFSKrnResult result;
    u32 key = extent->start;
    u32 data = extent->len / (1<<nBlkLog2Size) - 1;

    // the P-tree block is pinned.
    // one F-tree block (left child) is pinned.

    // insert the new control block to the freeblock allocator tree.
    result = EPTreeInsert(area, (MEPTreeItr*) iter, iter->fti.splitKey, iter->fti.newChild);
    if (result != WFSKRN_RESULT_OK)
    {
        goto INSERTNEWCHILD_ERROR;
    }

    iter->height = iter->eptreeHdr[0]->height; // the height may be changed.

    // insert the extent into the freeblock allocator tree.
    if (key < iter->fti.splitKey)
    {
        FTreeHdr* ftreeHdr = iter->ftreeHdr;
        result = FTreeInsert(area, ftreeHdr, &iter->fti, key, data);
        BCacheDirty(area->pVolInfo, area->nAbsStartBlkAdr + iter->ftreeAddr); // pass through the debug code in DirtyFtreeBlock().
    }
    else
    {
        void* cache;
        FTreeHdr* newChildHdr;
        result = FBAGetCache(area, iter->fti.newChild,
            BCACHE_FLAG_DIRTY | BCACHE_FLAG_READ | BCACHE_FLAG_PINNED, &cache);
        if (result != WFSKRN_RESULT_OK)
        {
            goto INSERTNEWCHILD_ERROR;
        }

        newChildHdr = GET_FTREEHDR(cache);
        iter->fti.subTree = FTreeSelectSubTree(newChildHdr, nBlkLog2Size); // [check]

        result = FTreeInsert(area, newChildHdr, &iter->fti, key, data);
        FBAUnpinCache(area, iter->fti.newChild); // unpin the new child block.
    }

    // remove the address of the new control block from the freeblock allocator tree,
    // to prevent the block from being allocated as a freeblock.
    result = FreeBlkDelete(area, iter,
        iter->fti.newChild, 1, FTreeBlkLog2Ofs[iter->fti.newChildSubTreeType]);

INSERTNEWCHILD_ERROR:
    // the P-tree block is pinned.
    // an F-tree block (left child) is pinned.
    return result;
}

static inline int GetSubTreeType(WFSBlkAdr len)
{
    int subTreeType;

    if (len < (1<<FTREE_LOG2_OFS_B))
    {
        WFSKRN_FTREE_ASSERT((1<<FTREE_LOG2_OFS_A) <= len);
        subTreeType = FTREE_TYPE_A;
    }
    else if (len < (1<<FTREE_LOG2_OFS_C))
    {
        subTreeType = FTREE_TYPE_B;
    }
    else if (len < (1<<FTREE_LOG2_OFS_D))
    {
        subTreeType = FTREE_TYPE_C;
    }
    else if (len < (1<<FTREE_LOG2_OFS_E))
    {
        subTreeType = FTREE_TYPE_D;
    }
    else if (len < (1<<FTREE_LOG2_OFS_F))
    {
        subTreeType = FTREE_TYPE_E;
    }
    else if (len < (1<<FTREE_LOG2_OFS_G))
    {
        subTreeType = FTREE_TYPE_F;
    }
    else
    {
        subTreeType = FTREE_TYPE_G;
    }

    return subTreeType;
}

static inline WFSKrnResult CopyFBAIter(AreaInfo* area, FBATreeItr* src, FBATreeItr* dst, u32 flags)
{
    WFSKrnResult result;
    void* cache;
    u8 ptreeHeight;
    u8 level;

    ptreeHeight = dst->eptreeHdr[0]->height;
    for (level = 1; level < ptreeHeight; ++level)
    {
        // unpin nodes.
        if (dst->eptreeHdr[level])
        {
            FBAUnpinCache(area, dst->eptreeHdr[level]->addr);
        }
    }
    if (dst->ftreeHdr)
    {
        FBAUnpinCache(area, dst->ftreeAddr);
    }

    *dst = *src;
    // the root P-tree block is pinned.

    for (level = 1; level < ptreeHeight; ++level)
    {
        // pin nodes.
        result = FBAGetCache(area, dst->eptreeHdr[level]->addr, flags, &cache);
    }
    return FBAGetCache(area, dst->ftreeAddr, flags, &cache);
}

static WFSKrnResult GetNeighbors(AreaInfo* area, FBATreeItr* iter, WFSBlkAdr nBlkAdr,
    WFSBlkAdr nBlkLen, FTreeExtentInfo* prev, FTreeExtentInfo* next, u32 nBlkLog2Size, WFSKrnResult resultLookup)
{
    WFSKrnResult result;
    FBAItrOpen(area, &prev->iter);
    FBAItrOpen(area, &next->iter);

    // search neighbors.
    result = CopyFBAIter(area, iter, &prev->iter, BCACHE_FLAG_READ | BCACHE_FLAG_PINNED);
    if (result != WFSKRN_RESULT_OK)
    {
        return result;
    }
    result = CopyFBAIter(area, iter, &next->iter, BCACHE_FLAG_READ | BCACHE_FLAG_PINNED);
    if (result != WFSKRN_RESULT_OK)
    {
        FBAItrClose(area, &prev->iter);
        return result;
    }

    next->len = 0;
    prev->len = 0;

    if (resultLookup != WFSKRN_RESULT_OK)
    {
        // prev extent cat be in the previous f-tree leaf.
        result = FreeBlkPrev(area, &prev->iter);
    }

    if (result == WFSKRN_RESULT_OK)
    {
        FTreeItr* fti = &prev->iter.fti;
        WFSBlkAdr len = (FTREE_GET_DATA(fti) + 1) * (1<<nBlkLog2Size);
        if (FTREE_GET_KEY(fti) + len == nBlkAdr)
        {
            prev->start = FTREE_GET_KEY(fti);
            prev->len = len;
            if (nBlkLog2Size == FTREE_LOG2_OFS_G &&
                (1<<FTREE_LOG2_OFS_END) < nBlkLen + prev->len)
            {
                prev->len = 0;
            }
        }
    }

    result = WFSKRN_RESULT_OK;
    if (resultLookup == WFSKRN_RESULT_OK ||
        resultLookup == WFSKRN_RESULT_FTREE_EMPTY)
    {
        result = FreeBlkNext(area, &next->iter);
    }
    /*
    else
    {           F-tree leaf                      F-tree leaf
             +---------------+               +---------------+
             |               |               |               |
             |1st_Key        |      or       |     empty     |
             +-A-------------+               +-A-------------+
           A   |                               |      A
           |   +------ iter has this key. -----+      |
        nBlkAdr                                    nBlkAdr
    }
    */
    if (result == WFSKRN_RESULT_OK)
    {
        FTreeItr* fti = &next->iter.fti;
        // next extent
        if (nBlkAdr + nBlkLen == FTREE_GET_KEY(fti))
        {
            next->start = FTREE_GET_KEY(fti);
            next->len = (FTREE_GET_DATA(fti) + 1) * (1<<nBlkLog2Size);
            if (nBlkLog2Size == FTREE_LOG2_OFS_G &&
                (1<<FTREE_LOG2_OFS_END) < nBlkLen + next->len)
            {
                next->len = 0;
            }
        }

    }

    // keep pinning the prev and next extent.

    return WFSKRN_RESULT_OK;
}

static inline WFSBool IsLastKey(FTreeItr* fti)
{
    u32 index = fti->ptree.aEntryIdx[0];
    return (WFSBool) (fti->subTree->nNumRecs == 0 ||
                      index == FTREE_LEAF_FAN_OUT - 1 || fti->pLeaf->aKey[index+1] == 0);
}

static WFSKrnResult InsertIntoThisSubTree(AreaInfo* area, FBATreeItr* iter, WFSBlkAdr nBlkLog2Size,
    FTreeExtent* newExtent, FTreeExtentInfo* prev, FTreeExtentInfo* next)
{
    WFSKrnResult result;
    FTreeItr* fti = &iter->fti;

    if (prev && 0 < prev->len)
    {
        if (next && 0 < next->len)
        {
            result = FTreeDelete(next->iter.ftreeHdr, &next->iter.fti, next->start);
            if (result != WFSKRN_RESULT_OK)
            {
                return result;
            }
            FBADirtyFtreeBlock(area, &next->iter);
        }
        result = FTreeModify(prev->iter.ftreeHdr, &prev->iter.fti, prev->start,
            prev->start, newExtent->len / (1<<nBlkLog2Size) - 1);
        FBADirtyFtreeBlock(area, &prev->iter);
    }
    else if (next && 0 < next->len)
    {
        if (IsLastKey(fti))
        {
            result = FTreeDelete(next->iter.ftreeHdr, &next->iter.fti, next->start);
            if (result != WFSKRN_RESULT_OK)
            {
                return result;
            }
            FBADirtyFtreeBlock(area, &next->iter);

#ifdef FBA_DEBUG_VERBOSE
            if (FBADebugVerbose)
            {
                osTPrintf("*** connect to the next extent ***\n");
            }
#endif // FBA_DEBUG_VERBOSE
            result = FreeBlkInsert(area, iter, newExtent->start, newExtent->len); // [check] recursive call.
        }
        else
        {
            WFSKRN_FTREE_ASSERT(newExtent->start < next->start);
            result = FTreeModify(iter->ftreeHdr, fti, next->start, newExtent->start, newExtent->len / (1<<nBlkLog2Size) - 1); // modify next start and len.
            FBADirtyFtreeBlock(area, iter);
        }
    }
    else
    {
        result = FTreeInsert(area, iter->ftreeHdr, fti, newExtent->start, newExtent->len / (1<<nBlkLog2Size) - 1); // [check]
        FBADirtyFtreeBlock(area, iter);
        if (result == WFSKRN_RESULT_FTREE_RETRY)
        {
            // WFSKRN_FTREE_ASSERT(FTreeVerifyTree(iter->ftreeHdr));

            // The F-tree block was full and split into 2 F-tree blocks.
            // newExtent is not yet inserted.

            // insert the newExtent.
            result = InsertNewChild(area, iter, newExtent, nBlkLog2Size);

            // WFSKRN_FTREE_ASSERT(FTreeVerifyTree(iter->ftreeHdr));
        }
    }
    return result;
}

static inline void ConnectToPrevExtent(AreaInfo* area, FBATreeItr* iter, FTreeExtent* newExtent,
    FTreeExtentInfo* prev, WFSBlkAdr minBlkLenUpperSubTuree)
{
    if (prev->alignment)
    {
        if (minBlkLenUpperSubTuree <= (prev->len + newExtent->len))
        {
            prev->len = 0; // cannot connect.
            return;
        }
    }

    newExtent->start = prev->start;
    newExtent->len += prev->len;
}

static inline void ConnextToNextExtent(AreaInfo* area, FBATreeItr* iter, FTreeExtent* newExtent,
    FTreeExtentInfo* next, WFSBlkAdr minBlkLenUpperSubTuree)
{
    if (next->alignment)
    {
        if (minBlkLenUpperSubTuree <= (next->len + newExtent->len))
        {
            next->len = 0; // cannot connect.
            return;
        }
    }

    newExtent->len += next->len;
}

static void CheckAlignment(FTreeExtentInfo* prev, FTreeExtentInfo* next, WFSBlkAdr minBlkLenUpperSubTuree)
{
    prev->alignment = (WFSBool) FALSE;
    next->alignment = (WFSBool) FALSE;

    if (0 < prev->len)
    {
        WFSBlkAdr end = prev->start + prev->len;
        if (end % minBlkLenUpperSubTuree == 0)
        {
            prev->alignment = (WFSBool) TRUE;
        }
    }

    if (0 < next->len)
    {
        WFSBlkAdr start = next->start;
        if (start % minBlkLenUpperSubTuree == 0)
        {
            next->alignment = (WFSBool) TRUE;
        }
    }
}

WFSKrnResult InsertIntoTopTree(AreaInfo* area, FBATreeItr* iter, FTreeExtent* newExtent,
    FTreeExtentInfo* prev, FTreeExtentInfo* next, u32 nBlkLog2Size, WFSBlkAdr minBlkLenUpperSubTuree)
{
    WFSKrnResult result = WFSKRN_RESULT_UNKNOWN;
    FTreeExtent extent = *newExtent;
    FTreeExtentInfo* prevExtent = NULL;
    FTreeExtentInfo* nextExtent = NULL;
    // insert this subtree.
    while (extent.start < newExtent->start + newExtent->len)
    {
        extent.len = (newExtent->start - extent.start) + newExtent->len;
        if (minBlkLenUpperSubTuree < extent.len)
        {
            WFSBlkAdr end;
            end = ROUNDDOWN(extent.start + minBlkLenUpperSubTuree, minBlkLenUpperSubTuree);
            extent.len = end - extent.start;
        }

        if (extent.start == newExtent->start && 0 < prev->len)
        {
            prevExtent = prev;
        }
        else
        {
            prevExtent = NULL;
        }
        if (extent.start + extent.len == newExtent->start + newExtent->len && 0 < next->len)
        {
            nextExtent = next;
        }
        else
        {
            nextExtent = NULL;
        }

        result = InsertIntoThisSubTree(area, iter, nBlkLog2Size, &extent, prevExtent, nextExtent);
        if (result != WFSKRN_RESULT_OK)
        {
            break;
        }
        extent.start += extent.len;
    }

    return result;
}

static WFSKrnResult Insert(AreaInfo* area, FBATreeItr* iter, WFSBlkAdr nBlkAdr, WFSBlkAdr nBlkLen, u32 nBlkLog2Size)
{
    WFSKrnResult result;
    WFSBlkAdr minBlkLenUpperSubTuree;
    u32 subTreeIdx;
    FTreeExtent newExtent;
    WFSBool InsertUpperTree = (WFSBool) FALSE;

    subTreeIdx = GetSubTreeType(nBlkLen);
    minBlkLenUpperSubTuree = (1<<FTreeBlkLog2Ofs[subTreeIdx+1]);

    result = FBALookupBeforeInsert(area, iter, nBlkAdr, nBlkLog2Size);
    if (result == WFSKRN_RESULT_OK ||
        result == WFSKRN_RESULT_NOT_FOUND ||
        result == WFSKRN_RESULT_FTREE_EMPTY)
    {
        FTreeExtentInfo prev;
        FTreeExtentInfo next;
        result = GetNeighbors(area, iter, nBlkAdr, nBlkLen, &prev, &next, nBlkLog2Size, result);
        if (result != WFSKRN_RESULT_OK)
        {
            return result;
        }
        CheckAlignment(&prev, &next, minBlkLenUpperSubTuree);

        newExtent.start = nBlkAdr;
        newExtent.len = nBlkLen;

        if (0 < prev.len)
        {
            if (0 < next.len)
            {
                if (next.alignment)
                {
                    ConnectToPrevExtent(area, iter, &newExtent, &prev, minBlkLenUpperSubTuree);
                    ConnextToNextExtent(area, iter, &newExtent, &next, minBlkLenUpperSubTuree);
                }
                else
                {
                    ConnextToNextExtent(area, iter, &newExtent, &next, minBlkLenUpperSubTuree);
                    ConnectToPrevExtent(area, iter, &newExtent, &prev, minBlkLenUpperSubTuree);
                }
            }
            else
            {
                ConnectToPrevExtent(area, iter, &newExtent, &prev, minBlkLenUpperSubTuree);
            }
        }
        else if (0 < next.len)
        {
            ConnextToNextExtent(area, iter, &newExtent, &next, minBlkLenUpperSubTuree);
        }

        if (nBlkLog2Size == FTREE_LOG2_OFS_G)
        {
            result = InsertIntoTopTree(area, iter, &newExtent, &prev, &next, nBlkLog2Size, minBlkLenUpperSubTuree);
        }
        else if (newExtent.len < minBlkLenUpperSubTuree)
        {
            result = InsertIntoThisSubTree(area, iter, nBlkLog2Size, &newExtent, &prev, &next);
        }
        else
        {
            if (0 < prev.len)
            {
                result = FTreeDelete(prev.iter.ftreeHdr, &prev.iter.fti, prev.start);
                FBADirtyFtreeBlock(area, &prev.iter);
            }
            if (0 < next.len)
            {
                result = FTreeDelete(next.iter.ftreeHdr, &next.iter.fti, next.start);
                FBADirtyFtreeBlock(area, &next.iter);
            }
            InsertUpperTree = (WFSBool) TRUE;
        }

        FBAItrClose(area, &prev.iter);
        FBAItrClose(area, &next.iter);
    }

    if (InsertUpperTree)
    {
        // insert the upper subtree.
        result = FreeBlkInsert(area, iter, newExtent.start, newExtent.len);
    }
    return result;
}

WFSKrnResult FreeBlkInsert(AreaInfo* area, FBATreeItr* iter, WFSBlkAdr nBlkAdr, WFSBlkAdr nBlkLen)
{
    WFSKrnResult result = WFSKRN_RESULT_UNKNOWN;
    u32 subTreeIdx;
    u32 nBlkLog2Size;
    WFSBlkAdr start;
    WFSBlkAdr end;
    WFSBlkAdr subTreeMinBlkLen;

    WFSKRN_FTREE_ASSERT(0 < nBlkLen);

#ifdef FBA_DEBUG_VERBOSE
    if (FBADebugVerbose)
    {
        osTPrintf("[INS] start %5u (len %4u)\n", nBlkAdr, nBlkLen);
    }
#endif //  FBA_DEBUG_VERBOSE

    // the EP-tree block is pinned.

    if (!iter->eptreeHdr)
    {
        return WFSKRN_RESULT_INVALID;
    }

    /*
        +------------------+                               +-------------+
        |   prev extent    |                               | next extent |
        +------------------+                               +-------------+
                                         A Insert
                                         |
                             alignment   |     alignment
                                  :      |        :
                             +----:---------------:------+
                             |head:     body      : tail |
                             +----:---------------:------+
                                  :               :
                                headEnd        tailStart

    */

    subTreeIdx = GetSubTreeType(nBlkLen);
    nBlkLog2Size = FTreeBlkLog2Ofs[subTreeIdx];
    subTreeMinBlkLen = (1<<nBlkLog2Size);

    start = ROUNDUP(nBlkAdr, subTreeMinBlkLen);
    end = ROUNDDOWN(nBlkAdr + nBlkLen, subTreeMinBlkLen);
    if (start < end)
    {
        // body.
        result = Insert(area, iter, start, end - start, nBlkLog2Size);
        if (result != WFSKRN_RESULT_OK)
        {
            // [check] rollback. // NYI.
            return result;
        }
    }

    if (start == nBlkAdr &&
        end == nBlkAdr + nBlkLen)
    {
        return result;
    }

    WFSBlkAdr headEnd = start;
    WFSBlkAdr tailStart = end;

    // insert head and tail into lower subtrees.
    while (subTreeIdx != 0)
    {
        --subTreeIdx;
        nBlkLog2Size = FTreeBlkLog2Ofs[subTreeIdx];
        subTreeMinBlkLen = (1<<nBlkLog2Size);

        // head
        start = ROUNDUP(nBlkAdr /* headStart */, subTreeMinBlkLen);
        WFSKRN_FTREE_ASSERT(headEnd == ROUNDDOWN(headEnd, subTreeMinBlkLen));
        if (start < headEnd)
        {
            result = Insert(area, iter, start, headEnd - start, nBlkLog2Size);
            if (result != WFSKRN_RESULT_OK)
            {
                // [check] rollback. // NYI.
                return result;
            }
            headEnd = start;
        }

        // tail
        WFSKRN_FTREE_ASSERT(tailStart == ROUNDUP(tailStart, subTreeMinBlkLen));
        end = ROUNDDOWN(nBlkAdr + nBlkLen /* tailEnd */, subTreeMinBlkLen);
        if (tailStart < end)
        {
            result = Insert(area, iter, tailStart, end - tailStart, nBlkLog2Size);
            if (result != WFSKRN_RESULT_OK)
            {
                // [check] rollback. // NYI.
                return result;
            }
            tailStart = end;
        }
    }

    iter->isInserted = (WFSBool) TRUE; // leaves or nodes are modified.

    // the P-tree block is pinned.
    // an F-tree block is pinned.

    WFSKRN_FTREE_ASSERT(FTreeVerifyTree(iter->ftreeHdr));
    return result;
}

static WFSKrnResult DivideBlock(WFSBlkAdr existingKey, WFSBlkAdr keyBlkLen, u32 nSubTreeBlkLog2Size,
    WFSBlkAdr nBlkAdr, WFSBlkAdr nBlkLen, RemainderExtents* remainder)
{
    WFSBlkAdr len;
    if (existingKey == nBlkAdr)
    {
        // no head.
        remainder->head.len = 0;
        remainder->head.start = 0;
        remainder->fractionHead.len = 0;
        remainder->fractionHead.start = 0;
    }
    else
    {
        remainder->head.start = existingKey;
        remainder->head.len = ROUNDDOWN(nBlkAdr - existingKey, (1<<nSubTreeBlkLog2Size));
        remainder->fractionHead.len = (nBlkAdr - existingKey) % (1<<nSubTreeBlkLog2Size);
        remainder->fractionHead.start = remainder->head.start + remainder->head.len;
    }

    len = (existingKey + keyBlkLen) - (nBlkAdr + nBlkLen);
    if (len == 0)
    {
        // no tail
        remainder->tail.len = 0;
        remainder->fractionTail.len = 0;
    }
    else
    {
        remainder->fractionTail.start = nBlkAdr + nBlkLen;

        remainder->fractionTail.len = ROUNDUP(nBlkAdr + nBlkLen, (1<<nSubTreeBlkLog2Size)) - (nBlkAdr + nBlkLen);
        remainder->tail.start = remainder->fractionTail.start + remainder->fractionTail.len;
        remainder->tail.len = existingKey + keyBlkLen - (remainder->tail.start);
    }

    return WFSKRN_RESULT_OK;
}

static WFSKrnResult DeleteFTreeBlock(AreaInfo* area, FBATreeItr* iter)
{
    FTreeItr* fti;
    EPTreeItr* epti;
    WFSKrnResult result = WFSKRN_RESULT_NOT_EMPTY;
    PTreeHdr* subTree;
    int index;
    u32 addrCb;
    FBATreeItr insertIter;

    epti = EPTreeGetLastIter((MEPTreeItr*) iter);
    if (epti->nKey == 0) // The 1st F-tree is never removed.
    {
        return WFSKRN_RESULT_NOT_EMPTY;
    }

    fti = &iter->fti;
    // check other sub-trees.
    for (index = 0; index < MAX_SUB_TREE; ++index)
    {
        subTree = &iter->ftreeHdr->tree[index];
        if (fti->subTree == subTree)
        {
            continue;
        }
        if (subTree->nNumRecs != 0)
        {
            return WFSKRN_RESULT_NOT_EMPTY;
        }
    }

    // all subtree are empty.
#ifdef FBA_DEBUG_VERBOSE
    if (FBADebugVerbose)
    {
        osTPrintf("DeleteFTreeBlock(): delete control block\n");
    }
#endif //FBA_DEBUG_VERBOSE
    index = epti->aEntryIdx[0];
    addrCb = epti->pLeaf->aData[index];

    WFSBlkAdr freeBlks[(EPTREE_MAX_HEIGHT-1) * 2];
    memset(freeBlks, 0, sizeof(freeBlks));
    result = EPTreeDelete(area, (MEPTreeItr*) iter, EPTREE_GET_KEY(epti), freeBlks);
    FBAFreeCache(area, addrCb);
    iter->ftreeHdr = 0;
    // unpin ftreeHdr here.

    // the P-tree block is pinned.
    FBAItrOpen(area, &insertIter);

    // insert  F-tree block into Freeblock allocator as a free block.
    result = FreeBlkInsert(area, &insertIter, addrCb, 1);
    for (index = 0; index < 2 * (EPTREE_MAX_HEIGHT-1) && result == WFSKRN_RESULT_OK; ++index)
    {
        if (freeBlks[index] != 0)
        {
            // no need to increment header->tatalBlks, because
            // both a meta-block and freeblocks are in FBA tree.
            result = FreeBlkInsert(area, &insertIter, freeBlks[index], 1);
        }
    }

    FBAItrClose(area, &insertIter);

    // must not increment header->totalBlks here,
    // when a control block is changed to a freeblock.
    return result;
}

static WFSKrnResult DeleteBody(AreaInfo* area, FBATreeItr* iter, u32 existingKey)
{
    WFSKrnResult result;
    FTreeItr* fti = &iter->fti;

    if (fti->subTree->nNumRecs == 1)
    {
        result = DeleteFTreeBlock(area, iter);
        if (result != WFSKRN_RESULT_NOT_EMPTY)
        {
            // If result == WFSKRN_RESULT_OK,
            // the control block including existingKey was removed, and
            // inserted the block address into the freeblock tree as a free block.
            return result;
        }
    }

    result = FTreeDelete(iter->ftreeHdr, fti, existingKey);
    if (result != WFSKRN_RESULT_OK)
    {
        WFSKRN_FTREE_ASSERT(FTreeVerifyTree(iter->ftreeHdr));
        return result;
    }
    FBADirtyFtreeBlock(area, iter);

    return result;
}

WFSKrnResult FreeBlkDelete(AreaInfo* area,
    FBATreeItr* iter, WFSBlkAdr nBlkAdr, WFSBlkAdr nBlkLen, u32 nBlkLog2Size)
{
    WFSKrnResult result;
    FTreeItr* fti = &iter->fti;
    u32 nSubTreeBlkLog2Size;

    // the P-tree block is pinned.
    // F-tree may be pinned.

    if (!iter->eptreeHdr)
    {
        return WFSKRN_RESULT_INVALID;
    }

    // find f-tree block.
    result = FBALookup(area, iter, nBlkAdr, nBlkLog2Size);
    if (result == WFSKRN_RESULT_OK)
    {
        RemainderExtents remainder;
        u32 existingKey = FTREE_GET_KEY(fti);
        u32 existingData = FTREE_GET_DATA(fti);
        WFSBlkAdr keyBlkLen;
        int treeType;

        /*
                         alignment          alignment          alignemnt          alienment          alignment
                             :                  :                  :                  :                  :
                             :        nBlkAdr   :                  :                  :                  :
                             :           |      :       nBlkLen    :                  :                  :
                             :           |<---------------------------------->:       :                  :
         key0                :           V      :                  :          :       :                  :
          |                  :           +------------------------------------+       :                  :
          V                  :           |      :                  :          |       :                  :
          +------------------:-----------|------:------------------:----------|-------:------------------+
          |//////////////////:///////////|//////://////////////////://////////|///////://////////////////|
          +------------------:-----------|------:------------------:----------|-------:------------------+
          :                  :           |      :                  :          |       :                  :
          :                  :           +------------------------------------+       :                  :
          :                  :           :      :                  :          :       :                  :
          :                  :           :      :                  :          :       :                  :
          :                  :           :      :                  :          :       :                  :
          <-----------------><----------><-----------------------------------><------><----------------->:
          A      head          fraction              to be deleted.           fraction       tail        :
          |                    head                                           tail                       :
          |                                                                                              :
          |<-------------------------------------------------------------------------------------------->:
          |                   (existingData + 1) * (1<<nSubTreeBlkLog2Size)
        existingKey

        */

        treeType = GET_SUBTREE_TYPE(iter->ftreeHdr, &iter->fti);
        nSubTreeBlkLog2Size = FTreeBlkLog2Ofs[treeType]; // [check] should refine.
        keyBlkLen = (existingData + 1) * (1<<nSubTreeBlkLog2Size);

        WFSKRN_FTREE_ASSERT(existingKey <= nBlkAdr);
        if (existingKey + keyBlkLen < nBlkAdr + nBlkLen)
        {
            result = WFSKRN_RESULT_NOT_FOUND;
            goto FBA_DELETE_RESULT;
        }

        DivideBlock(existingKey, keyBlkLen, nSubTreeBlkLog2Size, nBlkAdr, nBlkLen, &remainder);

        result = DeleteBody(area, iter, existingKey);
        if (result != WFSKRN_RESULT_OK)
        {
            goto FBA_DELETE_RESULT;
        }

        FBA_STRICT_ASSERT(FTreeVerifyTree(iter->ftreeHdr));

        if (0 < remainder.head.len)
        {
            // [check] should refine using FTreeInsert().
            result = FreeBlkInsert(area, iter, remainder.head.start, remainder.head.len);
        }

        if (0 < remainder.tail.len)
        {
            result = FreeBlkInsert(area, iter, remainder.tail.start, remainder.tail.len);
        }

        if (0 < remainder.fractionHead.len)
        {
            result = FreeBlkInsert(area, iter, remainder.fractionHead.start, remainder.fractionHead.len);
        }

        if (0 < remainder.fractionTail.len)
        {
            result = FreeBlkInsert(area, iter, remainder.fractionTail.start, remainder.fractionTail.len);
        }
    }

FBA_DELETE_RESULT:
    return result;
}

WFSKrnResult FreeBlkFirst(AreaInfo* area, FBATreeItr* iter, u32 nSubTreeBlkLog2Size)
{
    WFSKrnResult result;

    // the P-tree block is pinned.

    if (!iter->eptreeHdr)
    {
        return WFSKRN_RESULT_INVALID;
    }

    result = EPTreeFirst(area, (MEPTreeItr*) iter);
    while (result == WFSKRN_RESULT_OK)
    {
        FTreeItr* fti = &iter->fti;
        result = FBAGetAndPinFtreeBlock(area, iter, BCACHE_FLAG_READ | BCACHE_FLAG_PINNED);
        if (result != WFSKRN_RESULT_OK)
        {
            break;
        }

        // search F-tree.
        memset(fti->ptree.aEntryIdx, 0, sizeof(fti->ptree.aEntryIdx));
        fti->subTree = FTreeSelectSubTree(iter->ftreeHdr, nSubTreeBlkLog2Size);
        result = FTreeFirst(iter->ftreeHdr, fti);
        if (result == WFSKRN_RESULT_OK)
        {
            break;
        }

        // 1st F-tree may be empty.
        // (It must not be deleted even when it is empty.)
        result = EPTreeNext(area, (MEPTreeItr*) iter);
    }

    if (result == WFSKRN_RESULT_OK)
    {
        iter->isInserted = (WFSBool) FALSE;
    }
    else
    {
        iter->isInserted = (WFSBool) TRUE;
    }

    // the P-tree block is pinned.
    // an F-tree block is pinned.
    return result;
}

WFSKrnResult FreeBlkNext(AreaInfo* area, FBATreeItr* iter)
{
    WFSKrnResult result = WFSKRN_RESULT_FTREE_EMPTY;
    FTreeItr* fti = &iter->fti;
    int subTreeIndex;

    // the P-tree block is pinned.
    // an F-tree block is pinned.

    // FBA_STRICT_ASSERT(FTreeVerifyTree(iter->ftreeHdr));
    if (0 < iter->fti.subTree->nNumRecs)
    {
        result =  FTreeNext(iter->ftreeHdr, fti);
    }

    subTreeIndex = (fti->subTree - &iter->ftreeHdr->tree[0]);
    while (result == WFSKRN_RESULT_PTREE_ENTRY_NOT_FOUND ||
           result == WFSKRN_RESULT_NOT_FOUND ||
           result == WFSKRN_RESULT_FTREE_EMPTY)
    {

        // search the next F-tree.
        result = EPTreeNext(area, (MEPTreeItr*) iter);
        if (result == WFSKRN_RESULT_OK)
        {
            result = FBAGetAndPinFtreeBlock(area, iter, BCACHE_FLAG_READ | BCACHE_FLAG_PINNED);
            if (result == WFSKRN_RESULT_OK)
            {
                fti->subTree = &iter->ftreeHdr->tree[subTreeIndex];
                if (fti->subTree->nNumRecs == 0)
                {
                    // skip the empty block.
                    result = WFSKRN_RESULT_FTREE_EMPTY;
                    continue;
                }
                result = FTreeFirst(iter->ftreeHdr, fti);
            }
        }
        else
        {
            break;
        }
    }

    // the P-tree block is pinned.
    // an F-tree block is pinned.
    return result;
}

WFSKrnResult FreeBlkLast(AreaInfo* area, FBATreeItr* iter, u32 nSubTreeBlkLog2Size)
{
    WFSKrnResult result;
    FTreeItr* fti = &iter->fti;

    // the P-tree block is pinned.

    if (!iter->eptreeHdr)
    {
        return WFSKRN_RESULT_INVALID;
    }

    result = EPTreeLast(area, (MEPTreeItr*) iter);
    while (result == WFSKRN_RESULT_OK)
    {
        result = FBAGetAndPinFtreeBlock(area, iter, BCACHE_FLAG_READ | BCACHE_FLAG_PINNED);
        if (result == WFSKRN_RESULT_OK)
        {
            // search F-tree.
            memset(fti->ptree.aEntryIdx, 0, sizeof(fti->ptree.aEntryIdx));
            fti->subTree = FTreeSelectSubTree(iter->ftreeHdr, nSubTreeBlkLog2Size);
            if (fti->subTree->nNumRecs == 0)
            {
                iter->isInserted = (WFSBool) FALSE;
                result = EPTreePrev(area, (MEPTreeItr*) iter);
                continue;
            }

            result = FTreeLast(iter->ftreeHdr, fti);
            break;
        }
    }

    if (result == WFSKRN_RESULT_OK)
    {
        iter->isInserted = (WFSBool) FALSE;
    }
    else
    {
        iter->isInserted = (WFSBool) TRUE;
    }

    // the P-tree block is pinned.
    // an F-tree block is pinned.
    return result;
}

WFSKrnResult FreeBlkPrev(AreaInfo* area, FBATreeItr* iter)
{
    WFSKrnResult result;
    FTreeItr* fti = &iter->fti;

    // P-tree and F-tree blocks are pinned.

    // FBA_STRICT_ASSERT(FTreeVerifyTree(iter->ftreeHdr));
    if (iter->isInserted == (WFSBool) TRUE)
    {
        return WFSKRN_RESULT_FTREE_RETRY; // must call Find() or Last() before FreeBlkPrev().
    }

    result =  FTreePrev(iter->ftreeHdr, fti);
    while (result == WFSKRN_RESULT_PTREE_ENTRY_NOT_FOUND ||
           result == WFSKRN_RESULT_NOT_FOUND ||
           result == WFSKRN_RESULT_FTREE_EMPTY)
    {
        int subTreeIndex = (fti->subTree - &iter->ftreeHdr->tree[0]);

        // lookup the previous F-tree.
        result = EPTreePrev(area, (MEPTreeItr*) iter);
        if (result == WFSKRN_RESULT_OK)
        {
            result = FBAGetAndPinFtreeBlock(area, iter, BCACHE_FLAG_READ | BCACHE_FLAG_PINNED);
            if (result == WFSKRN_RESULT_OK)
            {
                memset(fti->ptree.aEntryIdx, 0, sizeof(fti->ptree.aEntryIdx));
                fti->subTree = &iter->ftreeHdr->tree[subTreeIndex];

                if (fti->subTree->nNumRecs == 0)
                {
                    result = WFSKRN_RESULT_FTREE_EMPTY;
                    continue; // skip. Because the subtree in this f-tree block is empty.
                }
                result = FTreeLast(iter->ftreeHdr, fti);
            }
        }
        else
        {
            // set 1st. entry.
            memset(fti->ptree.aEntryIdx, 0, sizeof(fti->ptree.aEntryIdx));
            break;
        }
    }

    // the P-tree block is pinned.
    // an F-tree block is pinned.
    return result;
}

static inline void AllocatorInit(PTreeAllocator* allocator, EPTreeHdr* eptreeHdr)
{
    allocator->pBase = GET_BLOCK_HEAD(eptreeHdr);
    allocator->fpAllocNode = FTreeAllocEntry;
    allocator->fpAllocNodes = FTreeAllocEntries;
    allocator->fpFreeNode = FTreeFreeEntry;
}

// public APIs.

WFSKrnResult FreeBlkFind(AreaInfo* area, FBATreeItr* iter, u32 nRefBlkAdr, u32 nSubTreeBlkLog2Size)
{
    WFSKrnResult result;

    // the P-tree block is pinned.

    if (!iter->eptreeHdr[0])
    {
        return WFSKRN_RESULT_INVALID;
    }

    result = FBALookup(area, iter, nRefBlkAdr, nSubTreeBlkLog2Size);

    if (result == WFSKRN_RESULT_OK)
    {
        return WFSKRN_RESULT_OK;
    }

    if (iter->fti.subTree->nNumRecs == 0)
    {
        while (result != WFSKRN_RESULT_OK)
        {
            result = EPTreeNext(area, (MEPTreeItr*) iter);
            if (result != WFSKRN_RESULT_OK)
            {
                break;
            }

            result = FBAGetAndPinFtreeBlock(area, iter, BCACHE_FLAG_READ | BCACHE_FLAG_PINNED);
            if (result == WFSKRN_RESULT_OK)
            {
                iter->fti.subTree = FTreeSelectSubTree(iter->ftreeHdr, nSubTreeBlkLog2Size);
                result = FTreeFirst(iter->ftreeHdr, &iter->fti);
            }
        }
    }
    else
    {
        result = WFSKRN_RESULT_OK;
    }

    // the P-tree block is pinned.
    // an F-tree block is pinned.
    return result;
}

WFSKrnResult FreeBlkFindStrict(AreaInfo* area, FBATreeItr* iter, u32 nBlkAdr, u32 nSubTreeBlkLog2Size)
{
    WFSKrnResult result;

    // the P-tree block is pinned.

    if (!iter->eptreeHdr)
    {
        return WFSKRN_RESULT_INVALID;
    }
    result = FBALookupStrict(area, iter, nBlkAdr, nSubTreeBlkLog2Size);

    // the P-tree block is pinned.
    // an F-tree block is pinned.
    return result;
}

static WFSKrnResult InsertFreeBlock(AreaInfo* area)
{
    WFSKrnResult result;
    FBATreeItr iter;
    FTreeExtent extent;

    // the P-tree block is pinned.

    extent.start = 2; // FBA_ROOT_BLK and FBA_INITIAL_FTREE_BLK are reserved.
    extent.len = area->ah.nNumBlks - extent.start;

    FBAItrOpen(area, &iter);

#if _DEBUG
    u32 pinCount = BCacheSkipCheckPin();
#endif

    result = FreeBlkInsert(area, &iter, extent.start, extent.len);
    WFSKRN_FTREE_ASSERT(result == WFSKRN_RESULT_OK);

    GET_FBATREE_HDR(area)->totalBlks = area->ah.nNumBlks;

    FBAItrClose(area, &iter);

#if _DEBUG
    WFSKRN_FTREE_ASSERT(BCacheStopSkipCheckPin(pinCount) == WFSKRN_RESULT_OK);
#endif

    // the P-tree block is pinned.
    return result;
}

WFSKrnResult FreeBlkInit(AreaInfo *pAreaInfo, u32 offset)
{
    FBATreeHdr* header;
    WFSKrnResult result;
    if (offset < sizeof(WFSMetaDataFlags) + sizeof(WFSHashCode))
    {
        // must not overwrite hash.
        return WFSKRN_RESULT_INVALID;
    }
#ifdef FBA_TEST
    offset += pAreaInfo->nBlkSize - offset - sizeof(EPTreeHdr) - 32 * 3;
#endif // _FBA_TEST
    // the P-tree block is pinned.

    header = GET_FBATREE_HDR(pAreaInfo);
    header->totalBlks = 0;
    header->ofsEPtreeHdr = pAreaInfo->nBlkSize - offset - sizeof(EPTreeHdr); // offset from pAreaInfo->pAflh.
    header->ptreeHeight = FBA_PTREE_MIN_HEIGHT; // This means only one ptree block in the FBA tree.

    result = EPTreeInit(pAreaInfo, offset);
    if (result != WFSKRN_RESULT_OK)
    {
        return result;
    }

    result = InsertFreeBlock(pAreaInfo);

    // the P-tree block is pinned.
    return result;
}

static WFSKrnResult FBAAllocInit(AreaInfo* area, WFSBlkAdr nRefBlkAdr,
    FBATreeItr* iter, FBATreeItr* iterPrev, CacheMap* workBuf, u32 nBlkLog2Size)
{
    WFSKrnResult result;
    WFSBlkAdr start;

    result = FreeBlkFind(area, iter, nRefBlkAdr, nBlkLog2Size);
    if (result != WFSKRN_RESULT_OK)
    {
        result = FreeBlkFirst(area, iter, nBlkLog2Size);
    }
    if (result == WFSKRN_RESULT_OK)
    {
        // prepare backward search.
        result = CopyFBAIter(area, iter, iterPrev, BCACHE_FLAG_READ | BCACHE_FLAG_PINNED);
        if (result != WFSKRN_RESULT_OK)
        {
            return result; // [check] error handling.
        }

        if (result == WFSKRN_RESULT_OK)
        {
            start = FBA_RESULT_KEY(iter);
            workBuf->diskAddr = start;
        }
    }

    return result;
}

static ExtentStorageHeader* InitWorkBuffer(AreaInfo* area, WFSBlkAdr nRefBlkAdr, FBATreeItr* iter, CacheMap* workBuf)
{
    WFSKrnResult result;
    ExtentStorageEntry* extents;
    ExtentStorageHeader* extentHeader;

    result = FBAGetCache(area, workBuf->diskAddr, BCACHE_FLAG_DIRTY | BCACHE_FLAG_PINNED, &workBuf->cache);
    if (result != WFSKRN_RESULT_OK)
    {
        return NULL; // [check] should return WFSKrnResult?
    }

    extentHeader = (ExtentStorageHeader*) (workBuf->cache);

    extents = (ExtentStorageEntry*) ((u8*) extentHeader + sizeof(ExtentStorageHeader));
    if (nRefBlkAdr <= FBA_RESULT_KEY(iter))
    {
        // next
        extentHeader->numPrev = 0;
        extentHeader->numNext = 1;
    }
    else
    {
        // prev
        int indexPrev = ((area->nBlkSize - sizeof(ExtentStorageHeader)) / sizeof(ExtentStorageEntry)) - 1;
        extentHeader->numPrev = 1;
        extentHeader->numNext = 0;
        extents = &extents[indexPrev];
    }


    extents->start = FBA_RESULT_KEY(iter);
    extents->treeType = (u16) GET_SUBTREE_TYPE(iter->ftreeHdr, &iter->fti);
    WFSKRN_FTREE_ASSERT(extents->treeType <= FTREE_TYPE_G);
    //WFSKRN_FTREE_ASSERT(FTREE_TYPE_A <= extents->treeType);
    extents->len = (FBA_RESULT_DATA(iter) + 1) * (1<<FTreeBlkLog2Ofs[extents->treeType]);

    extentHeader->prio = 1;
    return extentHeader;
}

static inline WFSBool IsFarawayFromRefAddr(AreaInfo* area, ExtentStorageHeader* extentHeader,
    ExtentStorageEntry* extent, int indexTail, int indexHead, WFSBlkAdr nRefBlkAdr, WFSBlkAdr nBlkLen)
{
    if (0 < extentHeader->numNext)
    {
        ASSERT(0 < extent[indexTail].start);
        if ((area->ah.nNumBlks>>1) < extent[indexTail].start - nRefBlkAdr) // [check] hard-coding.
        {
            return (WFSBool) TRUE;
        }
    }
    if (0 < extentHeader->numPrev)
    {
        ASSERT(0 < extent[indexHead].start);
        if ((area->ah.nNumBlks>>1) < nRefBlkAdr - extent[indexHead].start) // [check] hard-coding.
        {
            return (WFSBool) TRUE;
        }
    }
    return (WFSBool) FALSE;
}

/*
    +-------------+ <------ aBlkAdr
    | block addr0 | A
    +-------------+ |
    |             | |
    :             : | nStride
    |             | |
    +-------------+ V
    | block addr1 | A
    +-------------+ |
    |             | |
    :             : | nStride
    |             | |
    +-------------+ V
    | block addr2 |
    +-------------+
    :             :
    :             :
    :             :
    +-------------+
    | block addrN | N == nNumBlks - 1
    +-------------+
    :             :
    :             :
    +-------------+
                           disk block
    block addr m ----> +----------------+
                       |                |
                       | 2^nBlkLog2Size |
                       |                |
                       +----------------+
*/

static WFSKrnResult SelectFreeBlocks(AreaInfo* area, FBATreeItr* iter, FBATreeItr* iterPrev,
    ExtentStorageHeader* extentHeader, WFSBlkAdr nRefBlkAdr, WFSBlkAdr nBlkLen)
{
    WFSKrnResult result = WFSKRN_RESULT_OK;
    int indexPrev = ((area->nBlkSize - sizeof(ExtentStorageHeader)) / sizeof(ExtentStorageEntry)) - 1;
    int indexNext = 0;
    u32 reservedLenPrev = 0;
    u32 reservedLenNext = 0;
    u32 totalBlk = 0;
    u32 len;
    u32 dn = 0;
    u32 dp = 0;
    WFSBool stopSearchNext = (WFSBool) FALSE;
    WFSBool stopSearchPrev = (WFSBool) FALSE;
    ExtentStorageEntry* extent;

    extent = (ExtentStorageEntry*) ((u8*) extentHeader + sizeof(ExtentStorageHeader));
    if (extentHeader->numNext == 1)
    {
        dn = extent[indexNext].len;
        ++indexNext;
        totalBlk += extent->len;
    }
    else
    {
        WFSKRN_FTREE_ASSERT(extentHeader->numPrev == 1);
        dp = extent[indexPrev].len;
        totalBlk += extent[indexPrev].len;
        --indexPrev;
    }

    // search
    while (totalBlk < nBlkLen)
    {
        if (!stopSearchNext && (stopSearchPrev || dn <= dp))
        {
            if (0 < reservedLenNext)
            {
                ++indexNext;
                ++extentHeader->numNext;
                len = reservedLenNext;
                reservedLenNext = 0;
                result = WFSKRN_RESULT_OK;
            }
            else
            {
                result = FreeBlkNext(area, iter);
                if (result == WFSKRN_RESULT_OK)
                {
                    WFSKRN_FTREE_ASSERT(nRefBlkAdr < FBA_RESULT_KEY(iter));
                    dn += FBA_RESULT_KEY(iter) - nRefBlkAdr;
                    extent[indexNext].start = FBA_RESULT_KEY(iter);
                    extent[indexNext].treeType = (u16) GET_SUBTREE_TYPE(iter->ftreeHdr, &iter->fti);

                    len = (FBA_RESULT_DATA(iter) + 1) * (1<<FTreeBlkLog2Ofs[extent[indexNext].treeType]);
                    extent[indexNext].len = len;

                    if (dp < dn && !stopSearchPrev)
                    {
                        reservedLenNext = len;
                        continue;
                    }
                    else if (nBlkLen <= len)
                    {
                        extent[0] = extent[indexNext];
                        extentHeader->numNext = 1;
                        extentHeader->numPrev = 0;
                        break;
                    }
                    else
                    {
                        ++indexNext;
                        ++extentHeader->numNext;
                    }
                }
                else
                {
                    stopSearchNext = (WFSBool) TRUE;
                    continue;
                }
            }
        }
        else if (!stopSearchPrev)
        {
            if (0 < reservedLenPrev)
            {
                --indexPrev;
                ++extentHeader->numPrev;
                len = reservedLenPrev;
                reservedLenPrev = 0;
                result = WFSKRN_RESULT_OK;
            }
            else
            {
                result = FreeBlkPrev(area, iterPrev);
                if (result == WFSKRN_RESULT_OK)
                {
                    WFSKRN_FTREE_ASSERT(FBA_RESULT_KEY(iterPrev) - nRefBlkAdr);
                    dp += nRefBlkAdr - FBA_RESULT_KEY(iterPrev);
                    extent[indexPrev].start = FBA_RESULT_KEY(iterPrev);
                    extent[indexPrev].nNumBlk = FBA_RESULT_DATA(iterPrev);
                    extent[indexPrev].treeType = (u16) GET_SUBTREE_TYPE(iterPrev->ftreeHdr, &iterPrev->fti);
                    len = (FBA_RESULT_DATA(iterPrev) + 1) * (1<<FTreeBlkLog2Ofs[extent[indexPrev].treeType]);
                    extent[indexPrev].len = len;

                    if (dn <= dp && !stopSearchNext)
                    {
                        reservedLenPrev = len;
                        continue;
                    }
                    else if (nBlkLen <= len)
                    {
                        int indexPrev0 = ((area->nBlkSize - sizeof(ExtentStorageHeader)) / sizeof(ExtentStorageEntry)) - 1;
                        extent[indexPrev0] = extent[indexPrev];
                        extentHeader->numNext = 0;
                        extentHeader->numPrev = 1;
                        break;
                    }
                    else
                    {
                        --indexPrev;
                        ++extentHeader->numPrev;
                    }
                }
                else
                {
                    stopSearchPrev = (WFSBool) TRUE;
                    continue;
                }
            }
        }
        else
        {
            result = WFSKRN_RESULT_DIRECTORY_QUOTA;
            // FreeBlkDumpFreeblock(area, header);
            goto SEARCHBLK_RESULT;
        }

        if (indexPrev <= indexNext)
        {
            result = WFSKRN_RESULT_RESOURCE_LIMIT_EXCEEDED; // Work block (8KB) is too small to store all extents. In this case,
                                                            // WFSKRN_RESULT_RESOURCE_LIMIT_EXCEEDED seems to be suitable.
            goto SEARCHBLK_RESULT;
        }
        totalBlk += len;
    }

    if (IsFarawayFromRefAddr(area, extentHeader, extent, indexNext-1, indexPrev+1, nRefBlkAdr, nBlkLen))
    {
        ++extentHeader->prio;
    }

SEARCHBLK_RESULT:
    return result;
}

static WFSKrnResult SearchAndDelete(AreaInfo* area, FBATreeItr* iter,
    WFSBlkAdr nBlkAdr, WFSBlkAdr nNumBlks, u16 startSubTreeType)
{
    WFSKrnResult result = WFSKRN_RESULT_NOT_FOUND;
    u16 subTreeType;
    for (subTreeType = startSubTreeType; subTreeType < MAX_SUB_TREE; ++subTreeType)
    {
        osTPrintf("SearchAndDelete: subTree %u start %u len %u\n", subTreeType, nBlkAdr, nNumBlks);
        result = FreeBlkDelete(area, iter, nBlkAdr, nNumBlks, FTreeBlkLog2Ofs[subTreeType]);
        if (result == WFSKRN_RESULT_OK)
        {
            osTPrintf("result %d\n", result);
            break;
        }
    }
    return result;
}

static WFSKrnResult MakeFreeBlkListFromForwardSearch(AreaInfo* area, FBATreeItr* iter,
    ExtentStorageHeader* extentHeader,  WFSBlkAdr nRefBlkAdr, WFSBlkAdr* num,
        WFSBlkAdr* aBlkAdr, s32 nStride, u32 unitBlkLen)
{
    WFSKrnResult result = WFSKRN_RESULT_INVALID;
    ExtentStorageEntry* storage = (ExtentStorageEntry*) ((u8*) extentHeader + sizeof(ExtentStorageHeader));
    int index;
    WFSBlkAdr block;

    if (extentHeader->numNext == 0)
    {
        return WFSKRN_RESULT_OK;
    }

    // result of forward search.
    for (index = 0; index < extentHeader->numNext; ++index)
    {
        ExtentStorageEntry* extent = &storage[index];

        if (*num * unitBlkLen < extent->len)
        {
            if (extent->start <= nRefBlkAdr &&
                nRefBlkAdr + *num * unitBlkLen <= extent->start + extent->len)
            {
                extent->start = nRefBlkAdr;
            }
            extent->len = *num * unitBlkLen;
        }

        for (block = 0; block * unitBlkLen < extent->len && 0 < *num; ++block)
        {
            *aBlkAdr = extent->start + block * unitBlkLen;
            aBlkAdr = (WFSBlkAdr*) ((u8*) aBlkAdr + nStride);
            --*num;
        }

#ifdef FBA_DEBUG_VERBOSE
        if (FBADebugVerbose)
        {
            osTPrintf("[DEL] start %5u, len %4u, treeType %1d, extent[%4d] \n",
                extent->start, extent->len, extent->treeType, index);
        }
#endif //  FBA_DEBUG_VERBOSE

        result = FreeBlkDelete(area, iter, extent->start, block * unitBlkLen,
                FTreeBlkLog2Ofs[extent->treeType]);
        if (result != WFSKRN_RESULT_OK)
        {

            result = SearchAndDelete(area, iter, extent->start, block * unitBlkLen, extent->treeType + 1);
            if (result != WFSKRN_RESULT_OK)
            {
                // Extents are changed by releasing control blocks, delete blocks one by one.
                extent->start += unitBlkLen * (block - 1);
                while (block--)
                {
                    result = SearchAndDelete(area, iter, extent->start, unitBlkLen, extent->treeType);
                    if (result != WFSKRN_RESULT_OK)
                    {
                        break;
                    }
                    extent->start -= unitBlkLen;
                }
            }

            if (result != WFSKRN_RESULT_OK)
            {
                osTPrintf("MakeFreeBlkList() [2]: ***FreeBlkDelete error %d ***\n", result); // [check] restore. NYI.
                result = WFSKRN_RESULT_DIRECTORY_QUOTA; // [check] should return WFSKRN_RESULT_FATAL_ERROR?
                WFSKRN_FTREE_ASSERT(0); // [check] should not reach here.
                break;
            }
        }
    }
    return result;
}

static WFSKrnResult MakeFreeBlkListFromBackwardSearch(AreaInfo* area, FBATreeItr* iter,
    ExtentStorageHeader* extentHeader,  WFSBlkAdr nRefBlkAdr, WFSBlkAdr* num,
            WFSBlkAdr* aBlkAdr, s32 nStride, u32 unitBlkLen)
{
    WFSKrnResult result = WFSKRN_RESULT_INVALID;
    ExtentStorageEntry* storage = (ExtentStorageEntry*) ((u8*) extentHeader + sizeof(ExtentStorageHeader));
    int index;
    int indexPrevTop;
    int indexLast;
    WFSBlkAdr block;

    if (extentHeader->numPrev == 0)
    {
        return WFSKRN_RESULT_OK;
    }

    // result of backward search.
    indexLast = ((area->nBlkSize - sizeof(ExtentStorageHeader)) / sizeof(ExtentStorageEntry));
    indexPrevTop = indexLast - extentHeader->numPrev;
    for (index = indexPrevTop; index < indexPrevTop + extentHeader->numPrev; ++index)
    {
        ExtentStorageEntry* extent = &storage[index];

        if (*num * unitBlkLen < extent->len)
        {
            if (extent->start <= nRefBlkAdr &&
                nRefBlkAdr + *num * unitBlkLen <= extent->start + extent->len)
            {
                extent->start = nRefBlkAdr;
            }
            else
            {
                extent->start += (extent->len - *num * unitBlkLen);
            }
            extent->len = *num * unitBlkLen;
        }

        for (block = 0; block * unitBlkLen < extent->len && 0 < *num; ++block)
        {
            *aBlkAdr = extent->start + block * unitBlkLen;
            aBlkAdr = (WFSBlkAdr*) ((u8*) aBlkAdr + nStride);
            --*num;
        }

#ifdef FBA_DEBUG_VERBOSE
        if (FBADebugVerbose)
        {
            osTPrintf("[DEL] extent[%4d] len %4u, start %5u, treeType %1d\n",
                index, extent->len, extent->start, extent->treeType);
        }
#endif //  FBA_DEBUG_VERBOSE

        result = FreeBlkDelete(area, iter, extent->start,
                block * unitBlkLen, FTreeBlkLog2Ofs[extent->treeType]);
        if (result != WFSKRN_RESULT_OK)
        {
            u16 subTreeType;
            for (subTreeType = extent->treeType + 1; subTreeType < MAX_SUB_TREE; ++subTreeType)
            {
                osTPrintf("[2] subTree %u start %u len %u\n",
                    extent->treeType, extent->start, block * unitBlkLen);
                result = FreeBlkDelete(area, iter, extent->start,
                        block * unitBlkLen, FTreeBlkLog2Ofs[subTreeType]);
                if (result == WFSKRN_RESULT_OK)
                {
                    break;
                }
            }

            if (result != WFSKRN_RESULT_OK)
            {
                osTPrintf("MakeFreeBlkList() [1]: ***FreeBlkDelete error %d ***\n", result); // [check] restore. NYI.
                result = WFSKRN_RESULT_DIRECTORY_QUOTA; // [check] should return WFSKRN_RESULT_FATAL_ERROR?
                WFSKRN_FTREE_ASSERT(0); // [check] should not reach here.
                break;
            }
        }
    }
    return result;
}

static WFSKrnResult MakeFreeBlkList(AreaInfo* area, ExtentStorageHeader* extentHeader,
    WFSBlkAdr nRefBlkAdr, WFSBlkAdr nNumBlks, WFSBlkAdr* aBlkAdr, s32 nStride, u32 nBlkLog2Size)
{
    WFSKrnResult result = WFSKRN_RESULT_INVALID;
    FBATreeItr iter;
    WFSBlkAdr numToBeAllocated = nNumBlks;
    WFSBlkAdr unitBlkLen = (1<<nBlkLog2Size);

    FBAItrOpen(area, &iter);

    result = MakeFreeBlkListFromBackwardSearch(area, &iter, extentHeader, nRefBlkAdr, &numToBeAllocated, aBlkAdr, nStride, unitBlkLen);
    if (result == WFSKRN_RESULT_OK)
    {
        aBlkAdr = (WFSBlkAdr*) ((u8*) aBlkAdr + nStride * (nNumBlks - numToBeAllocated));
        result = MakeFreeBlkListFromForwardSearch(area, &iter, extentHeader, nRefBlkAdr, &numToBeAllocated, aBlkAdr, nStride, unitBlkLen);

        WFSKRN_FTREE_ASSERT(result != WFSKRN_RESULT_OK || numToBeAllocated == 0);
    }

    FBAItrClose(area, &iter);
    return result;
}

static WFSKrnResult ScareUpBlocks(AreaInfo* area,
    u32 nBlkLen, CacheMap* workBufs, int subTreeIdxMin, int subTreeIdxMax, u32* subTreeIdx)
{
    ExtentStorageHeader* extentHeader = NULL;
    u32 totalBlks = 0;
    int indexEnd = ((area->nBlkSize - sizeof(ExtentStorageHeader)) / sizeof(ExtentStorageEntry)) - 1;
    int indexPrev = indexEnd;
    int indexNext = 0;
    int idx;
    ExtentStorageHeader* eh;
    ExtentStorageEntry* extent;

    for (idx = subTreeIdxMax; subTreeIdxMin <= idx; --idx)
    {
        if (workBufs[idx].cache)
        {
            extentHeader = (ExtentStorageHeader*) (workBufs[idx].cache);
            *subTreeIdx = (u32) idx;
            indexPrev -= extentHeader->numPrev;
            indexNext += extentHeader->numNext;

            extent = (ExtentStorageEntry*) ((u8*) extentHeader + sizeof(ExtentStorageHeader));
            int i;
            for (i = 0; i < extentHeader->numNext; ++i)
            {
                totalBlks += extent[i].len;
            }
            for (i = 0; i < extentHeader->numPrev; ++i)
            {
                totalBlks += extent[indexEnd - i].len;
            }
            break;
        }
    }

    if (extentHeader == NULL)
    {
        return WFSKRN_RESULT_DIRECTORY_QUOTA;
    }
    if (nBlkLen <= totalBlks)
    {
        extentHeader->numNext = indexNext;
        extentHeader->numPrev = indexEnd - indexPrev;
        return WFSKRN_RESULT_OK;
    }

    extent = (ExtentStorageEntry*) ((u8*) extentHeader + sizeof(ExtentStorageHeader));

    for (idx = subTreeIdxMax; (int) subTreeIdxMin <= idx; --idx)
    {
        eh = (ExtentStorageHeader*) (workBufs[idx].cache);

        if (!eh || idx == *subTreeIdx)
        {
            continue;
        }

        ExtentStorageEntry* exNext = (ExtentStorageEntry*) ((u8*) eh + sizeof(ExtentStorageHeader));
        ExtentStorageEntry* exPrev = &exNext[indexEnd];

        int numNext = eh->numNext;
        int numPrev = eh->numPrev;
        for (;;)
        {
            if (0 < numNext)
            {
                extent[indexNext++] = *exNext;
                totalBlks += exNext->len;
                --numNext;
                ++exNext;
            }

            if (0 < numPrev)
            {
                extent[indexPrev--] = *exPrev;
                totalBlks += exPrev->len;
                --numPrev;
                --exPrev;
            }
            if (nBlkLen <= totalBlks)
            {
                extentHeader->numNext = indexNext;
                extentHeader->numPrev = indexEnd - indexPrev;
                return WFSKRN_RESULT_OK;
            }

            if (numPrev == 0 && numNext == 0)
            {
                break;
            }
        }
    }

    return WFSKRN_RESULT_DIRECTORY_QUOTA;
}

static WFSKrnResult FBASearchLowerSubtrees(AreaInfo* area,
    WFSBlkAdr nRefBlkAdr,WFSBlkAdr nNumBlks, u32 nBlkLog2Size, CacheMap* workBuf)
{
    WFSKrnResult result = WFSKRN_RESULT_NOT_FOUND;
    WFSBlkAdr nBlkLen;
    u32 subTreeIdx = FTREE_TYPE_END;
    int subTreeIdxMin;
    int subTreeIdxMax;
    u32 nSubTreeBlkLog2Size;
    int idx;
    CacheMap workBufs[MAX_SUB_TREE];

    FBATreeItr iter;
    FBATreeItr iterPrev;

    FBAItrOpen(area, &iter);
    FBAItrOpen(area, &iterPrev);

    iterPrev.ftreeAddr = 0xffffffff; // [check] this means the address is not specified.

    memset(workBufs, 0, sizeof(workBufs));

    ExtentStorageHeader* extentHeader = NULL;

    nBlkLen = (nNumBlks * (1<<nBlkLog2Size));
    subTreeIdxMax = GetSubTreeType(nBlkLen);
    subTreeIdxMin = GetSubTreeType(1<<nBlkLog2Size);

    for (idx = subTreeIdxMax; (int) subTreeIdxMin <= idx; --idx)
    {
        nSubTreeBlkLog2Size = FTreeBlkLog2Ofs[idx];

        result = FBAAllocInit(area, nRefBlkAdr, &iter, &iterPrev, &workBufs[idx], nSubTreeBlkLog2Size);
        if (result == WFSKRN_RESULT_OK)
        {
            extentHeader = InitWorkBuffer(area, nRefBlkAdr, &iter, &workBufs[idx]);
            if (extentHeader)
            {
                // search
                result = SelectFreeBlocks(area, &iter, &iterPrev, extentHeader, nRefBlkAdr,
                    nNumBlks * (1<<nBlkLog2Size));
                if (result == WFSKRN_RESULT_OK)
                {
                    subTreeIdx = (u32) idx;
                    break;
                }
            }
        }
        else if (idx == subTreeIdxMax)
        {
            --subTreeIdxMax;
        }
    }

    FBAItrClose(area, &iter);
    FBAItrClose(area, &iterPrev);

    if (result != WFSKRN_RESULT_OK && subTreeIdxMin <= subTreeIdxMax)
    {
        result = ScareUpBlocks(area, nBlkLen, workBufs, subTreeIdxMin, subTreeIdxMax, &subTreeIdx);
    }

    if (result == WFSKRN_RESULT_OK)
    {
        WFSKRN_FTREE_ASSERT(subTreeIdx < MAX_SUB_TREE);
        workBuf->diskAddr = workBufs[subTreeIdx].diskAddr;
        workBuf->cache = workBufs[subTreeIdx].cache;

        u32 threshold;
        if (nBlkLen < (1<<FTREE_LOG2_OFS_C))
        {
            threshold = 4;
        }
        else if (nBlkLen < (1<<FTREE_LOG2_OFS_D))
        {
            threshold = 6;
        }
        else
        {
            threshold = 8;
        }

        extentHeader = (ExtentStorageHeader*) (workBuf->cache);
        if (threshold <= extentHeader->numNext + extentHeader->numPrev)
        {
            ++extentHeader->prio;
        }
    }

    // choose
    for (idx = subTreeIdxMin; idx <= subTreeIdxMax; ++idx)
    {
        if (workBufs[idx].diskAddr && (idx != subTreeIdx || result != WFSKRN_RESULT_OK))
        {
            FBAFreeCache(area, workBufs[idx].diskAddr);
        }
    }

    return result;
}

static WFSKrnResult FBASearchUpperSubtrees(AreaInfo* area, WFSBlkAdr nRefBlkAdr, WFSBlkAdr nNumBlks, u32 nBlkLog2Size,
    FBASingleEntry* singleEntry)
{
    WFSBlkAdr nBlkLen;
    u32 subTreeIdx;
    u32 searchBlkLog2Size;
    WFSKrnResult result = WFSKRN_RESULT_UNKNOWN;
    ExtentStorageEntry alt; // alternative
    WFSBlkAdr start;
    WFSBlkAdr len;
    WFSBlkAdr dist;
    WFSBlkAdr altDist = 0;
    FBATreeItr iter;
    ExtentStorageEntry* entry = &singleEntry->entry;

    FBAItrOpen(area, &iter);

    nBlkLen = (nNumBlks * (1<<nBlkLog2Size));
    subTreeIdx = GetSubTreeType(nBlkLen);

    entry->len = 0;
    alt.len = 0; // indicates the extent not found.

    while (++subTreeIdx < MAX_SUB_TREE)
    {
        searchBlkLog2Size = FTreeBlkLog2Ofs[subTreeIdx];

        result = FreeBlkFind(area, &iter, nRefBlkAdr, searchBlkLog2Size);
        if (result == WFSKRN_RESULT_OK)
        {
            start = FBA_RESULT_KEY(&iter);
            len = (FBA_RESULT_DATA(&iter) + 1) * (1<<searchBlkLog2Size);

            // [check] should choose the best answer.
            if (start <= nRefBlkAdr &&
                nRefBlkAdr + nBlkLen < start + len)
            {
                entry->start = start;
                // entry->nNumBlk = FBA_RESULT_DATA(&iter) + 1;
                entry->treeType = subTreeIdx;
                entry->len = len;
                break;
            }

            if (start < nRefBlkAdr)
            {
                dist = nRefBlkAdr - (start + len);
            }
            else
            {
                dist = start - nRefBlkAdr;
            }

            if (alt.len == 0 || dist < altDist)
            {
                altDist = dist;
                alt.start = start;
                alt.nNumBlk = FBA_RESULT_DATA(&iter) + 1;
                alt.treeType = subTreeIdx;
                alt.len = len;
            }

            if (start + len < nRefBlkAdr)
            {
                // check the next extent.
                result = FreeBlkNext(area, &iter);
                if (result == WFSKRN_RESULT_OK)
                {
                    start = FBA_RESULT_KEY(&iter);
                    len = (FBA_RESULT_DATA(&iter) + 1) * (1<<searchBlkLog2Size);
                    dist = start - nRefBlkAdr;

                    if (start - nRefBlkAdr < nRefBlkAdr - (alt.start + alt.len))
                    {
                        if (start < nRefBlkAdr + nBlkLen)
                        {
                            entry->start = start;
                            entry->treeType = subTreeIdx;
                            entry->len = len;
                            singleEntry->header.prio = 0;
                            break;
                        }
                        altDist = start - (nRefBlkAdr + nBlkLen);
                        alt.start = start;
                        alt.nNumBlk = FBA_RESULT_DATA(&iter) + 1;
                        alt.treeType = subTreeIdx;
                        alt.len = len;
                        singleEntry->header.prio = 1;

                    }

                    if (alt.len == 0 || dist < altDist)
                    {
                        alt.start = start;
                        alt.nNumBlk = FBA_RESULT_DATA(&iter) + 1;
                        alt.treeType = subTreeIdx;
                        alt.len = len;
                    }

                    if (dist < nBlkLen)
                    {
                        break;
                    }
                }
            }
        }
    }

    singleEntry->header.prio = 0; // 0 is the highest.
    if (entry->len == 0 && alt.len != 0)
    {
        *entry = alt;
        result = WFSKRN_RESULT_OK;
        ++singleEntry->header.prio; // decrease the priority.
    }

    singleEntry->header.numPrev = 0;
    if (result == WFSKRN_RESULT_OK)
    {
        singleEntry->header.numNext = 1;
    }
    else
    {
        singleEntry->header.numNext = 0;
    }

    FBAItrClose(area, &iter);
    return result;
}


static void MergeExtents(AreaInfo* area, WFSBlkAdr nRefBlkAdr, u32 nNumBlks, u32 nBlkLog2Size, CacheMap* workBuf, FBASingleEntry* singleEntry)
{
    ExtentStorageHeader* extentHeader = (ExtentStorageHeader*) (workBuf->cache);
    ExtentStorageEntry* extent = (ExtentStorageEntry*) ((u8*) extentHeader + sizeof(ExtentStorageHeader));
    u32 requiredBlkLen = nNumBlks * (1<<nBlkLog2Size);
    if (nRefBlkAdr < singleEntry->entry.start)
    {
        int index;
        for (index = 0; index < extentHeader->numNext; ++index)
        {
            ExtentStorageEntry* entry = &extent[index];
            if (entry->start + entry->len == singleEntry->entry.start)
            {
                if (requiredBlkLen <= entry->len)
                {
                    return; // no need to merge.
                }
                /*
                  +-------+----------------------------+
                  | entry |       singleEntry          |
                  +-------+----------------------------+
                */
                extentHeader->numNext = 2;
                extentHeader->numPrev = 0;
                extent[0] = *entry;
                extent[1] = singleEntry->entry;
                return;
            }
        }
    }
    else
    {

        int indexEnd = ((area->nBlkSize - sizeof(ExtentStorageHeader)) / sizeof(ExtentStorageEntry)) - 1;
        int index;
        for (index = 0; index < extentHeader->numPrev; ++index)
        {
            ExtentStorageEntry* entry = &extent[indexEnd - index];
            if (entry->start == singleEntry->entry.start + singleEntry->entry.len)
            {
                if (requiredBlkLen <= entry->len)
                {
                    return; // no need to merge.
                }
                /*
                  +----------------------------+-------+
                  |       singleEntry          | entry |
                  +----------------------------+-------+
                */
                extentHeader->numNext = 0;
                extentHeader->numPrev = 2;
                singleEntry->entry.len -= entry->len;
                extent[indexEnd - 1] = *entry;
                extent[indexEnd] = singleEntry->entry;
                return;
            }
        }
    }
}

WFSKrnResult FreeBlkAlloc(AreaInfo *pAreaInfo, WFSBlkAdr nRefBlkAdr, WFSBlkAdr nNumBlks, WFSBlkAdr *aBlkAdr, s32 nStride, u32 nBlkLog2Size)
{
    WFSKrnResult resultUpper = WFSKRN_RESULT_NOT_FOUND;
    WFSKrnResult resultLower = WFSKRN_RESULT_NOT_FOUND;
    WFSKrnResult result;
    FBATreeHdr* header;
    CacheMap workBuf;
    ExtentStorageHeader* extentHeader;
    FBASingleEntry singleEntry;

    ASSERT(aBlkAdr);
    ASSERT(pAreaInfo);
    ASSERT(0 < nNumBlks);

    // the P-tree block is pinned.

#if _DEBUG
    u32 pinCount = BCacheSkipCheckPin();
#ifdef FBA_DEBUG_VERBOSE
    if (FBADebugVerbose)
    {
        osTPrintf("===================\n");
        osTPrintf("      ALLOC (%u, %u)\n", nRefBlkAdr, nNumBlks);
        osTPrintf("===================\n");
    }
#endif // FBA_DEBUG_VERBOSE
#endif
    header = GET_FBATREE_HDR(pAreaInfo);
    WFSKRN_FTREE_ASSERT(header); // header must be set before calling FreeBlkAlloc().

    nBlkLog2Size -= WFS_LOG2_SMALL_BLK_SIZE; // use offset from WFS_LOG2_SMALL_BLK_SIZE in FreeBlk APIs.

    // try to allocate blocks sequentially.
    if (nNumBlks == 1) // [check] hard cording
    {
        resultLower = FBASearchLowerSubtrees(pAreaInfo, nRefBlkAdr, nNumBlks, nBlkLog2Size, &workBuf);
        extentHeader = (ExtentStorageHeader*) (workBuf.cache);
        if (resultLower == WFSKRN_RESULT_OK &&
            extentHeader->prio <= 1)
        {
            result = MakeFreeBlkList(pAreaInfo, extentHeader, nRefBlkAdr, nNumBlks, aBlkAdr, nStride, nBlkLog2Size);
        }
        else
        {
            resultUpper = FBASearchUpperSubtrees(pAreaInfo, nRefBlkAdr, nNumBlks, nBlkLog2Size, &singleEntry);
            if (resultUpper == WFSKRN_RESULT_OK)
            {
                extentHeader = (ExtentStorageHeader*) &singleEntry.header;
                result = MakeFreeBlkList(pAreaInfo, extentHeader, nRefBlkAdr, nNumBlks, aBlkAdr, nStride, nBlkLog2Size);
            }
            else if (resultLower == WFSKRN_RESULT_OK)
            {
                result = MakeFreeBlkList(pAreaInfo, extentHeader, nRefBlkAdr, nNumBlks, aBlkAdr, nStride, nBlkLog2Size);
            }
            else
            {
                result = WFSKRN_RESULT_DIRECTORY_QUOTA;
            }
        }
    }
    else
    {
        resultUpper = FBASearchUpperSubtrees(pAreaInfo, nRefBlkAdr, nNumBlks, nBlkLog2Size, &singleEntry);
        if (resultUpper == WFSKRN_RESULT_OK &&
            singleEntry.header.prio == 0)
        {
            extentHeader = (ExtentStorageHeader*) &singleEntry.header;
            result = MakeFreeBlkList(pAreaInfo, extentHeader, nRefBlkAdr, nNumBlks, aBlkAdr, nStride, nBlkLog2Size);
        }
        else
        {
            resultLower = FBASearchLowerSubtrees(pAreaInfo, nRefBlkAdr, nNumBlks, nBlkLog2Size, &workBuf);
            extentHeader = (ExtentStorageHeader*) (workBuf.cache);
            if (resultLower == WFSKRN_RESULT_OK &&
                (extentHeader->prio <= 1 || resultUpper != WFSKRN_RESULT_OK))
            {
                if (resultUpper == WFSKRN_RESULT_OK)
                {
                    MergeExtents(pAreaInfo, nRefBlkAdr, nNumBlks, nBlkLog2Size, &workBuf, &singleEntry); // Merge singleEntry -> workBuf.
                }
                result = MakeFreeBlkList(pAreaInfo, extentHeader, nRefBlkAdr, nNumBlks, aBlkAdr, nStride, nBlkLog2Size);
            }
            else if (resultUpper == WFSKRN_RESULT_OK &&
                     singleEntry.header.prio == 1)
            {
                extentHeader = (ExtentStorageHeader*) &singleEntry.header;
                result = MakeFreeBlkList(pAreaInfo, extentHeader, nRefBlkAdr, nNumBlks, aBlkAdr, nStride, nBlkLog2Size);
            }
            else
            {
                result = WFSKRN_RESULT_DIRECTORY_QUOTA;
            }
        }
    }

    if (result == WFSKRN_RESULT_OK)
    {
        header->totalBlks -= nNumBlks * (1<<nBlkLog2Size);
    }

    if (resultLower == WFSKRN_RESULT_OK)
    {
        FBAFreeCache(pAreaInfo, workBuf.diskAddr);
    }

    if (result == WFSKRN_RESULT_NOT_FOUND ||
        result == WFSKRN_RESULT_PTREE_ENTRY_NOT_FOUND ||
        result == WFSKRN_RESULT_FTREE_EMPTY)
    {
#ifdef _DEBUG
        WFSKRN_FTREE_ASSERT(FreeBlkVerify(pAreaInfo));
#endif // _DEBUG

        result = WFSKRN_RESULT_DIRECTORY_QUOTA;
    }

#if _DEBUG
    WFSKRN_FTREE_ASSERT(BCacheStopSkipCheckPin(pinCount) == WFSKRN_RESULT_OK);
#endif

    return result;
}

WFSKrnResult FreeBlkFree(AreaInfo *pAreaInfo, WFSBlkAdr nNumBlks, WFSBlkAdr *aBlkAdr, s32 nStride, u32 nBlkLog2Size)
{
    FBATreeItr iter;
    u32 startExtent;
    u32 freeBlks; // the number of blocks to be freed.
    WFSKrnResult result = WFSKRN_RESULT_OK; // kondo_masahiro 08/08/15 // must return WFSKRN_RESULT_OK when nNumBlks == 0.
    FBATreeHdr* header;
    WFSBlkAdr* prevBlkAdr;
    WFSBlkAdr* endBlkAdr;
#ifdef _DEBUG
    u32 pinCount = BCacheSkipCheckPin();
#ifdef FBA_DEBUG_VERBOSE
    if (FBADebugVerbose)
    {
        osTPrintf("===================\n");
        osTPrintf("       FREE\n");
        osTPrintf("===================\n");
    }
#endif // FBA_DEBUG_VERBOSE
#endif // _DEBUG

    // the P-tree block is pinned.

    nBlkLog2Size -= WFS_LOG2_SMALL_BLK_SIZE; // use offset from WFS_LOG2_SMALL_BLK_SIZE in FreeBlk APIs.

    FBAItrOpen(pAreaInfo, &iter);

    header = GET_FBATREE_HDR(pAreaInfo);
    WFSKRN_FTREE_ASSERT(header); // header must be set before calling FreeBlkAlloc().

    startExtent = *aBlkAdr;
    prevBlkAdr = aBlkAdr;
    aBlkAdr = (WFSBlkAdr*) ((u8*) aBlkAdr + nStride);

    endBlkAdr = (WFSBlkAdr*) ((u8*) aBlkAdr + nStride * nNumBlks);

    while (1)
    {
        if (aBlkAdr == endBlkAdr)
        {
            freeBlks = *prevBlkAdr - startExtent;
            if (0 < freeBlks)
            {
                BCacheInvalidateBlks(pAreaInfo->pVolInfo, startExtent, freeBlks); // must call before FreeBlkInsert(),
                                                                                  // becase a block inserted can be used as a new control block in FreeBlkInsert().
                result = FreeBlkInsert(pAreaInfo, &iter, startExtent, freeBlks);
            }
            break;
        }

        if (*aBlkAdr - *prevBlkAdr != (1<<nBlkLog2Size)) // end of an extent.
        {
            freeBlks = *prevBlkAdr - startExtent + (1<<nBlkLog2Size);

            BCacheInvalidateBlks(pAreaInfo->pVolInfo, startExtent, freeBlks); // must call before FreeBlkInsert(),
                                                                              // becase a block inserted can be used as a new control block in FreeBlkInsert().
            result = FreeBlkInsert(pAreaInfo, &iter, startExtent, freeBlks);
            if (result != WFSKRN_RESULT_OK)
            {
                break;
            }
            startExtent = *aBlkAdr;
        }

        prevBlkAdr = aBlkAdr;
        aBlkAdr = (WFSBlkAdr*) ((u8*) aBlkAdr + nStride);
    }

    if (result == WFSKRN_RESULT_OK)
    {
        header->totalBlks += nNumBlks * (1<<nBlkLog2Size);
    }
    else
    {
        // [check] convert error. NYI.
    }

    FBAItrClose(pAreaInfo, &iter);

#ifdef _DEBUG
    WFSKRN_FTREE_ASSERT(BCacheStopSkipCheckPin(pinCount) == WFSKRN_RESULT_OK);
#endif // DEBUG
    // the P-tree block is pinned.
    return result;
}

static u32 GetNumMetaBlks(AreaInfo* area)
{
    u32 numMetaBlks;
    WFSKrnResult result;
    WFSBlkAdr nBlkAdr;
    WFSBlkAdr ftreeBlk;
    FBATreeItr iter;
    EPTreeItr* epti;
    u8 level;
    u8 ptreeHeight;

    WFSBlkAdr node[EPTREE_MAX_HEIGHT];

    // the P-tree block is pinned.
    FBAItrOpen(area, &iter);
    ptreeHeight = iter.eptreeHdr[0]->height;

    memset(node, 0, sizeof(node));
    node[0] = FBA_ROOT_BLK;
    numMetaBlks = 1;

    result = EPTreeFirst(area, (MEPTreeItr*) &iter);
    while (result == WFSKRN_RESULT_OK)
    {
        for (level = 1; level < ptreeHeight; ++level)
        {
            if (node[level] != iter.eptreeHdr[level]->addr)
            {
                node[level] = iter.eptreeHdr[level]->addr;
                ++numMetaBlks;
            }
        }

        epti = EPTreeGetLastIter((MEPTreeItr*) &iter);
        nBlkAdr = PTREE_GET_KEY(epti);
        ftreeBlk = PTREE_GET_DATA(epti);
        ++numMetaBlks;

        result = EPTreeNext(area, (MEPTreeItr*) &iter);
    }

    FBAItrClose(area, &iter);
    return numMetaBlks;
}

u32 FreeBlkGetFreeSize(AreaInfo *pAreaInfo)
{
    FBATreeHdr* header;

    header = GET_FBATREE_HDR(pAreaInfo);
    u32 numMetaBlks = GetNumMetaBlks(pAreaInfo);
    WFSKRN_FTREE_ASSERT(numMetaBlks <= header->totalBlks);

    return header->totalBlks - numMetaBlks; // unit is 8kB
}

//
// Debug functions.
//

WFSBool FreeBlkVerify(AreaInfo* area)
{
    WFSKrnResult result;
    WFSBlkAdr nBlkAdr;
    FBATreeItr iterSeq; // sequential
    FBATreeItr iter;
    EPTreeHdr* eptreeHdr;
    u32 total = 0;
    u32 totalFound = 0;
    u32 numMetaBlks;
    WFSBool ret;
    int i;

    // the P-tree block is pinned.

#if _DEBUG
    FBATreeHdr* header = GET_FBATREE_HDR(area);
    u32 pinCount = BCacheSkipCheckPin();
#endif

    eptreeHdr = GET_EPTREEHDR_FROM_AREAINFO(area);
    WFSKRN_FTREE_ASSERT(eptreeHdr->ptreeHdr.nRootHeight < PTREE_MAX_HEIGHT);

    FBAItrOpen(area, &iter);
    FBAItrOpen(area, &iterSeq);

    for (i = 0; i < MAX_SUB_TREE; ++i)
    {

        result = FreeBlkLast(area, &iterSeq, FTreeBlkLog2Ofs[i]);
        while (result == WFSKRN_RESULT_OK)
        {
            nBlkAdr = FBA_RESULT_KEY(&iterSeq);

            total += (FBA_RESULT_DATA(&iterSeq) + 1) * (1<<FTreeBlkLog2Ofs[i]);
            result = FreeBlkFindStrict(area, &iter, nBlkAdr, FTreeBlkLog2Ofs[i]);
            FBAItrClose(area, &iter);
            if (result != WFSKRN_RESULT_OK)
            {
                ret = (WFSBool) FALSE;
                goto FTREEBLK_VERIFY_RESULT;
            }
            totalFound += (FBA_RESULT_DATA(&iter) + 1) * (1<<FTreeBlkLog2Ofs[i]);
            result = FreeBlkPrev(area, &iterSeq);
        }
    }
    WFSKRN_FTREE_ASSERT(total == totalFound);
    numMetaBlks = GetNumMetaBlks(area);
    WFSKRN_FTREE_ASSERT(total == header->totalBlks - numMetaBlks);
    ret = (WFSBool) TRUE;

FTREEBLK_VERIFY_RESULT:
    FBAItrClose(area, &iterSeq);

#if _DEBUG
    WFSKRN_FTREE_ASSERT(BCacheStopSkipCheckPin(pinCount) == WFSKRN_RESULT_OK);
#endif
    // the P-tree block is pinned.
    return ret;
}

u32 FreeBlkGetFreeBlkLen(AreaInfo* area, u32 nBlkLog2SizeOfs)
{
    WFSKrnResult result;
    WFSBlkAdr nBlkAdr;
    FBATreeItr iter;
    u32 total = 0;
    EPTreeHdr* eptreeHdr;

#ifdef _DEBUG
    u32 pinCount = BCacheSkipCheckPin();
#endif // _DEBUG

    // the P-tree block is pinned.

    eptreeHdr = GET_EPTREEHDR_FROM_AREAINFO(area);
    WFSKRN_FTREE_ASSERT(eptreeHdr->ptreeHdr.nRootHeight < PTREE_MAX_HEIGHT);

    FBAItrOpen(area, &iter);

    result = FreeBlkLast(area, &iter, nBlkLog2SizeOfs);
    while (result == WFSKRN_RESULT_OK)
    {
        nBlkAdr = FBA_RESULT_KEY(&iter);
        total += (FBA_RESULT_DATA(&iter) + 1) * (1<<nBlkLog2SizeOfs);
        result = FreeBlkPrev(area, &iter);
    }

    FBAItrClose(area, &iter);

#ifdef _DEBUG
    WFSKRN_FTREE_ASSERT(BCacheStopSkipCheckPin(pinCount) == WFSKRN_RESULT_OK);
#endif // _DEBUG
    // the P-tree block is pinned.
    return total;
}

static void SetBit(WFSSrvFileHandle fh, u32 diskAddr)
{
    u32 byte = diskAddr / 8;
    u32 bit = diskAddr % 8;
    u8 flag = (1 << bit);

    u8 dst;

    WFSResult nResult;
    nResult = (WFSResult)WFSSrvReadFile(fh, &dst, byte, 1);
    if(nResult < WFS_RESULT_OK){
        osTPrintf("SetBit() error %d\n", nResult);
    }

    WFSKRN_FTREE_ASSERT(0 == (dst & flag));

    dst |= flag;
    nResult = WFSSrvWriteFile(fh, &dst, byte, 1, true);

}

static void SetBitArrayInternal(WFSSrvFileHandle fh, u8* pBuffer, u32 bufferSize, u32 diskAddr, u32 length){

    u32 byte = diskAddr / 8;
    u32 bit  = diskAddr % 8;
    u32 byteLength = (diskAddr + length - 1) / 8  - byte + 1;

    WFSResult nResult;

    // read
    nResult = (WFSResult)WFSSrvReadFile(fh, pBuffer, byte, byteLength);
    if(nResult < WFS_RESULT_OK){
        osTPrintf("SetBitArray() error %d\n", nResult);
    }

    // set bits
    int i;
    for(i=bit; i<bit+length; ++i){
        u8 flag = 1 << (i & 0x07);
        WFSKRN_FTREE_ASSERT( !(pBuffer[i/8] & flag) );
        pBuffer[i/8] |= flag;
    }

    // write
    nResult = WFSSrvWriteFile(fh, pBuffer, byte, byteLength, true);

}

static void SetBitArray(WFSSrvFileHandle fh, u8* pBuffer, u32 bufferSize, u32 diskAddr, u32 length){

    while(length > (bufferSize-1)*8){
        SetBitArrayInternal(fh, pBuffer, bufferSize, diskAddr, (bufferSize-1)*8);
        length   -= (bufferSize-1)*8;
        diskAddr += (bufferSize-1)*8;
    }
    SetBitArrayInternal(fh, pBuffer, bufferSize, diskAddr, length);
}


static WFSKrnResult FreeBlkGetMetablock(AreaInfo* area, WFSSrvFileHandle fh, u32 size)
{
    WFSKrnResult result;
    WFSBlkAdr nBlkAdr;
    WFSBlkAdr ftreeBlk;
    FBATreeItr iter;
    EPTreeItr* epti;
    u8 level;
    u8 ptreeHeight;

    WFSBlkAdr node[EPTREE_MAX_HEIGHT];

    // the P-tree block is pinned.
    FBAItrOpen(area, &iter);
    ptreeHeight = iter.eptreeHdr[0]->height;

    memset(node, 0, sizeof(node));
    node[0] = FBA_ROOT_BLK;
    SetBit(fh, FBA_ROOT_BLK);

    result = EPTreeFirst(area, (MEPTreeItr*) &iter);
    while (result == WFSKRN_RESULT_OK)
    {
        for (level = 1; level < ptreeHeight; ++level)
        {
            if (node[level] != iter.eptreeHdr[level]->addr)
            {
                node[level] = iter.eptreeHdr[level]->addr;
                SetBit(fh, node[level]);
            }
        }

        epti = EPTreeGetLastIter((MEPTreeItr*) &iter);
        nBlkAdr = PTREE_GET_KEY(epti);
        ftreeBlk = PTREE_GET_DATA(epti);
        SetBit(fh, ftreeBlk);

        result = EPTreeNext(area, (MEPTreeItr*) &iter);
    }

    FBAItrClose(area, &iter);

    // the P-tree block is pinned.
    return WFSKRN_RESULT_OK;
}

static WFSKrnResult FreeBlkGetFreeblock(AreaInfo* area, WFSSrvFileHandle fh, u8* pBuffer, u32 bufferSize, u32 size)
{
    WFSKrnResult result;
    WFSBlkAdr start;
    u32 len;
    int i;
    FBATreeItr iter;

    // the P-tree block is pinned.
    FBAItrOpen(area, &iter);

    for (i = 0; i < MAX_SUB_TREE; ++i)
    {
        result = FreeBlkFirst(area, &iter, FTreeBlkLog2Ofs[i]);
        while (result == WFSKRN_RESULT_OK)
        {
            start = FBA_RESULT_KEY(&iter);
            len = (FBA_RESULT_DATA(&iter) + 1) * (1<<FTreeBlkLog2Ofs[i]);
            WFSKRN_FTREE_ASSERT(start < area->ah.nNumBlks);
            WFSKRN_FTREE_ASSERT(start + len <= area->ah.nNumBlks);
            SetBitArray(fh, pBuffer, bufferSize, start, len);
            result = FreeBlkNext(area, &iter);
        }
    }
    FBAItrClose(area, &iter);

    // the P-tree block is pinned.
    return WFSKRN_RESULT_OK;
}

WFSKrnResult FreeBlkCheckDisk(AreaInfo* area, WFSSrvFileHandle fh, u8* pBuffer, u32 bufferSize, u32 size)
{
    // the P-tree block is pinned.

    WFSKrnResult result;

#ifdef _DEBUG
    EPTreeHdr* eptreeHdr = GET_EPTREEHDR_FROM_AREAINFO(area);
    u32 pinCount = BCacheSkipCheckPin();
#endif // DEBUG

    WFSKRN_FTREE_ASSERT(eptreeHdr->ptreeHdr.nRootHeight < PTREE_MAX_HEIGHT);
    result = FreeBlkGetMetablock(area, fh, size);
    if (result != WFSKRN_RESULT_OK)
    {
        FreeBlkDumpFreeblock(area);
        goto FTREEBLK_CHECKDISK_RESULT;
    }
    result = FreeBlkGetFreeblock(area, fh, pBuffer, bufferSize, size);
    if (result != WFSKRN_RESULT_OK)
    {
        FreeBlkDumpFreeblock(area);
        goto FTREEBLK_CHECKDISK_RESULT;
    }

FTREEBLK_CHECKDISK_RESULT:

#ifdef _DEBUG
    WFSKRN_FTREE_ASSERT(BCacheStopSkipCheckPin(pinCount) == WFSKRN_RESULT_OK);
#endif // DEBUG

    // the P-tree block is pinned.
    return result;
}

WFSKrnResult FreeBlkDumpFreeblock(AreaInfo* area)
{
    WFSKrnResult result;
    WFSBlkAdr start;
    u32 len;
    int i;
    FBATreeItr iter;

    // the P-tree block is pinned.
    FBAItrOpen(area, &iter);

    for (i = 0; i < MAX_SUB_TREE; ++i)
    {
        result = FreeBlkFirst(area, &iter, FTreeBlkLog2Ofs[i]);
        while (result == WFSKRN_RESULT_OK)
        {
            start = FBA_RESULT_KEY(&iter);
            len = FBA_RESULT_DATA(&iter) + 1;
#ifdef FBA_DEBUG_VERBOSE
            if (FBADebugVerbose)
            {
                osTPrintf("start %u (len %u)\n", start, len);
            }
#endif //  FBA_DEBUG_VERBOSE
            result = FreeBlkNext(area, &iter);
        }
    }

    FBAItrClose(area, &iter);

    // the P-tree block is pinned.
    return WFSKRN_RESULT_OK;
}

#endif // BTREE_BASED_ALLOCATOR
