/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfskrn_Handles.cpp - Module for file handles and search directory handles


  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_Handles.cpp,v $
  Revision 1.2  2008/10/07 06:00:57  ueno
  Modified WFSKrnFileInfoAlloc() to set WFSKRN_HANDLE_ALLOCATED in order to indicate the handle is allocated.

  Revision 1.1  2008/04/25 17:29:06  kondo_masahiro
  Initial check-in

  Revision 1.1  2008/04/23 00:27:17  paul
  Copied and modified from wfscli_Handles.cpp

*---------------------------------------------------------------------------*/

#include "wfskrn_Api.h"

static bool bMutexInitialized = false; // [check]

void WFSKrnInitHandles() {
    if (!bMutexInitialized)
    {
        bMutexInitialized = true;
        WFSKrnInitMutex(&wkg.hnd.mxAccessTheFileInfo);
        WFSKrnInitMutex(&wkg.hnd.mxAccessTheSearchDirInfo);
    }
    wkg.hnd.nFileHandleCounter        = 0;
    wkg.hnd.nSearchDirHandleCounter   = 0;
    wkg.hnd.nFirstFreeFileHandle      = 0;
    wkg.hnd.nFirstFreeSearchDirHandle = 0;
    wkg.hnd.nNumUsedFileHandles       = 0;
    wkg.hnd.nNumUsedSearchDirHandles  = 0;
    //wkg.hnd.nFirstUsedFileHandle      = WFSKRN_MAX_FILE_HANDLES;  // A value of WFSKRN_MAX_FILE_HANDLES is the list terminator
    //wkg.hnd.nFirstUsedSearchDirHandle = WFSKRN_MAX_SEARCH_DIR_HANDLES;  // A value of WFSKRN_MAX_SEARCH_DIR_HANDLES is the list terminator
    u32 nI;
    for(nI=0; nI<WFSKRN_MAX_FILE_HANDLES; nI++) {
        wkg.hnd.aFileInfo[nI].nNext = nI+1;
    }
    for(nI=0; nI<WFSKRN_MAX_SEARCH_DIR_HANDLES; nI++) {
        wkg.hnd.aSearchDirInfo[nI].nNext = nI+1;
    }
}

WFSKrnResult WFSKrnGetFileInfo(WFSKrnFileInfo **ppFi, WFSSrvFileHandle sfh) {
    *ppFi = &wkg.hnd.aFileInfo[sfh&(WFSKRN_MAX_FILE_HANDLES-1)];
    return ((*ppFi)->nHandle == sfh)?WFSKRN_RESULT_OK:WFSKRN_RESULT_INVALID;
}

WFSKrnResult WFSKrnGetSearchDirInfo(WFSKrnSearchDirInfo **ppSdi, WFSSrvSearchDirectoryHandle ssdh) {
    *ppSdi = &wkg.hnd.aSearchDirInfo[ssdh&(WFSKRN_MAX_SEARCH_DIR_HANDLES-1)];
    return ((*ppSdi)->nHandle == ssdh)?WFSKRN_RESULT_OK:WFSKRN_RESULT_INVALID;
}

WFSKrnResult WFSKrnFileInfoAlloc(WFSKrnFileInfo **ppFi) {
    WFSKrnLockMutex(&wkg.hnd.mxAccessTheFileInfo);
    if (wkg.hnd.nFirstFreeFileHandle == WFSKRN_MAX_FILE_HANDLES) {
        WFSKrnUnlockMutex(&wkg.hnd.mxAccessTheFileInfo);
        return WFSKRN_RESULT_MAX_HANDLES;
    }
    wkg.hnd.nNumUsedFileHandles++;
    u32 nIdx = wkg.hnd.nFirstFreeFileHandle;
    wkg.hnd.nFirstFreeFileHandle = wkg.hnd.aFileInfo[nIdx].nNext;
	wkg.hnd.aFileInfo[nIdx].nNext = WFSKRN_HANDLE_ALLOCATED;
    *ppFi = &wkg.hnd.aFileInfo[nIdx];
    wkg.hnd.aFileInfo[nIdx].nHandle = wkg.hnd.nFileHandleCounter | nIdx;
    wkg.hnd.nFileHandleCounter += WFSKRN_FILE_HANDLE_INC;
    wkg.hnd.nFileHandleCounter &= WFSKRN_HANDLE_COUNTER_MASK;
    WFSKrnUnlockMutex(&wkg.hnd.mxAccessTheFileInfo);
    return WFSKRN_RESULT_OK;
}

WFSKrnResult WFSKrnSearchDirInfoAlloc(WFSKrnSearchDirInfo **ppSdi) {

    WFSKrnLockMutex(&wkg.hnd.mxAccessTheSearchDirInfo);
    if (wkg.hnd.nFirstFreeSearchDirHandle == WFSKRN_MAX_SEARCH_DIR_HANDLES) {
        WFSKrnUnlockMutex(&wkg.hnd.mxAccessTheSearchDirInfo);
        return WFSKRN_RESULT_MAX_HANDLES;
    }
    wkg.hnd.nNumUsedSearchDirHandles++;
    u32 nIdx = wkg.hnd.nFirstFreeSearchDirHandle;
    wkg.hnd.nFirstFreeSearchDirHandle = wkg.hnd.aSearchDirInfo[nIdx].nNext;
	wkg.hnd.aSearchDirInfo[nIdx].nNext = WFSKRN_HANDLE_ALLOCATED;
    *ppSdi = &wkg.hnd.aSearchDirInfo[nIdx];
    wkg.hnd.aSearchDirInfo[nIdx].nHandle = wkg.hnd.nSearchDirHandleCounter | nIdx;
    wkg.hnd.nSearchDirHandleCounter += WFSKRN_SEARCH_DIR_HANDLE_INC;
    wkg.hnd.nSearchDirHandleCounter &= WFSKRN_HANDLE_COUNTER_MASK;
    WFSKrnUnlockMutex(&wkg.hnd.mxAccessTheSearchDirInfo);
    return WFSKRN_RESULT_OK;
}

void WFSKrnFileInfoFree(WFSKrnFileInfo *pFi) {
    WFSKrnLockMutex(&wkg.hnd.mxAccessTheFileInfo);
    u32 nIdx = (u32)(pFi - wkg.hnd.aFileInfo);
    pFi->nHandle ^= 0x5555; // ensures the deleted handle is invalid
    pFi->nNext = wkg.hnd.nFirstFreeFileHandle;
    wkg.hnd.nNumUsedFileHandles--;
    wkg.hnd.nFirstFreeFileHandle = nIdx;
    WFSKrnUnlockMutex(&wkg.hnd.mxAccessTheFileInfo);
}

void WFSKrnSearchDirInfoFree(WFSKrnSearchDirInfo *pSdi) {
    WFSKrnLockMutex(&wkg.hnd.mxAccessTheSearchDirInfo);
    u32 nIdx = (u32)(pSdi - wkg.hnd.aSearchDirInfo);
    pSdi->nHandle ^= 0x5555; // ensures the deleted handle is invalid
    pSdi->nNext = wkg.hnd.nFirstFreeSearchDirHandle;
    wkg.hnd.nNumUsedSearchDirHandles--;
    wkg.hnd.nFirstFreeSearchDirHandle = nIdx;
    WFSKrnUnlockMutex(&wkg.hnd.mxAccessTheSearchDirInfo);
}
