/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfskrn_PTree.cpp


  Copyright 2007-2008 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_PTree.cpp,v $
  Revision 1.5  2008/10/14 10:32:51  kondo_masahiro
  Added #define _DEBUG_BREAK_POINT

  Revision 1.4  2008/08/22 01:20:30  ueno
  Modified line feed.

  Revision 1.3  2008/08/22 01:07:54  ueno
  Fixed PTreeInsert().

  Revision 1.2  2008/07/17 23:00:28  paul
  PTree is now used by updateMap.
  Changed name of adapter functions to the Sba allocators.
  Fixed PTreeInsert() to update pPti->aEntryIdx[0].
  Added PTreeInsertOrUpdateRec().

  Revision 1.1  2008/07/09 01:43:05  paul
  u32->u32 mapping optimized B-Tree. (for use by new block allocator, and transaction block re-mapper)

*---------------------------------------------------------------------------*/

#undef  _KRN_FILE_
#define _KRN_FILE_  "PTree"

#include "wfskrn_Defs.h"
#include "wfskrn_Api.h"
#include "wfskrn_Area.h"
#include "wfskrn_Errors.h"
#include "wfskrn_PTree.h"
#include "randomlib.h"

#if _DEBUG_PTREE
    static u32 nCurKey;
    static u32 nNodeCount;
#endif


u16 PTreeAllocNodeFromSba(void *pBase) {
    return SbaAllocSubBlk((WFSSubBlkAllocHdr *)pBase, PTREE_LOG2_NODE_SIZE);
}

WFSKrnResult PTreeAllocNodesFromSba(void *pBase, u16 aNodeOfs[], u32 nNumNodes) {
    return SbaAllocSubBlks((WFSSubBlkAllocHdr *)pBase, PTREE_LOG2_NODE_SIZE, aNodeOfs, nNumNodes);
}

void PTreeFreeNodeToSba(void *pBase, u16 nOfs) {
    SbaFreeSubBlk((WFSSubBlkAllocHdr *)pBase, nOfs, PTREE_LOG2_NODE_SIZE);
}

//inline
PTreeNode *PTreeGetNode(PTreeAllocator *pAllocator, u32 nOfs) {
    return (PTreeNode *)((u8*)pAllocator->pBase + nOfs);
}

//inline
PTreeLeaf *PTreeGetLeaf(PTreeAllocator *pAllocator, u32 nOfs) {
    return (PTreeLeaf *)((u8*)pAllocator->pBase + nOfs);
}

/*inline*/
u32 PTreeGetOfs(void *pNode, PTreeAllocator *pAllocator) {
    return (u8*)pNode - (u8*)pAllocator->pBase;
}


void PTreeInit(PTreeHdr *pPTreeHdr, PTreeAllocator *pAllocator) {
    pPTreeHdr->nRootHeight = 0;
    pPTreeHdr->nNumRecs = 0;
}

void PTreeInitRootNode(PTreeHdr *pPTreeHdr, PTreeAllocator *pAllocator) {
    pPTreeHdr->nRootOfs = pAllocator->fpAllocNode(pAllocator->pBase);
    PTreeNode *pNode = PTreeGetNode(pAllocator, (u32)pPTreeHdr->nRootOfs);
    memset(pNode, 0, sizeof(PTreeNode));
}


inline u32 PTreeFindLeafOfs(PTreeHdr *pPTreeHdr, PTreeItr *pPti, u32 nKey, PTreeAllocator *pAllocator) {
    u32 nLevel = pPTreeHdr->nRootHeight;
    u32 *pEntryIdx = &pPti->aEntryIdx[nLevel];
    u32 nOfs = (u32)pPTreeHdr->nRootOfs;
    //  K |  Input      | Out  | K<=k[0] | K<=k[1] | K<=k[2] | K<=k[3] | K<=k[4] | 0==k[0] | 0==k[1] | 0==k[2] | 0==k[3] | 0==k[4] |
    // ---+-------------+------+---------+---------+---------+---------+---------+---------+---------+---------+---------+---------+
    //  K |  0 0 0 0 0  |  0   |    0    |    0    |    0    |    0    |    0    |    1    |    1    |    1    |    1    |    1    |
    //  K |  a 0 0 0 0  |  1   |    0    |    0    |    0    |    0    |    0    |    0    |    1    |    1    |    1    |    1    |
    //  K |  z 0 0 0 0  |  0   |    1    |    0    |    0    |    0    |    0    |    0    |    1    |    1    |    1    |    1    |
    //  K |  a b 0 0 0  |  2   |    0    |    0    |    0    |    0    |    0    |    0    |    0    |    1    |    1    |    1    |
    //  K |  a z 0 0 0  |  1   |    0    |    1    |    0    |    0    |    0    |    0    |    0    |    1    |    1    |    1    |
    //  K |  y z 0 0 0  |  0   |    1    |    1    |    0    |    0    |    0    |    0    |    0    |    1    |    1    |    1    |
    //  K |  a b c 0 0  |  3   |    0    |    0    |    0    |    0    |    0    |    0    |    0    |    0    |    1    |    1    |
    //  K |  a b z 0 0  |  2   |    0    |    0    |    1    |    0    |    0    |    0    |    0    |    0    |    1    |    1    |
    //  K |  a y z 0 0  |  1   |    0    |    1    |    1    |    0    |    0    |    0    |    0    |    0    |    1    |    1    |
    //  K |  x y z 0 0  |  0   |    1    |    1    |    1    |    0    |    0    |    0    |    0    |    0    |    1    |    1    |
    //  K |  a b c d 0  |  4   |    0    |    0    |    0    |    0    |    0    |    0    |    0    |    0    |    0    |    1    |
    //  K |  a b c z 0  |  3   |    0    |    0    |    0    |    1    |    0    |    0    |    0    |    0    |    0    |    1    |
    //  K |  a b y z 0  |  2   |    0    |    0    |    1    |    1    |    0    |    0    |    0    |    0    |    0    |    1    |
    //  K |  a x y z 0  |  1   |    0    |    1    |    1    |    1    |    0    |    0    |    0    |    0    |    0    |    1    |
    //  K |  w x y z 0  |  0   |    1    |    1    |    1    |    1    |    0    |    0    |    0    |    0    |    0    |    1    |
    //  K |  a b c d e  |  5   |    0    |    0    |    0    |    0    |    0    |    0    |    0    |    0    |    0    |    0    |
    //  K |  a b c d z  |  4   |    0    |    0    |    0    |    0    |    1    |    0    |    0    |    0    |    0    |    0    |
    //  K |  a b c y z  |  3   |    0    |    0    |    0    |    1    |    1    |    0    |    0    |    0    |    0    |    0    |
    //  K |  a b x y z  |  2   |    0    |    0    |    1    |    1    |    1    |    0    |    0    |    0    |    0    |    0    |
    //  K |  a w x y z  |  1   |    0    |    1    |    1    |    1    |    1    |    0    |    0    |    0    |    0    |    0    |
    //  K |  v w x y z  |  0   |    1    |    1    |    1    |    1    |    1    |    0    |    0    |    0    |    0    |    0    |
    //  --+-------------+------+---------+---------+---------+---------+---------+---------+---------+---------+---------+---------+
    //                              5         8         9         8         5         0         2         5         9         14
    while (nLevel) {
        PTreeNode *pNode = pPti->apNode[nLevel-1] = PTreeGetNode(pAllocator, nOfs);
        // Hard-coded binary split search which avoids need to store length of sorted array, by assuming a 0 terminated array
        if (pNode->aKey[3] == 0) {
            // 10 possibilities
            if (pNode->aKey[2] == 0) {
                // 6 possibilities
                if (pNode->aKey[1] != 0) {
                    // 3 possibilities
                    goto FindInRange0to2;
                }
                // 3 possibilities
                if (pNode->aKey[0] > nKey) {
                    *pEntryIdx = 0;
                } else {
                    if (pNode->aKey[0] == 0) {
                        *pEntryIdx = 0;
                    } else {
                        *pEntryIdx = 1;
                    }
                }
            } else {
                // 4 possibilities
                if (pNode->aKey[1] > nKey) {
                    goto FindInRange0to1;
                } else {
                    if (pNode->aKey[2] > nKey) {
                        *pEntryIdx = 2;
                    } else {
                        *pEntryIdx = 3;
                    }
                }
            }
        } else {
            // 11 possibilities
            if (pNode->aKey[4] == 0) {
                // 5 possibilities
                if (pNode->aKey[2] > nKey) {
                    // 3 possibilities
                    goto FindInRange0to2;
                }
                // 2 possibilities
                if (pNode->aKey[3] > nKey) {
                    *pEntryIdx = 3;
                } else {
                    *pEntryIdx = 4;
                }
            } else {
                // 6 possibilities
                if (pNode->aKey[2] > nKey) {
                    // 3 possibilities
FindInRange0to2:
                    if (pNode->aKey[1] > nKey) {
                        // 2 possibilities
FindInRange0to1:
                        if (pNode->aKey[0] > nKey) {
                            *pEntryIdx = 0;
                        } else {
                            *pEntryIdx = 1;
                        }
                    } else {
                        *pEntryIdx = 2;
                    }
                } else {
                    // 3 possibilities
                    if (pNode->aKey[4] > nKey) {
                        // 2 possibilities
                        if (pNode->aKey[3] > nKey) {
                            *pEntryIdx = 3;
                        } else {
                            *pEntryIdx = 4;
                        }
                    } else {
                        *pEntryIdx = 5;
                    }
                }
            }
        }
        nOfs = (u32)pNode->aOfs[*pEntryIdx];
        --pEntryIdx;
        --nLevel;
    }
    return nOfs;
}


WFSKrnResult PTreeFind(PTreeHdr *pPTreeHdr, PTreeItr *pPti, u32 nKey, PTreeAllocator *pAllocator) {
    if (pPTreeHdr->nNumRecs == 0) {
        return WFSKRN_RESULT_PTREE_ENTRY_NOT_FOUND;
    }
    u32 nOfs = PTreeFindLeafOfs(pPTreeHdr, pPti, nKey, pAllocator);
    PTreeLeaf *pLeaf = pPti->pLeaf = PTreeGetLeaf(pAllocator, nOfs);
    // 32 cases to distinguish
    //  Insert(K) : Different cases for leaf update
    //  K |  Input    | Output       | K<=k[0] | K>=k[0] | K<=k[1] | K>=k[1] | K<=k[2] | K>=k[2] | K<=k[3] | K>=k[3] | 0==k[0] | 0==k[1] | 0==k[2] | 0==k[3] |
    //  --+-----------+--------------+---------+---------+---------+---------+---------+---------+---------+---------+---------+---------+---------+---------+
    //  0 |  k 0 0 0  |  0 Found     |    1    |    1    |    1    |    1    |    1    |    1    |    1    |    1    |    1    |    1    |    1    |    1    |
    //  K |  k 0 0 0  |  0 Found     |    1    |    1    |    0    |    1    |    0    |    1    |    0    |    1    |    0    |    1    |    1    |    1    |
    //  0 |  k t 0 0  |  0 Found     |    1    |    1    |    1    |    0    |    1    |    1    |    1    |    1    |    1    |    0    |    1    |    1    |
    //  K |  k t 0 0  |  0 Found     |    1    |    1    |    1    |    0    |    0    |    1    |    0    |    1    |    0    |    0    |    1    |    1    |
    //  0 |  k r t 0  |  0 Found     |    1    |    1    |    1    |    0    |    1    |    0    |    1    |    1    |    1    |    0    |    0    |    1    |
    //  K |  k r t 0  |  0 Found     |    1    |    1    |    1    |    0    |    1    |    0    |    0    |    1    |    0    |    0    |    0    |    1    |
    //  0 |  k p r t  |  0 Found     |    1    |    1    |    1    |    0    |    1    |    0    |    1    |    0    |    1    |    0    |    0    |    0    |
    //  K |  k p r t  |  0 Found     |    1    |    1    |    1    |    0    |    1    |    0    |    1    |    0    |    0    |    0    |    0    |    0    |
    //  0 |  r t 0 0  |  0 Not Found |    1    |    0    |    1    |    0    |    1    |    1    |    1    |    1    |    0    |    0    |    1    |    1    |
    //  K |  r t 0 0  |  0 Not Found |    1    |    0    |    1    |    0    |    0    |    1    |    0    |    1    |    0    |    0    |    1    |    1    |
    //  0 |  p r t 0  |  0 Not Found |    1    |    0    |    1    |    0    |    1    |    0    |    1    |    1    |    0    |    0    |    0    |    1    |
    //  K |  p r t 0  |  0 Not Found |    1    |    0    |    1    |    0    |    1    |    0    |    0    |    1    |    0    |    0    |    0    |    1    |
    //  0 |  n p r t  |  0 Not Found |    1    |    0    |    1    |    0    |    1    |    0    |    1    |    0    |    0    |    0    |    0    |    0    |
    //  K |  n p r t  |  0 Not Found |    1    |    0    |    1    |    0    |    1    |    0    |    1    |    0    |    0    |    0    |    0    |    0    |
    //  0 |  t 0 0 0  |  0 Not Found |    1    |    0    |    1    |    1    |    1    |    1    |    1    |    1    |    0    |    1    |    1    |    1    |
    //  K |  t 0 0 0  |  0 Not Found |    1    |    0    |    0    |    1    |    0    |    1    |    0    |    1    |    0    |    1    |    1    |    1    |
    //  K |  b k t 0  |  1 Found     |    0    |    1    |    1    |    1    |    1    |    0    |    0    |    1    |    0    |    0    |    0    |    1    |
    //  K |  b k 0 0  |  1 Found     |    0    |    1    |    1    |    1    |    0    |    1    |    0    |    1    |    0    |    0    |    1    |    1    |
    //  K |  b k r t  |  1 Found     |    0    |    1    |    1    |    1    |    1    |    0    |    1    |    0    |    0    |    0    |    0    |    0    |
    //  K |  b r t 0  |  1 Not Found |    0    |    1    |    1    |    0    |    1    |    0    |    0    |    1    |    0    |    0    |    0    |    1    |
    //  K |  b 0 0 0  |  1 Not Found |    0    |    1    |    0    |    1    |    0    |    1    |    0    |    1    |    0    |    1    |    1    |    1    |
    //  K |  b t 0 0  |  1 Not Found |    0    |    1    |    1    |    0    |    0    |    1    |    0    |    1    |    0    |    0    |    1    |    1    |
    //  K |  b p r t  |  1 Not Found |    0    |    1    |    1    |    0    |    1    |    0    |    1    |    0    |    0    |    0    |    0    |    0    |
    //  K |  b d k 0  |  2 Found     |    0    |    1    |    0    |    1    |    1    |    1    |    0    |    1    |    0    |    0    |    0    |    1    |
    //  K |  b d k t  |  2 Found     |    0    |    1    |    0    |    1    |    1    |    1    |    1    |    0    |    0    |    0    |    0    |    0    |
    //  K |  b d 0 0  |  2 Not Found |    0    |    1    |    0    |    1    |    0    |    1    |    0    |    1    |    0    |    0    |    1    |    1    |
    //  K |  b d t 0  |  2 Not Found |    0    |    1    |    0    |    1    |    1    |    0    |    0    |    1    |    0    |    0    |    0    |    1    |
    //  K |  b d r t  |  2 Not Found |    0    |    1    |    0    |    1    |    1    |    0    |    1    |    0    |    0    |    0    |    0    |    0    |
    //  K |  b d f k  |  3 Found     |    0    |    1    |    0    |    1    |    0    |    1    |    1    |    1    |    0    |    0    |    0    |    0    |
    //  K |  b d f 0  |  3 Not Found |    0    |    1    |    0    |    1    |    0    |    1    |    0    |    1    |    0    |    0    |    0    |    1    |
    //  K |  b d f t  |  3 Not Found |    0    |    1    |    0    |    1    |    0    |    1    |    1    |    0    |    0    |    0    |    0    |    0    |
    //  K |  b d f j  |  4 Not Found |    0    |    1    |    0    |    1    |    0    |    1    |    0    |    1    |    0    |    0    |    0    |    0    |
    //  --+-----------+--------------+---------+---------+---------+---------+---------+---------+---------+---------+---------+---------+---------+---------+
    //                                   16        24        20        17        20        18        16        23         4         5        12        21
    u32 *pEntryIdx = pPti->aEntryIdx;
    if ((pLeaf->aKey[1] >= nKey) || (pLeaf->aKey[1] == 0)) {
        if (pLeaf->aKey[0] >= nKey) {
            *pEntryIdx = 0;
        } else {
            *pEntryIdx = 1;
        }
    } else {
        if (pLeaf->aKey[2] >= nKey) {
            *pEntryIdx = 2;
        } else {
            if (pLeaf->aKey[3] == 0) {
                if (pLeaf->aKey[2] == 0) {
                    *pEntryIdx = 2;
                } else {
                    *pEntryIdx = 3;
                }
            } else {
                if (pLeaf->aKey[3] < nKey) {
                    *pEntryIdx = 4;
                    return WFSKRN_RESULT_PTREE_ENTRY_NOT_FOUND;
                } else {
                    *pEntryIdx = 3;
                }
            }
        }
    }
    if (pLeaf->aKey[*pEntryIdx] == nKey) {
        pPti->nData = pLeaf->aData[*pEntryIdx];
        return WFSKRN_RESULT_PTREE_ENTRY_FOUND;
    }
    return WFSKRN_RESULT_PTREE_ENTRY_NOT_FOUND;
}


WFSKrnResult PTreeInsertIntoNode(PTreeHdr *pPTreeHdr, PTreeItr *pPti, u32 nKey, u32 *pChildOfs, PTreeAllocator *pAllocator) {
    // *pChildOfs [IN/OUT] = On entry, contains the address of the left child. On output, contains the offset of the right child.
    //  K |  Input      | Output                  | K<=k[0] | K<=k[1] | K<=k[2] | K<=k[3] | K<=k[4] | 0==k[0] | 0==k[1] | 0==k[2] | 0==k[3] | 0==k[4] |
    // ---+-------------+-------------------------+---------+---------+---------+---------+---------+---------+---------+---------+---------+---------+
    //  K |  0 0 0 0 0  |  a K 0 0 0              |    0    |    0    |    0    |    0    |    0    |    1    |    1    |    1    |    1    |    1    |
    //  K |  a 0 0 0 0  |  a K 0 0 0              |    0    |    0    |    0    |    0    |    0    |    0    |    1    |    1    |    1    |    1    |
    //  K |  z 0 0 0 0  |  K z 0 0 0              |    1    |    0    |    0    |    0    |    0    |    0    |    1    |    1    |    1    |    1    |
    //  K |  a b 0 0 0  |  a b K 0 0              |    0    |    0    |    0    |    0    |    0    |    0    |    0    |    1    |    1    |    1    |
    //  K |  a z 0 0 0  |  a K z 0 0              |    0    |    1    |    0    |    0    |    0    |    0    |    0    |    1    |    1    |    1    |
    //  K |  y z 0 0 0  |  K y z 0 0              |    1    |    1    |    0    |    0    |    0    |    0    |    0    |    1    |    1    |    1    |
    //  K |  a b c 0 0  |  a b c K 0              |    0    |    0    |    0    |    0    |    0    |    0    |    0    |    0    |    1    |    1    |
    //  K |  a b z 0 0  |  a b K z 0              |    0    |    0    |    1    |    0    |    0    |    0    |    0    |    0    |    1    |    1    |
    //  K |  a y z 0 0  |  a K y z 0              |    0    |    1    |    1    |    0    |    0    |    0    |    0    |    0    |    1    |    1    |
    //  K |  x y z 0 0  |  K x y z 0              |    1    |    1    |    1    |    0    |    0    |    0    |    0    |    0    |    1    |    1    |
    //  K |  a b c d 0  |  a b c d K              |    0    |    0    |    0    |    0    |    0    |    0    |    0    |    0    |    0    |    1    |
    //  K |  a b c z 0  |  a b c K z              |    0    |    0    |    0    |    1    |    0    |    0    |    0    |    0    |    0    |    1    |
    //  K |  a b y z 0  |  a b K y z              |    0    |    0    |    1    |    1    |    0    |    0    |    0    |    0    |    0    |    1    |
    //  K |  a x y z 0  |  a K x y z              |    0    |    1    |    1    |    1    |    0    |    0    |    0    |    0    |    0    |    1    |
    //  K |  w x y z 0  |  K w x y z              |    1    |    1    |    1    |    1    |    0    |    0    |    0    |    0    |    0    |    1    |
    //  K |  a b c d e  |  a b 0 0 0   e K 0 0 0  |    0    |    0    |    0    |    0    |    0    |    0    |    0    |    0    |    0    |    0    |
    //  K |  a b c d z  |  a b 0 0 0   K z 0 0 0  |    0    |    0    |    0    |    0    |    1    |    0    |    0    |    0    |    0    |    0    |
    //  K |  a b c y z  |  a b 0 0 0   y z 0 0 0  |    0    |    0    |    0    |    1    |    1    |    0    |    0    |    0    |    0    |    0    |
    //  K |  a b x y z  |  a b K 0 0   y z 0 0 0  |    0    |    0    |    1    |    1    |    1    |    0    |    0    |    0    |    0    |    0    |
    //  K |  a w x y z  |  a K w 0 0   y z 0 0 0  |    0    |    1    |    1    |    1    |    1    |    0    |    0    |    0    |    0    |    0    |
    //  K |  v w x y z  |  K v w 0 0   y z 0 0 0  |    1    |    1    |    1    |    1    |    1    |    0    |    0    |    0    |    0    |    0    |
    //  --+-------------+-------------------------+---------+---------+---------+---------+---------+---------+---------+---------+---------+---------+
    //                                                 5         8         9         8         5         0         2         5         9         14
    PTreeNode *pNode, *pNewNode;
    u32 nLeftChildOfs = *pChildOfs;
    u32 nLevel = 1;
    // Allocate the necessary number of new blocks first
    u32 nNumNewNodes = 1;   // One for new leaf, one for new parent
    PTreeNode **ppNode = pPti->apNode;
    PTreeNode **ppNodeEnd = ppNode + pPTreeHdr->nRootHeight;
    while(1) {
        if (ppNode == ppNodeEnd) {
            if ((ppNode - pPti->apNode) > PTREE_MAX_HEIGHT-2) {
                // We cannot split this node as it is the root node, and the tree is already at the maximum height
                return WFSKRN_RESULT_PTREE_FULL;
            }
            ++nNumNewNodes;
            break;
        }
        if ((*ppNode)->aKey[PTREE_NODE_FAN_OUT-2]==0) {
            break;
        }
        ++ppNode;
        ++nNumNewNodes;
    }
    u16 aNodeOfs[PTREE_MAX_HEIGHT];
    if (pAllocator->fpAllocNodes(pAllocator->pBase, aNodeOfs, nNumNewNodes) != WFSKRN_RESULT_OK) {
        return WFSKRN_RESULT_PTREE_FULL;
    }
    u32 nRightChildOfs = *pChildOfs = aNodeOfs[0];   // Offset of new leaf node
    u16 *pNewNodeOfs = aNodeOfs+1;
    while(1) {
        if (nLevel <= pPTreeHdr->nRootHeight) {
            pNode = pPti->apNode[nLevel-1];
        } else {
            // Add a new level
            ++pPTreeHdr->nRootHeight;
            pNode = pPti->apNode[nLevel-1] = PTreeGetNode(pAllocator, *pNewNodeOfs);
            pPTreeHdr->nRootOfs = (u16)*pNewNodeOfs;
            memset(&pNode->aKey[1], 0, sizeof(u32) * 4);
            pNode->aKey[0] = nKey;
            pNode->aOfs[0] = (u16)nLeftChildOfs;
            pNode->aOfs[1] = (u16)nRightChildOfs;
            return WFSKRN_RESULT_OK;
        }
        u32 *pEntryIdx = &pPti->aEntryIdx[nLevel];
        if (pNode->aKey[PTREE_NODE_FAN_OUT-2] == 0) {
            // Key can be inserted without requiring a node split
            switch(*pEntryIdx) {
                case 0: pNode->aKey[4] = pNode->aKey[3];
                        pNode->aKey[3] = pNode->aKey[2];
                        pNode->aKey[2] = pNode->aKey[1];
                        pNode->aKey[1] = pNode->aKey[0];
                        pNode->aKey[0] = nKey;
                        pNode->aOfs[5] = pNode->aOfs[4];
                        pNode->aOfs[4] = pNode->aOfs[3];
                        pNode->aOfs[3] = pNode->aOfs[2];
                        pNode->aOfs[2] = pNode->aOfs[1];
                        pNode->aOfs[1] = nRightChildOfs;
                        break;
                case 1: pNode->aKey[4] = pNode->aKey[3];
                        pNode->aKey[3] = pNode->aKey[2];
                        pNode->aKey[2] = pNode->aKey[1];
                        pNode->aKey[1] = nKey;
                        pNode->aOfs[5] = pNode->aOfs[4];
                        pNode->aOfs[4] = pNode->aOfs[3];
                        pNode->aOfs[3] = pNode->aOfs[2];
                        pNode->aOfs[2] = nRightChildOfs;
                        break;
                case 2: pNode->aKey[4] = pNode->aKey[3];
                        pNode->aKey[3] = pNode->aKey[2];
                        pNode->aKey[2] = nKey;
                        pNode->aOfs[5] = pNode->aOfs[4];
                        pNode->aOfs[4] = pNode->aOfs[3];
                        pNode->aOfs[3] = nRightChildOfs;
                        break;
                case 3: pNode->aKey[4] = pNode->aKey[3];
                        pNode->aKey[3] = nKey;
                        pNode->aOfs[5] = pNode->aOfs[4];
                        pNode->aOfs[4] = nRightChildOfs;
                        break;
                case 4: pNode->aKey[4] = nKey;
                        pNode->aOfs[5] = nRightChildOfs;
                        break;
            }
            return WFSKRN_RESULT_OK;
        } else {
            // Key insertion requires a node split
            // The node needs to be split before the new key can be inserted
            pNewNode = pPti->apNode[nLevel-1] = PTreeGetNode(pAllocator, *pNewNodeOfs);
            u32 nOldKey;
            switch(*pEntryIdx) {
                case 0: pNewNode->aKey[0] = pNode->aKey[3];
                        pNewNode->aKey[1] = pNode->aKey[4];
                        pNewNode->aKey[2] = pNewNode->aKey[3] = pNewNode->aKey[4] = 0;
                        pNewNode->aOfs[0] = pNode->aOfs[3];
                        pNewNode->aOfs[1] = pNode->aOfs[4];
                        pNewNode->aOfs[2] = pNode->aOfs[5];
                        nOldKey = nKey;
                        nKey = pNode->aKey[2];
                        pNode->aKey[2] = pNode->aKey[1];
                        pNode->aKey[1] = pNode->aKey[0];
                        pNode->aKey[0] = nOldKey;
                        pNode->aOfs[3] = pNode->aOfs[2];
                        pNode->aOfs[2] = pNode->aOfs[1];
                        pNode->aOfs[1] = nRightChildOfs;
                        break;
                case 1: pNewNode->aKey[0] = pNode->aKey[3];
                        pNewNode->aKey[1] = pNode->aKey[4];
                        pNewNode->aKey[2] = pNewNode->aKey[3] = pNewNode->aKey[4] = 0;
                        pNewNode->aOfs[0] = pNode->aOfs[3];
                        pNewNode->aOfs[1] = pNode->aOfs[4];
                        pNewNode->aOfs[2] = pNode->aOfs[5];
                        nOldKey = nKey;
                        nKey = pNode->aKey[2];
                        pNode->aKey[2] = pNode->aKey[1];
                        pNode->aKey[1] = nOldKey;
                        pNode->aOfs[3] = pNode->aOfs[2];
                        pNode->aOfs[2] = nRightChildOfs;
                        break;
                case 2: pNewNode->aKey[0] = pNode->aKey[3];
                        pNewNode->aKey[1] = pNode->aKey[4];
                        pNewNode->aKey[2] = pNewNode->aKey[3] = pNewNode->aKey[4] = 0;
                        pNewNode->aOfs[0] = pNode->aOfs[3];
                        pNewNode->aOfs[1] = pNode->aOfs[4];
                        pNewNode->aOfs[2] = pNode->aOfs[5];
                        nOldKey = nKey;
                        nKey = pNode->aKey[2];
                        pNode->aKey[2] = nOldKey;
                        pNode->aOfs[3] = nRightChildOfs;
                        break;
                case 3: pNewNode->aKey[0] = pNode->aKey[3];
                        pNewNode->aKey[1] = pNode->aKey[4];
                        pNewNode->aKey[2] = pNewNode->aKey[3] = pNewNode->aKey[4] = 0;
                        pNewNode->aOfs[0] = nRightChildOfs;
                        pNewNode->aOfs[1] = pNode->aOfs[4];
                        pNewNode->aOfs[2] = pNode->aOfs[5];
                        break;
                case 4: pNewNode->aKey[0] = nKey;
                        pNewNode->aKey[1] = pNode->aKey[4];
                        pNewNode->aKey[2] = pNewNode->aKey[3] = pNewNode->aKey[4] = 0;
                        pNewNode->aOfs[0] = pNode->aOfs[4];
                        pNewNode->aOfs[1] = nRightChildOfs;
                        pNewNode->aOfs[2] = pNode->aOfs[5];
                        nKey = pNode->aKey[3];
                        break;
                case 5: pNewNode->aKey[0] = pNode->aKey[4];
                        pNewNode->aKey[1] = nKey;
                        pNewNode->aKey[2] = pNewNode->aKey[3] = pNewNode->aKey[4] = 0;
                        pNewNode->aOfs[0] = pNode->aOfs[4];
                        pNewNode->aOfs[1] = pNode->aOfs[5];
                        pNewNode->aOfs[2] = nRightChildOfs;
                        nKey = pNode->aKey[3];
                        break;
            }
            pNode->aKey[3] = pNode->aKey[4] = 0;
            nRightChildOfs = *pNewNodeOfs++;
            nLeftChildOfs = PTreeGetOfs(pNode, pAllocator);
            ++nLevel;
        }
    }
//    return WFSKRN_RESULT_OK;
}


WFSKrnResult PTreeInsert(PTreeHdr *pPTreeHdr, PTreeItr *pPti, u32 nKey, u32 nData, PTreeAllocator *pAllocator) {
    u32 nOfs;
    PTreeLeaf *pLeaf;
    PTreeLeaf *pNewLeaf;
    u32 *pEntryIdx = pPti->aEntryIdx;

    // check if the P-tree is empty before PTreeFindLeafOfs(). 
    if (pPTreeHdr->nNumRecs == 0) {
        u16 nOfs = pAllocator->fpAllocNode(pAllocator->pBase);
        if (nOfs == 0) {
            return WFSKRN_RESULT_PTREE_FULL;
        }
        pPTreeHdr->nRootOfs = nOfs;
        pPTreeHdr->nRootHeight = 0;
        pLeaf = pPti->pLeaf = PTreeGetLeaf(pAllocator, (u32)pPTreeHdr->nRootOfs);
        *pEntryIdx = 0;
        memset(&pLeaf->aKey[1], 0, 3 * sizeof(u32));    // It is necessary to initialize to 0
        pLeaf->aKey[0] = nKey;
        pLeaf->aData[0] = nData;
        goto Done;
    }

    nOfs = PTreeFindLeafOfs(pPTreeHdr, pPti, nKey, pAllocator);
    pLeaf = pPti->pLeaf = PTreeGetLeaf(pAllocator, nOfs);

    // 32 cases to distinguish
    //  Insert(K) : Different cases for leaf update
    //  K |  Input    | Output              | K<=k[0] | K>=k[0] | K<=k[1] | K>=k[1] | K<=k[2] | K>=k[2] | K<=k[3] | K>=k[3] | 0==k[0] | 0==k[1] | 0==k[2] | 0==k[3] |
    //  --+-----------+---------------------+---------+---------+---------+---------+---------+---------+---------+---------+---------+---------+---------+---------+
    //  0 |  k 0 0 0  |  Error              |    1    |    1    |    1    |    1    |    1    |    1    |    1    |    1    |    1    |    1    |    1    |    1    |
    //  0 |  t 0 0 0  |  K t 0 0            |    1    |    0    |    1    |    1    |    1    |    1    |    1    |    1    |    0    |    1    |    1    |    1    |
    //  0 |  k t 0 0  |  Error              |    1    |    1    |    1    |    0    |    1    |    1    |    1    |    1    |    1    |    0    |    1    |    1    |
    //  0 |  r t 0 0  |  K r t 0            |    1    |    0    |    1    |    0    |    1    |    1    |    1    |    1    |    0    |    0    |    1    |    1    |
    //  0 |  k r t 0  |  Error              |    1    |    1    |    1    |    0    |    1    |    0    |    1    |    1    |    1    |    0    |    0    |    1    |
    //  0 |  p r t 0  |  K p r t            |    1    |    0    |    1    |    0    |    1    |    0    |    1    |    1    |    0    |    0    |    0    |    1    |
    //  0 |  k p r t  |  Error              |    1    |    1    |    1    |    0    |    1    |    0    |    1    |    0    |    1    |    0    |    0    |    0    |
    //  0 |  n p r t  |  K n 0 0   p r t 0  |    1    |    0    |    1    |    0    |    1    |    0    |    1    |    0    |    0    |    0    |    0    |    0    |
    //  K |  b 0 0 0  |  b K 0 0            |    0    |    1    |    0    |    1    |    0    |    1    |    0    |    1    |    0    |    1    |    1    |    1    |
    //  K |  k 0 0 0  |  Error              |    1    |    1    |    0    |    1    |    0    |    1    |    0    |    1    |    0    |    1    |    1    |    1    |
    //  K |  t 0 0 0  |  K t 0 0            |    1    |    0    |    0    |    1    |    0    |    1    |    0    |    1    |    0    |    1    |    1    |    1    |
    //  K |  b d 0 0  |  b d K 0            |    0    |    1    |    0    |    1    |    0    |    1    |    0    |    1    |    0    |    0    |    1    |    1    |
    //  K |  b k 0 0  |  Error              |    0    |    1    |    1    |    1    |    0    |    1    |    0    |    1    |    0    |    0    |    1    |    1    |
    //  K |  b t 0 0  |  b K t 0            |    0    |    1    |    1    |    0    |    0    |    1    |    0    |    1    |    0    |    0    |    1    |    1    |
    //  K |  k t 0 0  |  Error              |    1    |    1    |    1    |    0    |    0    |    1    |    0    |    1    |    0    |    0    |    1    |    1    |
    //  K |  r t 0 0  |  K r t 0            |    1    |    0    |    1    |    0    |    0    |    1    |    0    |    1    |    0    |    0    |    1    |    1    |
    //  K |  b d f 0  |  b d f K            |    0    |    1    |    0    |    1    |    0    |    1    |    0    |    1    |    0    |    0    |    0    |    1    |
    //  K |  b d k 0  |  Error              |    0    |    1    |    0    |    1    |    1    |    1    |    0    |    1    |    0    |    0    |    0    |    1    |
    //  K |  b d t 0  |  b d K t            |    0    |    1    |    0    |    1    |    1    |    0    |    0    |    1    |    0    |    0    |    0    |    1    |
    //  K |  b k t 0  |  Error              |    0    |    1    |    1    |    1    |    1    |    0    |    0    |    1    |    0    |    0    |    0    |    1    |
    //  K |  b r t 0  |  b K r t            |    0    |    1    |    1    |    0    |    1    |    0    |    0    |    1    |    0    |    0    |    0    |    1    |
    //  K |  k r t 0  |  Error              |    1    |    1    |    1    |    0    |    1    |    0    |    0    |    1    |    0    |    0    |    0    |    1    |
    //  K |  p r t 0  |  K p r t            |    1    |    0    |    1    |    0    |    1    |    0    |    0    |    1    |    0    |    0    |    0    |    1    |
    //  K |  b d f j  |  b d f 0   j K 0 0  |    0    |    1    |    0    |    1    |    0    |    1    |    0    |    1    |    0    |    0    |    0    |    0    |
    //  K |  b d f k  |  Error              |    0    |    1    |    0    |    1    |    0    |    1    |    1    |    1    |    0    |    0    |    0    |    0    |
    //  K |  b d f t  |  b d f 0   K t 0 0  |    0    |    1    |    0    |    1    |    0    |    1    |    1    |    0    |    0    |    0    |    0    |    0    |
    //  K |  b d k t  |  Error              |    0    |    1    |    0    |    1    |    1    |    1    |    1    |    0    |    0    |    0    |    0    |    0    |
    //  K |  b d r t  |  b d K 0   r t 0 0  |    0    |    1    |    0    |    1    |    1    |    0    |    1    |    0    |    0    |    0    |    0    |    0    |
    //  K |  b k r t  |  Error              |    0    |    1    |    1    |    1    |    1    |    0    |    1    |    0    |    0    |    0    |    0    |    0    |
    //  K |  b p r t  |  b K 0 0   p r t 0  |    0    |    1    |    1    |    0    |    1    |    0    |    1    |    0    |    0    |    0    |    0    |    0    |
    //  K |  k p r t  |  Error              |    1    |    1    |    1    |    0    |    1    |    0    |    1    |    0    |    0    |    0    |    0    |    0    |
    //  K |  n p r t  |  K n 0 0   p r t 0  |    1    |    0    |    1    |    0    |    1    |    0    |    1    |    0    |    0    |    0    |    0    |    0    |
    //  --+-----------+---------------------+---------+---------+---------+---------+---------+---------+---------+---------+---------+---------+---------+---------+
    //                                           16        24        20        17        20        18        16        23        4         5         12        21
    if (pLeaf->aKey[0] >= nKey) {
        // 16 possibilities remain
        *pEntryIdx = 0;
        if (pLeaf->aKey[0] == nKey) {
            // nKey matches aKey[0] - covers 8 cases
            // 0 |  k 0 0 0  |  Error              |    1    |    1    |    1    |    1    |    1    |    1    |    1    |    1    |    1    |    1    |    1    |    1    |
            // 0 |  k t 0 0  |  Error              |    1    |    1    |    1    |    0    |    1    |    1    |    1    |    1    |    1    |    0    |    1    |    1    |
            // 0 |  k r t 0  |  Error              |    1    |    1    |    1    |    0    |    1    |    0    |    1    |    1    |    1    |    0    |    0    |    1    |
            // 0 |  k p r t  |  Error              |    1    |    1    |    1    |    0    |    1    |    0    |    1    |    0    |    1    |    0    |    0    |    0    |
            // K |  k 0 0 0  |  Error              |    1    |    1    |    0    |    1    |    0    |    1    |    0    |    1    |    0    |    1    |    1    |    1    |
            // K |  k t 0 0  |  Error              |    1    |    1    |    1    |    0    |    0    |    1    |    0    |    1    |    0    |    0    |    1    |    1    |
            // K |  k r t 0  |  Error              |    1    |    1    |    1    |    0    |    1    |    0    |    0    |    1    |    0    |    0    |    0    |    1    |
            // K |  k p r t  |  Error              |    1    |    1    |    1    |    0    |    1    |    0    |    1    |    0    |    0    |    0    |    0    |    0    |
            return WFSKRN_RESULT_PTREE_ENTRY_FOUND;
        } else {
            if (pLeaf->aKey[3] == 0) {
                // insert nKey to position aKey[0] in leaf - covers 6 cases
                // 0 |  t 0 0 0  |  K t 0 0            |    1    |    0    |    1    |    1    |    1    |    1    |    1    |    1    |    0    |    1    |    1    |    1    |
                // 0 |  r t 0 0  |  K r t 0            |    1    |    0    |    1    |    0    |    1    |    1    |    1    |    1    |    0    |    0    |    1    |    1    |
                // K |  t 0 0 0  |  K t 0 0            |    1    |    0    |    0    |    1    |    0    |    1    |    0    |    1    |    0    |    1    |    1    |    1    |
                // K |  r t 0 0  |  K r t 0            |    1    |    0    |    1    |    0    |    0    |    1    |    0    |    1    |    0    |    0    |    1    |    1    |
                // 0 |  p r t 0  |  K p r t            |    1    |    0    |    1    |    0    |    1    |    0    |    1    |    1    |    0    |    0    |    0    |    1    |
                // K |  p r t 0  |  K p r t            |    1    |    0    |    1    |    0    |    1    |    0    |    0    |    1    |    0    |    0    |    0    |    1    |
                pLeaf->aKey[3] = pLeaf->aKey[2];
                pLeaf->aKey[2] = pLeaf->aKey[1];
                pLeaf->aKey[1] = pLeaf->aKey[0];
                pLeaf->aData[3] = pLeaf->aData[2];
                pLeaf->aData[2] = pLeaf->aData[1];
                pLeaf->aData[1] = pLeaf->aData[0];
            } else {
                // Split leaf 1:3 preparing to insert nKey at position aKey[0], and initizling unused aKey[] elements to 0  - covers 2 cases
                // 0 |  n p r t  |  K n 0 0   p r t 0  |    1    |    0    |    1    |    0    |    1    |    0    |    1    |    0    |    0    |    0    |    0    |    0    |
                // K |  n p r t  |  K n 0 0   p r t 0  |    1    |    0    |    1    |    0    |    1    |    0    |    1    |    0    |    0    |    0    |    0    |    0    |
                WFSKrnReturnOnError(PTreeInsertIntoNode(pPTreeHdr, pPti, pLeaf->aKey[1], &nOfs, pAllocator));
                pNewLeaf = PTreeGetLeaf(pAllocator, nOfs);
                pNewLeaf->aKey[0] = pLeaf->aKey[1];
                pNewLeaf->aKey[1] = pLeaf->aKey[2];
                pNewLeaf->aKey[2] = pLeaf->aKey[3];
                pNewLeaf->aData[0] = pLeaf->aData[1];
                pNewLeaf->aData[1] = pLeaf->aData[2];
                pNewLeaf->aData[2] = pLeaf->aData[3];
                pLeaf->aKey[1] = pLeaf->aKey[0];
                pLeaf->aData[1] = pLeaf->aData[0];
                pNewLeaf->aKey[3] = pLeaf->aKey[2] = pLeaf->aKey[3] = 0;
            }
            // inserting K to position aKey[0] - covers above 8 cases
            pLeaf->aKey[0] = nKey;
            pLeaf->aData[0] = nData;
        }
    } else {
        // 16 possibilities remain
        if (pLeaf->aKey[2] >= nKey) {
            // 8 possibilities remain
            if (pLeaf->aKey[1] < nKey) {
                // 4 possibilities remain
                *pEntryIdx = 2;
                if (pLeaf->aKey[2] == nKey) {
                    // nKey matches aKey[2] - covers 2 cases
                    // K |  b d k 0  |  Error              |    0    |    1    |    0    |    1    |    1    |    1    |    0    |    1    |    0    |    0    |    0    |    1    |
                    // K |  b d k t  |  Error              |    0    |    1    |    0    |    1    |    1    |    1    |    1    |    0    |    0    |    0    |    0    |    0    |
                    return WFSKRN_RESULT_PTREE_ENTRY_FOUND;
                } else {
                    if (pLeaf->aKey[3] == 0) {
                        // shift position 2 to position 3 - covers 1 case
                        // K |  b d t 0  |  b d K t            |    0    |    1    |    0    |    1    |    1    |    0    |    0    |    1    |    0    |    0    |    0    |    1    |
                        pLeaf->aKey[3] = pLeaf->aKey[2];
                        pLeaf->aData[3] = pLeaf->aData[2];
                    } else {
                        // split leaf into 2:2 and initialize unused elements to 0 - covers 1 case
                        // K |  b d r t  |  b d K 0   r t 0 0  |    0    |    1    |    0    |    1    |    1    |    0    |    1    |    0    |    0    |    0    |    0    |    0    |
                        WFSKrnReturnOnError(PTreeInsertIntoNode(pPTreeHdr, pPti, pLeaf->aKey[2], &nOfs, pAllocator));
                        pNewLeaf = PTreeGetLeaf(pAllocator, nOfs);
                        pNewLeaf->aKey[0] = pLeaf->aKey[2];
                        pNewLeaf->aKey[1] = pLeaf->aKey[3];
                        pNewLeaf->aData[0] = pLeaf->aData[2];
                        pNewLeaf->aData[1] = pLeaf->aData[3];
                        pNewLeaf->aKey[2] = pNewLeaf->aKey[3] = pLeaf->aKey[3] = 0;
                    }
                    // inserting K to position aKey[2] - covers above 2 cases
                    pLeaf->aKey[2] = nKey;
                    pLeaf->aData[2] = nData;
                }
            } else {
                // 4 possibilities remain
                *pEntryIdx = 1;
                if (pLeaf->aKey[1] == nKey) {
                    // nKey matches aKey[1] - covers 2 cases
                    // K |  b k t 0  |  Error              |    0    |    1    |    1    |    1    |    1    |    0    |    0    |    1    |    0    |    0    |    0    |    1    |
                    // K |  b k r t  |  Error              |    0    |    1    |    1    |    1    |    1    |    0    |    1    |    0    |    0    |    0    |    0    |    0    |
                    return WFSKRN_RESULT_PTREE_ENTRY_FOUND;
                } else {
                    // 2 possibilities remain
                    if (pLeaf->aKey[3] == 0) {
                        // shift positions 1 and 2 to to the right - covers 1 case
                        // K |  b r t 0  |  b K r t            |    0    |    1    |    1    |    0    |    1    |    0    |    0    |    1    |    0    |    0    |    0    |    1    |
                        pLeaf->aKey[3] = pLeaf->aKey[2];
                        pLeaf->aKey[2] = pLeaf->aKey[1];
                        pLeaf->aData[3] = pLeaf->aData[2];
                        pLeaf->aData[2] = pLeaf->aData[1];
                    } else {
                        // split leaf into 1:3 and initialize unused elements to 0 - covers 1 case
                        // K |  b p r t  |  b K 0 0   p r t 0  |    0    |    1    |    1    |    0    |    1    |    0    |    1    |    0    |    0    |    0    |    0    |    0    |
                        WFSKrnReturnOnError(PTreeInsertIntoNode(pPTreeHdr, pPti, pLeaf->aKey[1], &nOfs, pAllocator));
                        pNewLeaf = PTreeGetLeaf(pAllocator, nOfs);
                        pNewLeaf->aKey[0] = pLeaf->aKey[1];
                        pNewLeaf->aKey[1] = pLeaf->aKey[2];
                        pNewLeaf->aKey[2] = pLeaf->aKey[3];
                        pNewLeaf->aData[0] = pLeaf->aData[1];
                        pNewLeaf->aData[1] = pLeaf->aData[2];
                        pNewLeaf->aData[2] = pLeaf->aData[3];
                        pNewLeaf->aKey[3] = pLeaf->aKey[2] = pLeaf->aKey[3] = 0;
                    }
                    // inserting K to position aKey[1] - covers above 2 cases
                    pLeaf->aKey[1] = nKey;
                    pLeaf->aData[1] = nData;
                }
            }
        } else {
            // 8 possibilities remain
            if (pLeaf->aKey[2] == 0) {
                // 4 possibilities remain
                if (pLeaf->aKey[1] >= nKey) {
                    // 2 possibilities remain
                    *pEntryIdx = 1;
                    if (pLeaf->aKey[1] == nKey) {
                        // nKey matches aKey[1] in leaf - covers 1 case
                        // K |  b k 0 0  |  Error              |    0    |    1    |    1    |    1    |    0    |    1    |    0    |    1    |    0    |    0    |    1    |    1    |
                        return WFSKRN_RESULT_PTREE_ENTRY_FOUND;
                    } else {
                        // insert nKey to position aKey[1] - covers 1 case
                        // K |  b t 0 0  |  b K t 0            |    0    |    1    |    1    |    0    |    0    |    1    |    0    |    1    |    0    |    0    |    1    |    1    |
                        pLeaf->aKey[2] = pLeaf->aKey[1];
                        pLeaf->aData[2] = pLeaf->aData[1];
                        pLeaf->aKey[1] = nKey;
                        pLeaf->aData[1] = nData;
                    }
                } else {
                    // 2 possibilities remain
                    if (pLeaf->aKey[1] == 0) {
                        // append nKey to position aKey[1] - covers 1 case
                        // K |  b 0 0 0  |  b K 0 0            |    0    |    1    |    0    |    1    |    0    |    1    |    0    |    1    |    0    |    1    |    1    |    1    |
                        *pEntryIdx = 1;
                        pLeaf->aKey[1] = nKey;
                        pLeaf->aData[1] = nData;
                    } else {
                        // append nKey to position aKey[2] - covers 1 case
                        // K |  b d 0 0  |  b d K 0            |    0    |    1    |    0    |    1    |    0    |    1    |    0    |    1    |    0    |    0    |    1    |    1    |
                        *pEntryIdx = 2;
                        pLeaf->aKey[2] = nKey;
                        pLeaf->aData[2] = nData;
                    }
                }
            } else {
                // 4 possibilities remain
                if (pLeaf->aKey[3] >= nKey) {
                    // 2 possibilities remain
                    if (pLeaf->aKey[3] == nKey) {
                        // nKey matches aKey[3] in leaf - covers 1 case
                        // K |  b d f k  |  Error              |    0    |    1    |    0    |    1    |    0    |    1    |    1    |    1    |    0    |    0    |    0    |    0    |
                        *pEntryIdx = 3;
                        return WFSKRN_RESULT_PTREE_ENTRY_FOUND;
                    } else {
                        // split leaf 3:1 inserting K into position aKey[0] of the right hand side, and initializing unused keys to 0 - covers 1 case
                        // K |  b d f t  |  b d f 0   K t 0 0  |    0    |    1    |    0    |    1    |    0    |    1    |    1    |    0    |    0    |    0    |    0    |    0    |
                        WFSKrnReturnOnError(PTreeInsertIntoNode(pPTreeHdr, pPti, nKey, &nOfs, pAllocator));
                        pNewLeaf = PTreeGetLeaf(pAllocator, nOfs);
                        pNewLeaf->aKey[1] = pLeaf->aKey[3];
                        pNewLeaf->aData[1] = pLeaf->aData[3];
                        pNewLeaf->aKey[2] = pNewLeaf->aKey[3] = pLeaf->aKey[3] = 0;
                        *pEntryIdx = 0;
                        pNewLeaf->aKey[0] = nKey;
                        pNewLeaf->aData[0] = nData;
                    }
                } else {
                    // 2 possibilities remain
                    if (pLeaf->aKey[3] == 0) {
                        // append nKey to position aKey[3] - covers 1 case
                        // K |  b d f 0  |  b d f K            |    0    |    1    |    0    |    1    |    0    |    1    |    0    |    1    |    0    |    0    |    0    |    1    |
                        *pEntryIdx = 3;
                        pLeaf->aKey[3] = nKey;
                        pLeaf->aData[3] = nData;
                    } else {
                        // split leaf into 3:1 inserting K to position aKey[1] of the right hand side, and initializing unused keys to 0 - covers 1 case
                        // K |  b d f j  |  b d f 0   j K 0 0  |    0    |    1    |    0    |    1    |    0    |    1    |    0    |    1    |    0    |    0    |    0    |    0    |
                        WFSKrnReturnOnError(PTreeInsertIntoNode(pPTreeHdr, pPti, pLeaf->aKey[3], &nOfs, pAllocator));
                        pNewLeaf = PTreeGetLeaf(pAllocator, nOfs);
                        pNewLeaf->aKey[0] = pLeaf->aKey[3];
                        pNewLeaf->aData[0] = pLeaf->aData[3];
                        pNewLeaf->aKey[2] = pNewLeaf->aKey[3] = pLeaf->aKey[3] = 0;
                        *pEntryIdx = 1;
                        pNewLeaf->aKey[1] = nKey;
                        pNewLeaf->aData[1] = nData;
                    }
                }
            }
        }
    }
Done:
    ++pPTreeHdr->nNumRecs;
    return WFSKRN_RESULT_OK;
}


WFSKrnResult PTreeDelete(PTreeHdr *pPTreeHdr, PTreeItr *pPti, PTreeAllocator *pAllocator) {
#if _DEBUG_BREAK_POINT
    static int c=0;++c;if(c>=12805){
        int a;a=0;
    }
#endif
    --pPTreeHdr->nNumRecs;
    PTreeLeaf *pLeaf = pPti->pLeaf;
    u32 *pEntryIdx = pPti->aEntryIdx;
    switch(*pEntryIdx) {
        case 0: if (pLeaf->aKey[1]==0) {
                    break;
                }
                pLeaf->aKey[0] = pLeaf->aKey[1];
                pLeaf->aData[0] = pLeaf->aData[1];
        case 1: pLeaf->aKey[1] = pLeaf->aKey[2];
                pLeaf->aData[1] = pLeaf->aData[2];
        case 2: pLeaf->aKey[2] = pLeaf->aKey[3];
                pLeaf->aData[2] = pLeaf->aData[3];
        case 3: pLeaf->aKey[3] = 0;
                return WFSKRN_RESULT_OK;
    }
    // No records left in the leaf - delete the leaf
    u32 nOfs = PTreeGetOfs(pLeaf, pAllocator);
    pAllocator->fpFreeNode(pAllocator->pBase, nOfs);
    // Delete the entry from the parent node
    u32 nLevel = 1;
    while(nLevel <= pPTreeHdr->nRootHeight) {
        ++pEntryIdx;
        PTreeNode *pNode = pPti->apNode[nLevel-1];
        switch(*pEntryIdx) {
            case 0: if (pNode->aKey[0]==0) {
                        break;
                    }
                    pNode->aOfs[0] = pNode->aOfs[1];
            case 1: pNode->aKey[0] = pNode->aKey[1];
                    pNode->aOfs[1] = pNode->aOfs[2];
            case 2: pNode->aKey[1] = pNode->aKey[2];
                    pNode->aOfs[2] = pNode->aOfs[3];
            case 3: pNode->aKey[2] = pNode->aKey[3];
                    pNode->aOfs[3] = pNode->aOfs[4];
            case 4: pNode->aKey[3] = pNode->aKey[4];
                    pNode->aOfs[4] = pNode->aOfs[5];
            case 5: pNode->aKey[4] = 0;
                    return WFSKRN_RESULT_OK;
        }
        nOfs = PTreeGetOfs(pNode, pAllocator);
        pAllocator->fpFreeNode(pAllocator->pBase, nOfs);
        ++nLevel;
    }
    return WFSKRN_RESULT_OK;
}


WFSKrnResult PTreeFirst(PTreeHdr *pPTreeHdr, PTreeItr *pPti, PTreeAllocator *pAllocator) {
    if (pPTreeHdr->nNumRecs == 0) {
        return WFSKRN_RESULT_PTREE_ENTRY_NOT_FOUND;
    }
    u32 nLevel = pPTreeHdr->nRootHeight;
    u32 *pEntryIdx = &pPti->aEntryIdx[nLevel];
    u32 nOfs = (u32)pPTreeHdr->nRootOfs;
    memset(pPti->aEntryIdx, 0, sizeof(pPti->aEntryIdx));
    while (nLevel) {
        --nLevel;
        PTreeNode *pNode = pPti->apNode[nLevel] = PTreeGetNode(pAllocator, nOfs);
        nOfs = pNode->aOfs[0]; //*pEntryIdx
        --pEntryIdx;
    }
    PTreeLeaf *pLeaf = pPti->pLeaf = PTreeGetLeaf(pAllocator, nOfs);
    pPti->nKey = pLeaf->aKey[0];
    pPti->nData = pLeaf->aData[0];
    return WFSKRN_RESULT_OK;
}


WFSKrnResult PTreeNext(PTreeHdr *pPTreeHdr, PTreeItr *pPti, PTreeAllocator *pAllocator) {
    WFSKrnResult nResult = WFSKRN_RESULT_OK;
    u32 *pEntryIdx = pPti->aEntryIdx;
#if _DEBUG_PTREE
    if (pPti->pLeaf->aKey[*pEntryIdx] < nCurKey) {
        WFSKrnOutputErrorStr("Leaf key out of order wrt. parent");
    }
    nCurKey = pPti->pLeaf->aKey[*pEntryIdx];
#endif
    ++(*pEntryIdx);
    u32 *aKey = pPti->pLeaf->aKey;
    u32 *pKey = &aKey[*pEntryIdx];
    if ((*pEntryIdx==PTREE_LEAF_FAN_OUT) || (*pKey==0)) {
        u32 *pEntryIdxEnd = &pPti->aEntryIdx[pPTreeHdr->nRootHeight];
        PTreeNode **ppNode = pPti->apNode;
        while(1) {
            ++pEntryIdx;
            if (pEntryIdx > pEntryIdxEnd) {
                return WFSKRN_RESULT_PTREE_ENTRY_NOT_FOUND;
            }
            aKey = (*ppNode)->aKey;
            ++ppNode;
#if _DEBUG_PTREE
            ++nNodeCount;
#endif
            if (*pEntryIdx >= PTREE_NODE_FAN_OUT-1) {
                continue;
            }
            pKey = &aKey[*pEntryIdx];
#if _DEBUG_PTREE
            if (*pKey != 0) {
                //MyOSReport("0x%08x\n", *pKey);
                if (*pKey < nCurKey) {
                    WFSKrnOutputErrorStr("Parent key out of order wrt. previous child");
                }
            }
            nCurKey = *pKey;
#endif
            ++(*pEntryIdx);
            if (*pKey != 0) {
                break;
            }
        }
        --ppNode;
        u32 nOfs = (*ppNode)->aOfs[pKey - aKey + 1];
        while(1) {
            --pEntryIdx;
            *pEntryIdx = 0;
            if (ppNode == pPti->apNode) {
                break;
            }
            --ppNode;
            *ppNode = PTreeGetNode(pAllocator, nOfs);
            nOfs = (*ppNode)->aOfs[0];
        }
        pPti->pLeaf = PTreeGetLeaf(pAllocator, nOfs);
        aKey = pPti->pLeaf->aKey;
        pKey = &aKey[0];
    }
    pPti->nKey = *pKey;
    pPti->nData = pPti->pLeaf->aData[pKey - aKey];
    return nResult;
}


WFSKrnResult PTreeInsertOrUpdateRec(PTreeHdr *pPTreeHdr, u32 nKey, u32 nData, PTreeAllocator *pAllocator) {
    // This function either adds a new record, or if a record already exists with the specified key, updates it to the new data
    PTreeItr pti;
    WFSKrnResult nResult = PTreeInsert(pPTreeHdr, &pti, nKey, nData, pAllocator);
    if (nResult == WFSKRN_RESULT_PTREE_ENTRY_FOUND) {
        pti.pLeaf->aData[pti.aEntryIdx[0]] = nData;
    }
    return nResult;
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#if _DEBUG_PTREE

    void PTreeCheck(PTreeHdr *pPTreeHdr, PTreeAllocator *pAllocator, bool bReport) {
        nCurKey = 0;
        nNodeCount = 0;
        PTreeItr pti;
        u32 nNumRecs = 0;
        WFSKrnResult nResult = PTreeFirst(pPTreeHdr, &pti, pAllocator);
        if (nResult == WFSKRN_RESULT_OK) {
            ++nNodeCount;
            while(1) {
                ++nNumRecs;
                u32 nPrevKey = pti.nKey;
                if (bReport) {
                    MyOSReport("%4d. %04x 0x%08x  0x%08x\n", nNumRecs, PTreeGetOfs(pti.pLeaf, pAllocator), pti.nKey, pti.nData);
                }
                nResult = PTreeNext(pPTreeHdr, &pti, pAllocator);
                if (nResult != WFSKRN_RESULT_OK) {
                    break;
                }
                if (pti.nKey <= nPrevKey) {
                    WFSKrnOutputErrorStr("PTree: key sorting error");
                }
            }
        }
        if (bReport) {
            MyOSReport("\n");
        }
        if (nResult != WFSKRN_RESULT_PTREE_ENTRY_NOT_FOUND) {
            WFSKrnOutputErrorCode(nResult);
        }
        if (nNumRecs != pPTreeHdr->nNumRecs) {
            WFSKrnOutputErrorStr("Error! Not enough records\n");
        }
    }

    void PTreeChoosePath(PTreeHdr *pPTreeHdr, PTreeItr *pPti, u32 nPathDeterminer, PTreeAllocator *pAllocator) {
        u32 nLevel = pPTreeHdr->nRootHeight;
        u32 *pEntryIdx = &pPti->aEntryIdx[nLevel];
        u32 nOfs = (u32)pPTreeHdr->nRootOfs;
        while (nLevel) {
            PTreeNode *pNode = pPti->apNode[nLevel-1] = PTreeGetNode(pAllocator, nOfs);
            // Hard-coded binary split search which avoids need to store length of sorted array, by assuming a 0 terminated array
            if (pNode->aKey[3] == 0) {
                if (pNode->aKey[1] == 0) {
                    if (pNode->aKey[0] == 0) {
                        // There are 0 keys (1 child)
                        *pEntryIdx = 0;
                    } else {
                        // There is 1 key (2 children)
                        *pEntryIdx = nPathDeterminer&1; nPathDeterminer>>=1;
                    }
                } else {
                    if (pNode->aKey[2] == 0) {
                        // There are 2 keys (3 children)
                        *pEntryIdx = nPathDeterminer%3; nPathDeterminer/=3;
                    } else {
                        // There are 3 keys (4 children)
                        *pEntryIdx = nPathDeterminer&3; nPathDeterminer>>=2;
                    }
                }
            } else {
                if (pNode->aKey[4] == 0) {
                    // There are 4 keys (5 children)
                    *pEntryIdx = nPathDeterminer%5; nPathDeterminer/=5;
                } else {
                    // There are 5 keys (6 children)
                    *pEntryIdx = nPathDeterminer%6; nPathDeterminer/=6;
                }
            }
            nOfs = (u32)pNode->aOfs[*pEntryIdx];
            --pEntryIdx;
            --nLevel;
        }
        PTreeLeaf *pLeaf = pPti->pLeaf = PTreeGetLeaf(pAllocator, nOfs);
        if (pLeaf->aKey[2] == 0) {
            if (pLeaf->aKey[1] == 0) {
                // There is 1 key = (1 child)
                *pEntryIdx = 0;
            } else {
                // There are 2 keys (2 children)
                *pEntryIdx = nPathDeterminer&1;
            }
        } else {
            if (pLeaf->aKey[3] == 0) {
                // There are 3 keys (3 children)
                *pEntryIdx = nPathDeterminer%3;
            } else {
                // There are 4 keys (4 children)
                *pEntryIdx = nPathDeterminer&3;
            }
        }
        pPti->nKey = pLeaf->aKey[*pEntryIdx];
        pPti->nData = pLeaf->aData[*pEntryIdx];
    }

    #include <set>
    using namespace std;
    void PTreeTest() {
        set<u32> testSet;
        PTreeAllocator allocator;
        allocator.pBase = (void*)&wkg.pathCache.pRxtHdr->sbah;
        allocator.fpAllocNode  = (PTreeAllocNodeFunc) PTreeAllocNodeFromSba;
        allocator.fpAllocNodes = (PTreeAllocNodesFunc)PTreeAllocNodesFromSba;
        allocator.fpFreeNode   = (PTreeFreeNodeFunc)  PTreeFreeNodeToSba;
        PTreeHdr ptreeHdr;
        PTreeItr pti;
        PTreeInit(&ptreeHdr, &allocator);
        u32 aTarget[] = { 3, 0, 50, 0, 500, 0, 1000, 200, 800, 500, 1200, 0, 2000 };
        u32 nTargetIdx;
        u32 nIntialTotalSize = wkg.pathCache.pRxtHdr->sbah.nTotalFree;
        for(nTargetIdx=0; nTargetIdx<sizeof(aTarget)/sizeof(aTarget[0]); ++nTargetIdx) {
            u32 nTarget = aTarget[nTargetIdx];
            while(ptreeHdr.nNumRecs < nTarget) {
                u32 nKey  = RandomU32()&0xffff00;
                if ((nTargetIdx==0)&&(ptreeHdr.nNumRecs==1)) {
                    nKey=0;
                }
                u32 nData = (nKey>>16) | (nKey<<16);
                WFSKrnResult nResult = PTreeInsert(&ptreeHdr, &pti, nKey, nData, &allocator);
                switch(nResult) {
                    case WFSKRN_RESULT_OK:
                        break;
                    case WFSKRN_RESULT_PTREE_FULL:
                        goto Done;
                    case WFSKRN_RESULT_PTREE_ENTRY_FOUND:
                        if (testSet.find(nKey) == testSet.end()) {
                            WFSKrnOutputErrorStr("Error! PTreeInsert() detected erroneous duplicate");
                        }
                        continue;
                    default:
                        WFSKrnOutputErrorCode(nResult);
                }
                if (testSet.find(nKey) != testSet.end()) {
                    WFSKrnOutputErrorStr("Error! PTreeInsert() did not detect duplicate");
                }
                testSet.insert(nKey);
                PTreeCheck(&ptreeHdr, &allocator, false);
                u32 nTotalSize = wkg.pathCache.pRxtHdr->sbah.nTotalFree + (nNodeCount * (1<<PTREE_LOG2_NODE_SIZE));
                if (nTotalSize != nIntialTotalSize) {
                    MyOSReport("%5d = %5d + (%-3dx 32)\n", nTotalSize, wkg.pathCache.pRxtHdr->sbah.nTotalFree, nNodeCount);
                    WFSKrnOutputErrorStr("Memory leak detected in PTree\n");
                }
            }
            while(ptreeHdr.nNumRecs > nTarget) {
                u32 nRandomPathDeterminer = RandomU32();
                PTreeChoosePath(&ptreeHdr, &pti, nRandomPathDeterminer, &allocator);
                if (testSet.find(pti.nKey) == testSet.end()) {
                    WFSKrnOutputErrorStr("Error! PTreeChoosePath() picked a key which should not be in the tree");
                }
                testSet.erase(pti.nKey);
                PTreeDelete(&ptreeHdr, &pti, &allocator);
                PTreeCheck(&ptreeHdr, &allocator, false);
                u32 nTotalSize = wkg.pathCache.pRxtHdr->sbah.nTotalFree + (nNodeCount * (1<<PTREE_LOG2_NODE_SIZE));
                if (nTotalSize != nIntialTotalSize) {
                    MyOSReport("%5d = %5d + (%-3dx 32)\n", nTotalSize, wkg.pathCache.pRxtHdr->sbah.nTotalFree, nNodeCount);
                    WFSKrnOutputErrorStr("Memory leak detected in PTree\n");
                }
            }
            u32 nKey=0;
            for(; nKey<=0xFFFF00; nKey+=0x100) {
                WFSKrnResult nResult = PTreeFind(&ptreeHdr, &pti, nKey, &allocator);
                switch(nResult) {
                    case WFSKRN_RESULT_PTREE_ENTRY_NOT_FOUND:
                        if (testSet.find(nKey) != testSet.end()) {
                            WFSKrnOutputErrorStr("Error! PTreeFind() did not find an existing entry");
                        }
                        if (ptreeHdr.nNumRecs) {
                            if (pti.aEntryIdx[0]>PTREE_LEAF_FAN_OUT) {
                                WFSKrnOutputErrorStr("Error! entryIdx out of range\n"); break;
                            }
                            if (pti.aEntryIdx[0]!=0) {
                                if (pti.pLeaf->aKey[pti.aEntryIdx[0]-1] >= nKey) {
                                    WFSKrnOutputErrorStr("Error! entryIdx too high\n"); break;
                                }
                            }
                            if (pti.aEntryIdx[0]<PTREE_LEAF_FAN_OUT) {
                                u32 nHigh = pti.pLeaf->aKey[pti.aEntryIdx[0]];
                                if ((nHigh!=0)&&(nHigh<=nKey)) {
                                    WFSKrnOutputErrorStr("Error! entryIdx too low\n"); break;
                                }
                            }
                        }
                        break;
                    case WFSKRN_RESULT_PTREE_ENTRY_FOUND:
                        if (testSet.find(nKey) == testSet.end()) {
                            WFSKrnOutputErrorStr("Error! PTreeFind() found a non-existing entry");
                        }
                        break;
                    default:
                        WFSKrnOutputErrorCode(nResult);
                }
            }
        }
    Done:
        PTreeCheck(&ptreeHdr, &allocator, true);
        u32 nTotalSize = wkg.pathCache.pRxtHdr->sbah.nTotalFree + (nNodeCount * (1<<PTREE_LOG2_NODE_SIZE));
        if (nTotalSize != nIntialTotalSize) {
            MyOSReport("%5d = %5d + (%-3dx 32)\n", nTotalSize, wkg.pathCache.pRxtHdr->sbah.nTotalFree, nNodeCount);
            WFSKrnOutputErrorStr("Memory leak detected in PTree\n");
        }
    }

#endif //_DEBUG_PTREE
