/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfskrn_PathCache.cpp - Caches recently used path names plus root path names "/dev", "/vol"


  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_PathCache.cpp,v $
  Revision 1.23  2008/12/04 00:48:21  ooizumi
  Fixed variable name witch is defined twice.
  Changed definition name to enable permission.

  Revision 1.22  2008/11/05 15:08:20  ueno
  Modified DirItrMapBlks() to roll back the update map in case of error.

  Revision 1.21  2008/10/31 09:51:11  ueno
  Modified WFSSrvAPIs to return WFSKRN_RESULT_RESOURCE_LIMIT_EXCEEDED when path cache is short of memory.

  Revision 1.20  2008/10/14 04:13:07  ooizumi
  Modified PathCacheFindParentDir() to return WFSKRN_RESULT_NOT_DIRECTORY
  when device path is specified.

  Revision 1.19  2008/10/10 01:21:06  ooizumi
  Fixed to call PathCacheInitSlash() in PathCacheFindParentDir().

  Revision 1.18  2008/10/09 14:11:33  ueno
  Modified PathCacheIncreaseEntryOpenHandleCount() to return WFSKRN_RESULT_FILE_OPEN when writing a file which is already open.

  Revision 1.17  2008/10/09 04:15:11  ooizumi
  Fixed build error.

  Revision 1.16  2008/10/09 03:24:32  ooizumi
  Fixed to update volume root's nAccessListIdx in PathCacheData.

  Revision 1.15  2008/10/08 04:57:25  ooizumi
  Fixed a bug PathCacheFindPath doesn't return devAttr when device name is specified.

  Revision 1.14  2008/10/07 10:31:09  ooizumi
  Modified to update nObjCreator in PathCache.

  Revision 1.13  2008/08/28 06:04:38  ooizumi
  Modified to set 0 to pAreaInfo when the path had depth 1 is passed to PathCacheFindParentDir.
  Fixed an error message.

  Revision 1.12  2008/08/05 03:56:29  kondo_masahiro
  Added error handling for hash error and detach device.

  Revision 1.11  2008/07/30 21:10:15  paul
  Fixed a bug in PathCacheInsertCommon().

  Revision 1.10  2008/07/30 06:11:57  ooizumi
  Implemented initial version of permission functions.

  Revision 1.9  2008/07/28 22:21:56  paul
  Fixed a bug in PathCacheDeleteSubTree().

  Revision 1.8  2008/07/21 22:32:02  paul
  Added more consistency checking code for path cache entries.
  Fixed bugs related to the distinction between the RadixTree and the DirectoryTree - special handling of the '/' character.

  Revision 1.7  2008/07/17 22:56:54  paul
  Fixed some bugs. Changed PathCacheDecreaseEntryOpenHandleCount() to move from pinned list to MRU list.

  Revision 1.6  2008/07/17 05:33:38  kondo_masahiro
  Fixed codes for porting IOP.

  Revision 1.5  2008/07/09 00:49:01  paul
  Changes to make directory code work with block cache. Also a fix for device attach.

  Revision 1.4  2008/06/09 17:56:14  paul
  Rewrote PathCache to use a radix tree. Now actually provides PathCache functionality.
  Also stores a list of open file / directory search paths so that invalid WFSSrv calls can be detected.

  Revision 1.3  2008/05/12 22:35:32  kondo_masahiro
  Fixed code to accept RM compiler

  Revision 1.2  2008/05/12 19:16:41  paul
  Made naming more consistent

  Revision 1.1  2008/05/10 03:57:13  kondo_masahiro
  Initial check-in

*---------------------------------------------------------------------------*/

#undef  _KRN_FILE_
#define _KRN_FILE_  "PathCache"

#include "wfskrn_Config.h"
#include "wfskrn_PathCache.h"
#include "wfskrn_Defs.h"
#include "wfskrn_Api.h"
#include "wfskrn_DirRxTree.h"
#include "wfskrn_DirNodeStack.h"
#include "wfskrn_DirFind.h"
#include "wfskrn_Errors.h"


// This module is used for 4 distinct purposes:
// 1. To store the contents of the "/", "/dev", and "/vol", directories which only exist in memory, not on physical storage.
//    It assists in allowing these virtual directories to be searched like real directories.
// 2. To keep track of which Files and Directories have associated Open File Handles or Search Directory Handles.
//    (For example this is used to prevent renaming a directory which has one or more open files below it)
// 3. To speed up access to recently used directories and reduce the size of the block cache working set, by
//    avoiding the need to re-navigate directories along the path every time an operation is performed.
// 4. To store results of intermediate calculations in order to speed up the parsing of permissions chains
//    in the WFS Permissions API.
// ToDo: Purpose 4 is not yet implemented

#if _WIN32
    static FILE *fpOut;
#endif

void PathCacheDebugOutputList(PathCacheEntry *pAnchor, char cChar) {
    PathCacheEntry *pEntry = pAnchor->ml.pNext;
    while(pEntry != pAnchor) {
        WFSPathName path;
        RxtNodeHdr *pNodeHdr = (RxtNodeHdr *)((u8*)wkg.pathCache.pRxtHdr + pEntry->ah.nParentOfs);
        RxtGetString(wkg.pathCache.pRxtHdr, pNodeHdr, path.sStr, pEntry->pcd.nStrLen);
        MyOSReport("%p %c %2d %2d %s\n", pEntry->pcd.pData, cChar, pEntry->pcd.nPinCount, pEntry->pcd.nOpenHandleCount, path.sStr);
#if _WIN32
        fprintf(fpOut, "%p %c %2d %2d %s\n", pEntry->pcd.pData, cChar, pEntry->pcd.nPinCount, pEntry->pcd.nOpenHandleCount, path.sStr);
#endif
        pEntry = pEntry->ml.pNext;
    }
    MyOSReport("\n");
}

void PathCacheDebugOutput() {
#if _WIN32
    char sFileName[40];
    sprintf(sFileName, "PathCache.txt");
    //sprintf(sFileName, "PathCache_%d.txt", nDbgCommandCount);
    fpOut = fopen(sFileName, "w");
#endif
    PathCacheDebugOutputList(wkg.pathCache.pPinAnchor, 'P');
    PathCacheDebugOutputList(wkg.pathCache.pMruAnchor, 'C');
#if _WIN32
    fclose(fpOut);
#endif
}


#if _CHECK_PATH_CACHE
void PathCacheEntryCheck(u32 nRecIdx, u32 nNodeDepth, utf8 *sPath, void *pEntryPtr) {
    // This is a callback function called by RxtCheck() to check the contents of a radix tree entry (in this case a path cache entry)
    // Checks:
    //  * That OpenHandleCount matches the sum of immediate children's OpenHandleCount and the sum of all descendants' PinCount
    static utf8 sLastPath[WFS_MAX_PATH_NAME_SIZE];
    static u32 aPinCount[WFS_MAX_PATH_DEPTH];
    static u32 aSumOpenHandleCount[WFS_MAX_PATH_DEPTH];
    static u32 nLastPathDepth;
    utf8 *pSrcStr = sPath;
    utf8 *pDstStr = sLastPath;
    u32 nPathDepth = 0;
    PathCacheEntry *pEntry = (PathCacheEntry *)pEntryPtr;
    if (pEntry->pcd.nOpenHandleCount) {
        while(*pSrcStr) {
            if (*pDstStr != *pSrcStr) {
                break;
            }
            if (*pSrcStr == '/') {
                ++nPathDepth;
            }
            ++pDstStr;
            ++pSrcStr;
        }
        u32 nDiffDepth = nPathDepth;
        while(*pSrcStr) {
            if (*pSrcStr == '/') {
                if (nPathDepth > 0) {
                    aSumOpenHandleCount[nPathDepth] = aSumOpenHandleCount[nPathDepth-1];
                }
                ++nPathDepth;
            }
            *pDstStr++ = *pSrcStr++;
        }
        *pDstStr = 0;
        if (nDiffDepth < nPathDepth) {
            aSumOpenHandleCount[nPathDepth] = 0;
            nLastPathDepth = nPathDepth;
        } else {
            while(nLastPathDepth > nPathDepth) {
                aPinCount[nLastPathDepth] = 0;
                aSumOpenHandleCount[nLastPathDepth-1] = aSumOpenHandleCount[nLastPathDepth];
                --nLastPathDepth;
            }
        }
        if (nRecIdx==0) {
            aSumOpenHandleCount[nPathDepth] = 0;
        }
        aSumOpenHandleCount[nPathDepth] += pEntry->pcd.nOpenHandleCount;
        aPinCount[nPathDepth] = pEntry->pcd.nPinCount;
#if _DEBUG
        if (nPathDepth >= 2) {
            if (aSumOpenHandleCount[nPathDepth] > aSumOpenHandleCount[nPathDepth-1]) {
                PathCacheDebugOutput();
                WFSKrnOutputErrorStr("nOpenHandleCount inconsistent!");
            }
        }
        if (nPathDepth >= 4) {
            if ((aSumOpenHandleCount[nPathDepth] + aPinCount[nPathDepth-1]) > aSumOpenHandleCount[nPathDepth-1]) {
                PathCacheDebugOutput();
                WFSKrnOutputErrorStr("nOpenHandleCount + nPinCount inconsistent!");
            }
        }
#endif
    }
}

void PathCacheCheck() {
    RxtCheck(wkg.pathCache.pRxtHdr, 0, PathCacheEntryCheck); //WFS_LOG2_PATH_CACHE_SIZE, PathCacheEntryCheck);
#if _DEBUG_PATH_CACHE
    if ((nDirTest>=nDirTestBreakPoint-1)&&(nDirTest<=nDirTestBreakPoint)) {
        MyOSReport("\n");
    }
#endif //_DEBUG_PATH_CACHE
}
#endif //_CHECK_PATH_CACHE


void PathCacheInsertToEntryListAfterCursorEntry(PathCacheEntry *pCursorEntry, PathCacheEntry *pNewEntry) {
    pNewEntry->ml.pNext = pCursorEntry->ml.pNext;
    pCursorEntry->ml.pNext->ml.pPrev = pNewEntry;
    pNewEntry->ml.pPrev = pCursorEntry;
    pCursorEntry->ml.pNext = pNewEntry;
}


inline void PathCacheDeleteLink(PathCacheEntry *pEntry) {
    pEntry->ml.pPrev->ml.pNext   = pEntry->ml.pNext;
    pEntry->ml.pNext->ml.pPrev   = pEntry->ml.pPrev;
}


void PathCacheDeleteEntry(PathCacheEntry *pEntry) {
    PathCacheDeleteLink(pEntry);
    RxtItr rxi;
    rxi.nEntryIdx = pEntry->ah.nEntryIdx+1;
    rxi.pNodeHdr  = (RxtNodeHdr *)((u8*)wkg.pathCache.pRxtHdr + pEntry->ah.nParentOfs);
    SbaFreeSubBlk((WFSSubBlkAllocHdr *)wkg.pathCache.pRxtHdr, (u32)((u8*)pEntry - (u8*)wkg.pathCache.pRxtHdr), PATH_CACHE_LOG2_NODE_SIZE);
    RxtDelete(wkg.pathCache.pRxtHdr, &rxi);
}


WFSKrnResult PathCacheRecursiveDelete(RxtItr *pRxi) {
    // Deletes a sub-tree of entries from the path cach.
    RxtNodeHdr *pNodeHdr = pRxi->pNodeHdr;
    u32 nNodeOfs = (u8*)pNodeHdr - (u8*)wkg.pathCache.pRxtHdr;
    u32 nParentOfs = 0;
    utf8 *pChoiceChar;
    while(1) {
        u16 *pSrcNodeOfsTblEnd = (u16*)((u8*)pNodeHdr + (1<<pNodeHdr->nLog2Size));
        pRxi->nEntryIdx = 1;
        pChoiceChar = &pNodeHdr->aFirstChar[pNodeHdr->nStrLen];
        if (*pChoiceChar == 0) {
            // String terminator char indicates this is path cache entry
            u32 nAttrOfs = pSrcNodeOfsTblEnd[-1];
            PathCacheEntry *pEntry = (PathCacheEntry *)((u8*)wkg.pathCache.pRxtHdr + nAttrOfs);
            PathCacheDeleteLink(pEntry);
            SbaFreeSubBlk((WFSSubBlkAllocHdr *)wkg.pathCache.pRxtHdr, (u32)((u8*)pEntry - (u8*)wkg.pathCache.pRxtHdr), PATH_CACHE_LOG2_NODE_SIZE);
            --wkg.pathCache.pRxtHdr->nNumRecs;
            ++pRxi->nEntryIdx;
        }
        if (pRxi->nEntryIdx > pNodeHdr->nNumEntries) {
            do {
                pRxi->nEntryIdx = pNodeHdr->nEntryIdx + 2;
                nParentOfs = pNodeHdr->nParentOfs;
                SbaFreeSubBlk((WFSSubBlkAllocHdr *)wkg.pathCache.pRxtHdr, nNodeOfs, pNodeHdr->nLog2Size);
                nNodeOfs = nParentOfs;
                pNodeHdr = (RxtNodeHdr *)((u8*)wkg.pathCache.pRxtHdr + nNodeOfs);
                if (pRxi->nDepth == 0) {
                    // We reached the top of the stack, so this should be the last record
                    --pRxi->nEntryIdx;
                    return RxtDeleteNode(wkg.pathCache.pRxtHdr, pRxi, pRxi->nEntryIdx, nNodeOfs, pNodeHdr);
                }
                --pRxi->nDepth;
            } while(pRxi->nEntryIdx > pNodeHdr->nNumEntries);
            pChoiceChar = &pNodeHdr->aFirstChar[pNodeHdr->nStrLen + pRxi->nEntryIdx-1];
            if ((unsigned)pChoiceChar[-1] >= (unsigned)*pChoiceChar) {
                WFSKrnOutputErrorStr("PathCache: Choice out of order!");
            }
        }
        ++pRxi->nDepth;
        nParentOfs = nNodeOfs;
        nNodeOfs = RxtGetChildOfs(pRxi->nEntryIdx, pNodeHdr);
        pNodeHdr = (RxtNodeHdr *)((u8*)wkg.pathCache.pRxtHdr + nNodeOfs);
    }
//    return WFSKRN_RESULT_OK;
}


WFSKrnResult PathCacheDeleteSubTree(PathCacheEntry *pEntry) {
    RxtItr rxi;
    rxi.nEntryIdx = pEntry->ah.nEntryIdx+1;
    u32 nParentOfs = pEntry->ah.nParentOfs;
    rxi.pNodeHdr = (RxtNodeHdr *)((u8*)wkg.pathCache.pRxtHdr + nParentOfs);
    utf8 *pChoiceChar = &rxi.pNodeHdr->aFirstChar[rxi.pNodeHdr->nStrLen];
    rxi.nEntryIdx = 1;
    while(rxi.nEntryIdx <= rxi.pNodeHdr->nNumEntries) {
        switch(pChoiceChar[rxi.nEntryIdx-1]) {
            case 0: {
                u32 nAttrOfs = RxtGetChildOfs(rxi.nEntryIdx, rxi.pNodeHdr);
                PathCacheEntry *pTargetEntry = (PathCacheEntry *)((u8*)wkg.pathCache.pRxtHdr + nAttrOfs);
                PathCacheDeleteLink(pTargetEntry);
                --wkg.pathCache.pRxtHdr->nNumRecs;
                SbaFreeSubBlk((WFSSubBlkAllocHdr *)wkg.pathCache.pRxtHdr, (u32)((u8*)pEntry - (u8*)wkg.pathCache.pRxtHdr), PATH_CACHE_LOG2_NODE_SIZE);
                u32 nNumEntries = rxi.pNodeHdr->nNumEntries;
                if (nNumEntries > 1) {
                    utf8 cNextChoice = pChoiceChar[1];
                    WFSKrnResult nResult = RxtDeleteNode(wkg.pathCache.pRxtHdr, &rxi, rxi.nEntryIdx, nParentOfs, rxi.pNodeHdr);
                    if (nNumEntries == 2) {
                        if (cNextChoice == '/') {
                            rxi.nDepth = 0;
                            return PathCacheRecursiveDelete(&rxi);
                        }
                        return nResult;
                    }
                    pChoiceChar = &rxi.pNodeHdr->aFirstChar[rxi.pNodeHdr->nStrLen];
                    rxi.nEntryIdx = 0;
                } else {
                    rxi.nEntryIdx = rxi.pNodeHdr->nEntryIdx + 1;
                    u32 nNodeOfs = rxi.pNodeHdr->nParentOfs;
                    SbaFreeSubBlk((WFSSubBlkAllocHdr *)wkg.pathCache.pRxtHdr, nParentOfs, rxi.pNodeHdr->nLog2Size);
                    rxi.pNodeHdr = (RxtNodeHdr *)((u8*)wkg.pathCache.pRxtHdr + nNodeOfs);
                    return RxtDeleteNode(wkg.pathCache.pRxtHdr, &rxi, rxi.nEntryIdx, nNodeOfs, rxi.pNodeHdr);
                }
                break;
            }
            case'/':
                rxi.pNodeHdr = (RxtNodeHdr *)((u8*)wkg.pathCache.pRxtHdr + RxtGetChildOfs(rxi.nEntryIdx, rxi.pNodeHdr));
                rxi.nDepth = 0;
                return PathCacheRecursiveDelete(&rxi);
        }
        ++rxi.nEntryIdx;
    }
    return WFSKRN_RESULT_OK;
}


WFSKrnResult PathCacheDeleteLru() {
    PathCacheEntry *pEntry = wkg.pathCache.mruAnchor.pPrev;
    if (pEntry == wkg.pathCache.pMruAnchor) {
#if _DEBUG
        WFSKrnOutputErrorStr("Path Cache MRU list already empty\n");
#endif
        return WFSKRN_RESULT_RESOURCE_LIMIT_EXCEEDED; // MRU list is empty.
    }
#if _DEBUG_PATH_CACHE
    if (pEntry->pcd.nPinCount) {
        WFSKrnOutputErrorStr("PinCount != 0\n");
    }
#endif
    PathCacheDeleteEntry(pEntry);
    return WFSKRN_RESULT_OK;
}

inline WFSKrnResult PathCacheDeleteLruAndFixItr(RxtItr *pRxi) {
    // After deleting the LRU element, fix up pRxi which could otherwise be invalidated by the delete
    WFSKrnResult nResult = PathCacheDeleteLru();
    if (nResult == WFSKRN_RESULT_OK)
    {
        if (pRxi->pNodeHdr->nParentOfs) {
            RxtNodeHdr *pNodeHdr = (RxtNodeHdr *)((u8*)wkg.pathCache.pRxtHdr + pRxi->pNodeHdr->nParentOfs);
            pRxi->pOfs = RxtGetChildOfsPtr(pRxi->pNodeHdr->nEntryIdx+1, pNodeHdr);
        } else {
            pRxi->pOfs = &wkg.pathCache.pRxtHdr->nRootOfs;
        }
    }
    return nResult;
}


PathCacheEntry *PathCacheFind3(RxtItr *pRxi, u32 nStrLen, void **ppData) {
    WFSKrnResult nResult = RxtFind(wkg.pathCache.pRxtHdr, pRxi, nStrLen);
    if (nResult == WFSKRN_RESULT_DIR_ENTRY_FOUND) {
        PathCacheEntry *pEntry = (PathCacheEntry *)((u8*)wkg.pathCache.pRxtHdr + RxtGetChildOfs(1, pRxi->pNodeHdr));
        *ppData = pEntry->pcd.pData;
        return pEntry;
    }
    return 0;
}


void PathCacheInitSlash(PathCacheItr *pPci) {
    // This function fixes up the Pci to skip the root '/' so that PathCacheIncrementalFindEntry can be used
    pPci->rxi.nDepth = 0;
    pPci->rxi.pOfs = &wkg.pathCache.pRxtHdr->nRootOfs;
    pPci->rxi.pNodeHdr = (RxtNodeHdr *)((u8*)wkg.pathCache.pRxtHdr + *pPci->rxi.pOfs);
    pPci->rxi.pNodeStrPtr = pPci->rxi.pNodeHdr->aFirstChar;
}


WFSKrnResult PathCacheIncrementalFind(RxtItr *pRxi, u32 nStrLen) {
    RxtHdr *pRxtHdr = wkg.pathCache.pRxtHdr;
    WFSKrnResult nResult;
    utf8 *pStrEnd = pRxi->pStrPtr + nStrLen;
    RxtNodeHdr *pNodeHdr = pRxi->pNodeHdr;
    utf8 cSearchChar = '/';
    if (*pRxi->pStrPtr != cSearchChar) {
        --pRxi->pStrPtr;
    }
    goto EntryPoint;
    while(1) {
        pNodeHdr = (RxtNodeHdr *)((u8*)pRxtHdr + *pRxi->pOfs);
        u32 nNodeStrLen = pNodeHdr->nStrLen;
        pRxi->pNodeStrPtr = pNodeHdr->aFirstChar;
        u32 nSearchStrLen = pStrEnd - pRxi->pStrPtr;
        if (nSearchStrLen < nNodeStrLen) {
            while((nSearchStrLen) && (*pRxi->pStrPtr == *pRxi->pNodeStrPtr)) {
                ++pRxi->pStrPtr; ++pRxi->pNodeStrPtr; --nSearchStrLen;
            }
            nNodeStrLen -= (pRxi->pNodeStrPtr - pNodeHdr->aFirstChar);
            pRxi->nEntryIdx = -(s32)nNodeStrLen;
            nResult = WFSKRN_RESULT_DIR_NODE_STRING_MISMATCH;
            if (nSearchStrLen == 0) {
                nResult = WFSKRN_RESULT_DIR_NODE_STRING_PREFIX;
            }
            break;
        }
        while((nNodeStrLen) && (*pRxi->pStrPtr == *pRxi->pNodeStrPtr)) {
            ++pRxi->pStrPtr; ++pRxi->pNodeStrPtr; --nNodeStrLen;
        }
        if (nNodeStrLen) {
            pRxi->nEntryIdx = pRxi->pNodeStrPtr - pNodeHdr->aFirstChar;
            nResult = WFSKRN_RESULT_DIR_NODE_STRING_MISMATCH;
            break;
        }
        // matched all the characters from the prefix. pRxi->pNodeStrPtr should now be pointing to the array of choices following the prefix.
        cSearchChar = *pRxi->pStrPtr;
        if (pRxi->pStrPtr == pStrEnd){
            cSearchChar = 0;
        }
EntryPoint:
        if (!DirBinarySplitSearchChoices(pRxi->pNodeStrPtr, pNodeHdr->nNumEntries, cSearchChar, (u32*)&pRxi->nEntryIdx)) {
            nResult = WFSKRN_RESULT_DIR_NODE_CHOICE_NOT_FOUND;
            break;
        }
        // Found the search character.
        if (cSearchChar==0) {
            // We matched a 0 .. which means we have found the search string
            nResult = WFSKRN_RESULT_DIR_ENTRY_FOUND;
            break;
        }
        ++pRxi->nDepth;
        pRxi->pOfs = RxtGetChildOfsPtr((s32)pRxi->nEntryIdx, pNodeHdr);
        ++pRxi->pStrPtr;
    }
    pRxi->pNodeHdr = pNodeHdr;
    return nResult;
}


inline PathCacheEntry *PathCacheIncrementalFindEntry(RxtItr *pRxi, u32 nStrLen, void **ppData) {
    WFSKrnResult nResult = PathCacheIncrementalFind(pRxi, nStrLen);
    if (nResult == WFSKRN_RESULT_DIR_ENTRY_FOUND) {
        PathCacheEntry *pEntry = (PathCacheEntry *)((u8*)wkg.pathCache.pRxtHdr + RxtGetChildOfs(1, pRxi->pNodeHdr));
        *ppData = pEntry->pcd.pData;
        return pEntry;
    }
    return 0;
}


PathCacheEntry *PathCacheInsertCommon(WFSKrnResult nResult, PathCacheEntry *pCursorEntry, RxtItr *pRxi, utf8 *sStr, u32 nStrLen) {
#if _CHECK_PATH_CACHE
    PathCacheCheck();
#endif
    PathCacheEntry *pEntry;
    if (nResult == WFSKRN_RESULT_DIR_ENTRY_FOUND) {
        // The string already exists in the cache
        pEntry = (PathCacheEntry *)((u8*)wkg.pathCache.pRxtHdr + RxtGetChildOfs(1, pRxi->pNodeHdr));
        // Delete entry from current position on recency list and move it to the MRU position
        if (pEntry->pcd.nPinCount == 0) {
            pEntry->ml.pPrev->ml.pNext = pEntry->ml.pNext;
            pEntry->ml.pNext->ml.pPrev = pEntry->ml.pPrev;
            goto InsertToEntryList;
        }
        goto IncCount;
    }
    u32 nAttrOfs;
    while(1) {
        nAttrOfs = SbaAllocSubBlk((WFSSubBlkAllocHdr *)wkg.pathCache.pRxtHdr, PATH_CACHE_LOG2_NODE_SIZE);
        if (nAttrOfs) {
            pEntry = (PathCacheEntry *)((u8*)wkg.pathCache.pRxtHdr + nAttrOfs);
            pEntry->ah.nLog2Size = PATH_CACHE_LOG2_NODE_SIZE;
            pEntry->ah.nEntryIdx = 0;
            u32 nSuffixLen  = nStrLen - (pRxi->pStrPtr - sStr);
            nResult = RxtInsert(wkg.pathCache.pRxtHdr, pRxi, nSuffixLen, (RxtAttrHdr *)pEntry);
            if (nResult == WFSKRN_RESULT_OK) {
                break;
            }
            if (nResult != WFSKRN_RESULT_DIR_BLK_FULL) {
                WFSKrnOutputErrorCode(nResult);
                return 0;
            }
            SbaFreeSubBlk((WFSSubBlkAllocHdr *)wkg.pathCache.pRxtHdr, nAttrOfs, PATH_CACHE_LOG2_NODE_SIZE);
        }
        nResult = PathCacheDeleteLruAndFixItr(pRxi);
        if (nResult != WFSKRN_RESULT_OK)
        {
#if _DEBUG
            WFSKrnOutputErrorStr("PathCacheInsertCommon() : WFSKRN_RESULT_RESOURCE_LIMIT_EXCEEDED\n");
#endif
            return NULL; // [check]
        }
        pRxi->pStrPtr = sStr;
        RxtFind(wkg.pathCache.pRxtHdr, pRxi, nStrLen);
    }
    memset(&pEntry->pcd, 0, sizeof(PathCacheData));
    pEntry->pcd.nStrLen = nStrLen;
    // Calculate sum of OpenHandleCount from any pre-existing children
    u32 nSumOfOpenHandleCounts = 0;
    RxtNodeHdr *pBaseNodeHdr = pRxi->pNodeHdr;
    u32 nNumEntries = pBaseNodeHdr->nNumEntries;
    if (nNumEntries > 1) {
        utf8 *pChoiceChar;
        // Calculate the sum of open handle counts for any direct child paths
        u32 nEntryIdx = 1;
        while(nEntryIdx<=nNumEntries) {
            if (pBaseNodeHdr->aFirstChar[pBaseNodeHdr->nStrLen+nEntryIdx-1] == '/') {
                u32 nNodeOfs = RxtGetChildOfs(nEntryIdx, pBaseNodeHdr);
                RxtNodeHdr *pNodeHdr = (RxtNodeHdr *)((u8*)wkg.pathCache.pRxtHdr + nNodeOfs);
                u32 nRootDepth = pRxi->nDepth;
                u32 nParentOfs = (u8*)pBaseNodeHdr - (u8*)wkg.pathCache.pRxtHdr;
                while(1) {
                    u16 *pSrcNodeOfsTblEnd = (u16*)((u8*)pNodeHdr + (1<<pNodeHdr->nLog2Size));
                    utf8 *pStr = pNodeHdr->aFirstChar;
                    pChoiceChar = pStr + pNodeHdr->nStrLen;
                    pRxi->nEntryIdx = 1;
                    if (*pChoiceChar == 0) {
                        // String terminator char indicates this is a record, so increment the record count
                        PathCacheEntry *pChildEntry = (PathCacheEntry *)((u8*)wkg.pathCache.pRxtHdr + RxtGetChildOfs_ote(1, pSrcNodeOfsTblEnd));
                        nSumOfOpenHandleCounts += pChildEntry->pcd.nOpenHandleCount;
                        ++pRxi->nEntryIdx;
                        ++pChoiceChar;
                    }
                    if (*pChoiceChar == '/') {
                TryNextEntry:
                        ++pChoiceChar;
                        ++pRxi->nEntryIdx;
                    }
                    if (pRxi->nEntryIdx > pNodeHdr->nNumEntries) {
                        do {
                            pRxi->nEntryIdx = pNodeHdr->nEntryIdx + 2;
                            pNodeHdr = (RxtNodeHdr *)((u8*)wkg.pathCache.pRxtHdr + pNodeHdr->nParentOfs);
                            if (pRxi->nDepth == nRootDepth) {
                                // We reached the top of the stack, so this should be the last record
                                --pRxi->nEntryIdx;
                                goto BranchFinished;
                            }
                            --pRxi->nDepth;
                        } while(pRxi->nEntryIdx > pNodeHdr->nNumEntries);
                        pChoiceChar = &pNodeHdr->aFirstChar[pNodeHdr->nStrLen + pRxi->nEntryIdx-1];
                #if _DEBUG
                        if ((unsigned)pChoiceChar[-1] >= (unsigned)*pChoiceChar) {
                            WFSKrnOutputErrorStr("PathCache: Choice out of order!");
                        }
                #endif
                        if (*pChoiceChar == '/') {
                            goto TryNextEntry;
                        }
                    }
                    ++pRxi->nDepth;
                    nParentOfs = nNodeOfs;
                    nNodeOfs = RxtGetChildOfs(pRxi->nEntryIdx, pNodeHdr);
                    pNodeHdr = (RxtNodeHdr *)((u8*)wkg.pathCache.pRxtHdr + nNodeOfs);
                }
            }
        BranchFinished:
            ++nEntryIdx;
        }
    }
    pEntry->pcd.nOpenHandleCount = nSumOfOpenHandleCounts;
InsertToEntryList:
    PathCacheInsertToEntryListAfterCursorEntry(pCursorEntry, pEntry);
IncCount:
    if (pCursorEntry == wkg.pathCache.pPinAnchor) {
        ++pEntry->pcd.nPinCount;
#if _DEBUG_PATH_CACHE
        ++wkg.pathCache.nNumPinned;
        MyOSReport("of:%2d + od:%2d - pc:%2d = %2d  ::  ", wkg.hnd.nNumUsedFileHandles, wkg.hnd.nNumUsedSearchDirHandles, wkg.pathCache.nNumPinned,
            wkg.hnd.nNumUsedFileHandles + wkg.hnd.nNumUsedSearchDirHandles - wkg.pathCache.nNumPinned);
        MyOSReport("%s <- %d\n", sStr, pEntry->pcd.nPinCount);
#endif
    }
    return pEntry;
}


PathCacheEntry *PathCacheIncrementalInsertAfterCursorEntry(PathCacheEntry *pCursorEntry, RxtItr *pRxi, u32 nPrefixLen, u32 nSuffixLen) {
    u32 nStrLen = nPrefixLen + nSuffixLen;
    utf8 *sStr = &pRxi->pStrPtr[-(s32)nPrefixLen];
    WFSKrnResult nResult = PathCacheIncrementalFind(pRxi, nSuffixLen);
    return PathCacheInsertCommon(nResult, pCursorEntry, pRxi, sStr, nStrLen);
}


WFSKrnResult PathCacheIncrementalInsertPinnedEntry(PathCacheItr *pPci, u32 nStrLen) {
    u32 nSuffixLen = pPci->pNameEnd - pPci->rxi.pStrPtr;
    if (nSuffixLen) {
        pPci->pEntry = PathCacheIncrementalInsertAfterCursorEntry(wkg.pathCache.pPinAnchor, &pPci->rxi, nStrLen - nSuffixLen, nSuffixLen);
    } else {
        PathCacheEntry *pEntry = pPci->pEntry;
        if (pEntry->pcd.nPinCount == 0) {
            PathCacheDeleteLink(pEntry);
            PathCacheInsertToEntryListAfterCursorEntry(wkg.pathCache.pPinAnchor, pEntry);
        }
        ++pEntry->pcd.nPinCount;
    }

    if (pPci->pEntry)
    {
        return WFSKRN_RESULT_OK;
    }
    return WFSKRN_RESULT_RESOURCE_LIMIT_EXCEEDED;
}


#if _DEBUG_PATH_CACHE
 void PathCachePinEntry(PathCacheEntry *pEntry, utf8 *sStr) {
#else
 void PathCachePinEntry(PathCacheEntry *pEntry) {
#endif
    if (pEntry->pcd.nPinCount==0) {
        // Transfer to pinned list
        pEntry->ml.pPrev->ml.pNext = pEntry->ml.pNext;
        pEntry->ml.pNext->ml.pPrev = pEntry->ml.pPrev;
        PathCacheInsertToEntryListAfterCursorEntry(wkg.pathCache.pPinAnchor, pEntry);
    }
    ++pEntry->pcd.nPinCount;
#if _DEBUG_PATH_CACHE
        ++wkg.pathCache.nNumPinned;
        MyOSReport("of:%2d + od:%2d - pc:%2d = %2d  ::  ", wkg.hnd.nNumUsedFileHandles, wkg.hnd.nNumUsedSearchDirHandles, wkg.pathCache.nNumPinned,
            wkg.hnd.nNumUsedFileHandles + wkg.hnd.nNumUsedSearchDirHandles - wkg.pathCache.nNumPinned);
        MyOSReport("%s <- %d\n", sStr, pEntry->pcd.nPinCount);
#endif
}


PathCacheEntry *PathCacheInsertAfterCursorEntry3(PathCacheEntry *pCursorEntry, utf8 *sStr, u32 nStrLen) {
    RxtItr rxi;
    rxi.pStrPtr = sStr;
    WFSKrnResult nResult = RxtFind(wkg.pathCache.pRxtHdr, &rxi, nStrLen);
    return PathCacheInsertCommon(nResult, pCursorEntry, &rxi, sStr, nStrLen);
}


#if _DEBUG_PATH_CACHE
 WFSKrnResult PathCacheIncreaseEntryOpenHandleCount(PathCacheItr *pPci, WFSAccess nDesiredAccess, utf8 *sStr) {
#else
 WFSKrnResult PathCacheIncreaseEntryOpenHandleCount(PathCacheItr *pPci, WFSAccess nDesiredAccess) {
#endif
    PathCacheEntry *pEntry = pPci->pEntry;
    if (pEntry->pcd.nFlags & PATH_CACHE_FLAG_OPEN_FOR_WRITE) {
        --pEntry->pcd.nPinCount;    // Assuming the PinCount has been incremented and needs to be undone
#if _DEBUG_PATH_CACHE
        --wkg.pathCache.nNumPinned;
        MyOSReport("of:%2d + od:%2d - pc:%2d = %2d  ::  ", wkg.hnd.nNumUsedFileHandles, wkg.hnd.nNumUsedSearchDirHandles, wkg.pathCache.nNumPinned,
            wkg.hnd.nNumUsedFileHandles + wkg.hnd.nNumUsedSearchDirHandles - wkg.pathCache.nNumPinned);
        MyOSReport("%s <- %d\n", sStr, pEntry->pcd.nPinCount);
#endif
        return WFSKRN_RESULT_FILE_OPEN;
    }
    if (nDesiredAccess & WFS_ACCESS_WRITE) {
        if (1 < pEntry->pcd.nPinCount)
        {
            // This file is already open. // [check] right?
            return WFSKRN_RESULT_FILE_OPEN;
        }
        pEntry->pcd.nFlags |= PATH_CACHE_FLAG_OPEN_FOR_WRITE;
    }
    RxtNodeHdr *pNodeHdr = pPci->rxi.pNodeHdr;
    while(1) {
        ++pEntry->pcd.nOpenHandleCount;
        while(1) {
            if (pNodeHdr->nParentOfs == 0) {
                return WFSKRN_RESULT_OK;
            }
            u32 nEntryIdx = pNodeHdr->nEntryIdx;
            pNodeHdr = (RxtNodeHdr *)((u8*)wkg.pathCache.pRxtHdr + pNodeHdr->nParentOfs);
            utf8 *pStr = pNodeHdr->aFirstChar;
            utf8 *pStrEnd = pStr + pNodeHdr->nStrLen;
            if (pStrEnd[nEntryIdx] == '/') {
                goto FoundSlash;
            }
            while(pStr<pStrEnd) {
                if (*pStr++ == '/') {
                    goto FoundSlash;
                }
            }
        }
FoundSlash:
        while(pNodeHdr->aFirstChar[pNodeHdr->nStrLen] != 0) {
            if (pNodeHdr->nParentOfs == 0) {
                return WFSKRN_RESULT_OK;
            }
            pNodeHdr = (RxtNodeHdr *)((u8*)wkg.pathCache.pRxtHdr + pNodeHdr->nParentOfs);
        };
        pEntry = (PathCacheEntry *)((u8*)wkg.pathCache.pRxtHdr + RxtGetChildOfs(1, pNodeHdr));
    }
}


#if _DEBUG_PATH_CACHE
 WFSKrnResult PathCacheDecreaseEntryOpenHandleCount(PathCacheItr *pPci, utf8 *sStr) {
#else
 WFSKrnResult PathCacheDecreaseEntryOpenHandleCount(PathCacheItr *pPci) {
#endif
    PathCacheEntry *pEntry = pPci->pEntry;
    pPci->rxi.nEntryIdx = pEntry->ah.nEntryIdx+1;
    RxtNodeHdr *pNodeHdr = pPci->rxi.pNodeHdr  = (RxtNodeHdr *)((u8*)wkg.pathCache.pRxtHdr + pEntry->ah.nParentOfs);
    while(1) {
#if _DEBUG
        if (pEntry->pcd.nOpenHandleCount == 0) {
            WFSKrnOutputErrorStr("OpenHandleCount already 0!");
        }
#endif
        --pEntry->pcd.nOpenHandleCount;
        while(1) {
            if (pNodeHdr->nParentOfs == 0) {
                goto ExitLoop;
            }
            u32 nEntryIdx = pNodeHdr->nEntryIdx;
            pNodeHdr = (RxtNodeHdr *)((u8*)wkg.pathCache.pRxtHdr + pNodeHdr->nParentOfs);
            utf8 *pStr = pNodeHdr->aFirstChar;
            utf8 *pStrEnd = pStr + pNodeHdr->nStrLen;
            if (pStrEnd[nEntryIdx] == '/') {
                goto FoundSlash;
            }
            while(pStr<pStrEnd) {
                if (*pStr++ == '/') {
                    goto FoundSlash;
                }
            }
        }
FoundSlash:
        while(pNodeHdr->aFirstChar[pNodeHdr->nStrLen] != 0) {
            if (pNodeHdr->nParentOfs == 0) {
                goto ExitLoop;
            }
            pNodeHdr = (RxtNodeHdr *)((u8*)wkg.pathCache.pRxtHdr + pNodeHdr->nParentOfs);
        }
        pEntry = (PathCacheEntry *)((u8*)wkg.pathCache.pRxtHdr + RxtGetChildOfs(1, pNodeHdr));
    }
ExitLoop:
    pEntry = pPci->pEntry;
#if _DEBUG
    if (pEntry->pcd.nPinCount == 0) {
        WFSKrnOutputErrorStr("nPinCount already 0!");
    }
#endif
    --pEntry->pcd.nPinCount;
#if _DEBUG_PATH_CACHE
        --wkg.pathCache.nNumPinned;
        MyOSReport("of:%2d + od:%2d - pc:%2d = %2d  ::  ", wkg.hnd.nNumUsedFileHandles, wkg.hnd.nNumUsedSearchDirHandles, wkg.pathCache.nNumPinned,
            wkg.hnd.nNumUsedFileHandles + wkg.hnd.nNumUsedSearchDirHandles - wkg.pathCache.nNumPinned);
        MyOSReport("%s <- %d\n", sStr, pEntry->pcd.nPinCount);
#endif
    if (pEntry->pcd.nPinCount == 0) {
#if 0
        PathCacheDeleteEntry(pEntry);
#else
        // Move to MRU list
        pEntry->pcd.nFlags &=~PATH_CACHE_FLAG_OPEN_FOR_WRITE;
        PathCacheDeleteLink(pEntry);
        PathCacheInsertToEntryListAfterCursorEntry(wkg.pathCache.pMruAnchor, pEntry);
#endif
        return WFSKRN_RESULT_OK;
    }
    return WFSKRN_RESULT_NOT_EMPTY;
}


void PathCacheInit() {
    wkg.pathCache.pMruAnchor = (PathCacheEntry *)((u8*)&wkg.pathCache.mruAnchor - (size_t)&((PathCacheEntry *)0)->ml);
    wkg.pathCache.pMruAnchor->ml.pNext = wkg.pathCache.pMruAnchor->ml.pPrev = wkg.pathCache.pMruAnchor;
    wkg.pathCache.pPinAnchor = (PathCacheEntry *)((u8*)&wkg.pathCache.pinAnchor - (size_t)&((PathCacheEntry *)0)->ml);
    wkg.pathCache.pPinAnchor->ml.pNext = wkg.pathCache.pPinAnchor->ml.pPrev = wkg.pathCache.pPinAnchor;
    wkg.pathCache.pRxtHdr = (RxtHdr *)wkg.pathCache.aRxtBuf;
    RxtInit(wkg.pathCache.pRxtHdr, WFS_LOG2_PATH_CACHE_SIZE);
}


#if (!_IOP && _DEBUG)
    #include "randomlib.h"
    void VolumeConvertU32ToVolId(WFSVolumeId *pVolId, u32 nVolId);
    extern utf8 **aTestWords;

    void PathCacheTest() {
        PathCacheDebugOutputList(wkg.pathCache.pMruAnchor, 'C');
        u32 nDev;
        for(nDev=0; nDev<1000; ++nDev) {
            utf8 sVolPath[WFS_MAX_DEVICE_NAME_SIZE + 50];
            u32 nVolId = ((RandomU32()%24) * 10000) + 5000;
            WFSVolumeId volId;
            VolumeConvertU32ToVolId(&volId, nVolId);
            SNPRINTF(sVolPath, WFS_MAX_DEVICE_NAME_SIZE + 50,"vol/%s", aTestWords[nVolId]);
            PathCacheInsertAfterCursorEntry2(wkg.pathCache.pMruAnchor, sVolPath)->pcd.pData = &nDev;
            PathCacheDebugOutputList(wkg.pathCache.pMruAnchor, 'C');
        }
        PathCacheDebugOutputList(wkg.pathCache.pMruAnchor, 'C');
    }
#endif


WFSKrnResult PathCacheParseNextName(WFSFileName *pName, const utf8 **ppSrc) {
    const utf8 *pSrc = *ppSrc + 1;
    utf8 *pDst = pName->sStr;
    utf8 *pDstEnd = &pName->sStr[WFS_MAX_FILE_NAME_SIZE];
    while((*pSrc!='/')&&(*pSrc!=0)) {
        if (pDst == pDstEnd) {
            return WFSKRN_RESULT_INVALID;
        }
        *pDst++ = *pSrc++;
    }
    *pDst = 0;
    *ppSrc = pSrc;
    pName->nLen = pDst - pName->sStr;
    if ((*pSrc=='/')&&(pSrc[1]!=0)) {
        return WFSKRN_RESULT_OK;
    } else {
        return WFSKRN_RESULT_SRV_END_OF_PATH;
    }
}


void PathCacheSetItrToVolumeRootArea(PathCacheItr *pPci, VolumeInfo *pVolInfo) {
    pPci->di.tba.pAreaInfo = &pVolInfo->rootAreaInfo;
    pPci->di.tba.nBlkAdr = pPci->di.tba.pAreaInfo->ah.nRootBlkAdr;
    pPci->di.pAttrHdr = &pVolInfo->rootAttr;
    pPci->di.ppDnsRootParent = 0;
}


WFSKrnResult PathCacheFindParentDir(const WFSPathName *pAbsPathName, utf8 *sLowerCaseStr, PathCacheItr *pPci) {
    // This function searches for the parent directory of the input path name, and
    // sets up the directory iterator to the top of this directory.
    pPci->di.ppDnsRootParent = 0;
    pPci->pNameEnd = pPci->rxi.pStrPtr = sLowerCaseStr;
    WFSKrnResult nResult = PathCacheParseNextName(&pPci->di.name, &pPci->pNameEnd);
    if (nResult == WFSKRN_RESULT_SRV_END_OF_PATH) {
        // If we get here, it means the path had depth 1. (e.g. "/something")
        PathCacheInitSlash(pPci); // to skip the root '/'
        pPci->di.tba.pAreaInfo = 0;
        nResult = WFSKRN_RESULT_SRV_PATH_DEPTH_1;
        goto SetEntry;
    } else if (nResult != WFSKRN_RESULT_OK) {
        return nResult;
    }
    void *pRootData;
    if (!PathCacheFind3(&pPci->rxi, pPci->pNameEnd - pPci->rxi.pStrPtr, &pRootData)) {
        return WFSKRN_RESULT_NOT_FOUND;
    }
    // First name should be "/vol" or "/dev"
    nResult = PathCacheParseNextName(&pPci->di.name, &pPci->pNameEnd);
    if (nResult == WFSKRN_RESULT_SRV_END_OF_PATH) {
        // The path was found in the path cache (e.g. "/vol" or "/dev")
        pPci->di.tba.pAreaInfo = 0;
        nResult = (WFSKrnResult)((u32)pRootData);
        goto SetEntry;
    } else if (nResult != WFSKRN_RESULT_OK) {
        return nResult;
    }
    void *pData;
    pPci->pEntry = PathCacheIncrementalFindEntry(&pPci->rxi, pPci->pNameEnd - pPci->rxi.pStrPtr, &pData);
    if (pPci->pEntry) {
        if (((DeviceInfo*)pData >= wkg.aDevInfo) && ((DeviceInfo*)pData < &wkg.aDevInfo[WFSKRN_MAX_DEVICES])) {
            // Cannot directly access devices listed under /dev
            //return WFSKRN_RESULT_PERMISSION;
           // If we get here, the path included device name as directory.
            return WFSKRN_RESULT_NOT_DIRECTORY;
        }
        if (((VolumeInfo*)pData >= wkg.aVolInfo) && ((VolumeInfo*)pData < &wkg.aVolInfo[WFSKRN_MAX_VOLUMES])) {
            // Access the root directory of a volume
            PathCacheSetItrToVolumeRootArea(pPci, (VolumeInfo*)pData);
#ifdef PERMISSION_ENABLED
            pPci->pEntry->pcd.nAccessListIdx = pPci->di.pAttrHdr->nAccessListIdx;
#endif
            // On this path, the condition "nResult == WFSKRN_RESULT_SRV_END_OF_PATH" could not be true
            if (nResult == WFSKRN_RESULT_SRV_END_OF_PATH) {
                // If we get here, the path was a volume root directory e.g. "/vol/abcdefg")
                nResult = WFSKRN_RESULT_SRV_PATH_VOL_ROOT;
                goto SetEntry;
            }
            else if(pPci->di.tba.pAreaInfo->pVolInfo->pDevInfo->nFlags & WFS_DEVFLAG_UNUSABLE) {
                return WFSKRN_RESULT_DEV_UNUSABLE;
            }
            else if(pPci->di.tba.pAreaInfo->pVolInfo->pDevInfo->nFlags & WFS_DEVFLAG_WRITE_PROTECTED) {
                return WFSKRN_RESULT_WRITE_PROTECTED;
            }
            else if(pPci->di.tba.pAreaInfo->pVolInfo->pDevInfo->nFlags & WFS_DEVFLAG_NOT_INITIALIZED) {
                return WFSKRN_RESULT_DEV_NOT_INITIALIZED;
            }
        } else {
            return WFSKRN_RESULT_NOT_FOUND;
        }
    } else {
        return WFSKRN_RESULT_NOT_FOUND;
    }
    PathCacheEntry *pCursorEntry = wkg.pathCache.pMruAnchor;
    // Now we are ready to start parsing normal directories within a volume
    while(1) {
        pPci->nRootBlkAdr = pPci->di.tba.nBlkAdr;
        PathCacheParseNextName(&pPci->di.name, &pPci->pNameEnd);
        if ((pPci->pNameEnd[0]!='/')||(pPci->pNameEnd[1]==0)) {
            goto Done;
        }
        pPci->pEntry = PathCacheIncrementalInsertAfterCursorEntry(pCursorEntry, &pPci->rxi, pPci->rxi.pStrPtr - sLowerCaseStr, pPci->di.name.nLen+1);
        if (pPci->pEntry == NULL)
        {
            return WFSKRN_RESULT_RESOURCE_LIMIT_EXCEEDED;
        }

#if PATH_CACHE_ENABLED
        if (pPci->pEntry->pcd.pAreaInfo == 0) {
#endif
            if (pPci->di.name.nLen==0) {
                return WFSKRN_RESULT_INVALID;
            }
            nResult = DirFindRaw(&pPci->di);
            if (nResult != WFSKRN_RESULT_OK) {
                goto Exit;
            }
            if ((pPci->di.pAttrHdr->nAccessListIdx & WFS_FLAG_IS_A_DIRECTORY)==0) {
                nResult = WFSKRN_RESULT_NOT_DIRECTORY;
                goto Exit;
            }
            pPci->pEntry->pcd.pAreaInfo = pPci->di.tba.pAreaInfo;
            pPci->pEntry->pcd.nBlkAdr   = pPci->di.pAttrHdr->dir.nSubDirBlkAdr;
#ifdef PERMISSION_ENABLED
            pPci->pEntry->pcd.nAccessListIdx = pPci->di.pAttrHdr->nAccessListIdx;
            pPci->pEntry->pcd.nObjCreator    = pPci->di.pAttrHdr->nObjCreator;
#endif
            DirItrClose(&pPci->di);
            pPci->di.tba.nBlkAdr        = pPci->pEntry->pcd.nBlkAdr;
#if PATH_CACHE_ENABLED
        } else {
            pPci->di.tba.pAreaInfo = pPci->pEntry->pcd.pAreaInfo;
            pPci->di.tba.nBlkAdr   = pPci->pEntry->pcd.nBlkAdr;
        }
#endif
        //pPci->rxi.pStrPtr = const_cast<utf8*>(pPci->pNameEnd);
        pPci->rxi.pStrPtr = (utf8*)(pPci->pNameEnd); // IOP compiler
        if (pPci->pEntry->pcd.nPinCount == 0) {
            pCursorEntry = pPci->pEntry;
        }
    }
SetEntry:
    pPci->pEntry = (PathCacheEntry *)((u8*)wkg.pathCache.pRxtHdr + RxtGetChildOfs(1, pPci->rxi.pNodeHdr));
Done:
    ++pPci->rxi.pStrPtr;
    u32 nStrLen = pPci->pNameEnd - pPci->rxi.pStrPtr;
    memcpy(pPci->di.name.sStr, pPci->rxi.pStrPtr, nStrLen);
    pPci->pRawNameStart = pAbsPathName->sStr + (pPci->rxi.pStrPtr - sLowerCaseStr);
    return nResult;
Exit:
    DirItrClose(&pPci->di);
    return nResult;
}


bool WFSKrnNameMatchesPattern(utf8 *sName, utf8 *sPattern);


WFSKrnResult PathCacheFindName(PathCacheItr *pPci) {
    void *pData;
    RxtItr oldRxi = pPci->rxi;
    PathCacheEntry *pEntry = PathCacheIncrementalFindEntry(&pPci->rxi, pPci->pNameEnd - pPci->rxi.pStrPtr, &pData);
    if (pEntry) {
        pPci->pEntry = pEntry;
        return WFSKRN_RESULT_OK;
    } else {
        pPci->rxi = oldRxi;
        return WFSKRN_RESULT_NOT_FOUND;
    }
}


WFSKrnResult PathCacheFindPath(const WFSPathName *pAbsPathName, utf8 *sLowerCaseStr, PathCacheItr *pPci) {
    WFSKrnResult nResult = PathCacheFindParentDir(pAbsPathName, sLowerCaseStr, pPci);
    void *pData;
    switch(nResult) {
        case WFSKRN_RESULT_OK: {
            PathCacheFindName(pPci);
            nResult = DirFindRaw(&pPci->di);
            return nResult;
        }
        case WFSKRN_RESULT_SRV_PATH_DEPTH_1: {
            //PathCacheInitSlash(pPci);
            //pPci->di.tba.pAreaInfo = 0;
            s32 nStrLen = pPci->pNameEnd - pPci->rxi.pStrPtr;
            if (nStrLen == 0) {
                nResult = WFSKRN_RESULT_OK;
                pPci->di.pAttrHdr = &wkg.rootDirAttr;
                break;
            }
            if (PathCacheIncrementalFindEntry(&pPci->rxi, nStrLen, &pData)) {
                // The path was found in the path cache (e.g. "/vol" or "/dev")
                nResult = (WFSKrnResult)((u32)pData);
                if (nResult == WFSKRN_RESULT_SRV_PATH_DEV) {
                    pPci->di.pAttrHdr = &wkg.devDirAttr;
                } else {
                    pPci->di.pAttrHdr = &wkg.volDirAttr;
                }
                nResult = WFSKRN_RESULT_OK;
                break;
            }
            return WFSKRN_RESULT_NOT_FOUND;
        }
        case WFSKRN_RESULT_SRV_PATH_DEV:
            nResult = PathCacheFindName(pPci);
            if (nResult == WFSKRN_RESULT_OK) {
                pPci->di.pAttrHdr = &wkg.devAttr;
            }
            break;
        case WFSKRN_RESULT_SRV_PATH_VOL:
            if (PathCacheIncrementalFindEntry(&pPci->rxi, pPci->pNameEnd - pPci->rxi.pStrPtr, &pData)) {
                PathCacheSetItrToVolumeRootArea(pPci, (VolumeInfo*)pData);
                nResult = WFSKRN_RESULT_OK;
                break;
            }
            return WFSKRN_RESULT_NOT_FOUND;
        case WFSKRN_RESULT_SRV_PATH_VOL_ROOT:
            nResult = WFSKRN_RESULT_OK;
            break;
        default:
            return nResult;
     }
     return nResult;
}
