/*---------------------------------------------------------------------------*
  Project:  wfsKrn
  File:     wfsKrn_Permission.cpp - Functions for permissions.

  Copyright 2008 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_Permission.cpp,v $
  Revision 1.16  2008/12/08 04:15:52  ooizumi
  Clean up GetInheritedEntry().

  Revision 1.15  2008/12/05 04:09:27  ooizumi
  Fixed a bug failed to update self entry.

  Revision 1.14  2008/12/04 00:49:24  ooizumi
  Fixed.

  Revision 1.13  2008/11/19 05:05:59  ooizumi
  Fixed.

  Revision 1.12  2008/11/04 06:35:15  ooizumi
  Fixed a bug failed to check VolumeInfo specified as AreaInfo.

  Revision 1.11  2008/10/10 10:47:21  ooizumi
  Fixed a bug.

  Revision 1.10  2008/10/10 04:24:37  ooizumi
  Fixed a bug stacking in infinite loop when calculating everyone entry's creator's permission.

  Revision 1.9  2008/10/09 09:39:03  ooizumi
  Fixed bugs.
  - Failed to get parent directory's permissions.
  - Changed a routine to decide whether to create acl entry or not.

  Revision 1.8  2008/10/09 04:11:16  ooizumi
  Fixed build error.

  Revision 1.7  2008/10/09 03:33:15  ooizumi
  Fixed several bugs.
  - Failed to get/set permissions of volume root.
  - A caller which has negative CL value have several permissions.
  - Failed to create acl entry which has no permissions (0x00000000).

  Revision 1.6  2008/10/08 04:16:15  ooizumi
  Fixed not to check CL when it is no need to update acl.

  Revision 1.5  2008/10/07 10:42:51  ooizumi
  Supported inherited permissions without using pathcache.

  Revision 1.4  2008/07/31 23:59:19  ooizumi
  Fixed permission of special directories..

  Revision 1.3  2008/07/30 06:11:57  ooizumi
  Implemented initial version of permission functions.

  Revision 1.2  2008/07/14 00:42:00  ueno
  Commented out because the APIs in this file are obsolete.

  Revision 1.1  2008/04/22 04:39:08  kondo_masahiro
  Fixed code to transfer the wfskrn project to RM_SDK tree

  Revision 1.22  2008/04/03 07:06:39  ueno
  Fixed typos.

  Revision 1.20  2008/03/14 08:02:53  ueno
  Revised API prefixes (WFS -> WFSKrn).

  Revision 1.19  2008/03/11 11:39:01  ueno
  Cleanup.

  Revision 1.18  2008/03/11 11:10:47  ueno
  Moved accesslist functions into wfskrn_Permission_AccessList.cpp.

  Revision 1.17  2008/03/11 09:51:54  ueno
  Fixed EscapeAccessList() to make a filename from UTF-16 characters (0x3300-0x33ff).

  Revision 1.16  2008/03/11 06:53:05  ueno
  Added WFSAccessListSetPermission() and GetPermission().

  Revision 1.15  2008/03/07 05:43:59  ueno
  Cleanup.

  Revision 1.14  2008/03/07 05:32:07  ueno
  Modified WFSCreateAccessList() to be able to create an accesslist for directories.

  Revision 1.13  2008/03/06 12:25:27  ueno
  Revised cache freelist (signgle -> double link list).

  Revision 1.12  2008/03/06 11:01:38  ueno
  Fixed asserts.

  Revision 1.11  2008/03/06 10:45:44  ueno
  Revised accesslist cache.

  Revision 1.10  2008/02/29 07:29:10  ueno
  Added Windows api wrappers.

  Revision 1.9  2008/02/29 00:48:05  ueno
  Changed the attribute of index to inout (WFSAccessListAddEntry() and RemoveEntry()).

  Revision 1.8  2008/02/28 13:11:01  ueno
  Revised to keep accesslist cache and indexlist handle in WFSAreaHdr{}.

  Revision 1.7  2008/02/28 07:01:52  ueno
  Fixed escape sequence for accesslist filename.

  Revision 1.6  2008/02/28 04:36:32  ueno
  Cleanup.

  Revision 1.5  2008/02/27 10:05:36  ueno
  Revised to use WFSAreaHdr{} to specifiy WFS area.

  Revision 1.4  2008/02/27 04:29:24  ueno
  Modified WFSAccessListAddEntry() and RemoveEntry() to delete accesslist file when the reference count is zero.

  Revision 1.3  2008/02/26 14:11:43  ueno
  Added WFSAccessListAddEntry() and WFSAccessListRemoveEntry() to edit accesslist.

  Revision 1.2  2008/02/25 09:43:57  ueno
  Added support for accesslist cache.

  Revision 1.1  2008/02/22 05:16:13  ueno
  Initial Check-in.


*---------------------------------------------------------------------------*/
#include "wfskrn_Config.h"

#ifdef PERMISSION_ENABLED

#include "wfskrn_Api.h"
#include "wfskrn_DirRxTree.h"
#include "wfskrn_PathCache.h"

#include "wfskrn_Permission.h"
#include "wfskrn_Permission_AccessList.h"

/*---------------------------------------------------------------------------*
  Name:         BacktrackToSlash

  Description:  Backtrack radix tree iterator to slash.

  Arguments:    pRxi   Radix tree iterator

  Returns:      Result of operation.
 *---------------------------------------------------------------------------*/
static WFSKrnResult BacktrackToSlash(RxtItr *pRxi)
{
    RxtNodeHdr *pNodeHdr = pRxi->pNodeHdr;
    u32 nDepth = 0;
    u16 nLen   = 0;
    while (1)
    {
        if (pNodeHdr->nParentOfs == 0)
        {
            // If we get here, it means the path had depth 1. (e.g. "/something")
            return WFSKRN_RESULT_SRV_PATH_DEPTH_1;
        }
        // Searching parent node until '/' is found.
        ++nDepth;
        nLen += pNodeHdr->nStrLen + 1;
        u32  nEntryIdx = pNodeHdr->nEntryIdx;
        pNodeHdr = (RxtNodeHdr *)((u8*)wkg.pathCache.pRxtHdr + pNodeHdr->nParentOfs);
        utf8 *pStr    = pNodeHdr->aFirstChar;
        utf8 *pStrEnd = pStr + pNodeHdr->nStrLen;
        if (pStrEnd[nEntryIdx] == '/')
        {
            break;
        }
    }
    if (!*(pRxi->pStrPtr))
    {
        --nLen;
    }

    pRxi->pNodeHdr     = pNodeHdr;
    pRxi->pStrPtr     -= nLen;
    pRxi->pNodeStrPtr  = &pNodeHdr->aFirstChar[pNodeHdr->nStrLen];
    pRxi->nEntryIdx    = pNodeHdr->nEntryIdx;
    pRxi->nDepth      -= nDepth;

    pNodeHdr = (RxtNodeHdr *)((u8*)wkg.pathCache.pRxtHdr + pNodeHdr->nParentOfs);
    pRxi->pOfs = RxtGetChildOfsPtr(pRxi->nEntryIdx + 1, pNodeHdr);

    return WFSKRN_RESULT_OK;
}

/*---------------------------------------------------------------------------*
  Name:         GetParentPathCacheEntry

  Description:  Get parent path cache entry.

  Arguments:    pEntry  Path cache entry
                pRxi    Radix tree iterator

  Returns:      Parent entry. If failed, returns NULL.
 *---------------------------------------------------------------------------*/
static PathCacheEntry *GetParentPathCacheEntry(PathCacheEntry *pEntry, RxtItr *pRxi)
{
    if (*pRxi->pStrPtr || (BacktrackToSlash(pRxi) == WFSKRN_RESULT_OK))
    {
        PathCacheEntry *entry = (PathCacheEntry *)((u8*)wkg.pathCache.pRxtHdr + RxtGetChildOfs(1, pRxi->pNodeHdr));
        BacktrackToSlash(pRxi);
        return entry;
    }
    return 0;
}

// Control level converting functions
static inline s32 GetRelCL(u32 perm)
{
    switch (perm & WFS_PERM_CL_MASK)
    {
        case WFS_PERM_CL_MINUS_0: return 0;
        case WFS_PERM_CL_MINUS_1: return -1;
        case WFS_PERM_CL_MINUS_2: return -2;
        case WFS_PERM_CL_MINUS_3: return -3;
        default:                  return 1;
    }
}
static inline s32 GetAbsCL(u32 perm)
{
    switch (perm & WFS_PERM_CL_MASK)
    {
        case WFS_PERM_CL_0: return 0;
        case WFS_PERM_CL_1: return 1;
        case WFS_PERM_CL_2: return 2;
        case WFS_PERM_CL_3: return 3;
        default:            return -1;
    }
}
static inline u32 GetAbsCLFlag(s32 abs)
{
    switch (abs)
    {
        case 0:  return WFS_PERM_CL_0;
        case 1:  return WFS_PERM_CL_1;
        case 2:  return WFS_PERM_CL_2;
        case 3:  return WFS_PERM_CL_3;
        default: return 0;
    }
}

static inline u32 GetRelCLFlag(s32 rel)
{
    switch (rel)
    {
        case  0:  return WFS_PERM_CL_MINUS_0;
        case -1:  return WFS_PERM_CL_MINUS_1;
        case -2:  return WFS_PERM_CL_MINUS_2;
        case -3:  return WFS_PERM_CL_MINUS_3;
        default:  return 0;
    }
}

/*---------------------------------------------------------------------------*
  Name:         GetAclEntries

  Description:  Get acl entries for specified title. Title entry, self entry
                and everyone entry.

  Arguments:    pAreaInfo       Pointer to AreaInfo
                nAccessListIdx  Access list handle
                nTitleId        Key title id

                pTitleEntry     Pointer to WFSAccessListEntry buffer to
                pSelfEntry      store entry informations
                pEveryoneEntry

  Returns:      If failed, returns error code from WFSKrnGetAccessList.
 *---------------------------------------------------------------------------*/
static WFSKrnResult GetAclEntries(const AreaInfo* pAreaInfo, WFSKrnAclHandle nAccessListIdx, WFSTitleId nTitleId,
                                  WFSAccessListEntry* pTitleEntry, WFSAccessListEntry* pSelfEntry, WFSAccessListEntry* pEveryoneEntry)
{
    // to get access list entry
    WFSAccessListEntry  aAclEntry[ACL_MAX_ENTRIES]; // 12*12=144
    s32 num;

    WFSKRN_PERM_ASSERT(pTitleEntry);
    WFSKRN_PERM_ASSERT(pSelfEntry);
    WFSKRN_PERM_ASSERT(pEveryoneEntry);

    num = (WFSKrnResult)WFSKrnGetAccessList(pAreaInfo, nAccessListIdx, aAclEntry, ACL_MAX_ENTRIES);
    if (num < 0)
    {
        return (WFSKrnResult)num;
    }
    while (num)
    {
        if (aAclEntry[num-1].titleId == nTitleId)
        {
            if (aAclEntry[num-1].titleId == aAclEntry[num-1].entryCreatorId)
            {
                pSelfEntry->titleId         = aAclEntry[num-1].titleId;
                pSelfEntry->entryCreatorId  = aAclEntry[num-1].entryCreatorId;
                pSelfEntry->permissions     = aAclEntry[num-1].permissions;
            }
            else
            {
                pTitleEntry->titleId         = aAclEntry[num-1].titleId;
                pTitleEntry->entryCreatorId  = aAclEntry[num-1].entryCreatorId;
                pTitleEntry->permissions     = aAclEntry[num-1].permissions;
            }
        }
        if (aAclEntry[num-1].titleId == WFS_EVERYONE_ID)
        {
            pEveryoneEntry->titleId         = aAclEntry[num-1].titleId;
            pEveryoneEntry->entryCreatorId  = aAclEntry[num-1].entryCreatorId;
            pEveryoneEntry->permissions     = aAclEntry[num-1].permissions;
        }
        num--;
    }
    return WFSKRN_RESULT_OK;
}

/*---------------------------------------------------------------------------*
  Name:         UpdateParentPermForChild

  Description:  Update parent permissions for child entry by calculating
                object creator's privilege and WFS_PERM_DELETE_OR_MOVE.

  Arguments:    nFlags    Flags from parent entry
                bDir      Directory or not
                bCreator  Object creator's privilege is enabled or not

  Returns:      Updated nFlags value.
 *---------------------------------------------------------------------------*/
static u32 UpdateParentPermForChild(u32 nFlags, BOOL bDir, BOOL bCreator)
{
    if (bDir)
    {
        if (nFlags & WFS_PERM_ADD_OR_DELETE_SUBDIR)
        {
            if (bCreator)
            {
                nFlags |= WFS_PERM_DIR_FULL;
            }
        }
        else
        {
            nFlags &= ~WFS_PERM_DELETE_OR_MOVE;
        }
    }
    else
    {
        if (nFlags & WFS_PERM_ADD_OR_DELETE_FILE)
        {
            if (bCreator)
            {
                nFlags |= WFS_PERM_FILE_FULL;
            }
        }
        else
        {
            nFlags &= ~WFS_PERM_DELETE_OR_MOVE;
        }
    }
    return nFlags;
}

/*---------------------------------------------------------------------------*
  Name:         CalcPermissions

  Description:  Calculate permissions from base flags and entry permissions.

  Arguments:    nFlags        Base flags
                nPermissions  Affected entry's permissions

  Returns:      Updated flags value.
 *---------------------------------------------------------------------------*/
static u32 CalcPermissions(u32 nFlags, u32 nPermissions)
{
    if (((s32)GetAbsCL(nFlags) + GetRelCL(nPermissions)) < 0)
    {
        // if CL is a negative value, title have no permission.
        nFlags = 0;
    }
    else
    {
        nFlags =
            GetAbsCLFlag(GetAbsCL(nFlags) + GetRelCL(nPermissions)) |
                ((nPermissions & ~WFS_PERM_CL_MASK) & (nFlags & ~WFS_PERM_CL_MASK));
    }
    return nFlags;
}


#define IS_ENTRY_AVAILABLE(entry) ((entry)->titleId != 0)
#define IS_SPECIAL_DIR(area)      ((area) == (void*)0 || \
                                   (area) == (void*)WFSKRN_RESULT_SRV_PATH_DEV || \
                                   (area) == (void*)WFSKRN_RESULT_SRV_PATH_VOL)

/*---------------------------------------------------------------------------*
  Name:         GetInheritedEntry

  Description:  Get inherited acl entry.

  Arguments:    pEntry          Path cache entry
                pRxi            Radix tree iterator
                pAreaInfo       Area info
                nAccessListIdx  Access list handle
                nObjCreator     Object creator's title id
                pPerm           Specified target title id
                self            Enable self entry or not
                everyone        Enable everyone entry or not

  Returns:      WFSKrnResult code.
 *---------------------------------------------------------------------------*/
static WFSKrnResult GetInheritedEntry(PathCacheEntry *pEntry, RxtItr *pRxi,
                                      AreaInfo* pAreaInfo, u32 nAccessListIdx, WFSTitleId nObjCreator,
                                      PathCachePerm* pPerm,
                                      BOOL self, BOOL everyone)
{
    WFSKrnResult nResult = WFSKRN_RESULT_OK; // 4
    RxtItr rxi;                              // 24

    WFSTitleId titleId; // 4

    PathCachePerm entryCreatorPerm; // 16
    PathCachePerm parentPerm;       // 16
    PathCachePerm everyonePerm;     // 16

    PathCacheEntry* pParentEntry;    // 4
    AreaInfo*       pParentAreaInfo; // 4

    titleId  = pPerm->nTitleId;

    // root has full permission.
    if (titleId == WFS_ROOT_ID)
    {
        pPerm->nFlags          = WFS_PERM_CL_3 | WFS_PERM_DIR_FULL | WFS_PERM_FILE_FULL;
        pPerm->nEntryCreatorId = WFS_ROOT_ID;
        pPerm->nEntryCreatorCl = 3;
        nResult = WFSKRN_RESULT_OK;
        goto Exit;
    }

    // permissions for special dirs.
    if (IS_SPECIAL_DIR(pAreaInfo))
    {
        pPerm->nFlags          = WFS_PERM_LIST;
        pPerm->nEntryCreatorId = WFS_ROOT_ID;
        pPerm->nEntryCreatorCl = 0;
        nResult = WFSKRN_RESULT_OK;
        goto Exit;
    }

    memcpy(&rxi, pRxi, sizeof(RxtItr));

    // to get access list entry
    WFSAccessListEntry titleEntry;    // 12
    WFSAccessListEntry selfEntry;     // 12
    WFSAccessListEntry everyoneEntry; // 12
    titleEntry.titleId    = 0;
    selfEntry.titleId     = 0;
    everyoneEntry.titleId = 0;

    nResult = GetAclEntries(pAreaInfo, nAccessListIdx, titleId,
                            &titleEntry, &selfEntry, &everyoneEntry);
    if (nResult < 0)
    {
        goto Exit;
    }

    if (!self)
    {
        selfEntry.titleId = 0;
    }

    // get title's inherited entry at parent dir
    pParentEntry = GetParentPathCacheEntry(pEntry, &rxi);
    // if no parent entry is there, it doesn't reach heare.
    WFSKRN_PERM_ASSERT(pParentEntry);

    pParentAreaInfo = pParentEntry->pcd.pAreaInfo;
    if (!pParentEntry->pcd.nBlkAdr)
    {
        // Special directory has no nBlkAdr in PathCache
        void *pData = pParentEntry->pcd.pData;
        if (((VolumeInfo*)pData >= wkg.aVolInfo) && ((VolumeInfo*)pData < &wkg.aVolInfo[WFSKRN_MAX_VOLUMES]) &&
            !(((u32)pData - (u32)wkg.aVolInfo) % sizeof(VolumeInfo)))
        {
            // Access the root directory of a volume
            pParentAreaInfo = &((VolumeInfo*)pParentAreaInfo)->rootAreaInfo;
        }
        // ToDo: To handle error.
    }
    parentPerm.nTitleId = titleId;
    if (IS_SPECIAL_DIR(pParentAreaInfo))
    {
        parentPerm.nFlags          = 0;
        parentPerm.nEntryCreatorId = WFS_ROOT_ID;
        parentPerm.nEntryCreatorCl = 0;
    }
    else
    {
        nResult = GetInheritedEntry(pParentEntry, &rxi,
                                    pParentAreaInfo, pParentEntry->pcd.nAccessListIdx, pParentEntry->pcd.nObjCreator,
                                    &parentPerm,
                                    TRUE, TRUE);
        if (nResult < 0)
        {
            goto Exit;
        }
    }

    // update parent entry permissions for child,
    // by calculating object creator's privilege and WFS_PERM_DELETE_OR_MOVE
    parentPerm.nFlags = UpdateParentPermForChild(parentPerm.nFlags,
                                                 (nAccessListIdx & WFS_FLAG_IS_A_DIRECTORY),
                                                 (nObjCreator == titleId));

    // There are 4 patterns.
    // - No entry.
    // - Title entry only.
    // - Title entry and self entry.
    // - Self entry only.
    if (!IS_ENTRY_AVAILABLE(&titleEntry) && !IS_ENTRY_AVAILABLE(&selfEntry))
    {
        // Inherit parent permissions
        pPerm->nFlags          = parentPerm.nFlags;
        pPerm->nEntryCreatorId = parentPerm.nEntryCreatorId;
        pPerm->nEntryCreatorCl = parentPerm.nEntryCreatorCl;
    }
    else if (IS_ENTRY_AVAILABLE(&titleEntry) && !IS_ENTRY_AVAILABLE(&selfEntry))
    {
        // To get entry creator's permissions
        entryCreatorPerm.nTitleId = titleEntry.entryCreatorId;
        nResult = GetInheritedEntry(pEntry, pRxi,
                                    pAreaInfo, nAccessListIdx, nObjCreator,
                                    &entryCreatorPerm,
                                    TRUE,
                                    IS_ENTRY_AVAILABLE(&everyoneEntry) ? (everyoneEntry.entryCreatorId != titleId) : TRUE);

        // To compare entry creator's CL with parent inherited entry creator's CL.
        WFSKRN_PERM_REPORT("Entry Creator %d <-> Parent %d\n", GetAbsCL(entryCreatorPerm.nFlags), parentPerm.nEntryCreatorCl);
        if (GetAbsCL(entryCreatorPerm.nFlags) < parentPerm.nEntryCreatorCl)
        {
            pPerm->nFlags          = parentPerm.nFlags;
            pPerm->nEntryCreatorId = parentPerm.nEntryCreatorId;
            pPerm->nEntryCreatorCl = parentPerm.nEntryCreatorCl;
        }
        else
        {
            // At first, calculate title entry's inherited vaues.
            WFSKRN_PERM_REPORT("Entry Creator %d, Title Entry RelCL %d\n", GetAbsCL(entryCreatorPerm.nFlags), GetRelCL(titleEntry.permissions));
            pPerm->nFlags          = CalcPermissions(entryCreatorPerm.nFlags, titleEntry.permissions);
            pPerm->nEntryCreatorId = titleEntry.entryCreatorId;
            pPerm->nEntryCreatorCl = GetAbsCL(entryCreatorPerm.nFlags);
        }
    }
    else if (IS_ENTRY_AVAILABLE(&titleEntry) && IS_ENTRY_AVAILABLE(&selfEntry))
    {
        // To get entry creator's permissions
        entryCreatorPerm.nTitleId = titleEntry.entryCreatorId;
        nResult = GetInheritedEntry(pEntry, pRxi,
                                    pAreaInfo, nAccessListIdx, nObjCreator,
                                    &entryCreatorPerm,
                                    TRUE,
                                    IS_ENTRY_AVAILABLE(&everyoneEntry) ? (everyoneEntry.entryCreatorId != titleId) : TRUE);

        // To compare entry creator's CL with parent inherited entry creator's CL.
        WFSKRN_PERM_REPORT("Entry Creator %d <-> Parent %d\n", GetAbsCL(entryCreatorPerm.nFlags), parentPerm.nEntryCreatorCl);
        if (GetAbsCL(entryCreatorPerm.nFlags) < parentPerm.nEntryCreatorCl)
        {
            WFSKRN_PERM_REPORT("Entry Creator %d, Self Entry RelCL %d\n", GetAbsCL(parentPerm.nFlags), GetRelCL(selfEntry.permissions));
            pPerm->nFlags          = CalcPermissions(parentPerm.nFlags, selfEntry.permissions);
            pPerm->nEntryCreatorId = parentPerm.nEntryCreatorId;
            pPerm->nEntryCreatorCl = parentPerm.nEntryCreatorCl;
        }
        else
        {
            // At first, calculate title entry's inherited vaues.
            WFSKRN_PERM_REPORT("Entry Creator %d, Title Entry RelCL %d\n", GetAbsCL(entryCreatorPerm.nFlags), GetRelCL(titleEntry.permissions));
            pPerm->nFlags = CalcPermissions(entryCreatorPerm.nFlags, titleEntry.permissions);
            // Next, calculate self entry.
            WFSKRN_PERM_REPORT("Entry Creator %d, Self Entry RelCL %d\n", GetAbsCL(pPerm->nFlags), GetRelCL(selfEntry.permissions));
            pPerm->nFlags = CalcPermissions(pPerm->nFlags, selfEntry.permissions);

            pPerm->nEntryCreatorId = parentPerm.nEntryCreatorId;
            pPerm->nEntryCreatorCl = parentPerm.nEntryCreatorCl;
        }
    }
    else // == if (!IS_ENTRY_AVAILABLE(&titleEntry) && IS_ENTRY_AVAILABLE(&selfEntry))
    {
        WFSKRN_PERM_REPORT("Entry Creator %d, Self Entry RelCL %d\n", GetAbsCL(parentPerm.nFlags), GetRelCL(selfEntry.permissions));
        pPerm->nFlags          = CalcPermissions(parentPerm.nFlags, selfEntry.permissions);
        pPerm->nEntryCreatorId = parentPerm.nEntryCreatorId;
        pPerm->nEntryCreatorCl = parentPerm.nEntryCreatorCl;
    }

    // Skip to calculate everyone's permission at several situations.
    if (!everyone)
    {
        // Skip when specified from caller.
        goto Exit;
    }
    if (WFS_EVERYONE_ID == titleId)
    {
        // Skip when calculate itself.
        goto Exit;
    }
    if (IS_ENTRY_AVAILABLE(&everyoneEntry) && everyoneEntry.entryCreatorId == titleId)
    {
        // Skip when calculate everyone entry's creator's permission.
        goto Exit;
    }

    // to get everyone permissions.
    everyonePerm.nTitleId = WFS_EVERYONE_ID;
    nResult = GetInheritedEntry(pEntry, pRxi,
                                pAreaInfo, nAccessListIdx, nObjCreator,
                                &everyonePerm,
                                TRUE, TRUE);

    // To compare everyone entry creator's CL with current selected entry creator's CL.
    WFSKRN_PERM_REPORT("Everyone %d <-> Inherited %d\n", everyonePerm.nEntryCreatorCl, pPerm->nEntryCreatorCl);
    if (everyonePerm.nEntryCreatorCl < pPerm->nEntryCreatorCl)
    {
        // Everyone entry is disabled. No operation is needed.
    }
    else if (everyonePerm.nEntryCreatorCl == pPerm->nEntryCreatorCl)
    {
        // It is need to OR.
        pPerm->nFlags |= (everyonePerm.nFlags & ~WFS_PERM_CL_MASK);
    }
    else
    {
        // To use everyone entry.
        pPerm->nTitleId        = WFS_EVERYONE_ID;
        pPerm->nFlags          = everyonePerm.nFlags;
        pPerm->nEntryCreatorId = everyonePerm.nEntryCreatorId;
        pPerm->nEntryCreatorCl = everyonePerm.nEntryCreatorCl;
    }

Exit:
    // To mask 
    if (nAccessListIdx & WFS_FLAG_IS_A_DIRECTORY)
    {
        pPerm->nFlags &= (WFS_PERM_CL_MASK | WFS_PERM_DIR_FULL);
    }
    else
    {
        pPerm->nFlags &= (WFS_PERM_CL_MASK | WFS_PERM_FILE_FULL);
    }
    WFSKRN_PERM_REPORT("%02d:%08X(%d)->%08X %08X\n", pRxi->nDepth, pPerm->nEntryCreatorId, pPerm->nEntryCreatorCl, pPerm->nTitleId, pPerm->nFlags);
    return nResult;
}

WFSKrnResult WFSKrnGetInheritedPermissionForUserDi(PathCacheItr* pPci, u32* pFlags, WFSTitleId titleId)
{
    WFSKrnResult nResult = WFSKRN_RESULT_OK;

    AreaInfo* pAreaInfo;
    u32       nAccessListIdx;
    u32       nObjCreator;
    if (pPci->di.pAttrHdr == &wkg.rootDirAttr || // /
        pPci->di.pAttrHdr == &wkg.devDirAttr  || // /dev
        pPci->di.pAttrHdr == &wkg.volDirAttr)    // /vol
    {
        *pFlags = WFS_PERM_LIST;
        nResult = WFSKRN_RESULT_OK;
        goto Exit;
    }
    if (!pPci->di.tba.pAreaInfo)
    {
        // volume root
        if (titleId == WFS_ROOT_ID)
        {
            *pFlags = WFS_PERM_FULL;
            nResult = WFSKRN_RESULT_OK;
        }
        else
        {
            *pFlags = 0;
            nResult = WFSKRN_RESULT_OK;
        }
        goto Exit;
    }
    pAreaInfo      = pPci->di.tba.pAreaInfo;
    nAccessListIdx = pPci->di.pAttrHdr->nAccessListIdx;
    nObjCreator    = pPci->di.pAttrHdr->nObjCreator;

    WFSKRN_PERM_REPORT("access list 0x%08X.\n", nAccessListIdx);

    PathCachePerm perm;
    perm.nTitleId = titleId;
    GetInheritedEntry(pPci->pEntry, &pPci->rxi,
                      pAreaInfo, nAccessListIdx, nObjCreator,
                      &perm,
                      TRUE, TRUE);
    nResult = WFSKRN_RESULT_OK;
    *pFlags = perm.nFlags;
    WFSKRN_PERM_REPORT("get di     %08X(%d)->%08X %08X.\n", perm.nEntryCreatorId, perm.nEntryCreatorCl, titleId, *pFlags);
    nResult = WFSKRN_RESULT_OK;
Exit:
    // ToDo: remapping error code?
    return nResult;
}

WFSKrnResult WFSKrnGetInheritedPermissionForUserPci(PathCacheItr* pPci, u32* pFlags, WFSTitleId titleId)
{
    WFSKrnResult nResult = WFSKRN_RESULT_OK;

    AreaInfo* pAreaInfo;
    u32       nAccessListIdx;
    u32       nObjCreator;
    //WFSKRN_PERM_REPORT("pEntry 0x%08X.\n", pPci->pEntry);
    if (pPci->pEntry->pcd.pData == (void*)0 ||                          // /
        pPci->pEntry->pcd.pData == (void*)WFSKRN_RESULT_SRV_PATH_DEV || // /dev
        pPci->pEntry->pcd.pData == (void*)WFSKRN_RESULT_SRV_PATH_VOL)   // /vol
    {
        *pFlags = WFS_PERM_LIST;
        nResult = WFSKRN_RESULT_OK;
        goto Exit;
    }
    pAreaInfo      = pPci->pEntry->pcd.pAreaInfo;
    if (!pPci->pEntry->pcd.nBlkAdr)
    {
        // Special directory has no nBlkAdr in PathCache
        void *pData = pPci->pEntry->pcd.pData;
        if (((VolumeInfo*)pData >= wkg.aVolInfo) && ((VolumeInfo*)pData < &wkg.aVolInfo[WFSKRN_MAX_VOLUMES]) &&
            !(((u32)pData - (u32)wkg.aVolInfo) % sizeof(VolumeInfo)))
        {
            // Access the root directory of a volume
            pAreaInfo = &((VolumeInfo*)pAreaInfo)->rootAreaInfo;
        }
    }
    nAccessListIdx = pPci->pEntry->pcd.nAccessListIdx;
    nObjCreator    = pPci->pEntry->pcd.nObjCreator;

    WFSKRN_PERM_REPORT("access list 0x%08X.\n", nAccessListIdx);

    RxtItr rxi;
    memcpy(&rxi, &pPci->rxi, sizeof(RxtItr));
    BacktrackToSlash(&rxi);

    PathCachePerm perm;
    perm.nTitleId = titleId;
    GetInheritedEntry(pPci->pEntry, &rxi,
                      pAreaInfo, nAccessListIdx, nObjCreator,
                      &perm,
                      TRUE, TRUE);
    nResult = WFSKRN_RESULT_OK;
    *pFlags = perm.nFlags;
    WFSKRN_PERM_REPORT("get pci    %08X(%d)->%08X %08X.\n", perm.nEntryCreatorId, perm.nEntryCreatorCl, titleId, *pFlags);
    nResult = WFSKRN_RESULT_OK;
Exit:
    // ToDo: remapping error code?
    return nResult;
}

WFSKrnResult WFSKrnGetParentInheritedPermissionForUserPci(PathCacheItr* pPci, u32* pFlags, WFSTitleId titleId)
{
    WFSKrnResult nResult = WFSKRN_RESULT_OK;

    RxtItr parentRxi;
    memcpy(&parentRxi, &pPci->rxi, sizeof(RxtItr));
    PathCacheEntry* pParentEntry = GetParentPathCacheEntry(pPci->pEntry, &parentRxi);

    WFSKRN_PERM_REPORT("parent pEntry 0x%08X.\n", pParentEntry);
    AreaInfo* pAreaInfo;
    u32       nAccessListIdx;
    u32       nObjCreator;
    if (pParentEntry->pcd.pData == (void*)0 ||                          // /
        pParentEntry->pcd.pData == (void*)WFSKRN_RESULT_SRV_PATH_DEV || // /dev
        pParentEntry->pcd.pData == (void*)WFSKRN_RESULT_SRV_PATH_VOL)   // /vol
    {
        *pFlags = WFS_PERM_LIST;
        nResult = WFSKRN_RESULT_OK;
        goto Exit;
    }

    pAreaInfo      = pParentEntry->pcd.pAreaInfo;
    if (!pParentEntry->pcd.nBlkAdr)
    {
        // Special directory has no nBlkAdr in PathCache
        void *pData = pParentEntry->pcd.pData;
        if (((VolumeInfo*)pData >= wkg.aVolInfo) && ((VolumeInfo*)pData < &wkg.aVolInfo[WFSKRN_MAX_VOLUMES]) &&
            !(((u32)pData - (u32)wkg.aVolInfo) % sizeof(VolumeInfo)))
        {
            // Access the root directory of a volume
            pAreaInfo = &((VolumeInfo*)pAreaInfo)->rootAreaInfo;
        }
    }
    nAccessListIdx = pParentEntry->pcd.nAccessListIdx;
    nObjCreator    = pParentEntry->pcd.nObjCreator;

    WFSKRN_PERM_REPORT("parent access list 0x%08X.\n", nAccessListIdx);

    PathCachePerm perm;
    perm.nTitleId = titleId;
    GetInheritedEntry(pParentEntry, &parentRxi,
                      pAreaInfo, nAccessListIdx, nObjCreator,
                      &perm,
                      TRUE, TRUE);
    nResult = WFSKRN_RESULT_OK;
    *pFlags = perm.nFlags;
    WFSKRN_PERM_REPORT("parent :%08X(%d)->%08X %08X.\n", perm.nEntryCreatorId, perm.nEntryCreatorCl, titleId, *pFlags);

Exit:
    // ToDo: remapping error code?
    return nResult;
}

WFSKrnResult WFSKrnSetInheritedPermissionForUser(PathCacheItr* pPci, u32 nFlags, u32 nMask, WFSTitleId entryCreatorId, WFSTitleId titleId)
{
    WFSKrnResult nResult = WFSKRN_RESULT_OK;
    u32 mask;

    WFSKRN_PERM_REPORT("WFSKrnSetInheritedPermissionForUser start\n");

    mask = nMask & ((pPci->di.pAttrHdr->nAccessListIdx & WFS_FLAG_IS_A_DIRECTORY) ? (WFS_PERM_CL_MASK | WFS_PERM_DIR_FULL) : (WFS_PERM_CL_MASK | WFS_PERM_FILE_FULL));

    if (titleId == WFS_ROOT_ID)
    {
        if (entryCreatorId != WFS_ROOT_ID)
        {
            // invalid
            nResult = WFSKRN_RESULT_PERMISSION; // INVALID?
        }
        if (((nFlags & mask) & WFS_PERM_CL_MASK) == 0 && ((nFlags & mask) & WFS_PERM_CL_3) == 0 && ((nFlags & mask) & WFS_PERM_CL_MINUS_0) == 0 )
        {
            nResult = WFSKRN_RESULT_PERMISSION;
        }
        if ((WFS_PERM_FULL & mask & ~WFS_PERM_CL_MASK) != (nFlags & mask & ~WFS_PERM_CL_MASK))
        {
            nResult = WFSKRN_RESULT_PERMISSION;
        }
        goto Exit;
    }

    // to get access list entry
    WFSAccessListEntry  aAclEntry[ACL_MAX_ENTRIES]; // 12*12=144
    WFSAccessListEntry* pTitleEntry    = NULL;
    WFSAccessListEntry* pSelfEntry     = NULL;
    WFSAccessListEntry* pEveryoneEntry = NULL;

    nResult = (WFSKrnResult)WFSKrnGetAccessList(pPci->di.tba.pAreaInfo, pPci->di.pAttrHdr->nAccessListIdx, aAclEntry, ACL_MAX_ENTRIES);
    if (nResult < 0)
    {
        goto Exit;
    }
    while (nResult)
    {
        if (aAclEntry[nResult-1].titleId == titleId)
        {
            if (aAclEntry[nResult-1].titleId == aAclEntry[nResult-1].entryCreatorId)
                pSelfEntry  = &aAclEntry[nResult-1];
            else
                pTitleEntry = &aAclEntry[nResult-1];
        }
        if (aAclEntry[nResult-1].titleId == WFS_EVERYONE_ID)
        {
            pEveryoneEntry = &aAclEntry[nResult-1];
        }
        nResult = (WFSKrnResult)((s32)nResult - 1);
    }

    AreaInfo* pAreaInfo;

    u32             newFlags;
    RxtItr          parentRxi;
    PathCacheEntry* pParentEntry;
    AreaInfo*       pParentAreaInfo;

    PathCachePerm   entryCreatorPerm;
    PathCachePerm   titleCreatorPerm;
    PathCachePerm   everyonePerm;
    PathCachePerm   parentPerm;
    PathCachePerm   currentPerm;

    memcpy(&parentRxi, &pPci->rxi, sizeof(RxtItr));
    pParentEntry = GetParentPathCacheEntry(pPci->pEntry, &parentRxi);

    if (pPci->di.pAttrHdr->nAccessListIdx & WFS_FLAG_IS_A_DIRECTORY)
    {
        pAreaInfo = pPci->di.tba.pAreaInfo;
    }
    else
    {
        pAreaInfo = pParentEntry->pcd.pAreaInfo;
    }

    if (((VolumeInfo*)pAreaInfo >= wkg.aVolInfo) && ((VolumeInfo*)pAreaInfo < &wkg.aVolInfo[WFSKRN_MAX_VOLUMES]) &&
        !(((u32)pAreaInfo - (u32)wkg.aVolInfo) % sizeof(VolumeInfo)))
    {
        // Access the root directory of a volume
        pAreaInfo = &((VolumeInfo*)pAreaInfo)->rootAreaInfo;
    }

    entryCreatorPerm.nTitleId = entryCreatorId;
    GetInheritedEntry(pPci->pEntry, &pPci->rxi,
                      pAreaInfo, pPci->di.pAttrHdr->nAccessListIdx, pPci->di.pAttrHdr->nObjCreator,
                      &entryCreatorPerm,
                      (entryCreatorId == titleId) ? FALSE : TRUE, TRUE);
    WFSKRN_PERM_REPORT("creator:%08X(%d)->%08X %08X %08X.\n",
                       entryCreatorPerm.nEntryCreatorId, entryCreatorPerm.nEntryCreatorCl,
                       entryCreatorPerm.nTitleId, entryCreatorPerm.nFlags,
                       pPci->di.pAttrHdr->nAccessListIdx);

    pParentAreaInfo = pParentEntry->pcd.pAreaInfo;
    if (!pParentEntry->pcd.nBlkAdr)
    {
        // Special directory has no nBlkAdr in PathCache
        void *pData = pParentEntry->pcd.pData;
        if (((VolumeInfo*)pData >= wkg.aVolInfo) && ((VolumeInfo*)pData < &wkg.aVolInfo[WFSKRN_MAX_VOLUMES]) &&
            !(((u32)pData - (u32)wkg.aVolInfo) % sizeof(VolumeInfo)))
        {
            // Access the root directory of a volume
            pParentAreaInfo = &((VolumeInfo*)pParentAreaInfo)->rootAreaInfo;
        }
    }
    parentPerm.nTitleId = titleId;
    if (pParentAreaInfo == (void*)WFSKRN_RESULT_SRV_PATH_DEV || // /dev
        pParentAreaInfo == (void*)WFSKRN_RESULT_SRV_PATH_VOL)   // /vol
    {
        parentPerm.nFlags          = 0;
        parentPerm.nEntryCreatorId = WFS_ROOT_ID;
        parentPerm.nEntryCreatorCl = 0;
        nResult = WFSKRN_RESULT_OK;
    }
    else
    {
        GetInheritedEntry(pParentEntry, &parentRxi,
                          pParentAreaInfo, pParentEntry->pcd.nAccessListIdx, pParentEntry->pcd.nObjCreator,
                          &parentPerm,
                          TRUE, TRUE);
    }
    WFSKRN_PERM_REPORT("parent :%08X(%d)->%08X %08X %08X.\n",
                       parentPerm.nEntryCreatorId, parentPerm.nEntryCreatorCl,
                       parentPerm.nTitleId, parentPerm.nFlags,
                       pParentEntry->pcd.nAccessListIdx);

    if (pTitleEntry)
    {
        titleCreatorPerm.nTitleId = pTitleEntry->entryCreatorId;
        GetInheritedEntry(pPci->pEntry, &pPci->rxi,
                          pAreaInfo, pPci->di.pAttrHdr->nAccessListIdx, pPci->di.pAttrHdr->nObjCreator,
                          &titleCreatorPerm,
                          TRUE, pEveryoneEntry ? (pEveryoneEntry->entryCreatorId != titleId) : TRUE);
        WFSKRN_PERM_REPORT("title  :%08X(%d)->%08X %08X %08X.\n",
                           titleCreatorPerm.nEntryCreatorId, titleCreatorPerm.nEntryCreatorCl,
                           titleCreatorPerm.nTitleId, titleCreatorPerm.nFlags,
                           pPci->di.pAttrHdr->nAccessListIdx);
    }

    // wild card
    if (nFlags == WFS_PERM_FULL && nMask == WFS_PERM_FULL)
    {
        if (titleId == entryCreatorId)
        {
            nFlags = WFS_PERM_CL_MINUS_0 | (entryCreatorPerm.nFlags & ~WFS_PERM_CL_MASK);
        }
        else
        {
            nFlags = WFS_PERM_CL_MINUS_1 | (entryCreatorPerm.nFlags & ~WFS_PERM_CL_MASK);
        }
    }

    // calculate object creator's privilege and WFS_PERM_DELETE_OR_MOVE
    if (pPci->di.pAttrHdr->nAccessListIdx & WFS_FLAG_IS_A_DIRECTORY)
    {
        if (parentPerm.nFlags & WFS_PERM_ADD_OR_DELETE_SUBDIR)
        {
            if (pPci->di.pAttrHdr->nObjCreator == titleId)
                parentPerm.nFlags |= WFS_PERM_DIR_FULL;
        }
    }
    else
    {
        if (parentPerm.nFlags & WFS_PERM_ADD_OR_DELETE_FILE)
        {
            if (pPci->di.pAttrHdr->nObjCreator == titleId)
                parentPerm.nFlags |= WFS_PERM_FILE_FULL;
        }
    }

    currentPerm.nTitleId = titleId;
    if (pTitleEntry)
    {
        WFSKRN_PERM_REPORT("Entry Creator %d <-> Parent %d\n", GetAbsCL(titleCreatorPerm.nFlags), parentPerm.nEntryCreatorCl);
        if (GetAbsCL(titleCreatorPerm.nFlags) < parentPerm.nEntryCreatorCl)
        {
            if (pSelfEntry)
            {
                if (((s32)GetAbsCL(parentPerm.nFlags) + GetRelCL(pSelfEntry->permissions)) < 0)
                {
                    // if CL is a negative value, title have no permission.
                    currentPerm.nFlags = 0;
                }
                else
                {
                    currentPerm.nFlags =
                        GetAbsCLFlag(GetAbsCL(parentPerm.nFlags) + GetRelCL(pSelfEntry->permissions)) |
                        ((pSelfEntry->permissions & ~WFS_PERM_CL_MASK) & (parentPerm.nFlags & ~WFS_PERM_CL_MASK));
                }
            }
            else
            {
                currentPerm.nFlags = parentPerm.nFlags;
            }
            currentPerm.nEntryCreatorId = parentPerm.nEntryCreatorId;
            currentPerm.nEntryCreatorCl = parentPerm.nEntryCreatorCl;
        }
        else
        {
            if (((s32)GetAbsCL(titleCreatorPerm.nFlags) + GetRelCL(pTitleEntry->permissions)) < 0)
            {
                // if CL is a negative value, title have no permission.
                currentPerm.nFlags = 0;
            }
            else
            {
                // At first, culclate title entry's inherited values.
                currentPerm.nFlags =
                    GetAbsCLFlag(GetAbsCL(titleCreatorPerm.nFlags) + GetRelCL(pTitleEntry->permissions)) |
                    ((pTitleEntry->permissions & ~WFS_PERM_CL_MASK) & (titleCreatorPerm.nFlags & ~WFS_PERM_CL_MASK));
            }
            if (pSelfEntry)
            {
                if (((s32)GetAbsCL(currentPerm.nFlags) + GetRelCL(pSelfEntry->permissions)) < 0)
                {
                    // if CL is a negative value, title have no permission.
                    currentPerm.nFlags = 0;
                }
                else
                {
                    // Next, culclate self entry.
                    currentPerm.nFlags =
                        GetAbsCLFlag(GetAbsCL(currentPerm.nFlags) + GetRelCL(pSelfEntry->permissions)) |
                        ((pSelfEntry->permissions & ~WFS_PERM_CL_MASK) & (currentPerm.nFlags & ~WFS_PERM_CL_MASK));
                }
            }
            currentPerm.nEntryCreatorId = titleCreatorPerm.nTitleId;
            currentPerm.nEntryCreatorCl = GetAbsCL(titleCreatorPerm.nFlags);
        }
    }
    else
    {
        if (pSelfEntry)
        {
            if (((s32)GetAbsCL(parentPerm.nFlags) + GetRelCL(pSelfEntry->permissions)) < 0)
            {
                // if CL is a negative value, title have no permission.
                currentPerm.nFlags = 0;
            }
            else
            {
                currentPerm.nFlags =
                    GetAbsCLFlag(GetAbsCL(parentPerm.nFlags) + GetRelCL(pSelfEntry->permissions)) |
                    ((pSelfEntry->permissions & ~WFS_PERM_CL_MASK) & (parentPerm.nFlags & ~WFS_PERM_CL_MASK));
            }
        }
        else
        {
            currentPerm.nFlags = parentPerm.nFlags;
        }
        currentPerm.nEntryCreatorId = parentPerm.nEntryCreatorId;
        currentPerm.nEntryCreatorCl = parentPerm.nEntryCreatorCl;
    }

    // to calculate everyone permissions.
    if (WFS_EVERYONE_ID != titleId)
    {
        everyonePerm.nTitleId = WFS_EVERYONE_ID;
        nResult = GetInheritedEntry(pPci->pEntry, &pPci->rxi,
                                    pAreaInfo, pPci->di.pAttrHdr->nAccessListIdx, pPci->di.pAttrHdr->nObjCreator,
                                    &everyonePerm,
                                    TRUE, TRUE);
        WFSKRN_PERM_REPORT("everyone :%08X(%d)->%08X %08X %08X.\n",
                           everyonePerm.nEntryCreatorId, everyonePerm.nEntryCreatorCl,
                           everyonePerm.nTitleId, everyonePerm.nFlags,
                           pPci->di.pAttrHdr->nAccessListIdx);

        WFSKRN_PERM_REPORT("Everyone %d <-> Inherited %d\n", everyonePerm.nEntryCreatorCl, currentPerm.nEntryCreatorCl);
        if (everyonePerm.nEntryCreatorCl < currentPerm.nEntryCreatorCl)
        {
            // no need to update.
        }
        else if (everyonePerm.nEntryCreatorCl == currentPerm.nEntryCreatorCl)
        {
            // ORed
            currentPerm.nFlags |= (everyonePerm.nFlags & ~WFS_PERM_CL_MASK);
        }
        else
        {
            // to select everyone entry.
            currentPerm.nTitleId        = WFS_EVERYONE_ID;
            currentPerm.nFlags          = everyonePerm.nFlags;
            currentPerm.nEntryCreatorId = everyonePerm.nEntryCreatorId;
            currentPerm.nEntryCreatorCl = everyonePerm.nEntryCreatorCl;
        }
    }


    if (mask & WFS_PERM_CL_MASK)
    {
        if (GetRelCL(nFlags) <= 0)
        {
            if (((s32)GetAbsCL(entryCreatorPerm.nFlags) + GetRelCL(nFlags)) < 0)
            {
                WFSKRN_PERM_REPORT("caller doesn't have permission to set entry.\n");
                nResult = WFSKRN_RESULT_PERMISSION_CL;
                goto Exit;
            }
            newFlags = (currentPerm.nFlags & ~WFS_PERM_CL_MASK & ~mask) | (nFlags & ~WFS_PERM_CL_MASK & mask) | GetAbsCLFlag((s32)GetAbsCL(entryCreatorPerm.nFlags) + GetRelCL(nFlags));
            nFlags   = (currentPerm.nFlags & ~WFS_PERM_CL_MASK & ~mask) | (nFlags & ~WFS_PERM_CL_MASK & mask) | (nFlags & WFS_PERM_CL_MASK);
        }
        else if (GetAbsCL(nFlags) >= 0)
        {
            if ((GetAbsCL(nFlags) - (s32)GetAbsCL(entryCreatorPerm.nFlags)) > 0)
            {
                WFSKRN_PERM_REPORT("caller doesn't have permission to set entry.\n");
                nResult = WFSKRN_RESULT_PERMISSION_CL;
                goto Exit;
            }
            newFlags = (currentPerm.nFlags & ~WFS_PERM_CL_MASK & ~mask) | (nFlags & ~WFS_PERM_CL_MASK & mask) | (nFlags & WFS_PERM_CL_MASK);
            nFlags   = (currentPerm.nFlags & ~WFS_PERM_CL_MASK & ~mask) | (nFlags & ~WFS_PERM_CL_MASK & mask) | GetRelCLFlag(GetAbsCL(nFlags) - (s32)GetAbsCL(entryCreatorPerm.nFlags));
        }
        else
        {
            WFSKRN_PERM_REPORT("invalid control level argument.\n");
            nResult = WFSKRN_RESULT_INVALID;
            goto Exit;
        }
    }
    else
    {
        if ((GetAbsCL(currentPerm.nFlags) - (s32)GetAbsCL(entryCreatorPerm.nFlags)) > 0)
        {
            WFSKRN_PERM_REPORT("caller doesn't have permission to set entry.\n");
            nResult = WFSKRN_RESULT_PERMISSION_CL;
            goto Exit;
        }
        newFlags = (currentPerm.nFlags & ~mask) | (nFlags & mask);
        nFlags   = (currentPerm.nFlags & ~WFS_PERM_CL_MASK & ~mask) | (nFlags & mask) | GetRelCLFlag(GetAbsCL(currentPerm.nFlags) - (s32)GetAbsCL(entryCreatorPerm.nFlags));
    }

    WFSKRN_PERM_REPORT("current:%08X(%d)->%08X %08X %08X.\n",
                       currentPerm.nEntryCreatorId, currentPerm.nEntryCreatorCl,
                       currentPerm.nTitleId, currentPerm.nFlags,
                       pPci->di.pAttrHdr->nAccessListIdx);

    if ((newFlags != currentPerm.nFlags) ||
       (currentPerm.nEntryCreatorId == WFS_ROOT_ID && currentPerm.nEntryCreatorCl == 0)) // effective entry has not created yet.
    {
        if (titleId == entryCreatorId)
        {
            // cannot create self entry which has minused CL.
            if (GetRelCLFlag(nFlags) < 0)
            {
                WFSKRN_PERM_REPORT("cannot create self entry which has minused CL.\n");
                //nResult = WFSKRN_RESULT_INVALID;
                //goto Exit;
            }
            // Does creator have enough permissions to set entry?
            if ((nFlags & ~WFS_PERM_CL_MASK & mask) & ~(entryCreatorPerm.nFlags & ~WFS_PERM_CL_MASK))
            {
                WFSKRN_PERM_REPORT("caller doesn't have permission to set entry.\n");
                nResult = WFSKRN_RESULT_PERMISSION;
                goto Exit;
            }
        }
        else
        {
            if (GetAbsCL(entryCreatorPerm.nFlags) == 0)
            {
                WFSKRN_PERM_REPORT("caller has CL 0.\n");
                nResult = WFSKRN_RESULT_PERMISSION_CL;
                goto Exit;
            }
            // Does creator have enought CL to set entry?
            if (GetAbsCL(entryCreatorPerm.nFlags) < currentPerm.nEntryCreatorCl)
            {
                WFSKRN_PERM_REPORT("caller doesn't have enough cl to set entry.\n");
                nResult = WFSKRN_RESULT_PERMISSION_CL;
                goto Exit;
            }
            // Does creator have enough permissions to set entry?
            if ((nFlags & ~WFS_PERM_CL_MASK & mask) & ~(entryCreatorPerm.nFlags & ~WFS_PERM_CL_MASK))
            {
                WFSKRN_PERM_REPORT("caller doesn't have permission to set entry.\n");
                nResult = WFSKRN_RESULT_PERMISSION;
                goto Exit;
            }
        }

        WFSKRN_PERM_REPORT("change :%08X->%08X\n", currentPerm.nFlags, newFlags);

        // cannot delete other created entry.
        if (pTitleEntry && (titleId != entryCreatorId))
        {
            WFSKrnDeletePermissions(pAreaInfo, &(pPci->di.pAttrHdr->nAccessListIdx),
                                    pTitleEntry->entryCreatorId, pTitleEntry->titleId);
            WFSKRN_PERM_REPORT("entry %08X(X)->%08X %08X is deleted. %08X.\n",
                               pTitleEntry->entryCreatorId, pTitleEntry->titleId, pTitleEntry->permissions,
                               pPci->di.pAttrHdr->nAccessListIdx);
        }

        if (pSelfEntry)
        {
            WFSKrnDeletePermissions(pAreaInfo, &(pPci->di.pAttrHdr->nAccessListIdx),
                                    pSelfEntry->entryCreatorId, pSelfEntry->titleId);
            WFSKRN_PERM_REPORT("entry %08X(X)->%08X %08X is deleted. %08X.\n",
                               pSelfEntry->entryCreatorId, pSelfEntry->titleId, pSelfEntry->permissions,
                               pPci->di.pAttrHdr->nAccessListIdx);
        }

        // if already set new permissions.
        if (newFlags == parentPerm.nFlags &&
           !(parentPerm.nEntryCreatorId == WFS_ROOT_ID && parentPerm.nEntryCreatorCl == 0)) // effective entry has already created.
        {
            WFSKRN_PERM_REPORT("no need to create new entry.\n");
            nResult = WFSKRN_RESULT_OK;
            goto Exit;
        }

        WFSKrnExitOnError(WFSKrnSetAccessList(pAreaInfo, &(pPci->di.pAttrHdr->nAccessListIdx), nFlags,
                                              mask, entryCreatorId, titleId));
        WFSKRN_PERM_REPORT("new entry %08X(%d)->%08X %08X is created. %08X.\n",
                entryCreatorId, GetAbsCL(entryCreatorPerm.nFlags), titleId, nFlags, pPci->di.pAttrHdr->nAccessListIdx);
    }
    else
    {
        WFSKRN_PERM_REPORT("no need to create new entry.\n");
        nResult = WFSKRN_RESULT_OK;
    }

Exit:
    WFSKRN_PERM_REPORT("WFSKrnSetInheritedPermissionForUser end\n");
    // ToDo: remapping error code?
    return nResult;
}

#endif
