/*---------------------------------------------------------------------------*
  Project:  wfsKrn
  File:     wfsKrn_Permission_Iop.cpp - Functions for permissions.

  Copyright 2008 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_Permission_Iop.cpp,v $
  Revision 1.27  2008/10/10 04:05:05  kondo_masahiro
  Fixed change arguments of AreaFreeBlks

  Revision 1.26  2008/10/08 23:19:54  kondo_masahiro
  Added codes to access large size file data

  Revision 1.25  2008/09/24 07:33:21  ueno
  Added AclHdr{} to AreaHdr{} to save configuration of accesslists in a disk.

  Revision 1.24  2008/08/27 23:19:26  paul
  Added nStride to AreaAllocBlks, AreaFreeBlks

  Revision 1.23  2008/08/05 03:56:24  kondo_masahiro
  Added error handling for hash error and detach device.

  Revision 1.22  2008/08/04 04:02:27  ueno
  Modified to use TransInfo.

  Revision 1.21  2008/07/25 16:12:49  paul
  Changed nTransIdx to *pTransInfo.  Changed WFSTransBlkAdr to TransBlkAdr

  Revision 1.20  2008/07/23 04:17:14  ueno
  Reduced memory size required for accesslist.

  Revision 1.19  2008/07/23 02:01:09  ueno
  Modified to call TransGetBlk() to set read/write flag before writing or deleting a file.
  Modified to call TransUnpinBlk() instead of TransUnpinBlk3().

  Revision 1.18  2008/07/22 04:10:44  ueno
  Modified to call TransUnpinBlk3() after DirCreateAndInitBewBlk().

  Revision 1.17  2008/07/21 23:24:48  ueno
  Added debug code.

  Revision 1.16  2008/07/14 00:43:51  ueno
  Revised directory initialization.

  Revision 1.15  2008/06/05 14:05:22  ueno
  Modified to reserve accesslist index 0 as the default (empty) accesslist.

  Revision 1.14  2008/05/30 08:50:07  ueno
  Cleanup.

  Revision 1.13  2008/05/28 02:12:49  ueno
  Cleanup.

  Revision 1.12  2008/05/28 01:09:09  ueno
  Added AclCheckDisk() to check acl meta-data.

  Revision 1.11  2008/05/23 06:32:08  ueno
  Cleanup.

  Revision 1.10  2008/05/22 08:54:59  ueno
  Fixed debug routine.

  Revision 1.9  2008/05/22 08:32:23  ueno
  Fixed assert.

  Revision 1.8  2008/05/22 07:57:25  ueno
  Fixed debug flag.

  Revision 1.7  2008/05/22 06:36:12  ueno
  Ported permission APIs to IOP.

  Revision 1.6  2008/05/20 07:08:38  ueno
  Added debug code.

  Revision 1.5  2008/05/19 10:11:52  ueno
  Revised not to use hardlink.

  Revision 1.4  2008/05/19 05:30:50  ueno
  Defined accesslist meta-data types.

  Revision 1.3  2008/05/16 11:16:17  ueno
  Implemented permission APIs with directory module.

  Revision 1.2  2008/05/15 13:31:46  ueno
  Implemented initialization using directory module.

  Revision 1.1  2008/05/06 07:38:40  ueno
  Initial Check-in.

*---------------------------------------------------------------------------*/
#ifndef WFSDEV

#include "wfs_Defs.h"
#include "wfskrn_Permission_AccessList.h"
#include "wfskrn_Permission_Iop.h"
#include "wfskrn_Permission_File.h"

#include <string.h>

#define FILEHANDLE_DI(fh)   ((fh)->di)
#define FILEHANDLE_TYPE(fh) ((fh)->type)
#define FILEHANDLE_NAME(fh) (&(FILEHANDLE_DI(fh).name))

static void AclDirIterOpen(const AreaInfo* area, DirItr* file, u32 type, TransInfo* pTransInfo)
{
    WFSBlkAdr parentBlock = 0;
    switch (type)
    {
      case ACL_TYPE_INDEXLIST:
      case ACL_TYPE_NAMEFILE:
        parentBlock = NameBlock(area);
        break;
      case ACL_TYPE_ACLFILE:
        parentBlock = AclBlock(area);
        break;
      default:
        WFSKRN_ACL_REPORT("AclDirIterOpen(): invalid type (%u).\n", type);
        WFSKRN_ACL_ASSERT(0);
        return;
    }
    file->tba.pAreaInfo = (AreaInfo*) area; // area
    file->tba.nBlkAdr = parentBlock;        // parent directory
    file->tba.pTransInfo = pTransInfo;
    file->nBlkDepth = 0;                    // search from the root block.
    file->ppDnsRootParent = 0;
}

WFSResult SetFileName(WFSFileName *pFileName, const utf8 *sFileName) {
    WFSKRN_ACL_ASSERT(sFileName);
    u32 nStrLen = STRNLEN(sFileName, WFS_MAX_FILE_NAME_SIZE);
    WFSKRN_ACL_ASSERT(nStrLen < WFS_MAX_FILE_NAME_SIZE);
    pFileName->nLen = (WFSFileNameLengthType) nStrLen;
    memmove(pFileName->sStr, sFileName, nStrLen+1);
    return WFS_RESULT_OK;
}

static inline WFSResult SetFileNameLen(WFSFileName *pFileName) {
    u32 nStrLen = STRNLEN(pFileName->sStr, WFS_MAX_FILE_NAME_SIZE);
    WFSKRN_ACL_ASSERT(nStrLen < WFS_MAX_FILE_NAME_SIZE);
    pFileName->nLen = (WFSFileNameLengthType) nStrLen;
    return WFS_RESULT_OK;
}

//
// IOP Utilities
//

static WFSKrnResult AclDirFindRaw(DirItr* pDi)
{
    return DirFind(&pDi->name, pDi, DIR_MAX_BLK_DEPTH, DIR_FIND_MODE_NORMAL);
}

static WFSResult CreateFile(const AreaInfo* area, const AclFileInfo* fileInfo, WFSAclFileHandle fh)
{
    WFSKrnResult result;
    TransInfo* pTransInfo = 0; // [check] pTransInfo must be passed from WFSSrv APIs as an argument. NYI.

    AclDirIterOpen(area, &FILEHANDLE_DI(fh), FILEHANDLE_TYPE(fh), pTransInfo);

    result = DirInsertRaw(FILEHANDLE_NAME(fh), &FILEHANDLE_DI(fh), 256 + 8); // [check] size
    if (result != WFSKRN_RESULT_ALREADY_EXISTS &&
        result != WFSKRN_RESULT_OK)
    {
        return WFS_RESULT_INVALID;
    }

    if (result == WFSKRN_RESULT_OK)
    {
        FILEHANDLE_DI(fh).pAttrHdr->file.nSize = 0;

#ifdef _DEBUG
        switch (fileInfo->type)
        {
          case ACL_TYPE_NAMEFILE:
          case ACL_TYPE_INDEXLIST:
            // must not be overwritten.
            WFSKRN_ACL_ASSERT(AclBlock(area) != FILEHANDLE_DI(fh).tba.nBlkAdr);
            break;
          case ACL_TYPE_ACLFILE:
            WFSKRN_ACL_ASSERT(NameBlock(area) != FILEHANDLE_DI(fh).tba.nBlkAdr);
            break;
       }
#endif // _DEBUG
    }

    return WFS_RESULT_OK;
}

static WFSResult OpenFile(const AreaInfo* area, WFSAclFileHandle fh)
{
    WFSKrnResult result;
    TransInfo* pTransInfo = 0; // [check] pTransInfo must be passed from WFSSrv APIs as an argument. NYI.

    AclDirIterOpen(area, &FILEHANDLE_DI(fh), FILEHANDLE_TYPE(fh), pTransInfo);

    result = AclDirFindRaw(&FILEHANDLE_DI(fh));
    if (result != WFSKRN_RESULT_DIR_ENTRY_FOUND)
    {
        WFSKRN_ACL_ASSERT(result == WFSKRN_RESULT_DIR_NODE_CHOICE_NOT_FOUND ||
                          result == WFSKRN_RESULT_DIR_NODE_STRING_MISMATCH);
        return WFS_RESULT_NOT_FOUND;
    }
    return WFS_RESULT_OK;
}

static WFSResult CloseFile(WFSAclFileHandle fh)
{
    DirItrClose(&FILEHANDLE_DI(fh)); // [check] error handling.
    return WFS_RESULT_OK;
}

static WFSResult DeleteFile(const AreaInfo* area, WFSAclFileHandle fh)
{
    WFSResult result;
    result = OpenFile(area, fh);
    if (result == WFS_RESULT_OK)
    {
        TransGetBlk(&FILEHANDLE_DI(fh).tba, TRANS_FLAG_RW_BLK, 0); // set read/write before delete. // [check] error handling.
        result = (WFSResult) DirDeleteEntry(&FILEHANDLE_DI(fh));
    }
    CloseFile(fh);

    return result;
}

static int GetNamefile(const AreaInfo* area, WFSKrnAclHandle handle, AclFileInfo* fileInfo, WFSAclFileHandle* pFh)
{
    int ret;
    fileInfo->type = ACL_TYPE_NAMEFILE;
    ret = osSNPrintf(ACL_FILEINFO_NAME(fileInfo), ACL_NAMEFILE_LEN, "%08x", handle);
    if (ret <= 0 || ACL_NAMEFILE_LEN < ret)
    {
        return -1;
    }

    // set file handle.
    *pFh = fileInfo;
    FILEHANDLE_TYPE(*pFh) = ACL_TYPE_NAMEFILE;
    return 0;
}

int CreateNamefile(const AreaInfo* area, WFSKrnAclHandle handle, const AclNameFileHeader* header, const utf8* filename)
{
    WFSAclFileHandle fh;
    AclFileInfo aclFile;

    if (GetNamefile(area, handle, &aclFile, &fh) < 0)
    {
        return -1;
    }

    if (AclCreateFile(area, &aclFile, (WFSBool) TRUE, &fh) != WFS_RESULT_OK)
    {
        return -1;
    }

    if (AclWriteFile(area, fh, header, sizeof(AclNameFileHeader), 0) != WFS_RESULT_OK)
    {
#ifdef WIN32
        WFSKRN_ACL_REPORT("CreateNamefile(%d) Error %ld\n", handle, GetLastError());
#endif // WIN32
        return -1;
    }

    if (AclWriteFile(area, fh, filename, STRNLEN(filename, WFS_MAX_FILE_NAME_SIZE)+1, sizeof(AclNameFileHeader)) != WFS_RESULT_OK)
    {
#ifdef WIN32
        WFSKRN_ACL_REPORT("CreateNamefile(%d) Error %ld\n", handle, GetLastError());
#endif // WIN32
        return -1;
    }
    return 0;
}

int ReadNamefile(const AreaInfo* area, WFSKrnAclHandle handle, void* buf, WFSFileSize size, WFSFileSize offset)
{
    WFSAclFileHandle fh;
    AclFileInfo aclFile;

    if ((handle & ACL_TYPE_MASK) == 0)
    {
        handle = 0; // handle 0 represents both file 0 and directory 0x80000000.
    }
    if (GetNamefile(area, handle, &aclFile, &fh) < 0)
    {
        return -1;
    }
    SetFileNameLen(FILEHANDLE_NAME(fh));
    if (AclReadFile(area, fh, buf, size, offset) < 0)
    {
        return -1;
    }
    return 0;
}

int WriteNamefile(const AreaInfo* area, WFSKrnAclHandle handle, const void* buf, WFSFileSize size, WFSFileSize offset)
{
    WFSAclFileHandle fh;
    AclFileInfo aclFile;
    if (GetNamefile(area, handle, &aclFile, &fh) < 0)
    {
        return -1;
    }
    SetFileNameLen(FILEHANDLE_NAME(fh));
    if (AclWriteFile(area, fh, buf, size, offset) < 0)
    {
        return -1;
    }
    return 0;
}

int DeleteNamefile(const AreaInfo* area, WFSKrnAclHandle handle)
{
    WFSAclFileHandle fh;
    AclFileInfo aclFile;
    if (GetNamefile(area, handle, &aclFile, &fh) < 0)
    {
        return -1;
    }
    SetFileNameLen(FILEHANDLE_NAME(fh));
    if (DeleteFile(area, fh) != WFS_RESULT_OK)
    {
#ifdef WIN32
        WFSKRN_ACL_REPORT("DeleteNamefile(%d) Error %ld\n", handle, GetLastError());
#endif // WIN32
        return -1;
    }
    return 0;
}

/*
 * WFSKrnAccessListDirectoryInit() allocates blocks, which contain
 * an indexlist, accesslist files and accesslist name files (<handle>).
 */

int WFSKrnAccessListDirectoryInit(AreaInfo* area)
{
    DirBlkHdr* pDbh;
    TransBlkAdr tba;
    WFSKrnResult ret;
    TransInfo* pTransInfo = 0; // [check] transInfo muset be passed from WFSSrvAPIs. NYI.

    // [check] how to check if acl and name blocks already exist in the area?

    // allocate a disk block for acl directory.
    ret = AreaAllocBlks((AreaInfo*) area, 0, area->ah.nLog2BlkSize, 1, &AclBlock(area), sizeof(WFSBlkAdr));
    if (ret != WFSKRN_RESULT_OK)
    {
        return -1;
    }

    // allocate a disk block for name directory.
    ret = AreaAllocBlks((AreaInfo*) area, 0, area->ah.nLog2BlkSize, 1, &NameBlock(area), sizeof(WFSBlkAdr));
    if (ret != WFSKRN_RESULT_OK)
    {
        AreaFreeBlks((AreaInfo*) area, area->ah.nLog2BlkSize, 1, &AclBlock(area), sizeof(WFSBlkAdr));
        return -1;
    }

    // initialize acl directory.
    tba.pAreaInfo = (AreaInfo*) area;
    tba.nBlkAdr = AclBlock(area);
    tba.pTransInfo = pTransInfo;

    WFSKrnReturnOnError( DirCreateAndInitNewBlk(&tba, (WFSMetaDataFlags)(WFS_MDF_DIR_LEAF_BLK | WFS_MDF_DIR_ROOT_BLK), &pDbh) );
    DirBlkCreateRootNode(pDbh);
    ret = TransUnpinBlk(&tba);
    if (ret != WFSKRN_RESULT_OK)
    {
        return -1;
    }

    // initialize name directory.
    tba.pAreaInfo = (AreaInfo*) area;
    tba.nBlkAdr = NameBlock(area);
    tba.pTransInfo = pTransInfo;

    WFSKrnReturnOnError( DirCreateAndInitNewBlk(&tba, (WFSMetaDataFlags)(WFS_MDF_DIR_LEAF_BLK | WFS_MDF_DIR_ROOT_BLK), &pDbh) );
    DirBlkCreateRootNode(pDbh);
    ret = TransUnpinBlk(&tba);
    if (ret != WFSKRN_RESULT_OK)
    {
        return -1;
    }

    return 0;
}

//
// Wrapper
//

WFSResult AclGetFileSize(const AreaInfo* area, WFSAclFileHandle fh, WFSFileSize* fileSize)
{
    WFSResult result = OpenFile(area, fh);
    if (result == WFS_RESULT_OK)
    {
        // [check] file?
        *fileSize = FILEHANDLE_DI(fh).pAttrHdr->file.nSize;
    }
    CloseFile(fh);
    return result;
}

WFSResult AclWriteFile(const AreaInfo* area, WFSAclFileHandle fh, const void *pFileData, WFSFileSize nSize, WFSFileSize offset)
{
    WFSResult result = WFS_RESULT_INVALID;
    WFSKrnResult ret;
    DirFilePos fpos;

    if (OpenFile(area, fh) == WFS_RESULT_OK)
    {
        if (nSize + offset <= FILEHANDLE_DI(fh).pAttrHdr->nAllocSize)
        {
            TransGetBlk(&FILEHANDLE_DI(fh).tba, TRANS_FLAG_RW_BLK, 0); // set read/write. // [check] error handling.
            ret = DirFindFilePos(&FILEHANDLE_DI(fh), offset, &fpos, TRANS_FLAG_RW_BLK);
            WFSKRN_ACL_ASSERT(ret == WFSKRN_RESULT_OK);
            memcpy(fpos.pDataPtr, pFileData, nSize);

            if (FILEHANDLE_DI(fh).pAttrHdr->file.nSize < nSize + offset)
            {
                FILEHANDLE_DI(fh).pAttrHdr->file.nSize = nSize + offset;
            }
            result = WFS_RESULT_OK;
        }
    }
    CloseFile(fh);

#ifdef _DEBUG
    if (result == WFS_RESULT_OK)
    {
        // [check] for debug.
        WFSResult result;

        result = OpenFile(area, fh);
        WFSKRN_ACL_ASSERT(result == WFS_RESULT_OK);
        CloseFile(fh);

        if (nSize == sizeof(u32))
        {
            u32 buf;
            AclReadFile(area, fh, &buf, sizeof(u32), offset);
            WFSKRN_ACL_ASSERT(buf == *((u32*) pFileData));
        }
    }
#endif // DEBUG
    return result;
}

s32 AclReadFile(const AreaInfo* area, WFSAclFileHandle fh, void *pFileDataBuffer, s32 nSize, WFSFileSize offset)
{
    s32 result = -1;
    WFSKrnResult ret;
    DirFilePos fpos;

    if (OpenFile(area, fh) == WFS_RESULT_OK)
    {
        if (offset < FILEHANDLE_DI(fh).pAttrHdr->nAllocSize)
        {
            ret = DirFindFilePos(&FILEHANDLE_DI(fh), offset, &fpos, TRANS_FLAG_READ_BLK);
            WFSKRN_ACL_ASSERT(ret == WFSKRN_RESULT_OK);

            if (FILEHANDLE_DI(fh).pAttrHdr->file.nSize < offset)
            {
                result = 0;
            }
            else
            {
                if (FILEHANDLE_DI(fh).pAttrHdr->file.nSize < nSize + offset)
                {
                    nSize = FILEHANDLE_DI(fh).pAttrHdr->file.nSize - offset;
                }
                memcpy(pFileDataBuffer, fpos.pDataPtr, nSize);
                result = nSize;
            }
        }
    }
    CloseFile(fh);

#ifdef _DEBUG
    if (0 <= result)
    {
        // [check] for debug.
        WFSResult result;
        result = OpenFile(area, fh);
        WFSKRN_ACL_ASSERT(result == WFS_RESULT_OK);
        CloseFile(fh);
    }
#endif // DEBUG

    return result;
}

WFSResult AclCreateAccessListFile(const AreaInfo* area,
    const AclFileInfo* aclFile, WFSBool create, WFSAclFileHandle *pFh)
{
    WFSKRN_ACL_ASSERT(aclFile->type == ACL_TYPE_ACLFILE);

    *pFh = (WFSAclFileHandle) aclFile;
    FILEHANDLE_TYPE(*pFh) = ACL_TYPE_ACLFILE;
    return AclCreateFile(area, aclFile, create, pFh);
}

WFSResult AclCreateFile(const AreaInfo* area, const AclFileInfo* fileInfo, WFSBool create, WFSAclFileHandle *pFh)
{
    if (!create && (AclAccessListAlreadyExist(area, fileInfo) == (WFSBool) TRUE))
    {
        return WFS_RESULT_OK;
    }

    SetFileNameLen(FILEHANDLE_NAME(*pFh));
    if (CreateFile(area, fileInfo, *pFh) != WFS_RESULT_OK)
    {
        CloseFile(*pFh);
        return WFS_RESULT_INVALID;
    }
    CloseFile(*pFh);

#ifdef _DEBUG
    {
        // [check] for debug.
        WFSResult result;
        result = OpenFile(area, *pFh);
        WFSKRN_ACL_ASSERT(result == WFS_RESULT_OK);
        CloseFile(*pFh);
    }
#endif // _DEBUG

    return WFS_RESULT_OK;
}

WFSResult AclDeleteAccessListFile(const AreaInfo* area, const AclFileInfo* aclFile)
{
    WFSResult result;
    WFSAclFileHandle fh = (WFSAclFileHandle) aclFile;
    FILEHANDLE_TYPE(fh) = aclFile->type;

    SetFileNameLen(FILEHANDLE_NAME(fh));
    result = DeleteFile(area, fh);
    if (result != WFS_RESULT_OK)
    {
        WFSKRN_ACL_REPORT("AclDeleteAccessListFile(%d) Error %d\n", result);
        return WFS_RESULT_INVALID; // [check] error code.
    }
    return WFS_RESULT_OK;
}

WFSResult AclCloseFile(WFSAclFileHandle fh)
{
    return CloseFile(fh);
}

WFSBool AclAccessListAlreadyExist(const AreaInfo* area, const AclFileInfo* aclFile)
{
    WFSResult ret;
    WFSAclFileHandle fh = (WFSAclFileHandle) aclFile;
    FILEHANDLE_TYPE(fh) = aclFile->type;

    SetFileNameLen(FILEHANDLE_NAME(fh));

    ret = OpenFile(area, fh);
    CloseFile(fh);
    if (ret == WFS_RESULT_OK)
    {
        return (WFSBool) TRUE;
    }
    return (WFSBool) FALSE;
}

#ifdef _DEBUG

s32 AclGetNumDirEntry(const AreaInfo* area, WFSBlkAdr block)
{
    WFSKrnResult result;
    s32 num = 0;
    DirItr di;

    SetFileName(&di.name, "");

    di.tba.pAreaInfo = (AreaInfo*) area; // area
    di.tba.nBlkAdr = block;              // parent directory
    di.tba.pTransInfo = 0;                // used for journaling.
    di.nBlkDepth = 0;                    // search from the root block.
    di.ppDnsRootParent = 0;

    result = DirFind(&di.name, &di, DIR_MAX_BLK_DEPTH,
        DIR_FIND_MODE_PREFIX_SEARCH);

    while (result == WFSKRN_RESULT_OK)
    {
        ++num;
        result = DirNextRaw(&di.name, &di);
    }
    DirItrClose(&di);

    if (result != WFSKRN_RESULT_NOT_FOUND &&
        result != WFSKRN_RESULT_DIR_NODE_CHOICE_NOT_FOUND)
    {
        // unexpected error.
        num = (s32) result;
    }
    return num;
}

s32 AclDirList(const AreaInfo* area, WFSBlkAdr block)
{
    WFSKrnResult result;
    s32 num = 0;
    DirItr di;

    SetFileName(&di.name, "");

    di.tba.pAreaInfo = (AreaInfo*) area; // area
    di.tba.nBlkAdr = block;              // parent directory
    di.tba.pTransInfo = 0;                // used for journaling.
    di.nBlkDepth = 0;                    // search from the root block.
    di.ppDnsRootParent = 0;

    result = DirFind(&di.name, &di, DIR_MAX_BLK_DEPTH,
        DIR_FIND_MODE_PREFIX_SEARCH);

    while (result == WFSKRN_RESULT_OK)
    {
        ++num;
        if (block == NameBlock(area))
        {
            osTPrintf("%s\n", di.name.sStr);
        }
        else
        {
            utf8* p = di.name.sStr;
            int count = (int) STRNLEN(p, WFS_MAX_FILE_NAME_SIZE);
            while (count--)
            {
                osTPrintf("%02x", (u8) *p++);
            }
            osTPrintf("\n");
        }
        result = DirNextRaw(&di.name, &di);
    }
    DirItrClose(&di);

    if (result != WFSKRN_RESULT_NOT_FOUND &&
        result != WFSKRN_RESULT_DIR_NODE_CHOICE_NOT_FOUND)
    {
        // unexpected error.
        num = (s32) result;
    }
    return num;
}

WFSKrnResult AclCheckDisk(const AreaInfo* area)
{
    WFSKrnAclHandle handle;
    IndexListEntry entry;
    int ret;
    int i;
    WFSKrnResult result;
    int num;
    AclFileInfo aclFile;

    WFSAclFileHandle indexListHandle;
    AclFileInfo indexListInfo;

    if (OpenIndexList(area, &indexListHandle, &indexListInfo) < 0)
    {
        return WFSKRN_RESULT_ACL_FILE;
    }

    num = 0;
    for (i = 0; i < WFSKRN_ACCESSLIST_INDEXLIST_MAX_PAGE * WFSKRN_ACCESSLIST_INDEXES_IN_PAGE; ++i)
    {
        handle = (WFSKrnAclHandle) i;
        ret = GetAccessListIndex(area, indexListHandle, handle, &entry);
        if (ret < 0)
        {
            break;
        }
        if (entry.location != WFSKRN_ACCESSLIST_FREEINDEX)
        {
            AclNameFileHeader header;
            if (ReadNamefile(area, handle, &header, sizeof(header), 0) < 0)
            {
                result = WFSKRN_RESULT_NOT_FOUND;
                goto CHECK_DISK_ERROR;
            }
            if (ReadNamefile(area, handle, ACL_FILEINFO_NAME(&aclFile), WFS_MAX_FILE_NAME_SIZE, sizeof(AclNameFileHeader)) < 0)
            {
                result = WFSKRN_RESULT_ACL_FILENAME;
                goto CHECK_DISK_ERROR;
            }
            aclFile.type = ACL_TYPE_ACLFILE;
            if (!AclAccessListAlreadyExist(area, &aclFile))
            {
                result = WFSKRN_RESULT_ACL_FILE;
                goto CHECK_DISK_ERROR;
            }
            ++num;
        }
        else
        {
            AclNameFileHeader header;
            if (ReadNamefile(area, handle, &header, sizeof(header), 0) == 0)
            {
                result = WFSKRN_RESULT_ALREADY_EXISTS;
                goto CHECK_DISK_ERROR;
            }
        }
    }

    if (AclGetNumDirEntry(area, AclBlock(area)) != num)
    {
        result = WFSKRN_RESULT_ACL_ACLDIR_INCONSISTENT;
        goto CHECK_DISK_ERROR;
    }
    if (AclGetNumDirEntry(area, NameBlock(area)) != num + 1)
    {
        result = WFSKRN_RESULT_ACL_NAMEDIR_INCONSISTENT;
        goto CHECK_DISK_ERROR;
    }

    CloseIndexList(area, indexListHandle);
    return WFSKRN_RESULT_OK;

CHECK_DISK_ERROR:

    /* [check] debug code.
    osTPrintf("--------- Acl Directory ----------\n");
    AclDirList(area, AclBlock(area));
    osTPrintf("--------- Name file directory ----------\n");
    AclDirList(area, nameBlock(area));
    osTPrintf("\n");
    */

    CloseIndexList(area, indexListHandle);
    return result;
}
#endif // _DEBUG

#endif // ifndef WFSDEV
