/*---------------------------------------------------------------------------*
  Project:  wfsKrn
  File:     wfsKrn_Permission.cpp - Functions for permissions.

  Copyright 2008 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_Permission_Win32.cpp,v $
  Revision 1.13  2008/09/24 07:33:21  ueno
  Added AclHdr{} to AreaHdr{} to save configuration of accesslists in a disk.

  Revision 1.12  2008/06/24 18:20:00  ueno
  Modified to reserve accesslist index 0 as the default (empty) accesslist.

  Revision 1.11  2008/05/19 10:11:52  ueno
  Revised not to use hardlink.

  Revision 1.10  2008/05/19 05:30:50  ueno
  Defined accesslist meta-data types.

  Revision 1.9  2008/05/15 13:31:54  ueno
  Updated.

  Revision 1.8  2008/05/08 08:07:44  ueno
  Revised AclReadFile() and AclWriteFile() to set offset.

  Revision 1.7  2008/05/06 03:17:57  ueno
  Added AclGetFileSize() to wrap Windows API..

  Revision 1.6  2008/05/06 02:26:03  ueno
  Fixed ifdef.

  Revision 1.5  2008/04/29 06:25:20  ueno
  Cleanup.

  Revision 1.4  2008/04/25 07:49:17  ueno
  Cleanup.

  Revision 1.3  2008/04/25 06:22:55  ueno
  Cleanup.

  Revision 1.2  2008/04/25 04:40:07  ueno
  Added support for WFSDEV.

  Revision 1.20  2008/04/23 08:59:46  ueno
  Replaced WFSAreaHdr by AreaInfo.
  Revised directory permission (change -> add).

  Revision 1.19  2008/04/11 06:16:47  ueno
  Defined WFSAclFileHandle.

  Revision 1.18  2008/04/08 06:39:31  ueno
  Cleanup.

  Revision 1.17  2008/04/07 09:19:21  ueno
  Modified WFSKrnAccessListDirectoryInit() to return an integer.

  Revision 1.16  2008/04/03 11:40:43  ueno
  Replaced accesslist index by accesslist handle.

  Revision 1.15  2008/03/19 10:17:38  ueno
  Fixed Win32 content <-> filename conversion.

  Revision 1.14  2008/03/14 08:02:53  ueno
  Revised API prefixes (WFS -> WFSKrn).

  Revision 1.13  2008/03/11 11:10:57  ueno
  Moved accesslist functions into wfskrn_Permission_AccessList.cpp.

  Revision 1.12  2008/03/11 10:48:22  ueno
  Cleanup.

  Revision 1.11  2008/03/11 09:51:54  ueno
  Fixed EscapeAccessList() to make a filename from UTF-16 characters (0x3300-0x33ff).

  Revision 1.10  2008/03/11 06:53:06  ueno
  Added WFSAccessListSetPermission() and GetPermission().

  Revision 1.9  2008/03/07 05:32:07  ueno
  Modified WFSCreateAccessList() to be able to create an accesslist for directories.

  Revision 1.8  2008/03/06 10:45:28  ueno
  Added debug codes.

  Revision 1.7  2008/02/29 07:29:10  ueno
  Added Windows api wrappers.

  Revision 1.6  2008/02/28 13:11:01  ueno
  Revised to keep accesslist cache and indexlist handle in WFSAreaHdr{}.

  Revision 1.5  2008/02/28 04:36:32  ueno
  Cleanup.

  Revision 1.4  2008/02/27 13:12:32  ueno
  Revised to store a filename of an accesslist in a file (*.name).

  Revision 1.3  2008/02/27 11:17:41  ueno
  Cleanup.

  Revision 1.2  2008/02/27 10:05:36  ueno
  Revised to use WFSAreaHdr{} to specifiy WFS area.

  Revision 1.1  2008/02/22 05:16:14  ueno
  Initial Check-in.


*---------------------------------------------------------------------------*/
#include "wfs_Defs.h"
#include "wfskrn_Permission_AccessList.h"
#include "wfskrn_Permission_Win32.h"
#include "wfskrn_Permission_File.h"

//
// Win32 Utilities
//
#if _WIN32
#ifdef WFSDEV

#include "wfsDev/wfsdev_api.h"
#include <winbase.h>

WFSBool AlreadyExist(const LPWSTR path)
{
    DWORD attr = GetFileAttributesW(path);
    if (attr != -1)
    {
        return TRUE;
    }
    return FALSE;
}

// source string must be null-terminated.
int AclUtf8ToWchar(const utf8* utf8, LPWSTR wPath, u32 bytes)
{
    int ret;
    if (bytes < sizeof(LONGPATH_PREFIX) + sizeof(WCHAR))
    {
        return -1;
    }

    u32 maxWchar = bytes/sizeof(WCHAR);
    int prefixLen = 0;
    // add prefix \\?\ to enable long pathname.
    prefixLen = _snwprintf(wPath, maxWchar, L"%s", LONGPATH_PREFIX);
    if (prefixLen < 0)
    {
        return -1;
    }

    wPath += prefixLen;
    maxWchar -= prefixLen;

    ret = MultiByteToWideChar(CP_UTF8, 0, utf8, -1, wPath, maxWchar);
    if (0 < ret)
    {
        return ret + prefixLen;
    }
    WFSKRN_ACL_REPORT("Utf8ToWchar() Error %ld\n", GetLastError());
    return -1;
}

// source string must be null-terminated.
int AclWcharToUtf8(const LPWSTR wPath, utf8* utf8, u32 bytes)
{
    size_t count = wcslen(LONGPATH_PREFIX);
    if (wcsncmp(wPath, LONGPATH_PREFIX, count) != 0)
    {
        count = 0;
    }

    int ret = WideCharToMultiByte(CP_UTF8, 0, wPath + count, -1, utf8, bytes, 0, 0);
    if (0 < ret)
    {
        return ret;
    }
    WFSKRN_ACL_REPORT("WcharToUtf8() Error %ld\n", GetLastError());
    return -1;
}

// Get the directory which contains accesslists.
int GetAccessListPath(const AreaInfo* area, AclFileInfo* aclFile)
{
    WFSResult result = wfsdev_api::_getAccessListFolder(aclFile->path, area);
    if (WFS_RESULT_OK == result)
    {
        WFSKRN_ACL_ASSERT(strlen(aclFile->path) < WFSKRN_ACCESSLIST_MAX_PATH_LEN);
        return strlen(aclFile->path);
    }
    return -1;
}

// convert to little endian UTF-16 for Windows.
static int ConvertToUTF16(const utf8* accessListName, LPWSTR wPath, u32 size)
{
    const utf8* src = accessListName;
    utf8* p = (utf8*) wPath;
    const utf8* tail = p + size;
    WFSKRN_ACL_ASSERT(2 < size);

    while (*src && (p < tail - 2))
    {
        if (*src == ESCAPE)
        {
            *p++ = ESCAPE;
            *p++ = ESCAPE;
            *p++ = *(src+1);
            *p++ = ESCAPE;
            src += 2;
        }
        else
        {
            *p++ = *src++;
            *p++ = ESCAPE;
        }
    }

    if (tail <= p + 2)
    {
        return -1;
    }
    *p++ = 0;
    *p++ = 0;

    return 0;
}

int Utf8ToWcharAccessList(const AreaInfo* area, const utf8* path, LPWSTR wPath, u32 wPathLen)
{
    AclFileInfo dir;
    u32 dirLen;
    int len;

    if (WFSKrnAccessListGetAclFileInfo(area, "", ACL_TYPE_ACLFILE, &dir) < 0)
    {
        return -1;
    }

    dirLen = strlen(dir.path);
    if (memcmp(path, dir.path, dirLen) != 0)
    {
        return -1;
    }

    len = AclUtf8ToWchar(dir.path, wPath, wPathLen);
    if (len <= 0)
    {
        return -1;
    }

    if (ConvertToUTF16(&path[dirLen], &wPath[len-1], wPathLen - len) < 0)
    {
        return -1;
    }

    return 0;
}

static int GetLinkFilename(const AreaInfo* area, WFSKrnAclHandle handle, LPWSTR linkFile)
{
    utf8 filename[WFSKRN_ACCESSLIST_MAX_PATH_LEN];
    AclFileInfo fileInfo;
    _snprintf(filename, sizeof(filename), "%x", handle);

    if (WFSKrnAccessListGetAclFileInfo(area, filename, ACL_TYPE_NAMEFILE, &fileInfo) < 0)
    {
        return -1;
    }

    if (AclUtf8ToWchar(fileInfo.path, linkFile, WFSKRN_ACCESSLIST_MAX_PATH_LEN) < 0)
    {
        return -1;
    }
    return 0;
}

static int GetNamefile(const AreaInfo* area, WFSKrnAclHandle handle, LPWSTR nameFile)
{
    utf8 name[WFS_MAX_FILE_NAME_SIZE];
    AclFileInfo fileInfo;
    _snprintf(name, sizeof(name), "%08x", handle);

    if (WFSKrnAccessListGetAclFileInfo(area, name, ACL_TYPE_NAMEFILE, &fileInfo) < 0 ||
        AclUtf8ToWchar(fileInfo.path, nameFile, WFSKRN_ACCESSLIST_MAX_PATH_LEN) < 0)
    {
        return -1;
    }
    return 0;
}

int CreateNamefile(const AreaInfo* area, WFSKrnAclHandle handle, const AclNameFileHeader* header, const utf8* filename)
{
    WCHAR wPath[WFSKRN_ACCESSLIST_MAX_PATH_LEN];
    WFSAclFileHandle hFile;
    DWORD written;

    if (GetNamefile(area, handle, wPath) < 0)
    {
        return -1;
    }

    hFile = CreateFileW(wPath, GENERIC_READ | GENERIC_WRITE, 0,
                         NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
    if (hFile == INVALID_HANDLE_VALUE)
    {
        return -1;
    }

    if (SetFilePointer(hFile, 0, NULL, FILE_BEGIN) == INVALID_SET_FILE_POINTER ||
        WriteFile(hFile, header, sizeof(AclNameFileHeader), &written, NULL) == 0 ||
        sizeof(AclNameFileHeader) != written)
    {
        WFSKRN_ACL_REPORT("CreateNamefile(%d) Error %ld\n", handle, GetLastError());
        CloseHandle(hFile);
        return -1;
    }

    if (SetFilePointer(hFile, sizeof(AclNameFileHeader), NULL, FILE_BEGIN) == INVALID_SET_FILE_POINTER ||
        WriteFile(hFile, filename, strlen(filename)+1, &written, NULL) == 0 ||
        strlen(filename)+1 != written)
    {
        WFSKRN_ACL_REPORT("CreateNamefile(%d) Error %ld\n", handle, GetLastError());
        CloseHandle(hFile);
        return -1;
    }
    CloseHandle(hFile);

    return 0;
}

int ReadNamefile(const AreaInfo* area, WFSKrnAclHandle handle, void* buf, WFSFileSize size, WFSFileSize offset)
{
    WCHAR wPath[WFSKRN_ACCESSLIST_MAX_PATH_LEN];
    WFSAclFileHandle hFile;
    DWORD bytesRead;

    if ((handle & ACL_TYPE_MASK) == 0)
    {
        handle = 0; // handle 0 represents both file 0 and directory 0x80000000.
    }
    if (GetNamefile(area, handle, wPath) < 0)
    {
        return -1;
    }

    hFile = CreateFileW(wPath, GENERIC_READ, FILE_SHARE_READ,
                         NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
    if (hFile == INVALID_HANDLE_VALUE)
    {
        return -1;
    }

    if (SetFilePointer(hFile, offset, NULL, FILE_BEGIN) == INVALID_SET_FILE_POINTER ||
        !ReadFile(hFile, buf, size, &bytesRead, NULL))
    {
        WFSKRN_ACL_REPORT("ReadNamefile(%d) Error %ld\n", handle, GetLastError());
        CloseHandle(hFile);
        return -1;
    }
    CloseHandle(hFile);

    return 0;
}

int WriteNamefile(const AreaInfo* area, WFSKrnAclHandle handle, const void* buf, WFSFileSize size, WFSFileSize offset)
{
    WCHAR wPath[WFSKRN_ACCESSLIST_MAX_PATH_LEN];
    WFSAclFileHandle hFile;
    DWORD written;

    if (GetNamefile(area, handle, wPath) < 0)
    {
        return -1;
    }

    hFile = CreateFileW(wPath, GENERIC_WRITE, FILE_SHARE_READ,
                         NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
    if (hFile == INVALID_HANDLE_VALUE)
    {
        return -1;
    }

    if (SetFilePointer(hFile, offset, NULL, FILE_BEGIN) == INVALID_SET_FILE_POINTER ||
        WriteFile(hFile, buf, size, &written, NULL) == 0 || written != size)
    {
        WFSKRN_ACL_REPORT("WriteNamefile(%d) Error %ld\n", handle, GetLastError());
        CloseHandle(hFile);
        return -1;
    }
    CloseHandle(hFile);

    return 0;
}

int DeleteNamefile(const AreaInfo* area, WFSKrnAclHandle handle)
{
    WCHAR wPath[WFSKRN_ACCESSLIST_MAX_PATH_LEN];

    if (GetNamefile(area, handle, wPath) < 0)
    {
        return -1;
    }

    if (!DeleteFileW(wPath))
    {
        WFSKRN_ACL_REPORT("DeleteNamefile(%d) Error %ld\n", handle, GetLastError());
        return -1;
    }

    return 0;
}

static int CreateNameDir(const AreaInfo* area)
{
    WCHAR wPath[WFSKRN_ACCESSLIST_MAX_PATH_LEN];
    AclFileInfo fileInfo;

    if (WFSKrnAccessListGetAclFileInfo(area, "", ACL_TYPE_NAMEFILE, &fileInfo) < 0)
    {
        return -1;
    }

    if (AclUtf8ToWchar(fileInfo.path, wPath, WFSKRN_ACCESSLIST_MAX_PATH_LEN) < 0)
    {
        return -1;
    }

    if (CreateDirectoryW(wPath, 0))
    {
        return 0;
    }

    if (GetLastError() == ERROR_ALREADY_EXISTS)
    {
        return 0;
    }
    return -1;
}

int WFSKrnAccessListDirectoryInit(AreaInfo* area)
{
    WFSResult result = _createAccessListFolder( area );
    if (result != WFS_RESULT_OK)
    {
        return -1;
    }
    return CreateNameDir(area);
}

//
// Wrapper
//
WFSResult AclGetFileSize(const AreaInfo* area, WFSAclFileHandle fh, WFSFileSize* fileSize)
{
    DWORD size = GetFileSize(fh, NULL);
    if (size < 0)
    {
        return WFS_RESULT_INVALID; // [check]
    }
    *fileSize = (WFSFileSize) size;
    return WFS_RESULT_OK;
}

WFSResult AclWriteFile(const AreaInfo* area, WFSAclFileHandle fh, const void *pFileData, WFSFileSize nSize, WFSFileSize offset)
{
    DWORD written;
    if (SetFilePointer(fh, (u32) offset, NULL, FILE_BEGIN) == INVALID_SET_FILE_POINTER)
    {
        WFSKRN_ACL_REPORT("SetFilePointer: Error %ld\n", GetLastError());
        return WFS_RESULT_INVALID;
    }

    if (WriteFile(fh, pFileData, nSize, &written, NULL) == 0 || written != nSize)
    {
        WFSKRN_ACL_REPORT("WFSKrnAccessListWriteFile(): Error %ld\n", GetLastError());
        return WFS_RESULT_INVALID; // [check] error code.
    }
    return WFS_RESULT_OK;
}

s32 AclReadFile(const AreaInfo* area, WFSAclFileHandle fh, void *pFileDataBuffer, s32 nSize, WFSFileSize offset)
{
    DWORD bytesRead;
    if (SetFilePointer(fh, (u32) offset, NULL, FILE_BEGIN) == INVALID_SET_FILE_POINTER)
    {
        WFSKRN_ACL_REPORT("SetFilePointer: Error %ld\n", GetLastError());
        return WFS_RESULT_INVALID;
    }

    if (!ReadFile(fh, pFileDataBuffer, nSize, &bytesRead, NULL))
    {
        WFSKRN_ACL_REPORT("WFSKrnAccessListReadFile(): Error %ld\n", GetLastError());
        return WFS_RESULT_INVALID; // [check] error code.
    }
    return bytesRead;
}

WFSResult AclCreateAccessListFile(const AreaInfo* area,
    const AclFileInfo* aclFile, WFSBool create, WFSAclFileHandle *pFh)
{
    WCHAR wPath[WFSKRN_ACCESSLIST_MAX_PATH_LEN];
    WFSAclFileHandle hFile;
    DWORD dwCreationDisposition;

    if (Utf8ToWcharAccessList(area, aclFile->path, wPath, sizeof(wPath)) < 0)
    {
        return WFS_RESULT_INVALID; // [check] error code.
    }

    if (create)
    {
        dwCreationDisposition = OPEN_ALWAYS;
    }
    else
    {
        dwCreationDisposition = OPEN_EXISTING;
    }

    hFile = CreateFileW(wPath, GENERIC_READ | GENERIC_WRITE, 0,
                         NULL, dwCreationDisposition, FILE_ATTRIBUTE_NORMAL, NULL);

    if (hFile == INVALID_HANDLE_VALUE)
    {
        return WFS_RESULT_INVALID; // [check] error code.
    }

    *pFh = hFile;
    return WFS_RESULT_OK;
}

WFSResult AclCreateFile(const AreaInfo* area, const AclFileInfo* fileInfo, WFSBool create, WFSAclFileHandle *pFh)
{
    WCHAR wPath[WFSKRN_ACCESSLIST_MAX_PATH_LEN];
    WFSAclFileHandle hFile;
    DWORD dwCreationDisposition;

    if (AclUtf8ToWchar(fileInfo->path, wPath, sizeof(wPath)) < 0)
    {
        return WFS_RESULT_INVALID; // [check] error code.
    }

    if (create)
    {
        dwCreationDisposition = OPEN_ALWAYS;
    }
    else
    {
        dwCreationDisposition = OPEN_EXISTING;
    }

    hFile = CreateFileW(wPath, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ,
                         NULL, dwCreationDisposition, FILE_ATTRIBUTE_NORMAL, NULL);

    if (hFile == INVALID_HANDLE_VALUE)
    {
        return WFS_RESULT_INVALID; // [check] error code.
    }

    *pFh = hFile;
    return WFS_RESULT_OK;
}

WFSResult AclDeleteAccessListFile(const AreaInfo* area, const AclFileInfo* aclFile)
{
    WCHAR wPath[WFSKRN_ACCESSLIST_MAX_PATH_LEN];

    if (Utf8ToWcharAccessList(area, aclFile->path, wPath, sizeof(wPath)) < 0)
    {
        return WFS_RESULT_INVALID; // [check] error code.
    }

    if (!DeleteFileW(wPath))
    {
        return WFS_RESULT_INVALID; // [check] error code.
    }
    return WFS_RESULT_OK;
}

WFSResult AclCloseFile(WFSAclFileHandle fh)
{
    if (CloseHandle(fh))
    {
        return WFS_RESULT_OK;
    }
    return WFS_RESULT_INVALID; // [check] error code.
}

WFSBool AclAccessListAlreadyExist(const AreaInfo* area, const AclFileInfo* aclFile)
{
    WCHAR wPath[WFSKRN_ACCESSLIST_MAX_PATH_LEN];

    if (Utf8ToWcharAccessList(area, aclFile->path, wPath, sizeof(wPath)) < 0)
    {
        return (WFSBool) FALSE;
    }

    if (AlreadyExist(wPath))
    {
        return (WFSBool) TRUE;
    }
    return (WFSBool) FALSE;
}

#endif // WFSDEV
#endif // _WIN32

