/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfskrn_SubBlkAlloc.cpp


  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_SubBlkAlloc.cpp,v $
  Revision 1.8  2008/10/17 08:51:10  kondo_masahiro
  Added the new MEDIUM size category (40KB-320KB)

  Revision 1.7  2008/08/27 23:23:28  ooizumi
  Fixed newline to CR+LF(Windows format).

  Revision 1.6  2008/08/27 09:42:37  paul
  Removed nTotalFree from block header

  Revision 1.5  2008/07/09 00:50:19  paul
  Minor changes

  Revision 1.4  2008/06/09 17:24:21  paul
  Changed SbaBlkInit() to support blocks with different header sizes

  Revision 1.3  2008/05/10 03:59:44  kondo_masahiro
  Merged and fixed code to accept RM compiler

  Revision 1.2  2008/04/25 17:28:50  kondo_masahiro
  Fixed code to accept RM compiler

  Revision 1.1  2008/04/22 04:39:08  kondo_masahiro
  Fixed code to transfer the wfskrn project to RM_SDK tree

  Revision 1.5  2008/04/22 22:19:42  paul
  Fixed a bug in SbaAllocSubBlk

  Revision 1.4  2008/04/19 05:50:28  kondo_masahiro
  Fixed code to accept RM compiler

  Revision 1.3  2008/04/04 23:36:33  paul
  Added free space checking function. Some identifier name changes.

  Revision 1.2  2008/02/28 06:29:17  kondo_masahiro
  Fixed code to accept PPC compiler

  Revision 1.1  2008/02/21 00:05:49  paul
  *** empty log message ***


 *---------------------------------------------------------------------------*/


#include "wfskrn_SubBlkAlloc.h"
#include "wfskrn_Dir.h"
#include "wfskrn_Area.h"
#include "wfskrn_BitField.h"
#include "wfskrn_Heap.h"
#include "wfskrn_Api.h"
#include "wfskrn_Mutex.h"

#undef dbg
#if _DEBUG_SUB_BLK_ALLOC
    #include "randomlib.h"
    #define dbg(s) s
#else
    #define dbg(s)
#endif

void SbaCheckOfs(u16 nOfs) {
    // Check the intra block offset to ensure it is within the right range and properly aligned.
    if ((nOfs==0) ||
        //(nOfs>=wkg.areaInfoRoot.nBlkSize) || 
        (nOfs&((1<<SBA_MIN_LOG2_SUB_BLK_SIZE)-1))) {
        WFSKrnOutputErrorStr("Ofs Error!\n");
    }
}

void SbaRemoveSubBlkFromFreeList(WFSSubBlkAllocHdr *pBlkHdr, u16 nOfs, u32 nLog2Size) {
    WFSSubBlkFreeHdr *pSubBlkHdr = (WFSSubBlkFreeHdr*)&((u8*)pBlkHdr)[nOfs];
    pSubBlkHdr->nMagic = 0;
    WFSSubBlkFreeHdr *pPrev = (WFSSubBlkFreeHdr*)&((u8*)pBlkHdr)[pSubBlkHdr->nPrev];
    WFSSubBlkFreeHdr *pNext = (WFSSubBlkFreeHdr*)&((u8*)pBlkHdr)[pSubBlkHdr->nNext];
    pPrev->nNext = pSubBlkHdr->nNext;
    pNext->nPrev = pSubBlkHdr->nPrev;
    pBlkHdr->aSubBlkList[nLog2Size - SBA_MIN_LOG2_SUB_BLK_SIZE].nNumFree--;
    pBlkHdr->aSubBlkList[nLog2Size - SBA_MIN_LOG2_SUB_BLK_SIZE].nFirstFree = pSubBlkHdr->nNext;
}

u16 SbaAllocSubBlk(WFSSubBlkAllocHdr *pBlkHdr, u32 nLog2Size) {
    // This function allocates a contiguous sub block of size (1<<nLog2Size) from within the block
    u32 nOfs;
    if (pBlkHdr->aSubBlkList[nLog2Size - SBA_MIN_LOG2_SUB_BLK_SIZE].nNumFree == 0) {
        // Find a larger block and split it
        u32 nI = nLog2Size;
        do {
            ++nI;
            if (nI > SBA_MAX_LOG2_SUB_BLK_SIZE) {
                return 0;   // Unable to find a block
            }
        } while(pBlkHdr->aSubBlkList[nI - SBA_MIN_LOG2_SUB_BLK_SIZE].nNumFree == 0);
#if _DEBUG_SUB_BLK_ALLOC
        pBlkHdr->nTotalFree -= 1<<nLog2Size;
#endif
        nOfs = pBlkHdr->aSubBlkList[nI - SBA_MIN_LOG2_SUB_BLK_SIZE].nFirstFree;     // Get offset to first free block of size (1<<nI)
        SbaRemoveSubBlkFromFreeList(pBlkHdr, nOfs, nI);  // Remove this free block from the list
        //pFreeSubBlkHdr->nNext = 0;
        // Split the block in half repeatedly until we get a block of the requested size.
        // Starts as |aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa|  where 'a' is the free block
        // Ends as   |RRRRddddccccccccbbbbbbbbbbbbbbbb|  where 'R' is the newly allocated block, and 'b', 'c', 'd' are newly created free blocks
        u32 nOfsJ = nOfs;
        nOfsJ += (1<<nLog2Size);
        u32 nJ;
        for(nJ=nLog2Size; nJ<nI; nJ++) {
            // For each power of 2 between the requested size and the free block size
            // Set up a single element list 
            WFSSubBlkFreeHdr *pFreeSubBlkHdr = (WFSSubBlkFreeHdr*)&((u8*)pBlkHdr)[nOfsJ];
            pFreeSubBlkHdr->nMagic = SBA_FREE_SUB_BLK_MAGIC;
            pFreeSubBlkHdr->nLog2Size = (u16)nJ;
            WFSSubBlkList *pSubBlkListJ = &pBlkHdr->aSubBlkList[nJ - SBA_MIN_LOG2_SUB_BLK_SIZE];
            // This is the first free block of this size, so make a circular list with one element
            pFreeSubBlkHdr->nNext = // continued on next line
            pFreeSubBlkHdr->nPrev = (u16)((size_t)pFreeSubBlkHdr - (size_t)pBlkHdr);
            pSubBlkListJ->nFirstFree = (u16)nOfsJ;
            pSubBlkListJ->nNumFree = 1;
            nOfsJ += (1<<nJ);
        }
        return (u16)nOfs;
    }
#if _DEBUG_SUB_BLK_ALLOC
    pBlkHdr->nTotalFree -= 1<<nLog2Size;
#endif
    nOfs = pBlkHdr->aSubBlkList[nLog2Size - SBA_MIN_LOG2_SUB_BLK_SIZE].nFirstFree;
    SbaRemoveSubBlkFromFreeList(pBlkHdr, nOfs, nLog2Size);  // Remove this free block from the list
    return (u16)nOfs;
}


WFSKrnResult SbaAllocSubBlks(WFSSubBlkAllocHdr *pBlkHdr, u32 nLog2Size, u16 *pSubBlkOfs, u32 nNumSubBlks) {
    // This function allocs multiple sub blocks of the same size at once. If it cannot alloc all blocks, it does not alloc any.
    u32 nNumFreeBlks = pBlkHdr->aSubBlkList[nLog2Size - SBA_MIN_LOG2_SUB_BLK_SIZE].nNumFree;
    u32 nL2S = nLog2Size + 1;
    u32 nMultiple = 2;
    while(nNumFreeBlks < nNumSubBlks) {
        if (nL2S > SBA_MAX_LOG2_SUB_BLK_SIZE) {
            return WFSKRN_RESULT_OUT_OF_MEMORY;
        }
        nNumFreeBlks += nMultiple * pBlkHdr->aSubBlkList[nL2S - SBA_MIN_LOG2_SUB_BLK_SIZE].nNumFree;
        nMultiple += nMultiple;
        ++nL2S;
    }
    /* Enough free space was found to satisfy the request
    u16 nOfs;
    --nL2S;
    WFSSubBlkList *pSubBlkList = &pBlkHdr->aSubBlkList[nLog2Size - SBA_MIN_LOG2_SUB_BLK_SIZE];
    u32 nOfsStep = (1<<nLog2Size);
    nMultiple = 1;
    u32 nJ = nLog2Size;
    while(nJ < nL2S) {
        if (pSubBlkList->nNumFree) {
            nNumSubBlks -= pSubBlkList->nNumFree * nMultiple;
            nOfs = pBlkHdr->aSubBlkList[nJ - SBA_MIN_LOG2_SUB_BLK_SIZE].nFirstFree;
            WFSSubBlkFreeHdr *pFreeSubBlkHdr;
            do {
                pFreeSubBlkHdr = (WFSSubBlkFreeHdr*)&((u8*)pBlkHdr)[nOfs];
                u32 nI = nMultiple;
                *pSubBlkOfs++ = nOfs;
                while(--nI) {
                    nOfs += nOfsStep;
                    *pSubBlkOfs++ = nOfs;
                }
                nOfs = *pSubBlkOfs = pFreeSubBlkHdr->nNext;
            } while(--pSubBlkList->nNumFree);
        }
        nOfsStep += nOfsStep;
        nMultiple += nMultiple;
        ++nJ;
    }*/
    while(nNumSubBlks--) {
        *pSubBlkOfs++ = SbaAllocSubBlk(pBlkHdr, nLog2Size);
    }
    return WFSKRN_RESULT_OK;
}


void SbaFreeSubBlk(WFSSubBlkAllocHdr *pBlkHdr, u16 nOfs, u32 nLog2Size) {
#if _DEBUG_SUB_BLK_ALLOC
    memset(((u8*)pBlkHdr)+nOfs, 0xAA, 1<<nLog2Size);
    pBlkHdr->nTotalFree += 1<<nLog2Size;
#endif
    while(nLog2Size < SBA_MAX_LOG2_SUB_BLK_SIZE) {
        u32 nSize = (u32)(1<<nLog2Size);
        u32 nNeigbourOfs = nOfs^nSize;
        WFSSubBlkFreeHdr *pBuddy = (WFSSubBlkFreeHdr*)&((u8*)pBlkHdr)[nNeigbourOfs];
        if ((pBuddy->nMagic!=SBA_FREE_SUB_BLK_MAGIC) || (pBuddy->nLog2Size!=nLog2Size)) {
            break;
        }
        // Remove the buddy sub-block from its free list
        WFSSubBlkFreeHdr *pPrev = (WFSSubBlkFreeHdr*)&((u8*)pBlkHdr)[pBuddy->nPrev];
        WFSSubBlkFreeHdr *pNext = (WFSSubBlkFreeHdr*)&((u8*)pBlkHdr)[pBuddy->nNext];
        pPrev->nNext = pBuddy->nNext;
        pNext->nPrev = pBuddy->nPrev;
        pBlkHdr->aSubBlkList[nLog2Size - SBA_MIN_LOG2_SUB_BLK_SIZE].nNumFree--;
        pBlkHdr->aSubBlkList[nLog2Size - SBA_MIN_LOG2_SUB_BLK_SIZE].nFirstFree = pBuddy->nNext;
        if (nOfs > nNeigbourOfs) {
            nOfs = nNeigbourOfs;
        }
        nLog2Size++;
    }
    WFSSubBlkFreeHdr *pSubBlkHdr = (WFSSubBlkFreeHdr*)&((u8*)pBlkHdr)[nOfs];
    if (pBlkHdr->aSubBlkList[nLog2Size - SBA_MIN_LOG2_SUB_BLK_SIZE].nNumFree) {
        u32 nNext = pBlkHdr->aSubBlkList[nLog2Size - SBA_MIN_LOG2_SUB_BLK_SIZE].nFirstFree;
        WFSSubBlkFreeHdr *pNext = (WFSSubBlkFreeHdr*)&((u8*)pBlkHdr)[nNext];
        pSubBlkHdr->nPrev = pNext->nPrev;
        pSubBlkHdr->nNext = (u16)nNext;
        WFSSubBlkFreeHdr *pPrev = (WFSSubBlkFreeHdr*)&((u8*)pBlkHdr)[pNext->nPrev];
        pNext->nPrev = pPrev->nNext = (u16)nOfs;
        pBlkHdr->aSubBlkList[nLog2Size - SBA_MIN_LOG2_SUB_BLK_SIZE].nNumFree++;
    } else {
        pBlkHdr->aSubBlkList[nLog2Size - SBA_MIN_LOG2_SUB_BLK_SIZE].nNumFree = 1;
        pBlkHdr->aSubBlkList[nLog2Size - SBA_MIN_LOG2_SUB_BLK_SIZE].nFirstFree = (u16)nOfs;
        pSubBlkHdr->nNext = pSubBlkHdr->nPrev = (u16)nOfs;
    }
    pSubBlkHdr->nMagic = SBA_FREE_SUB_BLK_MAGIC;
    pSubBlkHdr->nLog2Size = (u16)nLog2Size;
}

void SbaReduceSubBlkSize(WFSSubBlkAllocHdr *pBlkHdr, u16 nOfs, u32 nOldLog2Size, u32 nNewLog2Size) {
    // This function reduces the size of an allocated block to a smaller power of 2.
    // The start portion of the block will be retained.
    // As an alternative to this function, it is usually better to copy the old block
    // to a new block, then remove the old one. This function can be used when that method
    // fails because there is not enough space to allocate a new block.
    nOfs += (1<<nNewLog2Size);
    u32 nLog2Size;
    for(nLog2Size=nNewLog2Size; nLog2Size<nOldLog2Size; nLog2Size++) {
        SbaFreeSubBlk(pBlkHdr, nOfs, nLog2Size);
        nOfs += (1<<nLog2Size);
    }
}

WFSSubBlkAllocHdr *SbaBlkInit(void *pBlk, u32 nLog2BlkSize, u32 nLog2HdrSize) {
    u8 *pBlk_ = (u8*)pBlk;
    u32 nNumMaxSizeBlks = (u32)(1<<(nLog2BlkSize - SBA_MAX_LOG2_SUB_BLK_SIZE));
    WFSSubBlkAllocHdr *pBlkHdr = (WFSSubBlkAllocHdr *)pBlk;
    _ASSERT(SBA_MAX_LOG2_SUB_BLK_SIZE<nLog2BlkSize);
    memset(pBlkHdr, 0, sizeof(DirBlkHdr));
    u32 nI;
    for(nI=0; nI<nNumMaxSizeBlks; nI++) {
        WFSSubBlkFreeHdr *pSubBlkHdr = (WFSSubBlkFreeHdr *)&pBlk_[nI<<SBA_MAX_LOG2_SUB_BLK_SIZE];
        pSubBlkHdr->nNext = (u16)(((nI+1)&(nNumMaxSizeBlks-1))<<SBA_MAX_LOG2_SUB_BLK_SIZE);
        pSubBlkHdr->nPrev = (u16)(((nI-1)&(nNumMaxSizeBlks-1))<<SBA_MAX_LOG2_SUB_BLK_SIZE);
        pSubBlkHdr->nLog2Size = SBA_MAX_LOG2_SUB_BLK_SIZE;
    }
    pBlkHdr->aSubBlkList[SBA_NUM_SUB_BLK_SIZES-1].nNumFree = (u16)nNumMaxSizeBlks;
#if _DEBUG_SUB_BLK_ALLOC
    pBlkHdr->nTotalFree = (u16)(1<<nLog2BlkSize);
#endif
    SbaAllocSubBlk(pBlkHdr, nLog2HdrSize);
    return pBlkHdr;
#if _DEBUG_DIR
    //DirBlkTestAllocFree(pAreaInfo, pBlkHdr);
#endif    
}


u32 SbaCountFreeSpace(WFSSubBlkAllocHdr *pBlkHdr) {
    u32 nLog2Size, nTotalFree=0;
    for(nLog2Size=SBA_MIN_LOG2_SUB_BLK_SIZE; nLog2Size<=SBA_MAX_LOG2_SUB_BLK_SIZE; nLog2Size++) {
        u32 nNumFreeSubBlks = pBlkHdr->aSubBlkList[nLog2Size - SBA_MIN_LOG2_SUB_BLK_SIZE].nNumFree;
        nTotalFree += (1<<nLog2Size) * nNumFreeSubBlks;
    }
#if _DEBUG_SUB_BLK_ALLOC
    if (nTotalFree != pBlkHdr->nTotalFree) {
        WFSKrnOutputErrorStr("TotalFree Size Error!");
    }
#endif
    return nTotalFree;
}


u32 SbaCheckFreeSpace(WFSSubBlkAllocHdr *pBlkHdr, u32 *aUtilization) {
    u32 nLog2Size, nTotalFree=0;
    for(nLog2Size=SBA_MIN_LOG2_SUB_BLK_SIZE; nLog2Size<=SBA_MAX_LOG2_SUB_BLK_SIZE; nLog2Size++) {
        u32 nNumFreeSubBlks = pBlkHdr->aSubBlkList[nLog2Size - SBA_MIN_LOG2_SUB_BLK_SIZE].nNumFree;
        if (nNumFreeSubBlks) {
            u32 nFirstOfs = pBlkHdr->aSubBlkList[nLog2Size - SBA_MIN_LOG2_SUB_BLK_SIZE].nFirstFree;
            u32 nOfs = nFirstOfs;
            WFSSubBlkFreeHdr *pFreeSubBlk = (WFSSubBlkFreeHdr*)&((u8*)pBlkHdr)[nOfs];
            u32 nI;
            for(nI=0; nI<=nNumFreeSubBlks; nI++) {
                u32 nPrevOfs = nOfs;
                nOfs = pFreeSubBlk->nNext;
                SbaCheckOfs(nOfs);
                pFreeSubBlk = (WFSSubBlkFreeHdr*)&((u8*)pBlkHdr)[nOfs];
                if (nI>0) {
                    nTotalFree += 1<<nLog2Size;
                    if (pFreeSubBlk->nPrev != nPrevOfs) {
                        WFSKrnOutputErrorStr("Free List Error!\n");
                    }
                    if ((nI==nNumFreeSubBlks)^(nPrevOfs==nFirstOfs)) {
                        WFSKrnOutputErrorStr("Free List Error!\n");
                    }
                    if (WfsTestAndSetBitField((unsigned *)aUtilization, (int)(nOfs>>SBA_MIN_LOG2_SUB_BLK_SIZE), 1<<(nLog2Size-SBA_MIN_LOG2_SUB_BLK_SIZE))) {
                        WFSKrnOutputErrorStr("Overlap!");
                    }
                }
            }
        }
    }
#if _DEBUG_SUB_BLK_ALLOC
    if (nTotalFree != pBlkHdr->nTotalFree) {
        WFSKrnOutputErrorStr("TotalFree Size Error!");
    }
#endif
    return nTotalFree;
}


#if _DEBUG_SUB_BLK_ALLOC

    typedef struct {
        u16 nOfs;
        u16 nL2S;
    } DirSubBlk;

    static
    int CompareOfs(const void *pA, const void *pB) {
        return (int)((DirSubBlk*)pA)->nOfs - (int)((DirSubBlk*)pB)->nOfs;
    }


    static
    void SbaDebugPrintAllocs(AreaInfo *pAreaInfo, WFSSubBlkAllocHdr *pBlkHdr, u32 nNumAllocs, DirSubBlk *aSubBlk) {
        u8 aMap[1<<(WFS_LOG2_MEDIUM_BLK_SIZE-SBA_MIN_LOG2_SUB_BLK_SIZE)];
        u32 nOfs, nAlloc=0;
        u8 aChar[] = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ";
        u8 *pMap = aMap;
        u32 nTotalAllocated = 1<<WFS_NUM_BITS_REQUIRED(sizeof(DirBlkHdr));
        u32 nTotalFree = 0;
        //DirSubBlk aSorted[1<<(WFS_LOG2_MEDIUM_BLK_SIZE-SBA_MIN_LOG2_SUB_BLK_SIZE)];
        //memcpy(aSorted, aSubBlk, sizeof(DirSubBlk) * nNumAllocs);
        //qsort(aSorted, nNumAllocs, sizeof(DirSubBlk), CompareOfs);
        memset(aMap, '.', sizeof(aMap));
        for(nOfs=0; nOfs<nTotalAllocated; nOfs+=(1<<SBA_MIN_LOG2_SUB_BLK_SIZE)) {
            *pMap++ = 'h';
        }
        for(nAlloc=0; nAlloc<nNumAllocs; nAlloc++) {
            pMap = &aMap[aSubBlk[nAlloc].nOfs>>SBA_MIN_LOG2_SUB_BLK_SIZE];
            if (aSubBlk[nAlloc].nL2S) {
                nTotalAllocated += 1<<aSubBlk[nAlloc].nL2S;
                for(nOfs=aSubBlk[nAlloc].nOfs; nOfs<aSubBlk[nAlloc].nOfs+(1<<aSubBlk[nAlloc].nL2S); nOfs+=(1<<SBA_MIN_LOG2_SUB_BLK_SIZE)) {
                    *pMap++ = aChar[nAlloc%(sizeof(aChar)-1)];
                }
            }
        }
        u32 nLog2Size;
        for(nLog2Size=SBA_MIN_LOG2_SUB_BLK_SIZE; nLog2Size<=SBA_MAX_LOG2_SUB_BLK_SIZE; nLog2Size++) {
            u32 nNumFreeSubBlks = pBlkHdr->aSubBlkList[nLog2Size - SBA_MIN_LOG2_SUB_BLK_SIZE].nNumFree;
            if (nNumFreeSubBlks) {
                u32 nFirstOfs = pBlkHdr->aSubBlkList[nLog2Size - SBA_MIN_LOG2_SUB_BLK_SIZE].nFirstFree;
                nOfs = nFirstOfs;
                WFSSubBlkFreeHdr *pFreeSubBlk = (WFSSubBlkFreeHdr*)&((u8*)pBlkHdr)[nOfs];
                u32 nI;
                for(nI=0; nI<=nNumFreeSubBlks; nI++) {
                    u32 nPrevOfs = nOfs;
                    nOfs = pFreeSubBlk->nNext;
                    pFreeSubBlk = (WFSSubBlkFreeHdr*)&((u8*)pBlkHdr)[nOfs];
                    if (nI>0) {
                        nTotalFree += 1<<nLog2Size;
                        if (pFreeSubBlk->nPrev != nPrevOfs) {
                            WFSKrnOutputErrorStr("Free List Error!\n");
                        }
                        if ((nI==nNumFreeSubBlks)^(nPrevOfs==nFirstOfs)) {
                            WFSKrnOutputErrorStr("Free List Error!\n");
                        }
                        pMap = &aMap[nOfs>>SBA_MIN_LOG2_SUB_BLK_SIZE];
                        u32 nMapOfs;
                        for(nMapOfs=nPrevOfs; nMapOfs<nPrevOfs+(1<<nLog2Size); nMapOfs+=(1<<SBA_MIN_LOG2_SUB_BLK_SIZE)) {
                            if (*pMap != '.') {
                                WFSKrnOutputErrorStr("Free Map Error!\n");
                            }
                            *pMap++ = (u8)('0' + nLog2Size - SBA_MIN_LOG2_SUB_BLK_SIZE);
                        }
                    }
                }
            }
        }
        if ((nTotalFree + nTotalAllocated) != pAreaInfo->nBlkSize) {
            MyOSReport("TotalFree(%d) + TotalAllocated(%d) != BlkSize(%d)%d\n", nTotalFree, nTotalAllocated, pAreaInfo->nBlkSize);
        }
        for(nOfs=0; nOfs<pAreaInfo->nBlkSize; nOfs+=(1<<SBA_MAX_LOG2_SUB_BLK_SIZE)) {
            u8 *pLineEnd = &aMap[(nOfs+(1<<SBA_MAX_LOG2_SUB_BLK_SIZE))>>SBA_MIN_LOG2_SUB_BLK_SIZE];
            u8 cTemp = *pLineEnd;
            *pLineEnd = 0;
            MyOSReport("%s\n", &aMap[nOfs>>SBA_MIN_LOG2_SUB_BLK_SIZE]);
            *pLineEnd = cTemp;
        }
        MyOSReport("\n");
    }

    static
    void SbaTestAllocFree(AreaInfo *pAreaInfo, WFSSubBlkAllocHdr *pBlkHdr) {
        // Test making random allocs and frees of sub blocks
        DirSubBlk aSubBlk[1<<(WFS_LOG2_MEDIUM_BLK_SIZE-SBA_MIN_LOG2_SUB_BLK_SIZE)];
        u32 aNumAllocsPerSize[WFS_LOG2_MEDIUM_BLK_SIZE+2];
        //u32 nMaxSubBlksPerBlk = (u32)(1<<(pAreaInfo->ah.nLog2BlkSize - SBA_MIN_LOG2_SUB_BLK_SIZE)); // unused
        u32 nTotalAllocated = 1<<WFS_NUM_BITS_REQUIRED(sizeof(DirBlkHdr));
        u32 nNumAllocs = 0;
        u32 nLog2Size;
        u16 nOfs;
        memset(aNumAllocsPerSize, 0, sizeof(aNumAllocsPerSize));
        u32 nTest;
        for(nTest=0; nTest<100000; nTest++) {
            u32 nAlloc = nNumAllocs;
            nNumAllocs += RandomInt(1, 20);
            for(; nAlloc<nNumAllocs;) {
                // Generate more small allocs and less large allocs with an exponential fall-off distribution:
                nLog2Size = (u32)(SBA_MAX_LOG2_SUB_BLK_SIZE + 1 - WfsCalculateFieldSizeInBits((u32)RandomInt(1, (2<<(SBA_MAX_LOG2_SUB_BLK_SIZE-SBA_MIN_LOG2_SUB_BLK_SIZE))-1 ) ));
                nOfs = SbaAllocSubBlk(pBlkHdr, nLog2Size);
                if (nOfs) {
                    aNumAllocsPerSize[nLog2Size]++;
                    nTotalAllocated += 1<<nLog2Size;
                    aSubBlk[nAlloc].nOfs = nOfs;
                    aSubBlk[nAlloc].nL2S = (u16)nLog2Size;
                    u32 *pPtr = (u32*)&((u8*)pBlkHdr)[nOfs];
                    u32 nI;
                    for(nI=0; nI<(1<<(nLog2Size-2)); nI++) {
                        *pPtr++ = (0x11110000 * nLog2Size) + nOfs;
                    }
                    nAlloc++;
                    SbaDebugPrintAllocs(pAreaInfo, pBlkHdr, nAlloc, aSubBlk);
                } else {
                    nNumAllocs--;
                }
            }
            u32 nNumFrees = (u32)RandomInt(1, 20);
            if (nNumFrees > nNumAllocs) {
                nNumFrees = nNumAllocs;
            }
            while(nNumFrees) {
                u32 nHalfNumFrees = (nNumFrees+1)>>1;
                u32 nFree=0;
                while(nFree<nHalfNumFrees) {
                    u32 nFreeIdx;
                    do {
                        nFreeIdx = (u32)RandomInt(0, (s32)(nNumAllocs-1));
                    } while(aSubBlk[nFreeIdx].nL2S==0);
                    nOfs = aSubBlk[nFreeIdx].nOfs;
                    nLog2Size = aSubBlk[nFreeIdx].nL2S;
                    u32 *pPtr = (u32*)&((u8*)pBlkHdr)[nOfs];
                    u32 nI;
                    for(nI=0; nI<(1<<(nLog2Size-2)); nI++) {
                        if (*pPtr != (0x11110000 * nLog2Size) + nOfs) {
                            WFSKrnOutputErrorStr("Alloc/Free Error!\n");
                        }
                        *pPtr++ = 0;
                    }
                    SbaFreeSubBlk(pBlkHdr, nOfs, nLog2Size);
                    aSubBlk[nFreeIdx].nL2S=0;
                    nTotalAllocated -= 1<<nLog2Size;
                    nFree++;
                    SbaDebugPrintAllocs(pAreaInfo, pBlkHdr, nNumAllocs, aSubBlk);
                }
                u32 nI=0;
                u32 nJ=0;
                for(nJ=0; nJ<nNumAllocs; nJ++) {
                    aSubBlk[nI] = aSubBlk[nJ];
                    if (aSubBlk[nI].nL2S) {
                        nI++;
                    }
                }
                nNumAllocs -= nHalfNumFrees;
                nNumFrees -= nHalfNumFrees;
            }
            u32 nJ;
            for(nJ=0; nJ<nNumAllocs; nJ++) {
                nOfs = aSubBlk[nJ].nOfs;
                nLog2Size = aSubBlk[nJ].nL2S;
                u32 *pPtr = (u32*)&((u8*)pBlkHdr)[nOfs];
                u32 nI;
                for(nI=0; nI<(1<<(nLog2Size-2)); nI++) {
                    if (*pPtr++ != (0x11110000 * nLog2Size) + nOfs) {
                        WFSKrnOutputErrorStr("Alloc/Free Error!\n");
                    }
                }
            }
        }
        u32 nI;
        for(nI=0; nI<20; nI++) {
            MyOSReport("%d: %d\n", nI, aNumAllocsPerSize[nI]);
        }
    }

#endif
