/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfskrn_Trans.c - Provides access to blocks for transactions including re-mapping functionality

  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_Trans.cpp,v $
  Revision 1.22  2008/12/18 09:03:10  kondo_masahiro
  Fixed function name and process from BCacheSetBlkAdr() to BCacheRemap()

  Revision 1.21  2008/12/18 06:31:56  ooizumi
  Modified to use debug title id when PERM_TEST flag is enabled.

  Revision 1.20  2008/12/17 01:37:27  kondo_masahiro
  Added function BCacheSetBlkAdr()

  Revision 1.19  2008/11/26 01:41:04  kondo_masahiro
  Fixed arguments of several functions in order to improve access speed.

  Revision 1.18  2008/11/05 05:53:11  kondo_masahiro
  Fixed the end of line made to newline.

  Revision 1.17  2008/10/21 09:49:36  kondo_masahiro
  Added new function WFSWriteFileAndDiscard

  Revision 1.16  2008/10/17 08:51:10  kondo_masahiro
  Added the new MEDIUM size category (40KB-320KB)

  Revision 1.15  2008/10/14 10:26:24  kondo_masahiro
  Added codes to access very large size file data

  Revision 1.14  2008/10/10 04:05:35  kondo_masahiro
  Fixed change arguments of AreaFreeBlks

  Revision 1.13  2008/10/09 12:01:11  kondo_masahiro
  Fixed a bug in TransWriteFileLargeBlks (a argument of BCacheInvalidateBlks)

  Revision 1.12  2008/10/09 04:16:53  kondo_masahiro
  Fixed member of AreaHdr and AreaInfo

  Revision 1.11  2008/10/09 01:44:22  ueno
  Added workarounds.

  Revision 1.10  2008/10/08 23:19:54  kondo_masahiro
  Added codes to access large size file data

  Revision 1.9  2008/10/08 11:46:31  ueno
  Fixed a typo.

  Revision 1.8  2008/10/08 11:45:36  ueno
  Added a workaround to TransGetBlk5() to avoid exception.

  Revision 1.7  2008/10/07 05:59:43  ueno
  Added error handling in TransGetBlk5().

  Revision 1.6  2008/10/03 08:37:56  kondo_masahiro
  Fixed codes for porting IOP

  Revision 1.5  2008/09/28 23:32:04  kondo_masahiro
  Fixed codes to use medium size user block.

  Revision 1.4  2008/08/27 09:44:15  paul
  Started coding to differentiate File/Meta-data blocks

  Revision 1.3  2008/08/05 03:56:16  kondo_masahiro
  Added error handling for hash error and detach device.

  Revision 1.2  2008/07/25 02:52:36  paul
  Added some functions for allocating TransInfo

  Revision 1.1  2008/07/09 00:50:57  paul
  First checkin

*---------------------------------------------------------------------------*/

#include "wfskrn_Config.h"
#include "wfskrn_Trans.h"
#include "wfskrn_Api.h"


void TransModuleInit() {
    dbgv(osTPrintf("TransModuleInit()\n"));
    // Active list is doubly linked and circular
    // Free list is singly linked and NULL terminated
    wkg.pActiveTransListAnchor = (TransInfo *)((u8*)&wkg.activeTransListAnchor - (size_t)&((TransInfo *)0)->link);
    wkg.pActiveTransListAnchor->link.pNext = wkg.pActiveTransListAnchor->link.pPrev = wkg.pActiveTransListAnchor;
    TransInfo *pTransInfo = wkg.pTransInfoFreeList = wkg.aTransInfo;
    TransInfo *pTransEnd = (TransInfo *)((u8*)pTransInfo + sizeof(wkg.aTransInfo));
    for(; pTransInfo<pTransEnd; ++pTransInfo) {
        pTransInfo->link.pNext = pTransInfo+1;
    }
    pTransInfo[-1].link.pNext = NULL;
}


WFSTitleId WfsGetRequestorTitleId() {
#if _WIN32 || PERM_TEST
    return wkg.nDebugTitleId;
#elif _IOP
    //return IOS_GetTitleId();
    return IOS_GetGid(); // ??? must be fixed
#endif
}


void TransAddToActiveList(TransInfo *pTransInfo) {
    pTransInfo->link.pNext = wkg.pActiveTransListAnchor->link.pNext;
    pTransInfo->link.pNext->link.pPrev = pTransInfo;
    wkg.pActiveTransListAnchor->link.pNext = pTransInfo;
    pTransInfo->link.pPrev = wkg.pActiveTransListAnchor;
}


void TransRemoveFromActiveList(TransInfo *pTransInfo) {
    pTransInfo->link.pPrev->link.pNext   = pTransInfo->link.pNext;
    pTransInfo->link.pNext->link.pPrev   = pTransInfo->link.pPrev;
}


TransInfo *TransBegin() {
    TransInfo *pTransInfo = wkg.pTransInfoFreeList;
    wkg.pTransInfoFreeList = pTransInfo->link.pNext;
    pTransInfo->nTitleId = WfsGetRequestorTitleId();
    TransAddToActiveList(pTransInfo);
    return pTransInfo;
}


void TransEnd(TransInfo *pTransInfo) {
    TransRemoveFromActiveList(pTransInfo);
    pTransInfo->link.pNext = wkg.pTransInfoFreeList;
    wkg.pTransInfoFreeList = pTransInfo;
}


WFSKrnResult TransGetBlk5(AreaInfo *pAreaInfo, WFSBlkAdr nBlkAdr, TransInfo *pTransInfo, u32 nFlags, void **ppBlkPtr) {
    if (!pAreaInfo)
    {
        return WFSKRN_RESULT_INVALID; // [check] workaround to avoid exception when DirGetBlk() is called for "/".
    }
    u32 nAbsBlkAdr = pAreaInfo->nAbsStartBlkAdr + nBlkAdr;

#if AREA_ENABLE_BLK_CACHE
    BCacheEntry* pBce;
    WFSKrnResult nResult = BCacheAllocMetaBlk(pAreaInfo->pVolInfo, nAbsBlkAdr, nFlags, &pBce);
    if(ppBlkPtr && nResult == WFSKRN_RESULT_OK)
    {
        *ppBlkPtr = pBce->pBlkPtr;
    }
    return nResult;
#else //AREA_ENABLE_BLK_CACHE
    *ppBlkPtr = AreaGetPhysicalBlk(pAreaInfo, nAbsBlkAdr);
    return WFSKRN_RESULT_OK;
#endif //AREA_ENABLE_BLK_CACHE
}


WFSKrnResult TransGetAndPinBlk5(AreaInfo *pAreaInfo, WFSBlkAdr nBlkAdr, TransInfo *pTransInfo, u32 nFlags, void **ppBlkPtr) {
    if (!pAreaInfo)
    {
        return WFSKRN_RESULT_INVALID; // [check] workaournd in case DirGetBlk() is called for "/".
    }
    // ToDo: Perform a mapping from physical block to remapped block as seen by a particular transaction
    u32 nAbsBlkAdr = pAreaInfo->nAbsStartBlkAdr + nBlkAdr;
#if AREA_ENABLE_BLK_CACHE
    BCacheEntry* pBce;
    WFSKrnResult nResult = BCacheAllocMetaBlk(pAreaInfo->pVolInfo, nAbsBlkAdr, nFlags | BCACHE_FLAG_PINNED, &pBce);
    if(ppBlkPtr) *ppBlkPtr = pBce->pBlkPtr;
    return nResult;
#else //AREA_ENABLE_BLK_CACHE
    return AreaGetPhysicalBlk(pAreaInfo, nAbsBlkAdr);
#endif //AREA_ENABLE_BLK_CACHE
}


WFSKrnResult TransSetChainId4(AreaInfo *pAreaInfo, WFSBlkAdr nBlkAdr, TransInfo *pTransInfo, u32 nChainId){
    if (!pAreaInfo){
        return WFSKRN_RESULT_INVALID; // [check] workaournd in case DirGetBlk() is called for "/".
    }
    u32 nAbsBlkAdr = pAreaInfo->nAbsStartBlkAdr + nBlkAdr;
    return BCacheSetChainId(pAreaInfo->pVolInfo, nAbsBlkAdr, nChainId);
}


WFSKrnResult TransGetChainId4(AreaInfo *pAreaInfo, WFSBlkAdr nBlkAdr, TransInfo *pTransInfo, u32 *pChainId){
    if (!pAreaInfo){
        return WFSKRN_RESULT_INVALID; // [check] workaournd in case DirGetBlk() is called for "/".
    }
    u32 nAbsBlkAdr = pAreaInfo->nAbsStartBlkAdr + nBlkAdr;
    return BCacheGetChainId(pAreaInfo->pVolInfo, nAbsBlkAdr, pChainId);
}



WFSKrnResult TransUnpinBlk3(AreaInfo *pAreaInfo, WFSBlkAdr nBlkAdr, TransInfo *pTransInfo) {
#if AREA_ENABLE_BLK_CACHE
    u32 nAbsBlkAdr = pAreaInfo->nAbsStartBlkAdr + nBlkAdr;
    WFSKrnResult nResult = BCacheUnpin(pAreaInfo->pVolInfo, nAbsBlkAdr);
  #if _CHECK_BLK_USAGE
    // if nPinCount becomes 0, the cache data is flushed. (for debug)
    if(!BCacheGetPinCount(pAreaInfo->pVolInfo, nAbsBlkAdr)){
        if(DeviceDebugGetPseudoDetachDevice() == 0){
            DeviceDebugSetPseudoDetachDevice(1);
        }
        if(DeviceDebugGetPseudoHashInconsistent() == 0){
            DeviceDebugSetPseudoHashInconsistent(1);
        }
        nResult = AreaBlkFlush(pAreaInfo, nBlkAdr);
    }
  #endif
    return nResult;
#else //AREA_ENABLE_BLK_CACHE
    return WFSKRN_RESULT_OK;
#endif //AREA_ENABLE_BLK_CACHE
}


WFSKrnResult TransDirtyBlk3(AreaInfo *pAreaInfo, WFSBlkAdr nBlkAdr, TransInfo *pTransInfo) {
#if AREA_ENABLE_BLK_CACHE
    u32 nAbsBlkAdr = pAreaInfo->nAbsStartBlkAdr + nBlkAdr;
    WFSKrnResult nResult = BCacheDirty(pAreaInfo->pVolInfo, nAbsBlkAdr);
    return nResult;
#else //AREA_ENABLE_BLK_CACHE
    return WFSKRN_RESULT_OK;
#endif //AREA_ENABLE_BLK_CACHE
}


#include "wfskrn_Volume.h"

WFSKrnResult TransFreeBlks6(AreaInfo *pAreaInfo, u32 nLog2BlkSize, u32 nNumBlks, WFSBlkAdr *pBlkAdr, s32 nStride, TransInfo *pTransInfo) {
    // This function frees the transaction blocks passed in.
    WFSKrnResult nResult = WFSKRN_RESULT_OK;
#if AREA_ENABLE_BLK_CACHE
    AreaFreeBlks(pAreaInfo, nLog2BlkSize, nNumBlks, pBlkAdr, nStride);
#else //AREA_ENABLE_BLK_CACHE
    u32 nI;
    for(nI=0; nI<nNumBlks; nI++) {
        WFSBlkAdr *pNextFreeBlock = (WFSBlkAdr *)AreaGetPhysicalBlk(pAreaInfo, *pBlkAdr);
        *pNextFreeBlock = pAreaInfo->ah.nFirstFreeBlkAdr;
        pAreaInfo->ah.nFirstFreeBlkAdr = *pBlkAdr;
        pBlkAdr = (WFSBlkAdr*)((size_t)pBlkAdr + nStride);
    }
#endif //AREA_ENABLE_BLK_CACHE
    return nResult;
}


WFSKrnResult TransUnpinAndFreeBlks(AreaInfo *pAreaInfo, u32 nLog2BlkSize, u32 nNumBlks, WFSBlkAdr aBlkAdr[], TransInfo *pTransInfo) {
    // This function frees the transaction blocks passed in.
    WFSKrnResult nResult = WFSKRN_RESULT_OK;
    u32 nI;
#if AREA_ENABLE_BLK_CACHE
    for(nI=0; nI<nNumBlks; nI++) {
        TransUnpinBlk3(pAreaInfo, aBlkAdr[nI], pTransInfo);
    }
    nResult = AreaFreeBlks(pAreaInfo, nLog2BlkSize, nNumBlks, aBlkAdr, sizeof(WFSBlkAdr));
#else //AREA_ENABLE_BLK_CACHE
    for(nI=0; nI<nNumBlks; nI++) {
        WFSBlkAdr *pNextFreeBlock = (WFSBlkAdr *)AreaGetPhysicalBlk(pAreaInfo, aBlkAdr[nI]);
        *pNextFreeBlock = pAreaInfo->ah.nFirstFreeBlkAdr;
        pAreaInfo->ah.nFirstFreeBlkAdr = aBlkAdr[nI];
    }
#endif //AREA_ENABLE_BLK_CACHE
    return nResult;
}


WFSKrnResult TransStoreBlk3(AreaInfo *pAreaInfo, WFSBlkAdr nBlkAdr, TransInfo *pTransInfo) {
    WFSKrnResult nResult = WFSKRN_RESULT_OK;
#if AREA_ENABLE_BLK_CACHE
    nResult = AreaBlkStore (pAreaInfo, nBlkAdr);
#endif //AREA_ENABLE_BLK_CACHE
    return nResult;
}


WFSKrnResult TransFlushBlk3(AreaInfo *pAreaInfo, WFSBlkAdr nBlkAdr, TransInfo *pTransInfo) {
    WFSKrnResult nResult = WFSKRN_RESULT_OK;
#if AREA_ENABLE_BLK_CACHE
    nResult = AreaBlkFlush (pAreaInfo, nBlkAdr);
#endif //AREA_ENABLE_BLK_CACHE
    return nResult;
}


WFSKrnResult TransInvalidate3(AreaInfo *pAreaInfo, WFSBlkAdr nBlkAdr, TransInfo *pTransInfo) {
    WFSKrnResult nResult = WFSKRN_RESULT_OK;
#if AREA_ENABLE_BLK_CACHE
    nResult = AreaBlkInvalidate(pAreaInfo, nBlkAdr);
#endif //AREA_ENABLE_BLK_CACHE
    return nResult;
}


WFSKrnResult TransSetValidSize4(AreaInfo *pAreaInfo, WFSBlkAdr nBlkAdr, TransInfo *pTransInfo, u32 nValidSize){
    WFSKrnResult nResult = WFSKRN_RESULT_OK;
#if AREA_ENABLE_BLK_CACHE
    nResult = AreaBlkSetValidSize(pAreaInfo, nBlkAdr, nValidSize);
#endif //AREA_ENABLE_BLK_CACHE
    return nResult;    
}


WFSKrnResult TransRemapCache6(AreaInfo *pAreaInfo, WFSBlkAdr nSrcBlkAdr, TransInfo *pTransInfo, WFSBlkAdr nDstBlkAdr, WFSBlkAdr nHashBlkAdr, u32 nHashOfs)
{
    WFSKrnResult nResult = WFSKRN_RESULT_OK;
#if AREA_ENABLE_BLK_CACHE
    nResult = AreaBlkRemapCache(pAreaInfo, nSrcBlkAdr, nDstBlkAdr, nHashBlkAdr, nHashOfs);
#endif //AREA_ENABLE_BLK_CACHE
    return nResult;    
}


WFSKrnResult TransGetAndPinFileBlk8(AreaInfo *pAreaInfo, WFSBlkAdr nParentBlkAdr, u16 nHashOfs, WFSBlkAdr nFileBlkAdr, TransInfo *pTransInfo, u32 nFlags, u32 nValidSize, u32 nBCacheGroupId, void **ppBlkPtr) {
    // This function ensures the file block and its parent meta data block are in cache, and pins the 
    // It returns the address of
    // Arguments:
    //   pAreaInfo        [In] .. Pointer to the AreaInfo struct which contains the blocks to be written
    //   nParentBlkAdr    [In] .. Block address of the parent meta data block
    //   nHashOfs         [In] .. Offset to the hash code within the parent meta data block
    //   nFileBlkAdr      [In] .. Block address of the file block
    //   pTransInfo       [In] .. Pointer to a TransInfo struct for the transaction (may be used to remap blocks).
    //   nFlags           [In] .. Flags TRANS_FLAG_* and BCACHE_FLAG_*
    //   ppBlkPtr        [Out] .. pointer to variable which will be overwritten with the address of the file block in the cache

    // ***ToDo: Implement this function properly including necessary changes to block cache***

    // ToDo: Perform a mapping from physical block to remapped block as seen by a particular transaction
    u32 nAbsBlkAdr  = pAreaInfo->nAbsStartBlkAdr + nFileBlkAdr;
    u32 nHashBlkAdr = pAreaInfo->nAbsStartBlkAdr + nParentBlkAdr;
#if AREA_ENABLE_BLK_CACHE
    BCacheEntry* pBce;
    WFSKrnResult nResult = BCacheAlloc(pAreaInfo->pVolInfo, nAbsBlkAdr, nFlags | BCACHE_FLAG_PINNED, nValidSize, nBCacheGroupId, nHashBlkAdr, nHashOfs, &pBce);
    if(nResult == WFS_RESULT_OK && ppBlkPtr){
        *ppBlkPtr = pBce->pBlkPtr;
    }
    return nResult;
#else //AREA_ENABLE_BLK_CACHE
    return AreaGetPhysicalBlk(pAreaInfo, nAbsBlkAdr);
#endif //AREA_ENABLE_BLK_CACHE

}


WFSKrnResult TransReadFileBlks7(u8 *pAlignedPtr, AreaInfo *pAreaInfo, u32 nLog2BlkSize, u32 nSize, WFSFileBlkPtr *pFileBlkPtr, s32 nStride, TransInfo *pTransInfo) {
    // This function reads, decrypts and hash checks a set of file blocks directly to a specified aligned buffer
    // Arguments:
    //   pAlignedPtr     [Out] .. Address of aligned buffer to overwrite with data from file blocks. (must be aligned to WFS_STATIC_MEM_ALIGNMENT bytes)
    //   pAreaInfo        [In] .. Pointer to the AreaInfo struct which contains the blocks to be written
    //   nNumBlks         [In] .. The number of blocks to read
    //   pFileBlkPtr      [In] .. Pointer to the first WFSFileBlkPtr struct, which specifies the address and hash code of the file block
    //   nStride          [In] .. How to get from the first WFSFileBlkPtr to the next one
    //   pTransInfo       [In] .. Pointer to a TransInfo struct for the transaction (may be used to remap blocks).
    //

    // ToDo: Check if any blocks have been remapped for this transaction
    // ToDo: Check if any (remapped) blocks are already in block cache, and read from there if so
    // ToDo: Determine how many blocks are sequential (after optional remapping), so they can be read with a single combined read

    DeviceInfo *pDevInfo = pAreaInfo->pVolInfo->pDevInfo;
    ASSERT(nSize);
    ASSERT(!(nSize&(pDevInfo->dh.nLog2SectorSize-1)));
    u32 nLog2SectorsPerBlk = pAreaInfo->ah.nLog2BlkSize - pDevInfo->dh.nLog2SectorSize;
    u32 nLog2SectorsPerReadBlk = nLog2BlkSize - pDevInfo->dh.nLog2SectorSize;
    u32 nLog2BlkPerReadBlk = nLog2SectorsPerReadBlk - nLog2SectorsPerBlk;
    u32 nNumBlks = ((nSize-1)>>nLog2BlkSize)+1;
    u32 nEndBlkSector = (nSize>>pDevInfo->dh.nLog2SectorSize)-((nNumBlks-1)<<nLog2SectorsPerReadBlk);
    u32 nStartIdx=0;
    while(nStartIdx < nNumBlks) {
        // Starting from the current block, look for a run of sequential blocks and count its length
        WFSFileBlkPtr *pStartFileBlkPtr = pFileBlkPtr;
        u32 nRunLength = 1;
        while(nStartIdx + nRunLength < nNumBlks) {
            pFileBlkPtr = (WFSFileBlkPtr *)((u8*)pFileBlkPtr + nStride);
            if (pFileBlkPtr->nBlkAdr != pStartFileBlkPtr->nBlkAdr + (nRunLength<<nLog2BlkPerReadBlk)) {
                break;
            }
            ++nRunLength;
        }
        // Read (nRunLength) blocks from the device to the destination buffer, then do decryption & hash check on each block
        WFSBlkAdr nAbsBlkAdr = pAreaInfo->nAbsStartBlkAdr + pStartFileBlkPtr->nBlkAdr;
#if _DEBUG_BREAK_POINT
        if(nAbsBlkAdr == 32248){
            int a; a=0;
        }
#endif
        u32 nReadSector;
        if(nStartIdx + nRunLength == nNumBlks){
            nReadSector = ((nRunLength-1)<<nLog2SectorsPerReadBlk)+nEndBlkSector;
        } else {
            nReadSector = nRunLength<<nLog2SectorsPerReadBlk;
        }
        WFSKrnResult nResult = DeviceRead(pDevInfo, nAbsBlkAdr<<nLog2SectorsPerBlk, pAlignedPtr, nReadSector, &pStartFileBlkPtr->hash, nLog2SectorsPerReadBlk, 0, nStride, 0);
        if(nResult == WFSKRN_RESULT_DEVICE_HASH_INCONSISTENT){
            osTPrintf("User Block Hash Error: %d\n", nAbsBlkAdr);
        }
        pAlignedPtr += nReadSector << pDevInfo->dh.nLog2SectorSize;
        nStartIdx += nRunLength;
    }
    return WFSKRN_RESULT_OK;
}


WFSKrnResult TransWriteFileBlks7(u8 *pAlignedPtr, AreaInfo *pAreaInfo, u32 nLog2BlkSize, u32 nSize, WFSFileBlkPtr *pFileBlkPtr, s32 nStride, TransInfo *pTransInfo, bool bDecryption) {
    // This function reads, decrypts and hash checs a set of file blocks directly to a specified aligned buffer
    //   Arguments:
    //   pAlignedPtr     [Out] .. Address of aligned buffer to overwrite with data from file blocks. (must be aligned to WFS_STATIC_MEM_ALIGNMENT bytes)
    //   pAreaInfo        [In] .. Pointer to the AreaInfo struct which contains the blocks to be written
    //   nNumBlks         [In] .. The number of blocks to read
    //   pFileBlkPtr      [In] .. Pointer to the first WFSFileBlkPtr struct, which specifies the address and hash code of the file block
    //   nStride          [In] .. How to get from the first WFSFileBlkPtr to the next one
    //   pTransInfo       [In] .. Pointer to a TransInfo struct for the transaction (may be used to remap blocks).

    // ToDo: Check if any blocks have been remapped for this transaction, and write to the remapped blocks 
    // ToDo: First, invalidate the (remapped) blocks we are writing to directly, to ensure we don't get stale data in the block cache
    // ToDo: Determine how many blocks are sequential (after optional remapping), so they can be written with a single combined write

    DeviceInfo *pDevInfo = pAreaInfo->pVolInfo->pDevInfo;
    ASSERT(nSize);
    ASSERT(!(nSize&(pDevInfo->dh.nLog2SectorSize-1)));
    u32 nLog2SectorsPerBlk = pAreaInfo->ah.nLog2BlkSize - pDevInfo->dh.nLog2SectorSize;
    u32 nLog2SectorsPerWriteBlk = nLog2BlkSize - pDevInfo->dh.nLog2SectorSize;
    u32 nLog2BlkPerWriteBlk = nLog2SectorsPerWriteBlk - nLog2SectorsPerBlk;
    u32 nNumBlks = ((nSize-1)>>nLog2BlkSize)+1;
    u32 nEndBlkSector = (nSize>>pDevInfo->dh.nLog2SectorSize)-((nNumBlks-1)<<nLog2SectorsPerWriteBlk);
    u32 nStartIdx=0;
    while(nStartIdx < nNumBlks) {
        // Starting from the current block, look for a run of sequential blocks and count its length
        WFSFileBlkPtr *pStartFileBlkPtr = pFileBlkPtr;
        u32 nRunLength = 1;
        while(nStartIdx + nRunLength < nNumBlks) {
            pFileBlkPtr = (WFSFileBlkPtr *)((u8*)pFileBlkPtr + nStride);
#if _DEBUG_BREAK_POINT
            if (pFileBlkPtr->nBlkAdr == 0x1c5a){
                int a; a=0;
            }
#endif
            if (pFileBlkPtr->nBlkAdr != pStartFileBlkPtr->nBlkAdr + (nRunLength<<nLog2BlkPerWriteBlk)) {
                break;
            }
            ++nRunLength;
        }
        // Write (nRunLength) blocks from the device to the destination buffer, then do decryption & hash check on each block
        WFSBlkAdr nAbsBlkAdr = pAreaInfo->nAbsStartBlkAdr + pStartFileBlkPtr->nBlkAdr;
#if _DEBUG_BREAK_POINT
        if(nAbsBlkAdr == 32248){
            int a; a=0;
        }
#endif
        u32 nWriteSector;
        if(nStartIdx + nRunLength == nNumBlks){
            nWriteSector = ((nRunLength-1)<<nLog2SectorsPerWriteBlk)+nEndBlkSector;
        } else {
            nWriteSector = nRunLength<<nLog2SectorsPerWriteBlk;
        }
        // The caches of the written blocks have to be Invalidated because BCache data would be old.
        WFSKrnReturnOnError(BCacheInvalidateBlks(pAreaInfo->pVolInfo, nAbsBlkAdr, (nRunLength<<nLog2BlkPerWriteBlk)));
        WFSKrnReturnOnError(DeviceWrite(pDevInfo, nAbsBlkAdr<<nLog2SectorsPerBlk, pAlignedPtr, nWriteSector, &pStartFileBlkPtr->hash, nLog2SectorsPerWriteBlk, 0, nStride, 0, bDecryption));
        pAlignedPtr += nWriteSector << pDevInfo->dh.nLog2SectorSize;
        nStartIdx += nRunLength;
    }
    return WFSKRN_RESULT_OK;
}


WFSKrnResult TransReadFileLargeBlks (u8 *pAlignedPtr, AreaInfo *pAreaInfo, u32 nSize, u32 nOffsetBlks, 
                                     WFSFileLargeBlkPtr *pFileLargeBlkPtr, s32 nStride, TransInfo *pTransInfo)
{
    // This function reads, decrypts and hash checks a set of file blocks directly to a specified aligned buffer
    // Arguments:
    //   pAlignedPtr     [Out] .. Address of aligned buffer to overwrite with data from file blocks. (must be aligned to WFS_STATIC_MEM_ALIGNMENT bytes)
    //   pAreaInfo        [In] .. Pointer to the AreaInfo struct which contains the blocks to be written
    //   nNumBlks         [In] .. The number of blocks to read 
    //   nOffsetBlks      [In] .. The number of offset blocks for first read block  
    //   pFileBlkPtr      [In] .. Pointer to the first WFSFileBlkPtr struct, which specifies the address and hash code of the file block
    //   nStride          [In] .. How to get from the first WFSFileLargeBlkPtr to the next one
    //   pTransInfo       [In] .. Pointer to a TransInfo struct for the transaction (may be used to remap blocks).

    DeviceInfo *pDevInfo = pAreaInfo->pVolInfo->pDevInfo;
    ASSERT(nSize);
    ASSERT(!(nSize&(pDevInfo->dh.nLog2SectorSize-1)));
    u32 nLog2SectorsPerBlk = pAreaInfo->ah.nLog2BlkSize - pDevInfo->dh.nLog2SectorSize;
    u32 nLog2SectorsPerMaxHashBlk = pAreaInfo->ah.nLog2MediumBlkSize - pDevInfo->dh.nLog2SectorSize;
    u32 nNumBlks = ((nSize-1)>>pAreaInfo->ah.nLog2MediumBlkSize)+1;
    u32 nEndBlkSector = (nSize>>pDevInfo->dh.nLog2SectorSize)-((nNumBlks-1)<<nLog2SectorsPerMaxHashBlk);
    u32 nStartIdx=0;
    u32 nOffsetBlksTmp = nOffsetBlks;
    while(0 < nNumBlks) {
        // Starting from the current block, look for a run of sequential blocks and count its length
        WFSFileLargeBlkPtr *pStartFileLargeBlkPtr = pFileLargeBlkPtr;
        u32 nRunLength = 0;
        u32 nReadSector;
        do{
            if(nNumBlks + nOffsetBlksTmp > (1<<pAreaInfo->nLog2MediumBlkPerLargeBlk)){
                nRunLength += (1<<pAreaInfo->nLog2MediumBlkPerLargeBlk) - nOffsetBlksTmp;
                nNumBlks -= (1<<pAreaInfo->nLog2MediumBlkPerLargeBlk) - nOffsetBlksTmp;
                nReadSector = nRunLength<<nLog2SectorsPerMaxHashBlk;
                nOffsetBlksTmp = 0;
            } else {
                nRunLength += nNumBlks;
                nReadSector = ((nRunLength-1)<<nLog2SectorsPerMaxHashBlk)+nEndBlkSector;
                nNumBlks = 0;
                break;
            }
            pFileLargeBlkPtr = (WFSFileLargeBlkPtr *)((u8*)pFileLargeBlkPtr + nStride);
        }while( pFileLargeBlkPtr->nBlkAdr == pStartFileLargeBlkPtr->nBlkAdr + ((nRunLength + nOffsetBlks)<<pAreaInfo->nLog2BlksPerMediumBlk));
        // Read (nRunLength) blocks from the device to the destination buffer, then do decryption & hash check on each block
        WFSBlkAdr nAbsBlkAdr = pAreaInfo->nAbsStartBlkAdr + pStartFileLargeBlkPtr->nBlkAdr + (nOffsetBlks<<pAreaInfo->nLog2BlksPerMediumBlk);
        WFSKrnResult nResult = DeviceRead(pDevInfo, nAbsBlkAdr<<nLog2SectorsPerBlk, pAlignedPtr, nReadSector, &pStartFileLargeBlkPtr->aHash[nOffsetBlks],
                                          nLog2SectorsPerMaxHashBlk, nOffsetBlks, sizeof(WFSHashCode), nStride-sizeof(pStartFileLargeBlkPtr->aHash));
        if(nResult == WFSKRN_RESULT_DEVICE_HASH_INCONSISTENT){
            osTPrintf("User Block Hash Error (Large): %d\n", nAbsBlkAdr);
        }
        if(nResult < WFSKRN_RESULT_OK){
            return nResult;
        }
        pAlignedPtr += nReadSector << pDevInfo->dh.nLog2SectorSize;
        nStartIdx += nRunLength;
        nOffsetBlks = 0;
    }                                     
    return WFSKRN_RESULT_OK;
}


WFSKrnResult TransWriteFileLargeBlks(u8 *pAlignedPtr, AreaInfo *pAreaInfo, u32 nSize, u32 nOffsetBlks, 
                                     WFSFileLargeBlkPtr *pFileLargeBlkPtr, s32 nStride, TransInfo *pTransInfo, bool bDecryption)
{
    // This function reads, decrypts and hash checs a set of file blocks directly to a specified aligned buffer
    //   Arguments:
    //   pAlignedPtr     [Out] .. Address of aligned buffer to overwrite with data from file blocks. (must be aligned to WFS_STATIC_MEM_ALIGNMENT bytes)
    //   pAreaInfo        [In] .. Pointer to the AreaInfo struct which contains the blocks to be written
    //   nNumBlks         [In] .. The number of blocks to write 
    //   nOffsetBlks      [In] .. The number of offset blocks for first write block  
    //   pFileBlkPtr      [In] .. Pointer to the first WFSFileBlkPtr struct, which specifies the address and hash code of the file block
    //   nStride          [In] .. How to get from the first WFSFileBlkPtr to the next one
    //   pTransInfo       [In] .. Pointer to a TransInfo struct for the transaction (may be used to remap blocks).

    DeviceInfo *pDevInfo = pAreaInfo->pVolInfo->pDevInfo;
    ASSERT(nSize);
    ASSERT(!(nSize&(pDevInfo->dh.nLog2SectorSize-1)));
    u32 nLog2SectorsPerBlk = pAreaInfo->ah.nLog2BlkSize - pDevInfo->dh.nLog2SectorSize;
    u32 nLog2SectorsPerMaxHashBlk = pAreaInfo->ah.nLog2MediumBlkSize - pDevInfo->dh.nLog2SectorSize;
    u32 nNumBlks = ((nSize-1)>>pAreaInfo->ah.nLog2MediumBlkSize)+1;
    u32 nEndBlkSector = (nSize>>pDevInfo->dh.nLog2SectorSize)-((nNumBlks-1)<<nLog2SectorsPerMaxHashBlk);
    u32 nStartIdx=0;
    u32 nOffsetBlksTmp = nOffsetBlks;
    while(0 < nNumBlks) {
        // Starting from the current block, look for a run of sequential blocks and count its length
        WFSFileLargeBlkPtr *pStartFileLargeBlkPtr = pFileLargeBlkPtr;
        u32 nRunLength = 0;
        u32 nWriteSector;
        do{
            if(nNumBlks + nOffsetBlksTmp > (1<<pAreaInfo->nLog2MediumBlkPerLargeBlk)){
                nRunLength += (1<<pAreaInfo->nLog2MediumBlkPerLargeBlk) - nOffsetBlksTmp;
                nNumBlks -= (1<<pAreaInfo->nLog2MediumBlkPerLargeBlk) - nOffsetBlksTmp;
                nWriteSector = nRunLength<<nLog2SectorsPerMaxHashBlk;
                nOffsetBlksTmp = 0;
            } else {
                nRunLength += nNumBlks;
                nWriteSector = ((nRunLength-1)<<nLog2SectorsPerMaxHashBlk)+nEndBlkSector;
                nNumBlks = 0;
                break;
            }
            pFileLargeBlkPtr = (WFSFileLargeBlkPtr *)((u8*)pFileLargeBlkPtr + nStride);
        }while( pFileLargeBlkPtr->nBlkAdr == pStartFileLargeBlkPtr->nBlkAdr + ((nRunLength + nOffsetBlks)<<pAreaInfo->nLog2BlksPerMediumBlk) );
        // Write (nRunLength) blocks from the device to the destination buffer, then do decryption & hash check on each block
        WFSBlkAdr nAbsBlkAdr = pAreaInfo->nAbsStartBlkAdr + pStartFileLargeBlkPtr->nBlkAdr + (nOffsetBlks<<pAreaInfo->nLog2BlksPerMediumBlk);
#if _DEBUG_BREAK_POINT
        if(nAbsBlkAdr == 0x2880){
            int a; a=0;
        }
#endif
        // The caches of the written blocks have to be Invalidated because BCache data would be old.
        WFSKrnReturnOnError(BCacheInvalidateBlks(pAreaInfo->pVolInfo, nAbsBlkAdr, nRunLength<<pAreaInfo->nLog2BlksPerMediumBlk));
        WFSKrnReturnOnError(DeviceWrite(pDevInfo, nAbsBlkAdr<<nLog2SectorsPerBlk, pAlignedPtr, nWriteSector, &pStartFileLargeBlkPtr->aHash[nOffsetBlks],
                                        nLog2SectorsPerMaxHashBlk, nOffsetBlks, sizeof(WFSHashCode), nStride-sizeof(pStartFileLargeBlkPtr->aHash), bDecryption));
        pAlignedPtr += nWriteSector << pDevInfo->dh.nLog2SectorSize;
        nStartIdx += nRunLength;
        nOffsetBlks = 0;
    }
    return WFSKRN_RESULT_OK;
}


WFSKrnResult TransTrancateFileBlk(TransBlkAdr *pTba, WFSBlkAdr nParentBlkAdr, u16 nHashOfs, WFSBlkAdr nFileBlkAdr, u32 nOldSize, u32 nNewSize)
{
    nOldSize = WFS_ROUND_UP(nOldSize, pTba->pAreaInfo->pVolInfo->pDevInfo->nSectorSize);
    nNewSize = WFS_ROUND_UP(nNewSize, pTba->pAreaInfo->pVolInfo->pDevInfo->nSectorSize);
    ASSERT(nOldSize<=pTba->pAreaInfo->nBlkSize && nNewSize<=pTba->pAreaInfo->nBlkSize);
    u8 *pBlkPtr;
    if(!nNewSize){
        WFSKrnReturnOnError(TransInvalidate3(pTba->pAreaInfo, nFileBlkAdr, pTba->pTransInfo))
    } else if(nNewSize != nOldSize) {
        WFSKrnReturnOnError(TransGetAndPinFileBlk(pTba, nParentBlkAdr, nHashOfs, nFileBlkAdr, TRANS_FLAG_RW_BLK, nOldSize, (void**)&pBlkPtr));
        WFSKrnReturnOnError(TransSetValidSize4(pTba->pAreaInfo, nFileBlkAdr, pTba->pTransInfo, nNewSize));
        WFSKrnReturnOnError(TransStoreBlk3(pTba->pAreaInfo, nFileBlkAdr, pTba->pTransInfo));
        WFSKrnReturnOnError(TransUnpinBlk3(pTba->pAreaInfo, nFileBlkAdr, pTba->pTransInfo));
    }
    return WFSKRN_RESULT_OK;
}


WFSKrnResult TransTrancateFileMediumBlk(TransBlkAdr *pTba, WFSBlkAdr nParentBlkAdr, u16 nHashOfs, WFSBlkAdr nFileBlkAdr, u32 nOldSize, u32 nNewSize)
{
    nOldSize = WFS_ROUND_UP(nOldSize, pTba->pAreaInfo->pVolInfo->pDevInfo->nSectorSize);
    nNewSize = WFS_ROUND_UP(nNewSize, pTba->pAreaInfo->pVolInfo->pDevInfo->nSectorSize);
    ASSERT(nOldSize<=pTba->pAreaInfo->nMediumBlkSize && nNewSize<=pTba->pAreaInfo->nMediumBlkSize);
    u8 *pBlkPtr;
    if(!nNewSize){
        WFSKrnReturnOnError(TransInvalidate3(pTba->pAreaInfo, nFileBlkAdr, pTba->pTransInfo));
    } else if(nNewSize != nOldSize){
        WFSKrnReturnOnError(TransGetAndPinFileMediumBlk(pTba, nParentBlkAdr, nHashOfs, nFileBlkAdr, TRANS_FLAG_RW_BLK, nOldSize, (void**)&pBlkPtr));
        WFSKrnReturnOnError(TransSetValidSize4(pTba->pAreaInfo, nFileBlkAdr, pTba->pTransInfo, nNewSize));
        WFSKrnReturnOnError(TransStoreBlk3(pTba->pAreaInfo, nFileBlkAdr, pTba->pTransInfo));
        WFSKrnReturnOnError(TransUnpinBlk3(pTba->pAreaInfo, nFileBlkAdr, pTba->pTransInfo));
    }
    return WFSKRN_RESULT_OK;
}

