/*---------------------------------------------------------------------------*
  Project:  wfs
  File:     wfskrn_Volume.cpp - Volume MR initialization and obtaining the MR.

  Copyright 2007 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: wfskrn_Volume.cpp,v $
  Revision 1.6  2008/09/28 23:32:04  kondo_masahiro
  Fixed codes to use medium size user block.

  Revision 1.5  2008/08/05 03:56:08  kondo_masahiro
  Added error handling for hash error and detach device.

  Revision 1.4  2008/07/17 05:33:46  kondo_masahiro
  Fixed codes for porting IOP.

  Revision 1.3  2008/07/09 00:56:32  paul
  Moved volume handling code from wfskrn_Api.cpp to here. Integrates volumes with path cache and block cache.

  Revision 1.2  2008/05/12 19:17:50  paul
  Merged in Wayne's previous changes

  Revision 1.1  2008/04/22 04:39:08  kondo_masahiro
  Fixed code to transfer the wfskrn project to RM_SDK tree

  Revision 1.6  2008/04/26 01:28:27  wayne.wong
  Moved area allocation functionality from Volume to Area module.
  Changed API appropriately.

  Revision 1.5  2008/04/18 01:59:14  wayne.wong
  Created separate Area types for persistent (hdr) and non-persistent (info) types.
  Changed function names to be consistent with kernel naming convention.
  Defines are included to support the previous naming, but will be deleted in
  the future.

  Revision 1.4  2008/03/24 10:17:40  wayne.wong
  Fixed a bug with area free that was not recursively coalescing correctly.

  Revision 1.3  2008/03/19 09:17:36  wayne.wong
  Fixed an allocation bug. Initial test set works correctly now.

  Revision 1.2  2008/03/19 07:43:50  wayne.wong
  Checkpoint. Added initial alloc/free area functions.

  Revision 1.1  2008/03/11 03:25:05  wayne.wong
  Checkpoint. Preliminary version.


*---------------------------------------------------------------------------*/

#include "wfskrn_Volume.h"
#include "wfskrn_BCache.h"
#include "wfskrn_Api.h"
#include "randomlib.h"


void VolumeModuleInit() {
    dbgv(OS_TPrintf("VolumeModuleInit()\n"));
    // Mounted list is doubly linked and circular
    // Free list is singly linked and NULL terminated
    wkg.pMountedVolListAnchor = (VolumeInfo *)((u8*)&wkg.mountedVolListAnchor - (size_t)&((VolumeInfo *)0)->vl);
    wkg.pMountedVolListAnchor->vl.pNext = wkg.pMountedVolListAnchor->vl.pPrev = wkg.pMountedVolListAnchor;
    VolumeInfo *pVolInfo = wkg.pFreeVolList = wkg.aVolInfo;
    VolumeInfo *pVolEnd = &pVolInfo[WFSKRN_MAX_VOLUMES];
    for(; pVolInfo<pVolEnd; ++pVolInfo) {
        pVolInfo->vl.pNext = pVolInfo+1;
    }
    pVolInfo[-1].vl.pNext = NULL;
}


VolumeInfo *VolumeAlloc() {
    dbgv(OS_TPrintf("VolumeAlloc()\n"));
    VolumeInfo *pVolInfo = wkg.pFreeVolList;
    wkg.pFreeVolList = wkg.pFreeVolList->vl.pNext;
    return pVolInfo;
}


void VolumeFree(VolumeInfo *pVolInfo) {
    dbgv(OS_TPrintf("VolumeFree()\n"));
#if _DEBUG
    memset(pVolInfo, 0, sizeof(*pVolInfo));
#endif
    pVolInfo->vl.pNext = wkg.pFreeVolList;
    wkg.pFreeVolList = pVolInfo;
}


void VolumeAddToMountedList(VolumeInfo *pVolInfo) {
    dbgv(OS_TPrintf("VolumeAddToMountedList()\n"));
    pVolInfo->vl.pNext = wkg.pMountedVolListAnchor->vl.pNext;
    pVolInfo->vl.pNext->vl.pPrev = pVolInfo;
    wkg.pMountedVolListAnchor->vl.pNext = pVolInfo;
    pVolInfo->vl.pPrev = wkg.pMountedVolListAnchor;
}


void VolumeRemoveFromMountedList(VolumeInfo *pVolInfo) {
    dbgv(OS_TPrintf("VolumeRemoveFromMountedList()\n"));
    pVolInfo->vl.pPrev->vl.pNext   = pVolInfo->vl.pNext;
    pVolInfo->vl.pNext->vl.pPrev   = pVolInfo->vl.pPrev;
}


WFSKrnResult VolumeInitRootArea(VolumeInfo *pVolInfo, u8 nLog2SmallBlkSize) {
    dbgv(OS_TPrintf("VolumeInitRootArea()\n"));
    pVolInfo->rootAreaInfo.pParent = 0;
    pVolInfo->rootAreaInfo.nRelStartBlkAdr = 0;
    pVolInfo->rootAreaInfo.ah.nLog2BlkSize = nLog2SmallBlkSize;
    pVolInfo->rootAreaInfo.pVolInfo = pVolInfo;
    return AreaInit(&pVolInfo->rootAreaInfo); // for debug wfskrn
}


void VolumeConvertU32ToVolId(WFSVolumeId *pVolId, u32 nVolId) {
    dbgv(OS_TPrintf("VolumeConvertU32ToVolId()\n"));
    utf8 *pStr = pVolId->sStr;
    u32 nI;
    for(nI=0; nI<7; nI++) {
        *pStr++ = "23456789abcdefghjklmnpqrstuvwxyz"[nVolId>>27];    // String has 32 characters
        nVolId<<=5;
    }
    *pStr = 0;
    pVolId->nLen = 7;
}


WFSKrnResult VolumeInit(VolumeInfo *pVolInfo, u32 nLog2BlkSize, u32 nNumBlks) {
    dbgv(OS_TPrintf("VolumeInit()\n"));
    // Arguments:
    // pVolInfo         [OUT] .. A pointer to the VolumeInfo struct to initialize.
    // nVolumeId         [IN] .. The randomly assigned volume identifier.
    // nLog2BlkSize      [IN] .. Size of small blocks (and meta-data blocks) in the root area.
    // nNumBlks          [IN] .. The total number of samll blocks contained in the volume.
    // Synopsis:
    // Initializes a new volume, updating the volume header on disk.
    // Note: BCache module should have been already enabled.
    pVolInfo->vh.mdh.nFlags  = WFS_MDF_VOLUME_HDR;
    pVolInfo->vh.nWfsVersion = (u16)((WFS_VERSION_MAJOR<<8) + WFS_VERSION_MINOR);
    pVolInfo->vh.nBuildDate  = BUILD_DATE;  // ToDo: each module should supply a build date, so that the latest can be chosen - otherwise we only get the date this particular source file was compiled
    utf8 sVolPath[WFS_MAX_VOLUME_ID_SIZE + 6];
    RxtItr rxi;
    rxi.pStrPtr = sVolPath;
    WFSVolumeId volId;
    do {
        pVolInfo->vh.nVolId = RandomU32();  // ToDo: Write a kernel specific random number generator which attempts to collect entropy from I/O, timing or other sources.
        VolumeConvertU32ToVolId(&volId, pVolInfo->vh.nVolId);
        osSNPrintf(sVolPath, WFS_MAX_VOLUME_ID_SIZE + 6, "/vol/%s", volId.sStr);
    } while(PathCacheFind1(&rxi));
    BCacheEntry* pBce;    
    WFSKrnResult nResult = BCacheAllocMetaBlk(pVolInfo, VOLUME_HDR_BLK_ADR, BCACHE_FLAG_PINNED | BCACHE_FLAG_DIRTY, &pBce);
    if(nResult < WFSKRN_RESULT_OK){
        return nResult;
    }
    VolumeHdr *pVh = (VolumeHdr*)pBce->pBlkPtr;
#if _DEBUG
    if (!pVh) {
        OS_TPrintf("Error: VolumeInit() WFSKRN_RESULT_VOLUME_BCACHE_ALLOC\n");
        return WFSKRN_RESULT_VOLUME_BCACHE_ALLOC;
    }
    if (nLog2BlkSize < pVolInfo->pDevInfo->dh.nLog2SectorSize) {
        OS_TPrintf("Error: VolumeInit() WFSKRN_RESULT_VOLUME_INVALID_PARAMETER\n");
        return WFSKRN_RESULT_VOLUME_INVALID_PARAMETER;
    }
#endif
    *pVh = pVolInfo->vh;
    BCacheUnpin(pVolInfo, VOLUME_HDR_BLK_ADR);
#if _DEBUG
    pVolInfo->volId = volId;
#endif
    pVolInfo->rootAreaInfo.pVolInfo        = pVolInfo;
    pVolInfo->rootAreaInfo.pParent         = 0;
    pVolInfo->rootAreaInfo.nRelStartBlkAdr = 0;
    pVolInfo->rootAreaInfo.ah.nLog2BlkSize = nLog2BlkSize;
    pVolInfo->rootAreaInfo.ah.nNumBlks     = nNumBlks;
    return AreaInit(&pVolInfo->rootAreaInfo); // for debug wfskrn
}


WFSKrnResult VolumeUnmount(VolumeInfo *pVolInfo) {
    dbgv(OS_TPrintf("VolumeUnmount()\n"));
    AreaClose(&pVolInfo->rootAreaInfo);
    BCacheFlushVolume(pVolInfo);
    pVolInfo->pDevInfo->pVolInfo = 0;
    VolumeRemoveFromMountedList(pVolInfo);
    VolumeFree(pVolInfo);
    return WFSKRN_RESULT_OK;
}
